/* _______              __ 
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::ifstream;
using std::ios;

#include "FABView.H"

#include "ParmParse.H"
#include "CH_HDF5.H"
#include "parstream.H"

#include "AMR.H"
#include "AMRLevel.H"
#include "AMRAdvectDiffuseFactory.H"
#include "AMRAdvectDiffuse.H"

#include "PatchAdvectDiffuse.H"

#include "TGASolver.H"
#include "BackwardEulerSolver.H"
#include "CrankNicolsonSolver.H"
#include "AdvectTestIBC.H"

#ifdef USE_ARRAYVIEW
#include "ArrayView.H"
extern "C" {
#include <fpu_control.h>
}
/* IM: Invalid operation mask
 * DM: Denormalized operand mask 
 * ZM: Zero-divide mask
 * OM: Overflow mask 
 * UM: Underflow mask 
 * PM: Precision (inexact result) mask 
  ---(pm is kinda stupid)
*/
static void __attribute__ ((constructor)) trapfpe(void)
{
  fpu_control_t cw = 
    _FPU_DEFAULT & ~(_FPU_MASK_IM | _FPU_MASK_ZM | _FPU_MASK_OM | _FPU_MASK_UM);
  _FPU_SETCW(cw);
}
#endif

// Available sample problems
#define PROBLEM_TEST  0
#define PROBLEM_GAUSSIAN  1
#define PROBLEM_SINUSOIDAL 2

// Available solver types
#define SOLVER_TGA  0
#define SOLVER_CN   1
#define SOLVER_BE   2

// amrGodunov is a function (as opposed to inline in main()) to get
// around MPI scoping problems
void amrGodunov();

// setupFixedGrids allows fixed grids to be read in and used in this AMR
// computation example
void setupFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                     const ProblemDomain&  a_domain,
                     int                   a_maxLevel,
                     int                   a_maxGridSize,
                     int                   a_blockFactor,
                     int                   a_verbosity,
                     std::string           a_gridFile);

// One more function for MPI
void dumpmemoryatexit();

int main(int a_argc, char* a_argv[])
{
#ifdef CH_MPI
  // Start MPI
  MPI_Init(&a_argc,&a_argv);
  setChomboMPIErrorHandler();
#endif

  // Check for an input file
  char* inFile = NULL;
#ifdef USE_ARRAYVIEW
  trapfpe();
#endif

  if (a_argc > 1)
  {
    inFile = a_argv[1];
  }
  else
  {
    pout() << "Usage:  amrGodunov...ex <inputfile>" << endl;
    pout() << "No input file specified" << endl;
    return -1;
  }

  // Parse the command line and the input file (if any)
  ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

  // Run amrGodunov, i.e., do the computation
  amrGodunov();

#ifdef CH_MPI
  // Exit MPI 
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}

void amrGodunov()
{
  // Read inputs that are prefixed with "godonouv."
  ParmParse ppgodunov("godunov");

  // Determine the sample problem specified
  int problem = -1;
  
  std::string problemString;

  if (ppgodunov.contains("problem"))
    {
      ppgodunov.query("problem",problemString);

      if (problemString == "advectTest")
        {
          problem = PROBLEM_TEST;
        }
      else if (problemString == "gaussian")
        {
          problem = PROBLEM_GAUSSIAN;
        } 
      else if (problemString == "sinusoidal")
        {
          problem = PROBLEM_SINUSOIDAL;
        }
      
      // The sample problem given isn't valid
      if (problem == -1)
        {
          pout() << "Invalid problem, \"" << problemString << "\", specified in input file" << endl << endl; 
          return;
        }
    }
  else
    {
      // A sample problem must be specified
      pout() << "\"godunov.problem\" not specified in input file" << endl << endl; 
      return;
    }
  
  
  
  // Determine the solver type to use (default is TGA)
  int solverType = SOLVER_TGA;
  std::string solverString;

  if (ppgodunov.contains("solver"))
    {
      ppgodunov.query("solver",solverString);

      if (solverString == "TGA")
        {
          solverType = SOLVER_TGA;
        }
      else if (solverString == "CrankNicolson")
        {
          solverType = SOLVER_CN;
        }
      else if (solverString == "BackwardEuler")
        {
          solverType = SOLVER_BE;
        }
      else 
        {
          pout() << "Invalid solver, \"" << solverString << "\", specified in input file" << endl << endl; 
          return;
        }
    }
  
  // diffusion coefficient
  Real diffusionCoeff = 0.0;
  ppgodunov.query("diffusionCoeff", diffusionCoeff);

  // default is to do implicit refluxing
  bool implicitReflux;
  int inImplicitReflux = 1;
  ppgodunov.query("doImplicitReflux", inImplicitReflux);
  implicitReflux = (inImplicitReflux == 1);

  // compute truncation errors?
  bool computeTruncationError = false;
  int tempBool = computeTruncationError;
  ppgodunov.query("computeTruncationError", tempBool);
  computeTruncationError = (tempBool == 1);

  
  // This determines the amount of diagnositic output generated
  int verbosity = 0;
  ppgodunov.query("verbosity",verbosity);
  CH_assert(verbosity >= 0);

  vector<Real> velocitypp(SpaceDim,0.0);
  RealVect velocity;

  ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
  for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];
  

  // Stop after this number of steps
  int nstop = 0;
  ppgodunov.query("max_step",nstop);

  // Stop when the simulation time get here
  Real stopTime = 0.0;
  ppgodunov.query("max_time",stopTime);

  // Set the physical size of the longest dimension of the domain
  Real domainLength = 1.0;
  ppgodunov.query("domain_length",domainLength);

  // Set the resolution of the coarsest level 
  vector<int> numCells(SpaceDim);
  for (int i = 0; i < SpaceDim; ++i) numCells[i]=0;
  ppgodunov.queryarr("num_cells",numCells,0,SpaceDim);

  CH_assert(D_TERM(   (numCells[0] > 0),
                   && (numCells[1] > 0),
                   && (numCells[2] > 0)));
  CH_assert(D_TERM(   (numCells[0] % 2 == 0),
                   && (numCells[1] % 2 == 0),
                   && (numCells[2] % 2 == 0)));

  // Determine which spatial directions are periodic
  vector<int> isPeriodica(SpaceDim,0);
  bool isPeriodic[SpaceDim];
  
  ppgodunov.queryarr("is_periodic",isPeriodica,0,SpaceDim);
  // convert periodic from int->bool
  for (int dim=0; dim<SpaceDim; dim++) 
    {
      isPeriodic[dim] = (isPeriodica[dim] == 1);
      if (isPeriodic[dim] && verbosity >= 2 && procID() == 0)
        pout() << "Using Periodic BCs in direction: " << dim << endl;
    }

  // Maximum AMR level limit
  int maxLevel = 0;
  ppgodunov.query("max_level",maxLevel);
  int numReadLevels = Max(maxLevel,1);

  // Refinement ratios between levels
  std::vector<int> refRatios;
  // Note: this requires a refRatio to be defined for the finest level
  // (even though it will never be used)
  ppgodunov.queryarr("ref_ratio",refRatios,0,numReadLevels+1);

  // Number of coarse time steps from one regridding to the next
  std::vector<int> regridIntervals;
  ppgodunov.queryarr("regrid_interval",regridIntervals,0,numReadLevels);

  // How far to extend refinement from cells newly tagged for refinement
  int tagBufferSize = 3;
  ppgodunov.query("tag_buffer_size",tagBufferSize);

  // Threshold that triggers refinement
  Real refineThresh = 0.3;
  ppgodunov.query ("refine_thresh",refineThresh);

  // Minimum dimension of a grid
  int blockFactor = 1;
  ppgodunov.query("block_factor",blockFactor);

  // Maximum dimension of a grid
  int maxGridSize = 32;
  ppgodunov.query("max_grid_size",maxGridSize);

  Real fillRatio = 0.75;
  ppgodunov.query("fill_ratio",fillRatio);

  // Do 4th order (1) or 2nd order (0) slope computations
  int inFourthOrderSlopes = 1;
  bool fourthOrderSlopes;
  ppgodunov.query("use_fourth_order_slopes",inFourthOrderSlopes);
  fourthOrderSlopes = (inFourthOrderSlopes == 1);

  // Do slope flattening - only valid with 4th order slopes
  int inFlattening = 0;
  bool flattening;
#if 0
  // at the moment, flattening is not implemented for the 
  // advection-diffusion example. so, in an act of cowardice,
  // disable the option
  ppgodunov.query("use_flattening",inFlattening);
#endif
  flattening = (inFlattening == 1);

  CH_assert(fourthOrderSlopes || !flattening);

  // default is to limit slopes
  int inLimitSlopes = 1;
  bool limitSlopes;
  ppgodunov.query("use_limitSlopes", inLimitSlopes);
  limitSlopes = (inLimitSlopes == 1);

  // Apply artificial viscosity
  int inUseArtificialViscosity = 1;
  bool useArtificialViscosity;
  ppgodunov.query("use_artificial_viscosity",inUseArtificialViscosity);
  useArtificialViscosity = (inUseArtificialViscosity == 1);

  // Artificial viscosity coefficient/multiplier
  Real artificialViscosity = 0.1;
  ppgodunov.query("artificial_viscosity",artificialViscosity);

  // Set up checkpointing
  int checkpointInterval = 0;
  ppgodunov.query("checkpoint_interval",checkpointInterval);

  // Set up plot file writing
  int plotInterval = 0;
  ppgodunov.query("plot_interval",plotInterval);

  // dump errors into plotfiles?
  bool plotErrors = false;
  tempBool = plotErrors;
  ppgodunov.query("plot_error", tempBool);
  plotErrors = (tempBool == 1);

  // CFL multiplier
  Real cfl = 0.8;
  ppgodunov.query("cfl",cfl);

  // Initial CFL multiplier
  Real initialCFL = 0.1;
  ppgodunov.query("initial_cfl",initialCFL);

  // Determine if a fixed or variable time step will be used
  Real fixedDt = -1;
  ppgodunov.query("fixed_dt",fixedDt);

  // Limit the time step growth
  Real maxDtGrowth = 1.1;
  ppgodunov.query("max_dt_growth",maxDtGrowth);

  // Let the time step grow by this factor above the "maximum" before
  // reducing it
  Real dtToleranceFactor = 1.1;
  ppgodunov.query("dt_tolerance_factor",dtToleranceFactor);

  Real solverTol = 1.0e-7;
  ppgodunov.query("solver_tolerance", solverTol);

  // Print the parameters
  if ( verbosity >= 2 )
    {
      pout() << "problem = " << problemString << endl;
      pout() << "solver = " << solverString << endl;
      pout() << "solver tolerance = " << solverTol << endl;

      //XXX -- gamma isn't a variable; it's a function in a gcc header file
      //XXXpout() << "gamma = " << gamma << endl;

      pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                        velocity[1] << "  " <<,
                                        velocity[2] << ) endl;

      pout() << "diffusionCoefficient = " << diffusionCoeff << endl;
      pout() << "do Implicit Refluxing = " << implicitReflux << endl;
      
      pout() << "maximum step = " << nstop << endl;
      pout() << "maximum time = " << stopTime << endl;

      pout() << "number of cells = " << D_TERM(numCells[0] << "  " <<,
                                               numCells[1] << "  " <<,
                                               numCells[2] << ) endl;

      pout() << "maximum level = " << maxLevel << endl;

      pout() << "refinement ratio = ";
      for (int i = 0; i < refRatios.size(); ++i) pout() << refRatios[i] << " ";
      pout() << endl;

      pout() << "regrid interval = ";
      for (int i = 0; i < regridIntervals.size(); ++i) pout() << regridIntervals[i] << " ";
      pout() << endl;

      pout() << "refinement threshold = " << refineThresh << endl;

      pout() << "blocking factor = " << blockFactor << endl;
      pout() << "max grid size = " << maxGridSize << endl;
      pout() << "fill ratio = " << fillRatio << endl;

      pout() << "checkpoint interval = " << checkpointInterval << endl;
      pout() << "plot interval = " << plotInterval << endl;
      pout() << "plot error = " << plotErrors << endl;

      pout() << "computeTruncation error = " 
             << computeTruncationError << endl;
      
      pout() << "CFL = " << cfl << endl;
      pout() << "initial CFL = " << initialCFL << endl;
      if (fixedDt > 0)
        {
          pout() << "fixed dt = " << fixedDt << endl;
        }
      pout() << "maximum dt growth = " << maxDtGrowth << endl;
      pout() << "dt tolerance factor = " << dtToleranceFactor << endl;
    }

  // Create initial and boundary condition (IBC) object and initialize
  OldPhysIBC* ibc;


  switch (problem)
    {
    case PROBLEM_TEST:
      {
        AdvectTestIBC* advectibc = new AdvectTestIBC;
        advectibc->advectionVel(velocity);
        // 1 is gaussian, 2 is sinusoidal
        int probtype = 1;
        advectibc->probType(probtype);
        ibc = advectibc;
        break;
      }

    case PROBLEM_GAUSSIAN:
      {
        AdvectTestIBC* advectibc = new AdvectTestIBC;
        advectibc->advectionVel(velocity);
        advectibc->probType(problem);
        ibc = advectibc;
        break;
      }

    case PROBLEM_SINUSOIDAL:
      {
        AdvectTestIBC* advectibc = new AdvectTestIBC;
        advectibc->advectionVel(velocity);
        advectibc->probType(problem);
        ibc = advectibc;
        break;
      }
    

    default:
      {
        MayDay::Error("amrGodunov \"problem\" (used for initial and boundary conditions) is invalid");
        break;
      }
    }
  
  ProblemDomain probDomain (IntVect::Zero,
                            IntVect(D_DECL(numCells[0]-1,
                                           numCells[1]-1,
                                           numCells[2]-1)),
                            isPeriodic);

  AMRAdvectDiffuseFactory amrGodFact;

  RefCountedPtr<OldPatchGodunov> patchAdvectDiffuse = static_cast<OldPatchGodunov*>
                                                   (new PatchAdvectDiffuse());

  // Set up the patch integrator for polytropic gas dynamics
  patchAdvectDiffuse->setPhysIBC(ibc);
  patchAdvectDiffuse->setSlopeParameters(fourthOrderSlopes,
                                         flattening,
                                         limitSlopes);

  patchAdvectDiffuse->setArtificialViscosity(useArtificialViscosity,
                                             artificialViscosity);


  // Set up the AMRLevel... factory
  amrGodFact.CFL(cfl);
  amrGodFact.domainLength(domainLength);
  amrGodFact.refinementThreshold(refineThresh);
  amrGodFact.tagBufferSize(tagBufferSize);
  amrGodFact.patchGodunov(patchAdvectDiffuse);
  amrGodFact.verbosity(verbosity);
  amrGodFact.initialDtMultiplier(initialCFL);
  amrGodFact.backgroundVelocity(velocity);
  amrGodFact.diffusionCoeff(diffusionCoeff);
  amrGodFact.doImplicitReflux(implicitReflux);
  amrGodFact.setSolverTolerance(solverTol);
  amrGodFact.computeTruncationError(computeTruncationError);
  amrGodFact.plotError(plotErrors);

  // use a TGA solver
  BaseHeatSolver* diffusionSolverPtr;
  if (solverType == SOLVER_TGA) 
    {
      diffusionSolverPtr = (BaseHeatSolver*) new TGASolver;
    }
  else if (solverType == SOLVER_BE)
    { 
      diffusionSolverPtr = (BaseHeatSolver*) new BackwardEulerSolver;
    }
  else if (solverType == SOLVER_CN)
    {
      diffusionSolverPtr = (BaseHeatSolver*) new CrankNicolsonSolver;
    }

  diffusionSolverPtr->setSolverTolerance(solverTol);
  amrGodFact.diffusionSolver(diffusionSolverPtr);
    

  AMR amr;

  // Set up the AMR object
  amr.define(maxLevel,refRatios,probDomain,&amrGodFact);

  if (fixedDt > 0)
    {
      amr.fixedDt(fixedDt);
    }

  // Set grid generation parameters
  amr.maxGridSize(maxGridSize);
  amr.blockFactor(blockFactor);
  amr.fillRatio(fillRatio);

  // The hyperbolic codes use a grid buffer of 1
  amr.gridBufferSize(1);

  // Set output parameters
  amr.checkpointInterval(checkpointInterval);
  amr.plotInterval(plotInterval);
  amr.regridIntervals(regridIntervals);
  amr.maxDtGrow(maxDtGrowth);
  amr.dtToleranceFactor(dtToleranceFactor);

  // Set up output files
  if (ppgodunov.contains("plot_prefix"))
    {
      std::string prefix;
      ppgodunov.query("plot_prefix",prefix);
      amr.plotPrefix(prefix);
    }

  if (ppgodunov.contains("chk_prefix"))
    {
      std::string prefix;
      ppgodunov.query("chk_prefix",prefix);
      amr.checkpointPrefix(prefix);
    }

  amr.verbosity(verbosity);

  // Set up input files
  if (!ppgodunov.contains("restart_file"))
    {
      if (!ppgodunov.contains("fixed_hierarchy"))
        {
          // initialize from scratch for AMR run
          // initialize hierarchy of levels
          amr.setupForNewAMRRun();
        }
      else
        {
          std::string gridFile;
          ppgodunov.query("fixed_hierarchy",gridFile);

          // initialize from a list of grids in "gridFile"
          Vector<Vector<Box> > amrGrids(maxLevel+1);
          setupFixedGrids(amrGrids,
                          probDomain,
                          maxLevel,
                          maxGridSize,
                          blockFactor,
                          verbosity,
                          gridFile);
          amr.setupForFixedHierarchyRun(amrGrids,1);
        }
    }
  else
    {
      std::string restartFile;
      ppgodunov.query("restart_file",restartFile);

#ifdef CH_USE_HDF5
      HDF5Handle handle(restartFile,HDF5Handle::OPEN_RDONLY);
      // read from checkpoint file
      amr.setupForRestart(handle);
      handle.close();
#else
      MayDay::Error("amrGodunov restart only defined with hdf5");
#endif
    }

  // Run the computation
  amr.run(stopTime,nstop);

  // Output the last plot file and statistics
  amr.conclude();
  
  // now clean up..
  if (diffusionSolverPtr != NULL)
    {
      delete diffusionSolverPtr;
    }

  if (ibc != NULL) 
    {
      delete ibc;
    }

}

// setupFixedGrids allows fixed grids to be read in and used in this AMR
// computation example
void setupFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                     const ProblemDomain&  a_domain,
                     int                   a_maxLevel,
                     int                   a_maxGridSize,
                     int                   a_blockFactor,
                     int                   a_verbosity,
                     std::string           a_gridFile)
{
  // Run this task on one processor
  if (procID() == uniqueProc(SerialTask::compute))
  {
    a_amrGrids.push_back(Vector<Box>(1,a_domain.domainBox()));

    // Read in predefined grids
    ifstream is(a_gridFile.c_str(), ios::in);

    if (is.fail())
    {
      MayDay::Error("Cannot open grids file");
    }
  
    // Format of file:
    //   number of levels, then for each level (starting with level 1):
    //   number of grids on level, list of boxes

    int inNumLevels;
    is >> inNumLevels;

    CH_assert (inNumLevels <= a_maxLevel+1);

    if (a_verbosity >= 3)
    {
      pout() << "numLevels = " << inNumLevels << endl;
    }

    while (is.get() != '\n');

    a_amrGrids.resize(inNumLevels);

    // Check to see if coarsest level needs to be broken up
    domainSplit(a_domain,a_amrGrids[0],a_maxGridSize,a_blockFactor);

    if (a_verbosity >= 3)
    {
      pout() << "level 0: ";
      for (int n = 0; n < a_amrGrids[0].size(); n++)
      {
        pout() << a_amrGrids[0][n] << endl;
      }
    }

    // Now loop over levels, starting with level 1
    int ngrid;
    for (int lev = 1; lev < inNumLevels; lev++)
    {
      is >> ngrid;

      if (a_verbosity >= 3)
      {
        pout() << "level " << lev << " numGrids = " << ngrid << endl;
        pout() << "Grids: ";
      }

      while (is.get() != '\n');

      a_amrGrids[lev].resize(ngrid);

      for (int i = 0; i < ngrid; i++)
      {
        Box bx;
        is >> bx;

        while (is.get() != '\n');

        // Quick check on box size
        Box bxRef(bx);

        if (bxRef.longside() > a_maxGridSize)
        {
          pout() << "Grid " << bx << " too large" << endl;
          MayDay::Error();
        }

        if (a_verbosity >= 3) 
        {
          pout() << bx << endl;
        }

        a_amrGrids[lev][i] = bx;
      } // End loop over boxes on this level
    } // End loop over levels
  }

  // Broadcast results to all the processors
  broadcast(a_amrGrids,uniqueProc(SerialTask::compute));
}
