/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _AMR_ADVECTDIFFUSE_FACTORY_H_
#define _AMR_ADVECTDIFFUSE_FACTORY_H_

#include "AMRLevelFactory.H"
#include "AMRAdvectDiffuse.H"

/// AMR Godunov factory
/**
 */
class AMRAdvectDiffuseFactory : public AMRLevelFactory
{
public:
  /// Null constructor
  /**
   */
  AMRAdvectDiffuseFactory();

  /// Virtual constructor
  /**
   */
  virtual AMRLevel* new_amrlevel() const;

  /// Destructor
  /**
   */
  virtual ~AMRAdvectDiffuseFactory();

  /// CFL number
  /**
   */
  virtual void CFL(Real a_cfl);

  /// Physical dimension of the longest side of the domain
  /**
   */
  virtual void domainLength(Real a_domainLength);

  ///
  /**
   */
  virtual void verbosity(const int& verbosity);

  /// Refinement threshold
  /**
   */
  virtual void refinementThreshold(Real a_refineThresh);

  /// Tag buffer size
  /**
   */
  void tagBufferSize(int a_tagBufferSize);

  /// Initial dt multiplier
  /**
   */
  void initialDtMultiplier(Real a_initialDtMultiplier);

  /// OldPatchGodunov object (used as a factory)
  /**
   */
  virtual void patchGodunov(RefCountedPtr<OldPatchGodunov> a_patchGodunov);

  /// background velocity for test problem
  virtual void backgroundVelocity(const RealVect& a_velocity);

  /// diffusion coefficient
  virtual void diffusionCoeff(const Real a_diffusionCoeff);

  /// set diffusion solver type
  virtual void diffusionSolver(const BaseHeatSolver* a_diffusionSolverPtr);

  /// set whether to use implicit refluxing 
  virtual void doImplicitReflux(bool a_do_implicit_reflux);

  /// set tolerance for implicit refluxing solve
  virtual void setSolverTolerance(Real a_solverTol);

  
  /// set whether to compute truncation errors
  virtual void computeTruncationError(bool a_compute_error);

  /// include error in plot files?
  /** This is obviously only useful when an exact solution exists...
   */
  virtual void plotError(bool a_plot_error);

  /// Check that everything is defined
  /**
   */
  bool isDefined() const;

protected:
  // Some default values
  void setDefaultValues();

  int m_verbosity;

  // CFL number
  Real m_cfl;
  bool m_cflSet;

  bool m_do_implicit_reflux;

  bool m_compute_truncation_error;
  bool m_plot_error;

  // background velocity
  RealVect m_backgroundVelocity;
  
  // diffusion coefficient
  Real m_diffusionCoeff;
  
  // diffusion solver to use
  BaseHeatSolver* m_diffusionSolverPtr;

  Real m_solver_tolerance;

  // Physical dimension of the longest side of the domain
  Real m_domainLength;
  bool m_domainLengthSet;

  // Refinement threshold for gradient
  Real m_refineThresh;
  bool m_refineThreshSet;

  // Tag buffer size
  int  m_tagBufferSize;
  bool m_tagBufferSizeSet;

  // Initial dt multiplier 
  Real m_initialDtMultiplier;
  bool m_initialDtMultiplierSet;

  // Patch integrator (used as a factory)
  RefCountedPtr<OldPatchGodunov> m_patchGodunov;
  bool                           m_patchGodunovSet;

private:
  // Disallowed for all the usual reasons
  void operator=(const AMRAdvectDiffuseFactory& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  AMRAdvectDiffuseFactory(const AMRAdvectDiffuseFactory& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
