/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "AMRLevel.H"
#include "AMRAdvectDiffuse.H"
#include "AMRAdvectDiffuseFactory.H"

AMRAdvectDiffuseFactory::AMRAdvectDiffuseFactory()
{
  setDefaultValues();
}

// Virtual constructor
AMRLevel* AMRAdvectDiffuseFactory::new_amrlevel() const
{
  // Make sure everything is defined
  CH_assert(isDefined());

  // Create a new AMRAdvectDiffuse
  AMRAdvectDiffuse* amrGodPtr = new AMRAdvectDiffuse();

  // Set up new object
  amrGodPtr->CFL(m_cfl);
  amrGodPtr->domainLength(m_domainLength);
  amrGodPtr->refinementThreshold(m_refineThresh);
  amrGodPtr->tagBufferSize(m_tagBufferSize);
  amrGodPtr->initialDtMultiplier(m_initialDtMultiplier);
  amrGodPtr->patchGodunov(m_patchGodunov);
  amrGodPtr->verbosity(m_verbosity);
  amrGodPtr->backgroundVelocity(m_backgroundVelocity);
  amrGodPtr->diffusionCoeff(m_diffusionCoeff);
  amrGodPtr->diffusionSolver(m_diffusionSolverPtr);
  amrGodPtr->doImplicitReflux(m_do_implicit_reflux);
  amrGodPtr->setSolverTolerance(m_solver_tolerance);
  amrGodPtr->computeTruncationError(m_compute_truncation_error);
  amrGodPtr->plotError(m_plot_error);

  // Return it
  return (static_cast <AMRLevel*> (amrGodPtr));
}

AMRAdvectDiffuseFactory::~AMRAdvectDiffuseFactory()
{
}

// CFL number
void AMRAdvectDiffuseFactory::CFL(Real a_cfl)
{
  m_cfl = a_cfl;
  m_cflSet = true;
}

void AMRAdvectDiffuseFactory::verbosity(const int& a_verbosity)
{
  m_verbosity = a_verbosity;
}

// Physical dimension of the longest side of the domain
void AMRAdvectDiffuseFactory::domainLength(Real a_domainLength)
{
  m_domainLength = a_domainLength;
  m_domainLengthSet = true;
}

// Refinement threshold
void AMRAdvectDiffuseFactory::refinementThreshold(Real a_refineThresh)
{
  m_refineThresh = a_refineThresh;
  m_refineThreshSet = true;
}

// Tag buffer size
void AMRAdvectDiffuseFactory::tagBufferSize(int a_tagBufferSize)
{
  m_tagBufferSize = a_tagBufferSize;
  m_tagBufferSizeSet = true;
}

// Initial dt multiplier
void AMRAdvectDiffuseFactory::initialDtMultiplier(Real a_initialDtMultiplier)
{
  m_initialDtMultiplier = a_initialDtMultiplier;
  m_initialDtMultiplierSet = true;
}

// OldPatchGodunov object (used as a factory)
void AMRAdvectDiffuseFactory::patchGodunov(RefCountedPtr<OldPatchGodunov> a_patchGodunov)
{
  m_patchGodunov = a_patchGodunov;
  m_patchGodunovSet = true;
}

// set background velocity
void AMRAdvectDiffuseFactory::backgroundVelocity(const RealVect& a_velocity)
{
  m_backgroundVelocity = a_velocity;
}

// set diffusion coefficient
void AMRAdvectDiffuseFactory::diffusionCoeff(const Real a_diffusionCoeff)
{
  m_diffusionCoeff = a_diffusionCoeff;
}

// set diffusion solver
void AMRAdvectDiffuseFactory::diffusionSolver(const BaseHeatSolver* a_solverPtr)
{
  m_diffusionSolverPtr = a_solverPtr->new_heatSolver();
}

// set whether to do implicit refluxing
void AMRAdvectDiffuseFactory::doImplicitReflux(bool a_do_implicit_reflux)
{
  m_do_implicit_reflux = a_do_implicit_reflux;
}

void 
AMRAdvectDiffuseFactory::setSolverTolerance(Real a_solverTol)
{
  m_solver_tolerance = a_solverTol;
}


//
void 
AMRAdvectDiffuseFactory::computeTruncationError(bool a_compute_error)
{
  m_compute_truncation_error = a_compute_error;
}

/// include error in plot files?
void
AMRAdvectDiffuseFactory::plotError(bool a_plot_error)
{
  m_plot_error = a_plot_error;
}

// Check that everything is defined
bool AMRAdvectDiffuseFactory::isDefined() const
{
  return (m_cflSet &&
          m_domainLengthSet &&
          m_refineThreshSet &&
          m_tagBufferSizeSet &&
          m_initialDtMultiplierSet &&
          m_patchGodunovSet);
}

// Some default values
void AMRAdvectDiffuseFactory::setDefaultValues()
{
  CFL(0.8);
  domainLength(1.0);
  refinementThreshold(0.2);
  tagBufferSize(2);
  initialDtMultiplier(0.1);
  m_patchGodunovSet = false;
  m_verbosity = 0;
  m_backgroundVelocity = RealVect(D_DECL(0,0,0));
  m_diffusionCoeff = 0.0;
  m_solver_tolerance = 1.0e-7;
  m_do_implicit_reflux = true;
  m_compute_truncation_error = false;
  m_plot_error = false;
}

