/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "PatchAdvectDiffuse.H"
#include "PatchAdvectDiffuseF_F.H"

#include "LoHiSide.H"

#include "LGintegrator.H"

PatchAdvectDiffuse::PatchAdvectDiffuse():OldPatchGodunov()
{
}

// Factory method - this object is its own factory.  It returns a pointer
// to new OldPatchGodunov object with its initial and boundary condtions, slope
// parameters, and artificial viscosity information defined.
OldPatchGodunov* PatchAdvectDiffuse::new_patchGodunov() const
{
  // Make the new object
  OldPatchGodunov* retval = 
    static_cast<OldPatchGodunov*>(new PatchAdvectDiffuse());

  // Pass the current initial and boundary condition (IBC) object to this new
  // patch integrator so it can create its own IBC object and define it as it
  // needs to (i.e. new domain and grid spacing if necessary)
  retval->setPhysIBC(getPhysIBC());

  // Set the slope and artificial viscosity parameters
  retval->setSlopeParameters(m_useFourthOrderSlopes,
                             m_useFlattening,
                             m_limitSlopes);
  retval->setArtificialViscosity(m_useArtificialViscosity,
                                 m_artificialViscosity);

  // Return the new object
  return retval;
}

// Compute the maximum wave speed
Real PatchAdvectDiffuse::getMaxWaveSpeed(const FArrayBox& a_U,
                                         const Box&       a_box)
{
  CH_assert(isDefined());
  CH_assert(a_U.contains(a_box));

  Real speed = 0.0;
  // in this code, a_U contains the velocities
  int xVelComp = UVELX;
  speed = a_U.norm(a_box, 0, xVelComp, SpaceDim);

  
  return speed;
}

// Number of conserved variables
int PatchAdvectDiffuse::numConserved()
{
  return UNUM;
}

// Names of the conserved variables
Vector<string> PatchAdvectDiffuse::stateNames()
{
  Vector<string> retval;

  retval.push_back("scalar");
  retval.push_back("x-velocity");
  if (SpaceDim >= 2)
    {
      retval.push_back("y-velocity");

      if (SpaceDim >= 3)
        {
          retval.push_back("z-velocity");
        }
    }
  return retval; 
}

// Number of flux variables
int PatchAdvectDiffuse::numFluxes()
{
  // In some computations there may be more fluxes than conserved variables
  return UFLUX;
}

// Number of primitive variables
int PatchAdvectDiffuse::numPrimitives()
{
  // This doesn't equal the number of conserved variables because
  // auxiliary/redundant variable may be computed and stored
  return WNUM;
}

//  Number of primitive variables for which slopes are computed
int PatchAdvectDiffuse::numSlopes()
{
  // This may be less than the number of primitive variables for the
  // reason given in numPrimitives() comments
  return WSLOPE;
}

// Compute the primitive variables from the conserved variables
void PatchAdvectDiffuse::consToPrim(FArrayBox&       a_W,
                                    const FArrayBox& a_U,
                                    const Box&       a_box)
{
  CH_assert(isDefined());
  CH_assert(a_U.box().contains(a_box));
  CH_assert(a_W.box().contains(a_box));
  
  // since the primitive and conserved variables are the same in 
  // this set of equations, even I can do this one...
  a_W.copy(a_U, a_box);
  

}

// Compute the conserved variables from the primitive variables
void PatchAdvectDiffuse::primToCons(FArrayBox&       a_U,
                                    const FArrayBox& a_W,
                                    const Box&       a_box)
{
  CH_assert(isDefined());
  CH_assert(a_U.box().contains(a_box));
  CH_assert(a_W.box().contains(a_box));

  // since primitive variables are conserved in this problem,
  // this is easy
  a_U.copy(a_W, a_box);


}

// Extrapolate the primitive variables to the cell faces
void PatchAdvectDiffuse::normalPred(FArrayBox&       a_WMinus,
                                    FArrayBox&       a_WPlus,
                                    const FArrayBox& a_W,
                                    const FArrayBox& a_dW,
                                    const Real&      a_scale,
                                    const int&       a_dir,
                                    const Box&       a_box)
{
  CH_assert(isDefined());

  FORT_NORMALPREDF(CHF_FRA(a_WMinus),
                  CHF_FRA(a_WPlus),
                  CHF_CONST_FRA(a_W),
                  CHF_CONST_FRA(a_dW),
                  CHF_CONST_REAL(a_scale),
                  CHF_CONST_INT(a_dir),
                  CHF_BOX(a_box));
}

// increment with source term
void PatchAdvectDiffuse::incrementWithSource(FArrayBox& a_W,
                                             const FArrayBox& a_S,
                                             const Real& a_scale,
                                             const Box& a_box)
{
  CH_assert (isDefined());

  CH_assert (a_W.box().contains(a_box));
  CH_assert (a_S.box().contains(a_box));
  // may eventually want to have version which only acts on a subset
  // of components, but for now, this is required (can always set non-active
  // components to 0):
  CH_assert (a_S.nComp() == a_W.nComp());

  // now we just add in the source term, right? (do i need to do 
  // cons->prim, increment w/ source, prim->cons? -- in this case, since
  // cons == prim, it's not important, but as an example, it would be
  // useful to actually have it correct in the code.
  // 

  // for the moment, skip the cons->prim->cons conversion
  // would be nice if FArrayBox had a scaleMult function which
  // worked on a sub-box (scale-mult functions only work on
  // entire domain) -- this is a problem because of the const-ness of 
  // a_S 
  // this is downright silly
  FArrayBox tempSrc(a_S.box(), a_S.nComp());
  tempSrc.copy(a_S);
  tempSrc *= a_scale;
  int startComp = 0;
  int numComp = a_S.nComp();
  a_W.plus(tempSrc, a_box, startComp, startComp, numComp);

  
}


// Compute a Riemann problem and generate fluxes at the faces
void PatchAdvectDiffuse::riemann(FArrayBox&       a_F,
                                 const FArrayBox& a_WLeft,
                                 const FArrayBox& a_WRight,
                                 const int&       a_dir,
                                 const Box&       a_box)
{
  CH_assert(isDefined());

  CH_assert(a_F.box().contains(a_box));

  // Get the numbers of relevant variables
  int numPrim = numPrimitives();
  int numFlux = numFluxes();

  CH_assert(a_F.nComp() == numFlux);
  CH_assert(a_WLeft.nComp() == numPrim);
  CH_assert(a_WRight.nComp() == numPrim);

  // Cast away "const" inputs so their boxes can be shifted left or right
  // 1/2 cell and then back again (no net change is made!)
  FArrayBox& shiftWLeft  = (FArrayBox&)a_WLeft;
  FArrayBox& shiftWRight = (FArrayBox&)a_WRight;

  // Solution to the Riemann problem
  FArrayBox Wgdnv(a_box,numPrim);

  // Shift the left and right primitive variable boxes 1/2 cell so they are
  // face centered
  shiftWLeft .shiftHalf(a_dir, 1);
  shiftWRight.shiftHalf(a_dir,-1);
  
  CH_assert(shiftWLeft.box().contains(a_box));
  CH_assert(shiftWRight.box().contains(a_box));

  // Riemann solver computes Wgdnv all edges that are not on the physical
  // boundary.
  FORT_RIEMANNF(CHF_FRA(Wgdnv),
                CHF_CONST_FRA(shiftWLeft),
                CHF_CONST_FRA(shiftWRight),
                CHF_CONST_INT(a_dir),
                CHF_BOX(a_box));
 
  // Shift the left and right primitive variable boxes back to their original
  // position
  shiftWLeft .shiftHalf(a_dir,-1);
  shiftWRight.shiftHalf(a_dir, 1);
  
  // Compute fluxes from the solution to the Riemann problem
  FORT_GETFLUXF(CHF_FRA(a_F),
               CHF_CONST_FRA(Wgdnv),
               CHF_CONST_INT(a_dir),
               CHF_BOX(a_box));
}

// Update the primitive variables using fluxes (of the conserved variables)
void PatchAdvectDiffuse::updatePrim(FArrayBox&       a_WMinus,
                                    FArrayBox&       a_WPlus,
                                    const FArrayBox& a_F,
                                    const Real&      a_scale,
                                    const int&       a_dir,
                                    const Box&       a_box)
{
  CH_assert(isDefined());
  CH_assert(a_WPlus.box().contains(a_box));
  CH_assert(a_WMinus.box().contains(a_box));

  // Number of conserved variables
  int numCons = numConserved();

  // Temporary storage for converting primitive variables to conserved
  // variables and back again (after update with the fluxes)
  FArrayBox U(a_box,numCons);

  // for W+-, convert primitive quantities to conservative quantities,
  // update with the fluxes, and convert back
  primToCons(U,a_WMinus,a_box);
  updateCons(U,a_F,a_scale,a_dir,a_box);
  consToPrim(a_WMinus,U,a_box);

  primToCons(U,a_WPlus,a_box);
  updateCons(U,a_F,a_scale,a_dir,a_box);
  consToPrim(a_WPlus,U,a_box);
}

// Update the conserved variable using fluxes and a scaling factor
void PatchAdvectDiffuse::updateCons(FArrayBox&       a_U,
                                    const FArrayBox& a_F,
                                    const Real&      a_scale,
                                    const int&       a_dir,
                                    const Box&       a_box)
{
  CH_assert(isDefined());

  FORT_UPDATECONSF(CHF_FRA(a_U),
                  CHF_CONST_FRA(a_F),
                  CHF_CONST_REAL(a_scale),
                  CHF_CONST_INT(a_dir),
                  CHF_BOX(a_box));
}

// Update the conserved variable using fluxes and a scaling factor
void PatchAdvectDiffuse::finalUpdate(FArrayBox&       a_U,
                                     const FArrayBox& a_F,
                                     const Real&      a_scale,
                                     const int&       a_dir,
                                     const Box&       a_box)
{
  // in this case, we just call updateCons..
  updateCons(a_U, a_F, a_scale, a_dir, a_box);
}



// Interval within the primitive variables corresponding to the velocities
// since this doesn't apply to this set of equations, return a bogus interval
Interval PatchAdvectDiffuse::velocityInterval()
{
  Interval retval(-1,-1);

  return retval;
}

// Component index within the primitive variables of the pressure
// since this doesn't apply to this set of equations, return a bogus value
int PatchAdvectDiffuse::pressureIndex()
{
  return -1;
}

// Component index within the primitive variables of the bulk modulus
// since this doesn't apply to this set of equations, return a bogus value
int PatchAdvectDiffuse::bulkModulusIndex()
{
  return -1;
}


