/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// NodeCFIVS.H
// petermc, 1 May 2001
#ifndef NODECFIVS_H
#define NODECFIVS_H

#include <iostream>
#include <cmath>
#include "SPACE.H"
#include <cstdlib>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "LoHiSide.H"
#include "DisjointBoxLayout.H"
#include "IntVectSet.H"

/// Class to get fine IntVectSet at coarse/fine interface
/** 
    Class to get IntVectSet of nodes (of a particular face of a
    particular box) that lie on interface with next coarser level.

    This class should be considered internal to AMRNodeSolver and should
    not be considered part of the Chombo API.
*/
class NodeCFIVS
{

public:

  /**
     \name Constructors, destructor and defines
  */
  /*@{*/

  /// 
  /** Default constructor.  User must subsequently call define().
   */
  NodeCFIVS();

  /// 
  /** Destructor.
   */
  ~NodeCFIVS();

  /// 
  /** Constructor calls setDefaultValues() and then
      calls define() with the same arguments.
  */
  NodeCFIVS(const Box& a_domain,
            const Box& a_box,
            const DisjointBoxLayout& a_levelBoxes, 
            int a_idir,
            Side::LoHiSide a_hiorlo);

  /// 
  /** Constructor calls setDefaultValues() and then
      calls define() with the same arguments.
  */
  NodeCFIVS(const ProblemDomain& a_domain,
            const Box& a_box,
            const DisjointBoxLayout& a_levelBoxes, 
            int a_idir,
            Side::LoHiSide a_hiorlo);

  ///
  /** Full define function.  The current level is taken to be the fine level.
  */
  void define(/// CELL-centered physical domain at this level
              const ProblemDomain& a_domain,
              /// the CELL-centered box at this level
              const Box& a_box,
              /// the set of all CELL-centered boxes at this level
              const DisjointBoxLayout& a_levelBoxes, 
              /// direction (0 to SpaceDim-1) of face of box on which we find interface nodes
              int a_idir,
              /// high or low side of box
              Side::LoHiSide a_hiorlo);

  ///
  /** Full define function.  The current level is taken to be the fine level.
  */
  void define(/// CELL-centered physical domain at this level
              const Box& a_domain,
              /// the CELL-centered box at this level
              const Box& a_box,
              /// the set of all CELL-centered boxes at this level
              const DisjointBoxLayout& a_levelBoxes, 
              /// direction (0 to SpaceDim-1) of face of box on which we find interface nodes
              int a_idir,
              /// high or low side of box
              Side::LoHiSide a_hiorlo);

  /*@}*/

  /**
     \name Access functions
  */
  /*@{*/
	      
  ///
  bool isDefined() const;

  /// 
  /** Returns <tt>true</tt> if this coarse-fine IntVectSet is empty.
  */
  bool isEmpty() const;

  /// 
  /** Returns <tt>true</tt> if this coarse-fine IntVectSet can be represented
      as just a Box, i.e., a user can perform a dense computation instead of
      a pointwise calculation using IVSIterator.

      If isPacked() then you can use packedBox() instead of getFineIVS().
  */
  bool isPacked() const;

  ///
  /** If isPacked() returns <tt>true</tt>,
      then the box returned by this function
      is a suitable substitute for this IntVectSet returned by getFineIVS().
  */
  const Box& packedBox() const;

  ///
  /** Returns indices of fine nodes, on the face of the box,
      that lie on the interface with the coarser level and where
      data need to be interpolated.

      This will be empty if isEmpty() returns <tt>true</tt>.
      If isPacked() then you can use packedBox() instead of getFineIVS().
  */
  const IntVectSet& getFineIVS() const ;

  /*@}*/

  static long long m_packCount, m_sparseCount;

protected:

  //for internal use
  void setDefaultValues();

  /** indices of fine nodes that need to be interpolated
   */
  IntVectSet m_fineInterpIVS;

  bool m_packed;

  bool m_empty;

  Box m_packedBox;

  bool m_isdefined;

private:
  void operator= (const NodeCFIVS& levcfs_in) {};

  NodeCFIVS(const NodeCFIVS& levcfs_in) {};
	
};


inline
bool NodeCFIVS::isEmpty() const
{
  return m_empty;
}


inline
bool NodeCFIVS::isPacked() const
{
  return m_packed;
}


inline
const Box& NodeCFIVS::packedBox() const
{
  return m_packedBox;
}

#endif
