/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// NodeDotProduct.H
// petermc, 11 June 2003

#ifndef NODEDOTPRODUCT_H
#define NODEDOTPRODUCT_H

#include "NodeFArrayBox.H"
#include "LevelData.H"
#include "IntVectSet.H"

/// Computes dot product of two sets of data on the same set of grids.
/** 
    The dot product computed by this function is the sum over
    all components and all points of the product
    of <i>a_dataOne</i> and <i>a_dataTwo</i>,
    weighted by trapezoidal integration rule weights.
 */
Real DotProductNodes(/// first set of data
                     const BoxLayoutData<NodeFArrayBox>& a_dataOne,
                     /// second set of data
                     const BoxLayoutData<NodeFArrayBox>& a_dataTwo,
                     /// CELL-centered layout, on the nodes of which we calculate the dot product
                     const BoxLayout& a_dblIn);


/// Computes dot product of two sets of data on the same set of grids.
/** 
    The dot product computed by this function is the sum over
    the components in <i>a_comps</i> and over all points of the product
    of <i>a_dataOne</i> and <i>a_dataTwo</i>,
    weighted by trapezoidal integration rule weights.
 */
Real DotProductNodes(/// first set of data
                     const BoxLayoutData<NodeFArrayBox>& a_dataOne,
                     /// second set of data
                     const BoxLayoutData<NodeFArrayBox>& a_dataTwo,
                     /// CELL-centered layout, on the nodes of which we calculate the dot product
                     const BoxLayout& a_dblIn,
                     /// components over which to take sum
                     const Interval& a_comps);


/// Computes dot product of two sets of data on the same set of grids.
/** 
    The dot product computed by this function is the sum over
    the components in a_comps and over all VALID nodes of the product
    of <i>a_dataOne</i> and <i>a_dataTwo</i>.

    The <i>a_IVSVext</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_dataOne.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_dataOne.getBoxes();
    </tt>
 */
Real DotProductNodes(/// first set of data
                     const LevelData<NodeFArrayBox>& a_dataOne,
                     /// second set of data
                     const LevelData<NodeFArrayBox>& a_dataTwo,
                     /// physical domain
                     const ProblemDomain& a_domain,
                     /// external boundary nodes
                     const LayoutData< Vector<IntVectSet> >& a_IVSVext,
                     /// components over which to take sum
                     const Interval& a_comps);


/// Computes dot product of two sets of data on the same set of grids.
/** 
    The dot product computed by this function is the sum over
    the components in a_comps and over all VALID nodes of the product
    of <i>a_dataOne</i> and <i>a_dataTwo</i>.

    The <i>a_IVSVext</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_dataOne.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_dataOne.getBoxes();
    </tt>
 */
Real DotProductNodes(/// first set of data
                     const LevelData<NodeFArrayBox>& a_dataOne,
                     /// second set of data
                     const LevelData<NodeFArrayBox>& a_dataTwo,
                     /// physical domain
                     const Box& a_domain,
                     /// external boundary nodes
                     const LayoutData< Vector<IntVectSet> >& a_IVSVext,
                     /// components over which to take sum
                     const Interval& a_comps);


/// Computes dot product of two sets of data on the same set of grids.
/** 
    The dot product computed by this function is the sum over
    all components and all points of the product
    of <i>a_dataOne</i> and <i>a_dataTwo</i>,
    weighted by trapezoidal integration rule weights.

    This function computes exterior boundary nodes.  It is more efficient
    to precompute these boundary-node objects and call the function that
    takes <i>a_IVSVext</i> as an argument.
 */
Real DotProductNodes(/// first set of data
                     const LevelData<NodeFArrayBox>& a_dataOne,
                     /// second set of data
                     const LevelData<NodeFArrayBox>& a_dataTwo,
                     /// physical domain
                     const ProblemDomain& a_domain,
                     /// components over which to take sum
                     const Interval& a_comps);

/// Computes dot product of two sets of data on the same set of grids.
/** 
    The dot product computed by this function is the sum over
    all components and all points of the product
    of <i>a_dataOne</i> and <i>a_dataTwo</i>,
    weighted by trapezoidal integration rule weights.

    This function computes exterior boundary nodes.  It is more efficient
    to precompute these boundary-node objects and call the function that
    takes <i>a_IVSVext</i> as an argument.
 */
Real DotProductNodes(/// first set of data
                     const LevelData<NodeFArrayBox>& a_dataOne,
                     /// second set of data
                     const LevelData<NodeFArrayBox>& a_dataTwo,
                     /// physical domain
                     const Box& a_domain,
                     /// components over which to take sum
                     const Interval& a_comps);

#endif
