/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// Norms.H
// petermc, 12 Aug 2002
// petermc, 17 Oct 2005, put integral functions into NodeIntegrals.H

#ifndef NORMS_H
#define NORMS_H

#include "LevelData.H"
#include "NodeFArrayBox.H"
#include "Vector.H"
#include "IntVectSet.H"

/// Computes L^p norm of a subbox of a NodeFArrayBox.
/** Computes L^p norm of the data in a subbox of <i>a_nfab</i>.
 */
Real norm(/// the data on which to take the norm
          const NodeFArrayBox& a_nfab,
          /// mesh spacing at this level
          const Real a_dx,
          /// CELL-centered subbox over which to take norm
          const Box& a_subbox,
          /// norm to take, or 0 for max norm
          const int a_p,
          /// starting component
          const int a_startComp = 0,
          /// number of components
          const int a_numComp = 1);

/// Computes max norm of a subbox of a NodeFArrayBox.
/** Computes max norm of the data in a subbox of <i>a_nfab</i>.
 */
Real maxnorm(/// the data on which to take the norm
             const NodeFArrayBox& a_nfab,
             /// CELL-centered subbox over which to take norm
             const Box& a_subbox,
             /// starting component
             const int a_startComp = 0,
             /// number of components
             const int a_numComp = 1);

/// Computes L^p norm of a BoxLayoutData<NodeFArrayBox>.
/** Computes L^p norm of <i>a_layout</i>. 
    Ignores ghost cells for LevelData.

    This is an internal function that should NOT normally
    be called by the user.
 */
Real norm(/// the data on which to take the norm
          const BoxLayoutData<NodeFArrayBox>& a_layout, 
          /// mesh spacing at this level
          const Real a_dx,
          /// norm to take, or 0 for max norm
          const int a_p,
          /// interval of components to use
          const Interval& a_interval,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);

/// Computes max norm of a BoxLayoutData<NodeFArrayBox>.
/** Computes max norm of <i>a_layout</i>.
    Ignores ghost cells for LevelData.

    This is an internal function that should NOT normally
    be called by the user.
 */
Real maxnorm(/// the data on which to take the norm
             const BoxLayoutData<NodeFArrayBox>& a_layout, 
             /// components to use in computing norm
             const Interval& a_interval,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);

/// Computes norm of multilevel array, counting only valid nodes at each level.
/** 
    Computes norm of multilevel array <i>a_phi</i>.
    Only the valid nodes at each level are counted.
    The valid nodes at a level are those that are not covered
    by the interior of any finer level.
 */
Real norm(/// multilevel data on which to take the norm
          const Vector<LevelData<NodeFArrayBox>* >& a_phi, 
          /// CELL-centered physical domain on each level
          const Vector<ProblemDomain>& a_domain,
          /// refinement ratios between successive levels
          const Vector<int>& a_nRefFine,
          /// mesh spacing at base level <i>a_lBase</i>
          const Real a_dxCrse,
          /// components to use in computing norm
          const Interval a_comps, 
          /// norm to take, or 0 for max norm
          const int a_p,
          /// index of base level to use in Vectors
          const int a_lBase,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);

/// Computes norm of multilevel array, counting only valid nodes at each level.
/** 
    Computes norm of multilevel array <i>a_phi</i>.
    Only the valid nodes at each level are counted.
    The valid nodes at a level are those that are not covered
    by the interior of any finer level.
 */
Real norm(/// multilevel data on which to take the norm
          const Vector<LevelData<NodeFArrayBox>* >& a_phi, 
          /// CELL-centered physical domain on each level
          const Vector<Box>& a_domain,
          /// refinement ratios between successive levels
          const Vector<int>& a_nRefFine,
          /// mesh spacing at base level <i>a_lBase</i>
          const Real a_dxCrse,
          /// components to use in computing norm
          const Interval a_comps, 
          /// norm to take, or 0 for max norm
          const int a_p,
          /// index of base level to use in Vectors
          const int a_lBase,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);

/// Computes norm of single-level array, counting only valid nodes.
/** 
    Computes norm of single-level array <i>a_phi</i>.
    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level that are not
    projections of interior nodes of the finer level, if any.

    This function computes exterior boundary nodes, and interior boundary
    nodes of the coarsened finer level, if any.  It is more efficient
    to precompute these boundary-node objects and call one of the other
    norm functions below.
 */
Real norm(/// the data on which to take the norm
          const LevelData<NodeFArrayBox>& a_phi, 
          /// CELL-centered physical domain on each level
          const ProblemDomain& a_domain,
          /// pointer to CELL-centered grids on the next finer level, or NULL if there is none.
          const DisjointBoxLayout* a_finerGridsPtr,
          /// refinement ratio to next finer level
          const int a_nRefFine,
          /// mesh spacing at this level
          const Real a_dx, 
          /// interval of components to use
          const Interval a_comps,
          /// norm to take, or 0 for max norm
          const int a_p,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);

/// Computes norm of single-level array, counting only valid nodes.
/** 
    Computes norm of single-level array <i>a_phi</i>.
    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level that are not
    projections of interior nodes of the finer level, if any.

    This function computes exterior boundary nodes, and interior boundary
    nodes of the coarsened finer level, if any.  It is more efficient
    to precompute these boundary-node objects and call one of the other
    norm functions below.
 */
Real norm(/// the data on which to take the norm
          const LevelData<NodeFArrayBox>& a_phi, 
          /// CELL-centered physical domain on each level
          const Box& a_domain,
          /// pointer to CELL-centered grids on the next finer level, or NULL if there is none.
          const DisjointBoxLayout* a_finerGridsPtr,
          /// refinement ratio to next finer level
          const int a_nRefFine,
          /// mesh spacing at this level
          const Real a_dx, 
          /// interval of components to use
          const Interval a_comps,
          /// norm to take, or 0 for max norm
          const int a_p,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);

/// Computes norm of single-level array, counting only valid nodes.
/** 
    Computes norm of single-level array <i>a_phi</i>.
    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level that are not
    projections of interior nodes of the finer level, if any.
    
    The <i>a_IVSVext</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_phi.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_phi.getBoxes();
    </tt>

    The <i>a_IVSVintFinerCoarsened</i> argument is obtained from the call:<br>
    <tt>
    interiorBoundaryNodes(a_IVSVintFinerCoarsened, a_phi.getBoxes(), a_finerGridsCoarsened, a_domain);
    </tt>
*/
Real norm(/// the data on which to take the norm
          const LevelData<NodeFArrayBox>& a_phi, 
          /// CELL-centered physical domain on each level
          const ProblemDomain& a_domain,
          /// CELL-centered grids on the next finer level
          const DisjointBoxLayout& a_finerGridsCoarsened,
          /// exterior boundary nodes of this level
          const LayoutData< Vector<IntVectSet> >& a_IVSVext,
          /// interior boundary nodes of coarsened finer-level grids
          const LayoutData< Vector<IntVectSet> >& a_IVSVintFinerCoarsened,
          /// refinement ratio to next finer level
          const int a_nRefFine,
          /// mesh spacing at this level
          const Real a_dx, 
          /// interval of components to use
          const Interval a_comps,
          /// norm to take, or 0 for max norm
          const int a_p,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);


/// Computes norm of single-level array, counting only valid nodes.
/** 
    Computes norm of single-level array <i>a_phi</i>.
    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level that are not
    projections of interior nodes of the finer level, if any.
    
    The <i>a_IVSVext</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_phi.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_phi.getBoxes();
    </tt>

    The <i>a_IVSVintFinerCoarsened</i> argument is obtained from the call:<br>
    <tt>
    interiorBoundaryNodes(a_IVSVintFinerCoarsened, a_phi.getBoxes(), a_finerGridsCoarsened, a_domain);
    </tt>
*/
Real norm(/// the data on which to take the norm
          const LevelData<NodeFArrayBox>& a_phi, 
          /// CELL-centered physical domain on each level
          const Box& a_domain,
          /// CELL-centered grids on the next finer level
          const DisjointBoxLayout& a_finerGridsCoarsened,
          /// exterior boundary nodes of this level
          const LayoutData< Vector<IntVectSet> >& a_IVSVext,
          /// interior boundary nodes of coarsened finer-level grids
          const LayoutData< Vector<IntVectSet> >& a_IVSVintFinerCoarsened,
          /// refinement ratio to next finer level
          const int a_nRefFine,
          /// mesh spacing at this level
          const Real a_dx, 
          /// interval of components to use
          const Interval a_comps,
          /// norm to take, or 0 for max norm
          const int a_p,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);


/// Computes norm of single-level array, counting only valid nodes, with no finer level.
/** 
    Computes norm of single-level array <i>a_phi</i>, where there is no
    finer level.  Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level.
    
    The <i>a_IVSVext</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_phi.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_phi.getBoxes();
    </tt>
*/
Real norm(/// the data on which to take the norm
          const LevelData<NodeFArrayBox>& a_phi, 
          /// CELL-centered physical domain on each level
          const ProblemDomain& a_domain,
          /// exterior boundary nodes of this level
          const LayoutData< Vector<IntVectSet> >& a_IVSVext,
          /// mesh spacing at this level
          const Real a_dx, 
          /// interval of components to use
          const Interval a_comps,
          /// norm to take, or 0 for max norm
          const int a_p,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);


/// Computes norm of single-level array, counting only valid nodes, with no finer level.
/** 
    Computes norm of single-level array <i>a_phi</i>, where there is no
    finer level.  Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level.
    
    The <i>a_IVSVext</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_phi.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_phi.getBoxes();
    </tt>
*/
Real norm(/// the data on which to take the norm
          const LevelData<NodeFArrayBox>& a_phi, 
          /// CELL-centered physical domain on each level
          const Box& a_domain,
          /// exterior boundary nodes of this level
          const LayoutData< Vector<IntVectSet> >& a_IVSVext,
          /// mesh spacing at this level
          const Real a_dx, 
          /// interval of components to use
          const Interval a_comps,
          /// norm to take, or 0 for max norm
          const int a_p,
          /// verbosity, whether to print norms of each box
          bool a_verbose = true);


/// Computes max norm of multilevel array, counting only valid nodes at each level.
/** 
    Computes max norm of multilevel array <i>a_phi</i>.
    Only the valid nodes at each level are counted.
    The valid nodes at a level are those that are not covered
    by the interior of any finer level.
 */
Real maxnorm(/// multilevel data on which to take the norm
             const Vector<LevelData<NodeFArrayBox>* >& a_phi, 
             /// CELL-centered physical domain on each level
             const Vector<ProblemDomain>& a_domain,
             /// refinement ratios between successive levels
             const Vector<int>& a_nRefFine,
             /// components to use in computing norm
             const Interval a_comps, 
             /// index of base level to use in Vectors
             const int a_lBase,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);

/// Computes max norm of multilevel array, counting only valid nodes at each level.
/** 
    Computes max norm of multilevel array <i>a_phi</i>.
    Only the valid nodes at each level are counted.
    The valid nodes at a level are those that are not covered
    by the interior of any finer level.
 */
Real maxnorm(/// multilevel data on which to take the norm
             const Vector<LevelData<NodeFArrayBox>* >& a_phi, 
             /// CELL-centered physical domain on each level
             const Vector<Box>& a_domain,
             /// refinement ratios between successive levels
             const Vector<int>& a_nRefFine,
             /// components to use in computing norm
             const Interval a_comps, 
             /// index of base level to use in Vectors
             const int a_lBase,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);

/// Computes max norm of single-level array, counting only valid nodes.
/** 
    Computes max norm of single-level array <i>a_phi</i>.
    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level that are not
    projections of interior nodes of the finer level, if any.

    This function computes exterior boundary nodes, and interior boundary
    nodes of the coarsened finer level, if any.  It is more efficient
    to precompute these boundary-node objects and call one of the other
    maxnorm functions below.
 */
Real maxnorm(/// the data on which to take the norm
             const LevelData<NodeFArrayBox>& a_phi, 
             /// CELL-centered physical domain on each level
             const ProblemDomain& a_domain,
             /// pointer to CELL-centered grids on the next finer level, or NULL if there is none.
             const DisjointBoxLayout* a_finerGridsPtr,
             /// refinement ratio to next finer level
             const int a_nRefFine,
             /// interval of components to use
             const Interval a_comps,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);

/// Computes max norm of single-level array, counting only valid nodes.
/** 
    Computes max norm of single-level array <i>a_phi</i>.
    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level that are not
    projections of interior nodes of the finer level, if any.

    This function computes exterior boundary nodes, and interior boundary
    nodes of the coarsened finer level, if any.  It is more efficient
    to precompute these boundary-node objects and call one of the other
    maxnorm functions below.
 */
Real maxnorm(/// the data on which to take the norm
             const LevelData<NodeFArrayBox>& a_phi, 
             /// CELL-centered physical domain on each level
             const Box& a_domain,
             /// pointer to CELL-centered grids on the next finer level, or NULL if there is none.
             const DisjointBoxLayout* a_finerGridsPtr,
             /// refinement ratio to next finer level
             const int a_nRefFine,
             /// interval of components to use
             const Interval a_comps,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);


/// Computes max norm of single-level array, counting only valid nodes.
/** 
    Computes max norm of single-level array <i>a_phi</i>.
    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level that are not
    projections of interior nodes of the finer level, if any.
    
    The <i>a_IVSVext</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_phi.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_phi.getBoxes();
    </tt>

    The <i>a_IVSVintFinerCoarsened</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(a_IVSVintFinerCoarsened, 
    a_phi.getBoxes(), a_finerGridsCoarsened, a_domain);
    </tt>
*/
Real maxnorm(/// the data on which to take the norm
             const LevelData<NodeFArrayBox>& a_phi, 
             /// CELL-centered physical domain on each level
             const ProblemDomain& a_domain,
             /// CELL-centered grids on the next finer level
             const DisjointBoxLayout& a_finerGridsCoarsened,
             /// exterior boundary nodes of this level
             const LayoutData< Vector<IntVectSet> >& a_IVSVext,
             /// interior boundary nodes of coarsened finer-level grids
             const LayoutData< Vector<IntVectSet> >& a_IVSVintFinerCoarsened,
             /// refinement ratio to next finer level
             const int a_nRefFine,
             /// interval of components to use
             const Interval a_comps,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);


/// Computes max norm of single-level array, counting only valid nodes.
/** 
    Computes max norm of single-level array <i>a_phi</i>.
    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level that are not
    projections of interior nodes of the finer level, if any.
    
    The <i>a_IVSVext</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_phi.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_phi.getBoxes();
    </tt>

    The <i>a_IVSVintFinerCoarsened</i> argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(a_IVSVintFinerCoarsened, 
    a_phi.getBoxes(), a_finerGridsCoarsened, a_domain);
    </tt>
*/
Real maxnorm(/// the data on which to take the norm
             const LevelData<NodeFArrayBox>& a_phi, 
             /// CELL-centered physical domain on each level
             const Box& a_domain,
             /// CELL-centered grids on the next finer level
             const DisjointBoxLayout& a_finerGridsCoarsened,
             /// exterior boundary nodes of this level
             const LayoutData< Vector<IntVectSet> >& a_IVSVext,
             /// interior boundary nodes of coarsened finer-level grids
             const LayoutData< Vector<IntVectSet> >& a_IVSVintFinerCoarsened,
             /// refinement ratio to next finer level
             const int a_nRefFine,
             /// interval of components to use
             const Interval a_comps,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);


/// Computes max norm of single-level array, counting only valid nodes, with no finer level.
/** 
    Computes max norm of single-level array <i>a_phi</i>, where there is no
    finer level.

    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level.
    
    The a_IVSVext argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_phi.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_phi.getBoxes();
    </tt>
*/
Real maxnorm(/// the data on which to take the norm
             const LevelData<NodeFArrayBox>& a_phi, 
             /// CELL-centered physical domain on each level
             const ProblemDomain& a_domain,
             /// exterior boundary nodes of this level
             const LayoutData< Vector<IntVectSet> >& a_IVSVext,
             /// interval of components to use
             const Interval a_comps,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);

/// Computes max norm of single-level array, counting only valid nodes, with no finer level.
/** 
    Computes max norm of single-level array <i>a_phi</i>, where there is no
    finer level.

    Only the valid nodes at this level are counted.
    The valid nodes are the interior nodes of this level.
    
    The a_IVSVext argument is obtained from the calls:<br>
    <tt>
    interiorBoundaryNodes(IVSVint, a_phi.getBoxes(), a_domain);<br>
    exteriorBoundaryNodes(a_IVSVext, IVSVint, a_phi.getBoxes();
    </tt>
*/
Real maxnorm(/// the data on which to take the norm
             const LevelData<NodeFArrayBox>& a_phi, 
             /// CELL-centered physical domain on each level
             const Box& a_domain,
             /// exterior boundary nodes of this level
             const LayoutData< Vector<IntVectSet> >& a_IVSVext,
             /// interval of components to use
             const Interval a_comps,
             /// verbosity, whether to print norms of each box
             bool a_verbose = true);

#endif
