/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// Norms.cpp
// petermc, 12 Aug 2002
// petermc, 9 Sep 2002, revised again:
// For L_p norms where p is finite, use trapezoidal rule.
// petermc, 28 Mar 2003, revised again:
// Use only valid nodes.
// For L_p norms where p is finite, weight by dx^SpaceDim.
// petermc, 17 Oct 2005, put the integral functions into NodeIntegrals.cpp

#include <cmath>
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include "Norms.H"
#include "NodeSetOperations.H"
#include "NodeDotProdF_F.H"

// ---------------------------------------------------------
// added 11 June 2003, removed from NodeFArrayBox
Real
norm(const NodeFArrayBox& a_nfab,
     const Real a_dx,
     const Box& a_subbox, // CELL-centered
     const int a_p,
     const int a_startComp,
     const int a_numComp)
{
  if (a_p == 0)
    return maxnorm(a_nfab, a_subbox, a_startComp, a_numComp);

  const FArrayBox& fab = a_nfab.getFab();
  Box bx(a_nfab.box());
  int nFabComp = fab.nComp();
  CH_assert(a_startComp + a_numComp <= nFabComp);
  CH_assert(bx.contains(a_subbox));

  Box nodeBox(surroundingNodes(a_subbox));

  // Set dv = dx^SpaceDim
  Real dv = 1.;
  for (int idir = 0; idir < SpaceDim; idir++) dv *= a_dx;

  // Weights for trapezoidal rule of integration.
  FArrayBox weight(nodeBox, 1);
  FORT_TRAPWEIGHTS(CHF_FRA1(weight, 0),
                   CHF_BOX(nodeBox),
                   CHF_CONST_REAL(dv));

  Real normTotal;
  FORT_NODENORM(CHF_REAL(normTotal),
                CHF_CONST_INT(a_p),
                CHF_CONST_FRA(fab),
                CHF_CONST_FRA1(weight, 0),
                CHF_BOX(nodeBox),
                CHF_CONST_INT(a_startComp),
                CHF_CONST_INT(a_startComp + a_numComp-1));

  return normTotal;
}


// ---------------------------------------------------------
// added 11 June 2003, removed from NodeFArrayBox
Real
maxnorm(const NodeFArrayBox& a_nfab,
        const Box& a_subbox, // CELL-centered
        const int a_startComp,
        const int a_numComp)
{
  const FArrayBox& fab = a_nfab.getFab();
  Box bx(a_nfab.box());

  int nFabComp = fab.nComp();
  CH_assert(a_startComp + a_numComp <= nFabComp);
  CH_assert(bx.contains(a_subbox));
  Box nodeBox(surroundingNodes(a_subbox));

  Real normTotal;
  FORT_NODEMAXNORM(CHF_REAL(normTotal),
                   CHF_CONST_FRA(fab),
                   CHF_BOX(nodeBox),
                   CHF_CONST_INT(a_startComp),
                   CHF_CONST_INT(a_startComp + a_numComp-1));

  return normTotal;
}


// ---------------------------------------------------------
// 27 March 2003:
// This is called by other functions, and should not be called directly.
Real
norm(const BoxLayoutData<NodeFArrayBox>& a_layout, 
     const Real a_dx,
     const int a_p,
     const Interval& a_interval,
     bool a_verbose)
{
  if (a_p == 0)
    return maxnorm(a_layout, a_interval, a_verbose);

  Real normTotal = 0.;
  
  for (DataIterator it = a_layout.dataIterator(); it.ok(); ++it)
    {
      const NodeFArrayBox& thisNfab = a_layout[it()];
      const Box& thisBox(a_layout.box(it())); // CELL-centered
      Real thisNfabNorm = norm(thisNfab, a_dx, thisBox, a_p,
                               a_interval.begin(), a_interval.size());
      if (a_verbose)
        cout << a_p << "norm(" << thisBox << ") = " << thisNfabNorm << endl;

      if (a_p == 1)
        {
          normTotal += thisNfabNorm;
        }
      else if (a_p == 2)
        {
          normTotal += thisNfabNorm * thisNfabNorm;
        }
      else
        {
          normTotal += pow(thisNfabNorm, Real(a_p));
        }
    }
# ifdef CH_MPI
  Real recv;
  // add up (a_p is not 0)
  int result = MPI_Allreduce(&normTotal, &recv, 1, MPI_CH_REAL,
                             MPI_SUM, Chombo_MPI::comm);
  if (result != MPI_SUCCESS)
    { //bark!!!
      MayDay::Error("sorry, but I had a communication error on norm");
    }
  normTotal = recv;
# endif
  // now do sqrt, etc
  if (a_p == 2)
    normTotal = sqrt(normTotal);
  else
    if ((a_p != 0) && (a_p != 1))
      normTotal = pow(normTotal, 1.0/Real(a_p));

  return normTotal;
}


// ---------------------------------------------------------
// 27 March 2003:
// This is called by other functions, and should not be called directly.
Real
maxnorm(const BoxLayoutData<NodeFArrayBox>& a_layout, 
        const Interval& a_interval,
        bool a_verbose)
{
  Real normTotal = 0.;
  // a_p == 0:  max norm
  for (DataIterator it = a_layout.dataIterator(); it.ok(); ++it)
    {
      const Box& thisBox(a_layout.box(it())); // CELL-centered
      const NodeFArrayBox& thisNfab = a_layout[it()];
      Real thisNfabNorm =
        maxnorm(thisNfab, thisBox, a_interval.begin(), a_interval.size());
      if (a_verbose)
        cout << "maxnorm(" << thisBox << ") = " << thisNfabNorm << endl;
      normTotal = Max(normTotal, thisNfabNorm);
    }
# ifdef CH_MPI
  Real recv;
  int result = MPI_Allreduce(&normTotal, &recv, 1, MPI_CH_REAL,
                             MPI_MAX, Chombo_MPI::comm);
  if (result != MPI_SUCCESS)
    { //bark!!!
      MayDay::Error("sorry, but I had a communication error on maxnorm");
    }
  normTotal = recv;
# endif
  return normTotal;
}


// --------------------------------------------------------------
Real
norm(const Vector<LevelData<NodeFArrayBox>* >& a_phi, 
     const Vector<Box>& a_domain,
     const Vector<int>& a_nRefFine,
     const Real a_dxCrse,
     const Interval a_comps, 
     const int a_p,
     const int a_lBase,
     bool a_verbose)
{
  int numlevels = a_domain.size();
  Vector<ProblemDomain> probdomain(numlevels);
  for (int ilev = 0; ilev < numlevels; ilev++)
    probdomain[ilev] = ProblemDomain(a_domain[ilev]);

  return norm(a_phi, probdomain, a_nRefFine,
              a_dxCrse, a_comps, a_p, a_lBase, a_verbose);
}


// --------------------------------------------------------------
// version of 27 March 2003
Real
norm(const Vector<LevelData<NodeFArrayBox>* >& a_phi, 
     const Vector<ProblemDomain>& a_domain,
     const Vector<int>& a_nRefFine,
     const Real a_dxCrse,
     const Interval a_comps, 
     const int a_p,
     const int a_lBase,
     bool a_verbose)
{
  // deleted 27 March 2003
  // if (a_p == 0)
  // return maxnorm(a_phi, a_domain, a_nRefFine, a_comps, a_lBase, a_verbose);

  Real dxLevel = a_dxCrse;
  int numLevels = a_phi.size();
  Real normTotal = 0.0;

  // loop over levels  
  for (int lev = a_lBase; lev < numLevels; lev++)
    {
      // in case there are extra levels which are not defined
      if (a_phi[lev] != NULL)
        {
          CH_assert(a_phi[lev]->isDefined());
      
          LevelData<NodeFArrayBox>& thisPhi = *(a_phi[lev]);
          const DisjointBoxLayout* finerGridsPtr = NULL;
      
          if (lev < numLevels-1)
            finerGridsPtr = &(a_phi[lev+1]->getBoxes());

          const ProblemDomain levDomain = a_domain[lev];
          Real normLevel = norm(thisPhi, levDomain, finerGridsPtr,
                                a_nRefFine[lev], dxLevel, a_comps, a_p, a_verbose);

          if (a_p == 0)
            {
              // max norm:  reset if new max found at this level
              if (normLevel > normTotal) 
                normTotal = normLevel;
            }
          else if (a_p == 1)
            normTotal += normLevel;
          else if (a_p == 2)
            normTotal += normLevel * normLevel;
          else
            normTotal += pow(normLevel, Real(a_p));
        }
      
      // update  dxLevel
      dxLevel = dxLevel / Real(a_nRefFine[lev]);
    }
  
  // shouldn't need to do broadcast/gather thing 

  // now do sqrt, etc
  if (a_p == 2)
    normTotal = sqrt(normTotal);
  else
    if ((a_p != 0) && (a_p != 1))
      normTotal = pow(normTotal, 1.0/Real(a_p));

  return normTotal;
}


// ------------------------------------------------------------
// version of 28 March 2003
Real 
norm(const LevelData<NodeFArrayBox>& a_phi, 
     const Box& a_domain,
     const DisjointBoxLayout* a_finerGridsPtr,
     const int a_nRefFine,
     const Real a_dx, 
     const Interval a_comps,
     const int a_p,
     bool a_verbose)
{
  ProblemDomain probdomain(a_domain);
  return norm(a_phi, probdomain, a_finerGridsPtr,
              a_nRefFine, a_dx, a_comps, a_p, a_verbose);
}


// ------------------------------------------------------------
// version of 27 March 2003
Real 
norm(const LevelData<NodeFArrayBox>& a_phi, 
     const ProblemDomain& a_domain,
     const DisjointBoxLayout* a_finerGridsPtr,
     const int a_nRefFine,
     const Real a_dx, 
     const Interval a_comps,
     const int a_p,
     bool a_verbose)
{
  const DisjointBoxLayout& grids = a_phi.getBoxes();

  LayoutData< Vector<IntVectSet> > IVSVint, IVSVext;
  interiorBoundaryNodes(IVSVint, grids, a_domain);
  exteriorBoundaryNodes(IVSVext, IVSVint, grids);
  
  Real normLevel = 0.;
  if (a_finerGridsPtr == NULL)
    {
      normLevel = norm(a_phi, a_domain,
                       IVSVext,
                       a_dx, a_comps, a_p, a_verbose);
    }
  else
    {
      DisjointBoxLayout coarsenedFinerGrids;
      coarsen(coarsenedFinerGrids, *a_finerGridsPtr, a_nRefFine);
      LayoutData< Vector<IntVectSet> > IVSVintFinerCoarsened;
      interiorBoundaryNodes(IVSVintFinerCoarsened, grids,
                            coarsenedFinerGrids, a_domain);

      normLevel = norm(a_phi, a_domain, coarsenedFinerGrids,
                       IVSVext, IVSVintFinerCoarsened,
                       a_nRefFine, a_dx, a_comps, a_p, a_verbose);
    }

  return normLevel;
}


// ------------------------------------------------------------
// version of 27 March 2003
Real 
norm(const LevelData<NodeFArrayBox>& a_phi, 
     const Box& a_domain,
     const DisjointBoxLayout& a_finerGridsCoarsened,
     const LayoutData< Vector<IntVectSet> >& a_IVSVext,
     const LayoutData< Vector<IntVectSet> >& a_IVSVintFinerCoarsened,
     const int a_nRefFine,
     const Real a_dx, 
     const Interval a_comps,
     const int a_p,
     bool a_verbose)
{
  ProblemDomain probdomain(a_domain);
  return norm(a_phi, probdomain, a_finerGridsCoarsened,
              a_IVSVext, a_IVSVintFinerCoarsened,
              a_nRefFine, a_dx, a_comps, a_p, a_verbose);
}


// ------------------------------------------------------------
// version of 27 March 2003
Real 
norm(const LevelData<NodeFArrayBox>& a_phi, 
     const ProblemDomain& a_domain,
     const DisjointBoxLayout& a_finerGridsCoarsened,
     const LayoutData< Vector<IntVectSet> >& a_IVSVext,
     const LayoutData< Vector<IntVectSet> >& a_IVSVintFinerCoarsened,
     const int a_nRefFine,
     const Real a_dx, 
     const Interval a_comps,
     const int a_p,
     bool a_verbose)
{
  // Idea:  copy a_phi to temp, then zero out temp on:
  // - exterior nodes of grids at this level;
  // - projections of interior nodes of the finer grids.
  int ncomps = a_comps.size();
  const DisjointBoxLayout& grids = a_phi.getBoxes();
  LevelData<NodeFArrayBox> temp(grids, ncomps);
  
  // Copy a_phi to temp.
  Interval newcomps(0, ncomps-1);
  for (DataIterator dit(grids.dataIterator()); dit.ok(); ++dit)
    {
      const NodeFArrayBox& nfab = a_phi[dit()];
      const Box& bx = grids.get(dit());
      temp[dit()].copy(bx, newcomps, bx, nfab, a_comps);
    }

  // Zero out temp on exterior nodes.
  zeroBoundaryNodes(temp, a_IVSVext);

  // Define zeroCoarsened to be all zero on the coarsened finer grids.
  LevelData<NodeFArrayBox>
    zeroCoarsened(a_finerGridsCoarsened, ncomps, IntVect::Zero);
  for (DataIterator dit(a_finerGridsCoarsened.dataIterator()); dit.ok(); ++dit)
    zeroCoarsened[dit()].getFab().setVal(0.);
  
  // Set temp to zero on interior nodes of coarsened finer grids.
  copyInteriorNodes(temp, zeroCoarsened, a_IVSVintFinerCoarsened);

  Real normLevel = norm(temp, a_dx, a_p, newcomps, a_verbose);
  
  return normLevel;
}


// ------------------------------------------------------------
// version of 27 March 2003:  no finer level
Real 
norm(const LevelData<NodeFArrayBox>& a_phi, 
     const Box& a_domain,
     const LayoutData< Vector<IntVectSet> >& a_IVSVext,
     const Real a_dx, 
     const Interval a_comps,
     const int a_p,
     bool a_verbose)
{
  ProblemDomain probdomain(a_domain);
  return norm(a_phi, probdomain,
              a_IVSVext,
              a_dx, a_comps, a_p, a_verbose);
}


// ------------------------------------------------------------
// version of 27 March 2003:  no finer level
Real 
norm(const LevelData<NodeFArrayBox>& a_phi, 
     const ProblemDomain& a_domain,
     const LayoutData< Vector<IntVectSet> >& a_IVSVext,
     const Real a_dx, 
     const Interval a_comps,
     const int a_p,
     bool a_verbose)
{
  // Idea:  copy a_phi to temp, then zero out temp on
  // exterior nodes of grids at this level.
  int ncomps = a_comps.size();
  const DisjointBoxLayout& grids = a_phi.getBoxes();
  LevelData<NodeFArrayBox> temp(grids, ncomps);
  
  // Copy a_phi to temp.
  Interval newcomps(0, ncomps-1);
  for (DataIterator dit(grids.dataIterator()); dit.ok(); ++dit)
    {
      const NodeFArrayBox& nfab = a_phi[dit()];
      const Box& bx = grids.get(dit());
      temp[dit()].copy(bx, newcomps, bx, nfab, a_comps);
    }

  // Zero out temp on exterior nodes.
  zeroBoundaryNodes(temp, a_IVSVext);

  Real normLevel = norm(temp, a_dx, a_p, newcomps, a_verbose);
  
  return normLevel;
}


// --------------------------------------------------------------
Real
maxnorm(const Vector<LevelData<NodeFArrayBox>* >& a_phi, 
        const Vector<Box>& a_domain,
        const Vector<int>& a_nRefFine,
        const Interval a_comps, 
        const int a_lBase,
        bool a_verbose)
{
  int numlevels = a_domain.size();
  Vector<ProblemDomain> probdomain(numlevels);
  for (int ilev = 0; ilev < numlevels; ilev++)
    probdomain[ilev] = ProblemDomain(a_domain[ilev]);

  Real normAll = maxnorm(a_phi, probdomain, a_nRefFine,
                         a_comps, a_lBase, a_verbose);

  return normAll;
}


// --------------------------------------------------------------
Real
maxnorm(const Vector<LevelData<NodeFArrayBox>* >& a_phi, 
        const Vector<ProblemDomain>& a_domain,
        const Vector<int>& a_nRefFine,
        const Interval a_comps, 
        const int a_lBase,
        bool a_verbose)
{
  Real dx = 1.;  // arbitrary dx isn't used anyway
  Real normAll = norm(a_phi, a_domain, a_nRefFine, 
                      dx, a_comps, 0, a_lBase, a_verbose);
  return normAll;
}


// ------------------------------------------------------------
// version of 28 March 2003
Real 
maxnorm(const LevelData<NodeFArrayBox>& a_phi, 
        const Box& a_domain,
        const DisjointBoxLayout* a_finerGridsPtr,
        const int a_nRefFine,
        const Interval a_comps,
        bool a_verbose)
{
  ProblemDomain probdomain(a_domain);
  return maxnorm(a_phi, probdomain, a_finerGridsPtr,
                 a_nRefFine, a_comps, a_verbose);
}


// ------------------------------------------------------------
// version of 27 March 2003
Real 
maxnorm(const LevelData<NodeFArrayBox>& a_phi, 
        const ProblemDomain& a_domain,
        const DisjointBoxLayout* a_finerGridsPtr,
        const int a_nRefFine,
        const Interval a_comps,
        bool a_verbose)
{
  Real dx = 1.;  // arbitrary dx isn't used anyway
  Real normLevel = norm(a_phi, a_domain, a_finerGridsPtr,
                        a_nRefFine, dx, a_comps, 0, a_verbose);
  return normLevel;
}


// ------------------------------------------------------------
Real 
maxnorm(const LevelData<NodeFArrayBox>& a_phi, 
        const ProblemDomain& a_domain,
        const DisjointBoxLayout& a_finerGridsCoarsened,
        const LayoutData< Vector<IntVectSet> >& a_IVSVext,
        const LayoutData< Vector<IntVectSet> >& a_IVSVintFinerCoarsened,
        const int a_nRefFine,
        const Interval a_comps,
        bool a_verbose)
{
  Real dx = 1.;  // arbitrary dx isn't used anyway
  Real normLevel = norm(a_phi, a_domain, a_finerGridsCoarsened,
                        a_IVSVext, a_IVSVintFinerCoarsened,
                        a_nRefFine, dx, a_comps, 0, a_verbose);
  return normLevel;
}


// ------------------------------------------------------------
Real 
maxnorm(const LevelData<NodeFArrayBox>& a_phi, 
        const Box& a_domain,
        const DisjointBoxLayout& a_finerGridsCoarsened,
        const LayoutData< Vector<IntVectSet> >& a_IVSVext,
        const LayoutData< Vector<IntVectSet> >& a_IVSVintFinerCoarsened,
        const int a_nRefFine,
        const Interval a_comps,
        bool a_verbose)
{
  ProblemDomain probdomain(a_domain);
  return maxnorm(a_phi, probdomain, a_finerGridsCoarsened,
                 a_IVSVext, a_IVSVintFinerCoarsened,
                 a_nRefFine, a_comps, a_verbose);
}


// ------------------------------------------------------------
// no finer level
Real 
maxnorm(const LevelData<NodeFArrayBox>& a_phi, 
        const ProblemDomain& a_domain,
        const LayoutData< Vector<IntVectSet> >& a_IVSVext,
        const Interval a_comps,
        bool a_verbose)
{
  Real dx = 1.;  // arbitrary dx isn't used anyway
  Real normLevel = norm(a_phi, a_domain,
                        a_IVSVext,
                        dx, a_comps, 0, a_verbose);
  return normLevel;
}


// ------------------------------------------------------------
// no finer level
Real 
maxnorm(const LevelData<NodeFArrayBox>& a_phi, 
        const Box& a_domain,
        const LayoutData< Vector<IntVectSet> >& a_IVSVext,
        const Interval a_comps,
        bool a_verbose)
{
  ProblemDomain probdomain(a_domain);
  return maxnorm(a_phi, probdomain,
                 a_IVSVext,
                 a_comps, a_verbose);
}
