/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "LevelOpacity.H"
#include "EllipsoidIF.H"
#include "GeometryShop.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "ParmParse.H"
#include "ImplicitFunctionBL.H"
#include "SphereIF.H"


void setMagneticFeild(LevelData<FArrayBox>& B,  LevelData<FArrayBox>& rho,
                      const ProblemDomain& domain,
                      const RealVect& spacing);

void makeGrids(DisjointBoxLayout& a_grids,
               ProblemDomain&     a_domain,
               RealVect&          a_dx);


int main(int argc, char* argv[])
{
#ifdef CH_MPI
  // Start MPI
  MPI_Init(&argc,&argv);
  setChomboMPIErrorHandler();
#endif

  {
    char* in_file = argv[1];
    ParmParse  pp(argc-2,argv+2,NULL,in_file);

    DisjointBoxLayout grids;
    ProblemDomain     domain;
    RealVect          spacing;
    makeGrids(grids, domain, spacing);
    LevelOpacity      opacityOp;
    opacityOp.define(grids, domain);


    LevelData<FArrayBox> tau(grids, 2, IntVect::Unit);
    LevelData<FArrayBox> B(grids, CH_SPACEDIM);
    LevelData<FArrayBox> rho(grids, 1, IntVect::Unit);
    
    setMagneticFeild(B, rho, domain, spacing);

    opacityOp.computeOpacity(tau, rho, B, spacing);
  }
  



#ifdef CH_MPI
  MPI_Finalize();
#endif
}
void setMagneticFeild(LevelData<FArrayBox>& B, LevelData<FArrayBox>& rho,
                      const ProblemDomain& domain,
                      const RealVect& spacing)
{
  for(DataIterator dit(B.dataIterator()); dit.ok(); ++dit)
    {
      B[dit].setVal(-0.2 , 0);
      B[dit].setVal(1.0 , 1);
#if CH_SPACEDIM == 3
      B[dit].setVal(0.0 , 2);
#endif
      rho[dit].setVal(1.0E-1, 0);
    }
}

void buildCoveredIntVectSet(IntVectSet& ivs, const GeometryService& levelSet,
                            const Box &region, const Box &domain, 
                            const RealVect &origin, const Real &dx)
{
  if(region.isEmpty()) return;
  if(levelSet.isCovered(region, domain, origin, dx))
    {
      return;
    }
  if(levelSet.isRegular(region, domain, origin, dx)) 
    {
      ivs|=region;
      return;
    }
  int longdir;
  int length = region.longside(longdir);
  if(length == 1)
    {
      ivs|= region;
      return;
    }
  int n = length/2;
  Box low(region), high;
  high = low.chop(longdir, region.smallEnd(longdir)+n);
  buildCoveredIntVectSet(ivs, levelSet, low,  domain, origin, dx);
  buildCoveredIntVectSet(ivs, levelSet, high, domain, origin, dx);
}


void makeGrids(DisjointBoxLayout& a_grids,
               ProblemDomain&     a_domain,
               RealVect&          a_dx)
{

  ParmParse pp;

  a_dx = RealVect(D_DECL(1,1,1));

  EllipsoidIF ellipsoid(RealVect(D_DECL(20,10,10)), RealVect(D_DECL(5,0,0)), true);
  
  //SphereIF sphere(10, RealVect(D_DECL(0,0,0)), true);

  Box d = Box(IntVect(D_DECL(-64,-64,-64)), IntVect(D_DECL(63, 63, 63)));
  a_domain.define(d);
  RealVect origin(D_DECL(0,0,0));

  ImplicitFunctionBL imp(ellipsoid);
  GeometryShop service(imp);
  IntVectSet ivs, pnd(d);
  buildCoveredIntVectSet(ivs, service, d, d, origin, a_dx[0]);

  Vector<int> refRatios(1); refRatios[0]=2;
  BRMeshRefine mesher(a_domain, refRatios, 0.85, 2, 2, 16);
  
  Vector<Box> boxes;

  mesher.makeBoxes (boxes, ivs, pnd, a_domain, 8);
  
  for(int i=0; i<boxes.size(); ++i)
    {
      boxes[i].refine(4);
    }
  a_domain.refine(4);

  Vector<int> procs;
  LoadBalance(procs, boxes);
  a_grids.define(boxes, procs);
}
