


#include "LevelOpacity.H"
#include "AMRIO.H"
#include "SPMD.H"
#include "BoxIterator.H"
#include "SPACE.H"



LevelOpacity::LevelOpacity(){;}

LevelOpacity::~LevelOpacity(){;}

void LevelOpacity::define(const DisjointBoxLayout& a_grids,
                          const ProblemDomain&     a_domain)
{
  // currently this breaks when the fine grids are adjacent to the
  // domain boundary.  This would seem bad, but not really.
  m_domain = a_domain;
  m_cfivs.define(a_grids);
  for(DataIterator dit(a_grids); dit.ok(); ++dit)
    {
      CFStruct& cfivsptr = m_cfivs[dit];
      Box b = a_grids.get(dit);
      for(int i=0; i<CH_SPACEDIM; ++i)
        {
          Box b = a_grids.get(dit);
          b.grow(1);
          b.grow(i, -1);
          cfivsptr.d[i]            .define(a_domain, b ,a_grids, i, Side::Lo);
          cfivsptr.d[i+CH_SPACEDIM].define(a_domain, b, a_grids, i, Side::Hi);
        }
    }
}


void LevelOpacity::computeOpacity(LevelData<FArrayBox>&       a_opacity,
                                  const LevelData<FArrayBox>& a_density,
                                  const LevelData<FArrayBox>& a_magneticField,
                                  const RealVect              a_dx) const 
{

  DisjointBoxLayout dbl = a_density.disjointBoxLayout();
  for(DataIterator dit(dbl); dit.ok(); ++dit)
    {
      FArrayBox& tau = a_opacity[dit];
      tau.setVal(-1.0);
      const CFStruct&  cfivs = m_cfivs[dit];
      for(int i = 0; i<2; ++i)
        {
          
          // first, set the value of opacity to zero on the CF interface.
          for(int dir=0; dir<CH_SPACEDIM; ++dir){
            const IntVectSet& ivs = cfivs.d[dir].getFineIVS();
            for(IVSIterator it(ivs); it.ok(); ++it)
              {
                tau(it(), i) = 0.0;
              }
            const IntVectSet& ivs2 =  cfivs.d[dir+CH_SPACEDIM].getFineIVS();
            for(IVSIterator it(ivs2); it.ok(); ++it)
              {
                tau(it(), i) = 0.0;
              }
          }
        }
    }
  bool finished = false;

  int counter[2] = {0,0};
  int totalCounter[2];
  int iteration = 0;
  int max_iter = dbl.size()*2;
  while(!finished)
    {
      
      for(DataIterator dit(dbl); dit.ok(); ++dit)
        {
          FArrayBox&       tau = a_opacity[dit];
          const FArrayBox& rho = a_density[dit];
          const FArrayBox& B   = a_magneticField[dit];
          const Box&      region = dbl.get(dit());

          for(int dir=0; dir<CH_SPACEDIM; dir++)
            {
              if(region.smallEnd(dir) == m_domain.domainBox().smallEnd(dir))
                {
                  Box b = region;
                  b.setRange(dir, region.smallEnd(dir), 1);
                  if(rho.max(b, 0) == 0)
                    {
                      b.shift(dir, -1);
                      b.grow(1);
                      b.grow(dir,-1);
                      tau.setVal(0, b, 0, 2);
                    }
                }
              if(region.bigEnd(dir) == m_domain.domainBox().bigEnd(dir))
                {
                  Box b = region;
                  b.setRange(dir, region.bigEnd(dir), 1);
                  if(rho.max(b, 0) == 0)
                    {
                      b.shift(dir, 1);
                      b.grow(1);
                      b.grow(dir,-1);
                      tau.setVal(0, b, 0, 2);
                    }
                }
            }
          for(int i = 0; i<2; ++i)
            {
              if(tau.min(region, i) == -1)
                {
                  integrateBox(tau, i, rho, B, region, a_dx);
                  //pout()<<"["<<i<<","<<tau.min(region, i)<<"] ";
                  if(tau.min(region, i) >= 0.0)
                    {
                      counter[i]++;
                    }
                }
              
            }
        }
      
#ifdef CH_MPI
      MPI_Allreduce(counter, totalCounter, 2 , MPI_INT, MPI_SUM, Chombo_MPI::comm);
#else
      totalCounter[0] = counter[0];
      totalCounter[1] = counter[1];
#endif
      if(totalCounter[0] == dbl.size() && totalCounter[1] == dbl.size())
        {
          finished = true;
        }
      a_opacity.exchange();
      //      char itername[40]; sprintf(itername, "iter_%02d.hdf5",iteration);
      //      writeLevelname(&a_opacity, itername);
      iteration++;
      if(iteration > max_iter) 
        {
          MayDay::Warning("LevelOpacity reached max iter");
          finished = true;
        }
    }
}


void LevelOpacity::integrateBox(FArrayBox& a_tau, 
                                int a_comp, 
                                const FArrayBox& a_rho, 
                                const FArrayBox& a_B, 
                                const Box& a_region, 
                                const RealVect& a_dx) const
{
  int sn = 1;
  if(a_comp == 1) sn=-1;

  bool updating = true;

  Real weight[STENSIZE];
  IntVect sten[STENSIZE];
  Real raylength;

  while(updating)
    {
      updating = false;
      for(BoxIterator bit(a_region); bit.ok(); ++bit)
        {
          RealVect B;
          const IntVect& iv = bit();
          if(a_tau(iv, a_comp) == -1)
            {
              for(int dir=0; dir<CH_SPACEDIM; dir++) 
                {  
                  B[dir] = a_B(iv, dir); 
                }
              B*=sn;
              IntVect ordering, upstream, sample(iv);
              maxOrder(ordering, upstream, sample, B);
              
              if(a_tau(sample, a_comp) == -1)
                {
                  sten[0]=sample; 
                  // optimization, just check one upstream stencil point, if
                  // it's -1, then there isn't any point computing the stencil
                  // weights and points, since the stencil will end up incomplete
                  // anyways.
                }
              else
                {
                  stencil(weight, sten, raylength, a_dx, ordering,
                          upstream, B, iv);
                }
              Real tauTr = 0;
              Real rhoTr = 0;
              int i;
              for(i=0; i<STENSIZE; ++i)
                {
                  if(a_tau(sten[i], a_comp) == -1)
                    {
                      i=STENSIZE+1;
                    }
                  else
                    {
                      tauTr += weight[i]*a_tau(sten[i], a_comp);
                      rhoTr += weight[i]*a_rho(sten[i], 0);
                    }
                }
              if(i==STENSIZE) // we had all upwind cells available for stencil
                {
                  Real rhoAv = (rhoTr + a_rho(iv, 0))/2.0;
                  // I'm assuming rho "ambient" is near zero compared to rhoAv here
                  Real val = raylength*rhoAv + tauTr;
                  
                  a_tau(iv, a_comp) = val;
                  updating = true;
                  
                }
            }
          
        }
      
    }
}

void  LevelOpacity::maxOrder(IntVect& a_order, 
                             IntVect& a_upstream, 
                             IntVect& a_sample, 
                             const RealVect& a_vector) const
{
  D_TERM(a_order[0] = 0;
         Real Bmax = fabs(a_vector[0]);,
         if(fabs(a_vector[1])>Bmax)
           { a_order[0]=1; Bmax = fabs(a_vector[1]);},
         if(fabs(a_vector[2])>Bmax)
           { a_order[0]=2; Bmax = fabs(a_vector[2]);});

  D_TERM( , a_order[1] = (a_order[0]+1)%CH_SPACEDIM;,
          a_order[2] = (a_order[1]+1)%CH_SPACEDIM);

  D_TERM(a_upstream[0] = (a_vector[a_order[0]]>=0) ? -1:1 ;,
         a_upstream[1] = (a_vector[a_order[1]]>=0) ? -1:1 ;,
         a_upstream[2] = (a_vector[a_order[2]]>=0) ? -1:1 );

  D_TERM(a_sample[a_order[0]] += a_upstream[0];,
         a_sample[a_order[1]] += a_upstream[1];,
         a_sample[a_order[2]] += a_upstream[2]);
}
 
void LevelOpacity::stencil(Real* a_weight, IntVect* a_iv,
                           Real& length,   const RealVect&  a_dx,   
                           const IntVect&  a_ordering,
                           const IntVect&  a_upstream, 
                           const RealVect& B,
                           const IntVect&  a_loc) const 
{

  Real Xweight1 = fabs(B[a_ordering[1]]/B[a_ordering[0]]);
  Real Xweight0 = 1.0 - Xweight1;
  Real x = a_dx[a_ordering[0]];
  Real y = Xweight1*a_dx[a_ordering[1]];
  length = sqrt(x*x + y*y);
  CH_assert(Xweight0<=1.0);
  CH_assert(Xweight1<=1.0);
  
#if CH_SPACEDIM == 2
  a_weight[0]=Xweight0;
  a_weight[1]=Xweight1;
  a_iv[0] = a_loc;
  a_iv[0][a_ordering[0]] += a_upstream[0];

  a_iv[1] = a_iv[0];
  a_iv[1][a_ordering[1]] += a_upstream[1];

#endif
#if CH_SPACEDIM == 3
  Real Yweight0 = fabs(B[a_ordering[2]]/B[a_ordering[0]]);
  Real Yweight1 = 1.0 - Yweight0;
  Real z = Yweight0*a_dx[a_ordering[2]];
  length = sqrt(length*length + z*z);

  a_weight[0] = Xweight0 * Yweight1;
  a_weight[1] = Xweight1 * Yweight1;
  a_weight[2] = Xweight0 * Yweight0;
  a_weight[3] = Xweight1 * Yweight0;

  a_iv[0] = a_loc;
  a_iv[0][a_ordering[0]] += a_upstream[0];
  
  a_iv[1] = a_iv[0];
  a_iv[1][a_ordering[1]] += a_upstream[1];

  a_iv[2] = a_iv[0];
  a_iv[2][a_ordering[2]] += a_upstream[2];
  
  a_iv[3] = a_iv[1];
  a_iv[3][a_ordering[2]] += a_upstream[2];
  

#endif
}
  
