/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// F. Miniati th, 21.10.04

#ifndef _GAUSS_BC_H_
#define _GAUSS_BC_H_

#include "Box.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "RealVect.H"
#include "LoHiSide.H"
#include "GhostBC.H"

/// Class to enforce proper boundary conditions for the spherical collapse of a cloud
/**
    We require that the potential on the boundary of the box is set by the 
    mass enclosed by the radius at which the cell find itself. We use this class
    for two problems:
    1) the collapse of a pressureless cloud in which case the mass referred to 
    above is the total mass in the box.
    2) the cosmological, secondary infall problem in which case the mass is due to 
    the overdensity with respect to the average.

    Recall that  BoxGhostBC is a class to encapsulate the 
    operations of ghost-cell boundary conditions at a face.  
    If the solution is phi and the face normal direction is x,
    the boundary conditions usually used can be expressed as
    A*phi + B*dphi/dx = C.
    Our conditions require, A=1, B=0, C=f(r)
    These functions are only meant to be called by DomainGhostBC
 */
class GaussBC: public BoxGhostBC
{
public:
  friend class DomainGhostBC;

  /// Null constructor
  GaussBC();

  /// Basic constructor -- sets Interval to (0,0)
  GaussBC(int dir,
          Side::LoHiSide sd);

  /// Full constructor
  GaussBC(int             dir,
          Side::LoHiSide  sd, 
          const Interval& a_comps);

  /// Destructor
  ~GaussBC();

  /// 
  virtual void setParameters(const Real& a_mass, const Real& a_lScale =1, 
                             const RealVect& a_center = RealVect(D_DECL(0.5,0.5,0.5)));

protected:
  /// 
  /**
     Fills coefficient arrays -- 
     fills neumann factor with 1.0
     and other two fields with 0.0.
   */
  virtual void fillBCValues(FArrayBox& a_neumfac,
                            FArrayBox& a_dircfac,
                            FArrayBox& a_inhmval,
                            Real       a_dx,
                            const Box& a_domain) const;

  ///
  /**
     Fills coefficient arrays --
     fills neumann factor with 1.0
     and other two fields with 0.0.
   */
  virtual void fillBCValues(FArrayBox&           a_neumfac,
                            FArrayBox&           a_dircfac,
                            FArrayBox&           a_inhmval,
                            Real                 a_dx,
                            const ProblemDomain& a_domain) const;

  ///
  /**
     Virtual constructor workaround.
  */
  BoxGhostBC* new_boxghostbc() const;

  /// mass that determines the potential at the boundary cells
  Real m_mass;

  /// scale to convert dx -> dl= scale*dx
  Real m_scale;

  /// location of the center of mass of m_mass
  RealVect m_center;

private:
  GaussBC(const GaussBC& a_neuin){;}
  virtual void operator=(const GaussBC& a_neuin){;}
};

#endif
