/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "Box.H"
#include "BoxIterator.H"
#include "FArrayBox.H"
#include "REAL.H"
#include "SPACE.H"
#include "Tuple.H"
#include "Vector.H"
#include "LoHiSide.H"

#include "GaussBC.H"
#include "MayDay.H"
#include "CONSTANTS_C.H"

GaussBC::GaussBC() : BoxGhostBC()
{
}

GaussBC::~GaussBC() 
{
}

void GaussBC::setParameters(const Real& a_mass, const Real& a_scale, 
		            const RealVect& a_center)
{
  m_mass = a_mass;
  m_scale = a_scale;
  m_center = a_center;
}

// set the BC according to the potential genearated by an interior spherical mass
// distribution
void GaussBC::fillBCValues(FArrayBox& a_neumfac,
                           FArrayBox& a_dircfac,
                           FArrayBox& a_inhmval,
                           Real       a_dx,
                           const Box& a_domain) const
{
  CH_assert(!a_neumfac.box().isEmpty());
  CH_assert(!a_dircfac.box().isEmpty());
  CH_assert(!a_inhmval.box().isEmpty());

  a_dircfac.setVal(1.0);
  a_neumfac.setVal(0.0);

  /*
  cout << endl;
  cout << " dir " << m_direction << endl;
  cout << " side " << m_side << endl;
  */
  const int edge = (m_side==Side::Lo ? 1 : 0);

  for (BoxIterator bi(a_inhmval.box()); bi.ok(); ++bi)
  {
    //    cout << " indx " << bi() << endl;

    Real r = 0.0;
    for (int dir=0; dir<SpaceDim; ++dir) 
    {
      // distance along this dir to the center of the cell face
      Real d;
      if (dir == m_direction)
        d = (bi()[dir]+edge)*a_dx - m_center[dir];
      else
	d = (bi()[dir]+half)*a_dx - m_center[dir];

      r += pow(d,2);
    }
    r = sqrt(r) * m_scale;

    a_inhmval(bi(),0) = -m_mass/r;
    if (a_inhmval(bi(),0) >= 0.0) cout << " phi > 0 : m_mass = " << m_mass << "  r = " << r << endl;
  }
}

void GaussBC::fillBCValues(FArrayBox&           a_neumfac,
                             FArrayBox&           a_dircfac,
                             FArrayBox&           a_inhmval,
                             Real                 a_dx,
                             const ProblemDomain& a_domain) const
{
  CH_assert(!a_neumfac.box().isEmpty());
  CH_assert(!a_dircfac.box().isEmpty());
  CH_assert(!a_inhmval.box().isEmpty());

  a_dircfac.setVal(1.0);
  a_neumfac.setVal(0.0);

  /*
  cout << endl;
  cout << " dir " << m_direction << endl;
  cout << " side " << m_side << endl;
  */
  const int edge = (m_side==Side::Lo ? 1 : 0);

  for (BoxIterator bi(a_inhmval.box()); bi.ok(); ++bi)
  {
    //    cout << " indx " << bi() << endl;

    Real r = 0.0;
    for (int dir=0; dir<SpaceDim; ++dir) 
    {
      // distance along this dir to the center of the cell face
      Real d;
      if (dir == m_direction)
        d = (bi()[dir]+edge)*a_dx - m_center[dir];
      else
	d = (bi()[dir]+half)*a_dx - m_center[dir];

      //      r += pow(d,2);
      r += d*d;
    }
    r = sqrt(r) * m_scale;

    a_inhmval(bi(),0) = -m_mass/r;
  }
}

BoxGhostBC* GaussBC::new_boxghostbc() const
{
  GaussBC* newop = new GaussBC();

  if (newop == NULL)
    {
      MayDay::Error("Out of Memory in GaussBC::new_boxghostbc");
    }

  newop->setParameters(m_mass,m_scale,m_center);

  return static_cast<BoxGhostBC*>(newop);
}

GaussBC::GaussBC(int             a_dir,
                     Side::LoHiSide  a_side,
                     const Interval& a_comps)
  :BoxGhostBC(a_dir,a_side,a_comps)
{
}

GaussBC::GaussBC(int a_dir,
                     Side::LoHiSide a_side)
  :BoxGhostBC(a_dir,a_side)
{
}
