

#ifndef  TESTSIBC_H_
#define  TESTSIBC_H_

#include  <iostream>

#include "LevelData.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "RealVect.H"
#include "AMRIO.H"

#include "PhysIBC.H"
#include "BoundariesF_F.H"

// Available problems
#define PROBLEM_EXPLOSION      0
#define PROBLEM_WINDCOLLISION  1
#define PROBLEM_EXPANDINGSHOCK 2
#define PROBLEM_JEANS          3

/// 
/**
   Parameters:
     a_gamma    - Gamma for polytropic, gamma-law gas
     a_radius   - Initial radius of the explosion
     a_velocity - Initial velocity of the gas
     a_artvisc  - Artificial viscosity coefficient

     a_rsTolerance - tolerance for the Riemann solver 
     a_max_Rs_iter - max number of iteration in the Riemann solver 
     a_Sthreshold  - min Eth/Etot to switch from Etot to entropy (S) eq.
 */
class TestsIBC : public PhysIBC//, public PartIBC
{
public:
  /// Null constructor
  TestsIBC();
  
  /// Constructor which defines parameters used by Fortran routines
  TestsIBC(const std::string a_problem,
           const Real&     a_gamma,
	   const Real&     a_ms,
           const RealVect& a_center,
           const Real&     a_size,
           const RealVect& a_velocity,
           const Real&     a_artvisc,
           const Real&     a_rsTolerance,
           const Real&     a_maxRsIter,
           const Real&     a_maxMach);

  /// Destructor
  ~TestsIBC() {};

  /// Sets parameters in a common block used by Fortran routines
  void setFortranCommon(const Real&     a_gamma,
			const Real&     a_ms,
                        const RealVect& a_center,
                        const Real&     a_size,
                        const RealVect& a_velocity,
                        const Real&     a_artvisc,
                        const Real&     a_rsTolerance,
                        const Real&     a_maxRsIter,
                        const Real&     a_maxMach);

  // set the specific test problem
  void setTestProblem(const std::string a_problem); 

  // set the specific test problem
  void setTestProblem(const int& a_problem);

  /// Set the flag m_isFortranCommonSet to true
  /**
     Set the flag m_isFortranCommonSet to true so that new IBCs made with
     new_physIBC() will have this flag set without calling setFortranCommon()
     (this is a clumsy design and should be improved).
   */
  void setFortranCommonSet();

  /// Factory method - this object is its own factory
  /**
     Return a pointer to a new PhysIBC object with m_isDefined = false (i.e.,
     its define() must be called before it is used) and m_isFortranCommonSet
     set to value of m_isFortranCommonset in the current (factory) object.
   */
  PhysIBC *new_physIBC();

  /// Set boundary primitive values.
  void primBC(FArrayBox&            a_WGdnv,
              const FArrayBox&      a_Wextrap,
              const FArrayBox&      a_W,
              const int&            a_dir,
              const Side::LoHiSide& a_side,
              const Real&           a_time);
  
  /// Set boundary slopes
  /**
     The boundary slopes in a_dW are already set to one sided difference
     approximations.  If this function doesn't change them they will be
     used for the slopes at the boundaries.
   */
  void setBdrySlopes(FArrayBox&       a_dW,
                     const FArrayBox& a_W,
                     const int&       a_dir,
                     const Real&      a_time);

  /// Adjust boundary fluxes to account for artificial viscosity
  void artViscBC(FArrayBox&       a_F,
                 const FArrayBox& a_U,
                 const FArrayBox& a_divVel,
                 const int&       a_dir,
                 const Real&      a_time);
  
  /// Set up initial conditions for gas and particles
  void initialize(LevelData<FArrayBox>& a_U);

  // does nothing in this implementation 
  //void initialize(LevelData<PBinFab>& a_DM);
  
protected:
  // True if the Fortran common block has been set
  bool m_isFortranCommonSet;

  // True if the problem string has been set
  bool m_isProblemSet;

  // specific test problem
  int m_problem;
};

#endif
