//  -*- Mode: C++; Modified: "Wed 19 Jan 2005 15:05:55 by dbs"; -*- 

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  v \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of 
// California, through Lawrence Berkeley National Laboratory (subject to 
// receipt of any required approvals from U.S. Dept. of Energy).  All 
// rights reserved.
// 
// Redistribution and use in source and binary forms, with or without 
// modification, are permitted provided that the following conditions are met:
// (1) Redistributions of source code must retain the above copyright 
// notice, this list of conditions and the following disclaimer.<br>
// (2) Redistributions in binary form must reproduce the above copyright 
// notice, this list of conditions and the following disclaimer in the 
// documentation and/or other materials provided with the distribution.<br>
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S. 
// Dept. of Energy nor the names of its contributors may be used to endorse 
// or promote products derived from this software without specific prior 
// written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// You are under no obligation whatsoever to provide any bug fixes, 
// patches, or upgrades to the features, functionality or performance of 
// the source code ("Enhancements") to anyone; however, if you choose to 
// make your Enhancements available either publicly, or directly to 
// Lawrence Berkeley National Laboratory, without imposing a separate 
// written license agreement for such Enhancements, then you hereby grant 
// the following license: a non-exclusive, royalty-free perpetual license 
// to install, use, modify, prepare derivative works, incorporate into 
// other computer software, distribute, and sublicense such Enhancements or 
// derivative works thereof, in binary and source code form.
// 
// TRADEMARKS. Product and company names mentioned herein may be the trademarks of
// their respective owners. Any rights not expressly granted herein are reserved. 

#include "LayoutIterator.H"
#include "ComputeEnergy.H"
#include "ComputeEnergyF_F.H"


////////////////////////////////////////////////////////////////

Real computeEnergy(const Vector<LevelData<FArrayBox>* >& a_phi, 
                   const Vector<LevelData<FArrayBox>* >& a_pi, 
                   const Vector<int>&                    a_nRefFine,
                   Real                                  a_dxCrse,
                   Real                                  a_lambda)
{
  int numLevels = a_phi.size();
  Real sum, sumLevel;

  sum = 0.0;

  // loop over levels  
  for (int lev = 0; lev < numLevels; lev++) {
    //in case there are extra levels which are not defined
    if (a_phi[lev] != NULL)
    {
      CH_assert(a_phi[lev]->isDefined());
      CH_assert(a_pi [lev]->isDefined());

      const DisjointBoxLayout* finerGridsPtr;

      if (lev < numLevels-1)
      {
        finerGridsPtr = &(a_phi[lev+1]->getBoxes());
      }
      else
      {
        finerGridsPtr = NULL;
      }
      // compute on current level with zeros in covered cells
      sumLevel = computeEnergy(*(a_phi[lev]),*(a_pi[lev]),finerGridsPtr,a_nRefFine[lev],
                               a_dxCrse, a_lambda);
      sum += sumLevel;

      a_dxCrse = a_dxCrse/a_nRefFine[lev];      
    }
  }
  
  return sum;
}

////////////////////////////////////////////////////////////////

Real computeEnergy(const LevelData<FArrayBox>& a_phi,
                   const LevelData<FArrayBox>& a_pi, 
                   const DisjointBoxLayout*    a_finerGridsPtr,
                   int                         a_nRefFine,
                   Real                        a_dx,
                   Real                        a_lambda)
{
  Real sum, sumLevel;
  sum = 0.0;

  const DisjointBoxLayout& levelGrids = a_phi.getBoxes();

  // compute the energy in box, overwrite the cells 
  // that are covered by finer levels with zero, then sum
  for (DataIterator dit = a_phi.dataIterator(); dit.ok(); ++dit) 
  {
    FArrayBox energy(levelGrids.get(dit()),1);

    FORT_COMPUTEENERGYINBOX(CHF_FRA1(energy,0),
                            CHF_CONST_FRA(a_phi[dit()]),
                            CHF_CONST_FRA(a_pi[dit()]),
                            CHF_BOX(levelGrids.get(dit())),
                            CHF_CONST_REAL(a_lambda), CHF_CONST_REAL(a_dx));

    if (a_finerGridsPtr != NULL)
    {  
      // now loop over fine boxes and set covered regions to 0
      for (LayoutIterator litFine = a_finerGridsPtr->layoutIterator(); litFine.ok(); ++litFine)
      {
        Box coveredBox(a_finerGridsPtr->get(litFine()));
        coveredBox.coarsen(a_nRefFine);
        coveredBox &= levelGrids.get(dit());
        if (!coveredBox.isEmpty())
        {
          energy.setVal(0.0,coveredBox,0);
        }
      }  // end loop over fine-grid boxes
    } // end if there is a finer level
    
    sumLevel = energy.sum(0); //[NOTE: redundant var to simplify debugging]
    sum += sumLevel;
  } // end loop over this level's grids


  // do broadcast/gather thing here
#ifdef CH_MPI
  Real recv;
  int result = MPI_Allreduce(&sum, &recv, 1, MPI_CH_REAL,
                             MPI_SUM, Chombo_MPI::comm);
  if (result != MPI_SUCCESS) 
    {
      MayDay::Error("Sorry, but I had a communication error in computeSum");
    }

  sum = recv;
#endif
  return sum;
}
