/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>

#include "ParmParse.H"
#include "CH_HDF5.H"
#include "parstream.H"

#include "EBExplosionIBCFactory.H"
#include "EBPatchPolytropicFactory.H"

#include "ArrayView.H"
#include "EBFABView.H"
#include "EBPatchPolytropic.H"

#include "EBAMRNSFactory.H"
#include "EBAMRNS.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "BaseIVFactory.H"
#include "RampBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "GeometryShop.H"
#include "PolyGeom.H"
#include "LevelData.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"
#include "VoFIterator.H"
#include "EBArith.H"
#include "AllRegularService.H"
#include "EBArrayView.H"
#include "EBLevelRedist.H"
#include "RedistStencil.H"
#include "SlabService.H"
#include "DebugDump.H"
#include "EBDebugDump.H"
#include "EBAMRIO.H"
#include "GodunovGeom.H"
#include "UnionIF.H"
#include "PlaneIF.H"
#include "IntersectionIF.H"
#include "ImplicitFunctionBL.H"
#include "ComplementIF.H"
#include "LatheIF.H"
 

IntVect ivdebug(D_DECL(10,54,0));
VolIndex vofdebug(ivdebug, 0);
/***************/

void get_polygons( Vector<Vector<RealVect> >& polygons, const char* name );

void amrNS( const Box& a_domain, 
            const Real& a_domainLength,
            const Real& a_dx )
{
// set up IBCFactory  --------------------------------------------------
  pout() << "explosion initial conditions" << endl;
  pout() << "solid wall boundary conditions" << endl;

  ParmParse pp;
  Real gamma = 1.4;
  pp.get("gamma",gamma);

  Real size = 0.25;
  pp.get("explosion_size",size);

  Real p0, r0, p1, r1;
  pp.get("explosion_p0", p0);
  pp.get("explosion_r0", r0);
  pp.get("explosion_p1", p1);
  pp.get("explosion_r1", r1);

  int geom=0;
  pp.get("geometry",geom );
  bool doRZCoords = (geom == 1);

  vector<Real> centerpp(SpaceDim,0.5);
  RealVect center;
  pp.getarr("explosion_center",centerpp,0,SpaceDim);
  for( int i = 0; i < SpaceDim; i++) 
     center[i] = centerpp[i];
      
  EBExplosionIBCFactory bcfactory( gamma, size, p0, r0, p1, r1, center, doRZCoords ); 

// set up patch integrator ------------------------------------------------
  int ifourth, iflatten, iartvisc;
  pp.get("use_fourth_order_slopes", ifourth);
  pp.get("use_flattening"         , iflatten);
  pp.get("use_art_visc"           , iartvisc);
  bool useFourthOrderSlopes = (ifourth  ==1);
  bool useFlattening        = (iflatten ==1);
  bool useArtificialVisc    = (iartvisc ==1);

  bool useLimiting = true;
  
  EBPatchPolytropicFactory patchGamma(&bcfactory,
                                      gamma, 
                                      useFourthOrderSlopes,
                                      useFlattening,
                                      useArtificialVisc,
                                      useLimiting,
                                      doRZCoords);

// set up EBAMRNSFactory ----------------------------------------------------
  Real cfl = 0.8;
  pp.get("cfl",cfl);

  Real initialCFL = 0.1;
  pp.get("initial_cfl",initialCFL);

  int redistRad = 0;
  pp.get("redist_radius",redistRad);

  Real refineThresh = 0.3;
  pp.get ("refine_thresh",refineThresh);

  int tagBufferSize = 3;
  pp.get("tag_buffer_size",tagBufferSize);

  int verbosity;
  pp.get("verbosity",verbosity);
  CH_assert(verbosity >= 0);

  int iusemassredist;
  pp.get("use_mass_redist", iusemassredist);
  bool useMassRedist    = (iusemassredist ==1);

  int dosmushing;
  pp.get("do_smushing", dosmushing);
  bool doSmushing = (dosmushing==1);
  
  //  bool hasSourceTerm = true;
  bool hasSourceTerm = doRZCoords;

  Real rom = 277.15;
  pp.get("R_over_M",rom);

  Real mu=0, lambda=0, kappa=0;
  pp.get("mu",mu);
  pp.get("lambda",lambda);
  pp.get("kappa",kappa);

  int force;
  pp.get("forcing_case",force);
  EBAMRNSFactory amrg_fact(     initialCFL,
                                cfl,
                                redistRad,
                                a_domainLength,
                                refineThresh,
                                tagBufferSize,
                                verbosity,
                                useMassRedist,
                                doSmushing,
                                doRZCoords,
                                hasSourceTerm,
                                &patchGamma,
				gamma, 
                                rom,
				mu,
				lambda,
				kappa,
				force );
  

// define AMR  -----------------------------------------------------------
  int max_level = 0;
  pp.get("max_level",max_level);

  int num_read_levels = Max(max_level,1);
  std::vector<int> ref_ratios; // (num_read_levels,1);
  // note this requires a ref_ratio to be defined for the 
  // finest level (even though it will never be used)
  pp.getarr( "ref_ratio", ref_ratios, 0, num_read_levels+1 );

  bool is_periodic[SpaceDim];
  for( int i = 0; i < SpaceDim; ++i ) 
     is_periodic[i] = false;
  ProblemDomain prob_domain(a_domain.smallEnd(),
                            a_domain.bigEnd(),
                            is_periodic);

  AMR amr;
  amr.define( max_level, ref_ratios, prob_domain, &amrg_fact );

//set AMR parameters -------------------------------------------------------
  Real fixed_dt = -1;
  pp.get("fixed_dt",fixed_dt);
  if (fixed_dt > 0)
    {
      amr.fixedDt(fixed_dt);
    }

  int max_grid_size = 32;
  pp.get("max_grid_size",max_grid_size);
  amr.maxGridSize(max_grid_size);

  int block_factor = 1;
  pp.get("block_factor",block_factor);
  amr.blockFactor(block_factor);


  Real fill_ratio = 0.75;
  pp.get("fill_ratio",fill_ratio);
  amr.fillRatio(fill_ratio);

  // the hyperbolic codes use a grid buffer of 1
  amr.gridBufferSize(1);

  // set output parameters
  int checkpoint_interval = 0;
  pp.get("checkpoint_interval",checkpoint_interval);
  amr.checkpointInterval(checkpoint_interval);

  int plot_interval = 0;
  pp.get("plot_interval",plot_interval);
  amr.plotInterval(plot_interval);

  std::vector<int> regrid_intervals; // (num_read_levels,1);
  pp.getarr("regrid_interval",regrid_intervals,0,num_read_levels);
  amr.regridIntervals(regrid_intervals);

  Real max_dt_growth = 1.1;
  pp.get("max_dt_growth",max_dt_growth);
  amr.maxDtGrow(max_dt_growth);

  Real dt_tolerance_factor = 1.1;
  pp.get("dt_tolerance_factor",dt_tolerance_factor);
  amr.dtToleranceFactor(dt_tolerance_factor);

  if (pp.contains("plot_prefix"))
    {
      std::string prefix;
      pp.get("plot_prefix",prefix);
      amr.plotPrefix(prefix);
    }

  if (pp.contains("chk_prefix"))
    {
      std::string prefix;
      pp.get("chk_prefix",prefix);
      amr.checkpointPrefix(prefix);
    }

  amr.verbosity(verbosity);

  if (!pp.contains("restart_file"))
    {
      if (!pp.contains("fixed_hierarchy"))
        {
          // initialize from scratch for AMR run
          // initialize hierarchy of levels
          amr.setupForNewAMRRun();
        }
      else
        {
          std::string grid_file;
          pp.get("fixed_hierarchy",grid_file);

          // initialize from a list of grids in "grid_file"
          Vector<Vector<Box> > amrGrids(max_level+1);
          godunovFixedGrids(amrGrids,
                            prob_domain,
                            max_level,
                            max_grid_size,
                            block_factor,
                            verbosity,
                            grid_file);
          amr.setupForFixedHierarchyRun(amrGrids,1);
        }
    }
  else
    {
      std::string restart_file;
      pp.get("restart_file",restart_file);
      pout() << "restarting from " << restart_file << endl;

#ifdef CH_USE_HDF5
      HDF5Handle handle(restart_file,HDF5Handle::OPEN_RDONLY);
      // read from checkpoint file
      amr.setupForRestart(handle);
      handle.close();
#else
      MayDay::Error("amrGodunov restart only defined with hdf5");
#endif
    }


// Run 
  int nstop = 0;
  pp.get("max_step",nstop);
  Real stopTime = 0.0;
  pp.get("max_time",stopTime);
  amr.run(stopTime,nstop);


// output last pltfile and statistics
  amr.conclude();

}


//-----------------------------------------------------------------------
//
//-----------------------------------------------------------------------
UnionIF* makeCrossSection(const Vector<Vector<RealVect> >& a_polygons)
{
  // The final result
  UnionIF* retval;

  // Get the number of polygons and make this inside of the domain
  // the inside of the polygons
  int numPolys = a_polygons.size();
  bool inside = true;

  // A list of all the polygons as implicit functions
  Vector<BaseImplicitFunction*> polytopes;
  polytopes.resize(0);

  // Process each polygon
  for (int p = 0; p < numPolys; p++)
  {
    // All the half planes/spaces used to make a polygon
    Vector<BaseImplicitFunction*> planes;
    planes.resize(0);

    // Get the current polygon (as a vector of points)
    const Vector<RealVect>& polygon = a_polygons[p];

    // Get the number of points in the polygon
    int numPts = polygon.size();

    // Process each pair of points
    for (int n = 0; n < numPts; n++)
    {
      // The normal and point is space used to specify each half plane/space
      RealVect normal(RealVect::Zero);
      RealVect point;

      // Set the normal remembering that the last point connects to the first
      // point.
      normal[0] = -(polygon[(n+1) % numPts][1] - polygon[n][1]);
      normal[1] =  (polygon[(n+1) % numPts][0] - polygon[n][0]);

      point = polygon[n];

      // Generate the appropriate half plane/space (as an implicit function)
      PlaneIF* plane;
      plane = new PlaneIF(normal,point,inside);

      // Save the result
      planes.push_back(plane);
    }

    // Intersect all the half planes/spaces to create an implicit function
    // that represents the polygon
    IntersectionIF* polygonIF = new IntersectionIF(planes);

    polytopes.push_back(polygonIF);
  }

  // Union all the polygon implicit functions to get the implicit function
  // returned
  retval = new UnionIF(polytopes);

  return retval;
}


//-----------------------------------------------------------------------
// Need on input file:  
//      n_cell           - ni, nj, nk for coarse domain
//      prob_lo, prob_hi - coordinates in 3D of lower left and upper
//                         right corners.
//      max_level        - scalar, number of levels.
//      ref_ratio        - one ratio for each level refinement (max_level-1)
// Outputs all parameters:
//-----------------------------------------------------------------------
void get_domain_info( Box&      a_coarsestDomain,
                      Box&      a_finestDomain,
                      Real&     a_domainLength,
                      RealVect& a_origin,
                      Real&     a_dxCoarsest,
                      Real&     a_dxFinest   )
{
  ParmParse pp;

  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);

  CH_assert(n_cell.size() == SpaceDim);

  IntVect lo = IntVect::Zero;
  IntVect hi;

  for (int ivec = 0; ivec < SpaceDim; ivec++)
  {
    if (n_cell[ivec] <= 0)
    {
      pout() << "Bogus number of cells input = " << n_cell[ivec];
      exit(1);
    }

    hi[ivec] = n_cell[ivec] - 1;
  }

  a_coarsestDomain.setSmall(lo);
  a_coarsestDomain.setBig(hi);

  Vector<Real> prob_lo(SpaceDim,0.0);
  Vector<Real> prob_hi(SpaceDim,1.0);

  pp.getarr("prob_lo",prob_lo,0,SpaceDim);
  pp.getarr("prob_hi",prob_hi,0,SpaceDim);

  a_domainLength = (prob_hi[0]-prob_lo[0]);
  a_dxCoarsest = (prob_hi[0]-prob_lo[0])/n_cell[0];

  int max_level = 0;
  std::vector<int> refRatios; // (num_read_levels,1);
  // note this requires a refRatio to be defined for the 
  // finest level (even though it will never be used)

  pp.get("max_level",max_level);
  int num_read_levels = Max(max_level,1);
  pp.getarr("ref_ratio",refRatios,0,num_read_levels+1);

  a_finestDomain = a_coarsestDomain;
  for(int ilev = 0; ilev < max_level; ilev++)
     a_finestDomain.refine(refRatios[ilev]);

  a_dxFinest = a_dxCoarsest;
  int ebMaxCoarsen = 0;
  for(int ilev = 0; ilev < max_level; ilev++)
    {
      a_dxFinest /= refRatios[ilev];
      ebMaxCoarsen += refRatios[ilev]/2;
    }

  for( int idir = 0; idir < SpaceDim; idir++ )
     a_origin[idir] = prob_lo[idir];
}


//-----------------------------------------------------------------------
// Makes the nozzle geometry.
// Needs polygon descriptions on input-file for xxx='body' and xxx='poppet':
//      'num_polygons_in_xxx' - Number of polygons forming the object
//      'points_in_xxx_polygons' - For each polygon the number of points in it.
//                                    (vector of num_polygons_in_xxx values)
//      'xxxptij'  - point j in polygon i for xxx (vector of three coordinates). 
//                               0 <= i < num_polygons_in_xxx
//                               0 <= j < points_in_xxx_polygons(i)
//
// Outputs the associated GeometryShop object.
//-----------------------------------------------------------------------
GeometryShop* make_nozzle()
{
  bool     insideRegular;

  Vector<Vector<RealVect> > polygonsBody;
  get_polygons( polygonsBody, "body");

  Vector<Vector<RealVect> > polygonsPoppet;
  get_polygons( polygonsPoppet, "poppet");

  // Make the vector of (convex) polygons (vectors of points) into a union
  // of convex polygons, each made from the intersection of a set of half
  // planes/spaces - all represented by implicit functions.
  Vector<Vector<RealVect> > polygons = polygonsBody;
  polygons.append(polygonsPoppet);

  UnionIF* crossSection = makeCrossSection(polygons);

  ImplicitFunctionBL *myGeom;

  if (SpaceDim == 2)
  {
    // In 2D use "as is"
    ComplementIF insideOut(*crossSection);
    //    retval = insideOut.newImplicitFunction();
    myGeom = new ImplicitFunctionBL(insideOut);
  }
  else
  {
    // In 3D rotate about the z-axis
    LatheIF implicit(*crossSection,insideRegular);
    //    retval = implicit.newImplicitFunction();
    myGeom = new ImplicitFunctionBL(implicit);
  }
  GeometryShop* workshop = new GeometryShop(*myGeom);
  delete myGeom;
  delete crossSection;
  return workshop;
}


//-----------------------------------------------------------------------
// Read one object consiting of polygons from input file.
//
// Input: name - The name of polygon (= 'xxx' in make_nozzle documentation)
//
// Output: polygons - A list of all polygons making 'xxx'.
//
//-----------------------------------------------------------------------
void get_polygons( Vector<Vector<RealVect> >& polygons, const char* name )
{
   ParmParse pp;
// Get number of polygons
  int n;
  char nname[100];
  sprintf(nname,"num_polygons_in_%s",name);
  pp.get(nname, n);
  polygons.resize(n);

// Get number of points in each polygon
  std::vector<int> points(n);
  sprintf(nname,"points_in_%s_polygons",name);
  pp.getarr(nname,points,0,n);

// Read the points
  for(int ipoly = 0; ipoly < n; ipoly++)
  {
     polygons[ipoly].resize(points[ipoly]);
     for(int ipt = 0; ipt < points[ipoly]; ipt++)
     {
         char pointname[100];
         sprintf(pointname,"%spt%d%d", name, ipoly, ipt );
         std::vector<Real> vecPt(SpaceDim);
         pp.getarr(pointname, vecPt, 0, SpaceDim);
         for(int idir = 0; idir < SpaceDim; idir++)
            polygons[ipoly][ipt][idir] = vecPt[idir];
     }
  }
}


//-----------------------------------------------------------------------
GeometryShop* make_cyl( int cylnr, const Real& dxf )
{
   RealVect c0(RealVect::Zero);
   if( cylnr == 1 )
   {
     c0[0] = 2.5;
     c0[1] = 2.5;
   }

   CylinderBL cyl( 0.5, 0.0, c0, 0, false, dxf ); 
   GeometryShop* workshop = new GeometryShop(cyl);
   return workshop;
}


//-----------------------------------------------------------------------
GeometryShop* make_rectangle( const Real& dxf )
{
   RealVect c0(RealVect::Unit*100.0);
   CylinderBL cyl( 0.5, 0.0, c0, 0, false, dxf ); 
   GeometryShop* workshop = new GeometryShop(cyl);
   return workshop;

}


//-----------------------------------------------------------------------
int 
main(int a_argc, char* a_argv[])
{
#ifdef CH_MPI
  MPI_Init(&a_argc,&a_argv);
  setChomboMPIErrorHandler();
#endif

  char* inFile = NULL;

  if( a_argc > 1 )
  {
     inFile = a_argv[1];
  }
  else
  {
      pout() << "Usage: <executable name> <inputfile>" << endl;
      pout() << "No input file specified" << endl;
      return -1;
  }

  // Parse the command line and the input file (if any)
  ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);


// Domain description:
  Box coarsestDomain, finestDomain;
  Real dxCoarsest, dxFinest, domainLength;
  RealVect origin;

  get_domain_info( coarsestDomain, finestDomain, domainLength, origin,
		   dxCoarsest, dxFinest );

  cout << "Coarsest domain is " << coarsestDomain << endl;
  cout << "Finest domain is   " << finestDomain << endl;
  cout << "Origin is          " << origin << endl;
  cout << "domain length is   " << domainLength << endl << endl;
  cout << "dx goes from " << dxCoarsest << " to " << dxFinest << endl;

// Generate embedded boundary info.
  int geom = 0;
  pp.get( "geometry", geom );

  GeometryShop* geo_ptr;
  if( geom == 1 )
     geo_ptr = make_nozzle();
  else if( geom == 2 )
     geo_ptr = make_cyl(0,dxFinest);
  else if( geom == 3 )
     geo_ptr = make_cyl(1,dxFinest);
  else if( geom == 4 )
     geo_ptr =  make_rectangle(dxFinest);

  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->define( finestDomain, origin, dxFinest, *geo_ptr );

  amrNS( coarsestDomain, domainLength, dxCoarsest );
  delete geo_ptr;

#ifdef CH_MPI
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}

