
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>

#include "ArrayToRaw.H"

void ArrayToRaw::output_array( const LevelData<EBCellFAB>& u, const EBISLayout& ebis, char* file )
{
   double *ar;
   int nc = u.nComp();
   int nboxes = ebis.getDisjointLayout().size();
   int fd=open(file, O_CREAT | O_TRUNC | O_WRONLY, 0660 );   
   write(fd,&nboxes,sizeof(int));

   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      IntVect dims = u[dex].getRegion().size();
      IntVect low  = u[dex].getRegion().smallEnd();

      int ni = dims[0];
      int nj = dims[1];
      ar = new double[ni*nj*nc];
      for( BoxIterator b(u[dex].getRegion()); b.ok(); ++b )
      {
        int i=b()[0]-low[0], j=b()[1]-low[1];
	if( !ebis[dex].isCovered(b()) )
	{
	   Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
           for( int vi=0 ; vi < vofs.size() ; vi++ )
	   {
	      for( int c= 0 ; c<nc ; c++ )
	         ar[c+i*nc+nc*ni*j] = u[dex](vofs[vi],c);
	   }
        }
	else
	   for( int c= 0 ; c<nc ; c++ )
	     ar[c+i*nc+nc*ni*j] = 0;
      }
      write(fd,&nc,sizeof(int));
      write(fd,&ni,sizeof(int));
      write(fd,&nj,sizeof(int));
      write(fd,ar,sizeof(double)*ni*nj*nc);
      delete[] ar;
   }
   close(fd);
}

void ArrayToRaw::output_level_array( const Vector<LevelData<EBCellFAB>*>& u,
 		         const Vector<EBISLayout>& ebis,
		         const Vector<RealVect> dx,
		         const RealVect x0, char* file )

{
   int fd=open(file, O_CREAT | O_TRUNC | O_WRONLY, 0660 );   
   int nlevels = u.size();
   write(fd,&nlevels,sizeof(int));
   write(fd,&x0[0],sizeof(double));
   write(fd,&x0[1],sizeof(double));
   for( int l=0 ; l < nlevels ; l++ )
   {
      write(fd,&dx[l][0],sizeof(double));
      write(fd,&dx[l][1],sizeof(double));
   }
   //   cout << "output_level_array" << endl;

   for( int l=0 ; l < nlevels ; l++ )
   {
     //      cout << "level " << l << endl;
      int nc = u[l]->nComp();
      for( DataIterator li=ebis[l].getDisjointLayout().dataIterator() ; li.ok() ; ++li )
      {
         DataIndex dex(li());
         IntVect dims = (*u[l])[dex].getRegion().size();
         IntVect low  = (*u[l])[dex].getRegion().smallEnd();
         int ni = dims[0];
         int nj = dims[1];
         int *flag    = new int[ni*nj];
         double *ar   = new double[ni*nj*nc];
	 //         cout << "past malloc " << endl;
         for( BoxIterator b((*u[l])[dex].getRegion()); b.ok(); ++b )
         {
            int i=b()[0]-low[0], j=b()[1]-low[1];
	    Vector<VolIndex> vofs = ebis[l][dex].getVoFs(b());
            for( int vi=0 ; vi < vofs.size() ; vi++ )
	    {
	       for( int c= 0 ; c<nc ; c++ )
	          ar[c+i*nc+nc*ni*j] = (*u[l])[dex](vofs[vi],c);
	    }
   	    if( ebis[l][dex].isCovered(b()) )
	       flag[i+ni*j] = 0;
	    else if( ebis[l][dex].isRegular(b()) )
	       flag[i+ni*j] = 1;
	    else
	       flag[i+ni*j] = 2;
	 }
	 //	 cout << "writing..." << endl;
         write(fd,&l, sizeof(int));
         write(fd,&nc,sizeof(int));
         write(fd,&ni,sizeof(int));
         write(fd,&nj,sizeof(int));
         write(fd,&low[0],sizeof(int));
         write(fd,&low[1],sizeof(int));
         write(fd,ar,sizeof(double)*ni*nj*nc);
         write(fd,flag,sizeof(int)*ni*nj);
         delete[] ar;
         delete[] flag;
      }
   }
   close(fd);
}


//-----------------------------------------------------------------------
void ArrayToRaw::print_ebcellfab(  const EBCellFAB& a_u, const char* name )
{
   const EBISBox& eb = a_u.getEBISBox();
   for( BoxIterator b(eb.getRegion()); b.ok(); ++b )
   {
      Vector<VolIndex> vofar = eb.getVoFs(b());
      if( eb.isCovered(b()) )
	 cout << name << "(" << b()[0] <<","<<b()[1] << ") is covered" << endl;	
      for( int iv=0 ; iv < vofar.size() ; iv++ )
      {
	 cout << name << "(" << b()[0] <<","<<b()[1] << ")";
	 cout << "["<<iv<<"]=" << a_u(vofar[iv],0) ;
	 if( vofar.size() >1)
	 {
	    RealVect cen=eb.centroid(vofar[iv]);
	    cout << " offset " << cen ;
	 }
	 cout << endl;
      }
   }
}

//-----------------------------------------------------------------------
void ArrayToRaw::chk_nan( const LevelData<EBCellFAB>& u, const EBISLayout& ebis, char* msg )
{
   int nc = u.nComp();
   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      for( BoxIterator b(u[dex].getRegion()); b.ok(); ++b )
      {
        int i=b()[0], j=b()[1];
	if( !ebis[dex].isCovered(b()) )
	{
	   Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
           for( int vi=0 ; vi < vofs.size() ; vi++ )
	   {
	      for( int c= 0 ; c<nc ; c++ )
	      {
		if( isnan(u[dex](vofs[vi],c)) || isinf(u[dex](vofs[vi],c)) )
		  cout << msg << " (" << c << ": " << i << "," << j << ")" << endl;
	      }
	   }
	}
      }
   }
}
