/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "BoxIterator.H"

#include "DirichletHelmholtzDomainBC.H"


//-----------------------------------------------------------------------
DirichletHelmholtzDomainBC::DirichletHelmholtzDomainBC():HelmholtzDomainBC()
{}


//-----------------------------------------------------------------------
DirichletHelmholtzDomainBC::~DirichletHelmholtzDomainBC()
{}


//-----------------------------------------------------------------------
void DirichletHelmholtzDomainBC::get_faceflux(
			        BaseFab<Real>&       a_faceFlux,
                                const BaseFab<Real>& a_phi,
                                const RealVect&      a_probLo,
                                const RealVect&      a_dx,
                                int                  a_idir,
                                Side::LoHiSide       a_side,
                                bool                 a_use_homogeneous,
				const Real           a_time )
{
   CH_assert( a_use_homogeneous || m_use_function || m_use_value );
   int iside   = a_side == Side::Lo ? 1 : -1 ;
   IntVect ofs = IntVect::Zero;
   ofs[a_idir] = iside;
   Real ih     = m_order == 2 ? iside/(3.0*a_dx[a_idir]) : 2.0*iside/a_dx[a_idir];

   if( !a_use_homogeneous && m_use_function )
   {
      BoxIterator bit(a_faceFlux.box());
      for( bit.begin(); bit.ok(); ++bit )
      {
	 IntVect iv = bit();
	 RealVect point;
	 for( int d=0;d<SpaceDim;d++ )
	    point[d] = (iv[d]+0.5*(1-ofs[d]))*a_dx[d]+a_probLo[d];
	 Vector<Real> value = m_func->get_value( point, a_time );
         Real fact = 1.0;
	 if( a_idir == 0 && m_rz )
	 {
	    Real r = static_cast<Real>(iv[0]);
	    r  = (r+0.5)*a_dx[0]+a_probLo[0];
	    Real rb=r-iside*a_dx[0]*0.5;
	    fact = rb/r;
	 }
	 for( int c=0 ; c <a_phi.nComp() ; c++ )
	    if( m_order == 2 )
	       a_faceFlux(iv,c) = ih*( 9*a_phi(iv,c) - a_phi(iv+ofs,c) - 8*value[c] )*fact;
	    else
	       a_faceFlux(iv,c) = ih*( a_phi(iv,c) - value[c] )*fact;
      }
   }
   else
   {
      Vector<Real> value(a_phi.nComp(),0.0);
      if( !a_use_homogeneous )
	 value = m_value;
      BoxIterator bit(a_faceFlux.box());
      for( bit.begin(); bit.ok(); ++bit )
      {
	 IntVect iv = bit();
         Real fact = 1.0;
	 if( a_idir == 0 && m_rz )
	 {
	    Real r = static_cast<Real>(iv[0]);
	    r  = (r+0.5)*a_dx[0]+a_probLo[0];
	    Real rb=r-iside*a_dx[0]*0.5;
	    fact = rb/r;
	 }
	 for( int c = 0 ; c < a_phi.nComp() ; c++ )
	    if( m_order == 2 )
	       a_faceFlux(iv,c) = ih*( 9*a_phi(iv,c)-a_phi(iv+ofs,c)-8*value[c] )*fact;
	    else
	       a_faceFlux(iv,c) = ih*( a_phi(iv,c)-value[c] )*fact;
      }
   }
}


//-----------------------------------------------------------------------
void DirichletHelmholtzDomainBC::get_faceflux(
					   Vector<Real>&    a_faceFlux,
                                           const VolIndex&  a_vof,
                                           const EBCellFAB& a_phi,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
                                           bool             a_use_homogeneous,
					   Real             a_time )
{

   CH_assert( a_use_homogeneous || m_use_function || m_use_value );
   int iside   = a_side == Side::Lo ? 1 : -1 ;
   IntVect ofs=IntVect::Zero;
   ofs[a_idir]  = iside;
   VolIndex voff;
   int order;
   const EBISBox& ebisBox = a_phi.getEBISBox();
   if( m_order == 2 )
       check_2nd_order( ebisBox, a_vof, ofs, order, voff );

// Find the function value at the boundary point
   Vector<Real> value(a_phi.nComp(),0.0);
   if( !a_use_homogeneous && m_use_function )
   {
      RealVect point = static_cast<RealVect>(a_vof.gridIndex());
      for( int d=0 ; d < SpaceDim ; d++ )
         point[d] = (point[d]+0.5*(1-ofs[d]))*a_dx[d] + a_probLo[d];
      value = m_func->get_value( point, a_time );
   }
   else if( !a_use_homogeneous )
      value = m_value;


// Compute the boundary flux
   Real ih = order == 2 ? iside/(3.0*a_dx[a_idir]) : 2.0*iside/a_dx[a_idir];
   Real fact = 1.0;
   if( m_rz )
   {
      Real r = static_cast<Real>(a_vof.gridIndex()[0])+ebisBox.centroid(a_vof)[0];
      r  = (r+0.5)*a_dx[0]+a_probLo[0];
      //      Real rb=r-iside*a_dx[0]*0.5;
      Real rb;
      if( a_idir == 0 )
         rb =( a_vof.gridIndex()[0] + 0.5*(1-iside) )*a_dx[0] + a_probLo[0];
      else
	 rb =( a_vof.gridIndex()[0] + ebisBox.centroid(ebisBox.getFaces(a_vof,a_idir,a_side)[0])[0] +0.5)*a_dx[0] + a_probLo[0];
      //      Real rb =( a_vof.gridIndex()[0] + 0.5*(1-iside) )*a_dx[0] + a_probLo[0];
      fact = rb/r;
   }
// a_faceFlux is d(phi)/dx at boundary (i.e., sign always +)
   for( int c = 0 ; c < a_phi.nComp() ; c++ )
      if( order == 2 )
	 a_faceFlux[c] = ih*( 9*a_phi(a_vof,c)-a_phi(voff,c)-8*value[c])*fact;
      else
	 a_faceFlux[c] = ih*( a_phi(a_vof,c) - value[c] )*fact;


// Reduce face area for boundary cells that are cut by the embedded boundary
   Real faceAreaFrac = 0.0;
   Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
   for (int i = 0; i < faces.size(); i++)
      faceAreaFrac += ebisBox.areaFrac(faces[i]);
   for( int c = 0 ; c < a_phi.nComp() ; c++ )
      a_faceFlux[c] *= faceAreaFrac;
}



//-----------------------------------------------------------------------
void DirichletHelmholtzDomainBC::check_2nd_order( const EBISBox&  a_ebis,
                                                  const VolIndex& a_vof, 
                                                  const IntVect&  a_ofs, 
                                                  int& a_order, 
                                                  VolIndex& a_voff )
{
   Vector<VolIndex> voflist = a_ebis.getVoFs(a_vof.gridIndex()+a_ofs);
   if( voflist.size() == 0 )
      a_order = 1;
   else if( voflist.size() > 1 )
   {
      bool found=false;
      for( int v=0 ; v < voflist.size() ; v++ )
	 if( a_ebis.isConnected( a_vof, voflist[v] ) )
	 {
	    found = true;
	    a_voff = voflist[v];
            a_order = 2;
	 }
      if( !found )
	 a_order = 1;
   }
   else
   {
      a_voff = voflist[0];
      a_order = 2;
   }
}


//-----------------------------------------------------------------------
void DirichletHelmholtzDomainBC::get_diagonal_weight( Vector<Real>& a_wgh,
						      const VolIndex& a_vof, 
						      const EBISBox&  a_ebis,
						      const RealVect& a_probLo,
						      const RealVect& a_dx,
						      int a_idir,
						      Side::LoHiSide a_side )
{
   int iside   = a_side == Side::Lo ? 1 : -1 ;
   IntVect ofs=IntVect::Zero;
   ofs[a_idir]  = iside;
   VolIndex voff;
   int order;
   if( m_order == 2 )
      check_2nd_order( a_ebis, a_vof, ofs, order, voff );

// Center element of boundary flux
   Real ih = order == 2 ? iside/(3.0*a_dx[a_idir]) : 2.0*iside/a_dx[a_idir];
   Real fact = 1;
   if( m_rz )
   {
      Real r = static_cast<Real>(a_vof.gridIndex()[0])+a_ebis.centroid(a_vof)[0];
      r  = (r+0.5)*a_dx[0]+a_probLo[0];
      Real rb;
      if( a_idir == 0 )
         rb =( a_vof.gridIndex()[0] + 0.5*(1-iside) )*a_dx[0] + a_probLo[0];
      else
	 rb =( a_vof.gridIndex()[0] + a_ebis.centroid(a_ebis.getFaces(a_vof,a_idir,a_side)[0])[0] +0.5)*a_dx[0] + a_probLo[0];
      fact = rb/r;
   }
   for( int c = 0 ; c < a_wgh.size() ; c++ )
      if( order == 2 )
	 a_wgh[c] = ih*9*fact;
      else
	 a_wgh[c] = ih*fact;

// Reduce face area for boundary cells that are cut by the embedded boundary
   Real faceAreaFrac = 0.0;
   Vector<FaceIndex> faces = a_ebis.getFaces(a_vof,a_idir,a_side);
   for (int i = 0; i < faces.size(); i++)
      faceAreaFrac += a_ebis.areaFrac(faces[i]);
   for( int c = 0 ; c < a_wgh.size() ; c++ )
      a_wgh[c] *= faceAreaFrac;
}


//-----------------------------------------------------------------------
//                               Factory class
//-----------------------------------------------------------------------
DirichletHelmholtzDomainBCFactory::DirichletHelmholtzDomainBCFactory():HelmholtzDomainBCFactory()
{}

//-----------------------------------------------------------------------
DirichletHelmholtzDomainBCFactory::~DirichletHelmholtzDomainBCFactory()
{}

//-----------------------------------------------------------------------
HelmholtzDomainBC* DirichletHelmholtzDomainBCFactory::create()
{
   DirichletHelmholtzDomainBC* newBC = new DirichletHelmholtzDomainBC();
   newBC->set_order( m_order );
   newBC->set_rz( m_rz );
   if( m_use_function )
      newBC->set_function( m_func );
   else if( m_use_value )
      newBC->set_value( m_value );
   return newBC;
}
