/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _DIRICHLETHELMHOLTZEBBC_H_
#define _DIRICHLETHELMHOLTZEBBC_H_

#include "HelmholtzEBBC.H"

/// Dirichlet boundary conditions on the embedded boundary for Poisson's equation in cylindrical coordinates.
/**
   Computes the boundary flux from a given function value on the boundary. 
   Cylindrical coordinates are used.
 */
class DirichletHelmholtzEBBC: public HelmholtzEBBC
{
public:
  /// Constructor.
  /**
   */
  DirichletHelmholtzEBBC( const ProblemDomain& a_domain,
			  const EBISLayout&    a_layout,
                          const RealVect&      a_dx );

  /// Destructor.
  /**
   */
  virtual ~DirichletHelmholtzEBBC();

  /// Set up internal data structures.
  /**
   */
  void define( const ProblemDomain& a_domain, const EBISLayout& a_layout, const RealVect& a_dx );


  void set_order( int a_order );

  void get_ebflux(Vector<Real>&               a_flux,
                  const VolIndex&             a_vof,
                  const LevelData<EBCellFAB>& a_phi,
                  const DataIndex&            a_dit,
                  const RealVect&             a_probLo,
                  const RealVect&             a_dx,
                  bool                        a_use_homogeneous,
                  Real                        a_time );

private:

  void getFirstOrderStencil(VoFStencil&     a_stencil,
                            Real&           a_weight,
                            const VolIndex& a_vof,
                            const EBISBox&  a_ebisBox,
                            const RealVect& a_dx);

  void getFirstOrderStencil(VoFStencil&     a_stencil,
                            Real&           a_weight,
                            const IntVect&  a_quadrant,
                            const VolIndex& a_vof,
                            const EBISBox&  a_ebisBox,
                            const RealVect& a_dx);
  
 void getSecondOrderStencil(VoFStencil&     a_stencil,
                            Real&           a_weight,
                            const VolIndex& a_vof,
                            const EBISBox&  a_ebisBox,
                            const RealVect& a_dx,
			    const IntVectSet& a_cfivs );

  void johanDropOrder(bool&          a_droppedOrder,
                      const IntVect  a_interpCen[2],
                      const int      a_planeDir[SpaceDim-1],
                      const EBISBox& a_ebisBox,
		      const IntVectSet& a_cfivs );

  VoFStencil BuildEQIStencil(const VolIndex a_vofs[3],
                             const Real&    a_distance);

  VoFStencil BuildEQIStencil(VoFStencil a_vofs[3],
                             const Real& a_distance);

  VoFStencil ivofStencil(const VoFStencil a_iPlaneStencil[2],
                         const Real&      a_distance1,
                         const Real&      a_distance2);

  ProblemDomain m_domain;
  EBISLayout m_layout;
  RealVect   m_dx;

  LayoutData<BaseIVFAB<Real> >       m_fluxWeight;
  LayoutData<BaseIVFAB<VoFStencil> > m_fluxStencil;
};


/// Class that can generate a DirichletHelmholtzEBBC object.
/**
 */
class DirichletHelmholtzEBBCFactory: public HelmholtzEBBCFactory
{
public:
   /// Constructor.
   /**
    */
   DirichletHelmholtzEBBCFactory();

   /// Destructor.
   /**
    */
   virtual ~DirichletHelmholtzEBBCFactory();

   /// Generate a DirichletHelmholtzEBBC object.
   /**
    */
   HelmholtzEBBC* create( const ProblemDomain& a_domain,
			  const EBISLayout&    a_layout,
	                  const RealVect&      a_dx );
};

#endif
