/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cmath>
#include "BoxIterator.H"
#include "EBArith.H"
#include "DirichletHelmholtzEBBC.H"


//-----------------------------------------------------------------------
DirichletHelmholtzEBBC::DirichletHelmholtzEBBC( const ProblemDomain& a_domain,
					        const EBISLayout&   a_layout,
						const RealVect&     a_dx )
  : HelmholtzEBBC()
{
   m_domain = a_domain;
   m_layout = a_layout;
   m_dx =    a_dx;
   define( a_domain, a_layout, a_dx );
}


//-----------------------------------------------------------------------
DirichletHelmholtzEBBC::~DirichletHelmholtzEBBC()
{
}


//-----------------------------------------------------------------------
void DirichletHelmholtzEBBC::set_order( int a_order )
{
   CH_assert(a_order >= 1);
   CH_assert(a_order <= 2);
   bool redefine = a_order != m_order;
   m_order = a_order;
   if( redefine )
      define( m_domain, m_layout, m_dx );
}


//-----------------------------------------------------------------------
void DirichletHelmholtzEBBC::define( const ProblemDomain& a_domain,
				     const EBISLayout& a_layout,
				     const RealVect&  a_dx )
{
   const DisjointBoxLayout& dbl = a_layout.getDisjointLayout();

   m_fluxStencil.define(dbl);
   m_fluxWeight.define(dbl);

// Figure out what the cfivs is.
   LayoutData<IntVectSet> cfivs;
   cfivs.define(dbl);
   for(DataIterator dit = dbl.dataIterator(); dit.ok(); ++dit)
   {
      Box grownBox = grow(dbl.get(dit()), 1);
      IntVectSet ivsGrid(grownBox);
      for(LayoutIterator lit = dbl.layoutIterator(); lit.ok(); ++lit)
          ivsGrid -= dbl.get(lit());
      ivsGrid &= a_domain.domainBox();
      cfivs[dit()] = ivsGrid;//this includes corners
   }

   for( DataIterator dit = dbl.dataIterator(); dit.ok(); ++dit )
   {
      const Box& curBox = dbl[dit()];
      const EBISBox& curEBISBox = a_layout[dit()];
      const EBGraph& curEBGraph = curEBISBox.getEBGraph();
      const IntVectSet& cfivsThisBox = cfivs[dit()];

      IntVectSet notRegular;

      notRegular |= curEBISBox.getIrregIVS  (curBox);
      notRegular |= curEBISBox.getMultiCells(curBox);

      BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[dit()];
      BaseIVFAB<Real>&       curWeightBaseIVFAB  = m_fluxWeight[dit()];

      curStencilBaseIVFAB.define(notRegular,curEBGraph,1);
      curWeightBaseIVFAB.define(notRegular,curEBGraph,1);

      for( VoFIterator vofit(notRegular,curEBGraph); vofit.ok(); ++vofit )
      {
         const VolIndex& vof = vofit();

         VoFStencil& curStencil = curStencilBaseIVFAB(vof,0);
         Real&       curWeight  = curWeightBaseIVFAB(vof,0);

         if( m_order == 1 )
            getFirstOrderStencil(curStencil,curWeight,vof,curEBISBox,a_dx);
         else if( m_order == 2 )
            getSecondOrderStencil( curStencil, curWeight, vof, curEBISBox, a_dx, cfivsThisBox );
         else
            MayDay::Error("DirichletHelmholtzEBBC::define stencil order not 1 or 2");
    }
  }
}


//-----------------------------------------------------------------------
void DirichletHelmholtzEBBC::get_ebflux(Vector<Real>&     a_flux,
                              const VolIndex&             a_vof,
                              const LevelData<EBCellFAB>& a_phi,
                              const DataIndex&            a_dit,
                              const RealVect&             a_probLo,
                              const RealVect&             a_dx,
                              bool                        a_use_homogeneous,
                              Real                        a_time )
{
  CH_assert( a_use_homogeneous || m_use_function || m_use_value );
  const EBCellFAB& curPhi     = a_phi[a_dit];
  const EBISBox    curEBISBox = curPhi.getEBISBox();

// Find function value on boundary
  Vector<Real> value(curPhi.nComp(),0.0);
  if( !a_use_homogeneous && m_use_function )
  {
     RealVect point = static_cast<RealVect>( a_vof.gridIndex() );
     for( int d=0 ; d < SpaceDim ; d++ )
         point[d] = (point[d]+0.5+curEBISBox.bndryCentroid(a_vof)[d])*a_dx[d]+a_probLo[d];
     value = m_func->get_value( point, a_time );
  }
  else if( m_use_value )
     value = m_value;

// Compute du/dn on boundary, store in 'a_flux'
  for( int c= 0 ; c < curPhi.nComp() ; c++ )
     a_flux[c] = 0.0;

  const BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[a_dit];
  const VoFStencil& curStencil = curStencilBaseIVFAB(a_vof,0);
  for( int i = 0; i < curStencil.size(); i++)
  {
     VolIndex curVoF = curStencil.vof(i);
     Real weight = curStencil.weight(i);
     if( curPhi.getRegion().contains(curVoF.gridIndex()) )
        for( int c= 0 ; c <a_phi.nComp() ; c++ )
           a_flux[c] += weight * curPhi(curVoF,c);
  }
  for( int c= 0 ; c <curPhi.nComp() ; c++ )
     a_flux[c] += m_fluxWeight[a_dit](a_vof,0)*value[c];

// Additional factor for cylindrical coordinates
  if( m_rz )
  {
     Real rb=(curEBISBox.bndryCentroid(a_vof)[0]+0.5+a_vof.gridIndex()[0])*a_dx[0]+a_probLo[0];
     Real rc=(curEBISBox.centroid(a_vof)[0]     +0.5+a_vof.gridIndex()[0])*a_dx[0]+a_probLo[0];
     //     cout << a_vof.gridIndex() << " " << curEBISBox.bndryCentroid(a_vof) << " "
     //	  << curEBISBox.centroid(a_vof) << endl;
     for( int c = 0 ; c < curPhi.nComp() ; c++ )
        a_flux[c] *= rb/rc;
  }

// Multiply length of side.
  for( int c = 0 ; c < curPhi.nComp() ; c++ )
     a_flux[c] *= curEBISBox.bndryArea(a_vof);
}


//-----------------------------------------------------------------------
void DirichletHelmholtzEBBC::getFirstOrderStencil(VoFStencil&     a_stencil,
                                                  Real&           a_weight,
                                                  const VolIndex& a_vof,
                                                  const EBISBox&  a_ebisBox,
                                                  const RealVect& a_dx)
{
  IntVect quadrant;
  const RealVect& normal = a_ebisBox.normal(a_vof);
  for(int idir=0; idir<SpaceDim; idir++)
    {
      if(normal[idir] < 0)
        {
          quadrant[idir]= -1;
        }
      else
        {
          quadrant[idir] = 1;
        }
    }
  getFirstOrderStencil(a_stencil, a_weight, quadrant, a_vof, a_ebisBox, a_dx);
  
  if(a_stencil.size()!=0)
    return;
  else
    {
      quadrant = IntVect::Unit;
      
      while(true) {
        getFirstOrderStencil(a_stencil, a_weight, quadrant, a_vof, a_ebisBox, a_dx);
        if(a_stencil.size() != 0)
          return;
        
        quadrant[0]-=2;
        for(int i=0; i<CH_SPACEDIM-1; i++){
          if(quadrant[i]==-3){quadrant[i]=1; quadrant[i+1]-=2;}
          if(quadrant[i+1]==-3) return;
        }
      } 
    }
}


//-----------------------------------------------------------------------
void DirichletHelmholtzEBBC::getFirstOrderStencil(
                                                VoFStencil&     a_stencil,
                                                Real&           a_weight,
                                                const IntVect&  a_quadrant,
                                                const VolIndex& a_vof,
                                                const EBISBox&  a_ebisBox,
                                                const RealVect& a_dx)
{
  RealVect normal   = a_ebisBox.normal(a_vof);
  RealVect centroid = a_ebisBox.bndryCentroid(a_vof);

  IntVect iv0 = a_vof.gridIndex();

#if   CH_SPACEDIM == 2
  int stenSize = 3;
#elif CH_SPACEDIM == 3
  int stenSize = 7;
#else
  tjlCompileTimeError;
#endif
  //  Box domainBox = m_domain.domainBox();
  Vector<IntVect> ivSten(stenSize);

  ivSten[0] = iv0 + a_quadrant[0]*BASISV(0)                                                    ;
  ivSten[1] = iv0                           + a_quadrant[1]*BASISV(1)                          ;
  ivSten[2] = iv0 + a_quadrant[0]*BASISV(0) + a_quadrant[1]*BASISV(1)                          ;
#if CH_SPACEDIM == 3
  ivSten[3] = iv0                                                     + a_quadrant[2]*BASISV(2);
  ivSten[4] = iv0 + a_quadrant[0]*BASISV(0)                           + a_quadrant[2]*BASISV(2);
  ivSten[5] = iv0                           + a_quadrant[1]*BASISV(1) + a_quadrant[2]*BASISV(2);
  ivSten[6] = iv0 + a_quadrant[0]*BASISV(0) + a_quadrant[1]*BASISV(1) + a_quadrant[2]*BASISV(2);
#endif

  bool dropOrder = false;

  Vector<VolIndex> volSten(stenSize);
  for (int isten = 0; isten < stenSize; isten++)
    {
      if (a_ebisBox.getDomain().contains(ivSten[isten]))
        {
          if (a_ebisBox.getVoFs(ivSten[isten]).size() > 0)
            {
              volSten[isten] = a_ebisBox.getVoFs(ivSten[isten])[0];
            }
          else
            {
              dropOrder = true;
              break;
            }
        }
      else
        {
          dropOrder = true;
          break;
        }
    }

  std::vector<bool> monotonePath(stenSize);

  for (int isten = 0; isten < stenSize; isten++)
    {
      monotonePath[isten] = EBArith::monotonePathVoFToCellVoF(volSten[isten],
                                                              a_vof,
                                                              ivSten[isten],
                                                              a_ebisBox);
      if (!monotonePath[isten])
        {
          dropOrder = true;
        }
    }

  if (!dropOrder)
    {
      RealVect x0;
      for (int idir = 0; idir < SpaceDim; idir++)
        {
          x0[idir] = a_dx[idir] * (0.5 + centroid[idir] + iv0[idir]);
        }

      Vector<RealVect> xp(stenSize);
      for (int isten = 0; isten < stenSize; isten++)
        {
          for (int idir = 0; idir < SpaceDim; idir++)
            {
              xp[isten][idir] = a_dx[idir] * (0.5 + ivSten[isten][idir]);
            }
        }

      Vector<RealVect> deltaX = xp;
      for (int isten = 0; isten < stenSize; isten++)
        {
          deltaX[isten] -= x0;
        }

      Vector<RealVect> aTransA(SpaceDim), invATransA(SpaceDim);
      for (int idir = 0; idir < SpaceDim; idir++)
        {
          for (int jdir = 0; jdir < SpaceDim; jdir++)
            {
              for (int isten = 0; isten < stenSize; isten++)
                {
                  aTransA[idir][jdir] = aTransA[idir][jdir]
                    + deltaX[isten][idir]*deltaX[isten][jdir];
                }
            }
        }

      Real det;
#if   CH_SPACEDIM == 2
      det = aTransA[0][0] * aTransA[1][1] - aTransA[0][1] * aTransA[1][0];

      invATransA[0][0] =  aTransA[1][1] / det;
      invATransA[0][1] = -aTransA[0][1] / det;
      invATransA[1][0] = -aTransA[1][0] / det;
      invATransA[1][1] =  aTransA[0][0] / det;
#elif CH_SPACEDIM == 3
      det = aTransA[0][0] * ( aTransA[1][1] * aTransA[2][2]
                              - aTransA[1][2] * aTransA[2][1])
        + aTransA[0][1] * ( aTransA[1][2] * aTransA[2][0]
                            - aTransA[1][0] * aTransA[2][2])
        + aTransA[0][2] * ( aTransA[1][0] * aTransA[2][1]
                            - aTransA[1][1] * aTransA[2][0]);

      invATransA[0][0] = ( aTransA[1][1] * aTransA[2][2]
                           - aTransA[1][2] * aTransA[2][1]) / det;
      invATransA[0][1] = ( aTransA[1][2] * aTransA[2][0]
                           - aTransA[1][0] * aTransA[2][2]) / det;
      invATransA[0][2] = ( aTransA[1][0] * aTransA[2][1]
                           - aTransA[1][1] * aTransA[2][0]) / det;
      invATransA[1][0] = ( aTransA[2][1] * aTransA[0][2]
                           - aTransA[2][2] * aTransA[0][1]) / det;
      invATransA[1][1] = ( aTransA[2][2] * aTransA[0][0]
                           - aTransA[2][0] * aTransA[0][2]) / det;
      invATransA[1][2] = ( aTransA[2][0] * aTransA[0][1]
                           - aTransA[2][1] * aTransA[0][0]) / det;
      invATransA[2][0] = ( aTransA[0][1] * aTransA[1][2]
                           - aTransA[0][2] * aTransA[1][1]) / det;
      invATransA[2][1] = ( aTransA[0][2] * aTransA[1][0]
                           - aTransA[0][0] * aTransA[1][2]) / det;
      invATransA[2][2] = ( aTransA[0][0] * aTransA[1][1]
                           - aTransA[0][1] * aTransA[1][0]) / det;
#else
      tjlCompileTimeError;
#endif

      Vector<RealVect> invATransAdeltaX(stenSize);
      for (int idir = 0; idir < SpaceDim; idir++)
        {
          for (int isten = 0; isten < stenSize; isten++)
            {
              for (int jdir = 0; jdir < SpaceDim; jdir++)
                {
                  invATransAdeltaX[isten][idir] = invATransAdeltaX[isten][idir]
                    + invATransA[idir][jdir] * deltaX[isten][jdir];
                }
            }
        }

      a_stencil.clear();
      a_weight = 0.0;

      for (int isten = 0; isten < stenSize; isten++)
        {
          Real dphidnWeight = 0.0;
          for (int idir = 0; idir < SpaceDim; idir++)
            {
              dphidnWeight -= invATransAdeltaX[isten][idir] * normal[idir];
            }

          a_stencil.add(volSten[isten],dphidnWeight);
          a_weight -= dphidnWeight;
        }
    }
  else
    {
      a_stencil.clear();
      a_weight = 0.0;
    }
}


//-----------------------------------------------------------------------
void DirichletHelmholtzEBBC::getSecondOrderStencil(VoFStencil&       a_stencil,
                                                 Real&             a_weight,
                                                 const VolIndex&   a_vof,
                                                 const EBISBox&    a_ebisBox,
                                                 const RealVect&   a_dx,
                                                 const IntVectSet& a_cfivs )
{
   bool droppedOrder = false;
   IntVect iv = a_vof.gridIndex();

  // Do Johansen-Colella
   RealVect bdCentroid = a_ebisBox.bndryCentroid(a_vof);
   RealVect normal     = a_ebisBox.normal(a_vof);

  // More complicated to prepare for anisotropic case
   RealVect vectDx = a_dx;

  // Find InterpDirection and hiLo
   Real nMax = 0.0;
   int nMaxDir;

   for( int idir = 0; idir < SpaceDim; ++idir )
   {
      if( Abs(normal[idir]) > nMax )
      {
         nMax = Abs(normal[idir]);
         nMaxDir = idir;
      }
   }

   int hiLo = 1;
   if( normal[nMaxDir] < 0.0 )
      hiLo = -1;

  // Find Intersection of planes and ray
   RealVect intersection[2];

  // We want the intersection in a coordinate system with dx = dy = dz = 1 and
  // the the cell center of a_vof is the origin

  // We know the planes are defined by nMaxDir = pm (1 or 2)
   intersection[0][nMaxDir] = 1.0*hiLo;
   intersection[1][nMaxDir] = 2.0*hiLo;

  // ScaledCentroid = centroid in an anisotropic coordinate system with
  // the the cell center of a_vof is the origin
   RealVect scaledCentroid;
   for (int idir = 0; idir < SpaceDim; ++idir)
      scaledCentroid[idir] = vectDx[idir]*bdCentroid[idir];

  // Get intersection points(relative to VectDx)
   for( int iPlane = 0; iPlane < 2; ++iPlane )
   {
      for( int idir = 0; idir < SpaceDim; ++idir )
      {
         if( idir != nMaxDir )
         {
              // We get intersection[iPlane][idir] in an anisotropic
              // coordinate system with the cell center of a_vof = origin
            intersection[iPlane][idir] = scaledCentroid[idir]
                + (normal[idir] * ( vectDx[nMaxDir]*intersection[iPlane][nMaxDir]
                                    - scaledCentroid[nMaxDir]) / normal[nMaxDir]);

              // We make intersection[iPlane][idir] in a coordinate system with
              // dx = dy = dz = 1 and the the cell center of a_vof is the origin
            intersection[iPlane][idir] /= vectDx[idir];
         }
      }
   }

  // Find interpCen
   IntVect interpCen[2];
   interpCen[0] = iv;
   interpCen[1] = iv;

  // Get ivs for cell containing intersection point
   for( int iPlane = 0; iPlane < 2; ++iPlane )
      for( int idir = 0; idir < SpaceDim; ++idir )
         interpCen[iPlane][idir] += (int)(rint(intersection[iPlane][idir]));

  // planeDir[0] and planeDir[1] are the directions != nMaxDir
  int planeDir[SpaceDim-1];
  planeDir[0] = (nMaxDir + 1) % SpaceDim;
#if CH_SPACEDIM == 3
  planeDir[1] = (planeDir[0] + 1) % SpaceDim;
#endif

  // Check do we have enough cells for a stencil
  johanDropOrder(droppedOrder,interpCen,planeDir,a_ebisBox,a_cfivs);

  if (!droppedOrder)
    {
      // We have enough cells, so interpolate

      //vofStencil for each point in iPlane
      VoFStencil iPlaneStencil[2];
      for (int iPlane = 0; iPlane < 2; ++iPlane)
        {
          VolIndex colVofs[3];

#if   CH_SPACEDIM == 3
          VoFStencil colStencil[3];

          for (int iRow = -1; iRow < 2; ++iRow)
            {
              IntVect center = interpCen[iPlane] + iRow*BASISV(planeDir[0]);
              IntVect low    = center - BASISV(planeDir[1]);
              IntVect high   = center + BASISV(planeDir[1]);

              VolIndex centerVof = a_ebisBox.getVoFs(center)[0];
              VolIndex lowVof    = a_ebisBox.getVoFs(low)[0];
              VolIndex highVof   = a_ebisBox.getVoFs(high)[0];

              colVofs[1] = centerVof;
              colVofs[0] = lowVof;
              colVofs[2] = highVof;

              // Distance (relative to dx)
              Real distance = intersection[iPlane][planeDir[1]] -
                (interpCen[iPlane][planeDir[1]] - iv[planeDir[1]]);

              // BuildStencil
              colStencil[iRow+1] = BuildEQIStencil(colVofs,distance);
            }
#elif CH_SPACEDIM == 2
          IntVect center = interpCen[iPlane];
          IntVect low    = center - BASISV(planeDir[0]);
          IntVect high   = center + BASISV(planeDir[0]);

          VolIndex centerVof = a_ebisBox.getVoFs(center)[0];
          VolIndex lowVof    = a_ebisBox.getVoFs(low)[0];
          VolIndex highVof   = a_ebisBox.getVoFs(high)[0];

          colVofs[1] = centerVof;
          colVofs[0] = lowVof;
          colVofs[2] = highVof;
#else
          tjlCompileTimeError;
#endif

          //distance (relative to dx)
          Real distance = intersection[iPlane][planeDir[0]] -
            (interpCen[iPlane][planeDir[0]] - iv[planeDir[0]]);

          // Build stencil
#if   CH_SPACEDIM == 3
          iPlaneStencil[iPlane] = BuildEQIStencil(colStencil,distance);
#elif CH_SPACEDIM == 2
          iPlaneStencil[iPlane] = BuildEQIStencil(colVofs,distance);
#else
          tjlCompileTimeError;
#endif
        }

      Real d1 = 0.0;
      Real d2 = 0.0;

      for (int idir = 0; idir < SpaceDim; ++idir)
        {
          d1 += (intersection[0][idir]*vectDx[idir] - scaledCentroid[idir])
            * (intersection[0][idir]*vectDx[idir] - scaledCentroid[idir]);

          d2 += (intersection[1][idir]*vectDx[idir]- scaledCentroid[idir])
            * (intersection[1][idir]*vectDx[idir]- scaledCentroid[idir]);
        }

      d1 = sqrt(d1);
      d2 = sqrt(d2);

      // Build stencil
      a_stencil.clear();
      a_stencil = ivofStencil(iPlaneStencil,d1,d2);

      a_weight = 1/(d2-d1) * (d2/d1 - d1/d2);
    }
  else
    {
      getFirstOrderStencil(a_stencil,a_weight,a_vof,a_ebisBox,a_dx);
    }
}


//-----------------------------------------------------------------------
void DirichletHelmholtzEBBC::johanDropOrder(bool&             a_droppedOrder,
                                          const IntVect     a_interpCen[2],
                                          const int         a_planeDir[SpaceDim-1],
                                          const EBISBox&    a_ebisBox,
                                          const IntVectSet& a_cfivs)
{
  // Return a_droppedOrder = true or false;
  a_droppedOrder = false;
  Box region = a_ebisBox.getRegion();

  Vector <VolIndex> interpVof[2];
  VolIndex dummyVof;

  // Check do we have enough cells for a stencil
  for (int iPlane = 0; iPlane < 2; ++iPlane)
    {
      if (!region.contains(a_interpCen[iPlane]) || 
          a_cfivs.contains(a_interpCen[iPlane]))
        {
          a_droppedOrder = true;
          return;
        }
    
      interpVof[iPlane] = a_ebisBox.getVoFs(a_interpCen[iPlane]);
    
      if (interpVof[iPlane].size() != 1)
        {
          a_droppedOrder = true;
          return;
        }
      else
        {
          int radius = 1;
          Vector<VolIndex> vofsStencil;
          IntVect timesMoved = IntVect::Zero;
          IntVect pathSign   = IntVect::Zero;
          EBArith::getAllVoFsInMonotonePath(vofsStencil,
                                            timesMoved,
                                            pathSign,
                                            interpVof[iPlane][0],
                                            a_ebisBox,
                                            radius);

          // Recall planeDir[0] and planeDir[1] are the directions != nMaxDir

          // In 2D check if 2 neighboring cells are in vofsStencil
          // In 3D check if all cells (!= interpCen) in a 3X3 neighborhood are in
          //                vofsStencil
          for (int iRow = -1; iRow < 2; ++iRow)
            {
              int numCols = SpaceDim - 2;
              for (int iColumn = -numCols; iColumn < numCols + 1; ++iColumn)
                {
                  IntVect cell = a_interpCen[iPlane] + iRow * BASISV(a_planeDir[0]);
#if CH_SPACEDIM == 3
                  cell += iColumn * BASISV(a_planeDir[1]);
#endif
                  if (cell != a_interpCen[iPlane])
                    {
                      if(a_cfivs.contains(cell))
                        {
                          a_droppedOrder = true;
                          return;
                        }
                      bool isVof = EBArith::isVoFHere(dummyVof,vofsStencil,cell);
                      if (!isVof)
                        {
                          a_droppedOrder = true;
                          return;
                        }
                    }
                }
            }
        }
    }

  return;
}


//-----------------------------------------------------------------------
VoFStencil DirichletHelmholtzEBBC::BuildEQIStencil(const VolIndex a_vofs[3],
                                          const Real&    a_distance)
{
  VoFStencil retval;

  Real x  = a_distance;
  Real x2 = x * x;

  retval.add(a_vofs[0], (x2 - x) / 2.0);
  retval.add(a_vofs[1],-(x2 - 1)      );
  retval.add(a_vofs[2], (x2 + x) / 2.0);

  return retval;
}


//-----------------------------------------------------------------------
VoFStencil DirichletHelmholtzEBBC::BuildEQIStencil(VoFStencil a_vofs[3],
                                          const Real& a_distance)
{
  VoFStencil retval;

  Real x  = a_distance;
  Real x2 = x * x;

  a_vofs[0] *=  (x2 - x) / 2.0;
  retval += a_vofs[0];

  a_vofs[1] *= -(x2 - 1);
  retval += a_vofs[1];

  a_vofs[2] *=  (x2 + x) / 2.0;
  retval += a_vofs[2];

  return retval;
}


//-----------------------------------------------------------------------
VoFStencil DirichletHelmholtzEBBC::ivofStencil(const VoFStencil a_iPlaneStencil[2],
                                      const Real&      a_distance1,
                                      const Real&      a_distance2)
{
  VoFStencil retvalTemp;
  VoFStencil retval;

  Real d1 = a_distance1;
  Real d2 = a_distance2;

  retvalTemp = a_iPlaneStencil[0];
  retvalTemp *= 1/(d2-d1) * (-d2/d1);
  retval += retvalTemp;

  retvalTemp = a_iPlaneStencil[1];
  retvalTemp *= 1/(d2-d1) * (d1/d2);
  retval += retvalTemp;

  return retval;
}

//-----------------------------------------------------------------------
//                  Factory class
//-----------------------------------------------------------------------
DirichletHelmholtzEBBCFactory::DirichletHelmholtzEBBCFactory()
  : HelmholtzEBBCFactory()
{}


//-----------------------------------------------------------------------
DirichletHelmholtzEBBCFactory::~DirichletHelmholtzEBBCFactory()
{}


//-----------------------------------------------------------------------
HelmholtzEBBC* DirichletHelmholtzEBBCFactory::create(
						     const ProblemDomain& a_domain,
						     const EBISLayout& a_layout,
						     const RealVect&   a_dx )
{
   DirichletHelmholtzEBBC* new_bc = new DirichletHelmholtzEBBC( a_domain, a_layout, a_dx );
   new_bc->set_order( m_order );
   new_bc->set_rz( m_rz );
   if( m_use_function )
      new_bc->set_function( m_func );
   else if( m_use_value )
      new_bc->set_value( m_value );
   return new_bc;
}
