/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _DIRICHLETNSVISCDOMAINBC_H_
#define _DIRICHLETNSVISCDOMAINBC_H_

#include "RefCountedPtr.H"
#include "BaseBCVectValue.H"
#include "RealVect.H"
#include "FArrayBox.H"

#include "VolIndex.H"
#include "EBCellFAB.H"
#include "EBFaceFAB.H"
#include "EBISLayout.H"

/// Impose boundary data on the domain boundary for the velocities in the Navier-Stokes operator.
/**
    The normal flux of the NS operator is computed from given Dirichlet data
    for the velocities. Planar boundary is assumed. The flux is given by (Cartesian case)
    \f[
       ((\lambda+2\mu)d(u_n)/dn+\lambda d(u_{t1})/dt1 + \lambda(u_{t2})/dt2)*e_n +
             \mu(d(u_n)/dt1+d(u_{t1})/dn)*e_{t1} + \mu(d(u_n)/dt2+d(u_{t2})/dn)*e_{t2}
    \f]
    where \f$ (\lambda, \mu)\f$ are viscosity coefficients, \f$ e_n\f$ is the unit
    vector normal to the boundary, \f$ e_{t1}\f$ and \f$ e_{t2}\f$ are unit vectors
    tangential to the boundary. 
    \f$ (e_n,e_{t1},e_{t2})\f$ are orthonormal vectors. \f$ (u_n,u_{t1},u_{t2})\f$ 
    are the velocity components in the basis \f$ (e_n,e_{t1},e_{t2})\f$.
    The input and output parameters of the methods are all assumed to be in
    Cartesian (x,y,z) or cylindrical (r,z) coordinates.

    When created, the object is in a state where only homogeneous
    (i.e. zero velocity) data can be imposed. The 'setFunction' method can be used
    to specify any given function for boundary data. The boundary function must
    be derived from the BaseBCVectValue class. The 'setValue' method is used to specify
    a constant possibly non-zero boundary value. The most recent call to 'setFunction'
    or 'setValue' overrides previous calls to these two functions.
    The formal order of accuracy of the boundary flux is by default 2, but can be
    changed by calling the 'set_order' method.

    The number of components in the fluxes should be 2 in two space dimensions, and
    3 in three space dimensions. The cylindrical case is assumed to be two dimensional.
 */
class DirichletNSviscDomainBC
{
public:
  ///  Constructor.
  /**
   */
  DirichletNSviscDomainBC();

  /// Destructor.
  /**
   */
  virtual ~DirichletNSviscDomainBC();

  /// Define the boundary value as a constant.
  /**
      Input a vector that specifies the values on the boundary.
   */
  virtual void setValue( Vector<Real> a_value );

  /// Define the boundary value as a function.
  /**
      Input a vector valued function that specifies the values on the boundary.
   */
  virtual void setFunction(RefCountedPtr<BaseBCVectValue> a_func);

  /// Set the order of accuracy of boundary flux approximation.
  /**
      Order 2 means the formula \f$(9*u_1-u_2-8*u_{1/2})/(3*h)\f$ is used to approximate
      normal derivatives on the boundary at \f$ x_{1/2}\f$. 
      Order 1 means that \f$ 2*(u_1-u_{1/2})/h \f$ is used. This translates into order 1
      and 0 respectively for the second derivative approximation. Upper sides are 
      approximated by similar formulas.
   */ 
  virtual void set_order( int a_ebOrder );

  /// Boundary fluxes at one side of a domain.
  /**
      - Input:
         - a_phi    - The function over the grid.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
         - a_idir   - Direction of flux (0 is I-direction, 1 is J-direction, 2 is K)
	 - a_side   - Upper or lower side.
         - a_useHomogeneous - If true, override the set boundary value and 
                                impose zero data.
	 - mu, la   - Viscosity coefficients, variable on the grid.

      - Output: a_faceFlux - The fluxes along the side specified by (a_idir,a_side).
   */ 
  virtual void getFaceFlux(BaseFab<Real>&       a_faceFlux,
                           const BaseFab<Real>& a_phi,
                           const RealVect&  a_probLo,
                           const RealVect&  a_dx,
                           int              a_idir,
                           Side::LoHiSide   a_side,
                           bool             a_useHomogeneous,
                           BaseFab<Real>& a_mu, BaseFab<Real>& a_la,
                           const Real a_time );

  /// Boundary fluxes at one cell.
  /**
      - Input:
         - a_vof    - The cell where the boundary flux is to be computed.
         - a_phi    - The function over the grid.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
         - a_idir   - Direction of flux (0 is I-direction, 1 is J-direction, 2 is K)
	 - a_side   - Upper or lower side.
         - a_useHomogeneous - If true, override the set boundary value and 
                                impose zero data.
	 - mu, la   - Viscosity coefficients, variable on the grid.

      - Output: a_faceFlux - The flux along the side in cell a_vof specified by (a_idir,a_side).
   */
  virtual void getFaceFlux(Vector<Real>&    a_faceFlux,
                           const VolIndex&  a_vof,
                           const EBCellFAB& a_phi,
                           const RealVect&  a_probLo,
                           const RealVect&  a_dx,
                           int              a_idir,
                           Side::LoHiSide   a_side,
                           bool             a_useHomogeneous,
                           Real mu, Real la, const Real a_time );


  /// Boundary fluxes at one cell, cylindrical coordinates.
  /**
      - Input:
         - a_vof    - The cell where the boundary flux is to be computed.
         - a_phi    - The function over the grid.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
         - a_idir   - Direction of flux (0 is I-direction, 1 is J-direction, 2 is K)
	 - a_side   - Upper or lower side.
         - a_useHomogeneous - If true, override the set boundary value and 
                                impose zero data.
	 - mu, la   - Viscosity coefficients, variable on the grid.

      - Output: a_faceFlux - The flux along the side in cell a_vof specified by (a_idir,a_side).
   */
  virtual void getFaceFlux_rz( Vector<Real>&    a_flux,
                           const VolIndex&  a_vof,
                           const EBCellFAB& a_phi,
                           const RealVect&  a_probLo,
                           const RealVect&  a_dx,
                           int              a_idir,
                           Side::LoHiSide   a_side,
                           bool             a_useHomogeneous,
                           Real mu, Real la, const Real a_time );

  /// Boundary fluxes at one cell, energy equation term.
  /**
      The contribution in the energy equation from the stress tensor is the
      scalar product of the momenum eq. flux with the wall velocity, so it is zero at
      a solid wall.
      - Input:
         - a_vof    - The cell where the boundary flux is to be computed.
         - a_phi    - The function over the grid.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
         - a_idir   - Direction of flux (0 is I-direction, 1 is J-direction, 2 is K)
	 - a_side   - Upper or lower side.
         - a_useHomogeneous - If true, override the set boundary value and 
                                impose zero data.
	 - mu, la   - Viscosity coefficients, variable on the grid.

      - Output: a_faceFlux - The flux along the side in cell a_vof specified by (a_idir,a_side).
   */
  void getFaceFlux_eeq(  Real&    a_faceFlux,
                           const VolIndex&  a_vof,
                           const EBCellFAB& a_phi,
                           const RealVect&  a_probLo,
                           const RealVect&  a_dx,
                           int              a_idir,
                           Side::LoHiSide   a_side,
                           bool             a_useHomogeneous,
                           Real mu, Real la, const Real a_time );

///
  void getFaceFlux_eeq_rz(  Real&    a_faceFlux,
                           const VolIndex&  a_vof,
                           const EBCellFAB& a_phi,
                           const RealVect&  a_probLo,
                           const RealVect&  a_dx,
                           int              a_idir,
                           Side::LoHiSide   a_side,
                           bool             a_useHomogeneous,
                           Real mu, Real la, const Real a_time );


  /// Contribution from boundary to the diagonal coefficient in the discretization
  /**
    When using a relaxation scheme, such as Jacobi or Gauss-Seidel, we need to
    divide by the diagonal of the discrete equation. In one space dimension for 
    example the diagonal element of \f$ (u_{i+1}-2u_i+u_{i-1})/h^2 = f_i \f$ is
    \f$ -2/h^2\f$, but at a boundary with flux \f$ 2*(u_1-u_{1/2})/h\f$, we get  
    the approximation \f$ (u_{2}-3u_1+2*u_{1/2})/h^2 = f_1\f$, i.e., the diagonal
    weight is \f$ -3/h^2\f$.
    This function computes the boundary contribution, in the example above it would
    be \f$ -2/h^2 \f$.
    The two (or three) equations are treated as independent scalar equations.
      - Input:
         - a_vof    - The cell where the diagonal weight is to be computed.
         - a_ebisbox- The EBISbox of the present grid.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
         - a_idir   - Direction of flux (0 is I-direction, 1 is J-direction, 2 is K)
	 - a_side   - Upper or lower side.
	 - mu, la   - Viscosity coefficients, variable on the grid.

      - Output: a_wghs - The contribution to the diagonal by the flux on the side 
        specified by (a_idir,a_side) in cell a_vof.
   */
  void get_uweight( Vector<Real>& a_wghs,
                                           const VolIndex&  a_vof,
                                           const EBISBox& a_ebisbox,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
		     Real mu, Real la );

  /// Contribution from boundary to the diagonal coefficient in the discretization
  /**
    This is the cylindrical version of 'get_uweight', see 'get_uweight' for documentation.   */
  void get_uweight_rz( Vector<Real>& a_wghs,
                                           const VolIndex&  a_vof,
                                           const EBISBox& a_ebisbox,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
		     Real mu, Real la );

protected:

  /// Compute the NS boundary flux from given normal derivatives.
  /**
   */
  void get_the_flux( int a_idir, Vector<Real> dns, RealVect point, 
		     Real mu, Real la, Vector<Real>& flx, bool a_useHomogeneous,
		     const Real a_time );

  /// Compute the NS boundary flux from given normal derivatives.
  /**
   */
  void get_the_flux_rz( int a_idir, Vector<Real> dns, Real drn, RealVect point, 
                      Real mu, Real la, Vector<Real>& flux_div,
		      Vector<Real>& flux_undiv, Real rb, Real uval,
		      bool rzero_bndry, bool a_useHomogeneous, Real vtnum,
                      const Real a_time );

  void tangential_derivative( IntVect of,
			      const VolIndex& vof,
			      const EBISBox& eb,
			      VoFStencil& vs,
			      const Real di,
			      const Box ubox );

  bool m_onlyHomogeneous;
  bool m_isFunctional;
  int  m_ncomp;

  Vector<Real> m_value;
  RefCountedPtr<BaseBCVectValue> m_func;

  int m_ebOrder;
};

/// Class that can generate a DirichletNSviscDomainBC object.
/**
 */
class DirichletNSviscDomainBCFactory 
{
public:
  /// Constructor.
  /**
   */
  DirichletNSviscDomainBCFactory();

  /// Destructor.
  /**
   */
  virtual ~DirichletNSviscDomainBCFactory();

  ///  Create a boundary flux object.
  /**
   */
  virtual DirichletNSviscDomainBC* create(const ProblemDomain& a_domain,
                                    const EBISLayout&    a_layout,
                                    const RealVect&      a_dx);

  ///  Define the boundary value as a constant.
  /**
   */
  virtual void setValue(Vector<Real> a_value);

  ///  Define the boundary value as a function.
  /**
   */
  virtual void setFunction( RefCountedPtr<BaseBCVectValue> a_func );

  /// Set the order of accuracy of boundary flux approximation.
  /**
   */
  virtual void set_order(int a_order);

protected:
  bool m_onlyHomogeneous;
  bool m_isFunctional;

  int m_ncomp;
  Vector<Real> m_value;
  RefCountedPtr<BaseBCVectValue> m_func;

  int m_order;
};
#endif
