/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "BoxIterator.H"
#include "DirichletNSviscDomainBC.H"
#include "EBArith.H"


//-----------------------------------------------------------------------
DirichletNSviscDomainBC::DirichletNSviscDomainBC()
{
  m_ncomp = SpaceDim;

  m_value.resize(m_ncomp);

  for( int c=0 ; c<m_ncomp ; c++ )
     m_value[c] = 12345.6789;
  
  m_func = NULL;

  m_onlyHomogeneous = true;
  m_isFunctional    = false;

  m_ebOrder = 2;
}


//-----------------------------------------------------------------------
DirichletNSviscDomainBC::~DirichletNSviscDomainBC()
{
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::setValue(Vector<Real> a_value)
{
  CH_assert( a_value.size() == m_value.size() );

  m_value = a_value;
  m_func  = NULL;
  m_onlyHomogeneous = false;
  m_isFunctional    = false;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::setFunction(RefCountedPtr<BaseBCVectValue> a_func)
{
  for( int c=0 ; c<m_ncomp ; c++ )
     m_value[c] = 12345.6789;

  m_func = a_func;
  m_onlyHomogeneous = false;
  m_isFunctional    = true;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::set_order( int a_ebOrder )
{
  CH_assert(a_ebOrder >= 1);
  CH_assert(a_ebOrder <= 2);

  m_ebOrder = a_ebOrder;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::getFaceFlux( BaseFab<Real>&       a_faceFlux,
                                           const BaseFab<Real>& a_phi,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
                                           bool             a_useHomogeneous,
                                           BaseFab<Real>& mu, BaseFab<Real>& la,
                                           const Real a_time )
{
  CH_assert(a_phi.nComp()      == m_ncomp );
  CH_assert(a_faceFlux.nComp() == m_ncomp );
  if( !a_useHomogeneous && !m_isFunctional && m_onlyHomogeneous )
          MayDay::Error("DirichletNSviscDomainBC::getFaceFlux called with undefined inhomogeneous BC");


  const Box& box = a_faceFlux.box();
  int iside;
  IntVect ofs=IntVect::Zero;

  if (a_side == Side::Lo)
  {
     iside = 1;
     ofs[a_idir]=1;
  }
  else
  {
     iside = -1;
     ofs[a_idir]=-1;
  }
  Real itdx = 1.0/(3.0*a_dx[a_idir]);
  Real ihdx = 2.0/(a_dx[a_idir]);

  Vector<Real> value(m_ncomp);
  if( a_useHomogeneous )
     for( int d =0 ; d < m_ncomp ; d++ )
	value[d] =  0;
  else
     for( int d =0 ; d < m_ncomp ; d++ )
	value[d] =  m_value[d];

  BoxIterator bit(box);
  for( bit.begin(); bit.ok(); ++bit )
  {
     const IntVect& iv = bit();
     RealVect point = RealVect::Zero;
     if( m_isFunctional && !a_useHomogeneous )
     {
        for( int d=0 ; d < SpaceDim ; d++ )
	   point[d] = (iv[d]-0.5*(ofs[d]-1))*a_dx[d]+a_probLo[d];
	//        point = static_cast<RealVect>(iv)-0.5*(ofs-1);
	//        point = point*a_dx + a_probLo;
        value = m_func->get_value( point, a_time );
     }
     Vector<Real> dns(m_ncomp);


     for( int comp = 0 ; comp < m_ncomp ; comp++ )
       if( m_ebOrder == 2 )
           dns[comp] = iside * itdx * ( 9*a_phi(iv,comp)-a_phi(iv+ofs,comp)-
                                                 8*value[comp]);
       else
           dns[comp] = iside * ihdx * ( a_phi(iv,comp)- value[comp]);
     Vector<Real> flx;           
     get_the_flux( a_idir, dns, point, mu(iv,0), la(iv,0), flx, a_useHomogeneous, a_time );
     for( int comp = 0 ; comp < m_ncomp ; comp++ )
        a_faceFlux(iv,comp) = flx[comp];
  }
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::getFaceFlux( Vector<Real>&    a_faceFlux,
                                           const VolIndex&  a_vof,
                                           const EBCellFAB& a_phi,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
                                           bool             a_useHomogeneous,
                                           Real mu, Real la, const Real a_time )
{
   CH_assert(a_phi.nComp() == m_ncomp );
   CH_assert( a_faceFlux.size() == m_ncomp );
   if( !a_useHomogeneous && !m_isFunctional && m_onlyHomogeneous )
          MayDay::Error("DirichletNSviscDomainBC::getFaceFlux called with undefined inhomogeneous BC");

   IntVect ofs=IntVect::Zero;
   int iside;
   if (a_side == Side::Lo)
   {
     iside = 1;
     ofs[a_idir]=1;
   }
   else
   {
     iside = -1;
     ofs[a_idir]=-1;
   }

   Real itdx = 1.0/(3.0*a_dx[a_idir]);
   Real ihdx = 2.0/(a_dx[a_idir]);
   const EBISBox& ebisBox = a_phi.getEBISBox();
   const IntVect& iv = a_vof.gridIndex();
   Vector<Real> value(m_ncomp);
   RealVect point = RealVect::Zero;
   if( a_useHomogeneous )
     for( int d =0 ; d < m_ncomp ; d++ )
	value[d] =  0;
   else if( !m_isFunctional )
     for( int d =0 ; d < m_ncomp ; d++ )
	value[d] =  m_value[d];
   else //   if( m_isFunctional )
   {    
      for( int d=0 ; d < SpaceDim ; d++ )
	 point[d] = (iv[d]-0.5*(ofs[d]-1))*a_dx[d]+a_probLo[d];
	//      point = (static_cast<RealVect>(iv)-0.5*(ofs-1))*a_dx+a_probLo;
     //     point = point*a_dx + a_probLo;
      value = m_func->get_value( point, a_time );
   }
   Vector<VolIndex> voflist = ebisBox.getVoFs(iv+ofs);
   VolIndex voff;
   int order = m_ebOrder;

   if( order == 2 )
   {
      if( voflist.size() == 0 )
         order = 1;
      else if( voflist.size() > 1 )
      {
         bool found=false;
         for( int v=0 ; v < voflist.size() ; v++ )
	    if( ebisBox.isConnected( a_vof, voflist[v] ) )
	    {
               found = true;
	       voff = voflist[v];
	    }
         if( !found )
	    order = 1;
      }
      else
         voff = voflist[0];
   }

   Vector<Real> dns(m_ncomp);
   for( int comp=0 ; comp < m_ncomp ; comp++ )
     if( order == 2 )
        dns[comp] = iside*itdx*( 9*a_phi(a_vof,comp)-a_phi(voff,comp)-
                                                 8*value[comp]);
     else
        dns[comp] = iside*ihdx*( a_phi(a_vof,comp)-value[comp]);
       


   get_the_flux( a_idir, dns, point, mu, la, a_faceFlux, a_useHomogeneous, a_time );
   
   Real faceAreaFrac = 0.0;
   Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
   for( int i = 0; i < faces.size(); i++ )
   {
     faceAreaFrac += ebisBox.areaFrac( faces[i] );
   }
   for( int d= 0 ; d<SpaceDim ;d++ )
      a_faceFlux[d] *= faceAreaFrac;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::getFaceFlux_rz( Vector<Real>& a_Faceflux,
                                           const VolIndex&  a_vof,
                                           const EBCellFAB& a_phi,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
                                           bool             a_useHomogeneous,
                                           Real mu, Real la, const Real a_time )
{
   CH_assert(a_phi.nComp() == m_ncomp );
   CH_assert( a_Faceflux.size() == m_ncomp );
   if( !a_useHomogeneous && !m_isFunctional && m_onlyHomogeneous )
          MayDay::Error("DirichletNSviscDomainBC::getFaceFlux_rz called with undefined inhomogeneous BC");

   IntVect ofs=IntVect::Zero;
   int iside;
   if( a_side == Side::Lo )
   {
     iside = 1;
     ofs[a_idir]=1;
   }
   else
   {
     iside = -1;
     ofs[a_idir]=-1;
   }

   Real itdx = 1.0/(3.0*a_dx[a_idir]);
   Real ihdx = 2.0/(a_dx[a_idir]);
   const EBISBox& ebisBox = a_phi.getEBISBox();
   const IntVect& iv = a_vof.gridIndex();

   Vector<Real> value(m_ncomp);
   RealVect point = RealVect::Zero;
   if( a_useHomogeneous )
      for( int d =0 ; d < m_ncomp ; d++ )
	 value[d] =  0;
   else if( !m_isFunctional )
      for( int d =0 ; d < m_ncomp ; d++ )
	 value[d] =  m_value[d];
   else //if( m_isFunctional )
   {    
      for( int d=0 ; d < SpaceDim ; d++ )
	 point[d] = (iv[d]-0.5*(ofs[d]-1))*a_dx[d]+a_probLo[d];
      value = m_func->get_value( point, a_time );
   }
   Vector<VolIndex> voflist = ebisBox.getVoFs(iv+ofs);
   VolIndex voff;
   int order = m_ebOrder;

   if( order == 2 )
   {
      if( voflist.size() == 0 )
         order = 1;
      else if( voflist.size() > 1 )
      {
         bool found=false;
         for( int v=0 ; v < voflist.size() ; v++ )
	    if( ebisBox.isConnected( a_vof, voflist[v] ) )
	    {
               found = true;
	       voff = voflist[v];
	    }
         if( !found )
	    order = 1;
      }
      else
         voff = voflist[0];
   }

   Vector<Real> dns(m_ncomp);
   for( int comp=0 ; comp < m_ncomp ; comp++ )
     if( order == 2 )
        dns[comp] = iside*itdx*( 9*a_phi(a_vof,comp)-a_phi(voff,comp)-8*value[comp]);
     else
        dns[comp] = iside*ihdx*( a_phi(a_vof,comp)-value[comp]);
   

   Real r  = (a_vof.gridIndex()[0]+ebisBox.centroid(a_vof)[0]+0.5)*a_dx[0]+a_probLo[0];
   Real rp;
   if( order == 2 )
      rp = (voff.gridIndex()[0]+ebisBox.centroid(voff)[0]+0.5)*a_dx[0]+a_probLo[0];

   bool rzero_bndry =  (a_vof.gridIndex()[0] == 0 && a_idir == 0 );
   //   bool rzero_bndry = false;

   Real rb;
   if( a_idir == 0 )
   {
      rb = ( a_vof.gridIndex()[0] + 0.5*(1-iside) )*a_dx[0] + a_probLo[0];
      //      rzero_bndry = rzero_bndry && (fabs(rb)<1e-12);
   }
   else
   {
      rb = (a_vof.gridIndex()[0] + 0.5 +
	    ebisBox.centroid(ebisBox.getFaces(a_vof,a_idir,a_side)[0])[0])*a_dx[0]+a_probLo[0];
   }

   Real drn;
   if( order == 2 )
      drn = iside*itdx*(9*a_phi(a_vof,0)*r-a_phi(voff,0)*rp-8*value[0]*rb);
   else
      drn = iside*ihdx*(  a_phi(a_vof,0)*r-value[0]*rb );


   Vector<Real> flux_div(2), flux_undiv(2);

   Real vt1 = 0;
   if( rzero_bndry )
   {
      VoFStencil vts;
      IntVect oftan = IntVect::Zero;
      oftan[1] = 1;
      tangential_derivative( oftan, a_vof, a_phi.getEBISBox(), vts,
			    1/a_dx[1], a_phi.getRegion() );
      for( int i=0; i<vts.size(); i++ )
         vt1 += a_phi(vts.vof(i),1)*vts.weight(i);
      if( order == 2 )
      {
          tangential_derivative( oftan, voff, a_phi.getEBISBox(), vts,
			    1/a_dx[1], a_phi.getRegion() );
          Real vt2 = 0;
          for( int i=0; i<vts.size(); i++ )
              vt2 += a_phi(vts.vof(i),1)*vts.weight(i);
          vt1 = 1.5*vt2 - 0.5*vt1;
      }
   }     
   get_the_flux_rz( a_idir, dns, drn, point, mu, la, flux_div, flux_undiv,
		    rb, value[0], rzero_bndry, a_useHomogeneous, vt1, a_time );

   Real rc = static_cast<Real>(a_vof.gridIndex()[0]) +
                                    ebisBox.centroid(a_vof)[0];

   rc = (rc+0.5)*a_dx[0]+a_probLo[0];
   a_Faceflux[0] = flux_div[0]/rc + flux_undiv[0];
   a_Faceflux[1] = flux_div[1]/rc + flux_undiv[1];
   //   if( rzero_bndry )
     //     cout << a_Faceflux[0] << " " << flux_undiv[0] << " " << point[1] << endl;

   Real faceAreaFrac = 0.0;
   Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
   for( int i = 0; i < faces.size(); i++ )
     faceAreaFrac += ebisBox.areaFrac( faces[i] );

   for( int d= 0 ; d<SpaceDim ; d++ )
      a_Faceflux[d]   *= faceAreaFrac;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::get_uweight_rz( Vector<Real>& a_wghs,
                                           const VolIndex&  a_vof,
				           const EBISBox&   ebisBox,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
                                           Real mu, Real la )
{
   CH_assert( a_wghs.size() == m_ncomp );

   IntVect ofs=IntVect::Zero;
   int iside;
   if (a_side == Side::Lo)
   {
     iside = 1;
     ofs[a_idir]=1;
   }
   else
   {
     iside = -1;
     ofs[a_idir]=-1;
   }

   Real itdx = 1.0/(3.0*a_dx[a_idir]);
   Real ihdx = 2.0/(a_dx[a_idir]);
   const IntVect& iv = a_vof.gridIndex();

   Vector<VolIndex> voflist = ebisBox.getVoFs(iv+ofs);
   int order = m_ebOrder;

   if( order == 2 )
   {
      if( voflist.size() == 0 )
         order = 1;
      else if( voflist.size() > 1 )
      {
         bool found=false;
         for( int v=0 ; v < voflist.size() ; v++ )
	    if( ebisBox.isConnected( a_vof, voflist[v] ) )
	    {
               found = true;
	    }
         if( !found )
	    order = 1;
      }
   }

   Real r = static_cast<Real>(a_vof.gridIndex()[0]);
   r  = (r+0.5)*a_dx[0]+a_probLo[0];

   bool rzero_bndry =  (a_vof.gridIndex()[0] == 0 && a_idir == 0 );
   //   bool rzero_bndry = false;

   Real rb, rp;
   if( a_idir == 0 )
   {
      rb = r-iside*a_dx[0]*0.5;
      rp = r+iside*a_dx[0];
      rzero_bndry = rzero_bndry && (fabs(rb)<1e-10);
   }
   else
   {
      rb = r;
      rp = r;
   }
   Real drn;

   if( order == 2 )
      drn = iside*itdx*(9*r);
   else
      drn = iside*ihdx*r;

   Vector<Real> dns(m_ncomp);
   for( int comp=0 ; comp < m_ncomp ; comp++ )
      if( order == 2 )
         dns[comp] = iside*itdx*9;
      else
         dns[comp] = iside*ihdx;

   Vector<Real> flux_div(2), flux_undiv(2);

   RealVect nv=RealVect::Zero;
   nv[a_idir] = 1;

   flux_div[0] = 2*mu*rb*nv[0]*(nv[0]*dns[0]);
   flux_div[1] =   mu*rb*(nv[0]*nv[1]*dns[0]+nv[0]*nv[0]*dns[1]);

   if( !rzero_bndry )
   {
      flux_undiv[0] = la/rb*nv[0]*(nv[0]*drn) + la*nv[0]*(nv[1]*dns[1]) + 
        mu*nv[1]*( nv[1]*dns[0] + nv[0]*dns[1] );
      flux_undiv[1] = (2*mu+la)*nv[1]*( nv[1]*dns[1] ) +
        la/rb*nv[1]*(nv[0]*drn);
   }
   else
   {
      flux_undiv[0] =  
        la*nv[0]*( nv[1]*dns[1] ) + 
        mu*nv[1]*( nv[1]*dns[0]+nv[0]*dns[1] );
      flux_undiv[1] = (2*mu+la)*nv[1]*( nv[1]*dns[1] );
   }
   Real rc = static_cast<Real>(a_vof.gridIndex()[0]) + ebisBox.centroid(a_vof)[0];
   rc = (rc+0.5)*a_dx[0]+a_probLo[0];
   a_wghs[0] = flux_div[0]/rc + flux_undiv[0];
   a_wghs[1] = flux_div[1]/rc + flux_undiv[1];

   Real faceAreaFrac = 0.0;
   Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
   for( int i = 0; i < faces.size(); i++ )
     faceAreaFrac += ebisBox.areaFrac( faces[i] );

   for( int d= 0 ; d<SpaceDim ; d++ )
      a_wghs[d]   *= faceAreaFrac;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::get_uweight( Vector<Real>&   a_wghs,
                                          const VolIndex&  a_vof,
				          const EBISBox&   ebisBox,
                                          const RealVect&  a_probLo,
                                          const RealVect&  a_dx,
                                          int              a_idir,
                                          Side::LoHiSide   a_side,
                                          Real mu, Real la )
{
   CH_assert( a_wghs.size() == m_ncomp );

   IntVect ofs=IntVect::Zero;
   int iside;
   if (a_side == Side::Lo)
   {
     iside = 1;
     ofs[a_idir]=1;
   }
   else
   {
     iside = -1;
     ofs[a_idir]=-1;
   }

   Real itdx = 1.0/(3.0*a_dx[a_idir]);
   Real ihdx = 2.0/(a_dx[a_idir]);
   const IntVect& iv = a_vof.gridIndex();

   Vector<VolIndex> voflist = ebisBox.getVoFs(iv+ofs);
   int order = m_ebOrder;

   if( order == 2 )
   {
      if( voflist.size() == 0 )
         order = 1;
      else if( voflist.size() > 1 )
      {
         bool found=false;
         for( int v=0 ; v < voflist.size() ; v++ )
	    if( ebisBox.isConnected( a_vof, voflist[v] ) )
	    {
               found = true;
	    }
         if( !found )
	    order = 1;
      }
   }


   Vector<Real> dns(m_ncomp);
   for( int comp=0 ; comp < m_ncomp ; comp++ )
     if( order == 2 )
        dns[comp] = iside*itdx*9;
     else
        dns[comp] = iside*ihdx;


   RealVect nv = RealVect::Zero;
   nv[a_idir] = 1;

   //   Real dundn = 0;
   //   for( int c=0; c<SpaceDim ;c++)
   //      dundn += dns[c]*nv[c];

   for( int c=0; c<SpaceDim ;c++)
     a_wghs[c] = (mu+la)*dns[a_idir]*nv[c]+mu*dns[c];

   Real faceAreaFrac = 0.0;
   Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
   for( int i = 0; i < faces.size(); i++ )
     faceAreaFrac += ebisBox.areaFrac( faces[i] );

   for( int d= 0 ; d<SpaceDim ;d++ )
      a_wghs[d] *= faceAreaFrac;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::getFaceFlux_eeq( Real&    a_faceFlux,
                                           const VolIndex&  a_vof,
                                           const EBCellFAB& a_phi,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
                                           bool             a_useHomogeneous,
                                           Real mu, Real la, const Real a_time )
{
   CH_assert(a_phi.nComp() == m_ncomp );
   a_faceFlux = 0;
   if( m_isFunctional && !a_useHomogeneous )
   {
      IntVect ofs=IntVect::Zero;

      if( a_side == Side::Lo )
         ofs[a_idir] = 1;
      else
         ofs[a_idir] = -1; 

      Vector<Real> flxx(m_ncomp,0);
      getFaceFlux(flxx,a_vof,a_phi,a_probLo,a_dx,a_idir,a_side,a_useHomogeneous,
		  mu,la,a_time);
      Vector<Real> value(m_ncomp,0);
      RealVect point;
      for( int d=0 ; d < SpaceDim ; d++ )
	 point[d] = (a_vof.gridIndex()[d]-0.5*(ofs[d]-1))*a_dx[d]+a_probLo[d];
      value=m_func->get_value( point, a_time );
      for( int c=0 ; c<m_ncomp ; c++ )
	 a_faceFlux += flxx[c]*value[c];
   }
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::getFaceFlux_eeq_rz( Real&    a_faceFlux,
                                           const VolIndex&  a_vof,
                                           const EBCellFAB& a_phi,
                                           const RealVect&  a_probLo,
                                           const RealVect&  a_dx,
                                           int              a_idir,
                                           Side::LoHiSide   a_side,
                                           bool             a_useHomogeneous,
                                           Real mu, Real la, const Real a_time )
{
   CH_assert(a_phi.nComp() == m_ncomp );
   a_faceFlux = 0;

   if( !a_useHomogeneous && !m_isFunctional && m_onlyHomogeneous )
      MayDay::Error("DirichletNSviscDomainBC::getFaceFlux_eeq_rz called with undefined inhomogeneous BC");

   IntVect ofs=IntVect::Zero;
   int iside;
   if (a_side == Side::Lo)
   {
     iside = 1;
     ofs[a_idir]=1;
   }
   else
   {
     iside = -1;
     ofs[a_idir]=-1;
   }
   Real itdx = 1.0/(3.0*a_dx[a_idir]);
   Real ihdx = 2.0/(a_dx[a_idir]);
   const EBISBox& ebisBox = a_phi.getEBISBox();
   const IntVect& iv = a_vof.gridIndex();

   Vector<Real> value(m_ncomp);
   RealVect point = RealVect::Zero;

   if( a_useHomogeneous )
     for( int d =0 ; d < m_ncomp ; d++ )
	value[d] =  0;
   else if( !m_isFunctional )
     for( int d =0 ; d < m_ncomp ; d++ )
	value[d] =  m_value[d];
   else 
   {    
      for( int d=0 ; d < SpaceDim ; d++ )
	 point[d] = (iv[d]-0.5*(ofs[d]-1))*a_dx[d]+a_probLo[d];
      value = m_func->get_value( point, a_time );
   }
   Vector<VolIndex> voflist = ebisBox.getVoFs(iv+ofs);
   VolIndex voff;
   int order = m_ebOrder;
   if( order == 2 )
   {
      if( voflist.size() == 0 )
         order = 1;
      else if( voflist.size() > 1 )
      {
         bool found=false;
         for( int v=0 ; v < voflist.size() ; v++ )
	    if( ebisBox.isConnected( a_vof, voflist[v] ) )
	    {
               found = true;
	       voff = voflist[v];
	    }
         if( !found )
	    order = 1;
      }
      else
         voff = voflist[0];
   }

   Vector<Real> dns(m_ncomp);
   for( int comp=0 ; comp < m_ncomp ; comp++ )
     if( order == 2 )
        dns[comp] = iside*itdx*( 9*a_phi(a_vof,comp)-a_phi(voff,comp)-
                                                 8*value[comp]);
     else
        dns[comp] = iside*ihdx*( a_phi(a_vof,comp)-value[comp]);
   Real drn;
   Real r = static_cast<Real>(a_vof.gridIndex()[0]);
   r  = (r+0.5)*a_dx[0]+a_probLo[0];

   bool rzero_bndry =  (a_vof.gridIndex()[0] == 0 && a_idir == 0 );
   //   bool rzero_bndry = false;


   Real rb, rp;
   if( a_idir == 0 )
   {
      rb = r-iside*a_dx[0]*0.5;
      rp = r+iside*a_dx[0];
      rzero_bndry = rzero_bndry && (fabs(rb)<1e-12);
   }
   else
   {
      rb = r;
      rp = r;
   }

   if( order == 2 )
      drn = iside*itdx*(9*a_phi(a_vof,0)*r-a_phi(voff,0)*rp-
                                                 8*value[0]*rb);
   else
      drn = iside*ihdx*( a_phi(a_vof,0)*r-value[0]*rb );

   Real dudt =0, dvdt=0, drudt=0;

   if( m_isFunctional && !a_useHomogeneous )
   {
      RealVect nv = RealVect::Zero, tv=RealVect::Zero;
      nv[a_idir] = 1;
      if( a_idir == 0 )
         tv[1] = 1;
      else
         tv[0] = 1;
      Vector<Real> ux(SpaceDim), uy(SpaceDim);
      ux = m_func->get_xderivatives( point, a_time );
      uy = m_func->get_yderivatives( point, a_time );
      if( SpaceDim == 2 )
      {
         dudt = tv[0]*ux[0]+tv[1]*uy[0];
         dvdt = tv[0]*ux[1]+tv[1]*uy[1];
         drudt = dudt*rb + nv[1]*value[0];
      }
   }
   if( rzero_bndry )
      a_faceFlux = 0;
   else
   {
      Real flx;
      if( a_idir == 0 )
      {
         flx = value[0]*( 2*rb*mu*dns[0] + la*drn + la*rb*dvdt ) +
               value[1]*mu*rb*( dudt+dns[1] );
         Real rc = static_cast<Real>(a_vof.gridIndex()[0]) +
                                     ebisBox.centroid(a_vof)[0];
         rc = (rc+0.5)*a_dx[0]+a_probLo[0];
         a_faceFlux = flx/rc;
      }
      else
      {
         a_faceFlux = value[0]*mu*( dns[0] + dvdt ) +
	              value[1]*( la*drudt/r + (2*mu+la)*dns[1]);
      }
   }
//   cout << order << " " << a_faceFlux << " ";

   Real faceAreaFrac = 0.0;
   Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
   for( int i = 0; i < faces.size(); i++ )
       faceAreaFrac += ebisBox.areaFrac( faces[i] );
   a_faceFlux *= faceAreaFrac;
   //   cout << a_faceFlux << endl;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::get_the_flux( int a_idir, Vector<Real> dns, 
                                            RealVect point, Real mu, Real la,
					    Vector<Real>& flx, bool a_useHomogeneous,
					    const Real a_time )
{
   RealVect nv = RealVect::Zero, tv=RealVect::Zero, bv=RealVect::Zero;
   nv[a_idir] = 1;
   if( a_idir == 0 )
     tv[1] = 1;
   else
     tv[0] = 1;
   if( SpaceDim == 2 )
   {
     bv[0] = 0;
     bv[1] = 0;
   }
   else
   {   
     bv[0] = nv[1]*tv[2]-nv[2]*tv[1];
     bv[1] = nv[2]*tv[0]-nv[0]*tv[2];
     bv[2] = nv[0]*tv[1]-nv[1]*tv[0];
   }
   Real dutdt=0, dubdb=0, dundt=0, dundb=0;
   if( m_isFunctional && !a_useHomogeneous )
   {
      Vector<Real> ux(SpaceDim), uy(SpaceDim), uz(SpaceDim);
      ux = m_func->get_xderivatives( point, a_time );
      uy = m_func->get_yderivatives( point, a_time );
      if( SpaceDim == 2 )
      {
         dutdt = tv[0]*(tv[0]*ux[0]+tv[1]*ux[1]) + tv[1]*(tv[0]*uy[0]+tv[1]*uy[1]);
         dubdb = bv[0]*(bv[0]*ux[0]+bv[1]*ux[1]) + bv[1]*(bv[0]*uy[0]+bv[1]*uy[1]);
         dundt = tv[0]*(nv[0]*ux[0]+nv[1]*ux[1]) + tv[1]*(nv[0]*uy[0]+nv[1]*uy[1]);
         dundb = bv[0]*(nv[0]*ux[0]+nv[1]*ux[1]) + bv[1]*(nv[0]*uy[0]+nv[1]*uy[1]);
      }
      else if( SpaceDim == 3 )
      {
         uz = m_func->get_zderivatives( point, a_time );
         dutdt = tv[0]*(tv[0]*ux[0]+tv[1]*ux[1]+tv[2]*ux[2]) + 
                 tv[1]*(tv[0]*uy[0]+tv[1]*uy[1]+tv[2]*uy[2]) +
                 tv[2]*(tv[0]*uz[0]+tv[1]*uz[1]+tv[2]*uz[2]);
         dubdb = bv[0]*(bv[0]*ux[0]+bv[1]*ux[1]+bv[2]*ux[2]) + 
                 bv[1]*(bv[0]*uy[0]+bv[1]*uy[1]+bv[2]*uy[2])+
                 bv[2]*(bv[0]*uz[0]+bv[1]*uz[1]+bv[2]*uz[2]);
         dundt = tv[0]*(nv[0]*ux[0]+nv[1]*ux[1]+nv[2]*ux[2]) + 
                 tv[1]*(nv[0]*uy[0]+nv[1]*uy[1]+nv[2]*uy[2])+
                 tv[2]*(nv[0]*uz[0]+nv[1]*uz[1]+nv[2]*uz[2]);
         dundb = bv[0]*(nv[0]*ux[0]+nv[1]*ux[1]+nv[2]*ux[2]) + 
                 bv[1]*(nv[0]*uy[0]+nv[1]*uy[1]+nv[2]*uy[2])+
                 bv[2]*(nv[0]*uz[0]+nv[1]*uz[1]+nv[2]*uz[2]);
      }
   }
   Real dundn = 0;
   for( int c=0; c<SpaceDim ;c++)
      dundn += dns[c]*nv[c];

   for( int c=0; c<SpaceDim ;c++)
   {
     Real tt = la*(dutdt+dubdb)*nv[c] + mu*dundt*tv[c] + mu*dundb*bv[c];
     flx[c] = (mu+la)*dundn*nv[c]+mu*dns[c]+tt;
   }
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::get_the_flux_rz( int a_idir,
					       Vector<Real> dns,
					       Real drn,
					       RealVect point, 
                                               Real mu,
					       Real la,
					       Vector<Real>& flux_div,
		                               Vector<Real>& flux_undiv,
					       Real rb,
					       Real uval,
					       bool rzero_bndry,
					       bool a_useHomogeneous,
					       Real vtnum,
					       const Real a_time )
{
   RealVect nv = RealVect::Zero, tv=RealVect::Zero;
   nv[a_idir] = 1;
   if( a_idir == 0 )
     tv[1] = 1;
   else
     tv[0] = -1;
   Real dudt=0, drt=0, dvdt=0;
   Real truncerr = 0;

   if( rzero_bndry )
      dvdt = vtnum;
   
   if( m_isFunctional && !a_useHomogeneous )
   {
      Vector<Real> ux(SpaceDim), uy(SpaceDim);
      ux = m_func->get_xderivatives( point, a_time );
      uy = m_func->get_yderivatives( point, a_time );
      if( SpaceDim == 2 )
      {
	 dudt = tv[0]*ux[0]+tv[1]*uy[0];
         if( !rzero_bndry )
            dvdt = tv[0]*ux[1]+tv[1]*uy[1];
         drt  = dudt*rb - nv[1]*uval;
	 //	 cout << dns[1] << "  " << nv[0]*ux[1]+nv[1]*uy[1] << endl;
	 //      cout << drn    << " " << rb*dns[0]+nv[0]*uval << endl;
      }
      //      Real dru3 = 6*cos(point[1])
	//      truncerr = h/24*dru3;
      //      truncerr = h/4*cos(point[1]);
   }
   flux_div[0] = 2*mu*rb*nv[0]*(nv[0]*dns[0]-nv[1]*dudt);
   flux_div[1] = mu*rb*(nv[0]*nv[1]*dns[0]+nv[0]*nv[0]*dudt+
			nv[0]*nv[0]*dns[1]-nv[0]*nv[1]*dvdt);
   if( !rzero_bndry )
   {
      flux_undiv[0] = la/rb*nv[0]*( nv[0]*drn - nv[1]*drt ) + 
        la*nv[0]*( nv[1]*dns[1] + nv[0]*dvdt ) + 
        mu*nv[1]*( nv[1]*dns[0] + nv[0]*(dudt+dns[1])-nv[1]*dvdt );
      flux_undiv[1] = (2*mu+la)*nv[1]*( nv[1]*dns[1]+nv[0]*dvdt ) +
        la/rb*nv[1]*(nv[0]*drn - nv[1]*drt);
   }
   else
   {
      flux_undiv[0] =  la*nv[0]*truncerr +
        la*nv[0]*( nv[1]*dns[1] + nv[0]*dvdt ) + 
        mu*nv[1]*( nv[1]*dns[0] + nv[0]*(dudt+dns[1])-nv[1]*dvdt );
      flux_undiv[1] = (2*mu+la)*nv[1]*( nv[1]*dns[1]+nv[0]*dvdt );
   }
}

//-----------------------------------------------------------------------
void DirichletNSviscDomainBC::tangential_derivative( IntVect of,
					   const VolIndex& vof,
					   const EBISBox& eb,
					   VoFStencil& vs,
					   const Real di,
					   const Box ubox )
{
  // try six cases.

  bool found = true;
  vs.clear();
  IntVect v=vof.gridIndex();
  
  Vector<VolIndex> allMonotoneVoFs;
  int radius = 1;
  IntVect timesMoved = IntVect::Zero;
  IntVect pathSign   = IntVect::Zero;
  EBArith::getAllVoFsInMonotonePath(allMonotoneVoFs,timesMoved,pathSign,
                                    vof,eb,radius);

  VolIndex vmm, vm, vp, vpp;  
  if( EBArith::isVoFHere(vp,allMonotoneVoFs,v+of) && 
      EBArith::isVoFHere(vm,allMonotoneVoFs,v-of) && 
      ubox.contains(vp.gridIndex()) &&
      ubox.contains(vm.gridIndex())   )
  {
    vs.add(vp,0.5);
    vs.add(vm,-0.5);
  }
  else if( EBArith::isVoFHere(vpp,allMonotoneVoFs,v+2*of) && 
           EBArith::isVoFHere(vp,allMonotoneVoFs,v+of) &&
      ubox.contains(vp.gridIndex()) &&
      ubox.contains(vpp.gridIndex())   )
  {
    vs.add(vof,-1.5);
    vs.add(vp,2.0);
    vs.add(vpp,-0.5);
  }
  else if( EBArith::isVoFHere(vmm,allMonotoneVoFs,v-2*of) && 
           EBArith::isVoFHere(vm,allMonotoneVoFs,v-of) &&
      ubox.contains(vm.gridIndex()) &&
      ubox.contains(vmm.gridIndex())   )
  {
    vs.add(vmm,0.5);
    vs.add(vm,-2.0);
    vs.add(vof,1.5);
  }
  else if( EBArith::isVoFHere(vp,allMonotoneVoFs,v+of) &&
	   ubox.contains(vp.gridIndex())  )
  {
    vs.add(vp,1.0);
    vs.add(vof,-1.0);
  }
  else if( EBArith::isVoFHere(vm,allMonotoneVoFs,v-of) &&
	   ubox.contains(vm.gridIndex())  )
  {
    vs.add(vm,-1.0);
    vs.add(vof,1.0);
  }
  else
  {
     found =false;
  }
  if( found )
     vs *= di;
}

//-----------------------------------------------------------------------
DirichletNSviscDomainBCFactory::DirichletNSviscDomainBCFactory()
{
  m_ncomp = SpaceDim;

  m_value.resize(m_ncomp);
  for( int c= 0 ; c < m_ncomp ; c++ )
     m_value[c] = 12345.6789;

  m_func = NULL;
  m_onlyHomogeneous = true;
  m_isFunctional    = false;
  m_order = 2;
}


//-----------------------------------------------------------------------
DirichletNSviscDomainBCFactory::~DirichletNSviscDomainBCFactory()
{
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBCFactory::setValue(Vector<Real> a_value)
{
  m_value = a_value;
  m_func = NULL;
  m_onlyHomogeneous = false;
  m_isFunctional    = false;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBCFactory::setFunction(RefCountedPtr<BaseBCVectValue> a_func)
{
  for( int c=0 ; c < m_ncomp ; c++ )
     m_value[c] = 12345.6789;

  m_func = a_func;

  m_onlyHomogeneous = false;
  m_isFunctional = true;
}


//-----------------------------------------------------------------------
void DirichletNSviscDomainBCFactory::set_order(int a_order)
{
  CH_assert(a_order >= 1);
  CH_assert(a_order <= 2);
  m_order = a_order;
}


//-----------------------------------------------------------------------
DirichletNSviscDomainBC* DirichletNSviscDomainBCFactory::create(
                     	       const ProblemDomain& a_domain,
                               const EBISLayout&    a_layout,
                               const RealVect&      a_dx )
{
  DirichletNSviscDomainBC* newBC = new DirichletNSviscDomainBC();
  newBC->set_order(m_order);
  if(!m_onlyHomogeneous)
    {
      if(m_isFunctional)
        {
          newBC->setFunction(m_func);
        }
      else
        {
          newBC->setValue(m_value);
        }
    }
  return newBC;
}
