/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _DIRICHLETNSVISCEBBC_H_
#define _DIRICHLETNSVISCEBBC_H_

#include "RefCountedPtr.H"
#include "BaseBCVectValue.H"

/// Impose boundary data on the embedded boundary for the velocities in the Navier-Stokes operator.
/**
  The normal flux of the NS operator is computed from given Dirichlet data for the 
  velocities. The embedded boundary is assumed piecewise planar. 
  The flux is given by (Cartesian case)
    \f[
       ((\lambda+2\mu)d(u_n)/dn+\lambda d(u_{t1})/dt1 + \lambda(u_{t2})/dt2)*e_n +
             \mu(d(u_n)/dt1+d(u_{t1})/dn)*e_{t1} + \mu(d(u_n)/dt2+d(u_{t2})/dn)*e_{t2}
    \f]
    where \f$ (\lambda, \mu)\f$ are viscosity coefficients, \f$ e_n\f$ is the unit
    vector normal to the boundary, \f$ e_{t1}\f$ and \f$ e_{t2}\f$ are unit vectors
    tangential to the boundary. 
    \f$ (e_n,e_{t1},e_{t2})\f$ are orthonormal vectors. \f$ (u_n,u_{t1},u_{t2})\f$ 
    are the velocity components in the basis \f$ (e_n,e_{t1},e_{t2})\f$.
    The input and output parameters of the methods are all assumed to be in
    Cartesian (x,y,z) or cylindrical (r,z) coordinates.

    When created, the object is in a state where only homogeneous
    (i.e. zero velocity) data can be imposed. The 'setFunction' method can be used
    to specify any given function for boundary data. The boundary function must
    be derived from the BaseBCVectValue class. The 'setValue' method is used to specify
    a constant possibly non-zero boundary value. The most recent call to 'setFunction'
    or 'setValue' overrides previous calls to these two functions.
    The formal order of accuracy of the boundary flux is by default 2, but can be
    changed by calling the 'set_order' method.

    The number of components in the fluxes should be 2 in two space dimensions, and
    3 in three space dimensions. The cylindrical case is assumed to be two dimensional.

    The class sets up stencils for the computation of the normal derivatives at the 
    embedded boundary for all irregular cells in one grid level. Therefore, a DataIndex
    is sometimes needed as input to the functions that compute fluxes.
    The stencils are set up by the 'define' method. Each flux computation function 
    checks if 'define' has been called, and calls it if necessary.
 */
class DirichletNSviscEBBC
{
public:
  /// Constructor.
  /**
    - Input:
      - a_domain - The entire domain (surrounding box) for this grid level.
      - a_layout - The grids on this grid level.
      - a_dx     - Discretization size on this grid level.
   */
  DirichletNSviscEBBC(const ProblemDomain& a_domain,
                const EBISLayout&    a_layout,
                const RealVect&      a_dx);

  /// Destructor.
  /**
   */
  virtual ~DirichletNSviscEBBC();

  /// Set up stencils for the normal derivative at the boundary.
  /**
     This function is called automatically when necessary (should perhaps have
     been 'private'? ).
   */
  virtual void define();

  /// Set the order of accuracy of boundary flux approximation.
  /**
      Order can be one or two, leading to order zero and one respectively when
      using the flux in a divergence update.
   */
  virtual void setOrder(int a_order);

  /// Define the boundary value as a constant.
  /** Input the value at the boundary in a vector.
   */ 
  virtual void setValue(Vector<Real> a_value);

  /// Define the boundary value as a function.
  /**
    Input a vector valued function that specifies the values on the boundary.
   */
  virtual void setFunction(RefCountedPtr<BaseBCVectValue> a_func);

  /// Boundary fluxes at one cell.
  /**
      - Input:
         - a_vof    - The cell where the boundary flux is to be computed.
         - a_phi    - The vector valued function (ie.velocities) over a set of 
                      grids on one grid level.
         - a_dit    - Data index for the grid where the a_vof is located.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
         - a_useHomogeneous - If true, override the set boundary value and 
                                impose zero data.
	 - mu, la   - Viscosity coefficients.
      - Output: a_flux - The flux along the embedded boundary in the a_vof.
   */
  virtual void getEBFlux(Vector<Real>&               a_flux,
                         const VolIndex&             a_vof,
                         const LevelData<EBCellFAB>& a_phi,
                         const DataIndex             a_dit,
                         const RealVect&             a_probLo,
                         const RealVect&             a_dx,
                         bool                        a_useHomogeneous,
                         const Real a_mu, 
                         const Real a_la,
			 const Real a_time );

  /// Boundary fluxes at one cell, in cylindrical coordinates.
  /**
      - Input:
         - a_vof    - The cell where the boundary flux is to be computed.
         - a_phi    - The vector valued function (ie.velocities) over a set of 
                      grids on one grid level.
         - a_dit    - Data index for the grid where the a_vof is located.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
         - a_useHomogeneous - If true, override the set boundary value and 
                                impose zero data.
	 - mu, la   - Viscosity coefficients.
      - Output: a_flux - The flux along the embedded boundary in the a_vof.
   */
 void getEBFlux_rz(Vector<Real>&      a_flux,
                  const VolIndex&             a_vof,
                  const LevelData<EBCellFAB>& a_phi,
                  const DataIndex             a_dit,
                  const RealVect&             a_probLo,
                  const RealVect&             a_dx,
                  bool                        a_useHomogeneous,
                  const Real a_mu,
                  const Real a_la,
		  const Real a_time );
  
  /// Boundary fluxes at one cell, energy equation term.
  /**
      - Input:
         - a_vof    - The cell where the boundary flux is to be computed.
         - a_phi    - The vector valued function (ie.velocities) over a set of 
                      grids on one grid level.
         - a_dit    - Data index for the grid where the a_vof is located.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
         - a_useHomogeneous - If true, override the set boundary value and 
                                impose zero data.
	 - mu, la   - Viscosity coefficients.
      - Output: a_flux - The flux along the embedded boundary in the a_vof.
   */
 void getEBFlux_eeq( Real&               a_flux,
                         const VolIndex&             a_vof,
                         const LevelData<EBCellFAB>& a_phi,
                         const DataIndex             a_dit,
                         const RealVect&             a_probLo,
                         const RealVect&             a_dx,
                         bool                        a_useHomogeneous,
                         const Real mu, 
                         const Real la,
		         const Real a_time );

 void getEBFlux_eeq_rz( Real&               a_flux,
                         const VolIndex&             a_vof,
                         const LevelData<EBCellFAB>& a_phi,
                         const DataIndex             a_dit,
                         const RealVect&             a_probLo,
                         const RealVect&             a_dx,
                         bool                        a_useHomogeneous,
                         const Real mu, 
                         const Real la,
			 const Real a_time );


/// Contribution from boundary to the diagonal coefficient in the discretization, cylindrical coordinates.
  /**
    When using a relaxation scheme, such as Jacobi or Gauss-Seidel, we need to
    divide by the diagonal of the discrete equation. In one space dimension for 
    example the diagonal element of \f$ (u_{i+1}-2u_i+u_{i-1})/h^2 = f_i \f$ is
    \f$ -2/h^2\f$, but at a boundary with flux \f$ 2*(u_1-u_{1/2})/h\f$, we get  
    the approximation \f$ (u_{2}-3u_1+2*u_{1/2})/h^2 = f_1\f$, i.e., the diagonal
    weight is \f$ -3/h^2\f$.
    This function computes the boundary contribution, in the example above it would
    be \f$ -2/h^2 \f$.
    The two equations are treated as independent scalar equations.
      - Input:
         - a_vof    - The cell where the diagonal weight is to be computed.
         - a_ebisbox- The EBISbox of the present grid.
         - a_dit    - DataIndex in the grid level for the present grid.
         - a_probLo - Origin of coordinate system.
	 - a_dx     - Spatial step size.
	 - mu, la   - Viscosity coefficients, variable on the grid.

      - Output: a_wghs - The contribution to the diagonal by the flux on the embedded
                 boundary in the cell a_vof.
   */
void get_uweight_rz( Vector<Real>& a_wghs,
		     const VolIndex&  a_vof,
		     const EBISBox&   ebisBox,
		     const DataIndex a_dit,
		     const RealVect&  a_probLo,
		     const RealVect&  a_dx,
		     Real mu, Real la );

private:
  virtual void getFirstOrderStencil(VoFStencil&     a_stencil,
                                    Real&           a_weight,
                                    const VolIndex& a_vof,
                                    const EBISBox&  a_ebisBox,
                                    const RealVect& a_dx);

  virtual void getSecondOrderStencil(VoFStencil&     a_stencil,
                                     Real&           a_weight,
                                     const VolIndex& a_vof,
                                     const EBISBox&  a_ebisBox,
                                     const RealVect& a_dx);

  virtual void johanDropOrder(bool&          a_droppedOrder,
                              const IntVect  a_interpCen[2],
                              const int      a_planeDir[SpaceDim-1],
                              const EBISBox& a_ebisBox);

  void get_the_flux( const RealVect& nv, const Vector<Real>& dns, 
                     const RealVect& point, const Real mu, 
                     const Real la, Vector<Real>& a_flux, bool a_useHomogeneous,
                     const Real a_time );

 void get_the_flux_rz( const RealVect& nv, const Vector<Real>& dns, 
                       const Real& drn, const RealVect& point, 
                       const Real mu, const Real la, 
	 	       Vector<Real>& a_flux_div, 
                       Vector<Real>& a_flux_undiv, 
		       Real rb, Real uval, bool a_useHomogeneous,
		       const Real a_time );
  
  void get_tanvects( const RealVect nv, RealVect& tv, RealVect& bv );

  VoFStencil BuildEQIStencil(const VolIndex a_vofs[3],
                             const Real&    a_distance);

  VoFStencil BuildEQIStencil(VoFStencil a_vofs[3],
                             const Real& a_distance);

  VoFStencil ivofStencil(const VoFStencil a_iPlaneStencil[2],
                         const Real&      a_distance1,
                         const Real&      a_distance2);

  bool m_isDefined;
  bool m_onlyHomogeneous;
  bool m_isFunction;

  int m_order;
  int m_ncomp;

  Vector<Real> m_value;
  RefCountedPtr<BaseBCVectValue> m_func;

  ProblemDomain m_domain;
  EBISLayout    m_layout;

  RealVect m_dx;

  LayoutData<BaseIVFAB<Real> >       m_fluxWeight;
  LayoutData<BaseIVFAB<VoFStencil> > m_fluxStencil;
};

/// Class that can generate a DirichletNSviscEBBC object.
/**
 */
class DirichletNSviscEBBCFactory
{
public:
  /// Constructor.
  /**
   */
  DirichletNSviscEBBCFactory();

  /// Destructor.
  /**
   */
  virtual ~DirichletNSviscEBBCFactory();

  /// Order of accuracy of boundary flux approximation in the to be generated object.
  /** Order can be one or two.
   */
  virtual void setOrder(int a_order);

  /// Define the boundary value as a constant in the to be generated object.
  /** Input the value at the boundary in a vector.
   */ 
  virtual void setValue(Vector<Real> a_value);

  /// Define the boundary value as a function in the to be generated object.
  /** Input a vector valued function that specifies the values on the boundary.
   */
  virtual void setFunction(RefCountedPtr<BaseBCVectValue> a_func);

  /// Creates (a pointer to) the boundary condition object.
  /** Input arguments are the same as to the constructor of DirichletNSviscEBBC.
    - Input:
      - a_domain - The entire domain (surrounding box) for this grid level.
      - a_layout - The grids on this grid level.
      - a_dx     - Discretization size on this grid level.
   */
  virtual DirichletNSviscEBBC* create(const ProblemDomain& a_domain,
                                const EBISLayout&    a_layout,
                                const RealVect&      a_dx);

private:
  bool m_onlyHomogeneous;
  bool m_isFunction;

  int m_order;
  int m_ncomp;
  Vector<Real> m_value;
  RefCountedPtr<BaseBCVectValue> m_func;
};

#endif
