/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cmath>
#include "BoxIterator.H"
#include "EBArith.H"
#include "DirichletNSviscEBBC.H"


//-----------------------------------------------------------------------
DirichletNSviscEBBC::DirichletNSviscEBBC(const ProblemDomain& a_domain,
					 const EBISLayout&    a_layout,
					 const RealVect&      a_dx)
{
   m_ncomp = SpaceDim;
   m_value.resize(m_ncomp);
   for( int c=0 ; c<m_ncomp ; c++ )
      m_value[c] = 12345.6789;

   m_func = NULL;

   m_order = 1;

   m_onlyHomogeneous = true;
   m_isFunction      = false;

   m_domain = a_domain;
   m_layout = a_layout;

   m_dx = a_dx;

   m_isDefined = false;
}


//-----------------------------------------------------------------------
DirichletNSviscEBBC::~DirichletNSviscEBBC()
{
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::define()
{
   const DisjointBoxLayout& dbl = m_layout.getDisjointLayout();
   m_fluxStencil.define( dbl );
   m_fluxWeight.define( dbl );

   for (DataIterator dit = dbl.dataIterator(); dit.ok(); ++dit)
   {
      const Box& curBox = dbl[dit()];
      //      cout << "Defining EBBC for " << curBox << endl;
      const EBISBox& curEBISBox = m_layout[dit()];
      const EBGraph& curEBGraph = curEBISBox.getEBGraph();

      IntVectSet notRegular;
      int nComps = 1;

      notRegular |= curEBISBox.getIrregIVS  (curBox);
      notRegular |= curEBISBox.getMultiCells(curBox);

      BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[dit()];
      BaseIVFAB<Real>&       curWeightBaseIVFAB  = m_fluxWeight[dit()];

      curStencilBaseIVFAB.define(notRegular,curEBGraph,nComps);
      curWeightBaseIVFAB.define(notRegular,curEBGraph,nComps);

      for (VoFIterator vofit(notRegular,curEBGraph); vofit.ok(); ++vofit)
      {
         const VolIndex& vof = vofit();
         VoFStencil& curStencil = curStencilBaseIVFAB(vof,0);
         Real&       curWeight  = curWeightBaseIVFAB(vof,0);
         if( m_order == 1 )
            getFirstOrderStencil(curStencil,curWeight,vof,curEBISBox,m_dx);
         else if( m_order == 2 )
            getSecondOrderStencil(curStencil,curWeight,vof,curEBISBox,m_dx);
         else
            MayDay::Error("DirichletNSviscEBBC::define stencil order not 1 or 2");
      }
      //      cout << "number of cut cells " << notRegular.numPts() << endl;
   }
   m_isDefined = true;
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::setOrder(int a_order)
{
  CH_assert(a_order >= 1 && a_order <= 2);

  if (m_order != a_order)
  {
    m_isDefined = false;
  }
  m_order = a_order;
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::setValue( Vector<Real> a_value )
{
  m_value = a_value;
  m_func = NULL;
  m_onlyHomogeneous = false;
  m_isFunction      = false;
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::setFunction(RefCountedPtr<BaseBCVectValue> a_func)
{
  for( int c= 0 ; c<m_ncomp ; c++ )
     m_value[c] = 12345.6789;

  m_func = a_func;

  m_onlyHomogeneous = false;
  m_isFunction      = true;
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::getEBFlux(Vector<Real>&         a_flux,
                              const VolIndex&             a_vof,
                              const LevelData<EBCellFAB>& a_phi,
                              const DataIndex             a_dit,
                              const RealVect&             a_probLo,
                              const RealVect&             a_dx,
                              bool                        a_useHomogeneous,
                              const Real a_mu,
                              const Real a_la,
			      const Real a_time )
{
  CH_assert(a_phi.nComp() == m_ncomp );

  const EBCellFAB& curPhi = a_phi[a_dit];
  const EBISBox    curEBISBox = curPhi.getEBISBox();

  RealVect point;

  if (!m_isDefined)
  {
    define();
  }

  Vector<Real> value(m_ncomp);

  if (a_useHomogeneous)
  {
    for( int c = 0 ; c < m_ncomp ; c++ )
       value[c] = 0.0;
  }
  else
  {
    if (m_isFunction)
    {
      const IntVect& iv = a_vof.gridIndex();

      point = RealVect::Unit;
      point *= 0.5;
      point += iv;

      point += curEBISBox.bndryCentroid(a_vof);
      point *= a_dx;

      point += a_probLo;

      value = m_func->get_value( point, a_time );
    }
    else
    {
      if (m_onlyHomogeneous)
      {
        MayDay::Error("DirichletNSviscEBBC::getFaceFlux called with undefined inhomogeneous BC");
      }

      value = m_value;
    }
  }

  Vector<Real> dns(m_ncomp,0);

  const BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[a_dit];
  const VoFStencil& curStencil = curStencilBaseIVFAB(a_vof,0);
  for (int i = 0; i < curStencil.size(); i++)
  {
     VolIndex curVoF = curStencil.vof(i);
     Real weight = curStencil.weight(i);
     for( int c=0 ; c < m_ncomp ; c++ )
        dns[c] += weight * curPhi(curVoF,c);
  }

  const BaseIVFAB<Real>& curWeightBaseIVFAB = m_fluxWeight[a_dit];
  const Real& curWeight = curWeightBaseIVFAB(a_vof,0);

  for( int c =0; c<m_ncomp ; c++ )
     dns[c] += curWeight*value[c];

  RealVect nv = curEBISBox.normal(a_vof);
  for( int d = 0 ; d <SpaceDim ; d++ )
    nv[d] = -nv[d];

  get_the_flux( nv, dns, point, a_mu, a_la, a_flux, a_useHomogeneous, a_time );

  const Real& areaFrac = curEBISBox.bndryArea(a_vof);
  for( int c= 0 ; c<m_ncomp ; c++ )
     a_flux[c] *= areaFrac;
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::getEBFlux_rz(Vector<Real>&      a_flux,
                              const VolIndex&             a_vof,
                              const LevelData<EBCellFAB>& a_phi,
                              const DataIndex             a_dit,
                              const RealVect&             a_probLo,
                              const RealVect&             a_dx,
                              bool                        a_useHomogeneous,
                              const Real mu,
                              const Real la,
			      const Real a_time )
{
  CH_assert(a_phi.nComp() == m_ncomp );
  CH_assert( SpaceDim == 2 );

  const EBCellFAB& curPhi = a_phi[a_dit];
  const EBISBox    curEBISBox = curPhi.getEBISBox();

  RealVect point;

  if (!m_isDefined)
  {
    define();
  }

  Vector<Real> value(m_ncomp,0);

  if (a_useHomogeneous)
  {
    for( int c = 0 ; c < m_ncomp ; c++ )
       value[c] = 0.0;
  }
  else
  {
    if (m_isFunction)
    {
      const IntVect& iv = a_vof.gridIndex();

      point = RealVect::Unit;
      point *= 0.5;
      point += iv;

      point += curEBISBox.bndryCentroid(a_vof);
      point *= a_dx;

      point += a_probLo;

      value = m_func->get_value( point, a_time );
    }
    else
    {
      if (m_onlyHomogeneous)
      {
        MayDay::Error("DirichletNSviscEBBC::getEBFlux_rz called with undefined inhomogeneous BC");
      }
      value = m_value;
    }
  }

  Vector<Real> dns(m_ncomp,0);
  Real drn=0;
  if( fabs(curEBISBox.bndryArea(a_vof)) > 0 )
  {
  const BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[a_dit];
  const VoFStencil& curStencil = curStencilBaseIVFAB(a_vof,0);
  for (int i = 0; i < curStencil.size(); i++)
  {
     VolIndex curVoF = curStencil.vof(i);
     Real weight = curStencil.weight(i);
     for( int c=0 ; c < m_ncomp ; c++ )
        dns[c] += weight * curPhi(curVoF,c);
     Real r = static_cast<Real>(curVoF.gridIndex()[0]);
     r = (0.5+r)*a_dx[0]+a_probLo[0];
     drn += weight*curPhi(curVoF,0)*r;
  }

  const BaseIVFAB<Real>& curWeightBaseIVFAB = m_fluxWeight[a_dit];
  const Real& curWeight = curWeightBaseIVFAB(a_vof,0);

  for( int c =0; c<m_ncomp ; c++ )
     dns[c] += curWeight*value[c];

  Real rb=(curEBISBox.bndryCentroid(a_vof)[0]+0.5+a_vof.gridIndex()[0])*a_dx[0]+a_probLo[0];
  drn += curWeight*value[0]*rb;

  RealVect nv = curEBISBox.normal(a_vof);
  for( int d = 0 ; d <SpaceDim ; d++ )
    nv[d] = -nv[d];

  Vector<Real> flux_div(2), flux_undiv(2);

  //  rb = (rb+0.5)*a_dx[0]+a_probLo[0];

  //  drn = rb*dns[0]+nv[0]*value[0];
  get_the_flux_rz( nv, dns, drn, point, mu, la, flux_div, flux_undiv, rb, value[0],
		   a_useHomogeneous, a_time );

  Real rc = static_cast<Real>(a_vof.gridIndex()[0]) + curEBISBox.centroid(a_vof)[0];

  rc = (rc+0.5)*a_dx[0]+a_probLo[0];

  a_flux[0] = flux_div[0]/rc + flux_undiv[0];
  a_flux[1] = flux_div[1]/rc + flux_undiv[1];

  const Real& areaFrac = curEBISBox.bndryArea(a_vof);
  for( int c= 0 ; c<m_ncomp ; c++ )
     a_flux[c] *= areaFrac;
  //  cout << "eb flux " << a_vof.gridIndex() << " " << a_flux << endl;
  }
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::getEBFlux_eeq(Real&         a_flux,
                              const VolIndex&             a_vof,
                              const LevelData<EBCellFAB>& a_phi,
                              const DataIndex             a_dit,
                              const RealVect&             a_probLo,
                              const RealVect&             a_dx,
                              bool                        a_useHomogeneous,
                              const Real mu,
                              const Real la,
			      const Real a_time )
{
   a_flux = 0;
   if (!m_isDefined)
   {
      define();
   }
   if( m_isFunction )
   {
     //      const EBCellFAB& curPhi = a_phi[a_dit];
      const EBISBox  curEBISBox = a_phi[a_dit].getEBISBox();

      Vector<Real> flxx(m_ncomp,0);
      getEBFlux( flxx, a_vof, a_phi, a_dit, a_probLo, a_dx, a_useHomogeneous,mu,la,a_time);
      RealVect point;
      for( int d = 0 ; d < SpaceDim ; d++ )
        point[d]=(curEBISBox.bndryCentroid(a_vof)[d]+0.5+a_vof.gridIndex()[d])*a_dx[d]+a_probLo[d];
      Vector<Real> value(m_ncomp);
      value = m_func->get_value(point,a_time);
      for( int d= 0 ; d < m_ncomp ; d++ )
	a_flux += flxx[d]*value[d];
   }
}


//-----------------------------------------------------------------------
 void DirichletNSviscEBBC::getEBFlux_eeq_rz(
 			      Real&         a_flux,
                               const VolIndex&             a_vof,
                               const LevelData<EBCellFAB>& a_phi,
                               const DataIndex             a_dit,
                               const RealVect&             a_probLo,
                               const RealVect&             a_dx,
                               bool                        a_useHomogeneous,
                               const Real mu,
                               const Real la,
 			      const Real a_time )
 {
    a_flux = 0;
    if (!m_isDefined)
    {
       define();
    }
    const EBCellFAB& curPhi = a_phi[a_dit];
    const EBISBox  curEBISBox = a_phi[a_dit].getEBISBox();
    RealVect point;
    for( int d = 0 ; d < SpaceDim ; d++ )
       point[d]=(curEBISBox.bndryCentroid(a_vof)[d]+0.5+a_vof.gridIndex()[d])*a_dx[d]+a_probLo[d];
    //   Real rb=(curEBISBox.bndryCentroid(a_vof)[0]+0.5+a_vof.gridIndex()[0])*a_dx[0]+a_probLo[0];
    Real rb = point[0];
    RealVect nv = curEBISBox.normal(a_vof);
    for( int d = 0 ; d <SpaceDim ; d++ )
       nv[d] = -nv[d];
   
    Vector<Real> value(m_ncomp,0);
    if( a_useHomogeneous )
      for( int d =0 ; d < m_ncomp ; d++ )
 	value[d] =  0;
    else if( !m_isFunction )
      for( int d =0 ; d < m_ncomp ; d++ )
 	value[d] =  m_value[d];
    else 
       value = m_func->get_value( point, a_time );

    Vector<Real> dns(m_ncomp,0);
    Real drn=0;
    if( fabs(curEBISBox.bndryArea(a_vof)) > 0 )
    {
      const BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[a_dit];
      const VoFStencil& curStencil = curStencilBaseIVFAB(a_vof,0);
      for (int i = 0; i < curStencil.size(); i++)
      {
         VolIndex curVoF = curStencil.vof(i);
         Real weight = curStencil.weight(i);
         for( int c=0 ; c < m_ncomp ; c++ )
            dns[c] += weight * curPhi(curVoF,c);
         Real r = static_cast<Real>(curVoF.gridIndex()[0]);
         r = (0.5+r)*a_dx[0]+a_probLo[0];
         drn += weight*curPhi(curVoF,0)*r;
      }

      const BaseIVFAB<Real>& curWeightBaseIVFAB = m_fluxWeight[a_dit];
      const Real& curWeight = curWeightBaseIVFAB(a_vof,0);
      for( int c =0; c<m_ncomp ; c++ )
         dns[c] += curWeight*value[c];
      drn += curWeight*value[0]*rb;

      RealVect tv, bv;
      get_tanvects( nv, tv, bv );
      Real dudt=0, dvdt=0, drudt=0;
      if( m_isFunction && !a_useHomogeneous )
      {
          Vector<Real> ux(SpaceDim), uy(SpaceDim);
          ux = m_func->get_xderivatives( point, a_time );
          uy = m_func->get_yderivatives( point, a_time );
          if( SpaceDim == 2 )
          {
 	    dudt = tv[0]*ux[0]+tv[1]*uy[0];
             dvdt = tv[0]*ux[1]+tv[1]*uy[1];
             drudt  = dudt*rb - nv[1]*value[0];
          }
      }
      Vector<Real> drs(m_ncomp);
      drs[0] = nv[0]*dns[0]-nv[1]*dudt;
      drs[1] = nv[0]*dns[1]-nv[1]*dvdt;
      Real drur   = nv[0]*drn-nv[1]*drudt;
      Real duz    = nv[1]*dns[0]+nv[0]*dudt;
      Real dvz    = nv[1]*dns[1]+nv[0]*dvdt;
     
      Real flxr = value[0]*( 2*rb*mu*drs[0] + la*drur + la*rb*dvz ) +
                   value[1]*mu*rb*( duz+drs[1] );
      Real flxz = value[0]*mu*rb*( drs[1] + duz  ) +
 	          value[1]*( la*drur + (2*mu+la)*rb*dvz );

      //     Real flxr = value[0]*( 2*rb*mu*dns[0] + la*drn + la*rb*dvdt ) +
      //                  value[1]*mu*rb*( dudt+dns[1] );
      //     Real flxz = value[0]*mu*rb*( dns[0] + dvdt ) +
      //	          value[1]*( la*drudt + (2*mu+la)*rb*dns[1]);

      Real rc = static_cast<Real>(a_vof.gridIndex()[0]) +
                                      curEBISBox.centroid(a_vof)[0];
      rc = (rc+0.5)*a_dx[0]+a_probLo[0];
      a_flux = (nv[0]*flxr + nv[1]*flxz)/rc*curEBISBox.bndryArea(a_vof);
    }
 }


 //-----------------------------------------------------------------------
 void DirichletNSviscEBBC::get_uweight_rz( Vector<Real>& a_wghs,
                                            const VolIndex&  a_vof,
 				           const EBISBox&   ebisBox,
 					   const DataIndex a_dit,
                                            const RealVect&  a_probLo,
                                            const RealVect&  a_dx,
                                            Real mu, Real la )
 {
   CH_assert(a_wghs.size() == m_ncomp );
   CH_assert( SpaceDim == 2 );
   if (!m_isDefined)
   {
      define();
   }

   //  const EBCellFAB& curPhi = a_phi[a_dit];
   //  const EBISBox    curEBISBox = curPhi.getEBISBox();

   Vector<Real> dns(m_ncomp,0);
   Real drn=0;

   const BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[a_dit];
   if( curStencilBaseIVFAB.getIVS().contains(a_vof.gridIndex()))
     {
   const VoFStencil& curStencil = curStencilBaseIVFAB(a_vof,0);
   for (int i = 0; i < curStencil.size(); i++)
   {
      VolIndex curVoF = curStencil.vof(i);
      if( curVoF == a_vof )
      {
         Real weight = curStencil.weight(i);
         for( int c=0 ; c < m_ncomp ; c++ )
 	   dns[c] += weight;
         Real r = static_cast<Real>(curVoF.gridIndex()[0]);
         r = (0.5+r)*a_dx[0]+a_probLo[0];
         drn += weight*r;
      }

   }
   //  cout << dns[0] << "  " << dns[1] << "  " << drn << endl;
   Real rb=(ebisBox.bndryCentroid(a_vof)[0]+0.5+a_vof.gridIndex()[0])*a_dx[0]+a_probLo[0];
   RealVect nv = ebisBox.normal(a_vof);
   for( int d = 0 ; d <SpaceDim ; d++ )
     nv[d] = -nv[d];

   Vector<Real> flux_div(2), flux_undiv(2);


   //  rb = (rb+0.5)*a_dx[0]+a_probLo[0];

   //  drn = rb*dns[0]+nv[0]*value[0];
   //  get_the_flux_rz( nv, dns, drn, point, mu, la, flux_div, flux_undiv, rb, value[0],
   //		   a_useHomogeneous );

    RealVect tv, bv;
    get_tanvects( nv, tv, bv );
    //   Real dudt=0, dvdt=0, drt=0;

    flux_div[0] = 2*mu*rb*nv[0]*(nv[0]*dns[0]);

    //   flux_div[1] = mu*rb*(nv[0]*nv[1]*dns[0]+nv[0]*nv[0]*dns[1]);
    flux_div[1] = mu*rb*(nv[0]*nv[0]*dns[1]);

    flux_undiv[0] = la/rb*(nv[0]*nv[0]*drn) + 
      mu*(nv[1]*nv[1]*dns[0]);
    //   flux_undiv[0] = la/rb*(nv[0]*nv[0]*drn) + 
    //     la*(nv[0]*nv[1]*dns[1]) + 
    //     mu*(nv[1]*nv[1]*dns[0]+nv[0]*nv[1]*(dns[1]));
    flux_undiv[1] = (2*mu+la)*(nv[1]*nv[1]*dns[1]);
    //   flux_undiv[1] = (2*mu+la)*(nv[1]*nv[1]*dns[1]) +
    //     la/rb*nv[1]*(nv[0]*drn);


   Real rc = static_cast<Real>(a_vof.gridIndex()[0]) + ebisBox.centroid(a_vof)[0];
   rc = (rc+0.5)*a_dx[0]+a_probLo[0];

   a_wghs[0] = flux_div[0]/rc + flux_undiv[0];
   a_wghs[1] = flux_div[1]/rc + flux_undiv[1];

   const Real& areaFrac = ebisBox.bndryArea(a_vof);
   for( int c= 0 ; c<m_ncomp ; c++ )
      a_wghs[c] *= areaFrac;
     }
 }


 //-----------------------------------------------------------------------
 void DirichletNSviscEBBC::getFirstOrderStencil(VoFStencil&     a_stencil,
                                          Real&           a_weight,
                                          const VolIndex& a_vof,
                                          const EBISBox&  a_ebisBox,
                                          const RealVect& a_dx)
 {
   RealVect normal   = a_ebisBox.normal(a_vof);
   RealVect centroid = a_ebisBox.bndryCentroid(a_vof);

   IntVect iv0 = a_vof.gridIndex();

 #if   CH_SPACEDIM == 2
   int stenSize = 3;
 #elif CH_SPACEDIM == 3
   int stenSize = 7;
 #else
   tjlCompileTimeError;
 #endif

   IntVect sign;
   for (int idir = 0; idir < SpaceDim; idir++)
   {
     if (normal[idir] < 0)
     {
       sign[idir]= -1;
     }
     else
     {
       sign[idir] = 1;
     }
   }

   Vector<IntVect> ivSten(stenSize);

   ivSten[0] = iv0 + sign[0]*BASISV(0)                                        ;
   ivSten[1] = iv0                     + sign[1]*BASISV(1)                    ;
   ivSten[2] = iv0 + sign[0]*BASISV(0) + sign[1]*BASISV(1)                    ;
 #if CH_SPACEDIM == 3
   ivSten[3] = iv0                                         + sign[2]*BASISV(2);
   ivSten[4] = iv0 + sign[0]*BASISV(0)                     + sign[2]*BASISV(2);
   ivSten[5] = iv0                     + sign[1]*BASISV(1) + sign[2]*BASISV(2);
   ivSten[6] = iv0 + sign[0]*BASISV(0) + sign[1]*BASISV(1) + sign[2]*BASISV(2);
 #endif

   bool dropOrder = false;

   Vector<VolIndex> volSten(stenSize);
   for (int isten = 0; isten < stenSize; isten++)
   {
     if (a_ebisBox.getDomain().contains(ivSten[isten]))
     {
       if (a_ebisBox.getVoFs(ivSten[isten]).size() > 0)
       {
         volSten[isten] = a_ebisBox.getVoFs(ivSten[isten])[0];
       }
       else
       {
         dropOrder = true;
         break;
       }
     }
     else
     {
       dropOrder = true;
       break;
     }
   }

   std::vector<bool> monotonePath(stenSize);

   for (int isten = 0; isten < stenSize; isten++)
   {
     monotonePath[isten] = EBArith::monotonePathVoFToCellVoF(volSten[isten],
                                                             a_vof,
                                                             ivSten[isten],
                                                             a_ebisBox);
     if (!monotonePath[isten])
     {
       dropOrder = true;
     }
   }

   if (!dropOrder)
   {
     RealVect x0;
     for (int idir = 0; idir < SpaceDim; idir++)
     {
       x0[idir] = a_dx[idir] * (0.5 + centroid[idir] + iv0[idir]);
     }

     Vector<RealVect> xp(stenSize);
     for (int isten = 0; isten < stenSize; isten++)
     {
       for (int idir = 0; idir < SpaceDim; idir++)
       {
         xp[isten][idir] = a_dx[idir] * (0.5 + ivSten[isten][idir]);
       }
     }

     Vector<RealVect> deltaX = xp;
     for (int isten = 0; isten < stenSize; isten++)
     {
       deltaX[isten] -= x0;
     }

     Vector<RealVect> aTransA(SpaceDim), invATransA(SpaceDim);
     for (int idir = 0; idir < SpaceDim; idir++)
     {
       for (int jdir = 0; jdir < SpaceDim; jdir++)
       {
         for (int isten = 0; isten < stenSize; isten++)
         {
           aTransA[idir][jdir] = aTransA[idir][jdir]
                               + deltaX[isten][idir]*deltaX[isten][jdir];
         }
       }
     }

     Real det;
 #if   CH_SPACEDIM == 2
     det = aTransA[0][0] * aTransA[1][1] - aTransA[0][1] * aTransA[1][0];

     invATransA[0][0] =  aTransA[1][1] / det;
     invATransA[0][1] = -aTransA[0][1] / det;
     invATransA[1][0] = -aTransA[1][0] / det;
     invATransA[1][1] =  aTransA[0][0] / det;
 #elif CH_SPACEDIM == 3
     det = aTransA[0][0] * ( aTransA[1][1] * aTransA[2][2]
                           - aTransA[1][2] * aTransA[2][1])
         + aTransA[0][1] * ( aTransA[1][2] * aTransA[2][0]
                           - aTransA[1][0] * aTransA[2][2])
         + aTransA[0][2] * ( aTransA[1][0] * aTransA[2][1]
                           - aTransA[1][1] * aTransA[2][0]);

     invATransA[0][0] = ( aTransA[1][1] * aTransA[2][2]
                        - aTransA[1][2] * aTransA[2][1]) / det;
     invATransA[0][1] = ( aTransA[1][2] * aTransA[2][0]
                        - aTransA[1][0] * aTransA[2][2]) / det;
     invATransA[0][2] = ( aTransA[1][0] * aTransA[2][1]
                        - aTransA[1][1] * aTransA[2][0]) / det;
     invATransA[1][0] = ( aTransA[2][1] * aTransA[0][2]
                        - aTransA[2][2] * aTransA[0][1]) / det;
     invATransA[1][1] = ( aTransA[2][2] * aTransA[0][0]
                        - aTransA[2][0] * aTransA[0][2]) / det;
     invATransA[1][2] = ( aTransA[2][0] * aTransA[0][1]
                        - aTransA[2][1] * aTransA[0][0]) / det;
     invATransA[2][0] = ( aTransA[0][1] * aTransA[1][2]
                        - aTransA[0][2] * aTransA[1][1]) / det;
     invATransA[2][1] = ( aTransA[0][2] * aTransA[1][0]
                        - aTransA[0][0] * aTransA[1][2]) / det;
     invATransA[2][2] = ( aTransA[0][0] * aTransA[1][1]
                        - aTransA[0][1] * aTransA[1][0]) / det;
 #else
     tjlCompileTimeError;
 #endif

     Vector<RealVect> invATransAdeltaX(stenSize);
     for (int idir = 0; idir < SpaceDim; idir++)
     {
       for (int isten = 0; isten < stenSize; isten++)
       {
         for (int jdir = 0; jdir < SpaceDim; jdir++)
         {
           invATransAdeltaX[isten][idir] = invATransAdeltaX[isten][idir]
                                         + invATransA[idir][jdir] * deltaX[isten][jdir];
         }
       }
     }

     a_stencil.clear();
     a_weight = 0.0;

     for (int isten = 0; isten < stenSize; isten++)
     {
       Real dphidnWeight = 0.0;
       for (int idir = 0; idir < SpaceDim; idir++)
       {
         dphidnWeight -= invATransAdeltaX[isten][idir] * normal[idir];
       }

       a_stencil.add(volSten[isten],dphidnWeight);
       a_weight -= dphidnWeight;
     }
   }
   else
   {
     a_stencil.clear();
     a_weight = 0.0;
   }
 }


 //-----------------------------------------------------------------------
 void DirichletNSviscEBBC::getSecondOrderStencil(VoFStencil&     a_stencil,
                                           Real&           a_weight,
                                           const VolIndex& a_vof,
                                           const EBISBox&  a_ebisBox,
                                           const RealVect& a_dx)
 {
   bool droppedOrder = false;
   IntVect iv = a_vof.gridIndex();

   // Do Johansen-Colella
   RealVect bdCentroid = a_ebisBox.bndryCentroid(a_vof);
   RealVect normal     = a_ebisBox.normal(a_vof);

   // More complicated to prepare for anisotropic case
   RealVect vectDx = a_dx;

   // Find InterpDirection and hiLo
   Real nMax = 0.0;
   int nMaxDir;

   for (int idir = 0; idir < SpaceDim; ++idir)
   {
     if (Abs(normal[idir]) > nMax)
     {
       nMax = Abs(normal[idir]);
       nMaxDir = idir;
     }
   }

   int hiLo = 1;
   if (normal[nMaxDir] < 0.0)
   {
     hiLo = -1;
   }

   // Find Intersection of planes and ray
   RealVect intersection[2];

   // We want the intersection in a coordinate system with dx = dy = dz = 1 and
   // the the cell center of a_vof is the origin

   // We know the planes are defined by nMaxDir = pm (1 or 2)
   intersection[0][nMaxDir] = 1.0*hiLo;
   intersection[1][nMaxDir] = 2.0*hiLo;

   // ScaledCentroid = centroid in an anisotropic coordinate system with
   // the the cell center of a_vof is the origin
   RealVect scaledCentroid;
   for (int idir = 0; idir < SpaceDim; ++idir)
   {
     scaledCentroid[idir] = vectDx[idir]*bdCentroid[idir];
   }

   // Get intersection points(relative to VectDx)
   for (int iPlane = 0; iPlane < 2; ++iPlane)
   {
     for (int idir = 0; idir < SpaceDim; ++idir)
     {
       if (idir != nMaxDir)
       {
         // We get intersection[iPlane][idir] in an anisotropic
         // coordinate system with the cell center of a_vof = origin
         intersection[iPlane][idir] = scaledCentroid[idir]
           + (normal[idir] * ( vectDx[nMaxDir]*intersection[iPlane][nMaxDir]
                             - scaledCentroid[nMaxDir]) / normal[nMaxDir]);

         // We make intersection[iPlane][idir] in a coordinate system with
         // dx = dy = dz = 1 and the the cell center of a_vof is the origin
         intersection[iPlane][idir] /= vectDx[idir];
       }
     }
   }

   // Find interpCen
   IntVect interpCen[2];
   interpCen[0] = iv;
   interpCen[1] = iv;

   // Get ivs for cell containing intersection point
   for (int iPlane = 0; iPlane < 2; ++iPlane)
   {
     for (int idir = 0; idir < SpaceDim; ++idir)
     {
       interpCen[iPlane][idir] += (int)(rint(intersection[iPlane][idir]));
     }
   }

   // planeDir[0] and planeDir[1] are the directions != nMaxDir
   int planeDir[SpaceDim-1];
   planeDir[0] = (nMaxDir + 1) % SpaceDim;
 #if CH_SPACEDIM == 3
   planeDir[1] = (planeDir[0] + 1) % SpaceDim;
 #endif

   // Check do we have enough cells for a stencil
   johanDropOrder(droppedOrder,interpCen,planeDir,a_ebisBox);

   if (!droppedOrder)
   {
     // We have enough cells, so interpolate

     //vofStencil for each point in iPlane
     VoFStencil iPlaneStencil[2];
     for (int iPlane = 0; iPlane < 2; ++iPlane)
     {
       VolIndex colVofs[3];

 #if   CH_SPACEDIM == 3
       VoFStencil colStencil[3];

       for (int iRow = -1; iRow < 2; ++iRow)
       {
         IntVect center = interpCen[iPlane] + iRow*BASISV(planeDir[0]);
         IntVect low    = center - BASISV(planeDir[1]);
         IntVect high   = center + BASISV(planeDir[1]);

         VolIndex centerVof = a_ebisBox.getVoFs(center)[0];
         VolIndex lowVof    = a_ebisBox.getVoFs(low)[0];
         VolIndex highVof   = a_ebisBox.getVoFs(high)[0];

         colVofs[1] = centerVof;
         colVofs[0] = lowVof;
         colVofs[2] = highVof;

         // Distance (relative to dx)
         Real distance = intersection[iPlane][planeDir[1]] -
           (interpCen[iPlane][planeDir[1]] - iv[planeDir[1]]);

         // BuildStencil
         colStencil[iRow+1] = BuildEQIStencil(colVofs,distance);
       }
 #elif CH_SPACEDIM == 2
       IntVect center = interpCen[iPlane];
       IntVect low    = center - BASISV(planeDir[0]);
       IntVect high   = center + BASISV(planeDir[0]);

       VolIndex centerVof = a_ebisBox.getVoFs(center)[0];
       VolIndex lowVof    = a_ebisBox.getVoFs(low)[0];
       VolIndex highVof   = a_ebisBox.getVoFs(high)[0];

       colVofs[1] = centerVof;
       colVofs[0] = lowVof;
       colVofs[2] = highVof;
 #else
       tjlCompileTimeError;
 #endif

       //distance (relative to dx)
       Real distance = intersection[iPlane][planeDir[0]] -
         (interpCen[iPlane][planeDir[0]] - iv[planeDir[0]]);

       // Build stencil
 #if   CH_SPACEDIM == 3
       iPlaneStencil[iPlane] = BuildEQIStencil(colStencil,distance);
 #elif CH_SPACEDIM == 2
       iPlaneStencil[iPlane] = BuildEQIStencil(colVofs,distance);
 #else
       tjlCompileTimeError;
 #endif
     }

     Real d1 = 0.0;
     Real d2 = 0.0;

     for (int idir = 0; idir < SpaceDim; ++idir)
     {
       d1 += (intersection[0][idir]*vectDx[idir] - scaledCentroid[idir])
           * (intersection[0][idir]*vectDx[idir] - scaledCentroid[idir]);

       d2 += (intersection[1][idir]*vectDx[idir]- scaledCentroid[idir])
           * (intersection[1][idir]*vectDx[idir]- scaledCentroid[idir]);
     }

     d1 = sqrt(d1);
     d2 = sqrt(d2);

     // Build stencil
     a_stencil.clear();
     a_stencil = ivofStencil(iPlaneStencil,d1,d2);

     a_weight = 1/(d2-d1) * (d2/d1 - d1/d2);
   }
   else
   {
     getFirstOrderStencil(a_stencil,a_weight,a_vof,a_ebisBox,a_dx);
   }
 }


 //-----------------------------------------------------------------------
 void DirichletNSviscEBBC::johanDropOrder(bool&          a_droppedOrder,
                                    const IntVect  a_interpCen[2],
                                    const int      a_planeDir[SpaceDim-1],
                                    const EBISBox& a_ebisBox)
 {
   // Return a_droppedOrder = true or false;
   a_droppedOrder = false;
   Box region = a_ebisBox.getRegion();

   Vector <VolIndex> interpVof[2];
   VolIndex dummyVof;

   // Check do we have enough cells for a stencil
   for (int iPlane = 0; iPlane < 2; ++iPlane)
   {
     if (!region.contains(a_interpCen[iPlane]))
     {
       a_droppedOrder = true;
       return;
     }

     interpVof[iPlane] = a_ebisBox.getVoFs(a_interpCen[iPlane]);

     if (interpVof[iPlane].size() != 1)
     {
       a_droppedOrder = true;
       return;
     }
     else
     {
       int radius = 1;
       Vector<VolIndex> vofsStencil;
       IntVect timesMoved = IntVect::Zero;
       IntVect pathSign   = IntVect::Zero;
       EBArith::getAllVoFsInMonotonePath(vofsStencil,
                                         timesMoved,
                                         pathSign,
                                         interpVof[iPlane][0],
                                         a_ebisBox,
                                         radius);

       // Recall planeDir[0] and planeDir[1] are the directions != nMaxDir

       // In 2D check if 2 neighboring cells are in vofsStencil
       // In 3D check if all cells (!= interpCen) in a 3X3 neighborhood are in
       //                vofsStencil
       for (int iRow = -1; iRow < 2; ++iRow)
       {
         int numCols = SpaceDim - 2;
         for (int iColumn = -numCols; iColumn < numCols + 1; ++iColumn)
         {
           IntVect cell = a_interpCen[iPlane] + iRow * BASISV(a_planeDir[0]);
 #if CH_SPACEDIM == 3
           cell += iColumn * BASISV(a_planeDir[1]);
 #endif
           if (cell != a_interpCen[iPlane])
           {
             bool isVof = EBArith::isVoFHere(dummyVof,vofsStencil,cell);
             if (!isVof)
             {
               a_droppedOrder = true;
               return;
             }
           }
         }
       }
     }
   }

   return;
 }


 //-----------------------------------------------------------------------
 VoFStencil DirichletNSviscEBBC::BuildEQIStencil(const VolIndex a_vofs[3],
                                           const Real&    a_distance)
 {
   VoFStencil retval;

   Real x  = a_distance;
   Real x2 = x * x;

   retval.add(a_vofs[0], (x2 - x) / 2.0);
   retval.add(a_vofs[1],-(x2 - 1)      );
   retval.add(a_vofs[2], (x2 + x) / 2.0);

   return retval;
 }


 //-----------------------------------------------------------------------
 VoFStencil DirichletNSviscEBBC::BuildEQIStencil(VoFStencil a_vofs[3],
                                           const Real& a_distance)
 {
   VoFStencil retval;

   Real x  = a_distance;
   Real x2 = x * x;

   a_vofs[0] *=  (x2 - x) / 2.0;
   retval += a_vofs[0];

   a_vofs[1] *= -(x2 - 1);
   retval += a_vofs[1];

   a_vofs[2] *=  (x2 + x) / 2.0;
   retval += a_vofs[2];

   return retval;
 }


 //-----------------------------------------------------------------------
 VoFStencil DirichletNSviscEBBC::ivofStencil(const VoFStencil a_iPlaneStencil[2],
                                       const Real&      a_distance1,
                                       const Real&      a_distance2)
 {
   VoFStencil retvalTemp;
   VoFStencil retval;

   Real d1 = a_distance1;
   Real d2 = a_distance2;

   retvalTemp = a_iPlaneStencil[0];
   retvalTemp *= 1/(d2-d1) * (-d2/d1);
   retval += retvalTemp;

   retvalTemp = a_iPlaneStencil[1];
   retvalTemp *= 1/(d2-d1) * (d1/d2);
   retval += retvalTemp;

   return retval;
 }


 //-----------------------------------------------------------------------
 void DirichletNSviscEBBC::get_the_flux( const RealVect& nv, const Vector<Real>& dns, 
                                         const RealVect& point, const Real mu, 
                                         const Real la, Vector<Real>& a_flux,
 					bool a_useHomogeneous, const Real a_time )
 {
    RealVect tv, bv;
    get_tanvects( nv, tv, bv );
    Real dutdt=0, dubdb=0, dundt=0, dundb=0;
    if( m_isFunction && !a_useHomogeneous )
    {
       Vector<Real> ux(SpaceDim), uy(SpaceDim), uz(SpaceDim);
       ux = m_func->get_xderivatives( point, a_time );
       uy = m_func->get_yderivatives( point, a_time );
       if( SpaceDim == 2 )
       {
          dutdt = tv[0]*(tv[0]*ux[0]+tv[1]*ux[1]) + tv[1]*(tv[0]*uy[0]+tv[1]*uy[1]);
          dubdb = bv[0]*(bv[0]*ux[0]+bv[1]*ux[1]) + bv[1]*(bv[0]*uy[0]+bv[1]*uy[1]);
          dundt = tv[0]*(nv[0]*ux[0]+nv[1]*ux[1]) + tv[1]*(nv[0]*uy[0]+nv[1]*uy[1]);
          dundb = bv[0]*(nv[0]*ux[0]+nv[1]*ux[1]) + bv[1]*(nv[0]*uy[0]+nv[1]*uy[1]);

 	 //         cout << dns[0]+nv[0]*ux[0]+nv[1]*uy[0] << endl;
       }
       else if( SpaceDim == 3 )
       {
         uz = m_func->get_zderivatives( point, a_time );
         dutdt = tv[0]*(tv[0]*ux[0]+tv[1]*ux[1]+tv[2]*ux[2]) + 
                 tv[1]*(tv[0]*uy[0]+tv[1]*uy[1]+tv[2]*uy[2]) +
                 tv[2]*(tv[0]*uz[0]+tv[1]*uz[1]+tv[2]*uz[2]);
         dubdb = bv[0]*(bv[0]*ux[0]+bv[1]*ux[1]+bv[2]*ux[2]) + 
                 bv[1]*(bv[0]*uy[0]+bv[1]*uy[1]+bv[2]*uy[2])+
                 bv[2]*(bv[0]*uz[0]+bv[1]*uz[1]+bv[2]*uz[2]);
         dundt = tv[0]*(nv[0]*ux[0]+nv[1]*ux[1]+nv[2]*ux[2]) + 
                 tv[1]*(nv[0]*uy[0]+nv[1]*uy[1]+nv[2]*uy[2])+
                 tv[2]*(nv[0]*uz[0]+nv[1]*uz[1]+nv[2]*uz[2]);
         dundb = bv[0]*(nv[0]*ux[0]+nv[1]*ux[1]+nv[2]*ux[2]) + 
                 bv[1]*(nv[0]*uy[0]+nv[1]*uy[1]+nv[2]*uy[2])+
                 bv[2]*(nv[0]*uz[0]+nv[1]*uz[1]+nv[2]*uz[2]);
      }
   }
   Real dundn = 0;
   for( int c=0; c<SpaceDim ;c++)
      dundn += dns[c]*nv[c];

   for( int c=0; c<SpaceDim ;c++)
   {
     Real tt = la*(dutdt+dubdb)*nv[c] + mu*dundt*tv[c] + mu*dundb*bv[c];
     a_flux[c] = (mu+la)*dundn*nv[c]+mu*dns[c]+tt;
   }
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::get_the_flux_rz( const RealVect& nv, const Vector<Real>& dns, 
                                           const Real& drn, const RealVect& point, 
                                           const Real mu, const Real la, 
					   Vector<Real>& a_flux_div, 
                                           Vector<Real>& a_flux_undiv, 
                                           Real rb, Real uval, bool a_useHomogeneous,
					   const Real a_time )
{
   RealVect tv, bv;
   get_tanvects( nv, tv, bv );
   Real dudt=0, dvdt=0, drt=0;

   if( m_isFunction && !a_useHomogeneous )
   {
      Vector<Real> ux(SpaceDim), uy(SpaceDim);
      ux = m_func->get_xderivatives( point, a_time );
      uy = m_func->get_yderivatives( point, a_time );
      if( SpaceDim == 2 )
      {
	 dudt = tv[0]*ux[0]+tv[1]*uy[0];
         dvdt = tv[0]*ux[1]+tv[1]*uy[1];
         drt  = dudt*rb - nv[1]*uval;
	 //	 cout <<  dns[0]-( nv[0]*ux[0]+nv[1]*uy[0]) << endl;
	 //	 cout <<  dns[1]-( nv[0]*ux[1]+nv[1]*uy[1) << endl;
	 //         cout <<  drn-  (rb*(nv[0]*ux[0]+nv[1]*uy[0])+nv[0]*uval) << endl;
      }
   }

   a_flux_div[0] = 2*mu*rb*nv[0]*(nv[0]*dns[0]-nv[1]*dudt);
   a_flux_div[1] = mu*rb*(nv[0]*nv[1]*dns[0]+nv[0]*nv[0]*dudt+
			  nv[0]*nv[0]*dns[1]-nv[0]*nv[1]*dvdt);
   a_flux_undiv[0] = la/rb*(nv[0]*nv[0]*drn-nv[0]*nv[1]*drt) + 
     la*(nv[0]*nv[1]*dns[1]+nv[0]*nv[0]*dvdt) + 
     mu*(nv[1]*nv[1]*dns[0]+nv[0]*nv[1]*(dudt+dns[1])-nv[1]*nv[1]*dvdt);
   a_flux_undiv[1] = (2*mu+la)*(nv[1]*nv[1]*dns[1]+nv[0]*nv[1]*dvdt) +
     la/rb*nv[1]*(nv[0]*drn-nv[1]*drt);
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBC::get_tanvects( const RealVect nv, RealVect& tv, RealVect& bv )
{
   if( SpaceDim == 2 )
   {
     tv[0] = -nv[1];
     tv[1] =  nv[0];
     bv[0] = 0;
     bv[1] = 0;
     
   }
   else if( SpaceDim == 3 )
   {
     if( fabs(nv[0]) > fabs(nv[2]) || fabs(nv[1]) > fabs(nv[2]))
     {
        Real n2i = 1/sqrt( nv[1]*nv[1]+nv[0]*nv[0]);
	tv[0] = -nv[1]*n2i;
	tv[1] =  nv[0]*n2i;
	tv[2] =  0;
     }
     else
     {
        Real n2i = 1/sqrt( nv[2]*nv[2]+nv[0]*nv[0]);
	tv[0] = -nv[2]*n2i;
	tv[1] =  0;
	tv[2] =  nv[0]*n2i;
     }       
     bv[0] = nv[1]*tv[2]-nv[2]*tv[1];
     bv[1] = nv[2]*tv[0]-nv[0]*tv[2];
     bv[2] = nv[0]*tv[1]-nv[1]*tv[0];
   }
}


//-----------------------------------------------------------------------
DirichletNSviscEBBCFactory::DirichletNSviscEBBCFactory()
{
  m_ncomp = SpaceDim;
  m_value.resize(m_ncomp);

  for( int c = 0 ; c<m_ncomp ;c++ )
     m_value[c] = 12345.6789;

  m_func = NULL;
  m_order = 1;
  m_onlyHomogeneous = true;
  m_isFunction = false;
}


//-----------------------------------------------------------------------
DirichletNSviscEBBCFactory::~DirichletNSviscEBBCFactory()
{
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBCFactory::setOrder(int a_order)
{
  CH_assert(a_order >= 1 && a_order <= 2);
  m_order = a_order;
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBCFactory::setValue(Vector<Real> a_value)
{
  m_value = a_value;
  m_func = NULL;

  m_onlyHomogeneous = false;
  m_isFunction = false;
}


//-----------------------------------------------------------------------
void DirichletNSviscEBBCFactory::setFunction(RefCountedPtr<BaseBCVectValue> a_func)
{
  for( int c = 0 ; c<m_ncomp ;c++ )
     m_value[c] = 12345.6789;
  m_func = a_func;
  m_onlyHomogeneous = false;
  m_isFunction = true;
}


//-----------------------------------------------------------------------
DirichletNSviscEBBC* DirichletNSviscEBBCFactory::create(const ProblemDomain& a_domain,
                                            const EBISLayout&    a_layout,
                                            const RealVect&      a_dx)
{
  DirichletNSviscEBBC* fresh = new DirichletNSviscEBBC(a_domain,a_layout,a_dx);

  fresh->setOrder(m_order);

  if (!m_onlyHomogeneous)
  {
    if (!m_isFunction)
    {
      fresh->setValue(m_value);
    }
    else
    {
      fresh->setFunction(m_func);
    }
  }

  return fresh;
}
