/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _EBAMRHELMHOLTZOP_H_
#define _EBAMRHELMHOLTZOP_H_

#include "REAL.H"
#include "Box.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "RefCountedPtr.H"

#include "AMRMultiGrid.H"

#include "EBIndexSpace.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"

#include "EBLevelDataOps.H"
//#include "BaseEBBC.H"
//#include "BaseDomainBC.H"
#include "HelmholtzDomainBC.H"
#include "HelmholtzEBBC.H"

#include "CFIVS.H"
#include "EBFluxRegister.H"
#include "EBMGAverage.H"
#include "EBMGInterp.H"

/// Helmholtz operator, inherited from AMRLevelOp
/**
  The class represents the operator
 \f[
     H(u) =  {\rm div}(\kappa\nabla u )-k^2\rho u
 \f]
  on one grid level, and can be use to solve the equation \f$ H(u) = f \f$, for 
  example by the class AMRMultiGrid. The coefficient \f$ \kappa \f$ can currently
  only be given a constant value, although represented internaly as a function.
  The operator can be used in both two and three space dimensions, in Cartesian
  coordinates. The class also supports cylindrical coordinates, \f$ (r,z)\f$, in 
  two dimensions. \f$ \rho \f$ is a given function of \f$ (x,y,z)\f$ ( or \f$ (r,z)\f$ in 
  the cylindrical case), and \f$ k^2 \f$ is a given positive constant.
 */
class EBAMRHelmholtzOp : public AMRLevelOp<LevelData<EBCellFAB> >
{
public:
  /// Destructor.
  ~EBAMRHelmholtzOp();

  /// Constructor.
  EBAMRHelmholtzOp();

  /** a_residual = a_rhs - L(a_phiFine, a_phi)   no coaser AMR level*/
  void AMRResidualNC(LevelData<EBCellFAB>&       a_residual,
                     const LevelData<EBCellFAB>& a_phiFine,
                     const LevelData<EBCellFAB>& a_phi,
                     const LevelData<EBCellFAB>& a_rhs,
                     bool a_homogeneousBC);

  /// Constructor, with all details.
  /** The class represents one grid level, but also needs to know about the 
      grid levels one refinement up and one refinement down. This is why most of
      the input data, should be given for three grid levels.
      
     Input: 
        - a_dblFine  Fine level disjoint box layout.
        - a_dbl      This level disjoint box layout.
        - a_dblCoar  Coarse level disjoint box layout.
	- a_ebislFine Fine level EBISLayout.
	- a_ebisl Fine This level EBISLayout.
	- a_ebislCoar Coarse level EBISLayout.
        - a_domainBCFine Pointer to domain boundary condition class on finer level.
        - a_domainBC Pointer to domain boundary condition class on this level.
        - a_domainBCCoar Pointer to domain boundary condition class on coarser level.
        - a_ebBCFine Pointer to embedded boundary condition class on finer level.
        - a_ebBC Pointer to embedded boundary condition class on this level.
        - a_ebBCCoar Pointer to embedded boundary condition class on coarser level.
	- a_dxFine Grid spacing at finer level.
	- a_dx     Grid spacing at this level.
	- a_dxCoar Grid spacing at coarser level.
	- a_domain The global domain 
        - a_origin Origin of coordinate systems (same on all levels).
        - a_k2  Constant \f$ k^2\f$ in operator.
        - a_refToFine
	- a_refToCoar
        - a_hasFiner
        - a_hasCoarser
	- a_orderEB
        - a_kappa dissipation parameter \f$ \kappa \f$ in operator.
        - a_rho function \f$ \rho \f$ in operator.
        - a_rz set to true for cylindrical coordinates.
   */  
  EBAMRHelmholtzOp(const DisjointBoxLayout&            a_dblFine,
                 const DisjointBoxLayout&            a_dbl,
                 const DisjointBoxLayout&            a_dblCoar,
                 const EBISLayout&                   a_ebislFine,
                 const EBISLayout&                   a_ebisl,
                 const EBISLayout&                   a_ebislCoar,
		   //                 const RefCountedPtr<BaseDomainBC>&  a_domainBCFine,
		   //                 const RefCountedPtr<BaseDomainBC>&  a_domainBC,
		   //                 const RefCountedPtr<BaseDomainBC>&  a_domainBCCoar,
		   //                 const RefCountedPtr<BaseEBBC>&      a_ebBCFine,
		   //                 const RefCountedPtr<BaseEBBC>&      a_ebBC,
		   //                 const RefCountedPtr<BaseEBBC>&      a_ebBCCoar,
                 const RefCountedPtr<HelmholtzDomainBC>&  a_domainBCFine,
                 const RefCountedPtr<HelmholtzDomainBC>&  a_domainBC,
                 const RefCountedPtr<HelmholtzDomainBC>&  a_domainBCCoar,
                 const RefCountedPtr<HelmholtzEBBC>&      a_ebBCFine,
                 const RefCountedPtr<HelmholtzEBBC>&      a_ebBC,
                 const RefCountedPtr<HelmholtzEBBC>&      a_ebBCCoar,
                 const RealVect&                     a_dxFine,
                 const RealVect&                     a_dx,
                 const RealVect&                     a_dxCoar,
                 const ProblemDomain&                a_domain,
                 const RealVect&                     a_origin,
	         const Real                          a_k2,
                 const int&                          a_refToFine,
                 const int&                          a_refToCoar,
                 const bool&                         a_hasFiner,
                 const bool&                         a_hasCoarser,
                 const int&                          a_relaxType,
                 const int&                          a_orderEB,
 	         const Real                          a_kappa,
                 const LevelData<EBCellFAB>*         a_rho,
	         const bool&                         a_rz,
		 const Real&                         a_time );

  //MGOp operations.  no finer or coarser

  ///
  /**
   */
  virtual void residual(LevelData<EBCellFAB>&       a_residual,
                        const LevelData<EBCellFAB>& a_phi,
                        const LevelData<EBCellFAB>& a_rhs,
                        bool                        a_homogeneous=false);

  ///
  /**
   */
  virtual void preCond(LevelData<EBCellFAB>&       a_opPhi,
                       const LevelData<EBCellFAB>& a_phi);

  ///
  /**
     This function assumes that coarse-fine boundary conditions have
     been dealt with.
   */
  virtual void applyOp(LevelData<EBCellFAB>&             a_opPhi,
                       const LevelData<EBCellFAB>&       a_phi,
                       const LevelData<EBCellFAB>* const a_phiCoarse,
                       bool                              a_homogeneousPhysBC,
                       bool                              a_homogeneousCFBC);

  ///
  /**
     get EB flux from leveldata if we have one.  otherwise use m_ebBC
     ignore input data in the case of homogeneous Phys BC
  */
  virtual void
  applyOp(LevelData<EBCellFAB>&                    a_opPhi,
          const LevelData<EBCellFAB>&              a_phi,
          const LevelData<EBCellFAB>* const        a_phiCoar,
          bool                                     a_homogeneousPhysBC,
          bool                                     a_homogeneousCFBC,
          const LevelData<BaseIVFAB<Real> >* const a_ebFluxBCLD);

  ///
  /**
     this is the linearop function.  CFBC is set to homogeneous.  phic is null
   */
  virtual void applyOp(LevelData<EBCellFAB>&             a_opPhi,
                       const LevelData<EBCellFAB>&       a_phi,
                       bool                              a_homogeneousPhysBC);

  ///
  /**
   */
  virtual void create(LevelData<EBCellFAB>&       a_lhs,
                      const LevelData<EBCellFAB>& a_rhs);

  ///
  virtual void createCoarsened(LevelData<EBCellFAB>&       a_lhs,
                               const LevelData<EBCellFAB>& a_rhs,
                               const int&                  a_refRat);

  Real
  AMRNorm(const LevelData<EBCellFAB>& a_coarResid,
          const LevelData<EBCellFAB>& a_fineResid,
          const int& a_refRat,
          const int& a_ord);
  ///
  /**
   */
  virtual void assign(LevelData<EBCellFAB>&       a_lhs,
                      const LevelData<EBCellFAB>& a_rhs);

  ///
  /**
   */
  virtual Real dotProduct(const LevelData<EBCellFAB>& a_1,
                          const LevelData<EBCellFAB>& a_2);

  ///
  /**
   */
  virtual void incr(LevelData<EBCellFAB>&       a_lhs,
                    const LevelData<EBCellFAB>& a_x,
                    Real                        a_scale);

  ///
  /**
   */
  virtual void axby(LevelData<EBCellFAB>&       a_lhs,
                    const LevelData<EBCellFAB>& a_x,
                    const LevelData<EBCellFAB>& a_y,
                    Real                        a_a,
                    Real                        a_b);

  ///
  /**
   */
  virtual void scale(LevelData<EBCellFAB>& a_lhs,
                     const Real&           a_scale);

  ///
  /**
   */
  virtual Real norm(const LevelData<EBCellFAB>& a_rhs,
                    int                         a_ord);

  ///
  /**
   */
  virtual void setToZero(LevelData<EBCellFAB>& a_lhs);

  ///
  /**
   */
  virtual void createCoarser(LevelData<EBCellFAB>&       a_coarse,
                             const LevelData<EBCellFAB>& a_fine,
                             bool                        a_ghosted);

  ///
  /**
   */
  virtual void relax(LevelData<EBCellFAB>&       a_e,
                     const LevelData<EBCellFAB>& a_residual,
                     int                         a_iterations);

  ///
  /**
     Calculate restricted residual:
       a_resCoarse[2h] = I[h->2h] (a_rhsFine[h] - L[h](a_phiFine[h]))
  */
  virtual void restrictResidual(LevelData<EBCellFAB>&       a_resCoarse,
                                LevelData<EBCellFAB>&       a_phiFine,
                                const LevelData<EBCellFAB>& a_rhsFine);

  ///
  /**
     Correct the fine solution based on coarse correction:
       a_phiThisLevel += I[2h->h] (a_correctCoarse)
  */
  virtual void prolongIncrement(LevelData<EBCellFAB>&       a_phiThisLevel,
                                const LevelData<EBCellFAB>& a_correctCoarse);

  ///
  /** Refinement ratio between this level and coarser level.
      Returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToCoarser();

  ///
  /** Refinement ratio between this level and coarser level.
      Returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToFiner();

  ///
  /** a_residual = a_rhs - L(a_phi, a_phiFine, a_phiCoarse) */
  virtual void AMRResidual(LevelData<EBCellFAB>& a_residual,
                           const LevelData<EBCellFAB>& a_phiFine,
                           const LevelData<EBCellFAB>& a_phi,
                           const LevelData<EBCellFAB>& a_phiCoarse,
                           const LevelData<EBCellFAB>& a_rhs,
                           bool a_homogeneousBC) ;

  ///
  /** a_residual = a_rhs - L(a_phi, a_phiCoarse)  */
  virtual void AMRResidualNF(LevelData<EBCellFAB>& a_residual,
                             const LevelData<EBCellFAB>& a_phi,
                             const LevelData<EBCellFAB>& a_phiCoarse,
                             const LevelData<EBCellFAB>& a_rhs,
                             bool a_homogeneousBC);

  ///
  /** a_resCoarse = I[h-2h] (a_residual - L(a_correction, a_coarseCorrection)) */
  virtual void AMRRestrict(LevelData<EBCellFAB>& a_resCoarse,
                           const LevelData<EBCellFAB>& a_residual,
                           const LevelData<EBCellFAB>& a_correction,
                           const LevelData<EBCellFAB>& a_coarseCorrection);

  ///
  /** a_correction += I[2h->h](a_coarseCorrection) */
  virtual void AMRProlong(LevelData<EBCellFAB>&       a_correction,
                          const LevelData<EBCellFAB>& a_coarseCorrection);

  ///
  /** a_residual = a_residual - L(a_correction, a_coarseCorrection) */
  virtual void AMRUpdateResidual(LevelData<EBCellFAB>&       a_residual,
                                 const LevelData<EBCellFAB>& a_correction,
                                 const LevelData<EBCellFAB>& a_coarseCorrection);

  /// If this level has a finer level, interpolate fluxes from the fine level
  void reflux(LevelData<EBCellFAB>& a_residual,
              const LevelData<EBCellFAB>& a_phiFine,
              const LevelData<EBCellFAB>& a_phi);

  void set_homogeneous_relax();

  void set_inhomogeneous_relax();

  Real get_diagonal_weight( const DataIndex& dex, const VolIndex& vof ) const;

  void set_rho( const LevelData<EBCellFAB>& a_rho );

  void set_rho_zero( );

  void set_k2( Real a_k2 );

  void set_time( Real a_time );

protected:

  void defineStencils();

  DisjointBoxLayout               m_dblFine;
  DisjointBoxLayout               m_dbl;
  DisjointBoxLayout               m_dblCoar;
  DisjointBoxLayout               m_dblCoarsenedFine;

  EBISLayout                      m_ebislFine;
  EBISLayout                      m_ebisl;
  EBISLayout                      m_ebislCoar;
  EBISLayout                      m_ebislCoarsenedFine;

  ProblemDomain                   m_domainFine;
  ProblemDomain                   m_domain;
  ProblemDomain                   m_domainCoar;

  //  RefCountedPtr<BaseDomainBC>     m_domainBCFine;
  //  RefCountedPtr<BaseDomainBC>     m_domainBC;
  //  RefCountedPtr<BaseDomainBC>     m_domainBCCoar;
  //  RefCountedPtr<BaseEBBC>         m_ebBCFine;
  //  RefCountedPtr<BaseEBBC>         m_ebBC;
  //  RefCountedPtr<BaseEBBC>         m_ebBCCoar;

  RefCountedPtr<HelmholtzDomainBC>     m_domainBCFine;
  RefCountedPtr<HelmholtzDomainBC>     m_domainBC;
  RefCountedPtr<HelmholtzDomainBC>     m_domainBCCoar;
  RefCountedPtr<HelmholtzEBBC>         m_ebBCFine;
  RefCountedPtr<HelmholtzEBBC>         m_ebBC;
  RefCountedPtr<HelmholtzEBBC>         m_ebBCCoar;

  RealVect                        m_dxFine;
  RealVect                        m_dx;
  RealVect                        m_invDx;
  RealVect                        m_invDx2;
  Real                            m_dxScale;
  RealVect                        m_dxCoar;
  RealVect                        m_origin;
  Real                            m_k2;

  int                             m_refToFine;
  int                             m_refToCoar;
  bool                            m_hasFine;
  bool                            m_hasInterpAve;
  bool                            m_hasCoar;
  int                             m_orderEB;
  int                             m_relaxType;
  bool                            m_split_u;
  Real                            m_time;

  //restriction object
  EBMGAverage                    m_ebAverage;
  //prolongation object
  EBMGInterp                     m_ebInterp;

  //stencils for operator evaluation
  LayoutData<BaseIVFAB<Real> >       m_weight;
  LayoutData<BaseIVFAB<VoFStencil> > m_opStencil;
  LayoutData<BaseIVFAB<Real> >       m_uweight;
  LayoutData<BaseIVFAB<VoFStencil> > m_uStencil;

  LayoutData<BaseIFFAB<FaceStencil> > m_iflux_stencil;
  LayoutData<BaseIFFAB<FaceStencil> > m_jflux_stencil;
  LayoutData<BaseIFFAB<FaceStencil> > m_kflux_stencil;

  // Coarse-fine stencils for homogeneous CFInterp
  LayoutData<CFIVS> m_loCFIVS[SpaceDim];
  LayoutData<CFIVS> m_hiCFIVS[SpaceDim];

  bool m_rz;  
  LevelData<EBCellFAB> m_rho;
  LevelData<EBCellFAB> m_kappa;
  LevelData<EBCellFAB> m_kappaFine;
  LevelData<EBCellFAB> m_wgh;

  //flux register with finer level
  EBFluxRegister m_fluxReg;

  //special mg objects for when we do not have
  //a coarser level or when the refinement to coarse
  //is greater than two
  //flag for when we need special MG objects
  bool                        m_hasMGObjects;
  //stuff below is only defined if m_hasMGObjects==true
  EBMGAverage                 m_ebAverageMG;
  EBMGInterp                  m_ebInterpMG;
  DisjointBoxLayout           m_dblCoarMG;
  EBISLayout                  m_ebislCoarMG;
  Box                         m_domainCoarMG;
  bool                        m_homogeneous_relax;

private:
  //internally useful but not for general consumption
  //lots of hidden assumptions and all that

  void incrementFRCoar(EBFluxRegister&             a_fluxReg,
                       const LevelData<EBCellFAB>& a_phiFine,
                       const LevelData<EBCellFAB>& a_phi);

  void incrementFRFine(EBFluxRegister&             a_fluxReg,
                       const LevelData<EBCellFAB>& a_phiFine,
                       const LevelData<EBCellFAB>& a_phi);

  void incrementFineBothRZ(  EBFluxRegister& a_fluxReg,
			     const EBFaceFAB&      a_fineFlux,
                             const Real&           a_scale,
                             const DataIndex&      a_fineDatInd,
                             const Interval&       a_variables,
                             const int&            a_dir,
                             const Side::LoHiSide& a_sd,
                             const Real&           a_dx );

  void incrementCoarseBothRZ(  EBFluxRegister& a_fluxReg,
			     const EBFaceFAB&      a_CoarFlux,
                             const Real&           a_scale,
                             const DataIndex&      a_CoarDatInd,
                             const Interval&       a_variables,
                             const int&            a_dir,
 		             const Real&           a_dx );


  void getFlux(EBFaceFAB&       a_flux,
               const EBCellFAB& a_phi,
               const Box&       a_box,
               const EBISBox&   a_ebisBox,
               const RealVect&  a_dx,
               const int&       a_idir,
	       const EBCellFAB& a_kappa );

  void applyCFBCs(LevelData<EBCellFAB>&             a_phi,
                  const LevelData<EBCellFAB>* const a_phiCoarse,
                  bool a_homogeneousCFBC);

  void getOpVoFStencil(VoFStencil&     a_stencil,
                       const EBISBox&  a_ebisbox,
                       const VolIndex& a_vof);

  void getOpFaceStencil(VoFStencil&             a_stencil,
                        const Vector<VolIndex>& a_allMonotoneVofs,
                        const EBISBox&          a_ebisbox,
                        const VolIndex&         a_vof,
                        int                     a_dir,
                        const Side::LoHiSide&   a_side,
                        const FaceIndex&        a_face);

  void get_ustencil( VoFStencil&     a_stencil,
                     Real&           a_weight,
                     const EBISBox&  a_ebisbox,
                     const VolIndex& a_vof);

  void levelMulticolorGS(LevelData<EBCellFAB>&       a_phi,
                         const LevelData<EBCellFAB>& a_rhs);

  void levelJacobi(LevelData<EBCellFAB>&       a_phi,
                   const LevelData<EBCellFAB>& a_rhs);

  void applyHomogeneousCFBCs(LevelData<EBCellFAB>&   a_phi);

  void applyHomogeneousCFBCs(EBCellFAB&            a_phi,
                             const DataIndex&      a_datInd,
                             int                   a_idir,
                             Side::LoHiSide        a_hiorlo);

  void update_irregular( EBCellFAB& a_res,
	                 const EBCellFAB& a_u,
			 const EBISBox& a_eb,
			 const BaseIFFAB<FaceStencil>& a_flux_stencil,
	                 const EBCellFAB& a_kappa,
			 const EBCellFAB& a_rho, int a_dir );

  void define_flux_stencils();

  void define_flux_stencil( int dir, BaseIFFAB<FaceStencil>& flux_stencil,
			    const EBISBox& eb, const Box& ubox );

  void compute_diagonal_weight();

  void compute_dw_irregular(	    EBCellFAB& a_wgh,
				    const EBISBox& a_eb,
				    const BaseIFFAB<FaceStencil>& a_flux_stencil,
				    const EBCellFAB& a_kappa,
				    int a_dir );

  void print_test_info(  const EBISBox& a_eb );
  
  void print_ebcellfab(  const EBCellFAB& a_u, const char* name );



private:
  //copy constructor and operator= disallowed for all the usual reasons
  EBAMRHelmholtzOp(const EBAMRHelmholtzOp& a_opin)
  {
    MayDay::Error("invalid operator");
  }

  void operator=(const EBAMRHelmholtzOp& a_opin)
  {
    MayDay::Error("invalid operator");
  }
};

/// Class that creates EBAMRHelmholtz operators
/**
 
 */
class EBAMRHelmholtzOpFactory: public AMRLevelOpFactory<LevelData<EBCellFAB> >
{
public:
  ///
  virtual ~EBAMRHelmholtzOpFactory();

  ///
  EBAMRHelmholtzOpFactory(const Vector<DisjointBoxLayout>&   a_grids,
                        const Vector<EBISLayout>&          a_ebisls,
                        const Vector<int>&                 a_refRatio,
                        const ProblemDomain&               a_domainCoarse,
                        const RealVect&                    a_dxCoarse,
                        const RealVect&                    a_origin,
		        const Real                         a_k2,
                        const int&                         a_orderEB,
                        const int&                         a_relaxType,
			  //                        RefCountedPtr<BaseDomainBCFactory> a_domainBCFactory,
			  //		        RefCountedPtr<BaseEBBCFactory>     a_ebBcFactory,
                        RefCountedPtr<HelmholtzDomainBCFactory> a_domainBCFactory,
		        RefCountedPtr<HelmholtzEBBCFactory>     a_ebBcFactory,
			  
                        const Real                         a_kappa,
	                const Vector<LevelData<EBCellFAB>*>& a_rho,
			const bool&                        a_rz );

  ///
  virtual EBAMRHelmholtzOp*
  MGnewOp(const ProblemDomain& a_FineindexSpace,
          int                  a_depth,
          bool                 a_homoOnly = true);

  EBAMRHelmholtzOp* createOperator(const DisjointBoxLayout& a_dblMGLevel,
                                 const EBISLayout&        a_ebislMGLevel,
                                 const ProblemDomain&     a_domainMGlevel,
                                 const RealVect&          a_dxMGLevel,
                                 const int&               a_whichLevel);
  ///
  virtual void reclaim(MGLevelOp<LevelData<EBCellFAB> >* a_reclaim);

  ///
  virtual EBAMRHelmholtzOp*
  AMRnewOp(const ProblemDomain& a_FineindexSpace);

  ///
  virtual void AMRreclaim(EBAMRHelmholtzOp* a_reclaim);

  ///
  /** Refinement ratio between this level and coarser level.
      Returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToFiner(const ProblemDomain& a_domain) const;

  void set_k2( Real a_k2, int a_level );

  void set_rho( LevelData<EBCellFAB>& a_rho, int a_level );
  
  void set_time( Real a_time );

  void dump_level( int a_level );

protected:
  void interp_frfine( LevelData<EBCellFAB>& a_uc,
		      const LevelData<EBCellFAB>& a_uf,
		      const Box& a_domainc,
		      const EBISLayout& a_ec,
		      const EBISLayout& a_ef,
		      const int a_rat );

  bool     m_isDefined;
  int      m_orderEB;
  int      m_relaxType;
  int      m_numLevels;
  Vector<Real>   m_k2;
  Vector< LevelData<EBCellFAB>*> m_rhoPtr;
  Real     m_kappa;
  bool     m_rz;
  Real     m_time;

  Vector<ProblemDomain>     m_domainVec;
  Vector<DisjointBoxLayout> m_gridVec;
  Vector<EBISLayout>        m_ebislVec;
  Vector<int>               m_refRatioVec;
  Vector<RealVect>          m_dxVec;
  RealVect                  m_origin;

  //  RefCountedPtr<BaseDomainBCFactory>   m_domainBCFactory;
  //  RefCountedPtr<BaseEBBCFactory>       m_ebBCFactory;
  RefCountedPtr<HelmholtzDomainBCFactory>   m_domainBCFactory;
  RefCountedPtr<HelmholtzEBBCFactory>       m_ebBCFactory;

private:
  ///weak construction bad
  EBAMRHelmholtzOpFactory()
  {
    MayDay::Error("invalid operator");
  }

  //copy constructor and operator= disallowed for all the usual reasons
  EBAMRHelmholtzOpFactory(const EBAMRHelmholtzOpFactory& a_opin)
  {
    MayDay::Error("invalid operator");
  }

  void operator=(const EBAMRHelmholtzOpFactory& a_opin)
  {
    MayDay::Error("invalid operator");
  }
};

  /*
    coarsen the dbl if we can
    otherwise,
    check to see if the domain is covered. if so
    make one big grid.  otherwise we fail
    returns false if not coarsenable
  */
extern bool
getCoarserLayouts(DisjointBoxLayout&       a_dblCoar,
                  EBISLayout&              a_ebislCoar,
                  Box&                     a_domainCoar,
                  const DisjointBoxLayout& a_dblFine,
                  const Box&               a_domainFine,
                  int                      a_refToCoar,
                  const EBIndexSpace*      ebisPtr);
#endif
