/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//


#include "LoadBalance.H"
#include "EBArith.H"

#include "EBAMRHelmholtzOp.H"
#include "EBEllipticLoHiCenter.H"
#include "EBQuadCFInterp.H"
#include "EBFluxRegister.H"

#include "EBAMRPoissonOpF_F.H"
#include "EBAMRHelmholtzOpF_F.H"
#include "EBCoarseAverage.H"

#include "ArrayToRaw.H"

//-----------------------------------------------------------------------
EBAMRHelmholtzOp::~EBAMRHelmholtzOp()
{
}


//-----------------------------------------------------------------------
EBAMRHelmholtzOp::EBAMRHelmholtzOp()
{
}

//-----------------------------------------------------------------------
EBAMRHelmholtzOp::EBAMRHelmholtzOp(
	       const DisjointBoxLayout&            a_dblFine,
               const DisjointBoxLayout&            a_dbl,
               const DisjointBoxLayout&            a_dblCoar,
               const EBISLayout&                   a_ebislFine,
               const EBISLayout&                   a_ebisl,
               const EBISLayout&                   a_ebislCoar,
	       //               const RefCountedPtr<BaseDomainBC>&  a_domainBCFine,
	       //               const RefCountedPtr<BaseDomainBC>&  a_domainBC,
	       //               const RefCountedPtr<BaseDomainBC>&  a_domainBCCoar,
	       //               const RefCountedPtr<BaseEBBC>&      a_ebBCFine,
	       //               const RefCountedPtr<BaseEBBC>&      a_ebBC,
	       //               const RefCountedPtr<BaseEBBC>&      a_ebBCCoar,
               const RefCountedPtr<HelmholtzDomainBC>&  a_domainBCFine,
               const RefCountedPtr<HelmholtzDomainBC>&  a_domainBC,
               const RefCountedPtr<HelmholtzDomainBC>&  a_domainBCCoar,
               const RefCountedPtr<HelmholtzEBBC>&      a_ebBCFine,
               const RefCountedPtr<HelmholtzEBBC>&      a_ebBC,
               const RefCountedPtr<HelmholtzEBBC>&      a_ebBCCoar,
               const RealVect&                     a_dxFine,
               const RealVect&                     a_dx,
               const RealVect&                     a_dxCoar,
               const ProblemDomain&                a_domain,
               const RealVect&                     a_origin,
               const Real                          a_k2,
               const int&                          a_refToFine,
               const int&                          a_refToCoar,
               const bool&                         a_hasFine,
               const bool&                         a_hasCoar,
               const int&                          a_relaxType,
               const int&                          a_orderEB,
	       const Real                          a_kappa,
               const LevelData<EBCellFAB>*         a_rho,
               const bool&                         a_rz,
	       const Real&                         a_time )
{

  CH_assert(a_k2 >= 0 );

  for(int idir = 1; idir < SpaceDim; idir++)
    {
      if(Abs((a_dx[idir]-a_dx[0])/a_dx[0]) > 0.01)
        {
          MayDay::Warning("Anisotropic problem");
        }
    }
  m_time           = a_time;
  m_hasFine        = a_hasFine  ;
  m_hasCoar        = a_hasCoar;
  m_relaxType      = a_relaxType ;
  m_orderEB        = a_orderEB   ;
  m_domain         = a_domain    ;
  m_dbl            = a_dbl       ;
  m_domainBC       = a_domainBC;
  m_ebBC           = a_ebBC    ;
  m_dx             = a_dx      ;
  m_origin         = a_origin  ;
  m_ebisl          = a_ebisl;
  m_k2             = a_k2;
  m_rz             = a_rz;
  m_split_u        = false;

  m_homogeneous_relax = true;

  EBCellFactory cfactory( m_ebisl );
  IntVect ghosts(IntVect::Unit);

  m_kappa.define( m_dbl, 1, ghosts, cfactory );
  m_rho.define( m_dbl, 1, ghosts, cfactory );
  bool rhodefed = a_rho != NULL;
  if( rhodefed )
     a_rho->copyTo( Interval(0,0), m_rho, Interval(0,0) );
  //     m_rho.define( *a_rho, cfactory );
  
  for( DataIterator di(m_dbl) ; di.ok() ; ++di )
  {
     m_kappa[di()].setVal(a_kappa);
     if( !rhodefed )
        m_rho[di()].setVal(1.0);
  }

  //  DataIterator di=m_rho.dataIterator();
  //  print_ebcellfab((*a_rho)[di],"arho");
  //  print_ebcellfab(m_rho[di],"mrho");

  //pre-compute 1/dx and 1/(dx^2)
  m_invDx  = 1.0/m_dx;
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      m_invDx2[idir] = m_invDx[idir]*m_invDx[idir];
    }

  //dxScale is used for unscaling the boundary area (see GeometryShop)
  // (it would be better if we could just use m_dx[0])
  Real maxDx = 0.0;
  Real dv    = 1.0;
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      dv *= m_dx[idir];
      if ( m_dx[idir] > maxDx )
        {
          maxDx = m_dx[idir];
        }
    }

  m_dxScale = pow(maxDx,SpaceDim-1)/dv;

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();

  if(m_hasFine)
  {
      m_dblFine        = a_dblFine;
      m_dxFine         = a_dxFine;
      m_ebislFine      = a_ebislFine;
      m_domainBCFine   = a_domainBCFine;
      m_ebBCFine       = a_ebBCFine;
      m_refToFine      = a_refToFine;
      m_domainFine     = refine(m_domain, m_refToFine);
      m_kappaFine.define( m_dblFine, 1, ghosts, EBCellFactory(m_ebislFine) );
      for( DataIterator di(m_dblFine) ; di.ok() ; ++di )
          m_kappaFine[di()].setVal(a_kappa);
  }
  else
  {
      m_ebBCFine     = NULL;
      m_domainBCFine = NULL;
      m_refToFine    = 1;
  }

  if(m_hasCoar)
    {
      m_dblCoar        = a_dblCoar     ;
      m_dxCoar         = a_dxCoar;
      m_ebislCoar      = a_ebislCoar   ;
      m_domainBCCoar   = a_domainBCCoar;
      m_ebBCCoar       = a_ebBCCoar    ;
      m_refToCoar      = a_refToCoar   ;
      m_domainCoar     = coarsen(m_domain, m_refToCoar);



      for (int idir = 0; idir < SpaceDim; idir++)
        {
          m_loCFIVS[idir].define(m_dbl);
          m_hiCFIVS[idir].define(m_dbl);

          for (DataIterator dit = m_dbl.dataIterator();  dit.ok(); ++dit)
            {
              m_loCFIVS[idir][dit()].define(m_domain, m_dbl.get(dit()),
                                            m_dbl, idir,Side::Lo);
              m_hiCFIVS[idir][dit()].define(m_domain, m_dbl.get(dit()),
                                            m_dbl, idir,Side::Hi);
            }
        }

      //if this fails, then the AMR grids violate proper nesting.
      Box domainCoarsenedFine;
      bool hasCoarser = getCoarserLayouts(m_dblCoarsenedFine,
                                          m_ebislCoarsenedFine,
                                          domainCoarsenedFine,
                                          m_dbl,
                                          m_domain.domainBox(),
                                          m_refToCoar,
					  ebisPtr);


      //should follow from coarsenable
      if(hasCoarser)
        {
          m_hasInterpAve = true;
          int ncomp = 1;
          m_ebInterp.define( m_dbl,     m_dblCoar,
                             m_ebisl, m_ebislCoar,
                             domainCoarsenedFine, m_refToCoar, ncomp,ebisPtr);
          m_ebAverage.define(m_dbl,     m_dblCoarsenedFine,
                             m_ebisl, m_ebislCoarsenedFine,
                             domainCoarsenedFine, m_refToCoar, ncomp,ebisPtr);
        }
      else
        {
          m_hasInterpAve = false;
        }
    }
  else
    {
      m_ebBCCoar     = NULL;
      m_domainBCCoar = NULL;
      m_refToCoar    = 1;
      m_domainCoar     = m_domain;
    }

  //special mg objects for when we do not have
  //a coarser level or when the refinement to coarse
  //is greater than two
  //flag for when we need special MG objects
  m_hasMGObjects = false;
  if((!m_hasCoar) || (m_refToCoar > 2))
    {
      int mgRef = 2;
      //don't want to go to 1x1
      int ncomp = 1;
      bool hasCoarser = getCoarserLayouts(m_dblCoarMG,
                                          m_ebislCoarMG,
                                          m_domainCoarMG,
                                          m_dbl,
                                          m_domain.domainBox(),
                                          mgRef,ebisPtr);

      if(hasCoarser)
        {
          m_hasMGObjects = true;
          m_ebInterpMG.define( m_dbl,     m_dblCoarMG,
                               m_ebisl, m_ebislCoarMG,
                               m_domainCoarMG, mgRef, ncomp,ebisPtr);
          m_ebAverageMG.define(m_dbl,     m_dblCoarMG,
                               m_ebisl, m_ebislCoarMG,
                               m_domainCoarMG, mgRef, ncomp,ebisPtr);
        }
      else
        {
          m_hasMGObjects = false;
        }
    }

  //define stencils for the operator
  defineStencils();

  DataIterator d = m_dbl.dataIterator() ; 

  //  print_test_info( m_ebisl[d()]);
  define_flux_stencils();
  compute_diagonal_weight();
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::defineStencils()
{
  m_uStencil.define( m_dbl);
  m_opStencil.define(m_dbl);
  m_weight.define(   m_dbl);
  m_uweight.define(  m_dbl);

  for (DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
  {
      const Box& curBox = m_dbl.get(dit());
      const EBISBox& curEBISBox = m_ebisl[dit()];
      const EBGraph& curEBGraph = curEBISBox.getEBGraph();

      IntVectSet notRegular = curEBISBox.getIrregIVS(curBox);

      BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_opStencil[dit()];
      BaseIVFAB<Real>&       curWeightBaseIVFAB  = m_weight[dit()];

      curStencilBaseIVFAB.define(notRegular,curEBGraph, 1);
      curWeightBaseIVFAB.define( notRegular,curEBGraph, 1);
      m_uStencil[dit()].define(notRegular,curEBGraph,1);
      m_uweight[dit()].define(notRegular,curEBGraph,1);
      for (VoFIterator vofit(notRegular,curEBGraph); vofit.ok(); ++vofit)
        {
          const VolIndex& VoF = vofit();
          get_ustencil( m_uStencil[dit()](VoF,0), m_uweight[dit()](VoF,0), curEBISBox, VoF );
          VoFStencil& curStencil = curStencilBaseIVFAB(VoF,0);

          getOpVoFStencil(curStencil,curEBISBox,VoF);

          Real& curWeight  = curWeightBaseIVFAB(VoF,0);
          curWeight = 0.0;
          for( int i = 0; i < curStencil.size(); i++ )
          {
             VolIndex curVoF = curStencil.vof(i);
             if( curVoF == VoF )
             {
                curWeight = curStencil.weight(i);
                break;
             }
          }

          const IntVect& iv = VoF.gridIndex();

          for (int idir = 0; idir < SpaceDim; idir++)
            {
              Box loSide = bdryLo(m_domain,idir);
              loSide.shiftHalf(idir,1);
              if (loSide.contains(iv))
                {
                  Real faceAreaFrac = 0.0;
                  Vector<FaceIndex> faces = curEBISBox.getFaces(VoF,idir,Side::Lo);
                  for (int i = 0; i < faces.size(); i++)
                    {
                      faceAreaFrac += curEBISBox.areaFrac(faces[i]);
                    }

                  curWeight += -1.0 * faceAreaFrac / (m_dx[idir]*m_dx[idir]);
                }

              Box hiSide = bdryHi(m_domain,idir);
              hiSide.shiftHalf(idir,-1);
              if (hiSide.contains(iv))
                {
                  Real faceAreaFrac = 0.0;
                  Vector<FaceIndex> faces = curEBISBox.getFaces(VoF,idir,Side::Hi);
                  for (int i = 0; i < faces.size(); i++)
                    {
                      faceAreaFrac += curEBISBox.areaFrac(faces[i]);
                    }

                  curWeight += -1.0 * faceAreaFrac / (m_dx[idir]*m_dx[idir]);
                }
            }

          if (curWeight == 0.0)
            {
              curWeight = -1.0;
            }
        }
    }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::define_flux_stencils()
{
  m_iflux_stencil.define( m_dbl);
  m_jflux_stencil.define( m_dbl);
  m_kflux_stencil.define( m_dbl);

  for( DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
  {
     const DataIndex& dex = dit();
     Box fluxbox = m_kappa[dex].box(); //m_dbl.get(dex);
     //     cout << "--------------------new stencil----------------------";
     //     print_test_info( eb );
     for( int dir=0; dir<SpaceDim; dir++ )
     {
        if( dir == 0 )
	   define_flux_stencil( dir, m_iflux_stencil[dex],
			        m_ebisl[dex], fluxbox );
        else if( dir == 1 )
           define_flux_stencil( dir, m_jflux_stencil[dex],
				m_ebisl[dex], fluxbox );
        else
           define_flux_stencil( dir, m_kflux_stencil[dex],
				m_ebisl[dex], fluxbox );
     }
  }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::define_flux_stencil( int dir,
					    BaseIFFAB<FaceStencil>& flux_stencil,
					    const EBISBox& eb,
					    const Box& ubox )
{
     const EBGraph& curEBGraph = eb.getEBGraph();
     IntVectSet notRegular     = eb.getIrregIVS(ubox);
     flux_stencil.define( notRegular, curEBGraph, dir, 1 );


// Set up stencils for the flux tangential interpolation
     for( FaceIterator it(notRegular,curEBGraph,dir,FaceStop::SurroundingNoBoundary); it.ok(); ++it)
     {
        FaceIndex fi=it();

        flux_stencil(fi,0).clear();

        RealVect center = eb.centroid(fi);
        RealVect normal = eb.normal(fi.getVoF(Side::Hi));

        RealVect wgh= RealVect::Zero;
	//        IntVect  of = IntVect::Zero;
        IntVect  of = IntVect::Zero;
        for( int d=0 ; d< SpaceDim ; d++ )
           if( d != dir )
  	   {
	      of[d]  =  normal[d] > 0 ? 1:-1;
              wgh[d] = fabs(center[d]);
	   }

        for( int d1=0 ; d1 < 2 ; d1++ )
           for( int d2=0 ; d2 < 2 ; d2++ )
              for( int d3=0 ; d3 < 2 ; d3++ )
       	      {
	         IntVect of2=IntVect::Zero;
  		 of2[0] = d1==0 ? 0 :of[0];
		 Real w1 = d1==0 ? 1-wgh[0]:wgh[0];
		 of2[1] = d2==0 ? 0 :of[1];
		 Real w2 = d2==0 ? 1-wgh[1]:wgh[1];
                 Real w3;
                 if( SpaceDim == 3 )
		 {
		    of2[2] = d3==0 ? 0 :of[2];
   		    w3 = d3==0 ? 1-wgh[2]:wgh[2];
		 }
		 else
		 {
		   w3 = 1-d3;
		 }
                 if( w1*w2*w3 != 0 )
		 {
                   Vector<VolIndex> vl, vu;
		   if( eb.getRegion().contains(fi.gridIndex(Side::Lo)+of2) &&
                       eb.getRegion().contains(fi.gridIndex(Side::Hi)+of2) )
		   {
		      vl=eb.getVoFs(fi.gridIndex(Side::Lo)+of2);
		      vu=eb.getVoFs(fi.gridIndex(Side::Hi)+of2);
                   }
		   else
		   {
		      vl=eb.getVoFs(fi.gridIndex(Side::Lo));
		      vu=eb.getVoFs(fi.gridIndex(Side::Hi));
                   }
                       
		       //                       Vector<VolIndex> vl=m_ebisl[dex].getVoFs(fi.getVoF(Side::Lo),dir,Side::Lo,1);
		       //		       Vector<VolIndex> vu=m_ebisl[dex].getVoFs(fi.getVoF(Side::Hi),dir,Side::Hi,1);
		   if( vl.size() >0 && vu.size() > 0 )
		   {
// find connections
                      Vector<FaceIndex> faces;
		      for( int i=0 ; i<vl.size() ; i++ )
                         for( int j=0 ;j < vu.size() ; j++ )
			   if( eb.isConnected(vl[i],vu[j]) )
			   {
                              faces.push_back(FaceIndex(vl[i],vu[j],dir)); 
			   }
// Is my face in the list, then use it
                      bool found=false;
		      for( int i=0 ; i<faces.size() ; i++ )
			 if( faces[i] == fi )
                         {
  		            flux_stencil(fi,0).add(faces[i],w1*w2*w3);
                            found = true;
			    break;
		         }
// otherwise just take another 
                      if( !found && faces.size() > 0 )
                      {
  		         flux_stencil(fi,0).add(faces[0],w1*w2*w3);
		      }
		   }

		 }
	      }
        flux_stencil(fi,0) *= eb.areaFrac(fi);
     }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::residual(LevelData<EBCellFAB>&       a_residual,
				const LevelData<EBCellFAB>& a_phi,
				const LevelData<EBCellFAB>& a_rhs,
				bool                        a_homogeneousPhysBC)
{
  //this is a multigrid operator so only homogeneous CF BC
  //and null coar level
  applyOp(a_residual,a_phi,NULL, a_homogeneousPhysBC, true);
  scale(a_residual,-1.0);
  incr(a_residual,a_rhs,1.0);
}

//-----------------------------------------------------------------------
//void EBAMRHelmholtzOp::preCond(LevelData<EBCellFAB>& a_lhs,
//			       const LevelData<EBCellFAB>& a_rhs)
//{
//   EBLevelDataOps::product( a_lhs, a_rhs, m_wgh );
//   for(int iter = 0; iter < 40; iter++)
//      levelJacobi( a_lhs, a_rhs );
//}

//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::preCond(LevelData<EBCellFAB>&       a_lhs,
			       const LevelData<EBCellFAB>& a_rhs)
{
   EBLevelDataOps::assign(a_lhs,a_rhs);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::applyOp(LevelData<EBCellFAB>& a_opPhi,
			       const LevelData<EBCellFAB>& a_phi,
			       bool a_homogeneousPhysBC)
{
  //homogeneous CFBCs because that is all we can do.
  applyOp(a_opPhi, a_phi, NULL, a_homogeneousPhysBC, true);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::applyOp(LevelData<EBCellFAB>&             a_opPhi,
			       const LevelData<EBCellFAB>&       a_phi,
			       const LevelData<EBCellFAB>* const a_phiCoar,
			       bool a_homogeneousPhysBC,
			       bool a_homogeneousCFBC)
{
  //apply operator without input EB inhomogeneous BC leveldata
  applyOp(a_opPhi, a_phi, a_phiCoar, a_homogeneousPhysBC, a_homogeneousCFBC, NULL);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::applyOp(LevelData<EBCellFAB>&       a_opPhi,
			       const LevelData<EBCellFAB>& a_phi,
			       const LevelData<EBCellFAB>* const a_phiCoar,
			       bool a_homogeneousPhysBC,
			       bool a_homogeneousCFBC,
			       const LevelData<BaseIVFAB<Real> >* const a_ebFluxBCLD)
{
  //if input EBFlux leveldata is null, then we have no input eb flux
  bool hasNoEBLevelData = (a_ebFluxBCLD == NULL);

  CH_assert(a_phi.nComp() == 1);
  CH_assert(a_opPhi.nComp() == 1);
  LevelData<EBCellFAB>& phi = (LevelData<EBCellFAB>&)a_phi;
  applyCFBCs(phi, a_phiCoar, a_homogeneousCFBC);
  phi.exchange(phi.interval());

  int nComps = 1;
  int comp = 0;
  int ibox = 0;

  for (DataIterator dit = phi.dataIterator(); dit.ok(); ++dit)
  {
     const DataIndex& dex = dit();


     const EBCellFAB& curPhiEBCellFAB = phi[dex];
     Box curPhiBox = curPhiEBCellFAB.box();
     const BaseFab<Real>& curPhiFAB = curPhiEBCellFAB.getSingleValuedFAB();

     EBCellFAB& curOpPhiEBCellFAB = a_opPhi[dex];
     Box curOpPhiBox = curOpPhiEBCellFAB.box();

     curOpPhiEBCellFAB.setVal(0.0);
     BaseFab<Real>& curOpPhiFAB = curOpPhiEBCellFAB.getSingleValuedFAB();

     Box loBox[SpaceDim],hiBox[SpaceDim];
     int hasLo[SpaceDim],hasHi[SpaceDim];

     for( int idir = 0; idir < SpaceDim; idir++ )
     {
        Box centerBox,entireBox;
        ebEllipticLoHiCenter(loBox[idir],hasLo[idir],hiBox[idir],hasHi[idir],
                               centerBox,entireBox,curOpPhiBox,curPhiBox,
                               m_domain, idir );
        if( m_rz )
           FORT_LAPLACIAN_RZ_1D(CHF_FRA1(curOpPhiFAB,comp),
				CHF_CONST_FRA1(curPhiFAB,comp),
				CHF_CONST_FRA1(m_kappa[dex].getSingleValuedFAB(),0),
				CHF_CONST_REAL(m_dx[idir]),
				CHF_CONST_REALVECT(m_origin),
				CHF_CONST_INT(idir),
				CHF_BOX(loBox[idir]),
				CHF_CONST_INT(hasLo[idir]),
                                CHF_BOX(hiBox[idir]),
				CHF_CONST_INT(hasHi[idir]),
				CHF_BOX(centerBox));
	else
           FORT_LAPLACIAN_XYZ_1D(CHF_FRA1(curOpPhiFAB,comp),
                                 CHF_CONST_FRA1(curPhiFAB,comp),
                                 CHF_CONST_REAL(m_dx[idir]),
                                 CHF_CONST_INT(idir),
                                 CHF_BOX(loBox[idir]),
                                 CHF_CONST_INT(hasLo[idir]),
				 CHF_BOX(hiBox[idir]),
				 CHF_CONST_INT(hasHi[idir]),
				 CHF_BOX(centerBox),
				 CHF_CONST_FRA1(m_kappa[dex].getSingleValuedFAB(),0));

	if( hasLo[idir] == 1 )
        {
	   BaseFab<Real> loFaceFlux(loBox[idir],nComps);
	   CH_assert(m_domainBC != NULL);
	   m_domainBC->get_faceflux( loFaceFlux, curPhiFAB, m_origin, m_dx,
				     idir, Side::Lo, a_homogeneousPhysBC, m_time );
	   int side = -1;
	   FORT_APPLYDOMAINFLUX(CHF_FRA1(curOpPhiFAB,comp),
				CHF_CONST_FRA1(loFaceFlux,comp),
				CHF_FRA1(m_kappa[dex].getSingleValuedFAB(),0),
				CHF_CONST_REAL(m_dx[idir]),
				CHF_CONST_INT(side),
				CHF_BOX(loBox[idir]) );
	}
	if( hasHi[idir] == 1 )
        {
           BaseFab<Real> hiFaceFlux(hiBox[idir],nComps);
	   CH_assert( m_domainBC != NULL );
	   m_domainBC->get_faceflux( hiFaceFlux, curPhiFAB, m_origin, m_dx,
				     idir, Side::Hi, a_homogeneousPhysBC, m_time);
	   int side = 1;
	   FORT_APPLYDOMAINFLUX(CHF_FRA1(curOpPhiFAB,comp),
				CHF_CONST_FRA1(hiFaceFlux,comp),
				CHF_FRA1(m_kappa[dex].getSingleValuedFAB(),0),
				CHF_CONST_REAL(m_dx[idir]),
				CHF_CONST_INT(side),
				CHF_BOX(hiBox[idir]));
	}
     }
     FORT_ADDHELMHOLTZ_RHO(CHF_FRA1(curOpPhiFAB,comp),
			   CHF_CONST_FRA1(curPhiFAB,comp),
			   CHF_CONST_FRA1(m_rho[dit()].getSingleValuedFAB(),comp),
			   CHF_BOX(curOpPhiBox),
			   CHF_CONST_REAL(m_k2));

     const EBISBox& curOpPhiEBISBox = curOpPhiEBCellFAB.getEBISBox();
     IntVectSet notRegular = m_ebisl[dex].getIrregIVS(curOpPhiEBCellFAB.box());

     for( VoFIterator vi(notRegular,m_ebisl[dex].getEBGraph());vi.ok();++vi)
         curOpPhiEBCellFAB(vi(),0) = 0;

     update_irregular( curOpPhiEBCellFAB, curPhiEBCellFAB, curOpPhiEBISBox,
			m_iflux_stencil[dex], m_kappa[dex], m_rho[dit()], 0 );
     update_irregular( curOpPhiEBCellFAB, curPhiEBCellFAB, curOpPhiEBISBox,
			m_jflux_stencil[dex], m_kappa[dex], m_rho[dit()], 1 );
     if( SpaceDim == 3 )
        update_irregular( curOpPhiEBCellFAB, curPhiEBCellFAB, curOpPhiEBISBox,
			   m_kflux_stencil[dex], m_kappa[dex], m_rho[dex], 2 );

     for( VoFIterator vofit(notRegular,m_ebisl[dex].getEBGraph()); vofit.ok(); ++vofit )
     {
        const VolIndex& vof = vofit();
        const IntVect& iv = vof.gridIndex();

        int comp = 0;

	 //         const VoFStencil& curStencil = curBaseIVFAB(vof,comp);
	 //         curOpPhiEBCellFAB(vof,comp) = 0.0;

	// 	bool interesting_vof = vof.gridIndex()[0] == 0 && vof.gridIndex()[1] == 3;
         for( int idir = 0; idir < SpaceDim; idir++ )
         {
            if( hasLo[idir] == 1 )
            {
               if( loBox[idir].contains(iv) )
               {
                  Vector<Real> flux(1);
                  m_domainBC->get_faceflux( flux, vof, curPhiEBCellFAB,
					   m_origin, m_dx, idir, Side::Lo,
					   a_homogeneousPhysBC, m_time );

		  //                  if( interesting_vof )
		  //		    cout << "low bc, before" << curOpPhiEBCellFAB(vof,0) << endl;
                  curOpPhiEBCellFAB(vof,comp) -=
		                   flux[0] * m_kappa[dex](vof,0)*m_invDx[idir];
		  //                  if( interesting_vof )
		  //		    cout << "low bc, after" << curOpPhiEBCellFAB(vof,0) << endl;
               }
            }
            if( hasHi[idir] == 1 )
            {
               if( hiBox[idir].contains(iv) )
               {
                  Vector<Real> flux(1);
                  m_domainBC->get_faceflux( flux, vof, curPhiEBCellFAB,
					   m_origin, m_dx,idir, Side::Hi,
					   a_homogeneousPhysBC, m_time );
		  //                  if( interesting_vof )
		  //		    cout << "high bc, before" << curOpPhiEBCellFAB(vof,0) << endl;
                  curOpPhiEBCellFAB(vof,comp) +=
		                    flux[0] * m_kappa[dex](vof,0)*m_invDx[idir];
		  //                  if( interesting_vof )
		  //		    cout << "high bc, after" << curOpPhiEBCellFAB(vof,0) << endl;

               }
            }
         }

          //get EB flux from leveldata if we have one.  otherwise use m_ebBC
          //ignore input data in the case of homogeneous Phys BC
         Vector<Real> ebFlux(1);
         if( hasNoEBLevelData )
         {
            m_ebBC->get_ebflux( ebFlux, vof, phi, dit(),
			       m_origin, m_dx,
			       a_homogeneousPhysBC, m_time );
         }
         else
         {
            const BaseIVFAB<Real>& ebInputFluxFAB = (*a_ebFluxBCLD)[dit()];
            CH_assert(ebInputFluxFAB.nComp() == 1);
            ebFlux[0] = ebInputFluxFAB(vof, 0);
         }
	 //                  if( interesting_vof )
	 //		    cout << "eb bc, before" << curOpPhiEBCellFAB(vof,0) << endl;

         curOpPhiEBCellFAB(vof,comp) += ebFlux[0]*m_kappa[dex](vof,0)* m_dxScale;
	 //                  if( interesting_vof )
	 //		    cout << "eb bc, after" << curOpPhiEBCellFAB(vof,0) << endl;

         Real hhterm=0;
         if( m_split_u )
         {
             const VoFStencil& u_stencil = m_uStencil[dit()](vof,comp);
             for (int i = 0; i < u_stencil.size(); i++)
                 hhterm += u_stencil.weight(i) *
		                       curPhiEBCellFAB(u_stencil.vof(i),comp);
	 }
	 else
	    hhterm = curPhiEBCellFAB( vof,comp);
	  
          curOpPhiEBCellFAB(vof,comp) -= m_k2*hhterm*m_rho[dit()](vof,0)*curOpPhiEBISBox.volFrac(vof);
        }
      ibox++;
      //      print_ebcellfab( curOpPhiEBCellFAB, "L(u)");
  }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::print_ebcellfab(  const EBCellFAB& a_u, const char* name )
{
   const EBISBox& eb = a_u.getEBISBox();
   for( BoxIterator b(eb.getRegion()); b.ok(); ++b )
   {
      Vector<VolIndex> vofar = eb.getVoFs(b());
      if( eb.isCovered(b()) )
	 cout << name << "(" << b()[0] <<","<<b()[1] << ") is covered" << endl;	
      for( int iv=0 ; iv < vofar.size() ; iv++ )
      {
	 cout << name << "(" << b()[0] <<","<<b()[1] << ")";
	 cout << "["<<iv<<"]=" << a_u(vofar[iv],0) ;
	 if( vofar.size() >1)
	 {
	    RealVect cen=eb.centroid(vofar[iv]);
	    cout << " offset " << cen ;
	 }
	 cout << endl;
      }
   }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::print_test_info(  const EBISBox& a_eb )
{  
  // enumerate cells
   EBCellFAB test(a_eb,a_eb.getRegion(),1);
   int vovnr = 1;
   for( BoxIterator b(a_eb.getRegion()); b.ok(); ++b )
   {
      Vector<VolIndex> vofar = a_eb.getVoFs(b());
      for( int iv=0 ; iv < vofar.size() ; iv++ )
	test(vofar[iv],0) = vovnr++;
   }

         for( BoxIterator b(a_eb.getRegion()); b.ok(); ++b )
         {
	    Vector<VolIndex> vofar = a_eb.getVoFs(b());
	    for( int iv=0 ; iv < vofar.size() ; iv++ )
	    {
	       cout << "---------------------------------------" << endl;
	       cout << b() << " u at vof is " << test(vofar[iv],0) << endl;
               cout << "Cell index is " << vofar[iv].cellIndex() << endl;

               Vector<VolIndex> vn = a_eb.getVoFs(vofar[iv],0,Side::Lo,1);
               for( int i=0 ; i<vn.size() ; i++ )
		  cout << "left neigbor is  " << test(vn[i],0) << endl;

               vn = a_eb.getVoFs(vofar[iv],0,Side::Hi,1);
               for( int i=0 ; i<vn.size() ; i++ )
		  cout << "Right neigbor is " << test(vn[i],0) << endl;

               vn = a_eb.getVoFs(vofar[iv],1,Side::Lo,1);
               for( int i=0 ; i<vn.size() ; i++ )
		  cout << "Down neigbor is  " << test(vn[i],0) << endl;

               vn = a_eb.getVoFs(vofar[iv],1,Side::Hi,1);
               for( int i=0 ; i<vn.size() ; i++ )
		  cout << "Up neigbor is    " << test(vn[i],0) << endl;

               cout << "Normal is " << a_eb.normal(vofar[iv]) << endl; 
               cout << "Cell is " ;
               if( a_eb.isCovered(vofar[iv].gridIndex()) )
		  cout << "c" ;
	       else if( a_eb.isIrregular(vofar[iv].gridIndex()) )
		  cout << "i";
	       else
		  cout << "r";
	       cout << endl;
              cout << "bndry centroid is  " << a_eb.bndryCentroid(vofar[iv]) << endl;
              cout << "bndry area   is    " << a_eb.bndryArea(vofar[iv]) << endl;
	      cout << "volume fraction is " << a_eb.volFrac(vofar[iv]) << endl;
               Vector<FaceIndex> faces = a_eb.getFaces(vofar[iv],1,Side::Lo);
               for( int i=0 ; i<faces.size() ; i++ )
	       {
 	           VolIndex vh=faces[i].getVoF(Side::Hi);
	           VolIndex vl=faces[i].getVoF(Side::Lo);

                   if( vh.cellIndex() >-1 && vl.cellIndex() > -1 )
		       cout << "Face between " << test(vh,0) << " and " << test(vl,0);
                   else if( vh.cellIndex() > -1 )
		       cout << "low domain bndry at " << test(vh,0);
		   else
		       cout << "high domain bndry at " << test(vl,0);
                   cout << "centroid " << a_eb.centroid (faces[i]) ;
                   cout << " area frac " << a_eb.areaFrac(faces[i]) << endl;
	       }         
               Vector<VolIndex> allMonotoneVoFs;
               int radius = 1;
               IntVect timesMoved = IntVect::Zero;
               IntVect pathSign   = IntVect::Zero;

               EBArith::getAllVoFsInMonotonePath(allMonotoneVoFs,timesMoved,pathSign,
                                    vofar[iv],a_eb,radius);
               cout << "Monotone vofs of radius "<< radius << endl;
         
               for( int i=0 ; i< allMonotoneVoFs.size() ; i++ )
		  cout <<  "         " << test(allMonotoneVoFs[i],0) << endl;

	    }
      	 }
         cout << "J-Fluxes: " << endl;
         IntVectSet all_points(a_eb.getRegion());
         for( FaceIterator it(all_points,a_eb.getEBGraph(), 1,
			      FaceStop::SurroundingNoBoundary); it.ok(); ++it)
	 {
	    VolIndex vh=it().getVoF(Side::Hi);
	    VolIndex vl=it().getVoF(Side::Lo);
            cout << " side between " << test(vh,0) << " and " << test(vl,0) ;
            cout << "centroid " << a_eb.centroid(it()) << endl;
	 }

         cout << " checking connects --------------------------------" << endl;
         for( BoxIterator b(a_eb.getRegion()); b.ok(); ++b )
         {
	    Vector<VolIndex> vofar = a_eb.getVoFs(b());
	    for( int iv=0 ; iv < vofar.size() ; iv++ )
	    {
	      for( BoxIterator c(a_eb.getRegion()) ; c.ok(); ++c )
	      {
		Vector<VolIndex> voff=a_eb.getVoFs(c());
		for( int w=0  ; w < voff.size() ; w++ )
		  if( a_eb.isConnected(voff[w],vofar[iv]) )
		    cout << test(voff[w],0) << " is connected to " <<
		      test(vofar[iv],0) << endl;
	      }
	    }
	 }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::update_irregular( EBCellFAB& a_res,
					 const EBCellFAB& a_u,
					 const EBISBox& a_eb,
					 const BaseIFFAB<FaceStencil>& a_flux_stencil,
					 const EBCellFAB& a_kappa,
					 const EBCellFAB& a_rho, int a_dir )
{
   const Real half=0.5;
   int testing_mode=0;

   Real dxi = 1/m_dx[a_dir];
    
   const EBGraph& curEBGraph = a_eb.getEBGraph();
   IntVectSet notRegular     = a_eb.getIrregIVS(a_u.box());
   IntVectSet irreg_res      = a_eb.getIrregIVS(a_res.box());
      
   EBCellFAB test;
   if( testing_mode == 2 || testing_mode == 1 )
   {
      test.define(a_eb,a_eb.getRegion(),1);
      int vovnr = 1;
      for( BoxIterator b(a_eb.getRegion()); b.ok(); ++b )
      {
         Vector<VolIndex> vofar = a_eb.getVoFs(b());
         for( int iv=0 ; iv < vofar.size() ; iv++ )
	    test(vofar[iv],0) = vovnr++;
      }
      if( testing_mode == 1 )
         print_test_info( a_eb );
   }

   for(FaceIterator it(notRegular,curEBGraph,a_dir,FaceStop::SurroundingNoBoundary); it.ok();++it)
   {
      FaceIndex fi=it();
      // Compute flux:
      bool interesting_case =  ( testing_mode ==2 &&
		     (test(fi.getVoF(Side::Lo),0) == 11 || test(fi.getVoF(Side::Hi),0) == 11)); 
      if( testing_mode == 1 || ( testing_mode ==2 && interesting_case ))
      {
         cout << "flux between " << test(fi.getVoF(Side::Lo),0) << " and " <<
	    test(fi.getVoF(Side::Hi),0) << endl;
      }
      Real flx = 0;
      for( int ii=0 ; ii < a_flux_stencil(fi,0).size() ; ii++ )
      {
         FaceIndex fic = a_flux_stencil(fi,0).face(ii);

         VolIndex voh = fic.getVoF(Side::Hi);
         VolIndex vol = fic.getVoF(Side::Lo);
         if( testing_mode == 1 || (testing_mode ==2 && interesting_case ) )
            cout << "flux-stencil contains  " << test(vol,0) << " and " << test(voh,0) << " "<< a_flux_stencil(fi,0).weight(ii) << endl;
         bool inh = a_u.box().contains( voh.gridIndex());
         bool inl = a_u.box().contains( vol.gridIndex());
         Real flxc=0;
	 if( inh && inl )
	 {
            flxc = (a_kappa(vol,0)+a_kappa(voh,0))*half*dxi*
		         (a_u(voh,0)-a_u(vol,0))*dxi;

	 //         if( m_rz && a_dir == 0 )
            if( m_rz )
   	    {
               Real rp= (voh.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
               Real r = (vol.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
               Real rh =(r+rp)*0.5;
 	       flxc *= rh;
	    }
	 }
         flx += flxc*a_flux_stencil(fi,0).weight(ii);
      }

// Add flux to residual
      const VolIndex& vohi = fi.getVoF(Side::Hi);
      if( irreg_res.contains(vohi.gridIndex()) )
      {
         if( testing_mode == 2 && test(vohi,0) == 11 )
	     cout << "vof 11 high, flx= " << flx << endl;
	 if( !m_rz  )
 	    a_res(vohi,0) -= flx;
	 else
	 {
            RealVect c=a_eb.centroid(vohi);
            Real rc=(c[0]+vohi.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
 	    a_res(vohi,0) -= flx/rc;
	   if( testing_mode == 2 && test(vohi,0) == 11 )
	     cout << "vof 11 high, rc= " << rc << " flx/rc = " << flx/rc << endl;
	 }
      }
      const VolIndex& volo = fi.getVoF(Side::Lo);
      if( irreg_res.contains(volo.gridIndex()) )
      { 
         if( testing_mode == 2 && test(volo,0) == 11 )
	     cout << "vof 11 low, flx= " << flx << endl;

	 if( !m_rz )
 	    a_res(volo,0) += flx;
         else
	 {
            RealVect c=a_eb.centroid(volo);
            Real rc=(c[0]+volo.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
	    a_res(volo,0) += flx/rc;

	   if( testing_mode == 2 && test(volo,0) == 11 )
	     cout << "vof 11 low, rc= " << rc << " flx/rc = " << flx/rc << endl;
	 }
      }
   }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::compute_dw_irregular(
				    EBCellFAB& a_wgh,
				    const EBISBox& a_eb,
				    const BaseIFFAB<FaceStencil>& a_flux_stencil,
				    const EBCellFAB& a_kappa,
				    int a_dir )
{
   const Real half=0.5;

   Real dxi = 1/m_dx[a_dir];

   const EBGraph& curEBGraph = a_eb.getEBGraph();
   IntVectSet irreg_wgh      = a_eb.getIrregIVS(a_wgh.box());
   for(FaceIterator it( irreg_wgh, curEBGraph, a_dir, FaceStop::SurroundingNoBoundary );
       it.ok();++it)
   {
      FaceIndex fi=it();

      const VolIndex& vohi = fi.getVoF(Side::Hi);
      const VolIndex& volo = fi.getVoF(Side::Lo);
      //      cout <<"dw between " << volo.gridIndex() << " and " << vohi.gridIndex();
      for( int ii=0 ; ii < a_flux_stencil(fi,0).size() ; ii++ )
      {
         FaceIndex fic = a_flux_stencil(fi,0).face(ii);

         VolIndex voh = fic.getVoF(Side::Hi);
         VolIndex vol = fic.getVoF(Side::Lo);

	 //         bool inh = irreg_wgh.contains(vohi.gridIndex()); 
	 //         bool inl = irreg_wgh.contains(volo.gridIndex());
         bool inh = a_wgh.box().contains(vohi.gridIndex());
         bool inl = a_wgh.box().contains(volo.gridIndex());
	 //         cout << " inh & inl " << inh <<" "<< inl<< " stencil size " << a_flux_stencil(fi,0).size() << " " << a_flux_stencil(fi,0).weight(ii) << endl;
         Real flfact=a_flux_stencil(fi,0).weight(ii);

	 //         if( m_rz && a_dir == 0 )
         if( m_rz  )
	 {
            Real rp= (voh.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
            Real r = (vol.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
            Real rh =(r+rp)*0.5;
 	    flfact *= rh;
	 }
	 if( inh && inl )
	 {
	    Real rc=1, rch=1;
	    //	    if( m_rz && a_dir == 0 )
	    if( m_rz  )
	    {
               RealVect c=a_eb.centroid(volo);
               rc  = (c[0]+volo.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
               c   = a_eb.centroid(vohi);
               rch = (c[0]+vohi.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
	    }
	    if( vol == volo )
 	       a_wgh(volo,0) -= (a_kappa(vol,0)+a_kappa(voh,0))*half*dxi*flfact*dxi/rc;
            if( vol == vohi )
       	       a_wgh(vohi,0) -= (a_kappa(vol,0)+a_kappa(voh,0))*half*dxi*flfact*dxi/rch;
            if( voh == vohi )
       	       a_wgh(vohi,0) -= (a_kappa(vol,0)+a_kappa(voh,0))*half*dxi*flfact*dxi/rch;
            if( voh == volo )
       	       a_wgh(volo,0) -= (a_kappa(vol,0)+a_kappa(voh,0))*half*dxi*flfact*dxi/rc;
	 }
	 //         if( volo.gridIndex() == IntVect(3,2) )
	   //	   cout << " added " << a_wgh(volo,0) << endl;
	   //         if( vohi.gridIndex() == IntVect(3,2) )
	   //	   cout << " added " << a_wgh(vohi,0) << endl;
      }
   }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::create( LevelData<EBCellFAB>&       a_lhs,
			       const LevelData<EBCellFAB>& a_rhs )
{
  int ncomp = a_rhs.nComp();
  EBCellFactory ebcellfact(m_ebisl);
  a_lhs.define(m_dbl, ncomp, a_rhs.ghostVect(), ebcellfact);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::createCoarsened(LevelData<EBCellFAB>&       a_lhs,
				       const LevelData<EBCellFAB>& a_rhs,
				       const int &                 a_refRat)
{
  int ncomp = a_rhs.nComp();
  IntVect ghostVect = a_rhs.ghostVect();

  CH_assert(m_dbl.coarsenable(a_refRat));

  //fill ebislayout
  DisjointBoxLayout dblCoarsenedFine;
  coarsen(dblCoarsenedFine, m_dbl, a_refRat);

  EBISLayout ebislCoarsenedFine;
  IntVect ghostVec = a_rhs.ghostVect();
  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  ebisPtr->fillEBISLayout(ebislCoarsenedFine, dblCoarsenedFine, m_domainCoar.domainBox(), ghostVec[0]);
  if(m_refToCoar > 1)
  {
      ebislCoarsenedFine.setMaxRefinementRatio(m_refToCoar,ebisPtr);
  }

  //create coarsened data
  EBCellFactory ebcellfactCoarsenedFine(ebislCoarsenedFine);
  a_lhs.define(dblCoarsenedFine, ncomp,ghostVec, ebcellfactCoarsenedFine);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::assign(LevelData<EBCellFAB>&       a_lhs,
			      const LevelData<EBCellFAB>& a_rhs)
{
  EBLevelDataOps::assign(a_lhs,a_rhs);
}

//-----------------------------------------------------------------------
Real EBAMRHelmholtzOp::dotProduct(const LevelData<EBCellFAB>& a_1,
				  const LevelData<EBCellFAB>& a_2)
{
  ProblemDomain domain;
  Real volume;

  return EBLevelDataOps::kappaDotProduct(volume,a_1,a_2,EBLEVELDATAOPS_ALLVOFS,domain);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::incr(LevelData<EBCellFAB>&       a_lhs,
			    const LevelData<EBCellFAB>& a_x,
			    Real                        a_scale)
{
  EBLevelDataOps::incr(a_lhs,a_x,a_scale);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::axby(LevelData<EBCellFAB>&       a_lhs,
			    const LevelData<EBCellFAB>& a_x,
			    const LevelData<EBCellFAB>& a_y,
			    Real                        a_a,
			    Real                        a_b)
{
  EBLevelDataOps::axby(a_lhs,a_x,a_y,a_a,a_b);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::scale(LevelData<EBCellFAB>& a_lhs,
			     const Real&           a_scale)
{
  EBLevelDataOps::scale(a_lhs,a_scale);
}


//-----------------------------------------------------------------------
Real EBAMRHelmholtzOp::norm(const LevelData<EBCellFAB>& a_rhs,
			    int                         a_ord)
{
  ProblemDomain domain;
  Real volume;

  return EBLevelDataOps::kappaNorm(volume,a_rhs,EBLEVELDATAOPS_ALLVOFS,domain,a_ord);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::setToZero(LevelData<EBCellFAB>& a_lhs)
{
  EBLevelDataOps::setToZero(a_lhs);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::createCoarser(LevelData<EBCellFAB>&       a_coar,
				     const LevelData<EBCellFAB>& a_fine,
				     bool                        a_ghosted)
{
  CH_assert(a_fine.nComp() == 1);
  DisjointBoxLayout dbl;
  Box coarDom = coarsen(m_domain.domainBox(), 2);
  // This is multigrid, so coarsen by 2
  if(a_fine.disjointBoxLayout().coarsenable(2))
    {
      coarsen(dbl,a_fine.disjointBoxLayout(),2);
    }
  else
    {
      //otherwise check to see if we are covering domain.
      //if so, make one box filling domain.
      //otherwise, SOL
      IntVectSet ivsDomain(m_domain.domainBox());
      for(LayoutIterator lit = m_dbl.layoutIterator(); lit.ok(); ++lit)
        {
          ivsDomain -= m_dbl.get(lit());
        }
      if(ivsDomain.isEmpty())
        {
          Vector<Box> boxes(1, coarDom);
          Vector<int> procs(1, 0);
          dbl.define(boxes, procs);
        }
      else
        {
          MayDay::Error("EBAMRP::createCoarse error---dbl not coarsenable");
        }
    }

  int nghost = a_fine.ghostVect()[0];
  EBISLayout coarEBISL;

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  ebisPtr->fillEBISLayout(coarEBISL,
                          dbl, coarDom, nghost);

  EBCellFactory ebcellfact(coarEBISL);
  a_coar.define(dbl, 1,a_fine.ghostVect(),ebcellfact);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::relax(LevelData<EBCellFAB>&       a_e,
			     const LevelData<EBCellFAB>& a_residual,
			     int                         a_iterations)
{
  CH_assert(a_e.nComp() == 1);
  CH_assert(a_residual.nComp() == 1);

  if( m_relaxType == 0 )
  {
     for( int i = 0; i < a_iterations; i++ )
        levelJacobi(a_e,a_residual);
  }
  else if( m_relaxType == 1 )
  {
     for( int i = 0; i < a_iterations; i++ )
        levelMulticolorGS(a_e,a_residual);
  }
  else
  {
     MayDay::Error("EBAMRHelmholtzOp::relax - invalid relaxation type");
  }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::restrictResidual(LevelData<EBCellFAB>&       a_resCoar,
					LevelData<EBCellFAB>&       a_phiThisLevel,
					const LevelData<EBCellFAB>& a_rhsThisLevel)
{

  CH_assert(a_resCoar.nComp() == 1);
  CH_assert(a_phiThisLevel.nComp() == 1);
  CH_assert(a_rhsThisLevel.nComp() == 1);

  LevelData<EBCellFAB> resThisLevel;
  bool homogeneous = true;

  EBCellFactory ebcellfactTL(m_ebisl);
  IntVect ghostVec = a_rhsThisLevel.ghostVect();

  resThisLevel.define(m_dbl, 1, ghostVec, ebcellfactTL);

  // Get the residual on the fine grid
  residual(resThisLevel,a_phiThisLevel,a_rhsThisLevel,homogeneous);

  // now use our nifty averaging operator
  Interval variables(0, 0);
  if(m_hasMGObjects)
    {
      m_ebAverageMG.average(a_resCoar, resThisLevel, variables);
    }
  else
    {
      CH_assert(m_hasInterpAve);
      m_ebAverage.average(a_resCoar, resThisLevel, variables);
    }
  //  cout << "-------------------------------------" << endl;
  //  cout << "Restricted " << endl;
  //  DataIterator dtt=resThisLevel.dataIterator();
  //  print_ebcellfab( resThisLevel[dtt], "res");
  //  cout << endl<< "To: " << endl;
  //  DataIterator dt=a_resCoar.dataIterator();
  //  print_ebcellfab( a_resCoar[dt], "rescoar");
  
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::prolongIncrement(LevelData<EBCellFAB>&       a_phiThisLevel,
					const LevelData<EBCellFAB>& a_correctCoar)
{
  Interval vars(0, 0);
  if(m_hasMGObjects)
  {
      m_ebInterpMG.pwcInterp(a_phiThisLevel, a_correctCoar, vars);
  }
  else
  {
      CH_assert(m_hasInterpAve);
      m_ebInterp.pwcInterp(a_phiThisLevel, a_correctCoar, vars);
  }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::applyCFBCs(LevelData<EBCellFAB>&             a_phi,
				  const LevelData<EBCellFAB>* const a_phiCoar,
				  bool a_homogeneousCFBC )
{
  CH_assert(a_phi.nComp() == 1);

  if(m_hasCoar)
    {
      if(!a_homogeneousCFBC)
        {
          if(a_phiCoar==NULL)
            {
              MayDay::Error("cannot enforce inhomogeneous CFBCs with NULL coar");
            }
          //define coarse fine interpolation object on the fly
          //because most operators do not need it
          CH_assert(a_phiCoar->nComp() == 1);
          int ncomp = 1;
          EBQuadCFInterp  quadCFIWithCoar(m_dbl,     m_dblCoar,
                                          m_ebisl, m_ebislCoar,
                                          m_domainCoar.domainBox(),
                                          m_refToCoar, ncomp);

          Interval interv(0,0);
          quadCFIWithCoar.interpolate(a_phi, *a_phiCoar, interv);
        }
      else
        {
          applyHomogeneousCFBCs(a_phi);
        }
    }
}

//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::applyHomogeneousCFBCs(LevelData<EBCellFAB>&   a_phi)
{
  CH_assert(a_phi.nComp() == 1);
  CH_assert( a_phi.ghostVect() >= IntVect::Unit );
  for (DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
      for (int idir = 0; idir < SpaceDim; idir++)
          for (SideIterator sit; sit.ok(); sit.next())
              applyHomogeneousCFBCs(a_phi[dit()],dit(),idir,sit());
}

//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::applyHomogeneousCFBCs(EBCellFAB&            a_phi,
					     const DataIndex&      a_datInd,
					     int                   a_idir,
					     Side::LoHiSide        a_hiorlo)
{
  CH_assert((a_idir >= 0) && (a_idir  < SpaceDim));
  CH_assert((a_hiorlo == Side::Lo )||(a_hiorlo == Side::Hi ));
  CH_assert(a_phi.nComp() == 1);
  int ivar = 0;

  const CFIVS* cfivsPtr = NULL;

  if (a_hiorlo == Side::Lo)
    {
      cfivsPtr = &m_loCFIVS[a_idir][a_datInd];
    }
  else
    {
      cfivsPtr = &m_hiCFIVS[a_idir][a_datInd];
    }

  const IntVectSet& interpIVS = cfivsPtr->getFineIVS();
  if (!interpIVS.isEmpty())
    {

      int ihilo = sign(a_hiorlo);

      Real halfdxcoar = m_dxCoar[a_idir]/2.0;
      Real halfdxfine = m_dx[a_idir]/2.0;
      Real xg = halfdxcoar -   halfdxfine;
      Real xc = halfdxcoar +   halfdxfine;
      Real xf = halfdxcoar + 3*halfdxfine;
      Real hf = m_dx[a_idir];
      Real denom = xf*xc*hf;

      const EBGraph&  ebgraph = m_ebisl[a_datInd].getEBGraph();
      for(VoFIterator vofit(interpIVS, ebgraph); vofit.ok(); ++vofit)
        {
          const VolIndex& VoFGhost = vofit();

          IntVect ivGhost  = VoFGhost.gridIndex();
          IntVect ivClose =  ivGhost;
          IntVect ivFar   =  ivGhost;

          ivClose[a_idir] -=  ihilo;
          ivFar[a_idir]   -= 2*ihilo;

          Vector<VolIndex> closeVoFs = ebgraph.getVoFs(ivClose);
          Vector<VolIndex>   farVoFs = ebgraph.getVoFs(ivFar);

          bool hasClose = (closeVoFs.size() == 1);
          bool hasFar   = (farVoFs.size()   == 1);

          Real phiGhost;
          if(hasClose && hasFar)
            {
              const VolIndex& VoFClose = closeVoFs[0];
              const VolIndex& VoFFar = farVoFs[0];

              Real phic = a_phi(VoFClose, 0);
              Real phif = a_phi(VoFFar, 0);

              // quadratic interpolation  phi = ax^2 + bx + c
              Real A = (phif*xc - phic*xf)/denom;
              Real B = (phic*hf*xf - phif*xc*xc + phic*xf*xc)/denom;

              phiGhost = A*xg*xg + B*xg;
            }
          else if(hasClose)
            {
              const VolIndex& VoFClose = closeVoFs[0];
              Real phic = a_phi(VoFClose, 0);
              //liner interpolation
              Real slope =  phic/xc;
              phiGhost   =  slope*xg;
            }
          else
            {
              phiGhost = 0.0; //nothing to interpolate from
            }
          a_phi(VoFGhost, ivar) = phiGhost;
        }
    }
}


//-----------------------------------------------------------------------
int EBAMRHelmholtzOp::refToCoarser()
{
  return m_refToCoar;
}


//-----------------------------------------------------------------------
int EBAMRHelmholtzOp::refToFiner()
{
  return m_refToFine;
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::AMRResidual(LevelData<EBCellFAB>&       a_residual,
				   const LevelData<EBCellFAB>& a_phiFine,
				   const LevelData<EBCellFAB>& a_phi,
				   const LevelData<EBCellFAB>& a_phiCoar,
				   const LevelData<EBCellFAB>& a_rhs,
				   bool a_homogeneousPhysBC)
{
  CH_assert( a_residual.nComp() == 1 );
  CH_assert( a_phi.nComp() == 1 );
  CH_assert( a_rhs.nComp() == 1 );

  setToZero( a_residual );
  applyOp( a_residual, a_phi, &a_phiCoar,  a_homogeneousPhysBC, false );
  if( m_hasFine )
     reflux(a_residual, a_phiFine, a_phi);

//   a_residual := rhs - L(phi)
   scale(a_residual,-1.0);
   incr(a_residual,a_rhs,1.0);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::reflux(LevelData<EBCellFAB>& a_residual,
			      const LevelData<EBCellFAB>& a_phiFine,
			      const LevelData<EBCellFAB>& a_phi)
{
  Interval interv(0,0);
  CH_assert(a_phiFine.nComp() == 1);
  int ncomp = 1;
  if( !m_fluxReg.isDefined() )
  {
     const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
     m_fluxReg.define(m_dblFine,     m_dbl,
                       m_ebislFine, m_ebisl,
                       m_domain.domainBox(),
                       m_refToFine, ncomp,ebisPtr);
  }
  m_fluxReg.setToZero();
  incrementFRCoar(m_fluxReg, a_phiFine, a_phi);
  incrementFRFine(m_fluxReg, a_phiFine, a_phi);
  for( int idir = 0; idir < SpaceDim; idir++ )
  {
     Real scale = 1.0/m_dx[idir];
     m_fluxReg.reflux(a_residual, interv, scale, idir);
  }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::incrementFRCoar(EBFluxRegister&             a_fluxReg,
				       const LevelData<EBCellFAB>& a_phiFine,
				       const LevelData<EBCellFAB>& a_phi)
{
  CH_assert(a_phiFine.nComp() == 1);
  CH_assert(a_phi.nComp() == 1);

  int ncomp = 1;
  Interval interv(0,0);

  for(DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
    {
      const EBCellFAB& coarfab = a_phi[dit()];
      const EBISBox& ebisBox = m_ebisl[dit()];
      const Box&  box = m_dbl.get(dit());
      for (int idir = 0; idir < SpaceDim; idir++)
        {
          EBFaceFAB coarflux(ebisBox, box, idir, ncomp);
          getFlux(coarflux, coarfab, box, ebisBox, m_dx, idir, m_kappa[dit()] );

          Real scale = 1.0;
          if( m_rz && idir == 0 )
	    incrementCoarseBothRZ( a_fluxReg, coarflux, scale,
                                   dit(), interv, idir, m_dx[0] );
	  else
            a_fluxReg.incrementCoarseBoth(coarflux, scale,
                                        dit(), interv, idir);
        }
    }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::incrementFRFine(EBFluxRegister&             a_fluxReg,
				       const LevelData<EBCellFAB>& a_phiFine,
				       const LevelData<EBCellFAB>& a_phi)
{
  CH_assert(a_phiFine.nComp() == 1);
  CH_assert(a_phi.nComp() == 1);
  CH_assert(m_hasFine);
  int ncomp = 1;
  Interval interv(0,0);

  // need to do cfinterpolation in aggregate before
  // any getFineData calls are made
  //define CF interpolation object with finer level
  //define coarse fine interpolation object on the fly
  //because most operators do not need it
  EBQuadCFInterp quadCFIWithFine(m_dblFine,     m_dbl,
                                 m_ebislFine, m_ebisl,
                                 m_domain.domainBox(),
                                 m_refToFine, ncomp);

  //ghost cells of phiFine need to be filled
  LevelData<EBCellFAB>& phiFine = (LevelData<EBCellFAB>&) a_phiFine;
  quadCFIWithFine.interpolate(phiFine, a_phi, interv);
  phiFine.exchange(interv);

  DataIterator ditf = a_phiFine.dataIterator();
  for( ditf.reset(); ditf.ok(); ++ditf )
  {
      const Box&     boxFine = m_dblFine.get(ditf());
      const EBISBox& ebisBoxFine = m_ebislFine[ditf()];
      const EBCellFAB& phiFine = a_phiFine[ditf()];
      for( int idir = 0; idir < SpaceDim; idir++ )
      {
          SideIterator sit;
          for (sit.begin(); sit.ok(); sit.next())
          {
              Box fabBox = adjCellBox(boxFine, idir, sit(), 1);
              fabBox.shift(idir, -sign(sit()));

              EBFaceFAB fluxFine(ebisBoxFine, fabBox, idir, ncomp);
            getFlux(fluxFine, phiFine, fabBox, ebisBoxFine, m_dxFine, idir, m_kappaFine[ditf()] );

              Real scale = 1.0;
              if( m_rz && idir == 0 )
                 incrementFineBothRZ( a_fluxReg, fluxFine, scale, ditf(),
                                      interv,  idir,    sit(), m_dxFine[0] );
              else
                 a_fluxReg.incrementFineBoth(fluxFine, scale, ditf(),
                                          interv,  idir,    sit());
          }
      }
  }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::incrementCoarseBothRZ(
				    EBFluxRegister& a_fluxReg,
				    const EBFaceFAB& a_coarFlux,
                                    const Real&      a_scale,
                                    const DataIndex& a_coarDatInd,
                                    const Interval&  a_variables,
                                    const int&       a_dir,
				    const Real&      a_dx)
{
  a_fluxReg.incrementCoarseRegulRZ(a_coarFlux, a_scale, a_coarDatInd, 
				     a_variables, a_dir, a_dx);
  
  IntVectSet ivs(a_coarFlux.getCellRegion());
  const EBISBox& ebisBoxCoar = a_coarFlux.getEBISBox();
  BaseIFFAB<Real> irregFlux(ivs, ebisBoxCoar.getEBGraph(), a_dir, a_coarFlux.nComp());

  for(FaceIterator faceit(ivs, ebisBoxCoar.getEBGraph(), a_dir,
                          FaceStop::SurroundingWithBoundary);
      faceit.ok(); ++faceit)
    {
      for(int icomp = 0; icomp < a_coarFlux.nComp(); icomp++)
        {
          irregFlux(faceit(), icomp) = a_coarFlux(faceit(), icomp);
        }
    }
  a_fluxReg.incrementCoarseIrregulRZ(irregFlux, a_scale, a_coarDatInd,
				       a_variables, a_dir, a_dx );
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::incrementFineBothRZ(
				  EBFluxRegister& a_fluxReg,
				  const EBFaceFAB&      a_fineFlux,
                                  const Real&           a_scale,
                                  const DataIndex&      a_fineDatInd,
                                  const Interval&       a_variables,
                                  const int&            a_dir,
                                  const Side::LoHiSide& a_sd,
                                  const Real&           a_dx )
{
  a_fluxReg.incrementFineRegulRZ(a_fineFlux, a_scale, a_fineDatInd,
				   a_variables, a_dir, a_sd, a_dx);

  IntVectSet ivs(a_fineFlux.getCellRegion());
  const EBISBox& ebisBoxFine = a_fineFlux.getEBISBox();
  BaseIFFAB<Real> irregFlux(ivs, ebisBoxFine.getEBGraph(), a_dir, a_fineFlux.nComp());

  for(FaceIterator faceit(ivs, ebisBoxFine.getEBGraph(), a_dir,
                          FaceStop::SurroundingWithBoundary);
      faceit.ok(); ++faceit)
    {
      for(int icomp = 0; icomp < a_fineFlux.nComp(); icomp++)
        {
          irregFlux(faceit(), icomp) = a_fineFlux(faceit(), icomp);
        }
    }
  a_fluxReg.incrementFineIrregulRZ(irregFlux, a_scale, a_fineDatInd,
				     a_variables, a_dir, a_sd, a_dx);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::getFlux(EBFaceFAB&       a_flux,
			       const EBCellFAB& a_phi,
			       const Box&       a_box,
			       const EBISBox&   a_ebisBox,
			       const RealVect&  a_dx,
			       const int&       a_idir,
			       const EBCellFAB& a_kappa )
{
  CH_assert(a_flux.nComp() == 1);
  CH_assert(a_phi.nComp() == 1);
  IntVectSet ivsCell(a_flux.getCellRegion());

  for( FaceIterator faceit(ivsCell, a_ebisBox.getEBGraph(), a_idir,
                     FaceStop::SurroundingNoBoundary); faceit.ok(); ++faceit)
                                                  
  {
     const FaceIndex& face = faceit();
     Real phiHi = a_phi(face.getVoF(Side::Hi), 0);
     Real phiLo = a_phi(face.getVoF(Side::Lo), 0);
     Real kap = 0.5*(a_kappa(face.getVoF(Side::Hi),0)+a_kappa(face.getVoF(Side::Lo),0));
     Real fluxFace = kap*(phiHi - phiLo)/a_dx[a_idir];
     a_flux(face, 0) = fluxFace;
     if( a_idir == 0 && m_rz )
     {
        IntVect ind= (face.getVoF(Side::Hi).gridIndex()+
		      face.getVoF(Side::Lo).gridIndex()  );
        Real rf = (ind[0]*0.5)*a_dx[0];
        a_flux(face,0) *= rf;
     }
  }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::AMRResidualNC(LevelData<EBCellFAB>&       a_residual,
				     const LevelData<EBCellFAB>& a_phiFine,
				     const LevelData<EBCellFAB>& a_phi,
				     const LevelData<EBCellFAB>& a_rhs,
				     bool a_homogeneousPhysBC)
{
  //dummy. there is no coarse when this is called
  LevelData<EBCellFAB> phiC;
  AMRResidual(a_residual, a_phiFine, a_phi, phiC, a_rhs, a_homogeneousPhysBC);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::AMRResidualNF(LevelData<EBCellFAB>&       a_residual,
				     const LevelData<EBCellFAB>& a_phi,
				     const LevelData<EBCellFAB>& a_phiCoar,
				     const LevelData<EBCellFAB>& a_rhs,
				     bool a_homogeneousPhysBC)
{
  applyOp(a_residual,a_phi, &a_phiCoar,  a_homogeneousPhysBC, false);
  scale(a_residual,-1.0);
  incr(a_residual,a_rhs,1.0);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::AMRRestrict(LevelData<EBCellFAB>&       a_resCoar,
				   const LevelData<EBCellFAB>& a_residual,
				   const LevelData<EBCellFAB>& a_correction,
				   const LevelData<EBCellFAB>& a_coarCorrection)
{
  CH_assert(a_residual.nComp() == 1);
  CH_assert(a_resCoar.nComp() == 1);
  CH_assert(a_correction.nComp() == 1);

  LevelData<EBCellFAB> resThisLevel;
  bool homogeneousPhys = true;
  bool homogeneousCF =   false;

  EBCellFactory ebcellfactTL(m_ebisl);
  IntVect ghostVec = a_residual.ghostVect();

  resThisLevel.define(m_dbl, 1, ghostVec, ebcellfactTL);

  //API says that we must average(a_residual - L(correction, coarCorrection))
  applyOp(resThisLevel, a_correction, &a_coarCorrection, homogeneousPhys, homogeneousCF);
  incr(resThisLevel, a_residual, -1.0);
  scale(resThisLevel,-1.0);

  //use our nifty averaging operator
  Interval variables(0, 0);
  CH_assert(m_hasInterpAve);
  m_ebAverage.average(a_resCoar, resThisLevel, variables);

}


//-----------------------------------------------------------------------
Real EBAMRHelmholtzOp::AMRNorm(const LevelData<EBCellFAB>& a_coarResid,
			       const LevelData<EBCellFAB>& a_fineResid,
			       const int& a_refRat,
			       const int& a_ord)

{
  int ncomp = a_coarResid.nComp();
  Interval interv(0, ncomp-1);
  const DisjointBoxLayout& coarGrids = a_coarResid.disjointBoxLayout();
  const DisjointBoxLayout& fineGrids = a_fineResid.disjointBoxLayout();
  CH_assert(coarGrids == m_dbl);

  //create temp and zero out under finer grids
  EBCellFactory ebcellfact(m_ebisl);
  LevelData<EBCellFAB> coarTemp(coarGrids, ncomp, IntVect::Zero, ebcellfact);
  a_coarResid.copyTo(interv, coarTemp, interv);

  for(DataIterator dit = coarGrids.dataIterator(); dit.ok(); ++dit)
    {
      EBCellFAB& coarTempFAB = coarTemp[dit()];
      const EBISBox& ebisBox = m_ebisl[dit()];

      LayoutIterator litFine = fineGrids.layoutIterator();
      for (litFine.reset(); litFine.ok(); ++litFine)
        {
          Box overlayBox = coarTempFAB.box();
          Box coarsenedGrid = coarsen(fineGrids[litFine()], a_refRat);

          overlayBox &= coarsenedGrid;
          if (!overlayBox.isEmpty())
            {
              IntVectSet ivsZero(overlayBox);

              for(VoFIterator vofit(ivsZero, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
                {
                  for(int ivar =0; ivar < ncomp; ivar++)
                    {
                      coarTempFAB(vofit(), ivar) = 0.0;
                    }
                }
            }
        }
    }
  //return norm of temp
  return norm(coarTemp, a_ord);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::AMRProlong(LevelData<EBCellFAB>&       a_correction,
				  const LevelData<EBCellFAB>& a_coarCorrection)
{
  //use cached interpolation object
  Interval variables(0, 0);
  CH_assert(m_hasInterpAve);
  m_ebInterp.pwcInterp(a_correction, a_coarCorrection, variables);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::AMRUpdateResidual(LevelData<EBCellFAB>&       a_residual,
					 const LevelData<EBCellFAB>& a_correction,
					 const LevelData<EBCellFAB>& a_coarCorrection)
{
  LevelData<EBCellFAB> lcorr;
  bool homogeneousPhys = true;
  bool homogeneousCF   = false;

  EBCellFactory ebcellfactTL(m_ebisl);
  IntVect ghostVec = a_residual.ghostVect();

  lcorr.define(m_dbl, 1, ghostVec, ebcellfactTL);

  applyOp(lcorr, a_correction, &a_coarCorrection, homogeneousPhys, homogeneousCF);

  incr(a_residual, lcorr, -1);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::getOpVoFStencil(VoFStencil&     a_stencil,
				       const EBISBox&  a_ebisbox,
				       const VolIndex& a_VoF)
{
  a_stencil.clear();

  Vector<VolIndex> allMonotoneVoFs;
  int radius = 1;
  IntVect timesMoved = IntVect::Zero;
  IntVect pathSign   = IntVect::Zero;


// This routine seems to find all vofs that can connect to a_VoF by radius
// moves along coord dirs, i.e., all points (i+l,j+m,k+n) where
//  |l| <= radius |m| <= radius |n| <= radius. timesMoved and pathSign seems
// to be variables that keeps track on the number of moves, think they
// normally should be set equal to zero.

  EBArith::getAllVoFsInMonotonePath(allMonotoneVoFs,timesMoved,pathSign,
                                    a_VoF,a_ebisbox,radius);

  for (int idir = 0; idir < SpaceDim; idir++)
    {
      for (SideIterator sit; sit.ok(); ++sit)
        {
          Side::LoHiSide side = sit();
          Vector<FaceIndex> faces;

          faces = a_ebisbox.getFaces(a_VoF,idir,side);

          for (int iface = 0; iface < faces.size(); iface++)
            {
              FaceIndex face = faces[iface];
              VoFStencil faceStencil;

              getOpFaceStencil(faceStencil,allMonotoneVoFs,a_ebisbox,a_VoF,
                               idir,side,face);

              a_stencil += faceStencil;
            }
        }
    }
}

//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::getOpFaceStencil(VoFStencil&             a_stencil,
					const Vector<VolIndex>& a_allMonotoneVoFs,
					const EBISBox&          a_ebisbox,
					const VolIndex&         a_VoF,
					int                     a_dir,
					const Side::LoHiSide&   a_side,
					const FaceIndex&        a_face)
{
  a_stencil.clear();

  const RealVect& faceCentroid = a_ebisbox.centroid(a_face);
  IntVect dirs = IntVect::Zero;
  for( int idir = 0; idir < SpaceDim; idir++ )
  {
     if (idir != a_dir)
        dirs[idir] = faceCentroid[idir] < 0 ? -1 : 1;
  }

  bool orderOne = true;
  if (m_orderEB == 0)
  {
     orderOne = false;
  }
  else
  {
      IntVect loVect = dirs;
      loVect.min(IntVect::Zero);

      IntVect hiVect = dirs;
      hiVect.max(IntVect::Zero);

      IntVect origin = a_VoF.gridIndex();

      Box fluxBox(loVect,hiVect);
      IntVectSet fluxIVS(fluxBox);

      IVSIterator ivsit(fluxIVS);
      for (ivsit.begin(); ivsit.ok(); ++ivsit)
        {
          bool curOkay;

          IntVect ivDelta = ivsit();

          IntVect iv1 = ivDelta;
          iv1 += origin;

          VolIndex VoF1;

// This call seems to take iv1 as input and return VoF1 if iv1 corresponds
// to a single VoF in the list all_montone_vofs.
          curOkay = EBArith::isVoFHere(VoF1,a_allMonotoneVoFs,iv1);

          IntVect iv2 = iv1;
          iv2[a_dir] += sign(a_side);

          VolIndex VoF2;

          curOkay = curOkay && EBArith::isVoFHere(VoF2,a_allMonotoneVoFs,iv2);

          orderOne = orderOne && curOkay;

          if (curOkay)
            {
              VoFStencil curPair;

              if( m_rz && a_dir == 0 )
	      {
		 Real rc1 = ( VoF1.gridIndex()[0]+0.5)*m_dx[0];
		 Real rc2 = ( VoF2.gridIndex()[0]+0.5)*m_dx[0];
		 Real rh =  (rc1+rc2)*0.5;
                 rc1 += m_dx[0]*a_ebisbox.centroid(VoF1)[0];
                 rc2 += m_dx[0]*a_ebisbox.centroid(VoF2)[0];
                 curPair.add(VoF1,-rh*m_invDx2[a_dir]/rc1);
                 curPair.add(VoF2, rh*m_invDx2[a_dir]/rc2);
	      }
	      else
	      {
                 curPair.add(VoF1,-m_invDx2[a_dir]);
                 curPair.add(VoF2, m_invDx2[a_dir]);
	      }
              for (int idir = 0; idir < SpaceDim; idir++)
                {
                  if (idir != a_dir)
                    {
                      if (ivDelta[idir] == 0)
                        {
                          curPair *= 1 - fabs(faceCentroid[idir]);
                        }
                      else
                        {
                          curPair *= fabs(faceCentroid[idir]);
                        }
                    }
                }
              a_stencil += curPair;
            }
        }
    }

  if (!orderOne)
  {
      VolIndex VoF1 = a_VoF;
      VolIndex VoF2 = a_face.getVoF(a_side);
      //      cout << "Order one failed " << a_VoF.gridIndex() << endl;
      if (a_ebisbox.getRegion().contains(VoF2.gridIndex()))
      {
          a_stencil.clear();
          if( m_rz && a_dir == 0 )
	  {
	     Real rc1 = ( VoF1.gridIndex()[0]+0.5)*m_dx[0];
	     Real rc2 = ( VoF2.gridIndex()[0]+0.5)*m_dx[0];
	     Real rh =  (rc1+rc2)*0.5;
             rc1 += m_dx[0]*a_ebisbox.centroid(VoF1)[0];
             rc2 += m_dx[0]*a_ebisbox.centroid(VoF2)[0];
             a_stencil.add(VoF1,-rh*m_invDx2[a_dir]/rc1);
             a_stencil.add(VoF2, rh*m_invDx2[a_dir]/rc2);
	  }
	  else
	  {
             a_stencil.add(VoF1,-m_invDx2[a_dir]);
             a_stencil.add(VoF2, m_invDx2[a_dir]);
	  }
      }
  }
  a_stencil *= a_ebisbox.areaFrac(a_face);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::get_ustencil( VoFStencil&     a_stencil,
				     Real&           a_wgh,
				     const EBISBox&  a_ebisbox,
				     const VolIndex& a_VoF )
{
  a_stencil.clear();
  Vector<VolIndex> allMonotoneVoFs;
  int radius = 1;
  IntVect timesMoved = IntVect::Zero;
  IntVect pathSign   = IntVect::Zero;

  EBArith::getAllVoFsInMonotonePath( allMonotoneVoFs,timesMoved,pathSign,
                                    a_VoF,a_ebisbox,radius);
  a_wgh  = 1;
  int no = 0;

  RealVect dists = a_ebisbox.centroid(a_VoF);
  for( int dir=0 ; dir < SpaceDim ; dir++ )   
  {
     Vector<VolIndex> vofl = a_ebisbox.getVoFs( a_VoF, dir, Side::Lo,1);
     Vector<VolIndex> vofh = a_ebisbox.getVoFs( a_VoF, dir, Side::Hi,1);
     if( vofl.size() > 0 && dists[dir] < 0 )
     {
	a_stencil.add(vofl[0],-dists[dir]);
	a_wgh += dists[dir];
        no++;
     }
     else if( vofh.size() > 0 && dists[dir] >= 0 )
     {
	a_stencil.add(vofh[0],dists[dir]);
	a_wgh -= dists[dir];
        no++;
     }  
  }
  //  if( no != SpaceDim )
  //  {
    //     MayDay::Warning("EBAMRHelmholtzOp::get_ustencil. Could not find u-stencil");
    //     cout << "no =    " << no << endl;
    //     cout << "a_wgh = " << a_wgh << endl;
  //  }
  a_stencil.add(a_VoF,a_wgh);


  //  IntVect  iv    = a_VoF.gridIndex();
  //  RealVect dists = a_ebisbox.centroid(a_VoF);

  //  for( int dir=0 ; dir < SpaceDim ; dir++ )   
  //  {
  //     IntVect of = IntVect::Zero;
  //     for( int ud=-1; ud <=1; ud+=2)
  //     {
  //        of[dir] = ud;
  //        VolIndex another_vof;
  //        if( EBArith::isVoFHere(another_vof,allMonotoneVoFs,iv+of) )
  //	{
  //	   a_stencil.add(another_vof,dists[dir]*ud);
  //	   a_wgh -= dists[dir]*ud;
  //           no++;
  //	   break;
  //	}
  //     }
  //  }
  //  if( no != SpaceDim )
    //     MayDay::Warning("EBAMRHelmholtz::get_ustencil. Could not find u-stencil");

  //  a_stencil.add(a_VoF,a_wgh);

  //  cout << "ustenc at " << a_VoF.gridIndex() << " center " << a_ebisbox.centroid(a_VoF) << " volfrac " << a_ebisbox.volFrac(a_VoF) << endl;
  //  for( int i=0 ; i<a_stencil.size() ; i++ )
  //    cout << a_stencil.vof(i).gridIndex() << " wgh= " << a_stencil.weight(i) << endl;
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::levelMulticolorGS(LevelData<EBCellFAB>&       a_phi,
					 const LevelData<EBCellFAB>& a_rhs)
{
  int comp  = 0;
  int ncomp = 1;

  LevelData<EBCellFAB> residLD;
  bool homogeneous = m_homogeneous_relax;

  EBCellFactory ebcellfact(m_ebisl);
  residLD.define(m_dbl, a_rhs.nComp(),a_rhs.ghostVect(), ebcellfact);

  IntVect color = IntVect::Zero;
  IntVect limit = IntVect::Unit;
  Real weight = 0.0;

  for (int idir = 0; idir < SpaceDim; idir++)
    {
      weight += -2.0 / (m_dx[idir]*m_dx[idir]);
    }

  // Loop over all possibilities (in all dimensions)
  while (color[SpaceDim-1] <= limit[SpaceDim-1])
    {
      // Get the residual everywhere
      residual(residLD,a_phi,a_rhs,homogeneous);
     cout << "GS it " << norm( residLD, 0 ) << endl;
      int ibox = 0;
      for (DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
        {
          EBCellFAB& phi = a_phi[dit()];

          const EBISBox& ebisBox = m_ebisl[dit()];
          const Box& box   = m_dbl.get(dit());

          EBCellFAB phiSave(phi.getEBISBox(),
                            box,
                            ncomp);
          phiSave.copy(phi);
	  
          const EBCellFAB& wgh = m_wgh[dit()];
	  //          EBCellFAB wgh(phi.getEBISBox(), box, ncomp );
	  //          wgh.setVal(1.0/(weight-m_k2));
	  //          for( VoFIterator vofit( m_ebisl[dit()].getIrregIVS(m_ebisl[dit()].getDomain()) ,m_ebisl[dit()].getEBGraph()) ; vofit.ok(); ++vofit )
	  //              wgh(vofit(),0) = 1.0/(m_weight[dit](vofit(),0)-m_ebisl[dit()].volFrac(vofit())*m_k2*m_uweight[dit()](vofit(),0));

          const EBCellFAB& resid = residLD[dit()];

          // Do the regular cells
          BaseFab<Real>& phiFAB         =     phi.getSingleValuedFAB();
          BaseFab<Real>& phiSaveFAB     = phiSave.getSingleValuedFAB();
          const BaseFab<Real>& residFAB =   resid.getSingleValuedFAB();

          IntVect loIV = box.smallEnd();
          IntVect hiIV = box.bigEnd();

          for (int idir = 0; idir < SpaceDim; idir++)
            {
              if (loIV[idir] % 2 != color[idir])
                {
                  loIV[idir]++;
                }
            }

          if (loIV <= hiIV)
            {
              Box coloredBox(loIV,hiIV);
              FORT_REGMULTICOLORGSHH(CHF_FRA1(phiFAB,comp),
                                   CHF_CONST_FRA1(phiSaveFAB,comp),
                                   CHF_CONST_FRA1(residFAB,comp),
                                   CHF_BOX(coloredBox), 
                                   CHF_CONST_FRA1(wgh.getSingleValuedFAB(),0) );
            }

          // Do the irregular cells
          const BaseIVFAB<Real>& curWeightBaseIVFAB = m_weight[dit()];
          const IntVectSet& curIrreg = curWeightBaseIVFAB.getIVS();
          for (VoFIterator vofit(curIrreg,ebisBox.getEBGraph()); vofit.ok(); ++vofit)
            {
              const VolIndex& VoF = vofit();
              const IntVect& iv = VoF.gridIndex();

              bool ok = true;
              for (int idir = 0; idir < SpaceDim; idir++)
                {
                  if (iv[idir] % 2 != color[idir])
                    {
                      ok = false;
                      break;
                    }
                }

              if (ok)
                {
		  //                  const Real& weight = curWeightBaseIVFAB(VoF,comp);

                  phi(VoF,comp) = phiSave(VoF,comp)
                    + wgh(VoF,comp) * resid(VoF,comp);
                }
            }
          ibox++;
        }

      color[0]++;

      // If greater that the maximum it can be...
      if (color[0] > limit[0])
        {
          // For all dimensions...
          for (int i = 0; i < SpaceDim-1; i++)
            {
              color[i] = 0;
              color[i+1]++;

              // If the next dimension's power is okay go on
              if (color[i+1] <= limit[i+1])
                {
                  break;
                }
            }
        }
    }
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::levelJacobi(LevelData<EBCellFAB>&       a_phi,
				   const LevelData<EBCellFAB>& a_rhs)
{
  LevelData<EBCellFAB> resid;
  bool homogeneous = m_homogeneous_relax;

  EBCellFactory ebcellfact(m_ebisl);
  resid.define(m_dbl, a_rhs.nComp(),a_rhs.ghostVect(), ebcellfact);
  LevelData<EBCellFAB> tmp(m_dbl,a_phi.nComp(),a_phi.ghostVect(),ebcellfact);

  for( int nit = 0 ; nit < 5 ; nit++ )
  {
     residual(resid,a_phi,a_rhs,homogeneous);
     //     cout << "Jacobi it " << norm( resid, 0 ) << " " << norm(a_phi,0) << endl;
     //    cout <<"-----------------New iteration--------------------"<<endl;
     //     cout << "k2 = " << m_k2 << endl;
     //     DataIterator d=a_rhs.dataIterator();

//     print_ebcellfab(m_kappa[d],"kap");
//     print_ebcellfab(m_wgh[d],"wgh");
     EBLevelDataOps::product( tmp, m_wgh, resid );
     EBLevelDataOps::sum( a_phi, a_phi, tmp );
     //     print_ebcellfab(a_phi[d],"u");
  }
}

//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::set_homogeneous_relax()
{
  m_homogeneous_relax = true;
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::set_inhomogeneous_relax()
{
  m_homogeneous_relax = false;
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::set_k2( Real a_k2 )
{
  m_k2 = a_k2;
  compute_diagonal_weight();
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::set_rho( const LevelData<EBCellFAB>& a_rho )
{
   EBLevelDataOps::assign( m_rho, a_rho );
   compute_diagonal_weight();
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::set_rho_zero( )
{
 EBLevelDataOps::setToZero( m_rho );
 compute_diagonal_weight();
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::compute_diagonal_weight()
{

// Compute coefficient in front of u_i in equation nr. i, needed
// for Jacobi- and Gauss-Seidel relaxation schemes.

   EBCellFactory cfactory( m_ebisl );
   m_wgh.define( m_dbl, 1, IntVect::Zero, cfactory );

   int ncomp = m_wgh.nComp();
   for (DataIterator dit = m_wgh.dataIterator(); dit.ok(); ++dit)
   {
      const DataIndex& dex = dit();
      const Box& bx = m_wgh[dex].box();
      const EBISBox& ebis = m_wgh[dex].getEBISBox();
      IntVectSet notRegular = m_ebisl[dex].getIrregIVS(m_dbl.get(dex));

// Define boundary boxes
      Box loBox[SpaceDim],hiBox[SpaceDim];
      int hasLo[SpaceDim],hasHi[SpaceDim];
      for (int idir = 0; idir < SpaceDim; idir++)
      {
         Box centerBox,entireBox;
         ebEllipticLoHiCenter(loBox[idir],hasLo[idir],hiBox[idir],hasHi[idir],
                              centerBox,entireBox,m_dbl.get(dex),bx,
                              m_domain,idir);
      }

// Compute weight in interior regular cells
      if( m_rz )
         FORT_GET_WGH_RZ(CHF_FRA1(m_wgh[dex].getSingleValuedFAB(),0),
			CHF_CONST_FRA1(m_kappa[dex].getSingleValuedFAB(),0),
			CHF_CONST_REALVECT(m_dx), CHF_CONST_REALVECT(m_origin),
			CHF_BOX(bx), CHF_CONST_REAL( m_k2 ),
			CHF_CONST_FRA1( m_rho[dex].getSingleValuedFAB(),0) );
      else
         FORT_GET_WGH(CHF_FRA1(m_wgh[dex].getSingleValuedFAB(),0),
		     CHF_CONST_FRA1(m_kappa[dex].getSingleValuedFAB(),0),
		     CHF_CONST_REALVECT(m_dx), CHF_BOX(bx), CHF_CONST_REAL(m_k2), 
		     CHF_CONST_FRA1( m_rho[dex].getSingleValuedFAB(),0) );

// Contribution from BC at regular cells
      for( int idir = 0; idir < SpaceDim; idir++ )
      {
         if( hasLo[idir] == 1 )
         {
            Vector<Real> wgh(ncomp);
            for( BoxIterator bi(loBox[idir]) ; bi.ok() ; ++bi )
	    {
	       Vector<VolIndex> vofs = ebis.getVoFs(bi());
               if( vofs.size() > 0 )
                  m_domainBC->get_diagonal_weight( wgh, vofs[0],ebis,m_origin,m_dx,idir,Side::Lo );
	       for( int c = 0 ; c < ncomp ; c++ )
	       {
	          m_wgh[dex].getSingleValuedFAB()(bi(),c) -= m_invDx[idir]*
	                            m_kappa[dex].getSingleValuedFAB()(bi(),0)*wgh[c];
	       }
	    }
         }
         if( hasHi[idir] == 1 )
         {
            Vector<Real> wgh(ncomp);
            for( BoxIterator bi(hiBox[idir]) ; bi.ok() ; ++bi )
   	    {
	      Vector<VolIndex> vofs = ebis.getVoFs(bi());
              if( vofs.size() > 0 )
                 m_domainBC->get_diagonal_weight( wgh,vofs[0],ebis, m_origin, m_dx, idir, Side::Hi );
	      for( int c = 0 ; c < ncomp ; c++ )
	      {
	         m_wgh[dex].getSingleValuedFAB()(bi(),c) += m_invDx[idir]*
	                            m_kappa[dex].getSingleValuedFAB()(bi(),0)*wgh[c];
	      }
	    }
         }
      }

// Weights at interior irregular cells
     EBCellFAB wgh(m_ebisl[dex],bx,1);
     wgh.setVal(0.0);
     compute_dw_irregular( wgh, m_ebisl[dex], m_iflux_stencil[dex],
                           m_kappa[dex], 0 );

     compute_dw_irregular( wgh, m_ebisl[dex], m_jflux_stencil[dex],
                           m_kappa[dex], 1 );
     if( SpaceDim == 3 )
        compute_dw_irregular( wgh, m_ebisl[dex], m_kflux_stencil[dex],
                           m_kappa[dex], 2 );

// Boundary irregular cells and the Helmholtz low order term:
     for (VoFIterator vofit(notRegular,m_ebisl[dex].getEBGraph()); vofit.ok(); ++vofit)
     {
        const VolIndex& vof = vofit();
	//        const IntVect& iv = vof.gridIndex();
        for( int idir = 0; idir < SpaceDim; idir++ )
        {
   	   Vector<Real> wghloc(ncomp);
           if( hasLo[idir] == 1 && loBox[idir].contains( vof.gridIndex()) )
           {
              m_domainBC->get_diagonal_weight( wghloc, vof, ebis, m_origin, m_dx,
					       idir, Side::Lo );
	      for( int c = 0 ; c < ncomp ; c++ )
	         wgh(vof,c) -= m_invDx[idir]*m_kappa[dex](vof,0)*wghloc[c];
	   }
           if( hasHi[idir] == 1 && hiBox[idir].contains( vof.gridIndex()) )
           {
              m_domainBC->get_diagonal_weight( wghloc, vof, ebis, m_origin, m_dx,
					       idir, Side::Hi );
	      for( int c = 0 ; c < ncomp ; c++ )
	         wgh(vof,c) += m_invDx[idir]*m_kappa[dex](vof,0)*wghloc[c];
	   }
        }
	if( m_split_u )
           m_wgh[dex](vof,0) =wgh(vof,0)-m_ebisl[dex].volFrac(vof)*
	                       m_k2*m_uweight[dex](vof,0)*m_rho[dex](vof,0);
	else
           m_wgh[dex](vof,0) =wgh(vof,0)-m_ebisl[dex].volFrac(vof)*
	                       m_k2*m_rho[dex](vof,0);
     }

// Transform weight into form suitable for relaxation schemes
     Real relpar;
     if( m_relaxType == 0 )
        relpar = 0.5;
     else
        relpar = 1.0;

     for( BoxIterator b(bx); b.ok(); ++b )
     {
        if( !m_ebisl[dex].isCovered(b()) )
        {
           Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
           for( int iv = 0 ; iv < vofs.size() ; iv++ )          
              m_wgh[dex](vofs[iv],0) = relpar/m_wgh[dex](vofs[iv],0);
        }
     }

  }
}


//-----------------------------------------------------------------------
Real EBAMRHelmholtzOp::get_diagonal_weight( const DataIndex& dex,
					    const VolIndex& vof ) const
{
   return m_wgh[dex](vof,0);
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOp::set_time( Real a_time )
{
   m_time = a_time;
}


//-----------------------------------------------------------------------
// EBAMR HELMHOLTZ FACTORY 
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
EBAMRHelmholtzOpFactory::~EBAMRHelmholtzOpFactory()
{
  for(int ilev = 0; ilev < m_numLevels; ilev++)
    delete m_rhoPtr[ilev];
}


//-----------------------------------------------------------------------
int EBAMRHelmholtzOpFactory::refToFiner(const ProblemDomain& a_domain) const
{
  int retval = -1;
  bool found = false;
  for(int ilev = 0; ilev < m_domainVec.size(); ilev++)
    {
      if(m_domainVec[ilev].domainBox() == a_domain.domainBox())
        {
          retval = m_refRatioVec[ilev];
          found = true;
        }
    }
  if(!found)
    {
      MayDay::Error("Domain not found in AMR hierarchy");
    }
  return retval;
}


//-----------------------------------------------------------------------
EBAMRHelmholtzOpFactory::EBAMRHelmholtzOpFactory(
		      const Vector<DisjointBoxLayout>&      a_grids,
                      const Vector<EBISLayout>&             a_ebisl,
                      const Vector<int>&                    a_refRatio,
                      const ProblemDomain&                  a_domainCoar,
                      const RealVect&                       a_dxCoar,
                      const RealVect&                       a_origin,
		      const Real                            a_k2,
                      const int&                            a_orderEB,
                      const int&                            a_relaxType,
		      //                      RefCountedPtr<BaseDomainBCFactory>    a_domainBCFactory,
		      // 		      RefCountedPtr<BaseEBBCFactory>        a_ebBCFactory,
                      RefCountedPtr<HelmholtzDomainBCFactory>    a_domainBCFactory,
 		      RefCountedPtr<HelmholtzEBBCFactory>        a_ebBCFactory,
                      const Real                            a_kappa,
	              const Vector<LevelData<EBCellFAB>*>&  a_rho,
                      const bool&                           a_rz )

{
  CH_assert(a_grids.size() <= a_refRatio.size());
  CH_assert( a_k2 >= 0 );

  m_time            = 0.0;
  m_numLevels       = a_grids.size();
  m_domainBCFactory = a_domainBCFactory;
  m_ebBCFactory     = a_ebBCFactory;
  m_gridVec         = a_grids;
  m_ebislVec        = a_ebisl;
  m_refRatioVec     = a_refRatio;
  m_orderEB         = a_orderEB;
  m_relaxType       = a_relaxType;
  m_origin          = a_origin;
  m_rz              = a_rz;
  m_kappa           = a_kappa;

  m_rhoPtr.resize(m_numLevels);
  m_k2.resize(m_numLevels);
  for( int i=0 ; i < m_numLevels ; i++ )
  {
     m_rhoPtr[i] = new LevelData<EBCellFAB>();
     EBCellFactory cfactory( a_ebisl[i] );
     m_rhoPtr[i]->define( a_grids[i], 1, IntVect::Unit, cfactory );
     EBLevelDataOps::assign( *(m_rhoPtr[i]), *(a_rho[i]) );
     m_k2[i] = a_k2;
  }
  m_dxVec.resize(m_numLevels);
  m_domainVec.resize(m_numLevels);
  m_domainVec[0] = a_domainCoar;
  m_dxVec[0]     = a_dxCoar;
  for( int ilev = 1; ilev < m_numLevels; ilev++ )
  {
     m_domainVec[ilev] = refine(m_domainVec[ilev-1], m_refRatioVec[ilev-1]);
     m_dxVec[ilev] = m_dxVec[ilev-1]/m_refRatioVec[ilev-1];
  }
}


//-----------------------------------------------------------------------
// called by MultiGrid. There is one MultiGrid-object for each AMR-level.
// a_domainFine is the domain where we want to solve the problem 
// (finest multi grid domain). a_depth is the number of coarsenings
// the returned operator should correspond to.
// The multi grid hierarchy is set up by successive calls with
// a_depth =0,1,2,.., but with the same a_domainFine.
//-----------------------------------------------------------------------
EBAMRHelmholtzOp* EBAMRHelmholtzOpFactory::MGnewOp(
			       const ProblemDomain& a_domainFine,
                               int                  a_depth,
                               bool                 a_homoOnly )
{

  //find out if there is a real starting point here.
  int whichlev;
  bool found = false;

  for( int ilev = 0; ilev < m_numLevels && !found; ilev++ )
  {
     if( a_domainFine.domainBox() == m_domainVec[ilev].domainBox() )
     {
        found    = true;
        whichlev = ilev;
     }
  }
  if(!found)
  {
     MayDay::Error("No corresponding AMRLevel to starting point of MGnewOp");
  }


  //multigrid operator.  coarn by two from depth  no
  //coarse or finer to worry about.
  ProblemDomain  domainMGLevel;
  DisjointBoxLayout dblMGLevel;
  EBISLayout      ebislMGLevel;
  RealVect           dxMGLevel;
  if(a_depth == 0)
  {
     domainMGLevel = m_domainVec[whichlev];
     dblMGLevel    = m_gridVec[whichlev];
     ebislMGLevel  = m_ebislVec[whichlev];
     dxMGLevel     = m_dxVec[whichlev];
  }
  else
  {
     int icoar = 1;
     for(int idep = 0; idep < a_depth; idep++)
     {
          icoar *= 2;
     }
     Box domainBoxMGLevel;
     bool coarsenable = true;

      //find out how much we coarsen
      //should never coarsen beyond next AMR level
      //only really works for lbase = 0
      //need to fix this
      //if((whichlev > 0) && (icoar >= m_refRatioVec[whichlev-1]))
      //  {
      //    coarsenable = false;
      //  }
     if(coarsenable)
     {
        dxMGLevel = m_dxVec[whichlev];
        dxMGLevel *= Real(icoar);
        const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
        coarsenable = getCoarserLayouts(dblMGLevel,
                                        ebislMGLevel,
                                        domainBoxMGLevel,
                                        m_gridVec[whichlev],
                                        m_domainVec[whichlev].domainBox(),
                                        icoar,ebisPtr );
     }
     if(!coarsenable)
     {
          //not coarsenable.
          //return null
        return NULL;
     }
     domainMGLevel = ProblemDomain(domainBoxMGLevel);
  }

  //creates coarse and finer info and bcs and all that
  EBAMRHelmholtzOp* op = createOperator( dblMGLevel, ebislMGLevel,
					 domainMGLevel, dxMGLevel, whichlev);
  return op;
}


//-----------------------------------------------------------------------
EBAMRHelmholtzOp* EBAMRHelmholtzOpFactory::AMRnewOp(const ProblemDomain& a_domainFine)
{
  //figure out which level we are at.
  int whichlev;
  bool found = false;

  for(int ilev = 0; ilev < m_numLevels && !found; ilev++)
    {
      if(a_domainFine.domainBox() == m_domainVec[ilev].domainBox())
        {
          found = true;
          whichlev = ilev ;
        }
    }
  if(!found)
    {
      MayDay::Error("No corresponding AMRLevel to starting point of AMRnewOp");
    }
  //creates coarse and finer info and bcs and all that
  ProblemDomain      domainMGLevel = m_domainVec[whichlev];
  DisjointBoxLayout  dblMGLevel    = m_gridVec[whichlev];
  EBISLayout         ebislMGLevel  = m_ebislVec[whichlev];
  RealVect           dxMGLevel     = m_dxVec[whichlev];

  EBAMRHelmholtzOp* op = createOperator(dblMGLevel, ebislMGLevel, domainMGLevel, dxMGLevel, whichlev);
  return op;
}


//-----------------------------------------------------------------------
EBAMRHelmholtzOp* EBAMRHelmholtzOpFactory::createOperator(
                                      const DisjointBoxLayout& a_dblMGLevel,
                                      const EBISLayout&        a_ebislMGLevel,
                                      const ProblemDomain&     a_domainMGLevel,
                                      const RealVect&          a_dxMGLevel,
                                      const int&               a_whichLevel)
{

  RefCountedPtr<HelmholtzEBBC>  ebbcMGLevel      = m_ebBCFactory->create( a_domainMGLevel, a_ebislMGLevel, a_dxMGLevel);
  RefCountedPtr<HelmholtzDomainBC> dombcMGLevel = m_domainBCFactory->create();

  //fine and coarse stuff undefined because this is a MG level
  DisjointBoxLayout  dblFine,     dblCoar;
  EBISLayout       ebislFine,   ebislCoar;
  ProblemDomain    domainFine, domainCoar;
  RealVect             dxFine,     dxCoar;
  //  RefCountedPtr<BaseEBBC>      ebbcCoar = NULL;
  //  RefCountedPtr<BaseEBBC>       ebbcFine  = NULL;
  //  RefCountedPtr<BaseDomainBC> dombcCoar = NULL;
  //  RefCountedPtr<BaseDomainBC>  dombcFine  = NULL;
  RefCountedPtr<HelmholtzEBBC>      ebbcCoar = NULL;
  RefCountedPtr<HelmholtzEBBC>       ebbcFine  = NULL;
  RefCountedPtr<HelmholtzDomainBC> dombcCoar = NULL;
  RefCountedPtr<HelmholtzDomainBC>  dombcFine  = NULL;
  int refToFine   = 1;
  int refToCoar = 1;
  bool hasFine   = ((a_whichLevel < m_numLevels - 1)  && (a_domainMGLevel.domainBox() == m_domainVec[a_whichLevel].domainBox()));
  bool hasCoar   = ((a_whichLevel > 0) && (a_domainMGLevel.domainBox() == m_domainVec[a_whichLevel].domainBox()));
  if(hasFine)
    {
      refToFine = m_refRatioVec[a_whichLevel];
      dblFine   =     m_gridVec[a_whichLevel+1];
      ebislFine =    m_ebislVec[a_whichLevel+1];
      domainFine =  m_domainVec[a_whichLevel+1];
      dxFine    =       m_dxVec[a_whichLevel+1];
      ebbcFine  =     m_ebBCFactory->create(domainFine,ebislFine, dxFine);
      dombcFine = m_domainBCFactory->create();
    }
  if(hasCoar)
    {
      refToCoar  =  m_refRatioVec[a_whichLevel-1];
      dblCoar   =       m_gridVec[a_whichLevel-1];
      ebislCoar =      m_ebislVec[a_whichLevel-1];
      domainCoar =    m_domainVec[a_whichLevel-1];
      dxCoar     =        m_dxVec[a_whichLevel-1];
      ebbcCoar   =     m_ebBCFactory->create(domainCoar,ebislCoar, dxCoar);
      dombcCoar  = m_domainBCFactory->create();
    }

  LevelData<EBCellFAB>* rho;
  bool alloced=false;
  if( a_domainMGLevel.domainBox() != m_domainVec[a_whichLevel].domainBox() )
  {
// interpolate rho from a_whichLevel
//     IntVect rat;
//     for( int d= 0 ; d<SpaceDim ; d++ )
//       rat[d] = a_domainMGLevel.domainBox().size()/
//	                m_domainVec[a_whichLevel].domainBox().size();

     int rat=1;
     Box bx = a_domainMGLevel.domainBox();
     while( bx != m_domainVec[a_whichLevel].domainBox() )
     {
        bx.refine(IntVect::Unit*2);
        rat *= 2;
     }
     rho = new LevelData<EBCellFAB>(a_dblMGLevel, 1, IntVect::Zero,
				    EBCellFactory(a_ebislMGLevel));
     alloced =true;
     interp_frfine( *rho, *(m_rhoPtr[a_whichLevel]), a_domainMGLevel.domainBox(),
		    a_ebislMGLevel, m_ebislVec[a_whichLevel], rat );
  }
  else
  {
     rho = m_rhoPtr[a_whichLevel];
  }
    
  //  cout << "create hh op, level  " << a_whichLevel << " " << m_k2[a_whichLevel] << endl;

  EBAMRHelmholtzOp* op = new EBAMRHelmholtzOp(dblFine ,      a_dblMGLevel,   dblCoar,
                                          ebislFine,   a_ebislMGLevel, ebislCoar,
                                          dombcFine,     dombcMGLevel, dombcCoar,
                                          ebbcFine ,      ebbcMGLevel,  ebbcCoar,
                                          dxFine ,        a_dxMGLevel,    dxCoar,
					  a_domainMGLevel,  m_origin, m_k2[a_whichLevel],
                                          refToFine, refToCoar,
                                          hasFine, hasCoar,
                                          m_relaxType, m_orderEB,
					  m_kappa, rho, m_rz, m_time );

  if( alloced )
    delete rho;
  return op;
}

//-----------------------------------------------------------------------
void EBAMRHelmholtzOpFactory::interp_frfine( LevelData<EBCellFAB>& a_uc,
					     const LevelData<EBCellFAB>& a_uf,
                                             const Box& a_domainc,
                                             const EBISLayout& a_ec,
                                             const EBISLayout& a_ef,
					     const int a_rat )
{
   int ncomp = a_uc.nComp();
   CH_assert( ncomp == a_uf.nComp() );
   Interval interval(0, ncomp-1);
   const EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
   if( a_uf.getBoxes().coarsenable(a_rat) )
   {
      EBCoarseAverage ebaverage(a_uf.getBoxes(), a_uc.getBoxes(),
                                a_ef, a_ec, a_domainc,
                                a_rat, ncomp, ebisPtr);
      ebaverage.average(a_uc, a_uf, interval);
   }
   else
   {
      Box domainFine = refine( a_domainc, a_rat );
      Vector<Box> fineBoxes(1,  domainFine);
      Vector<Box> coarBoxes(1,a_domainc);
      Vector<int> procs(1, 0);
      DisjointBoxLayout oneBoxDBLFine(fineBoxes, procs);
      DisjointBoxLayout oneBoxDBLCoar(coarBoxes, procs);

      CH_assert((oneBoxDBLFine.coarsenable(a_rat)));
       
      //make the one box data holders
      LevelData<EBCellFAB> u_box_coar,u_box_fine;
      const IntVect& ghostVecCoar = a_uc.ghostVect();
      const IntVect& ghostVecFine = a_uf.ghostVect();
      CH_assert((ghostVecCoar == ghostVecFine));
      int nghost = 1;
      EBISLayout oneBoxEBISLCoar, oneBoxEBISLFine;
      ebisPtr->fillEBISLayout( oneBoxEBISLCoar, oneBoxDBLCoar, a_domainc,  nghost );
      ebisPtr->fillEBISLayout( oneBoxEBISLFine, oneBoxDBLFine, domainFine, nghost );
     u_box_coar.define(oneBoxDBLCoar,ncomp,ghostVecCoar, EBCellFactory(oneBoxEBISLCoar));
     u_box_fine.define(oneBoxDBLFine,ncomp,ghostVecFine, EBCellFactory(oneBoxEBISLFine));

      //copy fine data over to the single box fine data
      a_uf.copyTo(u_box_fine);
 
      //now do the coarsening
      EBCoarseAverage ebaverage(oneBoxDBLFine,
                                oneBoxDBLCoar,
                                oneBoxEBISLFine,
                                oneBoxEBISLCoar,
                                a_domainc,
                                a_rat,
                                ncomp,
                                ebisPtr);
      ebaverage.average(u_box_coar, u_box_fine, interval );
      u_box_coar.copyTo(a_uc);
   }
}

//-----------------------------------------------------------------------
void EBAMRHelmholtzOpFactory::reclaim( MGLevelOp<LevelData<EBCellFAB> >* a_reclaim )
{
  delete a_reclaim;
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOpFactory::AMRreclaim( EBAMRHelmholtzOp* a_reclaim )
{
  delete a_reclaim;
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOpFactory::set_k2( Real a_k2, int a_level )
{
  //  cout <<"set k2 on level " << a_level << endl;
  m_k2[a_level] = a_k2;
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOpFactory::set_rho( LevelData<EBCellFAB>& a_rho, int a_level )
{
   EBLevelDataOps::assign(*m_rhoPtr[a_level],a_rho);
}

//-----------------------------------------------------------------------
void EBAMRHelmholtzOpFactory::set_time( Real a_time )
{
   m_time = a_time;
}


//-----------------------------------------------------------------------
void EBAMRHelmholtzOpFactory::dump_level( int a_level )
{
   cout << "Level  " << a_level << endl;
   cout << "k2  =  " << m_k2[a_level] << endl;
   cout << "kappa= " << m_kappa << endl;
   ArrayToRaw atr;
   atr.output_array( *m_rhoPtr[a_level], m_ebislVec[a_level], "rho.bin" );
}
