/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _EBAMRNS_H_
#define _EBAMRNS_H_

#include "AMRLevel.H"
#include "EBCellFAB.H"
#include "BaseIVFAB.H"
#include "LevelData.H"
#include "EBLevelRedist.H"
#include "EBCoarToCoarRedist.H"
#include "EBCoarToFineRedist.H"
#include "EBFineToCoarRedist.H"
#include "EBCoarseAverage.H"
#include "EBPWLFineInterp.H"
#include "EBFluxRegister.H"
#include "EBLevelGodunovNS.H"
#include "EBPatchGodunov.H"
#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "DisjointBoxLayout.H"

#include "EBAMRNSviscOp.H"
#include "EBAMRHelmholtzOp.H"
#include "Forcing.H"

///
/**
 */
class EBAMRNS : public AMRLevel
{
public:

  EBAMRNS();

  virtual ~EBAMRNS();

  void sumConserved(Real& a_sumcons,
                    const int& a_ivar) const;

  virtual void define(AMRLevel*  a_coarser_level_ptr,
              const Box& a_problem_domain,
              int        a_level,
              int        a_ref_ratio);

  virtual void define(AMRLevel*  a_coarser_level_ptr,
              const ProblemDomain& a_problem_domain,
              int        a_level,
              int        a_ref_ratio);

  /// advance by one timestep
  virtual Real advance();

  /// things to do after a timeste
  virtual void postTimeStep();

  /// setup fine->coarse stuff
  /**
     stuff that needs to be setup from the finer
     level.  A bunch of objects depend upon the layouts
     from both levels and the finer level changes more
     often from regrid so this needs to be called from the finer
     level
  */
  virtual void syncWithFineLevel();

  /// create tags
  virtual void tagCells(IntVectSet& a_tags) ;

  /// create tags at initialization
  virtual void tagCellsInit(IntVectSet& a_tags) ;

  /// regrid
  virtual void regrid(const Vector<Box>& a_new_grids);

  /// initialize grids
  virtual void initialGrid(const Vector<Box>& a_new_grids);

  /// initialize data
  virtual void initialData();

  /// things to do after initialization
  virtual void postInitialize();

  /// object for physics integration
  virtual void patchGodunov(const EBPatchGodunovFactory* const a_ebpatchGodunovFactory);

#ifdef CH_USE_HDF5
  /// write checkpoint header
  virtual void writeCheckpointHeader(HDF5Handle& a_handle) const;

  /// write checkpoint data for this level
  virtual void writeCheckpointLevel(HDF5Handle& a_handle) const;

  /// read checkpoint header
  virtual void readCheckpointHeader(HDF5Handle& a_handle);

  /// read checkpoint data for this level
  virtual void readCheckpointLevel(HDF5Handle& a_handle);

  /// write plotfile header
  virtual void writePlotHeader(HDF5Handle& a_handle) const;

  /// write plot file data for this level
  virtual void writePlotLevel(HDF5Handle& a_handle) const;

#endif

  ///
  virtual Real getDx() const;

  ///
  virtual Real getDt() const;

  /// compute dt
  virtual Real computeDt();

  /// compute dt with initial data
  virtual Real computeInitialDt();

  /// CFL
  virtual void CFL(Real a_cfl);

  ///
  virtual void doSmushing(bool a_doSmushing);

  ///
  virtual   void doRZCoords(bool a_doRZCoords);

  ///
  virtual   void hasSourceTerm(bool a_hasSourceTerm);

  ///redistribution radius
  virtual   void redistRadius(int a_redistRad);

  /// longest length of the domain in real space
  virtual   void domainLength(Real a_domainLength);

  /// refinement threshold
  virtual   void refinementThreshold(Real a_refine_thresh);

  /// tag buffer size
  virtual   void tagBufferSize(int a_tag_buffer_size);

  virtual   void useMassRedistribution(bool a_useMassRedist);

  virtual void levelSetup();

  const EBISLayout& getEBISLayout() const;
  virtual void fillConsAndPrim(LevelData<EBCellFAB>& a_data) const;

  void tagAll(bool a_tagAll);


  // state vector at old time
  LevelData<EBCellFAB>& getStateOld();

  // state vector at new time
  LevelData<EBCellFAB>& getStateNew();

  void set_gamma( Real a_gamma );

  void set_rom( Real a_rom );

  void set_viscosities( Real a_mu, Real a_lambda, Real a_kappa );

  void use_visc( bool a_use_visc );  

  void set_forcing( int a_force );

protected:

  void redefine_op_factories( Vector<EBAMRNS*> a_levels );

  void postRegrid( int a_base_level );

  void define_impops_all_levels();

  void redefine_implicit_op( int a_level );

  void compute_viscous_terms( LevelData<EBCellFAB>& visc_term,
			      Real a_gamma, Real a_rom );

  void divide_volume( LevelData<EBCellFAB>& u, const EBISLayout& ebis );
  
  void get_vel_and_T( LevelData<EBCellFAB>& a_vel, LevelData<EBCellFAB>& a_T,
		      int a_job, Real a_time, Real a_gamma, Real a_rom );

  void implicit_solve_for_visc( Real a_gamma, Real a_rom, bool a_all_levels );

  void define_and_set_visc( LevelData<EBCellFAB>*& rhs, LevelData<EBCellFAB>*& u,
                            bool save_rhs, bool use_saved_rhs );

  void insert_u( const LevelData<EBCellFAB>& u );

  void define_heat( LevelData<EBCellFAB>& u,
		    LevelData<EBCellFAB>*& uc,
		    LevelData<EBCellFAB>*& uf,
		    LevelData<EBCellFAB>*& T,
		    LevelData<EBCellFAB>*& rhsT,
		    Real a_gamma, Real a_rom,
                    bool save_rhs, bool use_saved_rhs );

  void fill_ghosts( EBCellFAB& u, const DataIndex& dex,
		    const Box ibox, const Box ubox );

  void insert_T( const LevelData<EBCellFAB>& T, Real a_gamma, Real a_rom );

  void transform_to_primitive( LevelData<EBCellFAB>& visc_term,
			       LevelData<EBCellFAB>& vel );

  void transform_to_conservative( LevelData<EBCellFAB>& visc_term,
				  const LevelData<EBCellFAB>& state );
  
  void errors_with_forcing();

  Real norm( const LevelData<EBCellFAB>& a_rhs, int a_ord );
			 
  
  bool m_tagAll;
  bool m_useMassRedist;
  bool m_impops_defined;
  
  Box m_domainBox;
  Real calculateMass() const;

  EBAMRNS* getCoarserLevel() const;
  EBAMRNS* getFinerLevel() const;

  DisjointBoxLayout m_grids;
  EBISLayout m_ebisl;

  // state vector at old time
  LevelData<EBCellFAB> m_stateOld;

  // state vector at new time
  LevelData<EBCellFAB> m_stateNew;

  // Sometimes, need to save right hand side in implicit solve:
  LevelData<EBCellFAB> m_rhs;
  LevelData<EBCellFAB> m_rhsT;

  // number of components of m_state
  int m_nComp;

  int m_nGhost ;
  // names of components
  Vector<string> m_stateNames;
  Vector<string> m_primNames;

  // grid spacing
  Real m_dx;

  // grid spacing
  Real m_domainLength;

  // cfl number
  Real m_cfl;

  // new time step
  Real m_dtNew;

  // averaging from coarse to fine level
  EBCoarseAverage m_ebCoarseAverage;

  // interpolation from fine to coarse level
  EBPWLFineInterp m_ebFineInterp;

  // level solver
  EBLevelGodunovNS m_ebLevelGodunov;

  // flux register
  EBFluxRegister m_ebFluxRegister;

  // patch integrator
  EBPatchGodunov* m_ebPatchGodunov;

  // patch integrator
  const EBPatchGodunovFactory* m_ebPatchGodunovFactory;

  // refinement threshold for gradient
  Real m_refineThresh;

  //redistribution mass
  LevelData<BaseIVFAB<Real> > m_massDiff;
  LayoutData<IntVectSet> m_sets;

  RedistStencil m_redStencil;
  EBCoarToFineRedist m_ebCoarToFineRedist;
  EBCoarToCoarRedist m_ebCoarToCoarRedist;
  EBFineToCoarRedist m_ebFineToCoarRedist;

  // tag buffer size
  int m_tagBufferSize;
  int m_redistRad;

  bool m_doRZCoords;
  bool m_hasSourceTerm;
  bool m_doSmushing;
  bool m_hasCoarser;
  bool m_hasFiner;
  bool m_isDefined;
  

// Viscous terms
  bool m_dovisc;
  EBAMRNSviscOp* m_ns_viscosity;
  EBAMRHelmholtzOp* m_heat_conduction;

// Forcing to test code
  bool m_doforce;
  Forcing* m_forcing;

//  ..and their factory variables (these are over all levels).
  EBAMRNSviscOpFactory *m_viscop_factory;
  EBAMRHelmholtzOpFactory *m_heatop_factory;
  
// viscosities, initially
  Real m_mu, m_lambda, m_kappa;

// Gas dynamic constants
  Real m_gamma, m_rom;
  
private:
  //disallowed for all the usual reasons
  void operator=(const EBAMRNS& a_input)
  {
    MayDay::Error("invalid operator");
  }
  EBAMRNS(const EBAMRNS& a_input)
  {
    MayDay::Error("invalid operator");
  }

};

#endif
