/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

//using std::cout;
//using std::cin;
//using std::cerr;
#include <cmath>
//#include <iomanip.h>

#include "parstream.H"
#include "ParmParse.H"
#include "PolyGeom.H"

#include "Box.H"
#include "Vector.H"
#include "IntVectSet.H"
#include "EBCellFAB.H"
#include "DisjointBoxLayout.H"
#include "LayoutIterator.H"
#include "CH_HDF5.H"
#include "SPMD.H"
#include "LoadBalance.H"
#include "ProblemDomain.H"
#include "BoxIterator.H"

#include "EBAMRIO.H"
#include "AMRLevel.H"
#include "EBAMRNS.H"
#include "EBCellFactory.H"
#include "BaseIVFactory.H"
#include "VoFIterator.H"
#include "EBPatchGodunovF_F.H"
#include "EBIndexSpace.H"
#include "EBArith.H"

#include "DirichletVectFunctionF.H"
#include "NeumannFunctionF.H"

#include "NeumannHelmholtzDomainBC.H"
#include "NeumannHelmholtzEBBC.H"
#include "DirichletHelmholtzDomainBC.H"
#include "DirichletHelmholtzEBBC.H"


//#include "NeumannPoissonDomainRZBC.H"
//#include "NeumannPoissonEBRZBC.H"
//#include "NeumannPoissonDomainBC.H"
//#include "NeumannPoissonEBBC.H"

#include "BiCGStabSolver.H"

#include "ArrayToRaw.H"


//-----------------------------------------------------------------------
void EBAMRNS::tagAll(bool a_tagAll)
{
  m_tagAll = a_tagAll;
}


//-----------------------------------------------------------------------
void EBAMRNS::doSmushing(bool a_doSmushing)
{
  m_doSmushing = a_doSmushing;
}


//-----------------------------------------------------------------------
void EBAMRNS::doRZCoords(bool a_doRZCoords)
{
  m_doRZCoords = a_doRZCoords;
}


//-----------------------------------------------------------------------
void EBAMRNS::hasSourceTerm(bool a_hasSourceTerm)
{
  m_hasSourceTerm = a_hasSourceTerm;
}


//-----------------------------------------------------------------------
void EBAMRNS::useMassRedistribution(bool a_useMassRedist)
{
  m_useMassRedist = a_useMassRedist;
}


//-----------------------------------------------------------------------
void EBAMRNS::use_visc(bool a_use_visc)
{
  m_dovisc = a_use_visc;
}


//-----------------------------------------------------------------------
void EBAMRNS::set_forcing( int a_force )
{
   if( a_force == 1 )
   {
      m_doforce = true;
      if( m_forcing != NULL )
	 delete m_forcing;
      m_forcing = new Forcing( 2.0, 3.0, m_mu, m_lambda, m_kappa, m_rom/(m_gamma-1) );
   }
   else
   {
      m_doforce = false;
      m_forcing = NULL;
   }
}


//-----------------------------------------------------------------------
EBAMRNS::EBAMRNS()
{
  m_gamma = 1.4;
  m_rom   = 8.3143/0.030;
  m_mu     = 0.1;
  m_lambda = 0.05;
  m_kappa  = 0.1;

  m_cfl = 0.8;
  m_tagAll = false;
  m_useMassRedist = true;
  m_doRZCoords = false;
  m_hasSourceTerm = false;
  m_doSmushing = true;
  m_domainLength = 1.0;
  m_refineThresh = 0.2;
  m_initial_dt_multiplier = 0.1;
  m_ebPatchGodunov = NULL;
  m_redistRad = 1;
  m_isDefined = false;
  m_viscop_factory = NULL;
  m_heatop_factory = NULL;
  m_ns_viscosity   = NULL;
  m_heat_conduction= NULL;
  m_impops_defined = false;
  m_dovisc  = true;
  m_doforce = false;
  m_forcing = NULL;
}


//-----------------------------------------------------------------------
void EBAMRNS::redistRadius(int a_redistRad)
{
  m_redistRad = a_redistRad;
}


//-----------------------------------------------------------------------
void EBAMRNS::set_viscosities( Real a_mu, Real a_lambda, Real a_kappa )
{
   m_mu = a_mu;
   m_lambda = a_lambda;
   m_kappa = a_kappa;
}


//-----------------------------------------------------------------------
EBAMRNS::~EBAMRNS()
{
  if(m_ebPatchGodunov != NULL)
    delete m_ebPatchGodunov;
}


//-----------------------------------------------------------------------
void EBAMRNS::define(AMRLevel*  a_coarser_level_ptr,
		     const Box& a_problem_domain,
		     int        a_level,
		     int        a_ref_ratio)
{
  MayDay::Error("EBAMRNS::define -\n\tShould never be called with a Box for a problem domain");
}


//-----------------------------------------------------------------------
void EBAMRNS::define(AMRLevel*  a_coarser_level_ptr,
		     const ProblemDomain& a_problem_domain,
		     int        a_level,
		     int        a_ref_ratio )
{
  if(s_verbosity >= 3)
  {
      pout() << "EBAMRNS::define, level=" << a_level << endl;
  }

  m_isDefined = true;
  AMRLevel::define(a_coarser_level_ptr,
                   a_problem_domain,
                   a_level,
                   a_ref_ratio);

  m_domainBox = m_problem_domain.domainBox();

  if (a_coarser_level_ptr != NULL)
    {
      EBAMRNS* amrg_ptr =
        dynamic_cast<EBAMRNS*>(a_coarser_level_ptr);
      if (amrg_ptr == NULL)
        {
          pout() << "EBAMRG::define:cannot cast  to derived class"
                 << endl;
          MayDay::Error();
        }

      m_cfl = amrg_ptr->m_cfl;
      m_domainLength = amrg_ptr->m_domainLength;
      m_refineThresh = amrg_ptr->m_refineThresh;
      m_tagBufferSize = amrg_ptr->m_tagBufferSize;
    }

  m_dx = m_domainLength/m_domainBox.size(0);

  m_nGhost = 0;

  if(m_ebPatchGodunov != NULL)
    delete m_ebPatchGodunov;

  m_ebPatchGodunov = m_ebPatchGodunovFactory->create();
  m_ebPatchGodunov->define(m_problem_domain, m_dx);

  m_nComp      = m_ebPatchGodunov->numConserved();
  m_stateNames = m_ebPatchGodunov->stateNames();
  m_primNames  = m_ebPatchGodunov->primNames();
}


//-----------------------------------------------------------------------
void EBAMRNS::set_gamma( Real a_gamma )
{
  m_gamma = a_gamma;
}


//-----------------------------------------------------------------------
void EBAMRNS::set_rom( Real a_rom )
{
  m_rom = a_rom;
}


//-----------------------------------------------------------------------
void EBAMRNS::redefine_op_factories( Vector<EBAMRNS*> a_levels )
{
   int n=a_levels.size();

   Vector<DisjointBoxLayout> dbls(n);
   Vector<EBISLayout> ebises(n);
   Vector<LevelData<EBCellFAB>*> rhos(n);
   Vector<int> ref_ratios(n);
   IntVect gh=IntVect::Unit;
   for( int i=0 ; i<n ; i++ )
   {
     //      static_cast<EBAMRNS*>(a_levels[i])->getEBISLayout().copyTo(ebises[i]) ;
      ebises[i] = static_cast<EBAMRNS*>(a_levels[i])->m_ebisl;
      dbls[i]   = static_cast<EBAMRNS*>(a_levels[i])->m_grids;
      EBCellFactory ebcellfactory(ebises[i]);
      rhos[i] = new LevelData<EBCellFAB>( dbls[i], 1, gh, ebcellfactory );
      ref_ratios[i] = a_levels[i]->refRatio();
      cout << "Define op factories: Box layout " << i << " : "; dbls[i].print();
   }
   ProblemDomain coarse_domain = static_cast<EBAMRNS*>(a_levels[0])->problemDomain();
   RealVect coarse_dx;
   RealVect x0;
   for( int d=0 ; d<SpaceDim ; d++ )
   {
     coarse_dx[d] = static_cast<EBAMRNS*>(a_levels[0])->getDx();
     x0[d] = 0;
   }
   int relaxtype = 0, bc_order = 2;
   Real k2=1.0;
   Vector<Real> c( n, 1.0 );
   const Vector<Real> zero(SpaceDim,0);

// Viscosity
   DirichletNSviscDomainBCFactory* domain_bc= new DirichletNSviscDomainBCFactory(); 
   if( m_doforce )
   {
      RefCountedPtr<BaseBCVectValue> bc_function = new DirichletVectFunctionF(m_forcing);
      domain_bc->setFunction( bc_function );
   }
   else
      domain_bc->setValue( zero );

   RefCountedPtr<DirichletNSviscDomainBCFactory> bc_factory=domain_bc;
   DirichletNSviscEBBCFactory* eb_bc= new DirichletNSviscEBBCFactory(); 

   if( m_doforce )
   {
      RefCountedPtr<BaseBCVectValue> bc_function = new DirichletVectFunctionF(m_forcing);
      eb_bc->setFunction( bc_function );
   }
   else
      eb_bc->setValue( zero );

   RefCountedPtr<DirichletNSviscEBBCFactory> bc_ebfactory=eb_bc;

   if( m_viscop_factory != NULL )
      delete m_viscop_factory;
   m_viscop_factory = new EBAMRNSviscOpFactory( dbls, ebises, ref_ratios, coarse_domain,
 	  			        coarse_dx, x0, bc_order, relaxtype,
					bc_factory, bc_ebfactory, c, m_mu, m_lambda,
					rhos, m_doRZCoords );

// Heat equation
   relaxtype = 0;

   bool dtdn = true;
   HelmholtzDomainBCFactory* bc_hh;
   HelmholtzEBBCFactory* bc_hheb;

   if( dtdn )
   {
      bc_hh   = new NeumannHelmholtzDomainBCFactory();
      bc_hheb = new NeumannHelmholtzEBBCFactory();
      if( m_doforce )
      {
         RefCountedPtr<BaseBCVectValue> bc_function = new NeumannFunctionF(m_forcing);
	 bc_hh->set_function( bc_function );
	 bc_hheb->set_function( bc_function );
      }
      else
      {
         Vector<Real> z(1,0.0);
         bc_hh->set_value( z );
         bc_hheb->set_value( z );
      }
   }
   else
   {
      bc_hh   = new DirichletHelmholtzDomainBCFactory();       
      bc_hheb = new DirichletHelmholtzEBBCFactory();
      Vector<Real> Twall(1,300.0);
      bc_hh->set_value( Twall );
      bc_hheb->set_value( Twall );
   }

   bc_hh->set_rz( m_doRZCoords );
   bc_hheb->set_rz( m_doRZCoords );

   RefCountedPtr<HelmholtzDomainBCFactory> bc_hh_p   = bc_hh;
   RefCountedPtr<HelmholtzEBBCFactory>     bc_hheb_p = bc_hheb;

   if( m_heatop_factory != NULL )
      delete m_heatop_factory;

   m_heatop_factory = new EBAMRHelmholtzOpFactory( dbls, ebises, ref_ratios, coarse_domain,
 	  			    coarse_dx, x0, k2, bc_order, relaxtype,
				    bc_hh_p, bc_hheb_p, m_kappa, rhos, m_doRZCoords );
   for( int i=0 ; i<n ; i++ )
     delete rhos[i];
}


//-----------------------------------------------------------------------
void EBAMRNS::postRegrid( int a_base_level )
{
   cout << "EBAMRNS: Post regrid, level " << a_base_level << " m_l "
	<< m_level << endl;

   EBAMRNS* l = this;
   EBAMRNS* lcoar;
   while( l != NULL )
   {
     lcoar = l;
     l = l->getCoarserLevel();
   }
   lcoar->define_impops_all_levels( );
}


//-----------------------------------------------------------------------
void EBAMRNS::define_impops_all_levels()
{
  //  if( getCoarserLevel() == NULL )
  {
// count levels
    int nlevels = 1;
    EBAMRNS* l=getFinerLevel();
    while( l != NULL )
    {
       nlevels++;
       l = l->getFinerLevel();
    }
// collect levels in vector
//    cout << "Nlevels = " << nlevels << endl;
    Vector<EBAMRNS*> levels(nlevels);
    l = this;
    for( int n=0 ; n<nlevels ; n++ )
    {
      levels[n] = l;
      l = l->getFinerLevel();
    }

    redefine_op_factories( levels );
    redefine_implicit_op( 0 );

// propagate factories to all levels
    for( int n=1 ; n<nlevels ; n++ )
    {
      levels[n]->m_viscop_factory = levels[0]->m_viscop_factory;
      levels[n]->m_heatop_factory = levels[0]->m_heatop_factory;
      levels[n]->redefine_implicit_op(n);
    }
  }
}

//-----------------------------------------------------------------------
void EBAMRNS::redefine_implicit_op( int a_level )
{


   cout << "in redefine_implicit_op" << endl;
   if( m_ns_viscosity != NULL )
      delete m_ns_viscosity;

   if( m_heat_conduction != NULL )
      delete m_heat_conduction;
   
   RealVect dxv = (RealVect::Unit)*m_dx;
   //   dxv *= m_dx;

   m_ns_viscosity    = m_viscop_factory->createOperator( m_grids, m_ebisl,
							 m_problem_domain,
						         dxv, a_level );
   m_heat_conduction = m_heatop_factory->createOperator( m_grids, m_ebisl,
							 m_problem_domain,
						         dxv, a_level );
   m_ns_viscosity->set_rho_zero();
   m_heat_conduction->set_k2( 0.0 );
}

//-----------------------------------------------------------------------
void EBAMRNS::compute_viscous_terms( LevelData<EBCellFAB>& visc_term,
				     Real a_gamma, Real a_rom )
{
// Evaluate viscosity and heat conduction:

// Temporary areas
//   IntVect ghosts=m_stateNew.ghostVect();
   IntVect ghosts=IntVect::Unit;
   LevelData<EBCellFAB> vel, velcoar, velfine;
   LevelData<EBCellFAB> T, Tcoar, Tfine;

   EBCellFactory cell_factory(m_ebisl);
   vel.define( m_grids, SpaceDim, ghosts, cell_factory );
   T.define( m_grids, 1, ghosts, cell_factory );

   get_vel_and_T( vel, T, 1, m_time, a_gamma, a_rom );
   
   //    ArrayToRaw atr;
    //   atr.chk_nan( vel, m_ebisl, "vel.bin" );
   //   atr.chk_nan( T,   m_ebisl, "T.bin" );

    //   cout << "coarse = " << m_hasCoarser << endl;
    //   cout << "fine = " << m_hasFiner << endl;

   if( m_hasCoarser )
   {
      EBCellFactory coar_factory((*getCoarserLevel()).m_ebisl);
      velcoar.define( (*getCoarserLevel()).m_grids, SpaceDim, ghosts, coar_factory );
      Tcoar.define( (*getCoarserLevel()).m_grids, 1, ghosts, coar_factory );
      (*getCoarserLevel()).get_vel_and_T( velcoar, Tcoar, 2, m_time, a_gamma, a_rom );
   }
   if( m_hasFiner )
   {
      EBCellFactory fine_factory((*getFinerLevel()).m_ebisl);
      velfine.define( (*getFinerLevel()).m_grids, SpaceDim, ghosts, fine_factory );
      Tfine.define( (*getFinerLevel()).m_grids, 1, ghosts, fine_factory );
      (*getFinerLevel()).get_vel_and_T( velfine, Tfine, 1, m_time, a_gamma, a_rom );
   }

   IntVect res_ghosts(IntVect::Zero);

// Residuals
   LevelData<EBCellFAB> vres, veeq, vh;
   vres.define( m_grids, SpaceDim, res_ghosts, cell_factory );
   veeq.define( m_grids, 1, res_ghosts, cell_factory );
   vh.define(  m_grids, 1, res_ghosts, cell_factory );


// Compute residuals
   LevelData<EBCellFAB> zero;
   zero.define( m_grids, SpaceDim, res_ghosts, cell_factory );
   EBLevelDataOps::setToZero(zero);

   m_ns_viscosity->set_rho_zero( );
   m_ns_viscosity->set_time( m_time );

   m_ns_viscosity->AMRResidual( vres, velfine, vel, velcoar, zero, false );

   //   atr.chk_nan( vres, m_ebisl, "vres" );
   //   atr.output_array(vel,m_ebisl,"uubb.bin");

   EBLevelDataOps::scale(vres,-1.0);

   //   atr.output_array( vres,   m_ebisl, "vres.bin" );
   vres.copyTo( Interval(0,SpaceDim-1), visc_term, Interval(1,SpaceDim) );

   EBLevelDataOps::setToZero(veeq);
   //   cout << "calling amrresidual" << endl;

   LevelData<EBCellFAB>* vf=NULL;
   LevelData<EBCellFAB>* vc=NULL;
   if( m_hasCoarser )
     vc = &velcoar;
   if( m_hasFiner )
     vf = &velfine;
   
   m_ns_viscosity->AMRResidual_eeq( veeq, vf, vel, vc, false );
   //   atr.output_array( veeq,   m_ebisl, "veeq.bin" );

   //   atr.chk_nan( vel, m_ebisl, "vel" );
   //   atr.chk_nan( velfine, (*getFinerLevel()).m_ebisl, "velfine" );

   LevelData<EBCellFAB> zeros;
   zeros.define( m_grids, 1, res_ghosts, cell_factory );
   EBLevelDataOps::setToZero(zeros);

   m_heat_conduction->set_k2( 0.0 );
   m_heat_conduction->set_time( m_time );
   m_heat_conduction->AMRResidual( vh, Tfine, T, Tcoar, zeros, false );
   EBLevelDataOps::scale(vh,-1.0);
   //   atr.output_array( vh,   m_ebisl, "vh.bin" );


// Energy equation viscosity:
   EBLevelDataOps::sum( veeq, veeq, vh );
   veeq.copyTo( Interval(0,0), visc_term, Interval(SpaceDim+1,SpaceDim+1) );

   //   atr.chk_nan( visc_term, m_ebisl, "vterm" );
   divide_volume( visc_term, m_ebisl );

   transform_to_primitive( visc_term, vel );
   //   atr.chk_nan( visc_term, m_ebisl, "vterm" );
  //   atr.output_array( visc_term,   m_ebisl, "visc.bin" );
   //   DataIterator d=visc_term.dataIterator();
   //   atr.print_ebcellfab(visc_term[d()],"visc");
}


//-----------------------------------------------------------------------
void EBAMRNS::divide_volume( LevelData<EBCellFAB>& u, const EBISLayout& ebis )
{
   int nc = u.nComp();
   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      for( BoxIterator b(u[dex].getRegion()); b.ok(); ++b )
	 if( !ebis[dex].isCovered(b()) )
	 {
	    Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
            for( int vi=0 ; vi < vofs.size() ; vi++ )
	    {
               if( vofs.size() > 1 )
		  cout << "voff voff" << endl;
               for( int c = 0 ; c < nc; c++ )
                  u[dex](vofs[vi],c) /= ebis[dex].volFrac(vofs[vi]);
	    }
	 }
   }
}


//-----------------------------------------------------------------------
void EBAMRNS::transform_to_primitive( LevelData<EBCellFAB>& visc_term,
				      LevelData<EBCellFAB>& vel )
{
  Real gm1 = m_gamma-1;
  for( DataIterator di=m_grids.dataIterator(); di.ok() ; ++di )
  {
     DataIndex dex = di();
     for( BoxIterator b(visc_term[dex].getRegion()); b.ok(); ++b )
     {
	Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
        for( int vi=0 ; vi < vofs.size() ; vi++ )
	{
	   Real rhoi = 1.0/m_stateNew[dex](vofs[vi],0);

           for( int c=0 ; c < SpaceDim ; c++ )
	   {
	      visc_term[dex](vofs[vi],SpaceDim+1) -=
		           vel[dex](vofs[vi],c)*visc_term[dex](vofs[vi],c);
	      visc_term[dex](vofs[vi],c) *= rhoi;

	   }
           visc_term[dex](vofs[vi],SpaceDim+1) *= gm1;
	}
     }
  }
}

//-----------------------------------------------------------------------
void EBAMRNS::transform_to_conservative( LevelData<EBCellFAB>& visc_term,
				        const LevelData<EBCellFAB>& state )
{
  Real gm1 = m_gamma-1;
  for( DataIterator di=m_grids.dataIterator(); di.ok() ; ++di )
  {
     DataIndex dex = di();
     for( BoxIterator b(visc_term[dex].getRegion()); b.ok(); ++b )
     {
	Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
        for( int vi=0 ; vi < vofs.size() ; vi++ )
	{
	   Real rho = state[dex](vofs[vi],0);
           visc_term[dex](vofs[vi],SpaceDim+1) /= gm1;

           for( int c=0 ; c < SpaceDim ; c++ )
	   {
 	      Real vel = state[dex](vofs[vi],c+1)/rho;
	      visc_term[dex](vofs[vi],SpaceDim+1) +=
		       rho*vel*visc_term[dex](vofs[vi],c);
	      visc_term[dex](vofs[vi],c) *= rho;

	   }

	}
     }
  }
}


//-----------------------------------------------------------------------
void EBAMRNS::get_vel_and_T( LevelData<EBCellFAB>& a_vel,
			     LevelData<EBCellFAB>& a_T,
			     int a_job,
			     Real a_time,
			     Real a_gamma,
			     Real a_rom )
{
  //  job==0 ==> compute from old state,
  //  job == 1 from new state, 
  //  job == 2 from average state.

  CH_assert( a_job >= 0 && a_job <= 2 );
  Real wgh;

  if( a_job == 2 )
  {
     wgh = (m_time-a_time)/m_dt;
  }

  for( DataIterator di=m_grids.dataIterator(); di.ok() ; ++di )
  {
     DataIndex dex = di();
     Box state_box = m_stateOld[dex].getRegion();
     for( BoxIterator b(state_box); b.ok(); ++b )
     {
	Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
        for( int vi=0 ; vi < vofs.size() ; vi++ )
	{
	   if( a_job == 0 )
	   {
   	      Real rho = m_stateOld[dex](vofs[vi],0);
              Real kinenergy = 0;
              for( int c= 0 ; c < SpaceDim ; c++ )
	      {
 	         a_vel[dex](vofs[vi],c) = m_stateOld[dex](vofs[vi],c+1)/rho;
                 kinenergy += a_vel[dex](vofs[vi],c)*a_vel[dex](vofs[vi],c);
	      }
              Real p = (a_gamma-1)*(m_stateOld[dex](vofs[vi],SpaceDim+1)-0.5*rho*
				  kinenergy);
              a_T[dex](vofs[vi],0) = p/(rho*a_rom);
	   }
           else if( a_job == 1 )
	   {
   	      Real rho = m_stateNew[dex](vofs[vi],0);
              Real kinenergy = 0;
              for( int c= 0 ; c < SpaceDim ; c++ )
	      {
 	         a_vel[dex](vofs[vi],c) = m_stateNew[dex](vofs[vi],c+1)/rho;
                 kinenergy += a_vel[dex](vofs[vi],c)*a_vel[dex](vofs[vi],c);
	      }
              Real p = (a_gamma-1)*(m_stateNew[dex](vofs[vi],SpaceDim+1)-0.5*rho*
				  kinenergy);
              a_T[dex](vofs[vi],0) = p/(rho*a_rom);
	   }
	   else 
	   {
	      // Average cons states first, then compute v and T. 
   	      Real rho  = (1-wgh)*m_stateNew[dex](vofs[vi],0)+
		            wgh*m_stateOld[dex](vofs[vi],0);
              Real kinenergy = 0;
              for( int c= 0 ; c < SpaceDim ; c++ )
	      {
 	         a_vel[dex](vofs[vi],c) = ((1-wgh)*m_stateNew[dex](vofs[vi],c+1)+
		   wgh*m_stateOld[dex](vofs[vi],c+1))/rho;
                 kinenergy += a_vel[dex](vofs[vi],c)*a_vel[dex](vofs[vi],c);
	      }

              Real p = (a_gamma-1)*((1-wgh)*m_stateNew[dex](vofs[vi],SpaceDim+1)+
				  wgh*m_stateOld[dex](vofs[vi],SpaceDim+1)-
				  0.5*rho*kinenergy);
	      a_T[dex](vofs[vi],0) = p/(rho*a_rom);
	   }
	}
     }
     fill_ghosts( a_T[dex],   dex, state_box, a_T[dex].getRegion() );
     fill_ghosts( a_vel[dex], dex, state_box, a_vel[dex].getRegion() );
  }
}


//-----------------------------------------------------------------------
Real EBAMRNS::advance()
{
  if( !m_impops_defined )
  {
     if( m_level == 0 )
        define_impops_all_levels();
     m_impops_defined = true;
  }
  if(s_verbosity >= 3)
    {
      pout() << " in EBAMRNS advance for level " << m_level << endl;
    }
  cout << " start time step of size " << m_dt << " at time " << m_time << endl;

  m_stateNew.copyTo(m_stateNew.interval(), m_stateOld, m_stateOld.interval());

  Real new_dt = 0.0;
  //set up arguments to step
  //undefined lfr in case we need it

  EBFluxRegister lfr;
  //undefined leveldata in case we need it

  const LevelData<EBCellFAB> ld;

  //set arguments to dummy arguments and
  //then fix if they are available
  EBFluxRegister* coarFR = &lfr;
  EBFluxRegister* fineFR = &lfr;
  const LevelData<EBCellFAB>* coarDataOld = &ld;
  const LevelData<EBCellFAB>* coarDataNew = &ld;

  Real told = 0.0;
  Real tnew = 0.0;
  if(m_hasCoarser)
    {
      EBAMRNS* coarPtr = getCoarserLevel();
      //recall that my flux register goes between my
      //level and the next finer level
      coarFR = &coarPtr->m_ebFluxRegister;
      coarDataOld = &coarPtr->m_stateOld;
      coarDataNew = &coarPtr->m_stateNew;
      tnew = coarPtr->m_time;
      told = tnew - coarPtr->m_dt;
      //time should never be greater than the newest coarse
      //time.  time might be very slightly smaller than
      //told because of the above subtraction.
      Real eps = 1.0e-10;
      if((m_time > tnew) || (m_time < (told - eps)))
        {
          MayDay::Error("out of bounds time input to AMRNS");
        }
      //correct for said floating-point nastiness
      m_time = Max(m_time, told);
    }
  if(m_hasFiner)
    {
      //recall that my flux register goes between my
      //level and the next finer level
      fineFR = &m_ebFluxRegister;
    }

#ifndef NDEBUG
  if(!m_hasCoarser && (s_verbosity > 1))
    {
      Real summass;
      int densityIndex = m_ebPatchGodunov->densityIndex();
      sumConserved(summass,  densityIndex);

      pout() << "sum mass = " << summass << endl;
    }
#endif

// Evaluate viscos operator:
  LevelData<EBCellFAB> viscos;
  viscos.define( m_grids, SpaceDim+2, IntVect::Zero, EBCellFactory(m_ebisl) );

  EBLevelDataOps::setToZero( viscos );
  
  ArrayToRaw atr;

  if( m_dovisc )
     compute_viscous_terms( viscos, m_gamma, m_rom );

  //  atr.output_array( viscos, m_ebisl, "viscos.bin" );
  //  atr.output_array( m_stateOld, m_ebisl, "oldstate.bin" );
 
  //  cout << "Level is " << m_level << endl;
  //  atr.chk_nan(m_stateOld,m_ebisl,"oldstate b is nan");
  //  atr.chk_nan(m_stateNew,m_ebisl,"newstate b is nan");
  //  atr.chk_nan(viscos,m_ebisl,"viscos b is nan");
  if( !m_doforce )
  {
     new_dt = m_ebLevelGodunov.step(m_stateNew,
                                    m_massDiff,
                                    *fineFR,
                                    *coarFR,
                                    *coarDataOld,
                                    *coarDataNew,
                                    m_time,
                                    told,
                                    tnew,
                                    m_dt,
				    viscos );
  
  }
  else 
  {
     LevelData<EBCellFAB> fo;
     fo.define( m_grids, SpaceDim+2, IntVect(0,0), EBCellFactory(m_ebisl) );
     RealVect problo = Chombo_EBIS::instance()->getOrigin();
     RealVect dxv = RealVect::Unit*m_dx;
     m_forcing->get_forcing( fo, m_ebisl, m_time, dxv, problo, false );
     EBLevelDataOps::incr( m_stateNew, fo, m_dt );
     new_dt = m_dt;
  }

  
  //  atr.chk_nan(m_stateOld,m_ebisl,"oldstate a is nan");
  atr.chk_nan( m_stateNew, m_ebisl, "newstate a is nan" );

  // flux register manipulation happens in levelgodunov
  // level redistribution happens in levelgodunov

  if( m_dovisc )
  {
    // Add explicit part of viscosity
     transform_to_conservative( viscos, m_stateOld );
     EBLevelDataOps::incr( m_stateNew, viscos, m_dt );
  }

  //  atr.chk_nan(m_stateNew,m_ebisl,"newstate av is nan");
  // Solve implicit NS-operator

  if( m_dovisc )
  {
    //     implicit_solve_for_visc( m_gamma, m_rom, false );
  }
  //  atr.output_array( m_stateNew, m_ebisl, "newstate.bin" );

  //  atr.chk_nan(m_stateNew,m_ebisl,"newstate avi is nan");  

  Interval interv(0, m_nComp-1);
  // increment redistribution register between this level
  // and next coarser level
  if(m_hasCoarser)
  {
      for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
      {
          m_ebFineToCoarRedist.increment(m_massDiff[dit()], dit(), interv);
      }
  }
  //  atr.chk_nan(m_stateNew,m_ebisl,"newstate coarse is nan");    
  //initialize redistirbution register between this level
  // and next finer level
  //this includes re-redistribution registers
  if(m_hasFiner)
  {

      m_ebCoarToFineRedist.setToZero();
      m_ebCoarToCoarRedist.setToZero();
      for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
      {
          BaseIVFAB<Real>& massDiffFAB = m_massDiff[dit()];
          m_ebCoarToFineRedist.increment(massDiffFAB, dit(), interv);
          m_ebCoarToCoarRedist.increment(massDiffFAB, dit(), interv);
      }
  }
  //  atr.chk_nan(m_stateNew,m_ebisl,"newstate fine is nan");    

  //deal with time and time step
  m_time += m_dt;
  Real return_dt = m_cfl * new_dt;

  //save stable timestep to save computational effort
  m_dtNew = return_dt;

  errors_with_forcing();
  
  return return_dt;
}


//-----------------------------------------------------------------------
void EBAMRNS::implicit_solve_for_visc( Real a_gamma, Real a_rom, bool a_all_levels )
{
   int levels_down  = 0;
   int l_max = 0;
   int l_min = 0;

   EBAMRNS* pcoar = this;
   EBAMRNS* p;

   if( a_all_levels )
   {
      p = getCoarserLevel();
// count the number of levels to solve over
      while( (p != NULL) && (fabs(p->m_time-m_time) < 1e-6*m_dt) )
      {
         pcoar = p;
         p = p->getCoarserLevel();
         levels_down++;
      }
      l_max = levels_down;
      l_min = 0;
// No need to to synchronization AMR solve if only one AMR level
      if( levels_down == 0 )
	return;
   }
   else
   {
     levels_down = 0;
     l_max = 0;
     l_min = 0;
     if( getCoarserLevel() != NULL )
     {
        pcoar = getCoarserLevel();
        levels_down = 1;
        l_max=1;
	l_min=0;
     }
   }
   int total_levels = levels_down+1;

   Vector<LevelData<EBCellFAB>*> rhs(total_levels), rhsT(total_levels);
   Vector<LevelData<EBCellFAB>*> u(total_levels), T(total_levels);

   //   ArrayToRaw atr;
// Set up levels and factory variable (done in define_and_set_visc).
   p = this;
   //   cout << "levels down " << levels_down << endl;

   for( int l=levels_down ; l>= 0 ; l-- )
   {
      p->define_and_set_visc( rhs[l], u[l], !a_all_levels, a_all_levels );
      p = p->getCoarserLevel();
   }

   //   atr.output_array( m_stateNew, m_ebisl, "ub.bin" );

// Set up multigrid solver
   AMRMultiGrid<LevelData<EBCellFAB> > mgsolver;
   BiCGStabSolver<LevelData<EBCellFAB> > bottomSolver;
   //   mgsolver.define( pcoar->problemDomain(), *m_viscop_factory, &bottomSolver,
   //		    levels_down+1 );
   //   cout << "mg with domain " << problemDomain().domainBox() << endl;
   mgsolver.define( pcoar->problemDomain(), *m_viscop_factory, &bottomSolver,
		    levels_down+1 );

// Do viscosity solve:   

   
//   if( m_level == 1 )
//   {
//     atr.output_array( (*u[0]), pcoar->m_ebisl, "ubef0.bin" );
//     atr.output_array( (*rhs[0]), pcoar->m_ebisl, "rhs0.bin" );
//     atr.output_array( (*u[1]), m_ebisl, "ubef1.bin" );
//     atr.output_array( (*rhs[1]), m_ebisl, "rhs1.bin" );
//   }
   cout << "Solving NS visc from " << l_min << " to " << l_max << endl;

   mgsolver.solve( u, rhs, l_max, l_min );

   //   if( m_level == 1 )
   //   {
   //      atr.output_array( (*u[0]), pcoar->m_ebisl, "uaft0.bin" );
   //      atr.output_array( (*u[1]), m_ebisl, "uaft1.bin" );
   //      exit(2);
      //   }
   //     atr.output_array( (*u[0]), m_ebisl, "u.bin" );

// Do heat equation solve:
   p = this;
   for( int l=levels_down ; l>= 0 ; l-- )
   {
      LevelData<EBCellFAB> *uc, *uf;
      if( l>0 )
 	 uc = u[l-1];
      else
	 uc = NULL;
      if( l<levels_down )
	 uf = u[l+1];
      else
	 uf = NULL;

      if( l_min <= l &&  l <= l_max ) 
          p->insert_u( *(u[l]) );
      p->define_heat( *(u[l]), uc, uf, T[l], rhsT[l], a_gamma, a_rom,
		      !a_all_levels, a_all_levels );
      p = p->getCoarserLevel();
   }

   for( int l=levels_down ; l>= 0 ; l-- )
   {
      delete u[l];
      delete rhs[l];
   }

   AMRMultiGrid<LevelData<EBCellFAB> > mgsolver2;
   BiCGStabSolver<LevelData<EBCellFAB> > bottomSolver2;
   mgsolver2.define(pcoar->problemDomain(), *m_heatop_factory, &bottomSolver2, levels_down+1 );


   //   ArrayToRaw atr;
   //   atr.output_array( (*T[0]), m_ebisl, "Tbef.bin" );
   //   atr.output_array( (*rhsT[0]), m_ebisl, "rhsT.bin" );
   //   m_heatop_factory->dump_level(0);
   //   exit(2);
   cout << "Solving heat eq. from " << l_min << " to " << l_max << endl;

   mgsolver2.solve(  T, rhsT, l_max, l_min );

   //     atr.output_array( (*T[0]), m_ebisl, "Taft.bin" );
   //     exit(2);
// Compute energy from the new temperature.
   p = this;
   for( int l=l_max ; l>= l_min ; l-- )
   {
      p->insert_T( *T[l], a_gamma, a_rom );
      p = p->getCoarserLevel();
   }
   //   atr.output_array( m_stateNew, m_ebisl, "ua.bin" );
   //   exit(2);

   for( int l = levels_down ; l >=0 ; l-- )
   {
      delete T[l];
      delete rhsT[l];
   }
   //   cout << "Done heat eq." << endl;
}


//-----------------------------------------------------------------------
void EBAMRNS::define_and_set_visc( LevelData<EBCellFAB>*& rhs,
				   LevelData<EBCellFAB>*& u,
				   bool save_rhs, bool use_saved_rhs )
{
   EBCellFactory cfactory( m_ebisl );

   u   = new LevelData<EBCellFAB>( m_grids, SpaceDim, IntVect::Unit,cfactory );
   rhs = new LevelData<EBCellFAB>( m_grids, SpaceDim, IntVect::Zero,cfactory );

   LevelData<EBCellFAB> rho( m_grids, 1, IntVect::Unit,cfactory );
   EBLevelDataOps::setToZero(*u);

   //   cout << "newstate ghosts " << m_stateNew.ghostVect() << endl;
   for( DataIterator di=m_grids.dataIterator(); di.ok() ; ++di )
   {
      DataIndex dex = di();
      Box state_box = m_stateNew[dex].box();
      Box u_box = (*u)[dex].box();

      for( BoxIterator b(state_box); b.ok(); ++b )
      {
	 Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
         for( int vi=0 ; vi < vofs.size() ; vi++ )
	 {
	    rho[dex](vofs[vi],0) = m_stateNew[dex](vofs[vi],0);
// Initial guess for u, and the right hand side for visc eq.
            for( int c=0 ; c < SpaceDim ; c++ )
	    {
   	       (*u)[dex](vofs[vi],c)   =  m_stateNew[dex](vofs[vi],c+1)/rho[dex](vofs[vi],0);
               (*rhs)[dex](vofs[vi],c) = -m_stateNew[dex](vofs[vi],c+1)*m_ebisl[dex].volFrac(vofs[vi])/(m_dt/2); 
	    }
	 }
      }
      fill_ghosts( (*u)[dex], dex, state_box, u_box );
      fill_ghosts( rho[dex], dex, state_box, u_box );
   }

   if( save_rhs )
       EBLevelDataOps::assign(m_rhs,*rhs);

   if( use_saved_rhs )
       EBLevelDataOps::assign(*rhs,m_rhs);

   m_viscop_factory->set_rho( rho, m_level );
   m_viscop_factory->set_coeff( m_dt/2, m_level );

   //   ArrayToRaw atr;
   //   atr.output_array( (*u), m_ebisl, "u.bin" );
   //   atr.output_array( (*rhs), m_ebisl, "rhs.bin" );
   //   atr.output_array( rho, m_ebisl, "rho.bin" );
   //   atr.output_array( m_stateNew, m_ebisl, "new.bin" );
   //   exit(2);

}


//-----------------------------------------------------------------------
void EBAMRNS::fill_ghosts( EBCellFAB& u, const DataIndex& dex,
			   const Box ibox, const Box ubox )
{
   int ncomp = u.nComp();

   Vector<IntVect> of(SpaceDim*4);

   of[0][0]=1;
   of[0][1]=0;

   of[1][0]=-1;
   of[1][1]=0;

   of[2][0]=0;
   of[2][1]=1;

   of[3][0]=0;
   of[3][1]=-1;

// corners
   of[4][0]=1;
   of[4][1]=1;

   of[5][0]=-1;
   of[5][1]=1;

   of[6][0]=1;
   of[6][1]=-1;

   of[7][0]=-1;
   of[7][1]=-1;

   if( SpaceDim == 3)
   {
      of[4][0]=0;
      of[4][1]=0;
      of[4][2]=1;

      of[5][0]=0;
      of[5][1]=0;
      of[5][2]=-1;

      of[0][2]=0;
      of[1][2]=0;
      of[2][2]=0;
      of[3][2]=0;
         
   }
   for( BoxIterator b(ubox); b.ok(); ++b )
   {
      if( !ibox.contains( b() ) )
      {
	 IntVect iv=b();
         int s = 0;
	 int found = 0;
         while( !found && s < SpaceDim*4 )
	 {
	    if( ibox.contains(iv+of[s]) )
	    {
     	       Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(iv);
     	       Vector<VolIndex> vofi = m_ebisl[dex].getVoFs(iv+of[s]);
	       //               CH_assert( vofi.size() > 0 );
               if( vofi.size() > 0 )
	       {
                  for( int vi=0 ; vi < vofs.size() ; vi++ )
  	          {
                     for( int c=0 ; c < ncomp ; c++ )
		        u(vofs[vi],c) = u(vofi[0],c);
	          }
	          found = 1;
	       }
	    }
	    s++;
	 }
         if( !found && !m_ebisl[dex].isCovered(iv) )
	 {
	   //	    cout << "Error in EBAMRNS::fill_ghosts" << endl;
	   //            cout << "ibox is " << ibox << endl;
	   //            cout << "ubox is " << ubox << endl;
	   //	    cout << "index is " << iv << endl;
	 }
      }
   }
}


//-----------------------------------------------------------------------
void EBAMRNS::insert_u( const LevelData<EBCellFAB>& u )
{
   for( DataIterator di=m_grids.dataIterator(); di.ok() ; ++di )
   {
      DataIndex dex = di();
      for( BoxIterator b(m_stateNew[dex].box()); b.ok(); ++b )
      {
	 Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
         for( int vi=0 ; vi < vofs.size() ; vi++ )
            for( int c=0 ; c < SpaceDim ; c++ )
	    {
   	       m_stateNew[dex](vofs[vi],c+1) =
		 u[dex](vofs[vi],c)*m_stateNew[dex](vofs[vi],0);
	    }
      }
   }
}

//-----------------------------------------------------------------------
void EBAMRNS::define_heat( LevelData<EBCellFAB>& u,
		 	   LevelData<EBCellFAB>*& uc,
			   LevelData<EBCellFAB>*& uf,
			   LevelData<EBCellFAB>*& T,
			   LevelData<EBCellFAB>*& rhsT,
			   Real a_gamma, Real a_rom,
                           bool save_rhs, bool use_saved_rhs )
{
   EBCellFactory cfactory(m_ebisl);
   T    = new LevelData<EBCellFAB>( m_grids, 1, IntVect::Unit, cfactory );
   rhsT = new LevelData<EBCellFAB>( m_grids, 1, IntVect::Zero, cfactory );

   EBLevelDataOps::setToZero(*T);   

   LevelData<EBCellFAB>  rho( m_grids, 1, IntVect::Unit, cfactory );
   LevelData<EBCellFAB> work( m_grids, 1, IntVect::Zero, cfactory );

   EBLevelDataOps::setToZero(work);

   m_ns_viscosity->AMRResidual_eeq( work, uf, u, uc, false );

   Real ith = 1/(m_dt/2);
   //   ArrayToRaw atr;

   for( DataIterator di=m_grids.dataIterator(); di.ok() ; ++di )
   {
      DataIndex dex = di();
      Box state_box = m_stateNew[dex].box();
      Box t_box = (*T)[dex].box();
      for( BoxIterator b(state_box); b.ok(); ++b )
      {
	 Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
         for( int vi=0 ; vi < vofs.size() ; vi++ )
	 {
	    rho[dex](vofs[vi],0) = m_stateNew[dex](vofs[vi],0);
// Initial guess for T, and the right hand side for energy eq.
            Real rh= 0;
            for( int c=0 ; c < SpaceDim ; c++ )
	    {
 	       rh += u[dex](vofs[vi],c)*u[dex](vofs[vi],c);
	    }
            (*rhsT)[dex](vofs[vi],0) = ( (rh*rho[dex](vofs[vi],0)*0.5-
					  m_stateNew[dex](vofs[vi],SpaceDim+1))*ith )*
				    m_ebisl[dex].volFrac(vofs[vi])-work[dex](vofs[vi],0);
            Real prorho = (a_gamma-1)*
	      (m_stateNew[dex](vofs[vi],SpaceDim+1)/rho[dex](vofs[vi],0)-0.5*rh);
            (*T)[dex](vofs[vi],0) = prorho/a_rom;
	 }

      }
      fill_ghosts( (*T)[dex], dex, state_box, t_box );
      fill_ghosts( rho[dex], dex, state_box, t_box );
   }

   //   atr.output_array( rho, m_ebisl, "rho.bin");
   //   atr.output_array( work, m_ebisl, "work.bin");
   //   atr.output_array( u, m_ebisl, "u.bin");
   //   exit(2);
   if( save_rhs )
       EBLevelDataOps::assign(m_rhsT,*rhsT);

   if( use_saved_rhs )
       EBLevelDataOps::assign(*rhsT,m_rhsT);

   m_heatop_factory->set_rho( rho, m_level );
   m_heatop_factory->set_k2( a_rom/(a_gamma-1)/(m_dt/2), m_level );
}


//-----------------------------------------------------------------------
void EBAMRNS::insert_T( const LevelData<EBCellFAB>& T, Real a_gamma, Real a_rom )
{
   for( DataIterator di=m_grids.dataIterator(); di.ok() ; ++di )
   {
      DataIndex dex = di();
      for( BoxIterator b(m_stateNew[dex].getRegion()); b.ok(); ++b )
      {
	 Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
         for( int vi=0 ; vi < vofs.size() ; vi++ )
	 {
	    Real rho = m_stateNew[dex](vofs[vi],0);
            Real rh = 0;
            for( int c=0 ; c < SpaceDim ; c++ )
 	       rh += m_stateNew[dex](vofs[vi],c+1)*m_stateNew[dex](vofs[vi],c+1);
            m_stateNew[dex](vofs[vi],SpaceDim+1) = a_rom/(a_gamma-1)*rho*T[dex](vofs[vi],0)+
	      0.5*rh/rho;
	 }
      }
   }
}


//-----------------------------------------------------------------------
void EBAMRNS::postTimeStep()
{

  if( m_level == 0 && m_dovisc )
     implicit_solve_for_visc( m_gamma, m_rom, true );
  
  if(s_verbosity >= 3)
    {
      pout() << " in EBAMRNS postTimeStep for level " << m_level << endl;
    }
  Interval interv(0, m_nComp-1);
  if(m_hasCoarser)
    {
      if(m_doSmushing)
        {
          //redistibute to coarser level
          EBAMRNS* coarPtr = getCoarserLevel();
          //if use mass weighting, need to
          //fix weights of redistribution object
          if(m_useMassRedist)
            {
              int densevar = m_ebPatchGodunov->densityIndex();
              m_ebFineToCoarRedist.resetWeights(coarPtr->m_stateNew, densevar);
            }
          m_ebFineToCoarRedist.redistribute(coarPtr->m_stateNew, interv);
        }
      m_ebFineToCoarRedist.setToZero();
    }
  if (m_hasFiner)
    {
      EBAMRNS* finePtr = getFinerLevel();
      // reflux from finer level solution

      Real scale = -1.0/m_dx;
      m_ebFluxRegister.reflux(m_stateNew, interv, scale);
      //the flux register must modify the redistribution
      //registers
      m_ebFluxRegister.incrementRedistRegister(m_ebCoarToFineRedist,
                                               interv, scale);

      m_ebFluxRegister.incrementRedistRegister(m_ebCoarToCoarRedist,
                                               interv, scale);

      if(m_doSmushing)
        {
          //if use mass weighting, need to
          //fix weights of redistribution object
          if(m_useMassRedist)
            {
              int densevar = m_ebPatchGodunov->densityIndex();
              m_ebCoarToFineRedist.resetWeights(m_stateNew, densevar);
              m_ebCoarToCoarRedist.resetWeights(m_stateNew, densevar);
            }
          //redistibute to finer level
          m_ebCoarToFineRedist.redistribute(finePtr->m_stateNew, interv);
          //do the re redistirubtion
          m_ebCoarToCoarRedist.redistribute(m_stateNew, interv);
        }

      // average from finer level data
      if(m_doRZCoords)
        {
          //debug do area weighted averaging
          finePtr->m_ebCoarseAverage.average(m_stateNew,
                                             finePtr->m_stateNew,
                                             interv);
          //end debug
//          finePtr->m_ebCoarseAverage.averaRZ(m_stateNew,
//                                             finePtr->m_stateNew,
//                                             interv,
//                                             finePtr->m_dx);
        }
      else
        {
          finePtr->m_ebCoarseAverage.average(m_stateNew,
                                             finePtr->m_stateNew,
                                             interv);
        }

      m_ebCoarToFineRedist.setToZero();
      m_ebCoarToCoarRedist.setToZero();
    }
}


//-----------------------------------------------------------------------
void EBAMRNS::tagCells(IntVectSet& a_tags)
{
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::tagCells for level " << m_level << endl;
    }

  // Create tags based on undivided gradient of density
  IntVectSet localTags;
  if(!m_tagAll)
    {
      int inoz  = 0;
      ParmParse pp;
      //      int queryVal = pp.query("nozzle_problem", inoz);
      //      bool nozzleProblem = ((queryVal != 0) && (inoz == 1));
      pp.get("geometry",inoz);
      bool nozzleProblem = inoz == 1;

      Real xmaxTag, ymaxTag, xminTag, yminTag;
      vector<Real> problo(SpaceDim);
      bool tagTheLot = false;
      if(nozzleProblem)
        {
          pp.get("xmax_tag", xmaxTag);
          pp.get("ymax_tag", ymaxTag);
          pp.getarr("prob_lo", problo, 0, SpaceDim);
          int queryVal = pp.query("nozzle_tag_the_lot", inoz);
          tagTheLot = ((queryVal != 0) && (inoz == 1));
          if(tagTheLot)
            {
              pp.get("xmin_tag", xminTag);
              pp.get("ymin_tag", yminTag);
            }
        }
      if(tagTheLot)
        {
          CH_assert(nozzleProblem);
          for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
            {
              const Box& b = m_grids.get(dit());
              const EBISBox& ebisBox = m_ebisl[dit()];

              // Tag all cells where in the range

              IntVectSet ivsTot(b);
              for(VoFIterator vofit(ivsTot, ebisBox.getEBGraph());
                  vofit.ok(); ++vofit)
                {
                  const VolIndex& vof = vofit();
                  const IntVect& iv = vof.gridIndex();
                  Real xval, yval;
                  if(nozzleProblem)
                    {
                      xval = (iv[0] + 0.5)*m_dx + problo[0];
                      yval = (iv[1] + 0.5)*m_dx + problo[1];
                    }

                  if((xval < xmaxTag) && (yval < ymaxTag) && (xval > xminTag) && (yval > yminTag))
                    {
                      localTags |= iv;
                    }
                }
            }
        }
      else //actually compute gradients and all that to compute tags
        {
          // If there is a coarser level interpolate undefined ghost cells
          //only interpolate the density
          int densityIndex = m_ebPatchGodunov->densityIndex();
          Interval intervDensity(densityIndex, densityIndex);
          EBCellFactory factory(m_ebisl);
          int nCons = m_ebPatchGodunov->numConserved();
          LevelData<EBCellFAB> consTemp(m_grids, nCons, IntVect::Unit, factory);
          Interval consInterv(0, nCons-1);
          m_stateNew.copyTo(consInterv, consTemp, consInterv);
          if (m_hasCoarser)
            {
              const EBAMRNS* amrGodCoarserPtr = getCoarserLevel();
              int refRatCrse = amrGodCoarserPtr->refRatio();
              int nghost = 1;
              EBPWLFillPatch patcher(m_grids,
                                     amrGodCoarserPtr->m_grids,
                                     m_ebisl,
                                     amrGodCoarserPtr->m_ebisl,
                                     amrGodCoarserPtr->m_domainBox,
                                     refRatCrse, m_nComp, nghost);

              Real coarTimeOld = 0.0;
              Real coarTimeNew = 1.0;
              Real fineTime    = 0.0;
              patcher.interpolate(consTemp,
                                  amrGodCoarserPtr->m_stateOld,
                                  amrGodCoarserPtr->m_stateNew,
                                  coarTimeOld,
                                  coarTimeNew,
                                  fineTime,
                                  intervDensity);
            }
          consTemp.exchange(intervDensity);

          // Compute undivided gradient
          for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
            {
              const Box& b = m_grids.get(dit());
              const EBISBox& ebisBox = m_ebisl[dit()];
              EBCellFAB gradFab(ebisBox, b, SpaceDim);
              const EBCellFAB& stateFab = consTemp[dit()];
              BaseFab<Real>& regGradFab = gradFab.getSingleValuedFAB();
              const BaseFab<Real>& regStateFab = stateFab.getSingleValuedFAB();

              for (int idir = 0; idir < SpaceDim; ++idir)
                {
                  const Box bCenter = b & grow(m_problem_domain,-BASISV(idir));
                  const Box bLo     = b & adjCellLo(bCenter,idir);
                  const int hasLo = ! bLo.isEmpty();
                  const Box bHi     = b & adjCellHi(bCenter,idir);
                  const int hasHi = ! bHi.isEmpty();

                  FORT_GETRELATIVEGRAD(CHF_FRA1(regGradFab,idir),
                                       CHF_CONST_FRA1(regStateFab,0),
                                       CHF_CONST_INT(idir),
                                       CHF_BOX(bLo),
                                       CHF_CONST_INT(hasLo),
                                       CHF_BOX(bHi),
                                       CHF_CONST_INT(hasHi),
                                       CHF_BOX(bCenter));

                  //do one-sided diffs where necessary at irregular cells.
                  IntVectSet ivsIrreg = ebisBox.getIrregIVS(b);
                  for(VoFIterator vofit(ivsIrreg, ebisBox.getEBGraph());
                      vofit.ok(); ++vofit)
                    {
                      const VolIndex& vof = vofit();
                      const IntVect&  iv = vof.gridIndex();
                      //one-sided diffs on domain bndry
                      bool onLeftDomain = iv[idir] == m_domainBox.smallEnd(idir);
                      bool onRighDomain = iv[idir] == m_domainBox.bigEnd(idir);
                      bool hasFacesLeft = (ebisBox.numFaces(vof, idir, Side::Lo) > 0) && !onLeftDomain;
                      bool hasFacesRigh = (ebisBox.numFaces(vof, idir, Side::Hi) > 0) && !onRighDomain;

                      Real valCent = stateFab(vof, densityIndex);
                      Real dpl, dpr, dpc;
                      //compute one-sided diffs where you have them
                      if(hasFacesLeft)
                        {
                          Vector<FaceIndex> facesLeft =
                            ebisBox.getFaces(vof, idir, Side::Lo);
                          Real valLeft = 0.0;
                          for(int iface = 0; iface <facesLeft.size(); iface++)
                            {
                              VolIndex vofLeft = facesLeft[iface].getVoF(Side::Lo);
                              valLeft += stateFab(vofLeft, densityIndex);
                            }
                          valLeft /= Real(facesLeft.size());
                          dpl = valCent - valLeft;
                        }
                      if(hasFacesRigh)
                        {
                          Vector<FaceIndex> facesRigh =
                            ebisBox.getFaces(vof, idir, Side::Hi);
                          Real valRigh = 0.0;
                          for(int iface = 0; iface <facesRigh.size(); iface++)
                            {
                              VolIndex vofRigh = facesRigh[iface].getVoF(Side::Hi);
                              valRigh += stateFab(vofRigh, densityIndex);
                            }
                          valRigh /= Real(facesRigh.size());
                          dpr = valRigh - valCent;
                        }
                      if(hasFacesLeft && hasFacesRigh)
                        {
                          dpc = 0.5*(dpl+dpr);
                        }
                      else if(!hasFacesLeft && !hasFacesRigh)
                        {
                          dpc = 0.0;
                        }
                      else if(hasFacesLeft && !hasFacesRigh)
                        {
                          dpc = dpl;
                        }
                      else if(hasFacesRigh && !hasFacesLeft)
                        {
                          dpc = dpr;
                        }

                      gradFab(vof, idir) = dpc/valCent;
                    }
                }

              EBCellFAB gradMagFab(ebisBox, b, 1);
              BaseFab<Real>& regGradMagFab = gradMagFab.getSingleValuedFAB();
              FORT_MAGNITUDE(CHF_FRA1(regGradMagFab,0),
                             CHF_CONST_FRA(regGradFab),
                             CHF_BOX(b));

              //pointwise op so just have to iterate over multivalued cells
              IntVectSet ivsMulti = ebisBox.getMultiCells(b);
              for(VoFIterator vofit(ivsMulti, ebisBox.getEBGraph());
                  vofit.ok(); ++vofit)
                {
                  const VolIndex& vof = vofit();
                  Real mag = 0.0;
                  for(int idir = 0; idir < SpaceDim; idir++)
                    {
                      Real graddir = gradFab(vof, idir);
                      mag += graddir*graddir;
                    }
                  mag = sqrt(mag);
                  gradMagFab(vof, 0) = mag;
                }

              // Tag where gradient exceeds threshold

              IntVectSet ivsTot(b);
              for(VoFIterator vofit(ivsTot, ebisBox.getEBGraph());
                  vofit.ok(); ++vofit)
                {
                  const VolIndex& vof = vofit();
                  const IntVect& iv = vof.gridIndex();
                  Real xval, yval;
                  if(nozzleProblem)
                    {
                      xval = (iv[0] + 0.5)*m_dx + problo[0];
                      yval = (iv[1] + 0.5)*m_dx + problo[1];
                    }

                  if((!nozzleProblem) || (xval < xmaxTag) && (yval < ymaxTag))
                    {
                      if (gradMagFab(vof, 0) >= m_refineThresh)
                        {
                          localTags |= iv;
                        }
                    }
                }

              //refine all irregular cells
              //this is probably not ideal.
              IntVectSet irregIVS = ebisBox.getIrregIVS(b);
              localTags |= irregIVS;
            }
        }

      localTags.grow(m_tagBufferSize);

      // Need to do this in two steps unless a IntVectSet::operator &=
      // (ProblemDomain) operator is defined
      Box localTagsBox = localTags.minBox();
      localTagsBox &= m_problem_domain;
      localTags &= localTagsBox;
    }//if !tagall
  else
    {
      localTags.makeEmpty();
      for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
        {
          const Box& b = m_grids.get(dit());
          const EBISBox& ebisBox = m_ebisl[dit()];
          IntVectSet ivsTot(b);
          for(VoFIterator vofit(ivsTot, ebisBox.getEBGraph());
              vofit.ok(); ++vofit)
            {
              localTags |= vofit().gridIndex();
            }
        }
    }

  Vector<IntVectSet> allTags;

  const int destProc = uniqueProc(SerialTask::compute);

  gather(allTags,localTags,destProc);

  if (procID() == uniqueProc(SerialTask::compute))
    {
      for (int i = 0; i < allTags.size(); ++i)
        {
          a_tags |= allTags[i];
        }
    }
}


//-----------------------------------------------------------------------
void EBAMRNS::tagCellsInit(IntVectSet& a_tags)
{
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::tagCellsInit for level " << m_level << endl;
    }

  // Since tags are calculated using only current time step data, use
  // the same tagging function for initialization and for regridding.
  tagCells(a_tags);
}


//-----------------------------------------------------------------------
void EBAMRNS::regrid(const Vector<Box>& a_new_grids)
{
  if(s_verbosity >= 3)
    {
      pout() << " in EBAMRNS regrid for level " << m_level << endl;
    }

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  // save data for later copy
  //not using m_ebisl because it gets wiped later
  //the ebisl has to know about the fact that we really have
  //four ghost cells.
  int nGhostEBISL = 6;

  EBISLayout ebislOld;
  ebisPtr->fillEBISLayout(ebislOld, m_grids, m_domainBox, nGhostEBISL);
  EBCellFactory factoryOld(ebislOld);
  IntVect ivGhost = m_nGhost*IntVect::Unit;
  LevelData<EBCellFAB> stateSaved(m_grids, m_nComp, ivGhost, factoryOld);
  Interval interv(0,m_nComp-1);
  stateSaved.define(m_grids, m_nComp, ivGhost, factoryOld);
  m_stateNew.copyTo(interv, stateSaved, interv);

  //create grids and ebis layouts
  //the ebisl has to know about the fact that we really have
  //four ghost cells.
  m_level_grids = a_new_grids;
  Vector<int> proc_map;
  if (procID() == uniqueProc(SerialTask::compute))
    {
      LoadBalance(proc_map,a_new_grids);
    }
  broadcast(proc_map,uniqueProc(SerialTask::compute));
  m_grids= DisjointBoxLayout(a_new_grids,proc_map);

  ebisPtr->fillEBISLayout(m_ebisl, m_grids, m_domainBox, nGhostEBISL);

  EBCellFactory factoryNew(m_ebisl);
  // reshape state with new grids
  m_stateNew.define(m_grids,m_nComp,ivGhost, factoryNew);
  m_stateOld.define(m_grids,m_nComp,ivGhost, factoryNew);

  m_rhs.define( m_grids,SpaceDim,IntVect::Zero,factoryNew);
  m_rhsT.define(m_grids,1,IntVect::Zero,factoryNew);

  // set up data structures
  levelSetup();

  // interpolate to coarser level
  if (m_hasCoarser)
    {
      EBAMRNS* coarPtr = getCoarserLevel();
      m_ebFineInterp.interpolate(m_stateNew,
                                 coarPtr->m_stateNew,
                                 interv);
    }

  // copy from old state
  stateSaved.copyTo(interv,m_stateNew, interv);
}

//-----------------------------------------------------------------------
void EBAMRNS::initialGrid(const Vector<Box>& a_new_grids)
{
  if(s_verbosity >= 3)
    {
      pout() << " in EBAMRNS initialGrid for level " << m_level << endl;
    }

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  m_level_grids = a_new_grids;

  // load balance and create boxlayout
  Vector<int> proc_map;
  LoadBalance(proc_map,a_new_grids);
  if(s_verbosity >= 3)
    {
      pout() << " just loadbalanced " << m_level << endl;
    }

  m_grids = DisjointBoxLayout(a_new_grids,proc_map);
  if(s_verbosity >= 3)
    {
      pout() << "EBAMRNS::initialgrid grids " << endl;
      dumpDBL(&m_grids);
    }

  //the ebisl has to know about the fact that we really have
  //four ghost cells.
  int nGhostEBISL = 6;
  if(s_verbosity >= 3)
    {
      pout() << " about to fill ebislayout  in EBAMRNS initialGrid for level " << m_level << endl;
    }

  ebisPtr->fillEBISLayout(m_ebisl, m_grids,  m_domainBox, nGhostEBISL);

  if(s_verbosity >= 3)
    {
      pout() << " done with filling ebislayout  in EBAMRNS initialGrid for level " << m_level << endl;
    }

  EBCellFactory factoryNew(m_ebisl);
  IntVect ivGhost = m_nGhost*IntVect::Unit;
  m_stateNew.define(m_grids,m_nComp, ivGhost, factoryNew);
  m_stateOld.define(m_grids,m_nComp, ivGhost, factoryNew);
  m_rhs.define( m_grids,SpaceDim,IntVect::Zero,factoryNew);
  m_rhsT.define(m_grids,1,IntVect::Zero,factoryNew);

  // set up data structures
  levelSetup();
}


//-----------------------------------------------------------------------
void EBAMRNS::initialData()
{

  if(s_verbosity >= 3)
      pout() << " in EBAMRNS initialData for level " << m_level << endl;

  if( m_doforce )
  {
     RealVect problo = Chombo_EBIS::instance()->getOrigin();
     RealVect dxv = RealVect::Unit;
     dxv *= m_dx;
     m_forcing->get_manufactured_solution( m_stateNew, m_ebisl, 0.0, dxv, problo );
     m_forcing->get_manufactured_solution( m_stateOld, m_ebisl, 0.0, dxv, problo );
  }
  else
  {
     const EBPhysIBC* const ebphysIBCPtr = m_ebPatchGodunov->getEBPhysIBC();
     ebphysIBCPtr->initialize(m_stateNew, m_ebisl);
     ebphysIBCPtr->initialize(m_stateOld, m_ebisl);
  }
}


//-----------------------------------------------------------------------
void EBAMRNS::postInitialize()
{
}


//-----------------------------------------------------------------------
void EBAMRNS::syncWithFineLevel()
{
  if(s_verbosity >= 3)
    {
      pout() << " in EBAMRNS syncWithFineLevel for level " << m_level << endl;
    }
  //stuff that needs to be setup from the finer
  //level.  A bunch of objects depend upon the layouts
  //from both levels and the finer level changes more
  //often from regrid so this needs to be called from the finer
  //level
  CH_assert(m_hasFiner);
  if(m_hasFiner)
    {
      EBAMRNS* finePtr = getFinerLevel();
      int nRefFine = refRatio();
      const DisjointBoxLayout& finer_m_grids = finePtr->m_grids;
      const EBISLayout& finer_m_ebisl = finePtr->m_ebisl;
      const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
      //define fine to coarse redistribution object
      //for now set to volume weighting
      m_ebCoarToFineRedist.define(finer_m_grids, m_grids,
                                  finer_m_ebisl,  m_ebisl,
                                  m_domainBox, nRefFine , m_nComp, 2, ebisPtr);
      //define coarse to coarse redistribution object
      m_ebCoarToCoarRedist.define(finer_m_grids, m_grids,
                                  finer_m_ebisl,  m_ebisl,
                                  m_domainBox, nRefFine , m_nComp);
      // maintain flux registers
      m_ebFluxRegister.define(finer_m_grids,
                              m_grids,
                              finer_m_ebisl,
                              m_ebisl,
                              m_domainBox,
                              nRefFine,
                              m_nComp,ebisPtr);

      //set all the registers to zero
      m_ebCoarToFineRedist.setToZero();
      m_ebCoarToCoarRedist.setToZero();
      m_ebFluxRegister.setToZero();
    }

}


//-----------------------------------------------------------------------
void EBAMRNS::patchGodunov(const EBPatchGodunovFactory* const a_ebPatchGodunovFactory)
{
  m_ebPatchGodunovFactory = a_ebPatchGodunovFactory;
}


//-----------------------------------------------------------------------
Real EBAMRNS::computeDt()
{
  Real newDt;
  newDt = m_dtNew;

  return newDt;
}


//-----------------------------------------------------------------------
Real EBAMRNS::computeInitialDt()
{
  Real maxwavespeed =  m_ebLevelGodunov.getMaxWaveSpeed(m_stateNew);
  Real newDT = m_initial_dt_multiplier * m_dx /maxwavespeed;

  return newDT;
}


//-----------------------------------------------------------------------
void EBAMRNS::CFL(Real a_cfl)
{
  m_cfl = a_cfl;
}


//-----------------------------------------------------------------------
void EBAMRNS::domainLength(Real a_domainLength)
{
  m_domainLength = a_domainLength;
}


//-----------------------------------------------------------------------
void EBAMRNS::refinementThreshold(Real a_refineThresh)
{
  m_refineThresh = a_refineThresh;
}


//-----------------------------------------------------------------------
void EBAMRNS::tagBufferSize(int a_tagBufferSize)
{
  m_tagBufferSize = a_tagBufferSize;
}


//-----------------------------------------------------------------------
void EBAMRNS::sumConserved(Real& a_sumcons,
                           const int& a_ivar ) const
{
  Real sumgrid = 0;
  for(DataIterator dit= m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBCellFAB& state = m_stateNew[dit()];
      Box thisBox = m_grids.get(dit());
      IntVectSet uberIVS(thisBox);
      const EBISBox& ebisBox = m_ebisl[dit()];
      for(VoFIterator vofit(uberIVS, ebisBox.getEBGraph());
          vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          Real consVal = state(vof, a_ivar);
          Real volFrac = ebisBox.volFrac(vof);
          Real volume = volFrac;
          if(m_doRZCoords)
            {
              Real cellVol, kvol;
              EBArith::getKVolRZ(kvol, cellVol, ebisBox, m_dx, vof);
              volume = cellVol*kvol;
            }
          sumgrid += consVal*volume;
        }
    }

  Vector<Real> all_sum;
  gather(all_sum,sumgrid,uniqueProc(SerialTask::compute));
  Real sumallgrid = 0.;
  if (procID() == uniqueProc(SerialTask::compute))
    {
      for (int i = 0; i < all_sum.size(); ++i)
        {
          sumallgrid += all_sum[i];
        }
    }
  broadcast(sumallgrid,uniqueProc(SerialTask::compute));
  a_sumcons = sumallgrid;
}


//-----------------------------------------------------------------------
EBAMRNS* EBAMRNS::getCoarserLevel() const
{
  EBAMRNS* retval = NULL;
  if(m_coarser_level_ptr != NULL)
    {
      retval = dynamic_cast <EBAMRNS*> (m_coarser_level_ptr);

      if (retval == NULL)
        {
          pout() << "EBAMRG::getCoarserLevel: dynamic cast failed"
                 << endl;
          MayDay::Error();
        }
    }
  return retval;
}


//-----------------------------------------------------------------------
EBAMRNS* EBAMRNS::getFinerLevel() const
{
  EBAMRNS* retval = NULL;
  if(m_finer_level_ptr != NULL)
    {
      retval = dynamic_cast <EBAMRNS*> (m_finer_level_ptr);

      if (retval == NULL)
        {
          pout() << "EBAMRG::getFinerLevel: dynamic cast failed"
                 << endl;
          MayDay::Error();
        }
    }
  return retval;
}


//-----------------------------------------------------------------------
void EBAMRNS::levelSetup()
{
  if(s_verbosity >= 3)
    {
      pout() << " in EBAMRNS levelSetup for level " << m_level << endl;
    }

  EBAMRNS* coarPtr = getCoarserLevel();
  EBAMRNS* finePtr = getFinerLevel();

  m_hasCoarser = (coarPtr != NULL);
  m_hasFiner   = (finePtr != NULL);

  if (m_hasCoarser)
    {
      int nRefCrse = m_coarser_level_ptr->refRatio();
      const DisjointBoxLayout& coarser_m_grids = coarPtr->m_grids;
      const EBISLayout& coarser_m_ebisl = coarPtr->m_ebisl;
      const Box& domainCoar = coarPtr->m_domainBox;
      const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();

      m_ebCoarseAverage.define(m_grids,
                               coarser_m_grids,
                               m_ebisl,
                               coarser_m_ebisl,
                               domainCoar,
                               nRefCrse,
                               m_nComp,ebisPtr);
      m_ebFineInterp.define(m_grids,
                            coarser_m_grids,
                            m_ebisl,
                            coarser_m_ebisl,
                            domainCoar,
                            nRefCrse,
                            m_nComp);

      // maintain levelgodunov
      m_ebLevelGodunov.define(m_grids,
                              coarser_m_grids,
                              m_ebisl,
                              coarser_m_ebisl,
                              m_problem_domain,
                              nRefCrse,
                              m_dx,
                              m_useMassRedist,
                              m_doSmushing,
                              m_doRZCoords,
                              m_hasSourceTerm,
                              m_ebPatchGodunovFactory,
                              m_hasCoarser,
                              m_hasFiner);

      //define fine to coarse redistribution object
      //for now set to volume weighting
      m_ebFineToCoarRedist.define(m_grids, coarser_m_grids,
                                  m_ebisl, coarser_m_ebisl,
                                  domainCoar, nRefCrse, m_nComp);
      m_ebFineToCoarRedist.setToZero();

      coarPtr->syncWithFineLevel();
    }
  else
    {
      m_ebLevelGodunov.define(m_grids,
                              DisjointBoxLayout(),
                              m_ebisl,
                              EBISLayout(),
                              m_problem_domain,
                              m_ref_ratio,
                              m_dx,
                              m_useMassRedist,
                              m_doSmushing,
                              m_doRZCoords,
                              m_hasSourceTerm,
                              m_ebPatchGodunovFactory,
                              m_hasCoarser,
                              m_hasFiner);
    }

  //set up mass redistribution array
  m_sets.define(m_grids);
  for(DataIterator dit = m_grids.dataIterator();
      dit.ok(); ++dit)
    {
      Box thisBox = m_grids.get(dit());
      m_sets[dit()] = m_ebisl[dit()].getIrregIVS(thisBox);
    }
  BaseIVFactory<Real> factory(m_ebisl, m_sets);
  //the ghost cells on the mass redistribution array
  //are tied to the redistribution radius
  int nghost = 2*m_redistRad;
  IntVect ivGhost = nghost*IntVect::Unit;
  m_massDiff.define(m_grids, m_nComp, ivGhost, factory);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      m_massDiff[dit()].setVal(0.0);
    }
}


//-----------------------------------------------------------------------
LevelData<EBCellFAB>& EBAMRNS::getStateNew()
{
  return m_stateNew;
}


//-----------------------------------------------------------------------
LevelData<EBCellFAB>& EBAMRNS::getStateOld()
{
  return m_stateOld;
}


//-----------------------------------------------------------------------
Real EBAMRNS::getDx() const
{
  return m_dx;
}


//-----------------------------------------------------------------------
Real EBAMRNS::getDt() const
{
  return m_dt;
}


//-----------------------------------------------------------------------
const EBISLayout& EBAMRNS::getEBISLayout() const
{
  return m_ebisl;
}


//-----------------------------------------------------------------------
void EBAMRNS::fillConsAndPrim(LevelData<EBCellFAB>& a_data) const
{
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::fillConsAndPrim" << endl;
    }
  //fill output data including a bunch of geometric crud
  int nCons = m_ebPatchGodunov->numConserved();
  int nPrim = m_ebPatchGodunov->numPrimitives() ;
  int consAndPrim = nCons + nPrim;

  EBCellFactory ebcellfact(m_ebisl);
  a_data.define(m_grids, consAndPrim, IntVect::Zero, ebcellfact);

  for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisbox = m_ebisl[dit()];
      const Box& grid = m_grids.get(dit());
      const EBCellFAB& consfab = m_stateNew[dit()];
      EBCellFAB primfab(ebisbox, grid, nPrim);
      //cfivs, time and timestep fake and not used here
      Real faket = 1.0;
      int logflag = 0;
      IntVectSet emptyivs;
      m_ebPatchGodunov->setValidBox(grid, ebisbox, emptyivs, faket, faket);
      m_ebPatchGodunov->consToPrim(primfab, consfab, grid,logflag);

      EBCellFAB& outputfab = a_data[dit()];

      Interval consIntervSrc(0, nCons-1);
      Interval consIntervDst(0, nCons-1);
      Interval primIntervSrc(0, nPrim-1);
      Interval primIntervDst(nCons, consAndPrim-1);

      // copy regular data
      outputfab.copy(grid, consIntervDst,  grid, consfab, consIntervSrc);
      outputfab.copy(grid, primIntervDst,  grid, primfab, primIntervSrc);
      Real coveredVal = -10.0;
      for(int ivar = 0; ivar < consAndPrim; ivar++)
        {
          outputfab.setInvalidData(coveredVal, ivar);
        }

    }//end loop over grids
}



#ifdef CH_USE_HDF5
//-----------------------------------------------------------------------
void EBAMRNS::writeCheckpointHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::writeCheckpointHeader" << endl;
    }

  //stuff in non-eb checkpoint header is already
  //set in the define function, such as
  // Setup the number of components
  // Setup the component names
  //so i will eliminate the middleman
}


//-----------------------------------------------------------------------
void EBAMRNS::readCheckpointHeader(HDF5Handle& a_handle)
{
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::readCheckpointHeader" << endl;
    }

  //stuff in non-eb checkpoint header is already
  //set in the define function, such as
  // Setup the number of components
  // Setup the component names
  // So i will eliminate the middleman.
}


//-----------------------------------------------------------------------
void EBAMRNS::writeCheckpointLevel(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::writeCheckpointLevel" << endl;
    }

  //reasons for deviations from non-eb stuff
  // the tag buffer size is set by the factory.
  // dx is set in the define function
  // the problem domain is set in the define function

  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  a_handle.setGroup(label);

  // Setup the level header information
  HDF5HeaderData header;

  header.m_int ["ref_ratio"]       = m_ref_ratio;
  header.m_real["dt"]              = m_dt;

  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::writeCheckpointLevel 2" << endl;
    }

  // Write the header for this level
  header.writeToFile(a_handle);

  // Write the data for this level
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::writeCheckpointLevel 3" << endl;
    }
  write(a_handle,m_grids);
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::writeCheckpointLevel 4" << endl;
    }
  write(a_handle,m_stateOld,"dataOld");
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::writeCheckpointLevel 5" << endl;
    }
  write(a_handle,m_stateNew,"dataNew");
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::writeCheckpointLevel 6" << endl;
    }
}


//-----------------------------------------------------------------------
void EBAMRNS::readCheckpointLevel(HDF5Handle& a_handle)
{
  CH_assert(m_isDefined);
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::readCheckpointLevel" << endl;
    }

  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  // Read the header for this level
  a_handle.setGroup(label);

  HDF5HeaderData header;
  header.readFromFile(a_handle);

  // Get the refinement ratio
  if (header.m_int.find("ref_ratio") == header.m_int.end())
    {
      MayDay::Error("::readCheckpointLevel: file does not contain ref_ratio");
    }
  m_ref_ratio = header.m_int["ref_ratio"];

  //reasons for deviations from non-eb stuff
  // the tag buffer size is set by the factory.
  // dx is set in the define function
  // the problem domain is set in the define function

  // Get dt
  if (header.m_real.find("dt") == header.m_real.end())
    {
      MayDay::Error("readCheckpointLevel: file does not contain dt");
    }
  m_dt = header.m_real["dt"];

  // Get the grids
  Vector<Box> vboxGrids;
  const int gridStatus = read(a_handle, vboxGrids);
  if (gridStatus != 0)
    {
      MayDay::Error("readCheckpointLevel: file has no grids");
    }

  Vector<int> proc_map;
  if (procID() == uniqueProc(SerialTask::compute))
    {
      LoadBalance(proc_map,vboxGrids);
    }
  broadcast(proc_map, uniqueProc(SerialTask::compute));
  m_grids= DisjointBoxLayout(vboxGrids,proc_map);

  //this keeps the order of the AMRLevel m_level_grids
  //consistent with m_grids
  LayoutIterator lit = m_grids.layoutIterator();
  for (lit.begin(); lit.ok(); ++lit)
    {
      Box b = m_grids.get(lit());
      m_level_grids.push_back(b);
    }

  //the ebisl has to know about the fact that we really have
  //four ghost cells.
  int nGhostEBISL = 6;
  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  ebisPtr->fillEBISLayout(m_ebisl, m_grids,
                          m_domainBox, nGhostEBISL);
  EBCellFactory factoryNew(m_ebisl);
  //m_nghost is set in define function
  IntVect ivGhost = m_nGhost*IntVect::Unit;
  m_stateNew.define(m_grids,m_nComp, ivGhost, factoryNew);
  m_stateOld.define(m_grids,m_nComp, ivGhost, factoryNew);
  m_rhs.define( m_grids,SpaceDim,IntVect::Zero,factoryNew);
  m_rhsT.define(m_grids,1,IntVect::Zero,factoryNew);

  //  Interval vars(0, m_nComp-1);
  //the false says to not redefine the data
  int dataStatusNew = read<EBCellFAB>(a_handle,
                                      m_stateNew,
                                      "dataNew",
                                      m_grids,
                                      Interval(),
                                      false);

  int dataStatusOld = read<EBCellFAB>(a_handle,
                                      m_stateOld,
                                      "dataOld",
                                      m_grids,
                                      Interval(),
                                      false);

  if ((dataStatusNew != 0) || (dataStatusOld != 0))
    {
      MayDay::Error("file does not contain state data");
    }
  // Set up data structures
  levelSetup();
}


//-----------------------------------------------------------------------
void EBAMRNS::writePlotHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRNS::writePlotHeader" << endl;
    }

  HDF5HeaderData header;
  // Setup the number of components
  //have to add in a lot of geometric crap.
  // 3 norms + 6 area fracs + 1 distance + 1 volFrac
  // =  11 extra components
  //forces 3d
  int nCons = m_ebPatchGodunov->numConserved();
  int nPrim = m_ebPatchGodunov->numPrimitives() ;
  int consAndPrim = nCons + nPrim;

  int indexVolFrac = consAndPrim;
  int indexAreaFrac = indexVolFrac+1;
  int indexNormal = indexAreaFrac+ 2*SpaceDim;
  int indexDist = indexNormal+SpaceDim;
  int nCompTotal = indexDist+1;
  CH_assert(nCompTotal == consAndPrim + 3*SpaceDim+2);

  Vector<string> names(nCompTotal);

  for (int i = 0; i < nCons; i++)
    {
      names[i] = m_stateNames[i];
    }
  for (int i = 0; i < nPrim; i++)
    {
      names[nCons + i] = m_primNames[i];
    }

  string volFracName("fraction-0");
  Vector<string> normName(3);
  Vector<string> areaName(6);
  string distName("distance-0");

  normName[0] = "xnormal-0";
  normName[1] = "ynormal-0";
  normName[2] = "znormal-0";

  areaName[0] = "xAreafractionLo-0";
  areaName[1] = "xAreafractionHi-0";
  areaName[2] = "yAreafractionLo-0";
  areaName[3] = "yAreafractionHi-0";
  areaName[4] = "zAreafractionLo-0";
  areaName[5] = "zAreafractionHi-0";

  names[indexVolFrac] = volFracName;

  for (int i=0; i < 2*SpaceDim; i++)
    {
      names[indexAreaFrac+i] = areaName[i];
    }

  for (int i=0; i < SpaceDim; i++)
    {
      names[indexNormal+i] = normName[i];
    }

  names[indexDist] = distName;

  //now output this into the hdf5 handle
  header.m_int["num_components"] = nCompTotal;
  // Setup the component names
  char compStr[30];
  for (int comp = 0; comp < nCompTotal; ++comp)
    {
      sprintf(compStr,"component_%d",comp);
      header.m_string[compStr] = names[comp];
    }

  // Write the header to the file
  header.writeToFile(a_handle);

  if (s_verbosity >= 4)
    {
      pout() << header << endl;
    }
}



//-----------------------------------------------------------------------
void EBAMRNS::writePlotLevel(HDF5Handle& a_handle) const
{

  if (s_verbosity >= 3)
    {
      pout() << "EBAMRGodunov::writePlotLevel" << endl;
    }
  //fill output data including a bunch of geometric crud
  int nCons = m_ebPatchGodunov->numConserved();
  int nPrim = m_ebPatchGodunov->numPrimitives() ;
  int consAndPrim = nCons + nPrim;
  int indexVolFrac = consAndPrim;
  int indexAreaFrac = indexVolFrac+1;
  int indexNormal = indexAreaFrac+ 2*SpaceDim;
  int indexDist = indexNormal+SpaceDim;
  int nCompTotal = indexDist+1;
  CH_assert(nCompTotal == consAndPrim + 3*SpaceDim+2);

  Vector<Real> coveredValuesCons(nCons, -10.0);
  Vector<Real> coveredValuesPrim(nPrim, -10.0);

  Vector<Real> coveredValues;
  coveredValues.append(coveredValuesCons);
  coveredValues.append(coveredValuesPrim);

  LevelData<FArrayBox> fabData(m_grids, nCompTotal, IntVect::Zero);

#ifdef MPI
    MPI_Barrier(Chombo_MPI::comm);
#endif
  for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisbox = m_ebisl[dit()];
      const Box& grid = m_grids.get(dit());
      const EBCellFAB& consfab = m_stateNew[dit()];
      EBCellFAB primfab(ebisbox, grid, nPrim);
      //cfivs, time and timestep fake and not used here
      Real faket = 1.0;
      IntVectSet emptyivs;
      int logflag = 1;
      m_ebPatchGodunov->setValidBox(grid, ebisbox, emptyivs, faket, faket);
      m_ebPatchGodunov->consToPrim(primfab, consfab, grid, logflag);

      FArrayBox& currentFab = fabData[dit()];

      // copy regular data
      currentFab.copy(consfab.getSingleValuedFAB(),0,0,nCons);
      currentFab.copy(primfab.getSingleValuedFAB(),0,nCons,nPrim);

      // set default volume fraction
      currentFab.setVal(1.0,indexVolFrac);

      // set default area fractions
      for (int i=0; i < 2*SpaceDim; i++)
        {
          currentFab.setVal(1.0,indexAreaFrac+i);
        }

      // set default normal
      for (int i=0; i < SpaceDim; i++)
        {
          currentFab.setVal(0.0,indexNormal+i);
        }

      // set default distance of EB from corner
      currentFab.setVal(0.0,indexDist);

      // set special values
      // iterate through the current grid
      // NOTE:  this is probably an inefficient way to do this
      for (BoxIterator bit(grid); bit.ok(); ++bit)
        {
          const IntVect& iv = bit();
          // set special values for covered cells
          if (ebisbox.isCovered(iv))
            {
              for (int icomp = 0; icomp < consAndPrim; icomp++)
                {
                  Real cval = coveredValues[icomp];

                  currentFab(iv,icomp) = cval;
                }
              // volume fraction is zero
              currentFab(iv,indexVolFrac) = 0.0;

              // area fractions are zero
              for (int i=0; i < 2*SpaceDim; i++)
                {
                  currentFab(iv,indexAreaFrac+i) = 0.0;
                }
            }

          // set special values for irregular cells
          if (ebisbox.isIrregular(iv))
            {
              Vector<VolIndex> vofs = ebisbox.getVoFs(iv);
              Real volFrac = ebisbox.volFrac(vofs[0]);
              RealVect normal = ebisbox.normal(vofs[0]);

              // set volume fraction
              currentFab(iv,indexVolFrac) = volFrac;

              // set area fractions--use only the first face you find
              for (int i=0; i < SpaceDim; i++)
                {
                  Vector<FaceIndex> faces;

                  faces = ebisbox.getFaces(vofs[0],i,Side::Lo);
                  if (faces.size() == 0)
                    {
                      currentFab(iv,indexAreaFrac+2*i) = 0.0;
                    }
                  else
                    {
                      currentFab(iv,indexAreaFrac+2*i) =
                        ebisbox.areaFrac(faces[0]);
                    }

                  faces = ebisbox.getFaces(vofs[0],i,Side::Hi);
                  if (faces.size() == 0)
                    {
                      currentFab(iv,indexAreaFrac+2*i+1) = 0.0;
                    }
                  else
                    {
                      currentFab(iv,indexAreaFrac+2*i+1) =
                        ebisbox.areaFrac(faces[0]);
                    }
                }

              // set normal
              for (int i=0; i < SpaceDim; i++)
                {
                  currentFab(iv,indexNormal+i) = normal[i];
                }

              // set distance unless the length of the normal is zero
              Real length = PolyGeom::dot(normal,normal);

              if (length > 0)
                {
                  Real dist = PolyGeom::computeAlpha(volFrac,normal)*m_dx;
                  currentFab(iv,indexDist) = -dist;
                }
            } //end if(isIrregular)
        }//end loop over cells
    }//end loop over grids

#ifdef MPI
    MPI_Barrier(Chombo_MPI::comm);
#endif
  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  a_handle.setGroup(label);

  // Setup the level header information
  HDF5HeaderData header;

  header.m_int ["ref_ratio"]   = m_ref_ratio;
  header.m_real["dx"]          = m_dx;
  header.m_real["dt"]          = m_dt;
  header.m_real["time"]        = m_time;
  header.m_box ["prob_domain"] = m_problem_domain.domainBox();

  // Write the header for this level
  header.writeToFile(a_handle);

  if (s_verbosity >= 4)
    {
      pout() << header << endl;
    }

  // Write the data for this level
  write(a_handle,fabData.boxLayout());
  write(a_handle,fabData,"data");
}

#endif


//-----------------------------------------------------------------------
void EBAMRNS::errors_with_forcing()
{
   RealVect problo = Chombo_EBIS::instance()->getOrigin();
   RealVect dxv = RealVect::Unit;
   dxv *= m_dx;
// use m_stateOld as temporary array
   m_forcing->get_manufactured_solution( m_stateOld, m_ebisl, m_time, dxv, problo );

   EBLevelDataOps::incr( m_stateOld, m_stateNew, -1.0 );
   cout << "L2-norm error   " << norm( m_stateOld,2) << endl;
   cout << "Linf-norm error " << norm( m_stateOld,0) << endl;
   Vector<DisjointBoxLayout> grids(1,m_grids);
   Vector<LevelData<EBCellFAB>* > data(1,&m_stateOld);
   Vector<string> names(4);
   Vector<Real> cover(4,1e-20);
   names[0] = "rho";
   names[1] = "mx";
   names[2] = "my";
   names[3] = "e";
   Vector<int> ratios(1,2);
   writeEBHDF5( "error1.hdf5",  grids, data, names, m_domainBox, m_dx,
		m_dt, m_time, ratios, 1, true, cover );
   exit(2);
   
}


//-----------------------------------------------------------------------
Real EBAMRNS::norm(const LevelData<EBCellFAB>& a_rhs, int a_ord )
{
  Vector<Real> nrm(a_rhs.nComp());
  Real volume = 0;
  for( int c = 0 ; c < a_rhs.nComp() ; c++ )
  {
     nrm[c] = 0;
     for( DataIterator d=a_rhs.getBoxes().dataIterator() ; d.ok() ; ++d )
        for( BoxIterator b(a_rhs[d()].box()) ; b.ok() ; ++b )
           if( !a_rhs[d()].getEBISBox().isCovered(b()) )
           {
	      Vector<VolIndex> vofs=a_rhs[d()].getEBISBox().getVoFs(b());
              for( int vi=0 ; vi < vofs.size() ; vi++ )
              {
		 if( c == 0 )
		     volume += a_rhs[d()].getEBISBox().volFrac(vofs[vi]);
		 if( a_ord == 0 )
		 {
		   if( nrm[c] < fabs(a_rhs[d()](vofs[vi],c)) )
		       nrm[c] = fabs(a_rhs[d()](vofs[vi],c));
		 }
		 else
	            nrm[c] += pow(a_rhs[d()](vofs[vi],c),a_ord)*a_rhs[d()].getEBISBox().volFrac(vofs[vi]);
	      }
	   }
  }
  if( volume > 0 && a_ord != 0 )
    for( int c = 0 ; c < a_rhs.nComp() ; c++ )
      nrm[c] /= volume;

  int norm_alternative = 0;
  double retval = 0;
  if( norm_alternative == 1 )
  {
     if( a_ord > 0 )
     {
        for( int c = 0 ; c < a_rhs.nComp() ; c++ )
           nrm[c] = pow(nrm[c],1.0/a_ord);
     }
     for( int c = 0 ; c < a_rhs.nComp() ; c++ )
     {
        if( retval < nrm[c] )
           retval = nrm[c];
     }
  }
  else
  {
     if( a_ord > 0 )
     {
        for( int c = 0 ; c < a_rhs.nComp() ; c++ )
	  retval += nrm[c];
        retval = pow(retval,1.0/a_ord);
     }
     else
        for( int c = 0 ; c < a_rhs.nComp() ; c++ )
	   if( retval < nrm[c] )
	      retval = nrm[c];
  }
  return retval;
}
