/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _EBAMRNSVISCOP_H_
#define _EBAMRNSVISCOP_H_

#include "REAL.H"
#include "Box.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "RefCountedPtr.H"

#include "AMRMultiGrid.H"

#include "EBIndexSpace.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"

#include "EBLevelDataOps.H"
#include "DirichletNSviscEBBC.H"
//#include "BaseEBBC.H"
//#include "BaseDomainBC.H"
#include "DirichletNSviscDomainBC.H"
#include "CFIVS.H"
#include "EBFluxRegister.H"
#include "EBMGAverage.H"
#include "EBMGInterp.H"

/// Navier-Stokes viscosity operator.
/**
   This class represents the operator
  \f[
      L(u) = \nabla( \lambda{\rm div}u) + 2\mu{\rm div}(D)
  \f]
   on one grid level. \f$ D \f$ is the symmetric part of the velocity gradient,
  \f[
     D=\frac{1}{2}(\nabla u + \nabla u^T)
  \f]
   Note that the operator is vector valued with two components in two 
   space dimensions and three components in three space dimensions.

   This class represents the operator on one grid level. If there are finer and/or 
   coarser grid levels present in the hierarchy, the operator can be evaluated with
   interpolation from finer and coarser levels.
   With fine and coarse level, the evaluation of the operator L(u), proceeds as
   follows:
   - If coarser levels are present, interpolate cell averages from coarse level
     to the ghost boundary points at the present level.
   - Exchange data with adjacent grids on the present level.
   - Compute fluxes on the present level.
   - If finer levels are present, evaluate fluxes on the fine grid, and interpolate
     these to the present level fluxes on the boundaries to the fine grids.

   By the above, the operator is a function of data on three levels.
   Symbolically we write this as L(phiC,phi,phiF). When no coarser level exists,
   we write L(phi,phiF). etc...
*/
class EBAMRNSviscOp: public AMRLevelOp<LevelData<EBCellFAB> >
{
public:
/// Destructor
  ~EBAMRNSviscOp();

/// Constructor
  EBAMRNSviscOp();

/// a_residual = a_rhs - L(a_phiFine, a_phi)   no coaser AMR level*/
  void AMRResidualNC(LevelData<EBCellFAB>&       a_residual,
                     const LevelData<EBCellFAB>& a_phiFine,
                     const LevelData<EBCellFAB>& a_phi,
                     const LevelData<EBCellFAB>& a_rhs,
                     bool a_homogeneousBC);

  /**
   */
  EBAMRNSviscOp( const DisjointBoxLayout&            a_dblFine,
                 const DisjointBoxLayout&            a_dbl,
                 const DisjointBoxLayout&            a_dblCoar,
                 const EBISLayout&                   a_ebislFine,
                 const EBISLayout&                   a_ebisl,
                 const EBISLayout&                   a_ebislCoar,
                 const RefCountedPtr<DirichletNSviscDomainBC>&  a_domainBCFine,
                 const RefCountedPtr<DirichletNSviscDomainBC>&  a_domainBC,
                 const RefCountedPtr<DirichletNSviscDomainBC>&  a_domainBCCoar,
                 const RefCountedPtr<DirichletNSviscEBBC>&      a_ebBCFine,
                 const RefCountedPtr<DirichletNSviscEBBC>&      a_ebBC,
                 const RefCountedPtr<DirichletNSviscEBBC>&      a_ebBCCoar,
                 const RealVect&                     a_dxFine,
                 const RealVect&                     a_dx,
                 const RealVect&                     a_dxCoar,
                 const ProblemDomain&                a_domain,
                 const RealVect&                     a_origin,
                 const int&                          a_refToFine,
                 const int&                          a_refToCoar,
                 const bool&                         a_hasFiner,
                 const bool&                         a_hasCoarser,
                 const int&                          a_relaxType,
                 const int&                          a_orderEB,
 	         const Real                          a_c,
	         const Real                          a_mu,
                 const Real                          a_lambda,
	         const LevelData<EBCellFAB>&         a_rho,
                 const bool&                         a_rz,
                 const Real                          a_time );

  //MGOp operations.  no finer or coarser

  ///
  /**
   */
  virtual void residual(LevelData<EBCellFAB>&       a_residual,
                        const LevelData<EBCellFAB>& a_phi,
                        const LevelData<EBCellFAB>& a_rhs,
                        bool                        a_homogeneous=false);

  ///
  /**
   */
  virtual void preCond(LevelData<EBCellFAB>&       a_opPhi,
                       const LevelData<EBCellFAB>& a_phi);

  ///
  /**
     This function assumes that coarse-fine boundary condtions have
     been dealt with.
   */
  virtual void applyOp(LevelData<EBCellFAB>&             a_opPhi,
                       const LevelData<EBCellFAB>&       a_phi,
                       const LevelData<EBCellFAB>* const a_phiCoarse,
                       bool                              a_homogeneousPhysBC,
                       bool                              a_homogeneousCFBC);

  ///
  /**
     get EB flux from leveldata if we have one.  otherwise use m_ebBC
     ignore input data in the case of homogeneous Phys BC
  */
  virtual void
  applyOp(LevelData<EBCellFAB>&                    a_opPhi,
          const LevelData<EBCellFAB>&              a_phi,
          const LevelData<EBCellFAB>* const        a_phiCoar,
          bool                                     a_homogeneousPhysBC,
          bool                                     a_homogeneousCFBC,
          const LevelData<BaseIVFAB<Real> >* const a_ebFluxBCLD);

  ///
  /**
     this is the linearop function.  CFBC is set to homogeneous.  phic is null
   */
  virtual void applyOp(LevelData<EBCellFAB>&             a_opPhi,
                       const LevelData<EBCellFAB>&       a_phi,
                       bool                              a_homogeneousPhysBC);

  ///
  /**
   */
  virtual void create(LevelData<EBCellFAB>&       a_lhs,
                      const LevelData<EBCellFAB>& a_rhs);

  ///
  virtual void createCoarsened(LevelData<EBCellFAB>&       a_lhs,
                               const LevelData<EBCellFAB>& a_rhs,
                               const int&                  a_refRat);

  Real
  AMRNorm(const LevelData<EBCellFAB>& a_coarResid,
          const LevelData<EBCellFAB>& a_fineResid,
          const int& a_refRat,
          const int& a_ord);
  ///
  /**
   */
  virtual void assign(LevelData<EBCellFAB>&       a_lhs,
                      const LevelData<EBCellFAB>& a_rhs);

  ///
  /**
   */
  virtual Real dotProduct(const LevelData<EBCellFAB>& a_1,
                          const LevelData<EBCellFAB>& a_2);

  ///
  /**
   */
  //  a_lhs := a_lhs + a_x*a_scale
  virtual void incr(LevelData<EBCellFAB>&       a_lhs,
                    const LevelData<EBCellFAB>& a_x,
                    Real                        a_scale);

  ///
  /**
   */
  virtual void axby(LevelData<EBCellFAB>&       a_lhs,
                    const LevelData<EBCellFAB>& a_x,
                    const LevelData<EBCellFAB>& a_y,
                    Real                        a_a,
                    Real                        a_b);

  ///
  /**
   */
  virtual void scale(LevelData<EBCellFAB>& a_lhs,
                     const Real&           a_scale);

  ///
  /**
   */
  virtual Real norm(const LevelData<EBCellFAB>& a_rhs,
                    int                         a_ord);

  ///
  /**
   */
  virtual void setToZero(LevelData<EBCellFAB>& a_lhs);

  ///
  /**
   */
  virtual void createCoarser(LevelData<EBCellFAB>&       a_coarse,
                             const LevelData<EBCellFAB>& a_fine,
                             bool                        a_ghosted);

  ///
  /**
   */
  virtual void relax(LevelData<EBCellFAB>&       a_e,
                     const LevelData<EBCellFAB>& a_residual,
                     int                         a_iterations);

  ///
  /**
     Calculate restricted residual:
       a_resCoarse[2h] = I[h->2h] (a_rhsFine[h] - L[h](a_phiFine[h]))
  */
  virtual void restrictResidual(LevelData<EBCellFAB>&       a_resCoarse,
                                LevelData<EBCellFAB>&       a_phiFine,
                                const LevelData<EBCellFAB>& a_rhsFine);

  ///
  /**
     Correct the fine solution based on coarse correction:
       a_phiThisLevel += I[2h->h] (a_correctCoarse)
  */
  virtual void prolongIncrement(LevelData<EBCellFAB>&       a_phiThisLevel,
                                const LevelData<EBCellFAB>& a_correctCoarse);

  ///
  /** Refinement ratio between this level and coarser level.
      Returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToCoarser();

  ///
  /** Refinement ratio between this level and coarser level.
      Returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToFiner();

  ///
  /** a_residual = a_rhs - L(a_phi, a_phiFine, a_phiCoarse) */
  virtual void AMRResidual(LevelData<EBCellFAB>& a_residual,
                           const LevelData<EBCellFAB>& a_phiFine,
                           const LevelData<EBCellFAB>& a_phi,
                           const LevelData<EBCellFAB>& a_phiCoarse,
                           const LevelData<EBCellFAB>& a_rhs,
                           bool a_homogeneousBC) ;

  ///
  /** a_residual = a_rhs - L(a_phi, a_phiCoarse)  */
  virtual void AMRResidualNF(LevelData<EBCellFAB>& a_residual,
                             const LevelData<EBCellFAB>& a_phi,
                             const LevelData<EBCellFAB>& a_phiCoarse,
                             const LevelData<EBCellFAB>& a_rhs,
                             bool a_homogeneousBC);

  ///
  /** a_resCoarse = I[h-2h] (a_residual - L(a_correction, a_coarseCorrection)) */
  virtual void AMRRestrict(LevelData<EBCellFAB>& a_resCoarse,
                           const LevelData<EBCellFAB>& a_residual,
                           const LevelData<EBCellFAB>& a_correction,
                           const LevelData<EBCellFAB>& a_coarseCorrection);

  ///
  /** a_correction += I[2h->h](a_coarseCorrection) */
  virtual void AMRProlong(LevelData<EBCellFAB>&       a_correction,
                          const LevelData<EBCellFAB>& a_coarseCorrection);

  ///
  /** a_residual = a_residual - L(a_correction, a_coarseCorrection) */
  virtual void AMRUpdateResidual(LevelData<EBCellFAB>&       a_residual,
                                 const LevelData<EBCellFAB>& a_correction,
                                 const LevelData<EBCellFAB>& a_coarseCorrection);

  void reflux(LevelData<EBCellFAB>& a_residual,
              const LevelData<EBCellFAB>& a_phiFine,
              const LevelData<EBCellFAB>& a_phi,
              int a_energy_eq );

 void get_energyeq_viscosity( LevelData<EBCellFAB>& a_ev,
			     const LevelData<EBCellFAB>& a_u,
        const LevelData<EBCellFAB>* const        a_uCoar,
        bool                                     a_homogeneousPhysBC,
        bool                                     a_homogeneousCFBC,
			     const LevelData<BaseIVFAB<Real> >* const a_ebFluxBCLD);

void AMRResidual_eeq(LevelData<EBCellFAB>&       a_residual,
            const LevelData<EBCellFAB>* a_phiFine,
            const LevelData<EBCellFAB>& a_phi,
            const LevelData<EBCellFAB>* a_phiCoar,
		     bool a_homogeneousPhysBC);

void get_uweight( const DataIndex dex, LevelData<EBCellFAB>& wgh );
void set_homogeneous_relax();
void set_inhomogeneous_relax();
void set_rho_zero();
void set_coeff( Real a_c );
void set_rho( const LevelData<EBCellFAB>& a_rho );
void set_time( Real a_time );

protected:


  DisjointBoxLayout               m_dblFine;
  DisjointBoxLayout               m_dbl;
  DisjointBoxLayout               m_dblCoar;
  DisjointBoxLayout               m_dblCoarsenedFine;

  EBISLayout                      m_ebislFine;
  EBISLayout                      m_ebisl;
  EBISLayout                      m_ebislCoar;
  EBISLayout                      m_ebislCoarsenedFine;

  ProblemDomain                   m_domainFine;
  ProblemDomain                   m_domain;
  ProblemDomain                   m_domainCoar;
  RefCountedPtr<DirichletNSviscDomainBC>     m_domainBCFine;
  RefCountedPtr<DirichletNSviscDomainBC>     m_domainBC;
  RefCountedPtr<DirichletNSviscDomainBC>     m_domainBCCoar;
  RefCountedPtr<DirichletNSviscEBBC>         m_ebBCFine;
  RefCountedPtr<DirichletNSviscEBBC>         m_ebBC;
  RefCountedPtr<DirichletNSviscEBBC>         m_ebBCCoar;

  LayoutData<IntVectSet>  m_irreg;
  LayoutData<IntVectSet>  m_irreg_w_ghost;

  RealVect                        m_dxFine;
  RealVect                        m_dx;
  RealVect                        m_invDx;
  RealVect                        m_invDx2;
  Real                            m_dxScale;

  RealVect                        m_dxCoar;
  RealVect                        m_origin;
  int                             m_refToFine;
  int                             m_refToCoar;
  bool                            m_hasFine;
  bool                            m_hasInterpAve;
  bool                            m_hasCoar;
  int                             m_orderEB;
  int                             m_relaxType;
  bool                            m_rz;
  bool                            m_homogeneous_relax;

  //restriction object
  EBMGAverage                    m_ebAverage;
  //prolongation object
  EBMGInterp                     m_ebInterp;

  // Coarse-fine stencils for homogeneous CFInterp
  LayoutData<CFIVS> m_loCFIVS[SpaceDim];
  LayoutData<CFIVS> m_hiCFIVS[SpaceDim];

// Diagonal weight in total discretization stencil, 
// needed for relaxation schemes.
  LayoutData<BaseIVFAB<Real> > m_weight;
// complete weight including boundary terms.
  LevelData<EBCellFAB> m_wgh;

// stencil for u (diagonal term).
  LayoutData<BaseIVFAB<VoFStencil> > m_uStencil;
  LayoutData<BaseIVFAB<Real> >   m_uweight;

  //stencils for flux evaluations
  LayoutData<BaseIFFAB<FaceStencil> > m_iflux_stencil;
  LayoutData<BaseIFFAB<FaceStencil> > m_jflux_stencil;
  LayoutData<BaseIFFAB<FaceStencil> > m_kflux_stencil;

 // stencils for flux-tangential velocities
  LayoutData<BaseIVFAB<VoFStencil> >  m_uyz_stencil;
  LayoutData<BaseIVFAB<VoFStencil> >  m_uxz_stencil;
  LayoutData<BaseIVFAB<VoFStencil> >  m_uxy_stencil;

// stencil for identity operator. L(u) = u
  bool m_split_u;
  Real m_time;

// Represents operator  L(u)-(rho/c)*u, where L(u) is viscos part of NS momentum eq.

// Viscosity coefficients
  LevelData<EBCellFAB> m_mu;
  LevelData<EBCellFAB> m_lambda;

// multiply op by this 'c' above, typcial value would be dt/2.
  Real m_c;

// Density field
  LevelData<EBCellFAB> m_rho;

  //flux register with finer level
  EBFluxRegister m_fluxReg;

  //special mg objects for when we do not have
  //a coarser level or when the refinement to coarse
  //is greater than two
  //flag for when we need special MG objects
  bool                        m_hasMGObjects;
  //stuff below is only defined if m_hasMGObjects==true
  EBMGAverage                 m_ebAverageMG;
  EBMGInterp                  m_ebInterpMG;
  DisjointBoxLayout           m_dblCoarMG;
  EBISLayout                  m_ebislCoarMG;
  Box                         m_domainCoarMG;

private:
  //internally useful but not for general consumption
  //lots of hidden assumptions and all that

  void incrementFRCoar(EBFluxRegister&             a_fluxReg,
                       const LevelData<EBCellFAB>& a_phiFine,
                       const LevelData<EBCellFAB>& a_phi,
                       int a_energy_eq );

  void incrementFRFine(EBFluxRegister&             a_fluxReg,
                       const LevelData<EBCellFAB>& a_phiFine,
                       const LevelData<EBCellFAB>& a_phi,
                       int a_energy_eq );

void incrementCoarseBothRZ(
				    EBFluxRegister& a_fluxReg,
				    const EBFaceFAB& a_coarFlux,
                                    const Real&      a_scale,
                                    const DataIndex& a_coarDatInd,
                                    const Interval&  a_variables,
                                    const int&       a_dir,
				    const Real&      a_dx);

 void incrementFineBothRZ(
				  EBFluxRegister& a_fluxReg,
				  const EBFaceFAB&      a_fineFlux,
                                  const Real&           a_scale,
                                  const DataIndex&      a_fineDatInd,
                                  const Interval&       a_variables,
                                  const int&            a_dir,
                                  const Side::LoHiSide& a_sd,
                                  const Real&           a_dx );


  void getFlux(EBFaceFAB&       a_flux,
               const EBCellFAB& a_phi,
               const Box&       a_box,
               const EBISBox&   a_ebisBox,
               const RealVect&  a_dx,
               const int&       a_idir,
	       const Real a_mu,
	       const Real a_lambda,
               int a_energy_eq );

void getFlux_rz(EBFaceFAB&       a_flux,
			    const EBCellFAB& a_phi,
			    const Box&       a_box,
			    const EBISBox&   a_ebisBox,
			    const RealVect&  a_dx,
			    const int&       a_idir,
			    const Real       a_mu,
			    const Real       a_lambda,
			       int              a_energy_eq );

  void applyCFBCs(LevelData<EBCellFAB>&             a_phi,
                  const LevelData<EBCellFAB>* const a_phiCoarse,
                  bool a_homogeneousCFBC);

  void define_flux_stencils( );

  void define_flux_stencil( int dir, BaseIFFAB<FaceStencil>& flux_stencil,
			    BaseIVFAB<VoFStencil>& utan_stencil,
			    const EBISBox& eb, const Box& ubox, const IntVectSet& irr );

  void tangential_derivative( IntVect of, const VolIndex& vof,
			      const EBISBox& eb, VoFStencil& vs,
			      const Real di, const Box );

  void define_ustencil( );

  void get_ustencil( VoFStencil&     a_stencil,
                     Real&           a_wgh,
                     const EBISBox&  a_ebisbox,
                     const VolIndex& a_VoF );


  void get_ns_uweight( BaseIVFAB<Real>& weight,
				     const EBISBox& eb,
				     const BaseIFFAB<FaceStencil>& flux_stencil, 
 	                             const EBCellFAB& a_mu,
				     const EBCellFAB& a_lambda,
				     int dir );

void get_ns_uweight_rz( BaseIVFAB<Real>& weight,
				     const EBISBox& eb,
				     const BaseIFFAB<FaceStencil>& flux_stencil, 
 	                             const EBCellFAB& a_mu,
				     const EBCellFAB& a_lambda,
			             int dir,
			             const IntVectSet& notRegular );

  void update_irregular( EBCellFAB& a_res, const EBCellFAB& a_u,
			 const EBISBox& eb, const BaseIFFAB<FaceStencil>& flux_stencil,
			 const BaseIVFAB<VoFStencil>& utan_stencil,
                         const EBCellFAB& a_mu, const EBCellFAB& a_lambda, int dir,
			 const IntVectSet& irr_w_ghost, const IntVectSet& irr );

void update_irregular_rz( EBCellFAB& a_res, const EBCellFAB& a_u, const EBISBox& eb, const BaseIFFAB<FaceStencil>& flux_stencil, const BaseIVFAB<VoFStencil>& utan_stencil,
			  const EBCellFAB& a_mu, const EBCellFAB& a_lambda, int dir,
			  const IntVectSet& irr_w_ghost, const IntVectSet& irr );
  
void update_irregular_eeq( EBCellFAB& a_res, const EBCellFAB& a_u, const EBISBox& eb, const BaseIFFAB<FaceStencil>& flux_stencil, const BaseIVFAB<VoFStencil>& utan_stencil,
			   const EBCellFAB& a_mu, const EBCellFAB& a_lambda, int dir );
void update_irregular_eeq_rz(
			      EBCellFAB& a_res,
			      const EBCellFAB& a_u,
			      const EBISBox& eb,
			      const BaseIFFAB<FaceStencil>& flux_stencil,
			      const BaseIVFAB<VoFStencil>& utan_stencil,
	                      const EBCellFAB& a_mu,
			      const EBCellFAB& a_lambda,
			      int dir,
			      const IntVectSet& irreg );
  
  void levelMulticolorGS( LevelData<EBCellFAB>&       a_phi,
                          const LevelData<EBCellFAB>& a_rhs);

  void levelJacobi( LevelData<EBCellFAB>&       a_phi,
                    const LevelData<EBCellFAB>& a_rhs );

  void applyHomogeneousCFBCs(LevelData<EBCellFAB>&   a_phi);

  void applyHomogeneousCFBCs(EBCellFAB&            a_phi,
                             const DataIndex&      a_datInd,
                             int                   a_idir,
                             Side::LoHiSide        a_hiorlo);

void dirichlet_tangential( RealVect nv, RealVect tv, RealVect bv, 
                           const VolIndex &a_vof, int side,
                           const EBCellFAB& la, const EBCellFAB& mu,
                           BaseFab<Real>& tterm, const Real faceAreaFrac,
			   RealVect dx, RealVect x0, bool homogeneousPhysBC );

void dirichlet_tangential( RealVect nv, RealVect tv, RealVect bv,
                                          const VolIndex& vof, const EBCellFAB& la,
                                          const EBCellFAB& mu, const EBISBox& ebis_box,
                                          RealVect tt, RealVect x0, RealVect dx,
			   bool homogeneousPhysBC );

  void get_tanvects( RealVect nv, RealVect& tv, RealVect& bv );
 
  void corner_cfbcs( LevelData<EBCellFAB>& a_phi, const EBISLayout& ebs );
  void get_irregular();

private:
  //copy constructor and operator= disallowed for all the usual reasons
  EBAMRNSviscOp(const EBAMRNSviscOp& a_opin)
  {
    MayDay::Error("invalid operator");
  }

  void operator=(const EBAMRNSviscOp& a_opin)
  {
    MayDay::Error("invalid operator");
  }
};


//-----------------------------------------------------------------------
class EBAMRNSviscOpFactory: public AMRLevelOpFactory<LevelData<EBCellFAB> >
{
public:
  ///
  virtual ~EBAMRNSviscOpFactory();

  ///
  EBAMRNSviscOpFactory(const Vector<DisjointBoxLayout>&   a_grids,
                        const Vector<EBISLayout>&          a_ebisls,
                        const Vector<int>&                 a_refRatio,
                        const ProblemDomain&               a_domainCoarse,
                        const RealVect&                    a_dxCoarse,
                        const RealVect&                    a_origin,
                        const int&                         a_orderEB,
                        const int&                         a_relaxType,
                        RefCountedPtr<DirichletNSviscDomainBCFactory> a_domainBCFactory,
		        RefCountedPtr<DirichletNSviscEBBCFactory>     a_ebBcFactory,
	                Vector<Real>                        a_c,
	                Real                                a_mu,
                        Real                                a_lambda,
                        const Vector<LevelData<EBCellFAB>*>& a_rho,
                        const bool&                         a_rz );

  ///
  virtual EBAMRNSviscOp* MGnewOp(const ProblemDomain& a_FineindexSpace,
				 int                  a_depth,
				 bool                 a_homoOnly = true);


  EBAMRNSviscOp* createOperator(const DisjointBoxLayout& a_dblMGLevel,
                                 const EBISLayout&        a_ebislMGLevel,
                                 const ProblemDomain&     a_domainMGlevel,
                                 const RealVect&          a_dxMGLevel,
                                 const int&               a_whichLevel);

  ///
  virtual void reclaim(MGLevelOp<LevelData<EBCellFAB> >* a_reclaim);

  ///
  virtual EBAMRNSviscOp* AMRnewOp(const ProblemDomain& a_FineindexSpace);

  ///
  virtual void AMRreclaim(EBAMRNSviscOp* a_reclaim);

  void set_coeff( Real a_cof, int a_level );

  void set_rho( const LevelData<EBCellFAB>& a_rho, int a_level );

  void set_time( const Real a_time );

  ///
  /** Refinement ratio between this level and coarser level.
      Returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToFiner(const ProblemDomain& a_domain) const;

protected:
  bool     m_isDefined;
  int      m_orderEB;
  int      m_relaxType;
  int      m_numLevels;

  Vector<ProblemDomain>     m_domainVec;
  Vector<DisjointBoxLayout> m_gridVec;
  Vector<EBISLayout>        m_ebislVec;
  Vector<int>               m_refRatioVec;
  Vector<RealVect>          m_dxVec;
  RealVect                  m_origin;

  RefCountedPtr<DirichletNSviscDomainBCFactory>   m_domainBCFactory;
  RefCountedPtr<DirichletNSviscEBBCFactory>       m_ebBCFactory;

  Vector<Real> m_c;
  Real m_mu;
  Real m_lambda;
  bool m_rz;
  Real m_time;

  Vector<LevelData<EBCellFAB>*> m_rhoPtr;

  void interp_frfine( LevelData<EBCellFAB>& a_uc,
		      const LevelData<EBCellFAB>& a_uf,
		      const Box& a_domainc,
		      const EBISLayout& a_ec,
		      const EBISLayout& a_ef,
		      const int a_rat );

private:
  ///weak construction bad
  EBAMRNSviscOpFactory()
  {
    MayDay::Error("invalid operator");
  }

  //copy constructor and operator= disallowed for all the usual reasons
  EBAMRNSviscOpFactory(const EBAMRNSviscOpFactory& a_opin)
  {
    MayDay::Error("invalid operator");
  }

  void operator=(const EBAMRNSviscOpFactory& a_opin)
  {
    MayDay::Error("invalid operator");
  }
};

  /*
    coarsen the dbl if we can
    otherwise,
    check to see if the domain is covered. if so
    make one big grid.  otherwise we fail
    returns false if not coarsenable
  */
extern bool
getCoarserLayouts(DisjointBoxLayout&       a_dblCoar,
                  EBISLayout&              a_ebislCoar,
                  Box&                     a_domainCoar,
                  const DisjointBoxLayout& a_dblFine,
                  const Box&               a_domainFine,
                  int                      a_refToCoar,
                  const EBIndexSpace*      ebisPtr);
#endif
