/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// dtg 1-23-2002

#ifndef _EBLEVELGODUNOVNS_H_
#define _EBLEVELGODUNOVNS_H_

#include "EBCellFAB.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "PiecewiseLinearFillPatch.H"
#include "EBPWLFillPatch.H"
#include "EBFluxRegister.H"
#include "ProblemDomain.H"
#include "EBPatchGodunovFactory.H"
#include "EBLevelRedist.H"

/// Level Godunov
/**
 */
class EBLevelGodunovNS
{
public:
  /// Default constructor
  /**
     Object requires define(..) to be called before all other functions.
  */
  EBLevelGodunovNS();

  /// Destructor
  /**
     destroys all objects created by define(..). Passed in
     data references of define(..) are left alone.
  */
  ~EBLevelGodunovNS();

  /// Actual constructor.
  /**
     For the coarsest level, an empty DisjointBoxLayout
     is passed in for coaserDisjointBoxLayout.
     Inside the routine, we cast away const-ness on the data members
     for the assignment.  The arguments passed in are maintained const.
     (coding standards).  a_nRefine is the refinement ratio between
     this level and the next coarser level.  a_numGhosts is the number
     of ghost cells in each direction.
  */
  void define(const DisjointBoxLayout&           a_thisDBL,
              const DisjointBoxLayout&           a_coarDBL,
              const EBISLayout&                  a_thisEBISL,
              const EBISLayout&                  a_coarEBISL,
              const ProblemDomain&               a_DProblem,
              const int&                         a_nRefine,
              const Real&                        a_dx,
              const bool&                        a_useMassRedist,
              const bool&                        a_doSmushing,
              const bool&                        a_doRZCoords,
              const bool&                        a_hasSourceTerm,
              const EBPatchGodunovFactory* const a_patchGodunov,
              const bool&                        a_hasCoarser,
              const bool&                        a_hasFiner);

  /// Take one timestep for this grid level.
  /**
     For the coarsest level, UcoarseOld and UCoarseNew are empty
     LevelData<EBCellFAB> objects.  Also, either UCoarseOld or
     UCoarseNew might be empty to indicate that t(nf) = t(nc) the
     one grid is at the current time and interpolation is not required
     for boundary condition generation.  U must be defined on the same
     DisjointBoxLayouts as were used in define(...)
     Coarse flux register is flux register with the coarse level.
     Fine flux register is the flux register with the fine level.
     To the finer level FR, this level is the coarse level.
     To the coarser level FR, this level is the fine level.
  */
  Real step(LevelData<EBCellFAB>&        a_consState,
            LevelData<BaseIVFAB<Real> >& a_massDiff,
            EBFluxRegister&              a_fineFluxRegister,
            EBFluxRegister&              a_coarFluxRegister,
            const LevelData<EBCellFAB>&  a_consStateCoarseOld,
            const LevelData<EBCellFAB>&  a_consStateCoarseNew,
            const Real&                  a_time,
            const Real&                  a_TCold,
            const Real&                  a_TCNew,
            const Real&                  a_dt,
            const LevelData<EBCellFAB>&  a_viscos );

  void
  getDrhoDtOverRho(LevelData<EBCellFAB>& a_drhoDt,
                   const LevelData<EBCellFAB>& a_rhoNew,
                   const LevelData<EBCellFAB>& a_rhoOld,
                   const Real& a_dt);

  /// Get maximum wave speed
  Real getMaxWaveSpeed(const LevelData<EBCellFAB>& a_state);

  bool isDefined() const;


protected:
  //these are not grown by one.
  LayoutData<IntVectSet> m_irregSetsSmall;

  //these are grown by one in the directions != idir
  LayoutData<IntVectSet> m_irregSetsGrown[SpaceDim];

  LevelData<BaseIFFAB<Real> > m_fluxInterpolants[SpaceDim];
  LevelData<BaseIVFAB<Real> > m_nonConsDivergence;
  LevelData<BaseIVFAB<Real> > m_ebIrregFaceFlux;
  LayoutData<IntVectSet>      m_cfIVS;
  bool              m_hasCoarser;
  bool              m_hasFiner;
  DisjointBoxLayout m_thisGrids;
  DisjointBoxLayout m_coarGrids;
  EBISLayout        m_thisEBISL;
  EBISLayout        m_coarEBISL;
  bool              m_isDefined;
  EBPatchGodunov*   m_ebPatchGodunov;
  Real              m_dx;
  ProblemDomain     m_domain;
  int               m_refRatCrse;
  int               m_nCons;
  int               m_nFlux;
  int               m_nGhost;
  EBPWLFillPatch    m_patcher;
  bool              m_useMassRedist;
  EBLevelRedist     m_ebLevelRedist;
  static int        s_timestep;
  bool m_doSmushing;
  bool m_doRZCoords;
  bool m_hasSourceTerm;
private:
  //disallowed for all the usual reasons
  void operator=(const EBLevelGodunovNS& a_input)
  {
    MayDay::Error("invalid operator");
  }
  EBLevelGodunovNS(const EBLevelGodunovNS& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
