/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "DebugOut.H"
#include "EBLevelGodunovNS.H"
#include "BaseIVFactory.H"
#include "BaseIFFactory.H"
#include "BaseIFFAB.H"
#include "EBFluxFAB.H"
#include "FaceIterator.H"
#include "REAL.H"
#include "EBCellFactory.H"
#include "FaceIterator.H"
#include <cstdio>
#include "EBAMRIO.H"
int EBLevelGodunovNS::s_timestep = 0;
/*****************************/
/*****************************/
EBLevelGodunovNS::EBLevelGodunovNS()
{
  m_isDefined = false;
  m_ebPatchGodunov = NULL;
}
/*****************************/
/*****************************/
bool
EBLevelGodunovNS::isDefined() const
{
  return m_isDefined;
}
/*****************************/
/*****************************/
EBLevelGodunovNS::~EBLevelGodunovNS()
{
  if(m_ebPatchGodunov != NULL)
    delete m_ebPatchGodunov;
}
/*****************************/
/*****************************/
void
EBLevelGodunovNS::define(const DisjointBoxLayout&      a_thisDBL,
                       const DisjointBoxLayout&      a_coarDBL,
                       const EBISLayout&             a_thisEBISL,
                       const EBISLayout&             a_coarEBISL,
                       const ProblemDomain&          a_domain,
                       const int&                    a_nRefine,
                       const Real&                   a_dx,
                       const bool&                   a_useMassRedist,
                       const bool&                   a_doSmushing,
                       const bool&                   a_doRZCoords,
                       const bool&                   a_hasSourceTerm,
                       const EBPatchGodunovFactory*  const a_patchGodunov,
                       const bool&                   a_hasCoarser,
                       const bool&                   a_hasFiner)
{
  CH_assert(a_dx > 0.0);
  CH_assert(a_nRefine > 0);

  m_isDefined = true;
  m_useMassRedist = a_useMassRedist;
  m_thisGrids = a_thisDBL;
  m_thisEBISL = a_thisEBISL;
  m_refRatCrse= a_nRefine;
  m_dx = a_dx;
  m_domain = a_domain;
  m_hasCoarser= a_hasCoarser;
  m_hasFiner= a_hasFiner;
  m_doSmushing = a_doSmushing;
  m_doRZCoords = a_doRZCoords;
  m_hasSourceTerm = a_hasSourceTerm;
  if(m_doRZCoords && !m_hasSourceTerm)
    {
      MayDay::Error("LG: RZ implies need of a source term--inconsistent inputs");
    }
  if(m_hasCoarser)
    {
      m_coarGrids = a_coarDBL;
      m_coarEBISL = a_coarEBISL;
    }

  if(m_ebPatchGodunov != NULL)
    delete m_ebPatchGodunov;

  m_ebPatchGodunov = a_patchGodunov->create();
  m_ebPatchGodunov->define(m_domain, m_dx);

  // Determing the number of ghost cells necessary here
  m_nGhost = 5;
  m_nCons = m_ebPatchGodunov->numConserved();
  m_nFlux = m_ebPatchGodunov->numFluxes();
  const Box& domainBox  = m_domain.domainBox();

  //define redistribution object for this level
  //for now set to volume weighting
  m_ebLevelRedist.define(m_thisGrids, m_thisEBISL,
                         domainBox, m_nCons);

  if(m_hasCoarser)
    {
      ProblemDomain domainCrse = coarsen(m_domain, m_refRatCrse);
      Box domainBoxCrse = domainCrse.domainBox();

      //patcher is defined with the number of conserved vars.
      m_patcher.define(m_thisGrids, m_coarGrids,
                       m_thisEBISL, m_coarEBISL,
                       domainBoxCrse, m_refRatCrse, m_nCons,
                       m_nGhost);
    }

  m_irregSetsSmall.define(m_thisGrids);
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      m_irregSetsGrown[faceDir].define(m_thisGrids);
    }
  for(DataIterator dit = m_thisGrids.dataIterator();
      dit.ok(); ++dit)
    {
      const EBISBox& ebisBox = m_thisEBISL[dit()];
      if(!ebisBox.isAllCovered())
        {
          const Box&     thisBox = m_thisGrids.get(dit());
          IntVectSet ivsIrreg = ebisBox.getIrregIVS(thisBox);
          m_irregSetsSmall[dit()] = ivsIrreg;

          for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
            {
              IntVectSet& grownIVS = m_irregSetsGrown[faceDir][dit()];
              //need to do this in case there is an irregular cells
              //just outside the box.
              Box grownBox2 = grow(thisBox, 2);
              grownBox2 &= domainBox;
              IntVectSet irregivs = ebisBox.getIrregIVS(grownBox2);
              grownIVS = irregivs;
              Box grownBox1 = thisBox;
              //grow in all directions != faceDir
              for(int jdir = 0; jdir < SpaceDim; jdir++)
                {
                  if (jdir != faceDir)
                    {
                      grownIVS.grow(jdir, 1);
                      grownBox1.grow(jdir, 1);
                    }
                }
              //restrict to domain
              grownBox1 &= domainBox;
              grownIVS &= grownBox1;
            }
        }
    }

  BaseIVFactory<Real> cellFactorySmall(m_thisEBISL, m_irregSetsSmall);
  m_nonConsDivergence.define(m_thisGrids, m_nCons,
                             IntVect::Zero, cellFactorySmall);
  m_ebIrregFaceFlux.define(m_thisGrids, m_nCons,
                           IntVect::Zero, cellFactorySmall);
  //define flux interpolant stuff
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      BaseIFFactory<Real> faceFactory(m_thisEBISL, m_irregSetsGrown[faceDir], faceDir);
      m_fluxInterpolants[faceDir].define(m_thisGrids, m_nFlux,
                                         3*IntVect::Unit, faceFactory);
    }
  //ebpatch needs coarse-fine IVS to know where to drop order for interpolation
  m_cfIVS.define(m_thisGrids);
  for(DataIterator ditOuter = m_thisGrids.dataIterator();
      ditOuter.ok(); ++ditOuter)
    {
      const EBISBox& ebisBox = m_thisEBISL[ditOuter()];
      if(!ebisBox.isAllCovered())
        {
          // make a grown box of the grid and then subtract off each grid in the
          //domain to make the coarse-fine IVS
          // CFIVS = grow(b, 1) - sum(bi)
          Box grownBox = grow(m_thisGrids.get(ditOuter()), 1);
          grownBox &= m_domain;
          IntVectSet complementIVS(grownBox);

          for(LayoutIterator litInner = m_thisGrids.layoutIterator();
              litInner.ok(); ++litInner)
            {
              Box innerBox = m_thisGrids.get(litInner());
              complementIVS -= innerBox;
            }

          m_cfIVS[ditOuter()] = complementIVS;
        }
    }
}
/*****************************/
/*****************************/
Real
EBLevelGodunovNS::step(LevelData<EBCellFAB>&         a_consState,
                     LevelData<BaseIVFAB<Real> >&  a_massDiff,
                     EBFluxRegister&               a_fineFluxRegister,
                     EBFluxRegister&               a_coarFluxRegister,
                     const LevelData<EBCellFAB>&   a_consStateCoarseOld,
                     const LevelData<EBCellFAB>&   a_consStateCoarseNew,
                     const Real&                   a_time,
                     const Real&                   a_coarTimeOld,
                     const Real&                   a_coarTimeNew,
                     const Real&                   a_dt,
                     const LevelData<EBCellFAB>&   a_viscsrc )
{
  CH_assert(isDefined());
  CH_assert(a_consState.disjointBoxLayout() == m_thisGrids);

  //create temp data with the correct number of ghost cells
  IntVect ivGhost = m_nGhost*IntVect::Unit;
  Interval consInterv(0, m_nCons-1);
  Interval fluxInterv(0, m_nFlux-1);

  EBCellFactory factory(m_thisEBISL);
  LevelData<EBCellFAB> consTemp(m_thisGrids, m_nCons, ivGhost, factory);
  for (DataIterator dit = m_thisGrids.dataIterator(); dit.ok(); ++dit)
    {
      consTemp[dit()].setVal(0.);
    }
  a_consState.copyTo(consInterv, consTemp, consInterv);
  m_ebLevelRedist.setToZero();
  // clear flux registers with fine level
  //remember this level is the coarse level to the fine FR
  if(m_hasFiner)
    {
      a_fineFluxRegister.setToZero();
    }

  // Fill ghost cells using fillInterp, and copyTo.
  if (m_hasCoarser)
    {
      m_patcher.interpolate(consTemp,
                            a_consStateCoarseOld,
                            a_consStateCoarseNew,
                            a_coarTimeOld,
                            a_coarTimeNew,
                            a_time,
                            consInterv);
    }

  // Exchange all the data between grids
  consTemp.exchange(consInterv);
  Box domainBox = m_domain.domainBox();

  //compute flattening coefficients. this saves a ghost cell. woo hoo.
  IntVect flatGhostIV = 3*IntVect::Unit;
  LevelData<EBCellFAB> flattening(m_thisGrids, 1, flatGhostIV, factory);
  int ibox = 0;
  for (DataIterator dit = m_thisGrids.dataIterator(); dit.ok(); ++dit)
    {
      EBCellFAB& consState = consTemp[dit()];
      const EBISBox& ebisBox = m_thisEBISL[dit()];
      if(!ebisBox.isAllCovered())
        {
          const Box& cellBox = m_thisGrids.get(dit());
          const IntVectSet& cfivs = m_cfIVS[dit()];
          m_ebPatchGodunov->setValidBox(cellBox, ebisBox, cfivs, a_time, a_dt);
          m_ebPatchGodunov->setCoveredConsVals(consState);
          int nPrim  = m_ebPatchGodunov->numPrimitives();
          EBCellFAB primState(ebisBox, consState.getRegion(), nPrim);
          int logflag = 0;
          bool verbose = false;
          //debug
          //verbose = true;
          //end debug
          m_ebPatchGodunov->consToPrim(primState, consState, consState.getRegion(), logflag, verbose);
          EBCellFAB& flatteningFAB = flattening[dit()];
          //this will set the stuff over the coarse-fine interface
          flatteningFAB.setVal(1.);
          if(m_ebPatchGodunov->usesFlattening())
            {
              m_ebPatchGodunov->computeFlattening(flatteningFAB,
                                                  primState,
                                                  cellBox);
            }
          ibox++;
        }
    }
  Interval zerointerv(0,0);
  flattening.exchange(zerointerv);

  //first pass to copy flux into interpolants
  ibox = 0;
  for (DataIterator dit = m_thisGrids.dataIterator(); dit.ok(); ++dit, ibox++)
    {
      const Box& cellBox = m_thisGrids.get(dit());
      const EBISBox& ebisBox = m_thisEBISL[dit()];
      if(!ebisBox.isAllCovered())
        {
          const IntVectSet& cfivs = m_cfIVS[dit()];

          EBCellFAB& consState = consTemp[dit()];
          m_ebPatchGodunov->setValidBox(cellBox, ebisBox, cfivs, a_time, a_dt);
          EBCellFAB source;
              //this is the real update  this is to do the actual
              //update and get the fluxes for the flux register
              //no source terms here so just send dummy FAB
          const Box& bigBox = consState.box();
          int nPrim = m_ebPatchGodunov->numPrimitives();
          source.define(ebisBox, bigBox, nPrim);
              //this setval is important
          source.setVal(0.);
          if(m_hasSourceTerm)
          {
              if(SpaceDim != 2)
                {
                  MayDay::Error("RZ only valid for SD==2");
                }
              m_ebPatchGodunov->setSource(source, consState, bigBox);
          }
          if( a_viscsrc[dit()].nComp() != source.nComp() )
	  {
	     EBCellFAB viscadd;
	     Box srcbox = a_viscsrc[dit()].box();
	     viscadd.define(ebisBox,bigBox,nPrim);
	     viscadd.setVal(0.0);
             Interval cons(0,SpaceDim-1);
             viscadd.copy( srcbox, cons, srcbox, a_viscsrc[dit()], cons );
             source += viscadd;
	  }
	  else
             source += a_viscsrc[dit()];

          EBFluxFAB flux(ebisBox, cellBox, m_nFlux);
          BaseIVFAB<Real>& nonConsDiv    = m_nonConsDivergence[dit()];
          BaseIVFAB<Real>& ebIrregFlux   = m_ebIrregFaceFlux[dit()];
          flux.setVal(7.89);
          ebIrregFlux.setVal(7.89);
          const IntVectSet& ivsIrreg     = m_irregSetsSmall[dit()];
          const EBCellFAB& flatteningFAB = flattening[dit()];
          bool verbose = false;
          //          IntVect ivdebug(D_DECL(72,11,23));
          //          if(cellBox.contains(ivdebug))
          //            {
          //              pout() << "ibox for ivdebug " << ivdebug  << " is = " << ibox << endl;
          //            }
          m_ebPatchGodunov->regularUpdate(consState, flux, ebIrregFlux,
                                          nonConsDiv,flatteningFAB,
                                          source, cellBox, ivsIrreg,
                                          verbose);

          //do fluxregister cha-cha
          /*
            Coarse flux register is flux register with the coarse level.
            Fine flux register is the flux register with the fine level.
            To the finer level FR, this level is the coarse level.
            To the coarser level FR, this level is the fine level.
          */
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              Real scale = a_dt;

              EBFaceFAB fluxRegFlux;
              if(m_doRZCoords)
                {
                  //gather fluxes into flux-register compatible form
                  fluxRegFlux.define(ebisBox, cellBox, idir, m_nCons);
                  m_ebPatchGodunov->assembleFluxReg(fluxRegFlux, flux[idir],
                                                    idir, cellBox);
                }
              if(m_hasFiner)
                {
                  if(m_doRZCoords)
                    {
                      a_fineFluxRegister.incrementCoarseRegulRZ(fluxRegFlux, scale, dit(),
                                                                consInterv, idir, m_dx);
                    }
                  else
                    {
                      a_fineFluxRegister.incrementCoarseRegular(flux[idir], scale,dit(),
                                                                consInterv, idir);
                    }
                }

              if(m_hasCoarser)
                {
                  for(SideIterator sit; sit.ok(); ++sit)
                    {
                      if(m_doRZCoords)
                        {

                          a_coarFluxRegister.incrementFineRegulRZ(fluxRegFlux,scale, dit(),
                                                                  consInterv, idir,sit(), m_dx);
                        }
                      else
                        {
                          a_coarFluxRegister.incrementFineRegular(flux[idir],scale, dit(),
                                                                  consInterv, idir,sit());
                        }
                    }
                }
            }

          //copy fluxes into sparse interpolant
          for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
            {
              IntVectSet ivsIrregGrown = m_irregSetsGrown[faceDir][dit()];
              ivsIrregGrown &= cellBox;
              FaceStop::WhichFaces stopCrit = FaceStop::SurroundingWithBoundary;

              BaseIFFAB<Real>& interpol = m_fluxInterpolants[faceDir][dit()];
              interpol.setVal(7.7777e7);
              EBFaceFAB& fluxDir = flux[faceDir];
              for(FaceIterator faceit(ivsIrregGrown, ebisBox.getEBGraph(),
                                      faceDir, stopCrit);
                  faceit.ok(); ++faceit)
                {
                  for(int ivar = 0; ivar < m_nFlux; ivar++)
                    {
                      interpol(faceit(), ivar) = fluxDir(faceit(), ivar);
                    }
                }

            }
        }
    }

  //exchange ghost cell data for flux interpolant
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      m_fluxInterpolants[faceDir].exchange(fluxInterv);
    }

  //now do the irregular update and the max wave speed
  Real maxWaveSpeed = 1.0e-12;
  ibox = 0;
  for (DataIterator dit = m_thisGrids.dataIterator(); dit.ok(); ++dit, ibox++)
    {
      const Box& cellBox = m_thisGrids.get(dit());
      const EBISBox& ebisBox = m_thisEBISL[dit()];
      if(!ebisBox.isAllCovered())
        {
          const IntVectSet& cfivs = m_cfIVS[dit()];

          EBCellFAB& consState = consTemp[dit()];
          BaseIVFAB<Real>& redMass = a_massDiff[dit()];

          m_ebPatchGodunov->setValidBox(cellBox, ebisBox, cfivs, a_time, a_dt);

          BaseIFFAB<Real> centroidFlux[SpaceDim];
          const BaseIFFAB<Real>* interpolantGrid[SpaceDim];
          const IntVectSet& ivsIrregSmall = m_irregSetsSmall[dit()];
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              const BaseIFFAB<Real>& interpol = m_fluxInterpolants[idir][dit()];
              interpolantGrid[idir] = &interpol;
              BaseIFFAB<Real>& fluxDir= centroidFlux[idir];
              fluxDir.define(ivsIrregSmall, ebisBox.getEBGraph(), idir, m_nFlux);
            }
          m_ebPatchGodunov->interpolateFluxToCentroids(centroidFlux,
                                                       interpolantGrid,
                                                       ivsIrregSmall);

          Real maxWaveSpeedGrid = 0.0;
          //update the state and interpolate the flux
          const BaseIVFAB<Real>& nonConsDiv = m_nonConsDivergence[dit()];
          const BaseIVFAB<Real>& ebIrregFlux = m_ebIrregFaceFlux[dit()];
          m_ebPatchGodunov->irregularUpdate(consState,
                                            maxWaveSpeedGrid, redMass,
                                            centroidFlux, ebIrregFlux, nonConsDiv,
                                            cellBox, ivsIrregSmall);

          m_ebLevelRedist.increment(redMass, dit(), consInterv);
          maxWaveSpeed = Max(maxWaveSpeed, maxWaveSpeedGrid);

          //do fluxregister mambo
          /*
            Coarse flux register is flux register with the coarse level.
            Fine flux register is the flux register with the fine level.
            To the finer level FR, this level is the coarse level.
            To the coarser level FR, this level is the fine level.
          */
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              Real scale = a_dt;

              BaseIFFAB<Real> fluxRegFlux;
              if(m_doRZCoords)
                {
                  //gather fluxes into flux-register compatible form
                  fluxRegFlux.define(ivsIrregSmall, ebisBox.getEBGraph(), idir, m_nCons);
                  m_ebPatchGodunov->assembleFluxIrr(fluxRegFlux, centroidFlux[idir],
                                                    idir,  cellBox, ivsIrregSmall);
                }

              if(m_hasFiner)
                {
                  if(m_doRZCoords)
                    {
                      a_fineFluxRegister.incrementCoarseIrregulRZ(fluxRegFlux,
                                                                  scale,dit(),
                                                                  consInterv, idir,
                                                                  m_dx);
                    }
                  else
                    {
                      a_fineFluxRegister.incrementCoarseIrregular(centroidFlux[idir],
                                                                  scale,dit(),
                                                                  consInterv, idir);
                    }
                }

              if(m_hasCoarser)
                {
                  for(SideIterator sit; sit.ok(); ++sit)
                    {
                      if(m_doRZCoords)
                        {

                          a_coarFluxRegister.incrementFineIrregulRZ(fluxRegFlux,
                                                                    scale, dit(),
                                                                    consInterv, idir,sit(),
                                                                    m_dx);
                        }
                      else
                        {
                          a_coarFluxRegister.incrementFineIrregular(centroidFlux[idir],
                                                                    scale, dit(),
                                                                    consInterv, idir,sit());
                        }
                    }
                }
            }
        }
    }// end of loop over grids.

  consTemp.exchange(consInterv);
  if(m_doSmushing)
    {
      if(m_useMassRedist)
        {
          //if use mass weighting, need to
          //fix weights of redistribution object
          int densityIndex = m_ebPatchGodunov->densityIndex();
          m_ebLevelRedist.resetWeights(consTemp, densityIndex);
        }

      //do fluxregister samba
      //redistribute mass at this level.  Redistribution to other
      //levels (including reredist) is handled in EBAMRGodunov::postTimeStep
      m_ebLevelRedist.redistribute(consTemp, consInterv);
    }

  //copy back into array with too few ghost cells.
  consTemp.copyTo(consInterv, a_consState, consInterv);

  // Find the minimum of dt's over this level
  Real dtNew = m_dx / maxWaveSpeed;
  //broadcast the minimum to processors
  // gather dtNew
  Vector<Real> all_dt;
  gather(all_dt,dtNew,uniqueProc(SerialTask::compute));
  if (procID() == uniqueProc(SerialTask::compute))
    {
      dtNew = all_dt[0];
      for (int i = 1; i < all_dt.size(); ++i)
        {
          dtNew = Min(dtNew,all_dt[i]);
        }
    }
  broadcast(dtNew,uniqueProc(SerialTask::compute));



  //return the maximum stable time step
  return dtNew;
}
/*****************************/
void
EBLevelGodunovNS::getDrhoDtOverRho(LevelData<EBCellFAB>& a_drhoDt,
                                 const LevelData<EBCellFAB>& a_rhoNew,
                                 const LevelData<EBCellFAB>& a_rhoOld,
                                 const Real& a_dt)
{
  for(DataIterator dit = m_thisGrids.dataIterator(); dit.ok(); ++dit)
    {
      EBCellFAB& drhodt = a_drhoDt[dit()];
      EBCellFAB& rhoold = (EBCellFAB&) a_rhoOld[dit()];
      EBCellFAB& rhonew = (EBCellFAB&) a_rhoNew[dit()];

      drhodt.setVal(0.);
      drhodt.setInvalidData(1.0, 0);
      rhonew.setInvalidData(1.0, 0);
      rhoold.setInvalidData(1.0, 0);

      drhodt += rhoold;
      drhodt -= rhonew;
      drhodt /= rhonew;
      drhodt /= a_dt;
    }
}
/*****************************/
Real EBLevelGodunovNS::getMaxWaveSpeed(const LevelData<EBCellFAB>& a_state)
{
  CH_assert(a_state.disjointBoxLayout() == m_thisGrids);
  Real speed = 0.0;
  for(DataIterator dit = m_thisGrids.dataIterator(); dit.ok(); ++dit)
    {
      const Box& validBox   = m_thisGrids.get(dit());
      const EBISBox& ebisBox= m_thisEBISL[dit()];
      if(!ebisBox.isAllCovered())
        {
          //place holder not used maxWaveSpeed calc
          Real time = 0.0;
          const IntVectSet& cfivs = m_cfIVS[dit()];

          m_ebPatchGodunov->setValidBox(validBox, ebisBox, cfivs, time, time);
          Real speedOverBox = m_ebPatchGodunov->getMaxWaveSpeed(a_state[dit()],
                                                                validBox);
          speed = Max(speed,speedOverBox);
        }
    }

  // gather speed
  Vector<Real> all_speeds;
  gather(all_speeds,speed,uniqueProc(SerialTask::compute));
  if (procID() == uniqueProc(SerialTask::compute))
    {
      speed = all_speeds[0];
      for (int i = 1; i < all_speeds.size (); ++i)
        {
          speed = Max(speed,all_speeds[i]);
        }
    }
  broadcast(speed,uniqueProc(SerialTask::compute));

  return speed;
}
/*****************************/
/*****************************/
