
#ifndef _FORCING_H
#define _FORCING_H

#include "LevelData.H"
#include "EBCellFAB.H"
#include "EBISLayout.H"

/// Forcing to produce a know solution for NS, aka. Method of Manufactured Solutions.
/**
Let the Navier-Stokes viscos operator be $NS(u)$, where $u=(\rho,\rho u, \rho v, \rho w, e)$.
If the forcing $f$ has been computed analytically as $f=w_t-NS(w)$, where $w$ is a 
known function, then the solution of 
$$
  u_t = NS(u) + f
$$
should be $u=w$. The implementation has some likelihood to be correct 
if the error $|u-w|$ converges to zero with the expected rate when the grid is refined.
This class implements the forcing, $f$, and the analytical solution, $w$. $w$ is composed
of trigonometric functions and low order polynomials. Only the constant viscosity case
has been implemented.
*/

class Forcing
{
public:
  /// Constructor
  /** Set up forcing. 
        Input: om1, om2 - frequencies, the manufactured solution
                     consists of terms of type sin(om1*x) or cos(om2*y), etc...
               mu, la, kappa - Viscosity coefficients, must be constants, same as
                               used when solving.
               konst         - The constant R/(M(gamma-1)), used to convert pressure
	                       to temperature.
  */
  Forcing( double om1, double om2, double mu, double la, double kappa, double konst );

  /// Select cylindrical (rz) (default) - 1, Cartesian 2D (xy) - 2, or Cartesian 3D (xyz) - 3
  void set_case( int a_case );

  /// Get the function w, for a grid level
  void get_manufactured_solution( LevelData<EBCellFAB>& ms, const EBISLayout& a_ebises,
				  const Real t, const RealVect& a_dx,
				  const RealVect& a_problo );

  /// Get the forcing corresponding to w, i.e., w_t-NS(w), for a grid level
  /** Input: a_ebises - EBISLayout at the level where a_fo is defined.
             t        - Time at which w is computed.
             a_dx     - Grid spacing at the level.
             a_problo - Origin for the grids.
             mult_vol - If true, a_fo is multiplied by the local cell volume.
      Output: a_fo  - The forcing function
  */
  void get_forcing( LevelData<EBCellFAB>& a_fo, const EBISLayout& a_ebises,
		    const Real t, const RealVect& a_dx, 
		    const RealVect& a_problo, const bool mult_vol );

  /// Get the function w at a point and at a time (needed for b.c.)
  Vector<Real> get_manufactured_solution( const RealVect& point, const Real& t );

  /// Get the forcing function at a point and at a time.
  Vector<Real> get_forcing( const RealVect& point, const Real& t );

  /// Get the velocities at a point and at a time.
  Vector<Real> get_vels( const RealVect& x, const Real& t );

  /// Get the temperature at a point and at a time.
  Real get_T( const RealVect& x, const Real& t );

  /// Get the temperature gradient at a point and at a time.
  Vector<Real> get_gradT( const RealVect& x, const Real& t );

  /// Get the x-direction velocity derivatives at a point and at a time.
  Vector<Real> get_ux( const RealVect& x, const Real& t );

  /// Get the y-direction velocity derivatives at a point and at a time.
  Vector<Real> get_uy( const RealVect& x, const Real& t );

  /// Get the z-direction velocity derivatives at a point and at a time.
  Vector<Real> get_uz( const RealVect& x, const Real& t );

private:

// Conserved variables routines
  void mansol_rz(double r, double z,double t,double konst,double om1,
	      double om2,double var[4] );
  void mansol_xy(double x, double y,double t,double konst,double om1,
	      double om2,double var[4] );
  void mansol_xyz(double x, double y, double z, double t,double konst,double om1,
	      double om2,double var[5] );

  void nsforce_rz(double r, double z, double t, double mu, double la, double kappa,
	       double konst, double om1, double om2, double forc[4] );
  void nsforce_xy(double x, double y, double t, double mu, double la, double kappa,
	       double konst, double om1, double om2, double forc[4] );
  void nsforce_xyz(double x, double y, double z, double t, double mu, double la,
	       double kappa, double konst, double om1, double om2, double forc[5] );

// Primitive variables routines
  void nsvar_rz(double r, double z, double t, double om1, double om2, double nsvar[2]);
  void nsvar_xy(double r, double z, double t, double om1, double om2, double nsvar[2]);
  void nsvar_xyz(double x, double y, double z, double t, double om1, double om2, double nsvar[3]);

  void ux_rz(double r,double z,double t,double om1,double om2,double ux[2]);
  void ux_xy(double x,double y,double t,double om1,double om2,double ux[2]);
  void ux_xyz(double x,double y,double z,double t,double om1,double om2,double ux[3]);

  void uy_rz(double r,double z,double t,double om1,double om2,double ux[2]);
  void uy_xy(double x,double y,double t,double om1,double om2,double ux[2]);
  void uy_xyz(double x,double y,double z,double t,double om1,double om2,double ux[3]);

  void uz_rz(double r,double z,double t,double om1,double om2,double ux[2]);
  void uz_xy(double x,double y,double t,double om1,double om2,double ux[2]);
  void uz_xyz(double x,double y,double z,double t,double om1,double om2,double ux[3]);

  double T_rz(double r, double z, double t, double om1, double om2 );
  double T_xy(double r, double z, double t, double om1, double om2 );
  double T_xyz(double x, double y, double z, double t, double om1, double om2 );

  void gradT_rz(double r,double z,double t,double om1, double om2, double gradt[2]);
  void gradT_xy(double x,double y,double t,double om1, double om2, double gradt[2]);
  void gradT_xyz(double x,double y,double z,double t,double om1, double om2, double gradt[3]);

  double m_om1, m_om2, m_mu, m_la, m_kappa, m_konst;
  int    m_case;

};

#endif
