#include <math.h>

#include "Forcing.H"

//-----------------------------------------------------------------------
Forcing::Forcing( double om1, double om2, double mu, double la, double kappa, double konst )
    : m_om1(om1), m_om2(om2), m_mu(mu), m_la(la), m_kappa(kappa), m_konst(konst)
{
// cases 1- axisym, 2-2D Cartesian, 3-3D Cartesian
     m_case = 1;
}


//-----------------------------------------------------------------------
void Forcing::set_case( int a_case )
{
  if( a_case > 0 && a_case < 4 )
     m_case = a_case;
}


//-----------------------------------------------------------------------
void Forcing::get_manufactured_solution( LevelData<EBCellFAB>& a_ms,
					 const EBISLayout& a_ebises,
					 const Real a_t, const RealVect& a_dx,
					 const RealVect& a_problo )
{
   double t = static_cast<double>(a_t);
   for( DataIterator di=a_ms.getBoxes().dataIterator(); di.ok() ; ++di )
   {
      DataIndex dex = di();
      Box box = a_ms[dex].getRegion();
      for( BoxIterator b(box); b.ok(); ++b )
      {
	 Vector<VolIndex> vofs = a_ebises[dex].getVoFs(b());
         for( int vi=0 ; vi < vofs.size() ; vi++ )
	 {
	    double vars[5];
            RealVect x= RealVect::Zero;
	    for( int d=0 ; d<SpaceDim ; d++ )
	       x[d] = (b()[d]+0.5)*a_dx[d] + a_problo[d];
            if( m_case == 1 )
   	       mansol_rz( x[0], x[1], t, m_konst, m_om1, m_om2, vars );
	    else if( m_case == 2 )
   	       mansol_xy( x[0], x[1], t, m_konst, m_om1, m_om2, vars );
            else if( m_case == 3 )
   	       mansol_xyz( x[0], x[1], x[2], t, m_konst, m_om1, m_om2, vars );
            a_ms[dex](vofs[vi],0) = vars[0];
            a_ms[dex](vofs[vi],1) = vars[1];
            a_ms[dex](vofs[vi],2) = vars[2];
            a_ms[dex](vofs[vi],3) = vars[3];
            if( m_case == 3 )
               a_ms[dex](vofs[vi],4) = vars[4];
	 }
      }
   }
}


//-----------------------------------------------------------------------
Vector<Real> Forcing::get_manufactured_solution( const RealVect& x, const Real& a_t )
{
   double t = static_cast<double>(a_t);
   Vector<Real> retval(SpaceDim,0);
   double forc[5];
   if( m_case == 1 )         	    
      mansol_rz(x[0], x[1], t, m_konst, m_om1, m_om2, forc );
   else if( m_case == 2 )
      mansol_xy(x[0], x[1], t, m_konst, m_om1, m_om2, forc );
   else if( m_case == 3 )
      mansol_xyz(x[0], x[1], x[2], t, m_konst, m_om1, m_om2, forc );
   for( int c=0 ; c<SpaceDim ; c++ )
      retval[c] = forc[c];
   return retval;
}


//-----------------------------------------------------------------------
Vector<Real> Forcing::get_vels( const RealVect& x, const Real& a_t )
{
   Vector<Real> retval(SpaceDim,0);
   double t=static_cast<double>(a_t);
   double val[5];
   if( m_case == 1 )
      nsvar_rz( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 2 )
      nsvar_xy( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 3 )
      nsvar_xyz( x[0], x[1], x[2], t, m_om1, m_om2, val );
   for( int c = 0; c<SpaceDim ; c++ )
     retval[c] = val[c];
   return retval;
}


//-----------------------------------------------------------------------
Real Forcing::get_T( const RealVect& x, const Real& a_t )
{
   Real retval;
   double t=static_cast<double>(a_t);
   if( m_case == 1 )
      retval = T_rz( x[0], x[1], t, m_om1, m_om2 );
   else if( m_case == 2 )
      retval = T_xy( x[0], x[1], t, m_om1, m_om2 );
   else if( m_case == 3 )
      retval = T_xyz( x[0], x[1], x[2], t, m_om1, m_om2 );
   return retval;
}


//-----------------------------------------------------------------------
Vector<Real> Forcing::get_gradT( const RealVect& x, const Real& a_t )
{
   Vector<Real> retval(SpaceDim,0);
   double t=static_cast<double>(a_t);
   double val[5];
   if( m_case == 1 )
      gradT_rz( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 2 )
      gradT_xy( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 3 )
      gradT_xyz( x[0], x[1], x[2], t, m_om1, m_om2, val );
   for( int c = 0; c<SpaceDim ; c++ )
     retval[c] = val[c];
   return retval;
}


//-----------------------------------------------------------------------
Vector<Real> Forcing::get_ux( const RealVect& x, const Real& a_t )
{
   Vector<Real> retval(SpaceDim,0);
   double t=static_cast<double>(a_t);
   double val[5];
   if( m_case == 1 )
      ux_rz( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 2 )
      ux_xy( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 3 )
      ux_xyz( x[0], x[1], x[2], m_om1, m_om2, t, val );
   for( int c = 0; c<SpaceDim ; c++ )
     retval[c] = val[c];
   return retval;
}


//-----------------------------------------------------------------------
Vector<Real> Forcing::get_uy( const RealVect& x, const Real& a_t )
{
   Vector<Real> retval(SpaceDim,0);
   double t=static_cast<double>(a_t);
   double val[5];
   if( m_case == 1 )
      uy_rz( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 2 )
      uy_xy( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 3 )
      uy_xyz( x[0], x[1], x[2], t, m_om1, m_om2, val );
   for( int c = 0; c<SpaceDim ; c++ )
     retval[c] = val[c];
   return retval;
}


//-----------------------------------------------------------------------
Vector<Real> Forcing::get_uz( const RealVect& x, const Real& a_t )
{
   Vector<Real> retval(SpaceDim,0);
   double t=static_cast<double>(a_t);
   double val[5];
   if( m_case == 1 )
      uz_rz( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 2 )
      uz_xy( x[0], x[1], t, m_om1, m_om2, val );
   else if( m_case == 3 )
      uz_xyz( x[0], x[1], x[2], t, m_om1, m_om2, val );
   for( int c = 0; c<SpaceDim ; c++ )
     retval[c] = val[c];
   return retval;
}


//-----------------------------------------------------------------------
void Forcing::get_forcing( LevelData<EBCellFAB>& a_fo,
			   const EBISLayout& a_ebises,
			   const Real t, const RealVect& a_dx,
			   const RealVect& a_problo, const bool mult_vol )
{
   for( DataIterator di=a_fo.getBoxes().dataIterator(); di.ok() ; ++di )
   {
      DataIndex dex = di();
      Box box = a_fo[dex].getRegion();
      for( BoxIterator b(box); b.ok(); ++b )
      {
	 Vector<VolIndex> vofs = a_ebises[dex].getVoFs(b());
         for( int vi=0 ; vi < vofs.size() ; vi++ )
	 {
	    double forc[5];
            RealVect x = RealVect::Zero;
            RealVect of= a_ebises[dex].centroid(vofs[vi]);
	    for( int d=0 ; d<SpaceDim ; d++ )
	       x[d] = (b()[d]+0.5+of[d])*a_dx[d] + a_problo[d];

            if( m_case == 1 )         	    
   	       nsforce_rz(x[0], x[1], t, m_mu, m_la, m_kappa, m_konst, m_om1, m_om2, forc );
	    else if( m_case == 2 )
   	       nsforce_xy(x[0], x[1], t, m_mu, m_la, m_kappa, m_konst, m_om1, m_om2, forc );
            else if( m_case == 3 )
   	       nsforce_xyz(x[0], x[1], x[2], t, m_mu, m_la, m_kappa, m_konst,
			   m_om1, m_om2, forc );
            Real fact=1;
            if( mult_vol )
	       fact = a_ebises[dex].volFrac(vofs[vi]);
            a_fo[dex](vofs[vi],0) = forc[0]*fact;
            a_fo[dex](vofs[vi],1) = forc[1]*fact;
            a_fo[dex](vofs[vi],2) = forc[2]*fact;
            a_fo[dex](vofs[vi],3) = forc[3]*fact;
            if( m_case == 3 )
               a_fo[dex](vofs[vi],4) = forc[4]*fact;
	 }
      }
   }
}


//-----------------------------------------------------------------------
Vector<Real> Forcing::get_forcing( const RealVect& x, const Real& a_t )
{
   Vector<Real> retval(SpaceDim,0);
   double t=static_cast<double>(a_t);
   double forc[5];
   if( m_case == 1 )         	    
      nsforce_rz(x[0], x[1], t, m_mu, m_la, m_kappa, m_konst, m_om1, m_om2, forc );
   else if( m_case == 2 )
      nsforce_xy(x[0], x[1], t, m_mu, m_la, m_kappa, m_konst, m_om1, m_om2, forc );
   else if( m_case == 3 )
      nsforce_xyz(x[0], x[1], x[2], t, m_mu, m_la, m_kappa, m_konst,
			   m_om1, m_om2, forc );
   for( int c=0 ; c<SpaceDim ; c++ )
      retval[c] = forc[c];
   return retval;
}


//-----------------------------------------------------------------------
//             Cartesian XY
//-----------------------------------------------------------------------
void Forcing::mansol_xy( double x,double y,double t,double konst,double om1,
		double om2,double var[4] )
{
  double t1;
  double t11;
  double t14;
  double t15;
  double t16;
  double t19;
  double t2;
  double t20;
  double t22;
  double t24;
  double t28;
  double t29;
  double t31;
  double t33;
  double t34;
  double t36;
  double t4;
  double t7;
  {
    t1 = om1*y;
    t2 = sin(t1);
    t4 = cos(om2*x);
    t7 = sin(om1*t);
    var[0] = 1.0+t2*t4*t7/2.0;
    t11 = sin(om1*x);
    t14 = cos(om2*y);
    t15 = t*t;
    t16 = sin(t15);
    var[1] = var[0]*t11*t14*t16;
    t19 = cos(t1);
    t20 = cos(t);
    var[2] = var[0]*t4*t19*t20;
    t22 = x*x;
    t24 = sin(x*y);
    t28 = t11*t11;
    t29 = t14*t14;
    t31 = t16*t16;
    t33 = t4*t4;
    t34 = t19*t19;
    t36 = t20*t20;
    var[3] = var[0]*(t22+1.0+t24)*konst+var[0]*(t28*t29*t31+t33*t34*t36)/2.0;
    return;
  }
}


//-----------------------------------------------------------------------
void Forcing::nsforce_xy(double x,double y,double t,double mu,double la,double kappa,
		         double konst,double om1,double om2,double forc[4])
{
  double t1;
  double t10;
  double t11;
  double t116;
  double t12;
  double t120;
  double t129;
  double t132;
  double t138;
  double t14;
  double t141;
  double t15;
  double t150;
  double t153;
  double t159;
  double t16;
  double t165;
  double t166;
  double t17;
  double t175;
  double t18;
  double t2;
  double t22;
  double t25;
  double t27;
  double t3;
  double t33;
  double t35;
  double t39;
  double t4;
  double t43;
  double t44;
  double t46;
  double t47;
  double t5;
  double t50;
  double t51;
  double t52;
  double t54;
  double t57;
  double t6;
  double t63;
  double t66;
  double t67;
  double t68;
  double t7;
  double t70;
  double t71;
  double t73;
  double t75;
  double t8;
  double t84;
  double t86;
  double t92;
  double t93;
  double t94;
  double t95;
  double t97;
  double t98;
  double t99;
  {
    t1 = om1*y;
    t2 = sin(t1);
    t3 = om2*x;
    t4 = cos(t3);
    t5 = t2*t4;
    t6 = om1*t;
    t7 = cos(t6);
    t8 = t7*om1;
    forc[0] = t5*t8/2.0;
    t10 = t5*t7;
    t11 = om1*x;
    t12 = sin(t11);
    t14 = om2*y;
    t15 = cos(t14);
    t16 = t*t;
    t17 = sin(t16);
    t18 = t15*t17;
    t22 = sin(t6);
    t25 = 1.0+t5*t22/2.0;
    t27 = cos(t16);
    t33 = 2.0*mu+la;
    t35 = om1*om1;
    t39 = sin(t3);
    t43 = cos(t);
    t44 = t2*om1*t43;
    t46 = t12*t15;
    t47 = om2*om2;
    t50 = t39*om2;
    t51 = t50*t44;
    t52 = -t46*t47*t17+t51;
    forc[1] = t10*om1*t12*t18/2.0+2.0*t25*t12*t15*t27*t+t33*t12*t35*t15*t17-la*
t39*om2*t44-mu*t52;
    t54 = t4*t4;
    t57 = cos(t1);
    t63 = sin(t);
    t66 = cos(t11);
    t67 = t66*om1;
    t68 = sin(t14);
    t70 = t68*om2*t17;
    t71 = t67*t70;
    t73 = t57*t43;
    t75 = -t71-t4*t47*t73;
    forc[2] = t2*t54*t7*om1*t57*t43/2.0-t25*t4*t57*t63-mu*t75+mu*t66*om1*t70+
t33*t4*t57*t35*t43;
    t84 = x*x;
    t86 = sin(x*y);
    t92 = t12*t12;
    t93 = t15*t15;
    t94 = t92*t93;
    t95 = t17*t17;
    t97 = t57*t57;
    t98 = t54*t97;
    t99 = t43*t43;
    t116 = t17*la;
    t120 = t67*t18-t5*om1*t43;
    t129 = t43*la;
    t132 = t4*t57;
    t138 = t66*t66;
    t141 = t35*t93*t95;
    t150 = om2*t17;
    t153 = -t12*t68*t150-t50*t73;
    t159 = mu*t12;
    t165 = mu*t54;
    t166 = t2*t2;
    t175 = y*y;
    forc[3] = t10*om1*(t84+1.0+t86)*konst/2.0+t5*t8*(t94*t95+t98*t99)/4.0+t25*(
4.0*t94*t17*t27*t-2.0*t98*t43*t63)/2.0-t67*t15*t116*t120-t46*t116*(-t12*t35*t18
+t51)+t5*om1*t129*t120-t132*t129*(-t71-t132*t35*t43)-2.0*mu*t138*t141+2.0*mu*
t92*t141+mu*t39*om2*t73*t153-mu*t4*t73*t75+t159*t68*t150*t153-t159*t18*t52-2.0*
t165*t166*t35*t99+2.0*t165*t97*t99*t35-kappa*(2.0-t86*t175-t86*t84);
    return;
  }
}


//-----------------------------------------------------------------------
void Forcing::nsvar_xy(double r, double z, double t, double om1, double om2, double nsvar[2])
{}


//-----------------------------------------------------------------------
double Forcing::T_xy(double r, double z, double t, double om1, double om2 )
{}


//-----------------------------------------------------------------------
void Forcing::gradT_xy(double x,double y,double t,double om1, double om2, double gradt[2])
{}


//-----------------------------------------------------------------------
void Forcing::ux_xy(double x,double y,double t,double om1,double om2,double ux[2])
{}


//-----------------------------------------------------------------------
void Forcing::uy_xy(double x,double y,double t,double om1,double om2,double uy[2])
{}


//-----------------------------------------------------------------------
void Forcing::uz_xy( double r, double z, double t, double om1, double om2, double val[2] )
{
  val[0] = val[1] = 0;
}


//-----------------------------------------------------------------------
//             Cartesian XYZ
//-----------------------------------------------------------------------
void Forcing::mansol_xyz(double x,double y,double z,double t,double konst,double om1,
		     double om2,double var[5] )
{}


//-----------------------------------------------------------------------
void Forcing::nsforce_xyz(double x, double y, double z, double t, double mu, double la,
	       double kappa, double konst, double om1, double om2, double forc[5] )
{}


//-----------------------------------------------------------------------
void Forcing::nsvar_xyz(double x, double y, double z, double t, double om1, double om2, double nsvar[3])
{}


//-----------------------------------------------------------------------
void Forcing::ux_xyz(double x,double y,double z,double t,double om1,double om2,double ux[3])
{}


//-----------------------------------------------------------------------
void Forcing::uy_xyz(double x,double y,double z,double t,double om1,double om2,double uy[3])
{}


//-----------------------------------------------------------------------
void Forcing::uz_xyz(double x,double y,double z,double t,double om1,double om2,double uz[3])
{}


//-----------------------------------------------------------------------
double Forcing::T_xyz(double x, double y, double z, double t, double om1, double om2 )
{}


//-----------------------------------------------------------------------
void Forcing::gradT_xyz(double x,double y,double z,double t,double om1, double om2, double gradt[3])
{}


//-----------------------------------------------------------------------
//             Axisymmetric RZ
//-----------------------------------------------------------------------
void Forcing::mansol_rz(double r,double z,double t,double konst,double om1,
	       double om2,double var[4] )
{
  double t1;
  double t11;
  double t14;
  double t15;
  double t16;
  double t2;
  double t20;
  double t21;
  double t23;
  double t25;
  double t29;
  double t30;
  double t32;
  double t35;
  double t36;
  double t38;
  double t4;
  double t7;
  {
    t1 = om1*z;
    t2 = sin(t1);
    t4 = cos(om2*r);
    t7 = sin(om1*t);
    var[0] = 1.0+t2*t4*t7/2.0;
    t11 = sin(om1*r);
    t14 = cos(om2*z);
    t15 = t*t;
    t16 = sin(t15);
    var[1] = var[0]*t11*t14*t16*r;
    t20 = cos(t1);
    t21 = cos(t);
    var[2] = var[0]*t4*t20*t21;
    t23 = r*r;
    t25 = cos(r*z);
    t29 = t11*t11;
    t30 = t14*t14;
    t32 = t16*t16;
    t35 = t4*t4;
    t36 = t20*t20;
    t38 = t21*t21;
    var[3] = var[0]*(t23+2.0+t25)*konst+var[0]*(t29*t30*t32*t23+t35*t36*t38)/
2.0;
    return;
  }
}


//-----------------------------------------------------------------------
void Forcing::nsforce_rz(double r,double z,double t,double mu,double la,double kappa,
                double konst,double om1,double om2,double forc[4])
{
  double t1;
  double t10;
  double t101;
  double t103;
  double t108;
  double t11;
  double t112;
  double t12;
  double t120;
  double t128;
  double t129;
  double t13;
  double t135;
  double t136;
  double t137;
  double t138;
  double t14;
  double t141;
  double t142;
  double t143;
  double t15;
  double t16;
  double t164;
  double t17;
  double t175;
  double t18;
  double t180;
  double t2;
  double t201;
  double t21;
  double t210;
  double t218;
  double t230;
  double t231;
  double t239;
  double t24;
  double t241;
  double t27;
  double t3;
  double t32;
  double t33;
  double t34;
  double t35;
  double t36;
  double t37;
  double t38;
  double t4;
  double t40;
  double t41;
  double t42;
  double t46;
  double t5;
  double t50;
  double t52;
  double t53;
  double t57;
  double t6;
  double t61;
  double t64;
  double t65;
  double t67;
  double t68;
  double t69;
  double t7;
  double t72;
  double t76;
  double t78;
  double t8;
  double t82;
  double t85;
  double t9;
  double t91;
  double t94;
  double t95;
  double t96;
  double t97;
  double t98;
  double t99;
  {
    t1 = om1*z;
    t2 = sin(t1);
    t3 = om2*r;
    t4 = cos(t3);
    t5 = t2*t4;
    t6 = om1*t;
    t7 = cos(t6);
    t8 = t7*om1;
    t9 = t5*t8;
    forc[0] = t9/2.0;
    t10 = om1*r;
    t11 = sin(t10);
    t12 = om2*z;
    t13 = cos(t12);
    t14 = t11*t13;
    t15 = t*t;
    t16 = sin(t15);
    t17 = t16*r;
    t18 = t14*t17;
    t21 = sin(t6);
    t24 = 1.0+t5*t21/2.0;
    t27 = cos(t15);
    t32 = cos(t10);
    t33 = t32*om1;
    t34 = t13*t16;
    t35 = t34*r;
    t36 = t33*t35;
    t37 = t14*t16;
    t38 = t36+t37;
    t40 = mu*r;
    t41 = om1*om1;
    t42 = t11*t41;
    t46 = -t42*t35+2.0*t33*t34;
    t50 = 1/r;
    t52 = r*r;
    t53 = t34*t52;
    t57 = -t42*t53+4.0*t36+2.0*t37;
    t61 = t33*t53+2.0*t18;
    t64 = sin(t3);
    t65 = t64*om2;
    t67 = cos(t);
    t68 = t2*om1*t67;
    t69 = t65*t68;
    t72 = om2*om2;
    t76 = -t14*t72*t16*r+t69;
    t78 = mu*t11;
    forc[1] = t9*t18/2.0+2.0*t24*t11*t13*t27*t*r-(2.0*mu*t38+2.0*t40*t46)*t50-
la*(t57*t50-t61/t52+t69)-mu*t76+2.0*t78*t34*t50;
    t82 = t4*t4;
    t85 = cos(t1);
    t91 = sin(t);
    t94 = sin(t12);
    t95 = t11*t94;
    t96 = om2*t16;
    t97 = t96*r;
    t98 = t95*t97;
    t99 = t85*t67;
    t101 = -t98-t65*t99;
    t103 = t33*t94;
    t108 = -t103*t97-t95*t96-t4*t72*t99;
    t112 = mu*t4;
    t120 = -t103*t96*t52-2.0*t98;
    forc[2] = t2*t82*t7*om1*t85*t67/2.0-t24*t4*t85*t91-(mu*t101+t40*t108)*t50+
2.0*t112*t85*t41*t67-la*(t120*t50-t4*t85*t41*t67);
    t128 = r*z;
    t129 = cos(t128);
    t135 = t11*t11;
    t136 = t13*t13;
    t137 = t135*t136;
    t138 = t16*t16;
    t141 = t85*t85;
    t142 = t82*t141;
    t143 = t67*t67;
    t164 = t52*mu;
    t175 = la*t32;
    t180 = la*t11;
    t201 = r*t101;
    t210 = 4.0*t40*t11*t34*t38+2.0*t164*t32*om1*t13*t16*t38+2.0*t164*t11*t34*
t46+t175*om1*t34*r*t61+t180*t34*t61+t180*t13*t17*t57-t175*t41*t13*t16*t52*t5*
t67-2.0*t180*t34*r*t4*t68+t180*t53*t69-mu*t64*om2*t99*t201+t112*t99*t101+t112*
t85*t67*r*t108;
    t218 = la*t4;
    t230 = (2.0*mu+la)*t82;
    t231 = t2*t2;
    t239 = sin(t128);
    t241 = z*z;
    forc[3] = t5*t7*om1*(t52+2.0+t129)*konst/2.0+t5*t8*(t137*t138*t52+t142*t143
)/4.0+t24*(4.0*t137*t16*t52*t27*t-2.0*t142*t67*t91)/2.0-t210*t50+t78*t94*t96*
t201-t78*t13*t17*t76+t218*t2*om1*t67*t50*t61-t218*t85*t67*t50*t120-t230*t231*
t41*t143+t230*t141*t143*t41-kappa*((2.0*r-t239*z+r*(2.0-t129*t241))*t50-t129*
t52);
    return;
  }
}


//-----------------------------------------------------------------------
void Forcing::nsvar_rz(double r, double z, double t, double om1, double om2, double nsvar[2])
{
  double t10;
  double t12;
  double t14;
  double t2;
  double t5;
  double t6;
  double t7;
  {
    t2 = sin(om1*r);
    t5 = cos(om2*z);
    t6 = t*t;
    t7 = sin(t6);
    nsvar[0] = r*t2*t5*t7;
    t10 = cos(om2*r);
    t12 = cos(om1*z);
    t14 = cos(t);
    nsvar[1] = t10*t12*t14;
    return;
  }
}


//-----------------------------------------------------------------------
double Forcing::T_rz(double r, double z, double t, double om1, double om2 )
{
  double t1;
  double t3;
  {
    t1 = r*r;
    t3 = cos(r*z);
    return(t1+2.0+t3);
  }
}

//-----------------------------------------------------------------------
void Forcing::gradT_rz(double r,double z,double t,double om1, double om2, double gradt[2])
{
  double t3;
  {
    t3 = sin(r*z);
    gradt[0] = 2.0*r-t3*z;
    gradt[1] = -t3*r;
    return;
  }
}

 
//-----------------------------------------------------------------------
void Forcing::ux_rz(double r,double z,double t,double om1,double om2,double ux[2])
{
  double t1;
  double t11;
  double t15;
  double t18;
  double t19;
  double t2;
  double t5;
  double t6;
  double t7;
  {
    t1 = om1*r;
    t2 = cos(t1);
    t5 = cos(om2*z);
    t6 = t*t;
    t7 = sin(t6);
    t11 = sin(t1);
    ux[0] = t2*om1*t5*t7*r+t11*t5*t7;
    t15 = sin(om2*r);
    t18 = cos(om1*z);
    t19 = cos(t);
    ux[1] = -t15*om2*t18*t19;
    return;
  }
}
 

//-----------------------------------------------------------------------
void Forcing::uy_rz(double r,double z,double t,double om1,double om2,double uy[2])
{
  double t12;
  double t14;
  double t16;
  double t2;
  double t4;
  double t6;
  double t7;
  {
    t2 = sin(om1*r);
    t4 = sin(om2*z);
    t6 = t*t;
    t7 = sin(t6);
    uy[0] = -t2*t4*om2*t7*r;
    t12 = cos(om2*r);
    t14 = sin(om1*z);
    t16 = cos(t);
    uy[1] = -t12*t14*om1*t16;
    return;
  }
}

//-----------------------------------------------------------------------
void Forcing::uz_rz( double r, double z, double t, double om1, double om2, double val[2] )
{
  val[0] = val[1] = 0;
}

