/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "BoxIterator.H"
#include "NeumannHelmholtzDomainBC.H"


//-----------------------------------------------------------------------
NeumannHelmholtzDomainBC::NeumannHelmholtzDomainBC()
   : HelmholtzDomainBC()
{}


//-----------------------------------------------------------------------
NeumannHelmholtzDomainBC::~NeumannHelmholtzDomainBC()
{}


//-----------------------------------------------------------------------
void NeumannHelmholtzDomainBC::get_faceflux( BaseFab<Real>&   a_faceFlux,
                                             const BaseFab<Real>& a_phi,
                                             const RealVect&  a_probLo,
                                             const RealVect&  a_dx,
                                             int              a_idir,
                                             Side::LoHiSide   a_side,
                                             bool             a_use_homogeneous,
				             Real             a_time)
{
   CH_assert( a_use_homogeneous || m_use_function || m_use_value );
   CH_assert( a_phi.nComp() == a_faceFlux.nComp() );

   const Box& box = a_faceFlux.box();
   int iside = a_side == Side::Lo ? 1 : -1;
   if( a_use_homogeneous )
     a_faceFlux.setVal(0.0);
   else 
   {
      BoxIterator bit(box);
      for( bit.begin(); bit.ok(); ++bit )
      {
         Vector<Real> value = m_value;
         if( m_use_function )
	 {
            RealVect point = static_cast<RealVect>( bit() );
            for( int d=0 ; d < SpaceDim ; d++ )
               point[d] = (point[d]+0.5)*a_dx[d]+a_probLo[d];
            point[a_idir] -= iside*0.5*a_dx[a_idir];
  	    RealVect normal = RealVect::Zero;
	    normal[a_idir] = iside;
            value = m_func->get_value( point, a_time, normal );
	 }
         for( int c = 0 ; c < a_phi.nComp() ; c++ )
	    a_faceFlux(bit(),c) = iside*value[c];
         if( m_rz && a_idir == 0 )
	 {
	    Real r = static_cast<Real>(bit()[0]);
	    r  = (r+0.5)*a_dx[0]+a_probLo[0];
	    Real rb = r-iside*a_dx[0]*0.5;
	    for( int c = 0 ; c < a_phi.nComp() ; c++ )
	       a_faceFlux(bit(),c) *= rb/r;
	 }
      }
   }
}



//-----------------------------------------------------------------------
void NeumannHelmholtzDomainBC::get_faceflux(Vector<Real>&   a_faceFlux,
                                            const VolIndex&  a_vof,
                                            const EBCellFAB& a_phi,
                                            const RealVect&  a_probLo,
                                            const RealVect&  a_dx,
                                            int              a_idir,
                                            Side::LoHiSide   a_side,
                                            bool             a_use_homogeneous,
              				    Real             a_time)
{
   CH_assert( a_use_homogeneous || m_use_function || m_use_value );
   CH_assert( a_faceFlux.size() == a_phi.nComp() );
   int iside = a_side == Side::Lo ? 1 : -1 ;
   const EBISBox& ebisBox = a_phi.getEBISBox();
   if( a_use_homogeneous )
       a_faceFlux.assign(0.0);
   else 
   {
      Vector<Real> flux = m_value;
      Vector<Real> totalMassFlux(a_phi.nComp(),0.0);
      Vector<FaceIndex> faces = ebisBox.getFaces( a_vof, a_idir, a_side );
      for( int i = 0; i < faces.size(); i++ )
      {
         RealVect point;
         if( m_use_function || m_rz )
	 {
            point = static_cast<RealVect>( a_vof.gridIndex() );
            for( int d=0 ; d < SpaceDim ; d++ )
               point[d] = ( point[d] + 0.5 )*a_dx[d] + a_probLo[d];
            point[a_idir] -= iside*0.5*a_dx[a_idir];
	 }
	 if( m_use_function )
	 {
            RealVect normal = RealVect::Zero;
            normal[a_idir]  = iside;
            flux = m_func->get_value( point, a_time, normal );
	 }
         if( m_rz )
	 {
            Real r = static_cast<Real>(a_vof.gridIndex()[0])+ebisBox.centroid(a_vof)[0];
            r  = (r+0.5)*a_dx[0]+a_probLo[0];
            Real rb;
            if( a_idir == 0 )
	       rb = point[0];
	    else
 	       rb = point[0] + ebisBox.centroid(faces[i])[0]*a_dx[0];
	    for( int c=0 ; c < a_phi.nComp() ; c++ )
	       totalMassFlux[c] += (rb/r)*ebisBox.areaFrac(faces[i])*flux[c];
	 }
	 else
	    for( int c = 0; c < a_phi.nComp() ; c++ )
	       totalMassFlux[c] += ebisBox.areaFrac(faces[i])*flux[c];
      }
      for( int c=0 ; c<a_phi.nComp() ; c++ )
	 a_faceFlux[c] = iside * totalMassFlux[c];
   }
}


//-----------------------------------------------------------------------
void NeumannHelmholtzDomainBC::get_diagonal_weight( Vector<Real>& a_wgh,
						    const VolIndex& a_vof, 
						    const EBISBox&  a_ebis,
						    const RealVect& a_probLo,
						    const RealVect& a_dx,
						    int a_idir,
						    Side::LoHiSide a_side )
{
   for( int c=0 ; c < a_wgh.size() ; c++ )
       a_wgh[c] = 0.0;
}


//-----------------------------------------------------------------------
//                               Factory class
//-----------------------------------------------------------------------
NeumannHelmholtzDomainBCFactory::NeumannHelmholtzDomainBCFactory():HelmholtzDomainBCFactory()
{}

//-----------------------------------------------------------------------
NeumannHelmholtzDomainBCFactory::~NeumannHelmholtzDomainBCFactory()
{}

//-----------------------------------------------------------------------
HelmholtzDomainBC* NeumannHelmholtzDomainBCFactory::create()
{
   HelmholtzDomainBC* newBC = new NeumannHelmholtzDomainBC();
   newBC->set_order( m_order );
   newBC->set_rz( m_rz );
   if( m_use_function )
      newBC->set_function( m_func );
   else if( m_use_value )
      newBC->set_value( m_value );
  return newBC;
}
