/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "NeumannHelmholtzEBBC.H"

NeumannHelmholtzEBBC::NeumannHelmholtzEBBC()
  : HelmholtzEBBC()
{}


//-----------------------------------------------------------------------
NeumannHelmholtzEBBC::~NeumannHelmholtzEBBC()
{}


//-----------------------------------------------------------------------
void NeumannHelmholtzEBBC::get_ebflux(Vector<Real>&     a_flux,
                            const VolIndex&             a_vof,
                            const LevelData<EBCellFAB>& a_phi,
                            const DataIndex&            a_dit,
                            const RealVect&             a_probLo,
                            const RealVect&             a_dx,
			    bool                        a_use_homogeneous,
                            Real                        a_time )
{
  CH_assert( a_use_homogeneous || m_use_function || m_use_value );

  const EBCellFAB& curPhi     = a_phi[a_dit];
  const EBISBox&   curEBISBox = curPhi.getEBISBox();

  RealVect point = static_cast<RealVect>(a_vof.gridIndex())+
	                                   curEBISBox.bndryCentroid( a_vof );
  for( int d=0 ; d < SpaceDim ; d++ )
     point[d] = (point[d]+0.5)*a_dx[d]+a_probLo[d];

  Vector<Real> value(a_phi.nComp(),0.0);
  if( !a_use_homogeneous && m_use_function )
  {
     RealVect normal = curEBISBox.normal(a_vof);
     value = m_func->get_value( point, a_time, normal );
  }
  else if( !a_use_homogeneous )
     value = m_value;
    
  if( m_rz )
  {
     Real rb = point[0];
     Real rc = (static_cast<RealVect>( a_vof.gridIndex() )+
	                                   curEBISBox.centroid( a_vof ))[0];
     rc = (rc+0.5)*a_dx[0]+a_probLo[0];
     for( int c = 0 ; c < a_phi.nComp() ; c++ )
        value[c] *= rb/rc;
  }
  a_flux = value;
  const Real& areaFrac = curEBISBox.bndryArea( a_vof );
  for( int c = 0 ; c< a_phi.nComp() ; c++ )
     a_flux[c] *= areaFrac;
}


//-----------------------------------------------------------------------
NeumannHelmholtzEBBCFactory::NeumannHelmholtzEBBCFactory()
  : HelmholtzEBBCFactory()
{}


//-----------------------------------------------------------------------
NeumannHelmholtzEBBCFactory::~NeumannHelmholtzEBBCFactory()
{}


//-----------------------------------------------------------------------
HelmholtzEBBC* NeumannHelmholtzEBBCFactory::create( const ProblemDomain& a_domain,
						    const EBISLayout&    a_layout,
						    const RealVect&      a_dx )
{
   HelmholtzEBBC* new_bc = new NeumannHelmholtzEBBC();
   new_bc->set_rz( m_rz );
   if( m_use_function )
      new_bc->set_function( m_func );
   else if( m_use_value )
      new_bc->set_value( m_value );
   return new_bc;
}
