/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "AMRLevel.H"

#include "AMRLevelMHDFactory.H"
#include "AMRLevelMHD.H"

AMRLevelMHDFactory::AMRLevelMHDFactory()
{
  m_godunovPhysics = NULL;
  m_isDefined = false;
}

AMRLevelMHDFactory::~AMRLevelMHDFactory()
{
  if (m_godunovPhysics != NULL)
  {
    delete m_godunovPhysics;
    m_godunovPhysics = NULL;
  }

  m_isDefined = false;
}

void AMRLevelMHDFactory::define(const Real&                 a_cfl,
                                const Real&                 a_domainLength,
                                const int&                  a_verbosity,
                                const Real&                 a_refineThresh,
                                const int&                  a_tagBufferSize,
                                const Real&                 a_initialDtMultiplier,
                                const GodunovPhysics* const a_godunovPhysics,
                                const int&                  a_normalPredOrder,
                                const bool&                 a_useFourthOrderSlopes,
                                const bool&                 a_usePrimLimiting,
                                const bool&                 a_useCharLimiting,
                                const bool&                 a_useFlattening,
                                const bool&                 a_useArtificialViscosity,
                                const Real&                 a_artificialViscosity,
                                const bool&                 a_doFilterBField)
{
  // Store the CFL number
  m_cfl = a_cfl;

  // Store the physical dimension of the longest side of the domain
  m_domainLength = a_domainLength;

  // Store the verbosity of the object
  m_verbosity = a_verbosity;

  // Store the refinement threshold for gradient
  m_refineThresh = a_refineThresh;

  // Store the tag buffer size
  m_tagBufferSize = a_tagBufferSize;

  // Store the initial dt multiplier
  m_initialDtMultiplier = a_initialDtMultiplier;

  // Delete any existing physics object
  if (m_godunovPhysics != NULL)
  {
    delete m_godunovPhysics;
    m_godunovPhysics = NULL;
  }

  // Store the object that supplies the physics needed by the integrator
  // (used as a factory)
  m_godunovPhysics = a_godunovPhysics->new_godunovPhysics();

  // Store the order of the normal predictor (1 -> PLM, 2 -> PPM)
  m_normalPredOrder = a_normalPredOrder;

  // Store the slope computation parameters
  m_useFourthOrderSlopes = a_useFourthOrderSlopes;
  m_usePrimLimiting      = a_usePrimLimiting;
  m_useCharLimiting      = a_useCharLimiting;
  m_useFlattening        = a_useFlattening;

  // Artificial viscosity coefficient must be greater than zero
  CH_assert(!a_useArtificialViscosity || (a_artificialViscosity >= 0.0));

  // Store the artificial viscosity flag and coefficient
  m_useArtificialViscosity = a_useArtificialViscosity;
  m_artificialViscosity    = a_artificialViscosity;

  //Filter magnetic field
  m_doFilterBField = a_doFilterBField;
  // The object is defined
  m_isDefined = true;
}

// Virtual constructor
AMRLevel* AMRLevelMHDFactory::new_amrlevel() const
{
  // Make sure everything is defined
  CH_assert(isDefined());

  // Create a new AMRLevelMHD
  AMRLevelMHD* amrGodPtr = new AMRLevelMHD();

  // Define the new object
  amrGodPtr->defineParams(m_cfl,
                          m_domainLength,
                          m_verbosity,
                          m_refineThresh,
                          m_tagBufferSize,
                          m_initialDtMultiplier,
                          m_godunovPhysics,
                          m_normalPredOrder,
                          m_useFourthOrderSlopes,
                          m_usePrimLimiting,
                          m_useCharLimiting,
                          m_useFlattening,
                          m_useArtificialViscosity,
                          m_artificialViscosity,
                          m_doFilterBField);

  // Return it
  return (static_cast <AMRLevel*> (amrGodPtr));
}

// Check that everything is defined
bool AMRLevelMHDFactory::isDefined() const
{
  return m_isDefined;
}
