/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// DTGraves, Mon, July 19, 1999

#ifndef _POISSONBC_H_
#define _POISSONBC_H_

#include "Box.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "LoHiSide.H"
#include "GhostBC.H"

/// Class to enforce homogeneous Neumann boundary conditions
/**
    BoxGhostBC class implentation to enforce homogeneous neumann
    boundary conditions
    Recall that  BoxGhostBC is a class to encapsulate the
    operations of ghost-cell boundary conditions at a face.
    If the solution is phi and the face normal direction is x,
    the boundary conditions usually used can be expressed as
    A*phi + B*dphi/dx = C.
    For homogeneous Neumann BCs, A=C=0, B= 1
    These functions are only meant to be called by DomainGhostBC
 */
class NeumannBC: public BoxGhostBC
{
public:
  friend class DomainGhostBC;

  /// Null constructor
  NeumannBC();

  /// Basic constructor -- sets Interval to (0,0)
  NeumannBC(int dir,
            Side::LoHiSide sd);

  /// Full constructor
  NeumannBC(int             dir,
            Side::LoHiSide  sd,
            const Interval& a_comps);

  /// Destructor
  ~NeumannBC();

  void setInhomVal(Real a_inhomVal);
protected:
  ///
  /**
     Fills coefficient arrays --
     fills neumann factor with 1.0
     and other two fields with 0.0.
   */
  virtual void fillBCValues(FArrayBox& a_neumfac,
                            FArrayBox& a_dircfac,
                            FArrayBox& a_inhmval,
                            Real       a_dx,
                            const Box& a_domain) const;

  ///
  /**
     Fills coefficient arrays --
     fills neumann factor with 1.0
     and other two fields with 0.0.
   */
  virtual void fillBCValues(FArrayBox&           a_neumfac,
                            FArrayBox&           a_dircfac,
                            FArrayBox&           a_inhmval,
                            Real                 a_dx,
                            const ProblemDomain& a_domain) const;

  ///
  /**
     Virtual constructor workaround.
  */
  BoxGhostBC* new_boxghostbc() const;

private:
  Real m_inhomVal;
  NeumannBC(const NeumannBC& a_neuin)
  {
  }

  virtual void operator=(const NeumannBC& a_neuin)
  {
  }
};

/// Class to enforce homogeneous Dirichlet boundary conditions
/**
    BoxGhostBC class implentation to enforce homogeneous dirchelet
    boundary conditions
    Recall that
    BoxGhostBC is a class to encapsulate the
    operations  of ghost-cell boundary conditions at a face.
    If the solution is phi and the face normal direction is x,
    the boundary conditions usually used can be expressed as
    A*phi + B*dphi/dx = C.
    For homogeneous Dirichlet BCs, B=C=0, A= 1
    These functions are only meant to be called by DomainGhostBC
 */
class DirichletBC: public BoxGhostBC
{
public:
  friend class DomainGhostBC;

  /// Null constructor
  DirichletBC();

  /// Simple constructor (sets Interval to (0,0)
  DirichletBC(int dir, Side::LoHiSide sd);

  /// Full constructor
  DirichletBC(int dir, Side::LoHiSide sd,
              const Interval& a_comps);

  /// Destructor
  ~DirichletBC();

  void setInhomVal(Real a_inhomVal);
private:
  ///
  /**
     fills Dirichlet factor with 1.0
     and other two fields with 0.0
   */
  virtual void fillBCValues(FArrayBox& a_neumfac,
                            FArrayBox& a_dircfac,
                            FArrayBox& a_inhmval,
                            Real       a_dx,
                            const Box& a_domain) const;

  ///
  /**
     fills Dirichlet factor with 1.0
     and other two fields with 0.0
   */
  virtual void fillBCValues(FArrayBox&           a_neumfac,
                            FArrayBox&           a_dircfac,
                            FArrayBox&           a_inhmval,
                            Real                 a_dx,
                            const ProblemDomain& a_domain) const;

  ///
  BoxGhostBC* new_boxghostbc() const;

private:
  Real m_inhomVal;

  DirichletBC(const DirichletBC& a_dircin)
  {
  }

  virtual void operator=(const DirichletBC& a_dircin)
  {
  }
};

#endif
