/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _AMR_H_
#define _AMR_H_

#include <iostream>
#include <string>
#include <ctime>

#include "REAL.H"
#include "Vector.H"
#include "AMRLevel.H"
#include "AMRLevelFactory.H"
#include "BRMeshRefine.H"
#include "ProblemDomain.H"
#include "Box.H"
#include "CH_HDF5.H"
#ifdef CH_USE_TIMER
#include "CH_Timer.H"
#endif

/// Framework for Berger-Oliger timestepping for AMR
/**
   This class is a framework for Berger-Oliger timestepping for
   adaptive mesh refinement of time-dependent problems.  It is
   applicable to both hyperbolic and parabolic problems.  It
   represents a hierarchy of levels of refinement as a collection of
   AMRLevel objects.

   The usage pattern is this:

   <ol>
     <li>Call define to define the stuff that does not change throughout
         the run (maxlevel, refinement ratios, domain, and operator).
     <li>Modify any parameters you like (blocking factor and so forth
         using access functions.
     <li>Call any one of the three setup functions so AMR can set up all its
         internal data structures.
     <li>Call run to run the calculation.
     <li>Call conclude to output (to stdout) how many cells were updated
         and so forth.
   </ol>
*/
class AMR
{
public:
  ///
  /**
     Default constructor.  Use must subsequently call a define() function
     and a setUp function.
  */
  AMR();

  ///
  /**
     Destructor.
  */
  ~AMR();

  ///
  /**
     Defines this object.  User must call a setup function before running.
     Arguments:

     <ul>
      <li>a_prob_domain (not modified): problem domain on the base level.
      <li>a_ref_ratios (not modified): refinement ratios.  There must be at
          least a_max_level+1 elements,  or an error will result. Element
          zero is the base level.
      <li>a_max_level (not modified): the maximum level allowed, where the
          base level is zero.  There will be a total of a_max_level+1
          levels, since level zero and level a_max_level will both
          exist.
      <li>a_amrlevel_ptr (not modified): pointer to a physics class factory
          object.  The object it points to is used to construct the
          collection of AMRLevel objects in this AMR as objects of the
          physics class type.  It can be destructed after this define()
          returns.
     </ul>
  */
  void define(int                          a_max_level,
              const Vector<int>&           a_ref_ratios,
              const Box&                   a_prob_domain,
              const AMRLevelFactory* const a_amrLevelFact);

  ///
  /**
     Defines this object.  User must call a setup function before running.
     Arguments:

     <ul>
      <li>a_prob_domain (not modified): problem domain on the base level.
      <li>a_ref_ratios (not modified): refinement ratios.  There must be at
          least a_max_level+1 elements,  or an error will result. Element
          zero is the base level.
      <li>a_max_level (not modified): the maximum level allowed, where the
          base level is zero.  There will be a total of a_max_level+1
          levels, since level zero and level a_max_level will both
          exist.
      <li>a_amrlevel_ptr (not modified): pointer to a physics class factory
          object.  The object it points to is used to construct the
          collection of AMRLevel objects in this AMR as objects of the
          physics class type.  It can be destructed after this define()
          returns.
     </ul>
  */
  void define(int                          a_max_level,
              const Vector<int>&           a_ref_ratios,
              const ProblemDomain&         a_prob_domain,
              const AMRLevelFactory* const a_amrLevelFact);

  ///
  /**
     Has a define function been called?  Lots of these
     functions will assert fail if not.
   */
  bool isDefined() const;

  ///
  /**
     Has a setUp function been called?  If not,
     you can't call run
   */
  bool isSetUp() const;

#ifdef CH_USE_HDF5
  ///
  /**
     Sets up this object from checkpointed data.  User must have
     previosuly called define.
     Need to call this or setupForNewAMRRun or setupforFixedHierarchyRun
     before you run.
  */
  void setupForRestart(HDF5Handle& a_handle);
#endif

  ///
  /**
     Sets up this object for cold start.  User must have previously
     called define().
     Need to call this or setupForRestart or setupforFixedHierarchyRun
     before you run.
   */
  void setupForNewAMRRun();

  /**
     This sets the hierarchy and sets regrid_intervals to -1
     (turns off regridding).  If you want to keep regridding
     on, reset regridIntervals after this call;
   */
  void setupForFixedHierarchyRun(const Vector<Vector<Box> >& a_amr_grids,
                                 int                         a_proper_nest = 1);

  ///
  /**
     You should call this last.  It writes the last
     checkpoint file and tells you how many cells you updated and
     all that.
   */
  void conclude() const;

  ///
  /**
     Runs the calculation.  User must have previously called
     both the define function and a setup function.
  */
  void run(Real a_max_time, int a_max_step);

  ///
  /**
     Sets the plot file prefix.
     Should be called after define
     and before setup.
  */
  void plotPrefix(const std::string& a_plotfile_prefix);

  ///
  /**
     Sets the checkpoint file prefix.
     Should be called after define
     and before setup.
  */
  void checkpointPrefix(const std::string& a_checkpointfile_prefix);

  ///
  /**
     Sets the interval to write plot files, in terms of the base level
     time step.
  */
  void plotInterval(int a_plot_interval);

  ///
  /**
     Sets the interval to write checkpoint files, in terms of the base level
     time step.
  */
  void checkpointInterval(int a_checkpoint_interval);

  ///
  /**
     Set the maximum grid size.  Should be called after define
     and before setup.
   */
  void maxGridSize(int a_max_grid_size);

  ///
  /**
     Set the maximum grid size for level 0 grids.
     Defaults to maxGridSize, so should be called after maxGridSize
     and before setup
   */
  void maxBaseGridSize(int a_max_base_grid_size);

  ///
  /**
     Set the factor by which the current dt must exceed the new (max) dt
     for time subcycling to occur (i.e., reduction of the current dt by
     powers of 2).
   */
  void dtToleranceFactor(Real a_dt_tolerance_factor);

  ///
  /**
     Set the fill ratio for meshrefinement.  Should be called after define
     and before setup.
   */
  void fillRatio(Real a_fillRat);

  ///
  /**
     Set the blocking factor for meshrefinement.  Should be called after define
     and before setup.
   */
  void blockFactor(int a_blockFactor);

  ///
  /**
     Set the buffering for the meshrefinement. Should be called after define
     and before setup.
  */
  void gridBufferSize(int a_grid_buffer_size);

  ///
  /**
     Returns the maximum grid size.
   */
  int maxGridSize() const;

  ///
  /**
     Returns the maximum grid size for level 0 grids.
   */
  int maxBaseGridSize() const;

  ///
  /**
     Sets verbosity level to a_verbosity.

     -inf to 0: print nothing.<br>
     1: prints a message every coarse time step.<br>
     2: also prints a message when regridding and writing files.<br>
     3: also prints function trace, and some more inf during file I/O.<br>
     4 to inf: also prints list of boxes and processor maps during regrid.<br>

     This should be OK to call any time after define and before run.
  */
  void verbosity (int a_verbosity);

  ///
  /**
     Sets the regridding intervals.  This should be OK to call any time
     after define and before run.
   */
  void regridIntervals(const Vector<int>& a_regridIntervals);

  ///
  /**
     Returns current verbosity level.

     -inf to 0: print nothing.<br>
     1: prints a message every coarse time step.<br>
     2: also prints a message when regridding and writing files.<br>
     3: also prints function trace, and some more inf during file I/O.<br>
     4 to inf: also prints list of boxes and processor maps during regrid.<br>
  */
  int verbosity () const;

  ///
  /** Set maximum factor by which a timestep can grow. */
  void maxDtGrow(Real a_dtGrowFactor);

  ///
  /** Set amount by which two times may differ and still be considered
      equal (used to determine whether we've reached stop time yet)
      In practice, this is multiplied by the current base-level timestep,
      so we consider ourselves finished if (stopTime - time) < timeEps*dt0
  */
  void timeEps(Real a_timeEps);

  ///
  /** Returns current maximum factor by which a timestep can grow. */
  Real maxDtGrow() const;

  ///
  /** Returns amount by which two times may differ and still be considered
      equal (used to determine whether we've reached stop time yet)
  */
  Real timeEps() const;

  ///
  /** Set a fixed timestep. */
  void fixedDt(Real a_dt);

  ///
  /** Returns the fixed timestep. */
  Real fixedDt() const;

  ///
  /**
     Returns a vector of all the AMR levels
   */
  Vector<AMRLevel*> getAMRLevels();

  /// 
  /** 
      Sets the initial time before starting compututation; this must be 
      called before calling setupForFixedHierarchyRun or setupForNewAMRRun.
      Note that this will be over-ridden by the checkpointfile value if 
      restarting using setupForRestart.
  */
  void initialTime(Real a_initialTime);

  ///
  /**
     Returns the current time
   */
  Real getCurrentTime() const;

#ifdef CH_USE_TIMER
  ///
  /**
     Returns the Chombo timer, and resets it if the arg is not NULL
  */
  Chombo::Timer * timer(Chombo::Timer *a_timer = NULL );
#endif

protected:
  // advance by dt on this level and all finer levels and return the
  // number of steps left - given the number of steps left on entry.
  int timeStep(int a_level, int a_stepsLeft, bool a_coarseTimeBoundary);

  //intenal use only
  void clearMemory();

  // make new grids.
  void regrid(int a_base_level);

  // make initial grids.
  // not an accurate name, since there are no grids to re.
  void initialGrid();

  void writePlotFile() const;

  void writeCheckpointFile() const;

  // computes maximum stable time step given the maximum stable time
  // step on the individual levels.
  void assignDt();

  // whether regridding should be done now.
  bool needToRegrid(int a_level, int a_numStepsLeft) const;

  void makeBaseLevelMesh (Vector<Box>& a_grids) const;

  void setDefaultValues();

  int  m_blockFactor;
  Real m_fillRatio;
  int  m_max_level;
  int  m_finest_level_old;
  int  m_finest_level;
  int  m_checkpoint_interval;
  int  m_plot_interval;
  int  m_max_grid_size;
  int  m_max_base_grid_size;
  Real m_dt_tolerance_factor;
  Real m_fixedDt;

  bool              m_isDefined;
  bool              m_isSetUp;
  Vector<AMRLevel*> m_amrlevels;
  Vector<int>       m_ref_ratios;
  Vector<int>       m_reduction_factor;
  Vector<int>       m_regrid_intervals;

  Real         m_dt_base;
  // New (maximum) dt
  Vector<Real> m_dt_new;
  // Current dt
  Vector<Real> m_dt_cur;
  Real         m_maxDtGrow;
  Real         m_time_eps;

  BRMeshRefine m_mesh_refine;
  bool         m_use_meshrefine;

  Vector<Vector<Box> > m_amr_grids;

  int m_cur_step;
  int m_restart_step;
  int m_lastcheck_step;

  Real         m_cur_time;
  Vector<int>  m_steps_since_regrid;
  Vector<long> m_cell_updates;

  std::string m_plotfile_prefix;
  std::string m_checkpointfile_prefix;

  int m_verbosity;
#ifdef CH_USE_TIMER
  Chombo::Timer *m_timer;  //assumes the application manages the memory
#endif
};

#endif
