/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _PATCHGODUNOV_H_
#define _PATCHGODUNOV_H_

#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "FluxBox.H"

#include "GodunovPhysics.H"
#include "GodunovUtilities.H"

///
/**
   The base class PatchGodunov provides an implementation of a second-order,
   unsplit Godunov method acting on a single grid/patch.  PatchGodunov
   provides an interface to the level integrator, LevelGodunov, which manages
   the entire level and flux corrections (via flux registers) to the coarser
   and finer levels.
 */
class PatchGodunov
{
public:
  /// Constructor
  /**
   */
  PatchGodunov();

  /// Destructor
  /**
   */
  virtual ~PatchGodunov();

  /// Define the object
  /**
   */
  virtual void define(const ProblemDomain&        a_domain,
                      const Real&                 a_dx,
                      const GodunovPhysics* const a_gdnvPhysicsPtr,
                      const int&                  a_normalPredOrder,
                      const bool&                 a_useFourthOrderSlopes,
                      const bool&                 a_usePrimLimiting,
                      const bool&                 a_useCharLimiting,
                      const bool&                 a_useFlattening,
                      const bool&                 a_useArtificialViscosity,
                      const Real&                 a_artificialViscosity);

  /// Set the current time before calling updateState, computeWHalf, computeUpdate.
  /**
   */
  virtual void setCurrentTime(const Real& a_currentTime);

  /// Set the current box before calling updateState, computeWHalf, computeUpdate.
  /**
   */
  virtual void setCurrentBox(const Box& a_currentBox);

  /// Update the conserved variables and return the fluxes used to do this.
  /**
     Update the conserved variables and return the final fluxes that were used
     for this.  Compute the fluxes using a second-order, unsplit Godunov method
     based on the input conserved variables, a_U, and source terms, a_S.  If
     there are no source terms then a_S should be null constructed.  Also
     return the maximum wave speed.
   */
  virtual void updateState(FArrayBox&       a_U,
                           FluxBox&         a_F,
                           Real&            a_maxWaveSpeed,
                           const FArrayBox& a_S,
                           const Real&      a_dt,
                           const Box&       a_box);

  /// Compute the time-centered values of the primitive variables on cell faces.
  /**
     This API is used in cases where some operation over the whole
     level must be performed on the face-centered variables prior to the
     final difference update. Examples include incompressible flow and MHD,
     in which it is necessary to compute the projection of a face-centered
     vector field on its divergence-free part. To complete the differencing,
     it is necessary to call the member function computeUpdate.
   */
  virtual void computeWHalf(FluxBox&         a_WHalf,
                            const FArrayBox& a_U,
                            const FArrayBox& a_S,
                            const Real&      a_dt,
                            const Box&       a_box);

  /// Compute the increment in the conserved variables from face variables.
  /**
   Compute dU = dt*dUdt, the change in the conserved variables over
   the time step. The fluxes are returned are suitable for use in refluxing.
   */
  virtual void computeUpdate(FArrayBox&       a_dU,
                             FluxBox&         a_Flux,
                             const FArrayBox& a_U,
                             const FluxBox&   a_WHalf,
                             const Real&      a_dt,
                             const Box&       a_box);

  /// Compute fluxes from primitive variables on faces
  /**
   */
  void computeFluxes(FluxBox&        a_F,
                     const FluxBox&  a_wHalf,
                     const Box&      a_box);

  /// PLM normal predictor
  /**
     Compute the increments in the characteristic amplitudes using PLM
   */
  void PLMNormalPred(FArrayBox&       a_WMinus,
                     FArrayBox&       a_WPlus,
                     const Real&      a_dt,
                     const Real&      a_dx,
                     const FArrayBox& a_W,
                     const FArrayBox& a_flat,
                     const int&       a_dir,
                     const Box&       a_box);

  /// PPM normal predictor
  /**
     Compute the increments in the characteristic amplitudes using PPM
   */
  void PPMNormalPred(FArrayBox&       a_WMinus,
                     FArrayBox&       a_WPlus,
                     const Real&      a_dt,
                     const Real&      a_dx,
                     const FArrayBox& a_W,
                     const FArrayBox& a_flat,
                     const int&       a_dir,
                     const Box&       a_box);

  /// Accessor to get at the GodunovPhysics object
  /**
   */
  GodunovPhysics* getGodunovPhysicsPtr();

  /// Is the object completely defined
  /**
     Return true if the object is completely defined.
   */
  virtual bool isDefined() const;


protected:
  // Problem domain and grid spacing
  ProblemDomain m_domain;
  Real          m_dx;

  // Object providing physics information
  GodunovPhysics*  m_gdnvPhysics;

  // Object containing various methods for the Godunov computation
  GodunovUtilities m_util;

  // Order of the normal predictor (1 -> PLM, 2-> PPM)
  int m_normalPredOrder;

  // Use 4th order slope computations (otherwise, use 2nd order)
  bool m_useFourthOrderSlopes;

  // Do slope limiting in the primative or characteristic variables, respect.
  bool m_usePrimLimiting;
  bool m_useCharLimiting;

  // Do slope flattening - MUST BE USING 4th order slopes
  bool m_useFlattening;

  // Apply artificial viscosity of a set value
  bool m_useArtificialViscosity;
  Real m_artificialViscosity;

  // Current time and has it been set
  Real m_currentTime;
  bool m_isCurrentTimeSet;

  // Current box and has it been set
  Box  m_currentBox;
  bool m_isCurrentBoxSet;

  // Has this object been defined
  bool m_isDefined;


private:
  // Disallowed for all the usual reasons
  void operator=(const PatchGodunov& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  PatchGodunov(const PatchGodunov& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
