/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// DTGraves, Fri, Dec 3, 1999

#include <fstream>

#include <string>
using std::fstream;
using std::string;

#include <cstdio>
#ifndef __IBMCPP__
using std::sprintf;
#endif
//using std::tempnam;

#include <cstdlib>
using std::system;

#include <cmath>

#ifdef CH_USE_HDF5
#include "CH_HDF5.H"
#endif

#include "AMRIO.H"
#include "BoxIterator.H"
#include "LayoutIterator.H"

#ifdef CH_USE_HDF5
/*
\\ write out hierarchy of amr data in HDF5 format
\\ filename,  == file to output to
\\ a_vectData == data at each level
\\ a_vectNames== names of variables
\\ a_domain == domain at coarsest level
\\ a_dx     == grid spacing at coarsest level
\\ a_dt     == time step at coarsest level
\\ a_time     == time
\\ a_vectRatio == refinement ratio at all levels
\\ (ith entry is refinement ratio between levels i and i + 1)
\\ a_numLevels == number of levels to output
*/
void
WriteAMRHierarchyHDF5(const string& filename,
                      const Vector<DisjointBoxLayout>& a_vectGrids,
                      const Vector<LevelData<FArrayBox>* > & a_vectData,
                      const Vector<string>& a_vectNames,
                      const Box& a_domain,
                      const Real& a_dx,
                      const Real& a_dt,
                      const Real& a_time,
                      const Vector<int>& a_refRatio,
                      const int& a_numLevels)
{
  HDF5Handle handle(filename.c_str(),  HDF5Handle::CREATE);

  WriteAMRHierarchyHDF5(handle, a_vectGrids, a_vectData, a_vectNames,
                        a_domain, a_dx, a_dt, a_time, a_refRatio, a_numLevels);

#ifdef CH_MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif

  handle.close();
}

void
WriteAMRHierarchyHDF5(HDF5Handle& handle,
                      const Vector<DisjointBoxLayout>& a_vectGrids,
                      const Vector<LevelData<FArrayBox>* > & a_vectData,
                      const Vector<string>& a_vectNames,
                      const Box& a_domain,
                      const Real& a_dx,
                      const Real& a_dt,
                      const Real& a_time,
                      const Vector<int>& a_refRatio,
                      const int& a_numLevels)
{
  CH_assert(a_numLevels > 0);
  CH_assert(a_vectData.size()  >= a_numLevels);
  CH_assert(a_refRatio.size() >= a_numLevels-1);

  HDF5HeaderData header;
  int nComp = a_vectNames.size();

  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       = a_numLevels;
  header.m_int ["num_components"]    = nComp;

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      header.m_string[label] = a_vectNames[ivar];
    }
  header.writeToFile(handle);

  Box domainLevel = a_domain;
  Real dtLevel = a_dt;
  Real dxLevel = a_dx;
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int refLevel = 1;
      if(ilev != a_numLevels -1)
          refLevel = a_refRatio[ilev];
      if(ilev != 0)
        {
          domainLevel.refine(a_refRatio[ilev-1]);
          dtLevel /= a_refRatio[ilev-1];
          dxLevel /= a_refRatio[ilev-1];
        }
      CH_assert(a_vectData[ilev] != NULL);
      const LevelData<FArrayBox>& dataLevel = *a_vectData[ilev];
      CH_assert(dataLevel.nComp() == nComp);
      Interval comps(0,nComp-1);
      IntVect ghostVect = a_vectData[0]->ghostVect();
      int eek = writeLevel(handle, ilev, dataLevel,
                           dxLevel, dtLevel,  a_time,
                           domainLevel, refLevel,ghostVect,comps);
      if(eek != 0)
        MayDay::Error("WriteAMRHierarchyHDF5: Error in writeLevel");
    }
}

void
WriteAMRHierarchyHDF5(const string& filename,
                      const Vector<DisjointBoxLayout>& a_vectGrids,
                      const Vector<LevelData<FArrayBox>* > & a_vectData,
                      const Box& a_domain,
                      const Vector<int>& a_refRatio,
                      const int& a_numLevels)
{

  HDF5Handle handle(filename.c_str(),  HDF5Handle::CREATE);
  WriteAMRHierarchyHDF5(handle, a_vectGrids, a_vectData,
                        a_domain, a_refRatio, a_numLevels);

#ifdef CH_MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  handle.close();
}

void
WriteAMRHierarchyHDF5(HDF5Handle& handle,
                      const Vector<DisjointBoxLayout>& a_vectGrids,
                      const Vector<LevelData<FArrayBox>* > & a_vectData,
                      const Box& a_domain,
                      const Vector<int>& a_refRatio,
                      const int& a_numLevels)
{
  CH_assert(a_numLevels > 0);
  CH_assert(a_vectData.size()  >= a_numLevels);
  CH_assert(a_refRatio.size() >= a_numLevels-1);
  Real dxin = 1.0;
  Real dtin = 1.0;
  Real time = 1.0;

  HDF5HeaderData header;
  int nComp = a_vectData[0]->nComp();

  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       = a_numLevels;
  header.m_int ["num_components"]    = nComp;

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      header.m_string[label] = label;
    }
  header.writeToFile(handle);

  Box domainLevel = a_domain;
  Real dtLevel = dtin;
  Real dxLevel = dxin;
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int refLevel = 1;
      if(ilev != a_numLevels -1)
          refLevel = a_refRatio[ilev];
      if(ilev != 0)
        {
          domainLevel.refine(a_refRatio[ilev-1]);
          dtLevel /= a_refRatio[ilev-1];
          dxLevel /= a_refRatio[ilev-1];
        }
      CH_assert(a_vectData[ilev] != NULL);
      const LevelData<FArrayBox>& dataLevel = *a_vectData[ilev];
      CH_assert(dataLevel.nComp() == nComp);
      Interval comps(0,nComp-1);
      IntVect ghostVect = a_vectData[0]->ghostVect();
      int eek = writeLevel(handle, ilev, dataLevel,
                           dxLevel, dtLevel,  time,
                           domainLevel, refLevel,ghostVect,comps);
      if(eek != 0)
        MayDay::Error("WriteAMRHierarchyHDF5: Error in writeLevel");
    }
}

//
/*
\\ Read in hierarchy of amr data in HDF5 format
\\ filename,  == file to output to
\\ a_vectData == data at each level
\\ a_vectNames== names of variables
\\ a_domain == domain at coarsest level
\\ a_dx     == grid spacing at coarsest level
\\ a_dt     == time step at coarsest level
\\ a_time     == time
\\ a_vectRatio == refinement ratio at all levels
\\ (ith entry is refinement ratio between levels i and i + 1)
\\ a_numLevels == number of levels to output

return values:
0: success
-1: bogus number of levels
-2: bogus number of components
-3: error in readlevel
-4: file open failed
*/
int
ReadAMRHierarchyHDF5(const string& filename,
                     Vector<DisjointBoxLayout>& a_vectGrids,
                     Vector<LevelData<FArrayBox>* > & a_vectData,
                     Vector<string>& a_vectNames,
                     Box& a_domain,
                     Real& a_dx,
                     Real& a_dt,
                     Real& a_time,
                     Vector<int>& a_refRatio,
                     int& a_numLevels)
{
  HDF5Handle handle;
  int err = handle.open(filename.c_str(),  HDF5Handle::OPEN_RDONLY);
  if( err < 0) return -4;
  int eekflag = ReadAMRHierarchyHDF5(handle, a_vectGrids, a_vectData,
                                     a_vectNames, a_domain, a_dx, a_dt,
                                     a_time, a_refRatio, a_numLevels);

#ifdef CH_MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  handle.close();

  return (eekflag);
}

int
ReadAMRHierarchyHDF5(HDF5Handle& handle,
                     Vector<DisjointBoxLayout>& a_vectGrids,
                     Vector<LevelData<FArrayBox>* > & a_vectData,
                     Vector<string>& a_vectNames,
                     Box& a_domain,
                     Real& a_dx,
                     Real& a_dt,
                     Real& a_time,
                     Vector<int>& a_refRatio,
                     int& a_numLevels)
{

  HDF5HeaderData header;
  header.readFromFile(handle);

  a_numLevels = header.m_int["num_levels"];
  if(a_numLevels <= 0)
  {
    MayDay::Warning("ReadAMRHierarchyHDF5: Bogus number of levels");
    return (-1);
  }
  a_vectData.resize(a_numLevels);
  a_refRatio.resize(a_numLevels);
  a_vectGrids.resize(a_numLevels);

  int nComp = header.m_int["num_components"];
  if(nComp <= 0)
  {
    MayDay::Warning("ReadAMRHierarchyHDF5: Bogus number of Components");
    return (-2);
  }
  a_vectNames.resize(nComp);

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      a_vectNames[ivar] = header.m_string[label];
    }
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int refLevel = 0;
      Box domainLevel;
      Real dtLevel;
      Real dxLevel;
      a_vectData[ilev] = new LevelData<FArrayBox>();
      int eek = readLevel(handle, ilev, *(a_vectData[ilev]),
                          dxLevel, dtLevel,  a_time,
                          domainLevel, refLevel, Interval(), true);
      if(eek != 0)
      {
        MayDay::Warning("ReadAMRHierarchyHDF5: readLevel failed");
        return (-3);
      }

      const DisjointBoxLayout& dbl = a_vectData[ilev]->getBoxes();
      a_vectGrids[ilev]= dbl;

      if(ilev == 0)
        {
          a_domain = domainLevel;
          a_dt = dtLevel;
          a_dx = dxLevel;
        }
      a_refRatio[ilev] = refLevel;
    }
  return (0);
}

int
ReadAMRHierarchyHDF5(const string& filename,
                     Vector<DisjointBoxLayout>& a_vectGrids,
                     Vector<LevelData<FArrayBox>* > & a_vectData,
                     Box& a_domain,
                     Vector<int>& a_refRatio,
                     int& a_numLevels)
{
  HDF5Handle handle;
  int err = handle.open(filename.c_str(),  HDF5Handle::OPEN_RDONLY);
  if( err < 0) return -4;

  int eekflag = ReadAMRHierarchyHDF5(handle, a_vectGrids, a_vectData,
                                     a_domain, a_refRatio, a_numLevels);

#ifdef CH_MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  handle.close();
  return (eekflag);
}

int
ReadAMRHierarchyHDF5(HDF5Handle& handle,
                     Vector<DisjointBoxLayout>& a_vectGrids,
                     Vector<LevelData<FArrayBox>* > & a_vectData,
                     Box& a_domain,
                     Vector<int>& a_refRatio,
                     int& a_numLevels)
{
  HDF5HeaderData header;
  header.readFromFile(handle);

  a_numLevels = header.m_int["num_levels"];
  if(a_numLevels <= 0)
  {
    MayDay::Warning("ReadAMRHierarchyHDF5: Bogus number of levels");
    return (-1);
  }
  a_vectData.resize(a_numLevels);
  a_refRatio.resize(a_numLevels);
  a_vectGrids.resize(a_numLevels);

  //  int nComp = header.m_int["num_components"];
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int refLevel = 0;
      Box domainLevel;
      Real dtLevel;
      Real dxLevel;
      Real time;
      a_vectData[ilev] = new LevelData<FArrayBox>();
      int eek = readLevel(handle, ilev, *(a_vectData[ilev]),
                          dxLevel, dtLevel,  time,
                          domainLevel, refLevel, Interval(), true);
      if(eek != 0)
      {
        MayDay::Warning("ReadAMRHierarchyHDF5: readLevel failed");
        return (-3);
      }

      const DisjointBoxLayout& dbl = a_vectData[ilev]->getBoxes();
      a_vectGrids[ilev]= dbl;

      if(ilev == 0)
        {
          a_domain = domainLevel;
        }
      a_refRatio[ilev] = refLevel;
    }

  return (0);
}

// put simple debugging functions here, at the end of the hdf5 stuff

void
writeFAB(const FArrayBox* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = "fab.hdf5";
  writeFABname(a_dataPtr, fname);
}

void
viewFAB(const FArrayBox* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = tempnam(NULL,NULL);
  writeFABname(a_dataPtr, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/bin/chombovis %s &",fname);
  system(command);
}

void
browseFAB(const FArrayBox* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = tempnam(NULL,NULL);
  writeFABname(a_dataPtr, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/bin/chombobrowser %s &",fname);
  system(command);
}

void
writeBFR(const BaseFab<Real>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  //copy the BaseFab over to a FArrayBox
  Box fabBox = (*a_dataPtr).box();
  int  ncomp = (*a_dataPtr).nComp();
  FArrayBox fab(fabBox,ncomp);
  fab.copy(*a_dataPtr);

  char* fname = "fab.hdf5";
  writeFABname(&fab, fname);
}

void
viewBFR(const BaseFab<Real>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  //copy the BaseFab over to a FArrayBox
  Box fabBox = (*a_dataPtr).box();
  int  ncomp = (*a_dataPtr).nComp();
  FArrayBox fab(fabBox,ncomp);
  fab.copy(*a_dataPtr);

  char* fname = tempnam(NULL,NULL);
  writeFABname(&fab, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/bin/chombovis %s &",fname);
  system(command);
}

void
writeFABname(const FArrayBox      * a_dataPtr,
             const char           * a_filename,
             const Vector<string> & a_compNames)
{
  if (a_dataPtr == NULL) return;

  const FArrayBox& data = *a_dataPtr;

  HDF5Handle handle(a_filename, HDF5Handle::CREATE);
  HDF5HeaderData header;

  int numlevels= 1;
  int nComp = data.nComp();

  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       = numlevels;
  header.m_int ["num_components"]    = nComp;

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      if( a_compNames.size() > ivar )
        {
          header.m_string[label] = a_compNames[ivar] ;
        }
      else
        {
          header.m_string[label] = label;
        }
    }
  header.writeToFile(handle);

  Box domainLevel = data.box();
  // put bogus numbers here
  Real dtLevel = 1.0;
  Real dxLevel = 1.0;
  Real time = 1.0;

  int refLevel = 1;

  // build bogus DisjointBoxLayout here
  Vector<Box> boxes(1,domainLevel);
  unsigned int myprocID= procID();
  Vector<int> procAssign(1,myprocID);
  DisjointBoxLayout grids(boxes, procAssign);
  LevelData<FArrayBox> ldf(grids, nComp);
  // now copy fab into ldf
  DataIterator dit = ldf.dataIterator();
  ldf[dit()].copy(data);

  int eek = writeLevel(handle, 0, ldf, dxLevel, dtLevel, time,
                       domainLevel, refLevel);
  if (eek != 0)
    MayDay::Error("writeFABname: error in writeLEvel");

  handle.close();
}

void
writeLevel(const LevelData<FArrayBox>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = "LDF.hdf5";
  writeLevelname(a_dataPtr, fname);
}

void
viewLevel(const LevelData<FArrayBox>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = tempnam(NULL,NULL);
  writeLevelname(a_dataPtr, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/bin/chombovis %s &",fname);
  system(command);
}

void
viewLevelNoGhost(const LevelData<FArrayBox>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  const DisjointBoxLayout& grids = a_dataPtr->getBoxes();
  int nVar = a_dataPtr->nComp();

  LevelData<FArrayBox> temp(grids, nVar);
  a_dataPtr->copyTo(a_dataPtr->interval(), temp, temp.interval());

  char* fname = tempnam(NULL,NULL);
  writeLevelname(&temp, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/bin/chombovis %s &",fname);
  system(command);
}

void
writeLevelname(const LevelData<FArrayBox>* a_dataPtr,
               const char*                 a_filename)
{
  if (a_dataPtr == NULL) return;

  const LevelData<FArrayBox>& data = *a_dataPtr;

  HDF5Handle handle(a_filename, HDF5Handle::CREATE);
  HDF5HeaderData header;

  int numlevels = 1;
  int nComp = data.nComp();

  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       = numlevels;
  header.m_int ["num_components"]    = nComp;

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      header.m_string[label] = label;
    }
  header.writeToFile(handle);

  // need to figure out what domain will contain this LevelData
  // This must be LayoutIterator instead of DataIterator because
  // we need domain over boxes in ALL procs.
  const DisjointBoxLayout& levelBoxes = data.getBoxes();
  LayoutIterator lit = levelBoxes.layoutIterator();
  lit.reset();
  Box domain = levelBoxes.get(lit());
  for (lit.reset(); lit.ok(); ++lit)
    {
      const Box thisBox = levelBoxes.get(lit());
      D_TERM(
             if (thisBox.smallEnd(0)<domain.smallEnd(0))
                domain.setSmall(0,thisBox.smallEnd(0)); ,
             if (thisBox.smallEnd(1)<domain.smallEnd(1))
                domain.setSmall(1,thisBox.smallEnd(1)); ,
             if (thisBox.smallEnd(2)<domain.smallEnd(2))
                domain.setSmall(2, thisBox.smallEnd(2)); );

      D_TERM(
             if (thisBox.bigEnd(0)>domain.bigEnd(0))
                domain.setBig(0,thisBox.bigEnd(0)); ,
             if (thisBox.bigEnd(1)>domain.bigEnd(1))
                domain.setBig(1,thisBox.bigEnd(1)); ,
             if (thisBox.bigEnd(2)>domain.bigEnd(2))
                domain.setBig(2, thisBox.bigEnd(2)); );

    } // end loop over boxes on level to determine "domain"

  // put bogus numbers here
  Real dtLevel = 1.0;
  Real dxLevel = 1.0;
  Real time = 1.0;

  int refLevel = 1;

  Interval comps(0,nComp-1);
  IntVect ghostVect = data.ghostVect();
  int eek = writeLevel(handle, 0, data, dxLevel, dtLevel, time,
                       domain, refLevel, ghostVect,comps);

  if (eek != 0)
    MayDay::Error("writeLDFname: error in writeLEvel");

  handle.close();
}

void
writeDBL(const DisjointBoxLayout* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = "DBL.hdf5";
  writeDBLname(a_dataPtr, fname);
}

void
viewDBL(const DisjointBoxLayout* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = tempnam(NULL,NULL);
  writeDBLname(a_dataPtr, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/bin/chombovis %s &",fname);
  system(command);
}

void
writeDBLname(const DisjointBoxLayout* a_dataPtr,
             const char*                 a_filename)
{
  if (a_dataPtr == NULL) return;

  const DisjointBoxLayout& grids = *a_dataPtr;

  // now create a LevelData<FArrayBox> based on the grids
  IntVect ghostVect(IntVect::Zero);
  LevelData<FArrayBox> data(grids, 1, ghostVect);

  // for now, initialize data to 1.
  Real dataVal = 1.0;
  DataIterator dit = data.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      data[dit()].setVal(dataVal);
    }

  writeLevelname(&data, a_filename);
}

void
WritePartialAMRHierarchyHDF5(const string& filename,
                             const Vector<DisjointBoxLayout>& a_vectGrids,
                             const Vector<LevelData<FArrayBox>* > & a_vectData,
                             const Vector<string>& a_vectNames,
                             const Box& a_baseDomain,
                             const Real& a_baseDx,
                             const Real& a_dt,
                             const Real& a_time,
                             const Vector<int>& a_vectRatio,
                             const Interval& a_levels)
{
  int numLevels = a_levels.size();

  // now make new dataholders which only have numLevels entries,
  // and which will move the baseLevel to level 0

  Vector<DisjointBoxLayout> newVectGrids(numLevels);
  Vector<LevelData<FArrayBox> * > newVectData(numLevels);
  Vector<int> newVectRatio(numLevels);

  int leveloffset = a_levels.begin();
  for (int srcLevel = a_levels.begin(); srcLevel <= a_levels.end(); srcLevel++)
    {
      int destLevel = srcLevel-leveloffset;
      newVectGrids[destLevel] = a_vectGrids[srcLevel];
      newVectData[destLevel] = a_vectData[srcLevel];
      newVectRatio[destLevel] = a_vectRatio[srcLevel];
    }

  WriteAMRHierarchyHDF5(filename, newVectGrids, newVectData, a_vectNames,
                        a_baseDomain, a_baseDx, a_dt, a_time,
                        newVectRatio, numLevels);
}

#endif // CH_USE_HDF5

void writeAFabASCII(ostream& os, const FArrayBox& fab)
{
    os << fab.box() << "  " << fab.nComp() << endl;;
    BoxIterator bit(fab.box());
    for(bit.begin(); bit.ok(); bit.next())
    {
        for(int ivar = 0; ivar < fab.nComp(); ivar++)
        {
            os << fab(bit(), ivar) << endl;
        }
    }

}
void readAFabASCII(istream& is, FArrayBox& fab)
{
    Box fabbox;
    int ncomp;
    is >> fabbox >> ncomp;
    fab.resize(fabbox, ncomp);
    BoxIterator bit(fabbox);
    for(bit.begin(); bit.ok(); bit.next())
    {
        for(int ivar = 0; ivar < fab.nComp(); ivar++)
        {
            Real temp;
            is >> temp;
            fab(bit(), ivar) = temp;
        }
    }

}

//this outputs stuff to the amrascii format very slowly
/*
\\ write out hierarchy of amr data in HDF5 format
\\ filename,  == file to output to
\\ a_vectData == data at each level
\\ a_vectNames== names of variables
\\ a_domain == domain at coarsest level
\\ a_dx     == grid spacing at coarsest level
\\ a_dt     == time step at coarsest level
\\ a_time     == time
\\ a_refRatio == refinement ratio at all levels
\\ (ith entry is refinement ratio between levels i and i + 1)
\\ a_numLevel s== number of levels to output
*/
void
WriteAMRHierarchyASCII(const string& filename,
                       const Vector<DisjointBoxLayout>& a_vectGrids,
                       const Vector<LevelData<FArrayBox>* > & a_vectData,
                       const Vector<string>& a_vectNames,
                       const Box& a_domain,
                       const Real& a_dx,
                       const Real& a_dt,
                       const Real& a_time,
                       const Vector<int>& a_refRatio,
                       const int& a_numLevels)
{
  CH_assert(a_numLevels > 0);
  CH_assert(a_vectData.size()  >= a_numLevels);
  CH_assert(a_refRatio.size() >= a_numLevels-1);

  //sanity checks
  int nvar = a_vectData[0]->nComp();
  for (int ivec = 0; ivec < a_vectData.size(); ivec++)
    CH_assert(nvar == a_vectData[ivec]->nComp());
  CH_assert(a_vectNames.size() == nvar);
  for (int ivec = 0; ivec < a_refRatio.size(); ivec++)
    CH_assert(a_refRatio[ivec] > 0);

  fstream os;
  os.open(filename.c_str(), ios::out);

  os <<  nvar           << endl; // Number of states dumped
  for(int iname = 0; iname < a_vectNames.size(); iname++)
    {
      os << a_vectNames[iname]  << endl; // Name of state
    }

  os << SpaceDim     << endl; // Dimension of data
  os << a_numLevels  << endl; // number of levels dumped

  for(int irat = 0; irat < a_numLevels-1; irat++)
    os << a_refRatio[irat] << " " ;          //refinement ratios
  os << endl;
  os << a_domain << endl;                   //coarse level domain
  os << a_dx     << endl;                   //coarse level grid spacing
  os << a_dt     << endl;                   //coarse level time step
  os << a_time   << endl;                   // Simulation time of dump

  for (int ilevel = 0; ilevel < a_numLevels; ilevel++)
    {
      os << ilevel << endl;  //level number
      const LevelData<FArrayBox>& mf_lev = *(a_vectData[ilevel]);
      const DisjointBoxLayout& bxa = a_vectGrids[ilevel];

      // dump actual boxes
      LayoutIterator lit = bxa.layoutIterator();
      os << bxa.size() << endl;              // number of boxes
      for(lit.reset(); lit.ok(); ++lit)
        os << bxa[lit()] << endl;              // For each grid, dump box

      // dump actual data
      DataIterator dit = mf_lev.dataIterator();
      for(dit.reset(); dit.ok(); ++dit)
        {
          const FArrayBox& dat = mf_lev[dit()];
          writeAFabASCII(os, dat);
        }
    }
  os.close();
}

//this reads stuff in  amrascii format very slowly
/*
\\ write out hierarchy of amr data in HDF5 format
\\ filename,  == file to output to
\\ a_vectData == data at each level
\\ a_vectNames== names of variables
\\ a_domain == domain at coarsest level
\\ a_dx     == grid spacing at coarsest level
\\ a_dt     == time step at coarsest level
\\ a_time     == time
\\ a_refRatio == refinement ratio at all levels
\\ (ith entry is refinement ratio between levels i and i + 1)
\\ a_numLevels == number of levels to output

return values:
0 = success
-1 = failure
*/
int
ReadAMRHierarchyASCII(const string& filename,
                      Vector<DisjointBoxLayout>& a_vectGrids,
                      Vector<LevelData<FArrayBox>* > & a_vectData,
                      Vector<string>& a_vectNames,
                      Box& a_domain,
                      Real& a_dx,
                      Real& a_dt,
                      Real& a_time,
                      Vector<int>& a_refRatio,
                      int& a_numLevels,
                      const IntVect& ghostVector)
{
  fstream is;
  if(numProc() > 1)
    {
      MayDay::Warning("ReadAMRHierarchyASCII is written for one processor");
    }
  is.open(filename.c_str(), ios::in);

  int nvar;
  is  >>  nvar;             // number of states
  a_vectNames.resize(nvar);
  for(int iname = 0; iname < a_vectNames.size(); iname++)
    {
      is >> a_vectNames[iname];   // Name of state
    }

  int dimdat;
  is >> dimdat; //dimension of problem
  if(dimdat != SpaceDim)
    {
      MayDay::Warning("ReadAMRHierarchyASCII: dimension of data != dimension of reader");
      return (-1);
    }
  is >> a_numLevels; // number of levels dumped

  a_refRatio.resize(a_numLevels);
  a_vectGrids.resize(a_numLevels);
  a_vectData.resize(a_numLevels);
  for(int irat = 0; irat < a_numLevels-1; irat++)
    is >> a_refRatio[irat];          //refinement ratios

  is >> a_domain;                //coarse level domain
  is >> a_dx;                   //coarse level grid spacing
  is >> a_dt;                   //coarse level time step
  is >> a_time;                 // Simulation time of dump

  for (int ilevel = 0; ilevel < a_numLevels; ilevel++)
    {
      int itemp;
      is  >> itemp;  //level number
      if(itemp != ilevel)
        {
          MayDay::Warning("ReadAMRHierarchyASCII: Input file wrong");
          return (-1);
        }

      DisjointBoxLayout& dbl = a_vectGrids[ilevel];
      int nboxes;
      is >> nboxes;
      for(int ibox = 0; ibox < nboxes; ibox++)
        {
          Box btemp;
          is >> btemp;
          dbl.addBox(btemp, 0);
        }
      dbl.close();
      LevelData<FArrayBox> mf_lev(dbl,nvar,ghostVector);
      DataIterator dit = mf_lev.dataIterator();
      // read actual data
      for(dit.reset(); dit.ok(); ++dit)
        {
          FArrayBox datnoGhost;
          readAFabASCII(is, datnoGhost);
          FArrayBox& dat = mf_lev[dit()];
          dat.copy(datnoGhost);
        }
      a_vectData[ilevel] = new LevelData<FArrayBox>();
      a_vectData[ilevel]->define(mf_lev);
    }
  is.close();

  return (0);
}

// Read in hierarchy of amr data in ASCII format
int
ReadEBAMRASCII(const string& filename,
               Vector<DisjointBoxLayout>& a_vectGrids,
               Vector<LevelData<FArrayBox>* > & a_vectData,
               Vector<string>& a_vectNames,
               Box& a_domain,
               Vector<IntVectSet>& a_coveredCells,
               Vector<IntVectSet>& a_multiValuedCells,
               Vector<int>& a_refRatio,
               int& a_numLevels,
               const IntVect& a_ghostVector)
{
  fstream is;
  if(numProc() > 1)
    {
      MayDay::Warning("ReadAMRHierarchyASCII is written for one processor");
    }
  is.open(filename.c_str(), ios::in);

  int nvar;
  is  >>  nvar;             // number of states
  a_vectNames.resize(nvar);
  for(int iname = 0; iname < a_vectNames.size(); iname++)
    {
      is >> a_vectNames[iname];   // Name of state
    }

  int dimdat;
  is >> dimdat; //dimension of problem
  if(dimdat != SpaceDim)
    {
      MayDay::Warning("ReadAMRHierarchyASCII: dimension of data != dimension of reader");
      return (-1);
    }
  is >> a_numLevels; // number of levels dumped

  a_coveredCells.resize(a_numLevels);
  a_multiValuedCells.resize(a_numLevels);
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int ivecsize;
      is >>  ivecsize; // number of covered boxes
      Box coveredBox;
      IntVectSet& coveredCells = a_coveredCells[ilev];
      coveredCells.makeEmpty();
      for(int ibox = 0; ibox < ivecsize; ibox++)
        {
          is >>  coveredBox; //input each covered box
          CH_assert(!coveredBox.isEmpty());
          coveredCells |= coveredBox;
        }

      IntVectSet& multiIVS = a_multiValuedCells[ilev];
      multiIVS.makeEmpty();
      is >> ivecsize ; //number of multi-valued cells
      for(int ivec = 0; ivec < ivecsize; ivec++)
        {
          IntVect iv;
          is >> iv;
          multiIVS |= iv;
        }
    }

  a_refRatio.resize(a_numLevels);
  a_vectGrids.resize(a_numLevels);
  a_vectData.resize(a_numLevels);
  for(int irat = 0; irat < a_numLevels-1; irat++)
    is >> a_refRatio[irat];          //refinement ratios

  is >> a_domain;                //coarse level domain
  for (int ilevel = 0; ilevel < a_numLevels; ilevel++)
    {
      int itemp;
      is  >> itemp;  //level number
      if(itemp != ilevel)
        {
          MayDay::Warning("ReadAMRHierarchyASCII: Input file wrong");
          return (-1);
        }

      DisjointBoxLayout& dbl = a_vectGrids[ilevel];
      int nboxes;
      is >> nboxes;
      for(int ibox = 0; ibox < nboxes; ibox++)
        {
          Box btemp;
          is >> btemp;
          dbl.addBox(btemp, 0);
        }
      dbl.close();
      LevelData<FArrayBox> mf_lev(dbl,nvar,a_ghostVector);
      DataIterator dit = mf_lev.dataIterator();
      // read actual data
      for(dit.reset(); dit.ok(); ++dit)
        {
          FArrayBox datnoGhost;
          //fab was written out as a farraybox.  no need
          //for special read function
          readAFabASCII(is, datnoGhost);
          FArrayBox& dat = mf_lev[dit()];
          dat.copy(datnoGhost);
        }
      a_vectData[ilevel] = new LevelData<FArrayBox>();
      a_vectData[ilevel]->define(mf_lev);
    }
  is.close();
  return (0);
}
