/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _CFIVS_H_
#define _CFIVS_H_

#include <iostream>
#include <cmath>
#include "SPACE.H"
#include <cstdlib>

#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "LoHiSide.H"
#include "DisjointBoxLayout.H"
#include "IntVectSet.H"
#include "ProblemDomain.H"

/// Internal Chombo class to get fine intvectset at CF interface
/**
    Class to get fine intvectset at cf interface for one side of a given
    fine-level Box.
    This class should be considered internal to AMRSolver and should
    not be considered part of the Chombo API.
*/
class CFIVS
{
public:
  /// Null constructor
  CFIVS();

  /// Destructor
  ~CFIVS();

  /// Deprecated constructor (The ProblemDomain class has replaced the Box domain)
  /**
     Replaces Box domain with a ProblemDomain and calls define fn.
  */
  CFIVS(const Box&               a_domain,
        const Box&               a_boxIn,
        const DisjointBoxLayout& a_fineBoxes,
        int                      a_direction,
        Side::LoHiSide           a_hiorlo);

  /// Full constructor
  /**
     Calls corresponding define function.
   */
  CFIVS(const ProblemDomain&     a_domain,
        const Box&               a_boxIn,
        const DisjointBoxLayout& a_fineBoxes,
        int                      a_direction,
        Side::LoHiSide           a_hiorlo);

  /// Deprecated define function
  /**
      Replaces Box domain with a ProblemDomain and calls define fn.
  */
  void define(const Box&               a_domain,
              const Box&               a_boxIn,
              const DisjointBoxLayout& a_fineBoxes,
              int                      a_direction,
              Side::LoHiSide           a_hiorlo);

  /// Full define function
  /**
   */
  void define (const ProblemDomain&     a_domain,    /// Problem domain at the fine level
               const Box&               a_boxIn,     /// fine-level grid box
               const DisjointBoxLayout& a_fineBoxes, /// fine-level grids
               int                      a_direction, /// direction of the face of interest
               Side::LoHiSide           a_hiorlo);   /// is the face on the hi or lo side of the box?

  /// Returns "true" if the object has been defined
  bool isDefined() const;

  /// Returns "true" if this CF IntVectSet can be represented as a simple Box.
  /**
      Returns 'true' if this coarse-fine IntVectSet can be represented
      as just a Box, ie.  a user can perform a dense computation instead of
      a pointwise calculation using IVSIterator.
  */
  bool isPacked() const;

  /// Returns packed box for when isPacked() is true.
  /** if isPacked() returns 'true', then the box returned by this function
      is a suitable substitute for this IntVectSet
  */
  const Box& packedBox() const;

  /// Get fine intvects which need to be interpolated.
  /**
      This will be empty if isEmpty() returns true
  */
  const IntVectSet& getFineIVS() const ;

  static long long m_packCount, m_sparseCount;


protected:
  // For internal use
  void  setDefaultValues();

  // The following variables are mainly for self-identification

  /**
     Fine intvects which need to be interpolated
  */
  IntVectSet m_fiinterpIVS;

  /**
     Can the coarse-fine IntVectSet be represented as a Box?
  */
  bool m_packed;

  /**
      If m_packed is true, the Box with which the coarse-fine
      IntVectSet can be represented
  */
  Box  m_packedBox;

  /**
     Has this object been defined?
  */
  bool m_isDefined;

private:
  void operator= (const CFIVS& a_levcfsIn)
  {
  }

  CFIVS(const CFIVS& a_levcfsIn)
  {
  }
};

inline bool CFIVS::isPacked() const
{
  return m_packed;
}

inline const Box& CFIVS::packedBox() const
{
  return m_packedBox;
}

#endif
