/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "LayoutIterator.H"
#include "DataIterator.H"

#include "CFIVS.H"

bool CFIVS::isDefined() const
{
  return m_isDefined;
}

const IntVectSet& CFIVS::getFineIVS() const
{
 CH_assert(m_isDefined);

  return m_fiinterpIVS;
}

void CFIVS::setDefaultValues()
{
  m_fiinterpIVS.define();
  m_packed = false;

  m_isDefined = false;
}

CFIVS::CFIVS()
{
  setDefaultValues();
}

CFIVS::~CFIVS()
{
  setDefaultValues();
}

CFIVS::CFIVS(const Box&               a_domain,
             const Box&               a_boxIn,
             const DisjointBoxLayout& a_fineBoxes,
             int                      a_direction,
             Side::LoHiSide           a_hiorlo)
{
  setDefaultValues();
  ProblemDomain probdomain(a_domain);

  define(probdomain, a_boxIn, a_fineBoxes, a_direction, a_hiorlo);

}

CFIVS::CFIVS(const ProblemDomain&     a_domain,
             const Box&               a_boxIn,
             const DisjointBoxLayout& a_fineBoxes,
             int                      a_direction,
             Side::LoHiSide           a_hiorlo)
{
  setDefaultValues();

  define(a_domain, a_boxIn, a_fineBoxes, a_direction, a_hiorlo);

}

void CFIVS::define(const Box&               a_domain,
                   const Box&               a_boxIn,
                   const DisjointBoxLayout& a_fineBoxes,
                   int                      a_direction,
                   Side::LoHiSide           a_hiorlo)
{
  ProblemDomain probdomain(a_domain);

  define(probdomain, a_boxIn, a_fineBoxes, a_direction, a_hiorlo);
}

void CFIVS::define(const ProblemDomain&     a_domain,
                   const Box&               a_boxIn,
                   const DisjointBoxLayout& a_fineBoxes,
                   int                      a_direction,
                   Side::LoHiSide           a_hiorlo)
{
  m_isDefined = true;

 CH_assert(a_direction >= 0);
 CH_assert(a_direction < SpaceDim);
 CH_assert(!a_domain.isEmpty());
 CH_assert(a_domain.contains(a_boxIn));
 CH_assert(a_fineBoxes.checkPeriodic(a_domain));
 CH_assert((a_hiorlo == Side::Lo) || (a_hiorlo == Side::Hi));

  // shift direction
  int hilo = sign(a_hiorlo);

  // create fine stencil
  Box finebox = a_boxIn;
  Box edgebox;

 CH_assert((hilo ==1) || (hilo == -1));

  if (hilo == -1)
    {
      edgebox = adjCellLo(finebox,a_direction,1);
    }
  else
    {
      edgebox = adjCellHi(finebox,a_direction,1);
    }

  edgebox &= a_domain;

  if (!edgebox.isEmpty())
    {
      m_fiinterpIVS.define(edgebox);

      LayoutIterator lit = a_fineBoxes.layoutIterator();
      Box periodicTestBox(a_domain.domainBox());

      if (a_domain.isPeriodic())
        {
          for (int idir = 0; idir < SpaceDim; idir++)
            {
              if (a_domain.isPeriodic(idir))
                periodicTestBox.grow(idir,-1);
            }
        }

      for (lit.reset(); lit.ok(); ++lit)
        {
          m_fiinterpIVS -= a_fineBoxes[lit()];

          // only do this IF we're periodic _and_ both boxes
          // adjoin the domain box boundary somewhere
          if (a_domain.isPeriodic() && !periodicTestBox.contains(edgebox)
              && !periodicTestBox.contains(a_fineBoxes[lit()]))
            {
              ShiftIterator shiftIt = a_domain.shiftIterator();
              IntVect shiftMult = a_domain.domainBox().size();
              Box shiftedBox = a_fineBoxes[lit()];

              for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                {
                  IntVect shiftVect(shiftMult*shiftIt());

                  shiftedBox.shift(shiftVect);
                  m_fiinterpIVS -= shiftedBox;
                  shiftedBox.shift(-shiftVect);
                }
            }
        }
    }

  m_fiinterpIVS.compact();

  if (m_fiinterpIVS.isEmpty())
    {
      m_packed = false;
    }
  else
  {
    Box min = m_fiinterpIVS.minBox();

    if (min.numPts() == m_fiinterpIVS.numPts())
    {
      m_packCount++;
      m_packed = true;
      m_packedBox = min;
    }
    else
    {
      m_sparseCount++;
      m_packed = false;
    }
  }
}

long long CFIVS::m_packCount = 0;

long long CFIVS::m_sparseCount = 0;
