/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _CFSTENCIL_H_
#define _CFSTENCIL_H_

#include <iostream>
#include <math.h>
#include "SPACE.H"
#include <stdlib.h>
#include "DisjointBoxLayout.H"
#include "Box.H"
#include "LoHiSide.H"
#include "IntVectSet.H"

/// Class to encapsulate coarse-fine information for AMR operations.
/**
    CFStencil is a wrapper around a bunch of things
    that allow one to do coarse-fine interpolation.
    This class should be considered internal to QuadCFInterp
    and should not be considered part of the Chombo API.
  */
class CFStencil
{

private:

protected:
  //for internal use
  void  setDefaultValues();

  //the following variables are mainly for self-identification

  // direction normal to interface
  int m_direction;

  // interface is on high or low side of grid
  Side::LoHiSide m_hiorlo;

  //
  DataIndex m_dataIndex;

  //the following variables are to be used by others

  /* fine intvects which need to be interpolated */
  IntVectSet m_fineIVS;

  /* coar intvects where slopes can be computed
      == coarsened version of fiinterp ivs*/
  IntVectSet m_coarIVS;

  bool m_isDefined;

public:
  /// {\bf Access functions}

  ///
  /** returns true if CFStencil is defined */
  bool isDefined() const;

  ///
  /**
     are there any interpolation points?  \\
     returns false if so.*/
  bool isEmpty() const;

  ///
  /** get fine intvects which need to be interpolated  \\
      This will be empty if isEmpty() returns true*/
  const IntVectSet& getFineIVS() const ;

  ///
  /** get coarse intvects that underly fiinterpivs.
      This will be empty if isEmpty() returns true
      this is the same as getCoarInterpIVS
  */
  const IntVectSet& getCoarIVS() const ;

  ///
  CFStencil& operator= (const CFStencil& cfs_in);

  ///
  CFStencil(const CFStencil& cfs_in);

  /*dummy constructor so i can use this with
     LocalBoxLayoutData
   */
  CFStencil(const Box& a_b, int a_nComp);

  void define(const Box& a_b, int a_nComp);

  ///
  CFStencil();

  ///
  ~CFStencil();

  ///
  CFStencil(
            const Box& FineDomain,
            const Box& grid,
            const DisjointBoxLayout& Levboxes,
            const DisjointBoxLayout& LevCBoxes,
            int RefRatio,
            int Direction,
            Side::LoHiSide hiorlo);

  ///
  CFStencil(
            const ProblemDomain& FineDomain,
            const Box& grid,
            const DisjointBoxLayout& Levboxes,
            const DisjointBoxLayout& LevCBoxes,
            int RefRatio,
            int Direction,
            Side::LoHiSide hiorlo);

  ///
  void   define(
                const Box& FineDomain,
                const Box& grid,
                const DisjointBoxLayout& fine_boxes,
                const DisjointBoxLayout& coar_boxes,
                int Refratio,
                int Direction,
                Side::LoHiSide hiorlo);

  ///
  void   define(
                const ProblemDomain& FineDomain,
                const Box& grid,
                const DisjointBoxLayout& fine_boxes,
                const DisjointBoxLayout& coar_boxes,
                int Refratio,
                int Direction,
                Side::LoHiSide hiorlo);

  ///
  void clear();

};
#endif
