/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "BaseFab.H"
#include "REAL.H"
#include "FArrayBox.H"
#include "DataIterator.H"
#include "LayoutIterator.H"
#include "parstream.H"
#include "AverageF_F.H"

#include "CoarseAverage.H"

CoarseAverage::CoarseAverage()
  :
  is_defined(false)
{
}

CoarseAverage::~CoarseAverage()
{
}

CoarseAverage::CoarseAverage(const DisjointBoxLayout& a_fine_domain,
                             int a_numcomps,
                             int a_ref_ratio)
  :
  is_defined(false), m_is_copier_defined(false)
{
  define(a_fine_domain, a_numcomps, a_ref_ratio);
}

CoarseAverage::CoarseAverage(const DisjointBoxLayout& a_fine_domain,
                             const DisjointBoxLayout& a_crse_domain,
                             int a_numcomps,
                             int a_ref_ratio)
  :
  is_defined(false), m_is_copier_defined(false)
{
  define(a_fine_domain, a_crse_domain, a_numcomps, a_ref_ratio);
}

void
CoarseAverage::define(const DisjointBoxLayout& a_fine_domain,
                      int a_numcomps,
                      int a_ref_ratio)
{
  m_ref_ratio = a_ref_ratio;
  DisjointBoxLayout coarsened_fine_domain;
  coarsen (coarsened_fine_domain, a_fine_domain, m_ref_ratio);
  m_coarsened_fine_data.define (coarsened_fine_domain, a_numcomps);

  is_defined = true;
}

void
CoarseAverage::define(const DisjointBoxLayout& a_fine_domain,
                      const DisjointBoxLayout& a_crse_domain,
                      int a_numcomps,
                      int a_ref_ratio)
{
  m_ref_ratio = a_ref_ratio;
  DisjointBoxLayout coarsened_fine_domain;
  coarsen (coarsened_fine_domain, a_fine_domain, m_ref_ratio);
  m_coarsened_fine_data.define (coarsened_fine_domain, a_numcomps);

  // also can pre-define copier here
  // note that since CoarseAverage only operates on interior
  // data, there is no need to worry about whether the domain
  // is periodic.
  m_copier.define(coarsened_fine_domain, a_crse_domain);
  m_is_copier_defined = true;

  is_defined = true;
}

bool
CoarseAverage::isDefined() const
{
  return ( is_defined );
}

void
CoarseAverage::averageToCoarse(LevelData<FArrayBox>& a_coarse_data,
                               const LevelData<FArrayBox>& a_fine_data)
{
 CH_assert(is_defined);
  // it would be nice if this could check for validity of a_fine_data.
  // this could be done with a redundant DisjointBoxLayout
  DataIterator dit = a_fine_data.boxLayout().dataIterator();
  for (dit.begin(); dit.ok(); ++dit)

    {
      BaseFab<Real>& coarsened_fine = m_coarsened_fine_data[dit()];
      const BaseFab<Real>& fine = a_fine_data[dit()];
      // coarsenGridData coarsens from the entire fine grid onto the entire
      // coarse grid.
      averageGridData(coarsened_fine,
                      fine,
                      m_ref_ratio);
    }

  if (m_is_copier_defined)
    {
      // we can use the pre-defined copier to make things faster
      m_coarsened_fine_data.copyTo(m_coarsened_fine_data.interval(),
                                   a_coarse_data,
                                   a_coarse_data.interval(),
                                   m_copier);
    }
  else
    {
      m_coarsened_fine_data.copyTo(m_coarsened_fine_data.interval(),
                                   a_coarse_data,
                                   a_coarse_data.interval() );
    }
}

void
CoarseAverage::averageGridData(BaseFab<Real>& a_coarse,
                               const BaseFab<Real>& a_fine,
                               int a_ref_ratio)
  const
{
  const Box& b = a_coarse.box();
  Box refbox(IntVect::Zero,
             (a_ref_ratio-1)*IntVect::Unit);
  FORT_AVERAGE( CHF_FRA(a_coarse),
                CHF_CONST_FRA(a_fine),
                CHF_BOX(b),
                CHF_CONST_INT(a_ref_ratio),
                CHF_BOX(refbox)
                );
}
