/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "DerivStencil.H"

DerivStencil::DerivStencil()
{
    clear();
}
void
DerivStencil::define()
{
    clear();
}
DerivStencil::DerivStencil(const DerivStencil& a_dsin)
{
    clear();
    *this = a_dsin;
}

///make derivstencil empty
void
DerivStencil::clear()
{
    m_vectIV.resize(0);
    m_vectWgt.resize(0);
    isdefined = true;
}
DerivStencil::~DerivStencil()
{
    clear();
}

///return length of vectors
int
DerivStencil::size() const
{
    CH_assert(isdefined);
    CH_assert(m_vectWgt.size() == m_vectIV.size());
    return(m_vectIV.size());
}

///get iv at ivec
const IntVect&
DerivStencil::getIndex(int a_ivec) const
{
    CH_assert(isdefined);
    CH_assert(a_ivec >= 0);
    CH_assert(a_ivec < m_vectIV.size());
    CH_assert(m_vectWgt.size() == m_vectIV.size());
    return(m_vectIV[a_ivec]);
}

///get weight at ivec
const Real&
DerivStencil::getWeight(int a_ivec) const
{
    CH_assert(isdefined);
    CH_assert(a_ivec >= 0);
    CH_assert(a_ivec < m_vectIV.size());
    CH_assert(m_vectIV.size()==m_vectWgt.size());
    return(m_vectWgt[a_ivec]);
}

/**
    add another set if the BAI is not in the
    stencil already.  \\
    **Add the weight to the existing weight otherwise**
    */
void
DerivStencil::accumulate(
    const IntVect& a_iv,
    Real a_weight
    )
{
    CH_assert(isdefined);
    CH_assert(m_vectIV.size()==m_vectWgt.size());
    bool isalreadythere = false;
    int ithere= -1;
    for(int ivect = 0; ivect < m_vectIV.size(); ivect++)
    {
        if(a_iv == m_vectIV[ivect])
        {
            isalreadythere = true;
            ithere = ivect;
            break;
        }
    }
    if(!isalreadythere)
    {
        m_vectWgt.push_back(a_weight);
        m_vectIV.push_back(a_iv);
    }
    else
    {
        m_vectWgt[ithere] += a_weight;
    }
}
const DerivStencil&
DerivStencil::operator=(const DerivStencil& a_dsin)
{
    CH_assert(isdefined);
    if(&a_dsin != this)
    {
        m_vectIV = a_dsin.m_vectIV;
        m_vectWgt = a_dsin.m_vectWgt;
    }
    return *this;
}

const DerivStencil&
DerivStencil::operator+=(Real a_facin)
{
    CH_assert(isdefined);
    CH_assert(m_vectIV.size()==m_vectWgt.size());
    for(int ivec = 0; ivec < m_vectWgt.size(); ivec++)
    {
        m_vectWgt[ivec] += a_facin;
    }
    return *this;
}

const DerivStencil&
DerivStencil::operator-=(Real a_facin)
{
    CH_assert(isdefined);
    CH_assert(m_vectIV.size()==m_vectWgt.size());
    for(int ivec = 0; ivec < m_vectWgt.size(); ivec++)
    {
        m_vectWgt[ivec] -= a_facin;
    }
    return *this;
}

const DerivStencil&
DerivStencil::operator*=(Real a_facin)
{
    CH_assert(isdefined);
    CH_assert(m_vectIV.size()==m_vectWgt.size());
    for(int ivec = 0; ivec < m_vectWgt.size(); ivec++)
    {
        m_vectWgt[ivec] *= a_facin;
    }
    return *this;
}

const DerivStencil&
DerivStencil::operator/=(Real a_facin)
{
    CH_assert(isdefined);
    CH_assert(m_vectIV.size()==m_vectWgt.size());
    for(int ivec = 0; ivec < m_vectWgt.size(); ivec++)
    {
        m_vectWgt[ivec] /= a_facin;
    }
    return *this;
}
