/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _QUADCFSTENCIL_H_
#define _QUADCFSTENCIL_H_

#include <iostream>
#include <math.h>
#include "SPACE.H"
#include <stdlib.h>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "FArrayBox.H"
#include "IntVectSet.H"
#include "CFStencil.H"
#include "DerivStencil.H"

/// class to encapsulate CF info for quadratic interpolation
/**
    QuadCFStencil is a wrapper around a bunch of things
    that allow LevelOp to do coarse-fine interpolation
    on a particular face of a grid.
  */
class QuadCFStencil
{

public:
  /// {\bf Derivative functions}

  ///
  /**
     compute second derivative in devdir_a direction
     at coarse point a_ivin
     Uses centered finite diff approximation if
     ivs_standard_m.contains(a_ivin ).
     Otherwise, it uses the stencil from second_m
     Asserts that a_derivdir != direction of face
  */
  Real computeSecondDerivative(
                               const BaseFab<Real> & a_phic,
                               int a_derivdir,
                               int a_ivar,
                               const IntVect& a_ivin,
                               Real a_dx) const;

  ///
  /**
     compute first derivative in devdir_a direction
     at coarse point a_ivin
     Uses centered finite diff approximation if
     ivs_standard_m.contains(a_ivin ).
     Otherwise, it uses the stencil from firstd_m
     Asserts that a_derivdir != direction of face
  */
  Real computeFirstDerivative(
                              const BaseFab<Real> & a_phic,
                              int a_derivdir,
                              int a_ivar,
                              const IntVect& a_ivin,
                              Real a_dx) const;

  ///
  /**
     compute mixed derivative (direction is unambiguous.
     x and y are the two directions tangential to face)
     at coarse point a_ivin. \\
     In two dimensions, always returns 0. \\
     Uses centered finite diff approximation if
     ivs_standard_m.contains(a_ivin ).
     It uses the stencil from mixed_sten_m.\\
     Returns 0 if SpaceDim != 3
  */
  Real computeMixedDerivative(
                              const BaseFab<Real> & a_phic,
                              int a_ivar,
                              const IntVect& a_ivin,
                              Real a_dx) const;

  ///
  QuadCFStencil();

  ///
  ~QuadCFStencil();

  ///
  QuadCFStencil(
                const Box& a_fine_domain,
                const Box& a_grid,
                const DisjointBoxLayout& a_fineBoxes,
                const DisjointBoxLayout& a_coarBoxes,
                int a_refRatio,
                int a_direction,
                Side::LoHiSide a_hiorlo);

  ///
  QuadCFStencil(
                const ProblemDomain& a_fine_domain,
                const Box& a_grid,
                const DisjointBoxLayout& a_fineBoxes,
                const DisjointBoxLayout& a_coarBoxes,
                int a_refRatio,
                int a_direction,
                Side::LoHiSide a_hiorlo);

  ///
  void   define(
                const Box& a_fine_domain,
                const Box& a_grid,
                const DisjointBoxLayout& a_fineBoxes,
                const DisjointBoxLayout& a_coarBoxes,
                int a_refRatio,
                int a_direction,
                Side::LoHiSide a_hiorlo);

  ///
  void   define(
                const ProblemDomain& a_fine_domain,
                const Box& a_grid,
                const DisjointBoxLayout& a_fineBoxes,
                const DisjointBoxLayout& a_coarBoxes,
                int a_refRatio,
                int a_direction,
                Side::LoHiSide a_hiorlo);

  /// {\bf Access functions}

  ///
  /**
     has full define function been called?  return true if so
  */
  bool isDefined() const;

  ///
  /**
     are there any interpolation points?  \\
     returns false if so.*/
  bool isEmpty() const
  {
    CH_assert(m_baseCFS.isDefined());
    return m_baseCFS.isEmpty();
  }

  ///
  /** get fine intvects which need to be interpolated  \\
      This will be empty if isEmpty() returns true*/
  const IntVectSet& getFineIVS() const
  {
    CH_assert(m_baseCFS.isDefined());
    return m_baseCFS.getFineIVS();
  }

  ///
  /** get coarse intvects that underly fiinterpivs.
      This will be empty if isEmpty() returns true*/
  const IntVectSet& getCoarIVS() const
  {
    CH_assert(m_baseCFS.isDefined());
    return m_baseCFS.getCoarIVS();
  }


protected:

  //has define fcn been called?
  bool m_isDefined;

  //direction of face
  int m_direction;

  ///
  /**
     set of all points which have ALL their
     derivatives taken normally.  This means
     that no stencil either leaves the domain or
     intersects the next finer level.
  */
  IntVectSet m_ivsStandard;

  ///
  /**
     set of all points which have at least
     one derivative shifted
  */
  IntVectSet m_ivsQuadd;

  /**
     Derivative stencils.  Only defined at intvects where there
     is at least one non-standard derivative to take
     (ie. at all points of ivs_Quadd)
  */

  ///
  /**
     stencil information for mixed derivatives
     undefined if SpaceDim != 3
  */
  BaseFab<DerivStencil> m_mixedSten;

  ///
  /**
     stencil information for unmixed derivatives
     (second derivatives)
  */
  BaseFab<DerivStencil> m_second[SpaceDim];

  ///
  /**
     stencil information for unmixed derivatives
     (first derivatives)
  */
  BaseFab<DerivStencil> m_firstD[SpaceDim];

  ///
  /**
     whether to drop order at any
     one-sided point
  */
  BaseFab<bool> m_dropOrd;

  ///
  CFStencil m_baseCFS;


private:
  // internally useful but not for public consumption

  //
  void  setDefaultValues();

  /*
    add fab to the stencil.  For internal use only.
  */
  void addFabToSten(const BaseFab<Real> & fabin_a,
                    DerivStencil&  sten_a);

  // there is no assignment operator for this class
  void operator= (const QuadCFStencil& cfs_in)
  {
  }

  // there is no copy constructor for this class
  QuadCFStencil(const QuadCFStencil& cfs_in)
  {
  }

};

#endif
