/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _TENSORFINESTENCILSET_H_
#define _TENSORFINESTENCILSET_H_

#include "SPACE.H"
#include <stdlib.h>
#include "IntVectSet.H"
#include "ProblemDomain.H"
#include "QuadCFStencil.H"

/// Class to encapsulate fine-level tangential gradient stencil computation
/** This is a helper class which essentially wraps three IntVectSets which
    define which type of stencil to use in computing fine-level tangential 
    gradients.  This is only relevant in the presence of a physical boundary
    (since the only place where the fine-level stencil is shifted is at a
    physical boundary; in a periodic world, all differences are centered.)
*/
class TensorFineStencilSet
{
public:
  /// constructors, destructors, defines
  
  /// default constructor -- leaves class in undefined state
  TensorFineStencilSet();

  /// full constructor -- defines stencil IntVectSets
  TensorFineStencilSet(const IntVectSet& a_fineIVS,
                       const ProblemDomain& a_domain,
                       int a_normalDir);

  /// destructor
  ~TensorFineStencilSet();

  /// define function
  void define(const IntVectSet& a_fineIVS,
              const ProblemDomain& a_domain,
              int a_dir);

  /// accessor 
  /** accessor -- returns cells where fine-level tangential gradient
      in a_tanDir direction will use a centered-difference stencil.  If
      a_tanDir is normal to the face (i.e, isn't a tangential direction 
      after all), will return an empty IntVectSet.
  */
  const IntVectSet& getCenteredStencilSet(int a_tanDir) const;

  /// accessor
  /** accessor -- returns cells where fine-level tangential gradient
      in a_tanDir will use a forward-difference stencil.  If a_tanDir 
      is normal to the face (i.e. it isn't a tangential direction 
      after all), will return an empty IntVectSet.
  */
  const IntVectSet& getForwardStencilSet(int a_tanDir) const;

  /// accessor
  /** accessor -- returns cells where fine-level tangential gradient
      in a_tanDir will use a backward-difference stencil.  If a_tanDir
      is normal to the face, will return an empty IntVectSet.
  */
  const IntVectSet& getBackwardStencilSet(int a_tanDir) const;

  /// has this object been defined?
  bool isDefined() const;

  /// does this contain any non-empty IntVectSets?
  bool isEmpty() const;

protected:
  ///
  bool m_isDefined;
  
  ///
  bool m_isEmpty;

  ///
  int m_normalDir;

  ///
  int m_tangentialDirections[SpaceDim-1];

  /// cells which will use centered diffs
  IntVectSet m_centeredSet[SpaceDim-1];
  
  /// cells which will use forward diffs
  IntVectSet m_forwardSet[SpaceDim-1];
  
  /// cells which will use backward diffs
  IntVectSet m_backwardSet[SpaceDim-1];

};

#endif

