/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "TensorFineStencilSet.H"

TensorFineStencilSet::TensorFineStencilSet()
  : m_isDefined(false), m_isEmpty(true)
{
}

TensorFineStencilSet::TensorFineStencilSet(const IntVectSet& a_fineIVS,
                                           const ProblemDomain& a_domain,
                                           int a_normalDir) 
  : m_isDefined(true), m_isEmpty(false)
{
  define(a_fineIVS, a_domain, a_normalDir);
}



TensorFineStencilSet::~TensorFineStencilSet()
{
}

void
TensorFineStencilSet::define(const IntVectSet& a_fineIVS,
                             const ProblemDomain& a_domain,
                             int a_normalDir)
{
 CH_assert (!a_domain.isEmpty());
  
  m_normalDir = a_normalDir;

  // set up indexing
  int index = 0;
  for (int tanDir=0; tanDir<SpaceDim; tanDir++)
    {
      if (tanDir != a_normalDir)
        {
          m_tangentialDirections[index] = tanDir;
          index++;
        }
    }

  

  // are there any cells to deal with here?
  if (!a_fineIVS.isEmpty())
    {
      m_isEmpty = false;
      
      for (int tanDirIndex=0; tanDirIndex<SpaceDim-1; tanDirIndex++) 
        {
          int tanDir = m_tangentialDirections[tanDirIndex];
          m_centeredSet[tanDirIndex] = a_fineIVS;
          // only potentially need to modify anything if domain isn't periodic
          // if periodic in tanDir, then use centered diffs everywhere
          if (!a_domain.isPeriodic(tanDir))
            {
              // this set of operations results in the box over which 
              // the centered difference stencil can be used (shrunk 
              // by one in non-periodic direction)              
              Box centeredStencilBox = a_domain.domainBox();
              centeredStencilBox.grow(tanDir, 1);
              centeredStencilBox &= a_domain;
              centeredStencilBox.grow(tanDir, -1);
              
              m_centeredSet[tanDirIndex] &= centeredStencilBox;
              
              Box forwardDiffBox = adjCellLo(a_domain.domainBox(), tanDir);
              forwardDiffBox.shift(tanDir,1);
              m_forwardSet[tanDirIndex] = a_fineIVS;
              m_forwardSet[tanDirIndex] &= forwardDiffBox;
              
              Box backwardDiffBox = adjCellHi(a_domain.domainBox(),tanDir);
              backwardDiffBox.shift(tanDir,-1);
              m_backwardSet[tanDirIndex] = a_fineIVS;
              m_backwardSet[tanDirIndex] &= backwardDiffBox;
            } // end modifications if domain is not periodic
        } // end loop over tangential directions
    
    } // end if there is anything to do here
  
  m_isDefined = true;

}



const IntVectSet &
TensorFineStencilSet::getCenteredStencilSet(int a_tanDir) const
{
 CH_assert (isDefined());
 CH_assert (a_tanDir != m_normalDir);

  int index = -1;

  // probably need to re-think this indexing strategy... (kinda clunky)
  for (int comp=0; comp<SpaceDim-1; comp++)
    if (m_tangentialDirections[comp] == a_tanDir) index = comp;
  
  return m_centeredSet[index];
}

const IntVectSet&
TensorFineStencilSet::getBackwardStencilSet(int a_tanDir) const
{
 CH_assert (isDefined());
 CH_assert (a_tanDir != m_normalDir);

  int index = -1;

  // probably need to re-think this indexing strategy... (kinda clunky)
  for (int comp=0; comp<SpaceDim-1; comp++)
    if (m_tangentialDirections[comp] == a_tanDir) index = comp;
  
  return m_backwardSet[index];
}

const IntVectSet&
TensorFineStencilSet::getForwardStencilSet(int a_tanDir) const
{
 CH_assert (isDefined());
 CH_assert (a_tanDir != m_normalDir);

  int index = -1;

  // probably need to re-think this indexing strategy... (kinda clunky)
  for (int comp=0; comp<SpaceDim-1; comp++)
    if (m_tangentialDirections[comp] == a_tanDir) index = comp;
  
  return m_forwardSet[index];
}

bool
TensorFineStencilSet::isDefined() const
{
  return m_isDefined;
}


bool
TensorFineStencilSet::isEmpty() const
{
  return m_isEmpty;
}
