/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cmath>
using std::pow;
using std::sqrt;
#ifdef CH_MPI
#include <mpi.h>
#endif

#include "LayoutIterator.H"
#include "BoxLayoutData.H"

#include "computeNorm.H"

Real computeNorm(const Vector<LevelData<FArrayBox>* >& a_phi,
                 const Vector<int>&                    a_nRefFine,
                 const Real                            a_dxCrse,
                 const Interval                        a_comps,
                 const int                             a_p,
                 const int                             a_lBase)
{
  Real dxLevel = a_dxCrse;
  int numLevels = a_phi.size();
  Real norm;
  Real normLevel;

  // it is often the case that while a_phi has many possible
  // levels, only a subset of them are defined -- check that
  // just to be sure
  if (a_phi[numLevels-1] == NULL)
  {
    int lev = numLevels-1;
    while (a_phi[lev] == NULL)
    {
      lev--;
    }
    numLevels = lev+1;
  }

  norm = 0.0;

  // loop over levels
  for (int lev = a_lBase; lev < numLevels; lev++)
  {
    //in case there are extra levels which are not defined
    if (a_phi[lev] != NULL)
    {
      CH_assert(a_phi[lev]->isDefined());

      LevelData<FArrayBox>& thisPhi = *(a_phi[lev]);
      const DisjointBoxLayout* finerGridsPtr = NULL;

      if (lev < numLevels-1)
      {
        finerGridsPtr = &(a_phi[lev+1]->getBoxes());
      }

      normLevel = computeNorm(thisPhi, finerGridsPtr, a_nRefFine[lev],
                              dxLevel, a_comps, a_p);

      if (a_p == 2)
      {
        normLevel = normLevel*normLevel;
      }

      if (a_p != 0)
      {
        norm += normLevel;
      }
      else if (normLevel > norm)
      {
        norm = normLevel;
      }
    }

    // update  dxLevel
    dxLevel = dxLevel/a_nRefFine[lev];

  }

  // shouldn't need to do broadcast/gather thing

  // now do sqrt, etc
  if (a_p == 2)
  {
    norm = sqrt(norm);
  }

  return norm;
}

Real computeNorm(const LevelData<FArrayBox>& a_phi,
                 const DisjointBoxLayout*    a_finerGridsPtr,
                 const int                   a_nRefFine,
                 const Real                  a_dx,
                 const Interval              a_comps,
                 const int                   a_p)
{
  Real normLevel;
  normLevel = 0.0;

  const DisjointBoxLayout& levelGrids = a_phi.getBoxes();

  LevelData<FArrayBox> temp(levelGrids, a_comps.size());

  Interval tempComps(0,a_comps.size()-1);

  DataIterator dit = a_phi.dataIterator();

  for (dit.reset(); dit.ok(); ++dit)
  {
    temp[dit()].copy(temp[dit()].box(), tempComps,
                     temp[dit()].box(), a_phi[dit()], a_comps);

    if (a_finerGridsPtr != NULL)
    {
      LayoutIterator litFine = a_finerGridsPtr->layoutIterator();

      // now loop over fine boxes and set covered regions to 0
      for (litFine.reset(); litFine.ok(); ++litFine)
      {
        Box coveredBox(a_finerGridsPtr->get(litFine()));
        coveredBox.coarsen(a_nRefFine);
        coveredBox &= temp[dit()].box();

        if (!coveredBox.isEmpty())
        {
          temp[dit()].setVal(0.0, coveredBox, 0, tempComps.size());
        }
      }  // end loop over fine-grid boxes
    } // end if there is a finer level
  } // end loop over this level's grids

  // compute unscaled norm
  normLevel = norm(temp, tempComps, a_p);

  if (a_p != 0)
  {
    Real exponent = SpaceDim;
    exponent /= a_p;
    Real scale = pow(a_dx,exponent);
    normLevel *= scale;
  }

  return normLevel;
}

Real computeMax(const Vector<LevelData<FArrayBox>* >& a_phi,
                const Vector<int>&                    a_nRefFine,
                const Interval                        a_comps,
                const int                             a_lBase)
{
  int numLevels = a_phi.size();

  // it is often the case that while a_phi has many possible
  // levels, only a subset of them are defined -- check that
  // just to be sure
  if (a_phi[numLevels-1] == NULL)
  {
    int lev= numLevels-1;

    while (a_phi[lev] == NULL)
    {
      lev--;
    }

    numLevels = lev+1;
  }

  Real max;
  Real maxOnLevel;

  max = -88888888888.0;

  // loop over levels
  for (int lev = a_lBase; lev < numLevels; lev++)
  {
    //in case there are extra levels which are not defined
    if (a_phi[lev] != NULL)
    {
      CH_assert(a_phi[lev]->isDefined());

      LevelData<FArrayBox>& thisPhi = *(a_phi[lev]);
      const DisjointBoxLayout* finerGridsPtr = NULL;

      if (lev < numLevels-1)
      {
        finerGridsPtr = &(a_phi[lev+1]->getBoxes());

        maxOnLevel = computeMax(thisPhi, finerGridsPtr, a_nRefFine[lev],
                                a_comps);
      }
      else
      {
        int bogusRefRatio = 100000;
        maxOnLevel = computeMax(thisPhi, finerGridsPtr, bogusRefRatio,
                                a_comps);
      }

      if (maxOnLevel > max)
      {
        max = maxOnLevel;
      }
    }
  }

  // shouldn't need to do broadcast/gather thing
  return max;
}

Real computeMax(const LevelData<FArrayBox>& a_phi,
                const DisjointBoxLayout*    a_finerGridsPtr,
                const int                   a_nRefFine,
                const Interval              a_comps)
{

  Real levelMax;
  Real thisMax;

  levelMax = -99999999999.9;

  const DisjointBoxLayout& levelGrids = a_phi.getBoxes();

  LevelData<FArrayBox> temp(levelGrids, a_comps.size());

  Interval tempComps(0,a_comps.size()-1);

  DataIterator dit = a_phi.dataIterator();

  for (dit.reset(); dit.ok(); ++dit)
  {
    temp[dit()].copy(temp[dit()].box(), tempComps,
                     temp[dit()].box(), a_phi[dit()], a_comps);

    if (a_finerGridsPtr != NULL)
    {
      LayoutIterator litFine = a_finerGridsPtr->layoutIterator();

      // now loop over fine boxes and set covered regions to 0
      for (litFine.reset(); litFine.ok(); ++litFine)
      {
        Box coveredBox(a_finerGridsPtr->get(litFine()));
        coveredBox.coarsen(a_nRefFine);
        coveredBox &= temp[dit()].box();

        if (!coveredBox.isEmpty())
        {
          temp[dit()].setVal(levelMax, coveredBox, 0, tempComps.size());
        }
      }  // end loop over fine-grid boxes
    } // end if there is a finer level

    // while we're looping over the grids, get max as well
    // need to loop over comps
    for (int comp = tempComps.begin(); comp <= tempComps.end(); ++comp)
    {
      thisMax = temp[dit()].max(comp);
      if (thisMax > levelMax)
      {
        levelMax = thisMax;
      }
    }
  } // end loop over this level's grids

  // need to do MPI gather here
#ifdef CH_MPI
  Real recv;
  int result = MPI_Allreduce(&levelMax, &recv, 1, MPI_CH_REAL,
                             MPI_MAX, Chombo_MPI::comm);

  if (result != MPI_SUCCESS)
  {
    MayDay::Error("Sorry, but I had a communication error in computeNorm");
  }

  levelMax = recv;
#endif

  return levelMax;
}

Real computeMin(const Vector<LevelData<FArrayBox>* >& a_phi,
                const Vector<int>&                    a_nRefFine,
                const Interval                        a_comps,
                const int                             a_lBase)
{
  int numLevels = a_phi.size();

  // it is often the case that while a_phi has many possible
  // levels, only a subset of them are defined -- check that
  // just to be sure
  if (a_phi[numLevels-1] == NULL)
  {
    int lev = numLevels-1;

    while (a_phi[lev] == NULL)
    {
      lev--;
    }

    numLevels = lev+1;
  }

  Real min;
  Real minOnLevel;

  min = 88888888888.0;

  // loop over levels
  for (int lev = a_lBase; lev < numLevels; lev++)
  {
    //in case there are extra levels which are not defined
    if (a_phi[lev] != NULL)
    {
      CH_assert(a_phi[lev]->isDefined());

      LevelData<FArrayBox>& thisPhi = *(a_phi[lev]);
      const DisjointBoxLayout* finerGridsPtr = NULL;

      if (lev < numLevels-1)
      {
        finerGridsPtr = &(a_phi[lev+1]->getBoxes());

        minOnLevel = computeMin(thisPhi, finerGridsPtr, a_nRefFine[lev],
                                a_comps);
      }
      else
      {
        int bogusRefRatio = 1000000;
        minOnLevel = computeMin(thisPhi, finerGridsPtr, bogusRefRatio,
                                a_comps);
      }

      if (minOnLevel < min)
      {
        min = minOnLevel;
      }
    }
  }

  // shouldn't need to do broadcast/gather thing
  return min;
}

Real computeMin(const LevelData<FArrayBox>& a_phi,
                const DisjointBoxLayout*    a_finerGridsPtr,
                const int                   a_nRefFine,
                const Interval              a_comps)
{

  Real levelMin;
  Real thisMin;

  levelMin = 99999999999.9;

  const DisjointBoxLayout& levelGrids = a_phi.getBoxes();

  LevelData<FArrayBox> temp(levelGrids, a_comps.size());

  Interval tempComps(0,a_comps.size()-1);

  DataIterator dit = a_phi.dataIterator();

  for (dit.reset(); dit.ok(); ++dit)
  {
    temp[dit()].copy(temp[dit()].box(), tempComps,
                     temp[dit()].box(), a_phi[dit()], a_comps);

    if (a_finerGridsPtr != NULL)
    {
      LayoutIterator litFine = a_finerGridsPtr->layoutIterator();

      // now loop over fine boxes and set covered regions to 0
      for (litFine.reset(); litFine.ok(); ++litFine)
      {
        Box coveredBox(a_finerGridsPtr->get(litFine()));
        coveredBox.coarsen(a_nRefFine);
        coveredBox &= temp[dit()].box();

        if (!coveredBox.isEmpty())
        {
          temp[dit()].setVal(levelMin, coveredBox, 0, tempComps.size());
        }
      }  // end loop over fine-grid boxes
    } // end if there is a finer level

    // while we're looping over the grids, get min as well
    // need to loop over comps
    for (int comp = tempComps.begin(); comp <= tempComps.end(); ++comp)
    {
      thisMin = temp[dit()].min(comp);
      if (thisMin < levelMin)
      {
        levelMin = thisMin;
      }
    }
  } // end loop over this level's grids

#ifdef CH_MPI
  Real recv;
  int result = MPI_Allreduce(&levelMin, &recv, 1, MPI_CH_REAL,
                             MPI_MIN, Chombo_MPI::comm);

  if (result != MPI_SUCCESS)
    {
      MayDay::Error("Sorry, but I had a communication error in computeNorm");
    }

  levelMin = recv;
#endif

  return levelMin;
}

Real norm(const BoxLayoutData<FluxBox>& a_layout,
          const Interval&               a_interval,
          const int&                    a_p)
{
  const BoxLayout& validRegion = a_layout.boxLayout();
  DataIterator dit = validRegion.dataIterator();
  Real norm = 0;

  if (a_p == 0)
  {
    for (dit.reset(); dit.ok(); ++dit)
    {
      const FluxBox& thisFlux = a_layout[dit()];
      for (int dir = 0; dir < SpaceDim; dir++)
      {
        norm = Max(norm, thisFlux[dir].norm(validRegion[dit()],
                                            0,
                                            a_interval.begin(),
                                            a_interval.size()));
      }
    }

#ifdef CH_MPI
    Real recv;
    int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                               MPI_MAX, Chombo_MPI::comm);

    if (result != MPI_SUCCESS)
    {
      MayDay::Error("sorry, but I had a communcation error in norm");
    }

    norm = recv;
#endif

  }
  else if (a_p == 1) // abs sum norm
  {
    for (dit.reset(); dit.ok(); ++dit)
    {
      const FluxBox& thisFlux = a_layout[dit()];

      for (int dir = 0; dir < SpaceDim; ++dir)
      {
        norm += thisFlux[dir].norm(validRegion[dit()],
                                   1,
                                   a_interval.begin(),
                                   a_interval.size());
      }
    }

#ifdef CH_MPI
      Real recv;
      int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                                 MPI_SUM, Chombo_MPI::comm);

      if (result != MPI_SUCCESS)
      {
        MayDay::Error("sorry, but I had a communcation error in norm");
      }

      norm = recv;
#endif
  }
  else
  {
    for (dit.reset(); dit.ok(); ++dit)
      {
        const FluxBox& thisFlux = a_layout[dit()];

        for (int dir = 0; dir < SpaceDim; dir++)
        {
          norm+=thisFlux[dir].sumPow(validRegion[dit()],
                                     a_p,
                                     a_interval.begin(),
                                     a_interval.size());
        }
      }

#ifdef CH_MPI
    Real recv;

    int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                               MPI_SUM, Chombo_MPI::comm);

    if(result != MPI_SUCCESS)
    {
      MayDay::Error("sorry, but I had a communcation error on norm");
    }

    norm = recv;
#endif

    Real invpwr = 1.0/a_p;
    norm = pow(norm, invpwr);
  }

  return norm;
}
