/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstring>
#include "BitSet.H"
#include <cstdlib>
#include <malloc.h>
#include "MayDay.H"
#include "SPMD.H"

BitSet BitSetIterator::emptyBitSet = BitSet();

BITSETWORD BitSet::trueMasks[BITSETWORDSIZE];

int BitSet::initialize()
{
  trueMasks[BITSETWORDSIZE-1] = 1;
  for(int i=BITSETWORDSIZE-2; i>=0 ; --i)
    {
      trueMasks[i] = trueMasks[i+1] << 1;
    }
  return 0;
}

BitSet::BitSet()
  :
  m_bits(NULL),
  m_size(0),
  m_length(0)
{}

long int BitSet::bytes = BitSet::initialize();
long int BitSet::peak = 0;

int BitSet::linearSize() const
{
  return 2* ::linearSize<int>(0) + m_length* sizeof(BITSETWORD);
}

void BitSet::linearIn(const void* const inBuf)
{
  if(m_bits != NULL)
    {
      free(m_bits);
      bytes -= m_length*sizeof(BITSETWORD);
    }

  char* buf = (char*)inBuf;
  ::linearIn<int>(m_size, buf);
  buf+=::linearSize<int>(m_size);
  ::linearIn<int>(m_length, buf);
  buf+=::linearSize<int>(m_length);
  m_bits = (BITSETWORD*)malloc(m_length*sizeof(BITSETWORD));
  memcpy(m_bits, buf, m_length*sizeof(BITSETWORD));
  bytes += m_length*sizeof(BITSETWORD);

}

void BitSet::linearOut(void* const a_outBuf) const
{
  char* buf = (char*)a_outBuf;
  ::linearOut<int>(buf, m_size);
  buf+=::linearSize<int>(m_size);
  ::linearOut<int>(buf, m_length);
  buf+=::linearSize<int>(m_length);
  memcpy(buf, m_bits, m_length*sizeof(BITSETWORD));
}

BitSet::BitSet(int bits, bool init)
  :
  m_bits(NULL),
  m_size(bits),
  m_length(0)
{
  define(bits, init);
}

bool BitSet::operator<( const BitSet& rhs ) const
{
  if( m_length < rhs.m_length )
  {
    return true;
  } else
  if( m_length > rhs.m_length )
  {
    return false;
  }

  for( int i=0; i<m_length; ++i )
  {
    if( m_bits[i] < rhs.m_bits[i] )
    {
      return true;
    } else
    if( m_bits[i] > rhs.m_bits[i] )
    {
      return false;
    }
  }

  return false;
}


void BitSet::define(int bits, bool init)
{
  if(m_bits != NULL){
    free(m_bits);
    bytes-=m_length*sizeof(BITSETWORD);
  }
  m_size = bits;
 CH_assert(bits >= 0);
  m_length = bits/BITSETWORDSIZE;
  if(bits == 0) { m_bits = NULL; return;}
  if(m_length*BITSETWORDSIZE != bits) m_length +=1;
  m_bits = (BITSETWORD*)malloc(m_length*sizeof(BITSETWORD));
  bytes+= m_length*sizeof(BITSETWORD);
  if(bytes > peak) peak = bytes;

  if(m_bits == NULL)
    {
      MayDay::Error("Memory Error in BitSet::BitSet(int bits, bool init)");
    }
  if(init)
    {
      memset(m_bits, ~0, (m_length-1)*sizeof(BITSETWORD));
      BITSETWORD* word = m_bits+(m_length-1);
      *word = 0;
      int count = m_size - (m_length-1)*BITSETWORDSIZE;
      for(int i=0; i<count; ++i)
        {
          *word = *word | trueMasks[i];
        }

    }
  else
    {
      memset(m_bits, 0, m_length*sizeof(BITSETWORD));
    }
}

BitSet::BitSet(const BitSet& rhs)
  :
  m_bits(NULL),
  m_size(rhs.m_size),
  m_length(rhs.m_length)
{
  if(m_length == 0) return;
  m_bits = (BITSETWORD*)malloc(m_length*sizeof(BITSETWORD));
  bytes+= m_length*sizeof(BITSETWORD);
  if(bytes > peak) peak = bytes;
  if(m_bits == NULL)
    {
      MayDay::Error("Memory Error in BitSet::BitSet(const BitSet& rhs)");
    }

  memcpy(m_bits, rhs.m_bits, m_length*sizeof(BITSETWORD));
}

bool BitSet::isEmpty() const
{
  if(m_length == 0) return true;
  for(int i=0; i<m_length; ++i)
    {
      if(m_bits[i] > 0) return false;
    }
  /*
  for(int i=(m_length-1)*BITSETWORDSIZE; i<m_size; ++i)
    {
      if (this->operator[](i)) return false;
    }
  */
  return true;
}

bool BitSet::isFull() const
{
  BITSETWORD g=0;
  g=~g;
  for(int i=0; i<m_length-1; ++i)
    {
      if(m_bits[i] != g) return false;
    }
  for(int i=(m_length-1)*BITSETWORDSIZE; i<m_size; ++i)
    {
      if (!this->operator[](i)) return false;
    }
  return true;
}

BitSet& BitSet::operator=(const BitSet& rhs)
{
  if(m_bits != NULL)
    {
      free(m_bits);
      bytes -= m_length*sizeof(BITSETWORD);
      m_bits = NULL;
    }
  m_length=rhs.m_length;
  m_size =rhs.m_size;

  if(rhs.m_bits != NULL)
    {
      m_bits = (BITSETWORD*)malloc(m_length*sizeof(BITSETWORD));
      bytes+= m_length*sizeof(BITSETWORD);
      if(bytes > peak) peak = bytes;
    }
  else
    {
      m_bits = NULL;
      return *this;
    }
  if(m_bits == NULL)
    {
      MayDay::Error("Memory Error in BitSet::operator=(const BitSet& rhs)");
    }

  memcpy(m_bits, rhs.m_bits, m_length*sizeof(BITSETWORD));
  return *this;
}

BitSet::~BitSet()
{
  if(m_bits != NULL)
    {
      free(m_bits);
      bytes-=m_length*sizeof(BITSETWORD);
    }
}

void BitSet::setFalse(int i)
{
  // BITSETWORD* tm = trueMasks;
 CH_assert(i>=0);
 CH_assert(i<m_size);
  int index = i/BITSETWORDSIZE;
 CH_assert(index < m_length);
  int remainder = i-BITSETWORDSIZE*index;
  BITSETWORD* word = m_bits+index;

  *word = *word & ~(trueMasks[remainder]);
}

void BitSet::setTrue(int i)
{
 CH_assert(i>=0);
 CH_assert(i<m_size);
  int index = i/BITSETWORDSIZE;
 CH_assert(index < m_length);
  int remainder = i-BITSETWORDSIZE*index;
  BITSETWORD* word = m_bits+index;

  *word = *word | trueMasks[remainder];
}

void BitSetIterator::end()
{
  m_index = m_length;
  m_remainder = m_partialBits;
}
void BitSetIterator::begin()
{
  m_index = 0;
  m_remainder = 0;
}
