/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cmath>
using std::pow;

#include "BoxLayoutData.H"

#ifdef CH_MPI
#include <mpi.h>
#endif

Real norm(const BoxLayoutData<FArrayBox>& a_layout,
          const Interval& interval,
          const int& p)
{
  const BoxLayout validRegion = a_layout.boxLayout();
  DataIterator it = validRegion.dataIterator();
  Real norm = 0;
  if(p == 0)  // max norm
    {
      for(;it.ok(); ++it)
        {
          norm = Max(norm, a_layout[it()].norm(validRegion[it()],
                                               0,
                                               interval.begin(),
                                               interval.size()));
        }
#     ifdef CH_MPI
      Real recv;
      int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                                 MPI_MAX, Chombo_MPI::comm);
      if(result != MPI_SUCCESS){ //bark!!!
        MayDay::Error("sorry, but I had a communcation error on norm");
      }
      norm = recv;
#     endif
    }
  else if(p == 1) // abs sum norm
    {
      for(;it.ok(); ++it)
        {
          norm += a_layout[it()].norm(validRegion[it()],
                                      1,
                                      interval.begin(),
                                      interval.size());
        }
#     ifdef CH_MPI
      Real recv;
      int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                                 MPI_SUM, Chombo_MPI::comm);
      if(result != MPI_SUCCESS){ //bark!!!
        MayDay::Error("sorry, but I had a communcation error on norm");
      }
      norm = recv;
#     endif
    }
  else
    {
      for(;it.ok(); ++it)
        {
          norm+=a_layout[it()].sumPow(validRegion[it()],
                                      p,
                                      interval.begin(),
                                      interval.size());
        }
#     ifdef CH_MPI
      Real recv;
      int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                                 MPI_SUM, Chombo_MPI::comm);
      if(result != MPI_SUCCESS){ //bark!!!
        MayDay::Error("sorry, but I had a communcation error on norm");
      }
      norm = recv;
#     endif
      Real invpwr = 1.0/p;
      if(p == 2) norm = sqrt(norm);
      else       norm = pow(norm, invpwr);
    }

  return norm;
}

template <>
BaseFab<int>* DefaultDataFactory<BaseFab<int> >::create(const Box& box,
                                                        int ncomps,
                                                        const DataIndex& a_datInd) const
{
  return new BaseFab<int>(box, ncomps);
}

template <>
FArrayBox* DefaultDataFactory<FArrayBox>::create(const Box& box, int ncomps,
                                          const DataIndex& a_datInd) const
{
  return new FArrayBox(box, ncomps);
}

FABAliasDataFactory::FABAliasDataFactory(const LayoutData<Real*>& aliases)
{
  define(aliases);
}

void FABAliasDataFactory::define(const LayoutData<Real*>& aliases)
{
  aliasPtrs.define(aliases.boxLayout());
  DataIterator dit = aliases.dataIterator();
  for(dit.begin(); dit.ok(); ++dit)
    {
      aliasPtrs[dit()] = aliases[dit()];
    }
}

FArrayBox* FABAliasDataFactory::create(const Box& box, int ncomps, const DataIndex& a_datInd) const
{
  const Box& b = aliasPtrs.box(a_datInd);
  if(b != box)
    {
      MayDay::Error("Aliased data holder dimensions do not match across LevelData const.");
    }
  FArrayBox* rtn = new FArrayBox(box, ncomps, aliasPtrs[a_datInd]);
  return rtn;
}
