/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _LAYOUTITERATOR_H_
#define _LAYOUTITERATOR_H_

#ifndef WRAPPER

#include "DataIndex.H"
#include "BoxLayout.H"

#endif

///An Iterator based on a BoxLayout object.
/**
   An Iterator based on a BoxLayout object.  It does not
   support a dereferencing operation(1), since it is intended
   to work with all of BoxLayouts, DisjointBoxLayouts, BoxLayoutDatas
   LevelData's, and any object that is built on top
   of a BoxLayout object.  LayoutIterator accesses the data in a BoxLayout-based
   object in a NON-data-parallel manner (i.e. every processor iterates through
   \b all the Boxes in the BoxLayout).  This differs from the DataIterator class.

   BoxLayout-based objects can act as the Factory for the LayoutIterator.

 (1) STL-speak.  not critical for comprehension, but can help people
 familiar with STL iterators and expecting similar behaviour.

*/
class LayoutIterator
{
public:
  /// a null constructed LayoutIterator will return false on ok()
  LayoutIterator()
  {}

  LayoutIterator(const BoxLayout& a_layout)
  {
    *this = a_layout.layoutIterator();
  }

  // Default copy and null constructor should be fine.
  // Not useful to someone using the iterator.  Used by
  // other classes in Chombo.  Could make it private and
  // let BoxLayout have access....

  ~LayoutIterator()
  {}

  /// return the index that this iterator is at
  const LayoutIndex& operator () () const;

  /// return a copy of the index that this iterator is at
  LayoutIndex i() const
  {
    return this->operator()();
  }

  /// move the iterator to the next Box in the layout
  void operator ++ ();

  /// move the iterator to the next Box in the layout
  void incr()
  {
    ++(*this);
  }

  /// return true if this iterator is still in its Layout
  bool ok() const;

  /// initialize this iterator to the first Box in its Layout
  void begin();

  /// same as begin()
  void reset();

  /// move this iterator to after the last Box in the layout
  /** The iterator will be !ok() afterwards. */
  void end();


protected:
  friend class BoxLayout;

  LayoutIterator(const BoxLayout& a_boxlayout,
                 const int*       a_layoutID);

  BoxLayout    m_layout;
  unsigned int m_index;

  LayoutIndex  m_current;
};

#ifndef WRAPPER

inline const LayoutIndex& LayoutIterator::operator () () const
{
 CH_assert(ok());
  return m_current;
}

inline void LayoutIterator::operator ++ ()
{
  ++m_index;

  if (ok())
  {
    m_current.m_index = (*(m_layout.m_boxes))[m_index].index;
  }
}

inline bool LayoutIterator::ok() const
{
  return m_index < m_layout.size();
}

inline void LayoutIterator::begin()
{
  m_index = 0;

  if (ok())
  {
    m_current.m_index = (*(m_layout.m_boxes))[m_index].index;
  }
}
#endif /*WRAPPER*/

#endif
