/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <cstring>
using namespace std;

#include "MayDay.H"
#include "CHOMBO_VERSION.H"
#ifdef CH_MPI
#include "mpi.h"
#endif

//
// The definition of our NULL string used as default argument.
//
const char * const MayDay::m_nullString = "";

//
// The definition of our version string.
//
#define bl_str(s) # s
#define bl_xstr(s) bl_str(s)

const char * const MayDay::version = "MayDay version " bl_xstr(CHOMBO_VERSION) " built " __DATE__ " at " __TIME__ ;

#undef bl_str
#undef bl_xstr


//
// This is used by MayDay::Error(), MayDay::Abort() and MayDay::Warning()
// to ensure that when writing the message to stderr, that no additional
// heap-based memory is allocated.
//
static void write_to_stderr_without_buffering (const char * const a_str)
{
  //
  // Flush all buffers.
  //
  fflush(NULL);

  if (a_str)
  {
    //
    // Add some `!'s and a newline to the string.
    //
    const char * const end = " !!!\n";

    fwrite(a_str, strlen(a_str), 1, stderr);
    fwrite(end  , strlen(end  ), 1, stderr);
  }
}

void MayDay::Error(const char * const a_msg, int exit_code)
{
  write_to_stderr_without_buffering(a_msg);
#ifdef CH_MPI
  MPI_Abort(MPI_COMM_WORLD ,exit_code);
  // this shouldn't return, but if it does, exit serially
#endif
  exit(exit_code);
}

void MayDay::Abort(const char * const a_msg)
{
  write_to_stderr_without_buffering(a_msg);
#ifdef CH_MPI
  MPI_Abort(MPI_COMM_WORLD ,CH_DEFAULT_ERROR_CODE);
  // this shouldn't return, but if it does, abort serially
#endif
  abort();
}

void MayDay::Warning(const char * const a_msg)
{
  write_to_stderr_without_buffering(a_msg);
}


#include "FORT_PROTO.H"

extern "C" {

  void
  FORTRAN_NAME(MAYDAYERROR,maydayerror) (void)
  {
    MayDay::Error( "A ChomboFortran routine called MAYDAYERROR().  Rerun with the debugger to find where." ) ;
  }

  void
  FORTRAN_NAME(MAYDAYABORT,maydayabort) (void)
  {
    MayDay::Abort( "A ChomboFortran routine called MAYDAYABORT().  Rerun with the debugger to find where." ) ;
  }

  void
  FORTRAN_NAME(MAYDAYEXIT,maydayexit) (int* exit_code)
  {
    MayDay::Error( "A ChomboFortran routine called MAYDAYEXIT().  Rerun with the debugger to find where." ,*exit_code ) ;
  }

  void
  FORTRAN_NAME(MAYDAY_ERROR,mayday_error) (void)
  {
    MayDay::Error( "A ChomboFortran routine called MAYDAY_ERROR().  Rerun with the debugger to find where." ) ;
  }

  void
  FORTRAN_NAME(MAYDAY_ABORT,mayday_abort) (void)
  {
    MayDay::Abort( "A ChomboFortran routine called MAYDAY_ABORT().  Rerun with the debugger to find where." ) ;
  }

  void
  FORTRAN_NAME(MAYDAY_EXIT,mayday_exit) (int* exit_code)
  {
    MayDay::Error( "A ChomboFortran routine called MAYDAY_EXIT().  Rerun with the debugger to find where." ,*exit_code ) ;
  }

}
