/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// Merge boxes on lines class
// this is used for line solves with AMR
// ---------
//  class MergeBoxesOnLines
//
///////////////////////////////////////////////////////////////////////////////

// Include files:

#include "Box.H"
#include "MayDay.H"
#include "parstream.H"
#include "MergeBoxesOnLines.H"
void MergeBoxesOnLines::
mergeBoxes(Vector<Vector<Box> >& a_vvbox,
           const int&   a_dir)
{
  for(int ilev = 0; ilev < a_vvbox.size(); ilev++)
    {
      mergeBoxes(a_vvbox[ilev],a_dir);
    }
}
void MergeBoxesOnLines::
mergeBoxes(Vector<Box>& a_vboxin,
           const int&   a_dir)
{

  //loop through all the boxes and chop them based on their neighbors
  bool keepChopping = (a_vboxin.size()>1);
  while(keepChopping)
    {
      keepChopping = false;
      for (SideIterator sit; sit.ok(); ++sit)
        {
          for(int dir = 0; dir < SpaceDim; dir++)
            {
              if(dir != a_dir)
                {
                  Vector<Box> vboxtemp;
                  for(int ibox = 0; ibox < a_vboxin.size(); ibox++)
                    {
                      Box& boxThis = a_vboxin[ibox];
                      Box growThis = boxThis;
                      growThis.grow(1);
                      growThis.grow(a_dir,-1);
                      for(int jbox = 0; jbox < a_vboxin.size(); jbox++)
                        {
                          if(ibox!=jbox)
                            {
                              const Box& boxNeigh = a_vboxin[jbox];

                              //check if this is a chop candidate (only chop based on a_dir neighbor boxes)
                              if(!boxNeigh.intersects(growThis))
                                {
                                  //get an adjacent box to the neighbor in the tangent direction dir
                                  Box sideNeigh = adjCellBox(boxNeigh,dir,sit(),1);
                                  //grow this in the line direction (a_dir)
                                  sideNeigh.grow(a_dir,1);
                                  if(sideNeigh.intersects(boxThis))
                                    {
                                      int chop_pnt = sideNeigh.bigEnd(dir);
                                      if(chop_pnt>boxThis.smallEnd(dir))
                                        {
                                          //chop this box based on the neighbor (results in two boxes)
                                          //pout() << "Chopping ibox = " << ibox << std::endl;
                                          Box newBox = boxThis.chop(dir,chop_pnt);
                                          if(sit()==Side::Lo)
                                            {
                                              boxThis.growHi(dir,1);
                                              newBox.growLo(dir,-1);
                                            }
                                          //add the new box to vboxtemp
                                          if(!newBox.isEmpty())
                                            {
                                              keepChopping = true;
                                              vboxtemp.push_back(newBox);
                                            }
                                        }
                                    }
                                }
                            }
                        }
                      //add this box to the new list
                      vboxtemp.push_back(boxThis);
                    }
                  a_vboxin = vboxtemp;
                }
            }
        }
    }

  //loop through all the boxes and merge them based on the neighbors
  bool keepMerging = (a_vboxin.size()>1);
  while(keepMerging)
    {
      keepMerging = false;
      Vector<Box> vboxtemp;
      for(int ibox = 0; ibox < a_vboxin.size(); ibox++)
        {
          Box& boxThis = a_vboxin[ibox];
          Box newBoxThis = boxThis;
          for(int jbox = 0; jbox < a_vboxin.size(); jbox++)
            {
              if(ibox<jbox)
                {
                  Box& boxNeigh = a_vboxin[jbox];
                  if(!boxNeigh.isEmpty())
                    {
                      //get an adjacent box to the neighbor in the tangent direction dir
                      Box growNeigh = boxNeigh;
                      //grow this in the line direction (a_dir)
                      growNeigh.grow(a_dir,1);
                      if(growNeigh.intersects(boxThis))
                        {
                          //pout() << "Merging ibox = " << ibox << std::endl;
                          keepMerging = true;
                          //merge this box with the neighbor, keeping the neighbor
                          newBoxThis = minBox(newBoxThis,boxNeigh);
                          boxNeigh = Box();//makes the neighbor empty
                        }
                    }
                }
            }
          //add this box to the new list only if it's not empty
          if(!newBoxThis.isEmpty())
            {
              vboxtemp.push_back(newBoxThis);
            }
        }
      a_vboxin = vboxtemp;
    }
}
      
MergeBoxesOnLines::MergeBoxesOnLines()
{
}
MergeBoxesOnLines::~MergeBoxesOnLines()
{
}

