/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// NodeFArrayBox.H
// adapted from dMartin/Chombo-IAMR/util/FluxBox.H
// by Dan Martin, Fri, Jan 14, 2000
// petermc, 1 Nov 2000
// 8 Jan 2001 added 

#ifndef NODEFARRAYBOX_H
#define NODEFARRAYBOX_H

#include "FArrayBox.H"

/// A wrapper for an FArrayBox to contain NODE-centered data.
/** This is a class to contain NODE-centered data on a box.<br>
    <b>Question</b>:  Why not just use NODE-centered FArrayBox?<br>
    <b>Answer</b>:  We want to use LevelData<NodeFArrayBox>,
    which requires that the underlying BoxLayout be a DisjointBoxLayout.
    Adjacent NODE-centered boxes intersect, but adjacent CELL-centered
    boxes do not.<br>
    The underlying DisjointBoxLayout will be CELL-centered.

    In a LevelData<NodeFArrayBox>, some of the grids may share nodes.
    The data on shared nodes must be the same.

    Example:  A four-grid layout is shown on the left.  As shown on
    the right, the nodes marked "2" are shared by two grids, and
    those marked "3" are shared by three grids.
<pre>
                  +-----+                     YYYY--+
                  |     |                     |     |
                  |     |                     |     |
                  |     |                     |     |
              +---+--+--+---+             +---2223222---+
              |      |      |             |      2      |
              |      |      |             |      2      |
              |      |      |             |      2      |
          +---+------+------+         X---22222223------X
          |          |                |          |
          |          |                |          |
          |          |                |          |
          +----------+                +-------YYYY
</pre>
    Additional nodes may be shared if the boxes extend to the faces
    of a domain that is periodic in one or more directions.
    In the example above, these nodes are marked "X" and "Y".<br>
    Horizontally periodic domain:  node "X" shared by two grids.<br>
    Vertically periodic domain:  nodes "Y" shared by two grids.

    One of the most useful features of LevelData<NodeFArrayBox> is the
    copyTo() function.  THE copyTo() FUNCTION MUST BE USED WITH CAUTION
    if the source and destination have different layouts.

    Consider an example where A and B are on layouts of one box
    each, and these two boxes abut:
<pre>
      +-------+
      |       |
      |   A   |
      |       |
      =========
      |       |
      |   B   |
      |       |
      +-------+
</pre>
    If we do A.copyTo(B), then the data on the nodes of the interface
    (marked "=") are NOT copied, because the underlying CELL-centered
    DisjointBoxLayouts of the LevelDatas do not intersect.

    So whenever we want to use src.copyTo(dest), unless we are sure
    that src and dest have the same layout, we do the following:
    - define BoxLayoutData<NodeFArrayBox> destExpanded on the BoxLayout of
    dest with each box expanded by one in each dimension;
    - do src.copyTo(destExpanded);
    - box by box, copy destExpanded to dest.
*/
class NodeFArrayBox
{

public:

  /**
     \name Constructors, destructor and defines
  */
  /*@{*/

  ///
  /** Default constructor.  User must subsequently call define().
   */
  NodeFArrayBox();

  /// 
  /** Constructs NodeFArrayBox on CELL-centered box <i>a_bx</i>
      with <i>a_nComp</i> components using the aliased data space.
  */
  NodeFArrayBox(const Box& a_bx, int a_nComp, Real* a_alias = NULL);

  /// 
  /** Constructs an aliased NodeFArrayBox.
  */
  NodeFArrayBox(const Interval& a_comps, NodeFArrayBox& a_original);

  ///
  /** Destructor.
   */
  ~NodeFArrayBox();

  ///
  /** Defines NodeFArrayBox on CELL-centered box <i>a_bx</i>
      with <i>a_nComp</i> components.
   */
  void define(const Box& a_bx, int a_nComp=1);

  /*@}*/

  /**
     \name Accessors
  */
  /*@{*/

  ///
  /** Returns the CELL-centered domain where the array is defined.
   */
  const Box& box() const;

  ///
  /** Returns a modifiable reference to the NODE-centered FArrayBox
      containing the data.
  */
  FArrayBox& getFab();

  /// 
  /** Returns a constant reference to the NODE-centered FArrayBox
      containing the data.
  */
  const FArrayBox& getFab() const;

  /*@}*/

  /**
     \name Data modification functions
  */
  /*@{*/

  ///
  /** Modifies the data in this NodeFArrayBox by copying data
      from <i>a_src</i> into it, over the nodes that they have in common.<br>
      If this NodeFArrayBox and <i>a_src</i> have abutting grids,
      as shown:
<pre>
      +--------+
      |        |
      |  this  |
      |        |
      ==========   <- data copied here from a_src to this
      |        |
      |  a_src |
      |        |
      +--------+
</pre>
      then the data in this NodeFArrayBox WILL be modified along the
      edge nodes where they abut -- even though this.box() and
      <i>a_src</i>.box(), being CELL-centered, do NOT intersect.

      All components are copied.
  */
  void copy(const NodeFArrayBox& a_src);

  ///
  /** Modifies the data in this NodeFArrayBox by copying the data
      from <i>a_src</i> into it, over the intersection of:
      - the nodes surrounding the CELL-centered box <i>a_regionFrom</i>;
      - the nodes surrounding the CELL-centered box <i>a_regionTo</i>;
      - the nodes containing data in this NodeFArrayBox;
      - the nodes containing data in <i>a_src</i>.

      The components in the interval <i>a_Csrc</i> in <i>a_src</i>
      are copied to the components in the interval <i>a_Cdest</i>
      in this NodeFArrayBox.

      This function is required in order to have BoxLayoutData<NodeFArrayBox>.
  */
  void copy(const Box& a_regionFrom,
            const Interval& a_Cdest,
            const Box& a_regionTo,
            const NodeFArrayBox& a_src,
            const Interval& a_Csrc);

  /*@}*/

  /**
     \name Linearization functions
  */
  /*@{*/
  
  ///
  /** Returns size, in number of bytes, of a flat linear
      representation of the data in this object in the area
      defined by the nodes in CELL-centered box <i>a_R</i> and the
      component Interval <i>a_comps</i>.

      This function is required in order to have BoxLayoutData<NodeFArrayBox>.
  */
  int size(const Box& a_R, const Interval& a_comps) const;

  ///
  /** Writes into <i>a_buf</i> a linear representation of the internal data for
      the nodes surrounding CELL-centered box <i>a_R</i>, over the component
      Interval <i>a_comps</i>.

      Assumes that sufficient
      memory for the buffer has already been allocated by the caller.

      This function is required in order to have BoxLayoutData<NodeFArrayBox>.
  */
  void linearOut(void* a_buf, const Box& a_R, const Interval& a_comps) const;

  ///
  void linearIn(void* a_buf, const Box& a_R, const Interval& a_comps);


  /// These functions are required for broadcast & gather.
  void linearOut(void* a_buf) const;
  ///
  void linearIn(const void* const a_buf);
  ///
  int linearSize(void) const;

  ///
  static int preAllocatable() {return 0;}

  /*@}*/
  
protected:

  // CELL-centered box. data are defined on the surrounding nodes.
  Box m_box;

  // NODE-centered data
  FArrayBox m_fab;

private:
  /// these are disallowed  (huh?)
  
  NodeFArrayBox (const NodeFArrayBox&);
  NodeFArrayBox& operator = (const NodeFArrayBox&);

};

#endif
