/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// NodeFArrayBox.cpp
// from dMartin/Chombo-IAMR/util/FluxBox.cpp by Dan Martin, Fri, Jan 14, 2000
// petermc, 1 Nov 2000
// petermc, 26 Mar 2002 added using std::cout and using std::endl

#include "NodeFArrayBox.H"
#include "SPMD.H"
#include "DebugDump.H"
using std::cout;
using std::cerr;
using std::endl;

// first do simple access functions

// ---------------------------------------------------------
const Box&
NodeFArrayBox::box() const
{
  return m_box;
}

// ---------------------------------------------------------
FArrayBox& 
NodeFArrayBox::getFab()
{
  return m_fab;
}

// ---------------------------------------------------------
const FArrayBox& 
NodeFArrayBox::getFab() const
{
  return m_fab;
}


// ---------------------------------------------------------
// constructors and destructors
// ---------------------------------------------------------
// ": mfab()" means initialize m_fab with default constructor.
NodeFArrayBox::NodeFArrayBox() : m_fab()
{
}

// ---------------------------------------------------------
// initialize the fab to be aliased to the input array
NodeFArrayBox::NodeFArrayBox(const Box& a_bx, int a_nComp, Real* a_alias )
  : m_fab( surroundingNodes(a_bx) ,a_nComp ,a_alias )
{
  CH_assert(a_nComp > 0);
  CH_assert(!a_bx.isEmpty());
  m_box = a_bx;
}

// ---------------------------------------------------------
NodeFArrayBox::NodeFArrayBox(const Interval& a_comps,
                             NodeFArrayBox& a_original)
  : m_fab( a_comps, a_original.m_fab )
{
  m_box = a_original.m_box;
}

// ---------------------------------------------------------
NodeFArrayBox::~NodeFArrayBox()
{

}


// ---------------------------------------------------------
// define function
void
NodeFArrayBox::define(const Box& a_bx, int a_nComp)
{
 CH_assert(a_nComp > 0);
 CH_assert(!a_bx.isEmpty());
  m_box = a_bx;
  Box nodeBox(surroundingNodes(a_bx));
  m_fab.define(nodeBox, a_nComp);
}
      

// ---------------------------------------------------------
void 
NodeFArrayBox::copy(const NodeFArrayBox& a_src)
{
  m_fab.copy(a_src.m_fab);
}

// ---------------------------------------------------------
void 
NodeFArrayBox::copy(const Box& a_regionFrom,
                    const Interval& a_Cdest,
                    const Box& a_regionTo,
                    const NodeFArrayBox& a_src,
                    const Interval& a_Csrc)
{
 CH_assert(a_regionFrom.sameSize(a_regionTo));
  IntVect translator(a_regionTo.smallEnd() - a_regionFrom.smallEnd());
  const FArrayBox& srcFab = a_src.m_fab;
  // srcFab.box() should contain surroundingNodes(a_R)
  //CH_assert(srcFab.box().contains(surroundingNodes(a_R)));
  // Box copyBox = surroundingNodes(a_R);
  Box nodesFrom = surroundingNodes(a_regionFrom);
 CH_assert(srcFab.box().contains(nodesFrom));
  Box nodesTo = surroundingNodes(a_regionTo);
  //CH_assert(m_fab.box().contains(nodesTo));
  // added by petermc, 15 Nov 2002, based on FluxBox::copy:
  // safety check -- due to node centering,
  // nodesTo may not be contained in m_fab.box().
  Box thisBox = m_fab.box();
  nodesTo &= thisBox;
  // Also need this so that nodesTo and nodesFrom are same size.
  nodesFrom &= srcFab.box(); // previously &= thisBox 
  Box nodesToTranslated(nodesTo);
  nodesToTranslated.shift(-translator);
  nodesFrom &= nodesToTranslated;
  m_fab.copy(nodesFrom, a_Cdest, nodesTo, srcFab, a_Csrc);
}


// ---------------------------------------------------------
int 
NodeFArrayBox::size(const Box& a_R, const Interval& a_comps) const
{
  Box nodeBox(surroundingNodes(a_R));
  return m_fab.size(nodeBox, a_comps);
}

// ---------------------------------------------------------
void 
NodeFArrayBox::linearIn(void* a_buf, const Box& a_R, const Interval& a_comps)
{
  Box nodeBox(surroundingNodes(a_R));
  // added by petermc, 5 Dec 2002, to avoid problems with copyTo().
  Box thisBox = m_fab.box();
  nodeBox &= thisBox;
  m_fab.linearIn(a_buf, nodeBox, a_comps);
}

// ---------------------------------------------------------
// this is for broadcast & gather; the Box is in the msg
void 
NodeFArrayBox::linearIn(const void* const a_buf)
{
  // first get the box, then the number of components, then the FAB
  char * bufptr = static_cast<char*>(const_cast<void*>(a_buf)) ;
  ::linearIn( m_box ,bufptr )  ;  bufptr += ::linearSize( m_box );
  int ncomps ;
  ::linearIn( ncomps ,bufptr ) ;  bufptr += sizeof( ncomps );
  Box nodeBox = surroundingNodes( m_box ) ;
  m_fab.resize( nodeBox ,ncomps );
  m_fab.linearIn( bufptr, nodeBox, Interval( 0,ncomps-1 ) );
}


// ---------------------------------------------------------
void
NodeFArrayBox::linearOut(void* a_buf, const Box& a_R, const Interval& a_comps) const
{
  Box nodeBox(surroundingNodes(a_R));
  // added by petermc, 5 Dec 2002, to avoid problems with copyTo().
  Box thisBox = m_fab.box();
  nodeBox &= thisBox;
  m_fab.linearOut(a_buf, nodeBox, a_comps);
}

// ---------------------------------------------------------
// this is for broadcast & gather; the Box is in the msg
void
NodeFArrayBox::linearOut(void* const a_outBuf) const
{
  char * bufptr = static_cast<char*>(a_outBuf) ;
  ::linearOut( bufptr ,m_box )          ; bufptr += ::linearSize( m_box ) ;
  ::linearOut( bufptr ,m_fab.nComp() )  ; bufptr += sizeof(int) ;
  m_fab.linearOut( bufptr ,m_fab.box() ,m_fab.interval() );
}

// ---------------------------------------------------------
int
NodeFArrayBox::linearSize( ) const
{ // the linearization contains the cell-centered Box, #components, and FAB data
  return ::linearSize(m_box) + sizeof(int) + m_fab.size( m_fab.box() ,m_fab.interval() );
}
