/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _POOL_H_
#define _POOL_H_

#include <cstdlib>
#include <list>

#include "Vector.H"

class Pool;
typedef std::list<Pool*> PoolList;

///  Pool is a class to optimize memory allocation.
/**
   Pool is a class to optimize memory allocation.  It is specialized to
   allocate fixed size chunks of memory specified by ptrSize in the
   constructor.  Its operation is analogous to malloc, not new.  It
   does not initialize the memory in any way.  The constructor can
   optionally specify an initial pool size, and memory alignment.  The
   pool size will grow as needed by calling ::new.  The pool size never
   shrinks.  Memory will be reclaimed at ~Pool().  The units of poolSize
   are number-of-ptrSize-chunks.  The units of alignment are bytes.

   Pool can only be used for objects of fixed size.  Typically used in
   situations where a class or struct is being constantly constructed and
   deleted.  Objects returned by value from functions, elements in a database,
   etc.  Pool's tend to be allocated statically.

   (from Copier.cpp)
    \code

    // static data member s_motionIemPool getting constructed
    Pool Copier::s_motionItemPool(sizeof(MotionItem), "Copier::MotionItem");

    // note the use of a 'placement new'
    MotionItem* item = new (s_motionItemPool.getPtr()) MotionItem(fromdi, todi, box);

    // if your object does requires it's destructor to be called.
    item->~MotionItem();

    // then return the memory chunk to your pool
    s_motionItemPool.returnPtr(item)

    \endcode

Technical note
   In the event of multi-threading Chombo, we will have to make
   pool access serialized (locks, single thread access, etc) or implement a
   fast lock-free version, or go back to new/delete and let the OS be clever
   again about memory management.

*/
class Pool {
public:
  ///
  /**
     @param a_ptrSize Size of fixed memory allocations needed

     @param a_poolSize Size of allocations (a_ptrSize*a_poolSize) made
     of the operating system.  This controls the granularity of the
     Pool allocations.  Smaller values result in less wasted memory,
     at the cost of more calls to the operating system for heap memory.

     @param a_name optional name for this Pool.  Used in reporting memory
     by the memory tracking system in Chombo.
  */
  Pool(int         a_ptrSize,
       const char* a_name = "unnamed",
       int         a_poolSize = 100,
       int         a_alignment = sizeof(int));

  ///
  ~Pool();

  /// request a section of memory of  ptrSize_  contiguous bytes.
  void* getPtr();

  /// return memory previous acquired with the getPtr() function.
  void returnPtr(void* a_ptr);

  /// report how much memory this Pool is currently using.
  /**
     memUsage for a Pool only grows until Pool destruction. The Pool
     object has no knowledge of what pieces of memory it has parcelled out
     to a user, so it keeps it all available. The user is responsible for
     not leaking Pool memory.
  */
  long memUsage() const;

  /**
     undocumented function, call this at own risk.  You must
     be absolutely positive that you have returned all the
     ptr's you've asked for, or you can have major seg faults.
  */
  void clear();

  // not for public consumption. used in memory tracking code.
  static PoolList* m_poolList_;
  char             m_name_[64];


protected:


private:
  ///
  Vector<char*> m_pool_;

  ///
  int m_ptrSize_;

  ///
  int m_poolSize_;

  ///
  int m_alignment_;

  ///
  void* m_next_;

  ///
  void* getMoreMemory();

  void* getMore();

  ///
  /// Not implemented.  Compiler will not let you copy a Pool
  Pool(const Pool& a_rhs);

  /// Not implemented. Compiler will not let you copy a Pool
  const Pool& operator = (const Pool& a_rhs);

  static void clearAllPools();

  friend void dumpmemoryatexit();
};

// Stroustrup (S. 11.5.1) says a declaration in the global scope is needed
void dumpmemoryatexit();

#endif
