/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "Pool.H"
#include "MayDay.H"

PoolList* Pool::m_poolList_ = NULL;

bool PoolListInit = false;

// Constructor sets Pool parameters, but doesn't actually allocate any memory.
Pool::Pool(int         a_ptrSize,
           const char* a_name,
           int         a_poolSize,
           int         a_alignment)
  :
  m_pool_(0),                // An array of pointers to pools
  m_ptrSize_(a_ptrSize),     // Chunk size (bytes) in each pool
  m_poolSize_(a_poolSize),   // Number of chunks in each pool
  m_alignment_(a_alignment), // allocated memory to be aligned on this boundary.
  m_next_(0)                 // Points to next available chunk or is null.
{
  if (m_poolSize_ < 1)
  {
    MayDay::Error("Pool::Pool non positive poolSize.");
  }

  // Minimum chunk size is size of pointer.
  if ((unsigned int)m_ptrSize_ < sizeof(void*))
  {
    m_ptrSize_ = sizeof(void*);
  }

  // m_alignment_ checked out only for 4 and 8.
  if ((unsigned int)m_alignment_ < sizeof(int))
  {
    m_alignment_ = sizeof(int);
  }

  // chunk size must be multiple of m_alignment_
  while ((m_ptrSize_ % m_alignment_) != 0)
  {
    m_ptrSize_++;
  }

  if (!PoolListInit)
  {
    m_poolList_ = new PoolList;
    PoolListInit = true;
  }

  m_poolList_->insert(m_poolList_->begin(), this);
  strcpy(m_name_, a_name);
}

Pool::~Pool()
{
  for (int i = 0; i < m_pool_.size(); i++)
  {
    delete [] m_pool_[i];
  }

  m_pool_.resize(0);
  m_poolList_->remove(this);
}

/// getPtr and returnPtr must be as fast as possible!
void* Pool::getPtr()
{
  if (m_next_ == 0)
  {
    m_next_ = getMoreMemory();
  }

  void* result = m_next_;      // result points to first free chunk in list
  m_next_ = *(void**)m_next_;  // point m_next_ to next free chunk in list.

  return result;
}

void Pool::returnPtr(void* ptr)
{
  *(void**)ptr = m_next_;  // Stamp *ptr with the first free chunk.
  m_next_ = ptr;           // Make ptr the first free chunk.
}

//  notes on performing getPtr and returnPtr in a lock-free synchronized manner.
//  currently this design will suffer from the 'ABA' problem of CAS1-based
//  synchronization.  I need to use a CAS2 or 'double-compare-single-swap'

// void* Pool::getPtrAtomic()
// {
//   void* newNext = 0;
//   if (m_next_ == 0)
//     {
//       newNext = getMore();
//       if( CAS(&m_next_, 0, newNext))
//         m_pool_.push_back(po); // technically we should use a stack with atomic push
//       else
//         delete newNext; //some other thread beat me there, throw this one away.
//     }

//   void* result;
//   void* next;
//   void* check;
//   do{
//     result = m_next_;
//     next = *(void**)m_next_;
//   }while(CAS(&m_next_, result, next));

//   return result;
// }

// void Pool::returnPtrAtomic(void* ptr)
// {
//   void* next;
//   do{
//     next = m_next_;
//     *(void**)ptr = next;  // Stamp *ptr with the first free chunk.
//   }while (CAS( &m_next_, next,  ptr)); // Make ptr the first free chunk.
// }

long Pool::memUsage() const
{
  return m_ptrSize_ * m_poolSize_ * m_pool_.size();
}

void Pool::clear()
{
  for (int i = 0; i < m_pool_.size(); i++)
  {
    delete [] m_pool_[i];
  }

  m_next_ = 0;
  m_pool_.resize(0);
}

void Pool::clearAllPools()
{
  PoolList::iterator   p;
  for (p = Pool::m_poolList_->begin(); p != Pool::m_poolList_->end(); p++)
    {
      (*p)->clear();
    }
}

void* Pool::getMoreMemory()
{
  void* po = getMore();
  m_pool_.push_back((char*)po);
  return po;
}

void* Pool::getMore()
{
  // Create another pool
  char* po;

  if (m_alignment_ == sizeof(int))
  {
    po = (char*)new int [m_ptrSize_ * m_poolSize_ / sizeof(int)];
  }
  else if (m_alignment_ == sizeof(double))
  {
    po = (char*)new double [m_ptrSize_ * m_poolSize_ / sizeof(double)];
  }
  else
  {
    MayDay::Error("Pool::Pool : alignments other than int or double not supported");
  }

  // Thread pool:
  // The first few bytes in each chunk is interpreted as a pointer
  // which points to the next free chunk.  With multiple pools, and random
  // news and deletes, the pools will become threaded together.  This is ok.
  char* p = po;

  for (int i = 0; i < m_poolSize_ - 1; i++)
  {
    *(char**)p = p + m_ptrSize_;  // Chunk at i points to chunk at i+1.
    p += m_ptrSize_;
  }

  *(char**)p = 0;  // Chunk at end of list points to null
  return po;       // return first chunk.
}
