/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _REFCOUNTEDPTR_H_
#define _REFCOUNTEDPTR_H_

#include "MayDay.H"

///  A reference-counting handle class.
/**
        This is to be used as a pointer to class T.
        This will feel and smell just like a built-in pointer except:
        -#  There is no need to call delete on the pointer.
        -#  The default copy constructor and assignment implement ref-counting.
        -#  The user may call isNonUnique to determine if this pointer is
            the only pointer to the data.  This can be used to hide the
            ref-counting behavior of a class.
        -#  Checks for dereference of a null pointer.

This class is completely inlined.

typical usage:

\code

{
  Box b;
  IntVect a;
  //refCount() == 1
  RefCounterPtr<IntVectSet> set(new IntVectSet());

  // still just one IntVectSet, but refCount()==2
  RefCountedPtr<IntVectSet>  otherSet(set);

  // Like a pointer, modifying one modifies the other
  otherSet->define(b);
  (*set)|=a;
  {
    RefCountedPtr<IntVectSet> anotherSet; // null
    anotherSet = otherSet ;  //now all three have refCount()==3
  }//anotherSet out of scope, so deleted.  IntVectSet NOT deleted.

  // set.refCount() == 2
  // otherSet.refCount() == 2;
  // otherset == set;

}
// when all RefCountedPtr's for a given object are deleted, the last
// one calls 'delete' on the member pointer.
\endcode

*/
template <class T>
class RefCountedPtr {
public:
  //
  inline RefCountedPtr(T* ptr = 0);
  //
  inline RefCountedPtr(const RefCountedPtr<T>& other);
  //
  inline ~RefCountedPtr();
  /// assignement operator.  copies pointer member
  /** copies pointer member and integer pointer, decreases refcount of rhs member
      before assignment.  this refcount increased my one. */
  inline const RefCountedPtr<T>& operator =(const RefCountedPtr<T>& rhs);
  /// dereference access operator.  use like a pointer derefence access function.
  inline T* operator ->();
  inline const T* operator ->() const;
  //
  inline bool isNull() const;
  //
  /// pointer dereference.
  inline T& operator *();
  inline const T& operator *() const;
  //
  /// auto conversion to regular const pointer where required.
  inline operator const T* () const;

  /// true when refcount is one.
  inline bool isNonUnique() const;

  inline int refCount() const;

  inline void swap(RefCountedPtr<T>& b);

protected:
  T* ptr_;
  int* refCount_;
private:
};

template <class T> inline
RefCountedPtr<T>::RefCountedPtr(T* ptr)
  : ptr_(ptr),
    refCount_(0)
{
  if (ptr_)
    {
      refCount_ = new int;
      if (refCount_ == 0)
        MayDay::Error("RefCountedPtr::RefCountedPtr(T* ptr) out of memory");
      *refCount_ = 1;
    }
}

template <class T> inline
RefCountedPtr<T>::RefCountedPtr(const RefCountedPtr<T>& other)
  : ptr_(other.ptr_),
    refCount_(other.refCount_)
{
  if (refCount_ != 0)
    ++(*refCount_);
}

template <class T> inline
RefCountedPtr<T>::~RefCountedPtr() {
  if (refCount_ != 0 && --(*refCount_) == 0)
    {
      delete ptr_;
      ptr_ = 0;
      delete refCount_;
      refCount_ = 0;
    }
}

template <class T> inline
const RefCountedPtr<T>& RefCountedPtr<T>::operator =(const RefCountedPtr<T>& rhs) {
  if (ptr_ != rhs.ptr_)
    {
      if (refCount_ != 0 && --(*refCount_) == 0)
        {
          delete ptr_;
          delete refCount_;
        }
      ptr_ = rhs.ptr_;
      refCount_ = rhs.refCount_;
      if (refCount_ != 0)
        ++(*refCount_);
    }
  return *this;
}

template <class T> inline
bool RefCountedPtr<T>::isNull() const {
  return (ptr_ == 0);
}

template <class T> inline
T* RefCountedPtr<T>::operator ->() {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator ->() on null pointer");
  return ptr_;
}

template <class T> inline
const T* RefCountedPtr<T>::operator ->() const {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator ->() on null pointer");
  return ptr_;
}

template <class T> inline
T& RefCountedPtr<T>::operator *() {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator *() on null pointer");
  return *ptr_;
}

template <class T> inline
const T& RefCountedPtr<T>::operator *() const {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator *() on null pointer");
  return *ptr_;
}

template <class T> inline
RefCountedPtr<T>::operator const T* () const {
  return ptr_;
}

template <class T> inline
bool RefCountedPtr<T>::isNonUnique() const {
  return refCount_ == 0 ? false : *refCount_ != 1;
}

template <class T> inline
int RefCountedPtr<T>::refCount() const {
  return refCount_ == 0 ? 0 : *refCount_;
}

template <class T> inline
void RefCountedPtr<T>::swap(RefCountedPtr<T>& b)
{
  T* s = ptr_;
  ptr_ = b.ptr_;
  b.ptr_ = s;

  int* r = refCount_;
  refCount_ = b.refCount_;
  b.refCount_ = r;
}
#endif
