#ifdef CH_LANG_CC
/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//
#endif

#ifndef _SPACE_H_
#define _SPACE_H_

#ifdef CH_LANG_CC
/**
@ingroup macros
compile-time cpp macro that replaces 'assert()' and aborts
correctly when running in parallel. */
# include "MayDay.H"
# ifndef NDEBUG
#  define CH_assert_str(s) # s
#  define CH_assert_xstr(s) CH_assert_str(s)
#  define CH_assert(cond) if(!(cond)){ MayDay::Abort( __FILE__ ":" CH_assert_xstr(__LINE__) ": Assertion `" #cond "' failed."); }
# else
#  define CH_assert(cond) (void)0
#endif
#endif //CH_LANG_CC

#ifdef  CH_LANG_FORT
#ifndef NDEBUG
#  define CH_assert(cond) if(.NOT. cond) call MAYDAYERROR()
#else
#  define CH_assert(cond)
#endif
#endif //CH_LANG_FORT

#ifdef DOXYGEN
/** \defgroup macros Chombo Macros
 */

/**
@ingroup macros
compile-time cpp macro that controls the dimensionality of
    the Chombo package.  */
#define CH_SPACEDIM
#undef  CH_SPACEDIM
#define CH_SPACEDIM 3
#endif

#if ! defined(CH_SPACEDIM)
#error CH_SPACEDIM must be defined
#endif

#if (CH_SPACEDIM != 1 && CH_SPACEDIM !=2 && CH_SPACEDIM != 3)
#error CH_SPACEDIM must be either 1, 2, or 3
#endif

#if !defined(CH_LANG_FORT) || defined(CH_LANG_CC)

/**
  The SpaceDim C++ integer constant specifies the dimension of the
  Chombo library to be built.  It is initialized to the value of
  the macro CH_SPACEDIM.  Hence it will have one of the values 1, 2, or 3.

  C++ code that depends on the dimensionality of the problem that
  you're trying to solve should be cast in terms of this constant,
  instead of the CH_SPACEDIM macro, if possible, as constants are
  visible to debuggers while macros are not.
*/
   const int SpaceDim = CH_SPACEDIM;
#endif

#if CH_SPACEDIM==1

#  define D_EXPR(a,b,c) ((void)((a),0))

#  define D_DECL(a,b,c) a

#  define D_TERM(a,b,c) a
#elif CH_SPACEDIM==2
#  define D_EXPR(a,b,c) ((void)((a),(b),0))
#  define D_DECL(a,b,c) a,b
#  define D_TERM(a,b,c) a b
#elif CH_SPACEDIM==3

#ifdef DOXYGEN
/**
@ingroup macros

  The D_EXPR(a,b,c) macro expands to a comma expression that will
  evaluate SpaceDim of its arguments, counting from the left; i.e.
  CH_SPACEDIM==1 implies only `a' is evaluated, CH_SPACEDIM==2
  implies both `a' and `b', and CH_SPACEDIM==3 implies all three
  arguments are evaluated.  This macro can be used to enhance
  the portability of code that works with structures whose size
  depends on CH_SPACEDIM.

  For example:

     D_EXPR(vect[0] *= s, vect[1] *= s, vect[2] *= s);

  More explicitly, this macro evaluates to one of

     ((void)((a),0))

     ((void)((a),(b),0))

     ((void)((a),(b),(c),0))

  corresponding to CH_SPACEDIM values of 1, 2, or 3.
*/
#endif
#  define D_EXPR(a,b,c) ((void)((a),(b),(c),0))

#ifdef DOXYGEN
/**
@ingroup macros

  The D_DECL(a,b,c) macro expands to a comma-separated list of
  1, 2, or all 3 of the arguments of the call, depending on the
  value of CH_SPACEDIM.  This can be used to write portable
  function calls that depend on 1, 2, or 3 arguments, corresponding
  to the value of CH_SPACEDIM.

  For example:

     return IntVect(D_DECL(p[0] + s, p[1] + s, p[2] + s));

  More explicitly, this macro evaluates to one of

     a

     a,b

     a,b,c

  corresponding to CH_SPACEDIM values of 1, 2, or 3.
*/
#endif
#  define D_DECL(a,b,c) a,b,c

#ifdef DOXYGEN
/**
@ingroup macros
   define D_TERM(a,b,c) a b c

  More explicitly, this macro evaluates to one of

     a

     a b

     a b c

    corresponding to CH_SPACEDIM values of 1, 2, or 3.

    handy for situations like:

    Real norm = sqrt(D_TERM(u[0]*u[0],+u[1]*u[1],+u[2]*u[2]));
*/
#endif
#  define D_TERM(a,b,c) a b c
#endif

#ifdef DOXYGEN
#undef CH_SPACEDIM
#endif

#endif /*CH_SPACE_H*/
