/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <fstream>
#include <string>
#include <cstdio>
#include <cmath>
using std::fstream;
using std::string;

#include "CH_HDF5.H"
#include "UGIO.H"
#include "SPMD.H"

#ifdef CH_USE_HDF5

void WriteUGHDF5(const string&               a_filename,
                 const DisjointBoxLayout&    a_grids,
                 const LevelData<FArrayBox>& a_data,
                 const Box&                  a_domain)
{
  HDF5Handle handle(a_filename.c_str(), HDF5Handle::CREATE);
  WriteUGHDF5(handle, a_grids, a_data, a_domain);

#ifdef CH_MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif

  handle.close();
}

void WriteUGHDF5(HDF5Handle&                 a_handle,
                 const DisjointBoxLayout&    a_grids,
                 const LevelData<FArrayBox>& a_data,
                 const Box&                  a_domain)
{
  HDF5HeaderData header;
  int nComp = a_data.nComp();

  int numlevels = 1;
  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       =  numlevels;
  header.m_int ["num_components"]    = nComp;

  for (int ivar = 0; ivar < nComp; ivar++)
  {
    char labelChSt[80];
    sprintf(labelChSt, "component_%d", ivar);
    string label(labelChSt);
    header.m_string[label] = label;
  }

  header.writeToFile(a_handle);

  Real dtLevel = 1.0;
  Real dxLevel = 1.0;
  Real time = 1.0;
  int refLevel = 1;
  int ilev = 0;

  int eek = writeLevel(a_handle, ilev, a_data, dxLevel, dtLevel, time,
                       a_domain, refLevel);

  if (eek != 0)
  {
    MayDay::Error("WriteUGHDF5: Error in writeLevel");
  }
}

int ReadUGHDF5(const string&         a_filename,
               DisjointBoxLayout&    a_grids,
               LevelData<FArrayBox>& a_data,
               Box&                  a_domain)
{
  HDF5Handle handle;
  int err = handle.open(a_filename.c_str(),  HDF5Handle::OPEN_RDONLY);

  if (err < 0)
  {
    return -4;
  }

  int eekflag = ReadUGHDF5(handle, a_grids, a_data, a_domain);

#ifdef CH_MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif

  handle.close();
  return eekflag;
}

int ReadUGHDF5(HDF5Handle&           a_handle,
               DisjointBoxLayout&    a_grids,
               LevelData<FArrayBox>& a_data,
               Box&                  a_domain)
{
  HDF5HeaderData header;
  header.readFromFile(a_handle);

  int numLevels = header.m_int["num_levels"];
  if (numLevels != 1)
  {
    MayDay::Warning("Tried to read multilevel data using ReadUGHDF5");
    return -5;
  }

  int nComp = header.m_int["num_components"];
  if (nComp <= 0)
  {
    MayDay::Warning("ReadUGHDF5: Bogus number of Components");
    return -2;
  }

  int ilev = 0;
  Real dtLevel;
  Real dxLevel;
  Real time;
  int refLevel =1;

  int eek = readLevel(a_handle, ilev, a_data, dxLevel, dtLevel, time,
                      a_domain, refLevel, Interval(), true);

  if (eek != 0)
  {
    MayDay::Warning("ReadUGHDF5: readLevel failed");
    return -3;
  }

  a_grids = a_data.getBoxes();

  return 0;
}

#endif
