/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// BVS, June 18, 2003

// We can assume that template class T has null construction.

#ifndef _AMRMULTIGRID_H_
#define _AMRMULTIGRID_H_

#include "MultiGrid.H"
#include "REAL.H"
#include "Box.H"
#include "NoOpSolver.H"
#include "parstream.H"

template <class T>
class AMRLevelOp : public MGLevelOp<T>
{
public:
  virtual ~AMRLevelOp(){;}
  /** returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToCoarser() = 0;

  ///
  /**
      a_residual = a_rhs - L(a_phiFine, a_phi, a_phiCoarse)

  */
  virtual void AMRResidual(T& a_residual, const T& a_phiFine, const T& a_phi,
                           const T& a_phiCoarse, const T& a_rhs,
                           bool a_homogeneousDomBC) = 0;

  ///
  /** a_residual = a_rhs - L(a_phi, a_phiCoarse)  */
  virtual void AMRResidualNF(T& a_residual, const T& a_phi, const T& a_phiCoarse,
                             const T& a_rhs, bool a_homogeneousBC) = 0;

  ///
  /** a_residual = a_rhs - L(a_phiFine, a_phi)   no coaser AMR level*/
  virtual void AMRResidualNC(T& a_residual, const T& a_phiFine, const T& a_phi,
                             const T& a_rhs, bool a_homogeneousBC) = 0;


  ///
  /** a_resCoarse = I[h-2h]( a_residual - L(a_correction, a_coarseCorrection)) */
  virtual void AMRRestrict(T& a_resCoarse, const T& a_residual, const T& a_correction,
                           const T& a_coarseCorrection) = 0;

  ///
  /** a_correction += I[2h->h](a_coarseCorrection) */
  virtual void AMRProlong(T& a_correction, const T& a_coarseCorrection) = 0;

  ///
  /** a_residual = a_residual - L(a_correction, a_coarseCorrection) */
  virtual void AMRUpdateResidual(T& a_residual, const T& a_correction,
                                 const T& a_coarseCorrection) = 0;
  ///
  /**
     compute norm over all cells on coarse not covered by finer
   */
  virtual Real AMRNorm(const T&   a_coarseResid,
                       const T&   a_fineResid,
                       const int& a_refRat,
                       const int& a_ord) = 0;

  virtual void createCoarsened(T&       a_lhs,
                               const T& a_rhs,
                               const int&     a_refRat) = 0;

};

template <class T>
class AMRLevelOpFactory : public MGLevelOpFactory<T>
{
public:
  virtual ~AMRLevelOpFactory(){;}
  virtual AMRLevelOp<T>* AMRnewOp(const ProblemDomain& a_indexSpace)=0;
  virtual int refToFiner(const ProblemDomain& a_indexSpace) const =0;

};

template <class T>
class AMRMultiGrid
{
public:
  AMRMultiGrid();
  virtual ~AMRMultiGrid();
  virtual void define(const ProblemDomain& a_coarseDomain,
                      AMRLevelOpFactory<T>& a_factory,
                      LinearSolver<T>* a_bottomSolver,
                      int a_maxAMRLevels);

  virtual void solve(Vector<T*>& a_phi, const Vector<T*>& a_rhs,
                     int l_max, int l_base);

  void relaxOnly(Vector<T*>& a_phi, const Vector<T*>& a_rhs,
                 int l_max, int l_base);
  
  virtual void AMRVCycle(Vector<T*>& a_correction,
                         Vector<T*>& a_residual,
                         int l, int l_max, int l_base);

  void setMGCycle(int a_numMG);

  void init(const Vector<T*>& a_phi, const Vector<T*>& a_rhs);

  Real m_eps, m_hang, m_normThresh;
  int m_imin, m_iterMax, m_verbosity, m_exitStatus;
  int m_pre, m_post, m_bottom, m_numMG;

  AMRLevelOp<T>& levelOp(int level);

  Real computeAMRResidual(Vector<T*>&       a_resid,
                          Vector<T*>&       a_phi,
                          const Vector<T*>& a_rhs,
                          int               l_max, 
                          int               l_base,
                          bool              a_homogeneousBC=false);

  //for changing coefficients
  Vector< MGLevelOp<T>* > getAllOperators();
protected:

  void relax(T& phi, T& R, int depth);

  void computeAMRResidualLevel(Vector<T*>&       a_resid,
                               Vector<T*>&       a_phi,
                               const Vector<T*>& a_rhs,
                               int l_max, int l_base, int ilev, 
                               bool a_homogeneousBC);

  Vector<AMRLevelOp<T>*>          m_op;
  Vector<MultiGrid<T> >           m_mg;

  Vector<T*>  m_corrSave;
  Vector<T*>  m_correction;
  Vector<T*>  m_dCorr;
  Vector<T*>  m_residual;
  Vector<T*>  m_resC;

  NoOpSolver<T>    m_nosolve;

  LinearSolver<T>* m_bottomSolver;

  void clear();

};

//*******************************************************
// AMRMultigrid Implementation
//*******************************************************

//===================================================================

template <class T>
Vector< MGLevelOp<T> * > 
AMRMultiGrid<T>::getAllOperators()
{
  Vector< MGLevelOp<T>* > retval;
  for(int iop = 0;  iop < m_op.size(); iop++)
    {
      MGLevelOp<T>* operPtr = (MGLevelOp<T>*) m_op[iop];
      retval.push_back(operPtr);
    }
  
  for(int img = 0; img < m_mg.size(); img++)
    {
      Vector< MGLevelOp<T>* > mgOps = m_mg[img].getAllOperators();
      retval.append(mgOps);
    }
  return retval;
}

template <class T>
AMRMultiGrid<T>::AMRMultiGrid():m_eps(1E-6), m_hang(1E-15), m_normThresh(1E-30),
                                m_imin(5), m_iterMax(20), m_verbosity(3),
                                m_pre(2), m_post(2), m_bottom(2),
                                m_numMG(1),
                                m_bottomSolver(NULL){}

template <class T>
void AMRMultiGrid<T>::setMGCycle(int a_numMG)
{
  for(int ilev=0; ilev<m_op.size(); ilev++)
    {
      m_mg[ilev].m_numMG = a_numMG;
    }
  m_numMG = a_numMG;
}
template <class T>
void AMRMultiGrid<T>::relax(T& a_correction, T& a_residual, int depth)
{
  m_op[depth]->relax(a_correction, a_residual, m_pre);  //numSmoothDown

  if(m_op[depth]->refToCoarser() > 2)
    {
      int intermediateDepth = 0;
      int r = m_op[depth]->refToCoarser();
      while(r>2)
        {
          r/=2;
          intermediateDepth++;
        }
      int tmp = m_mg[depth].m_depth;
      m_mg[depth].m_depth = intermediateDepth;
      //ok, there is an intermediate multigrid level that is
      // not an AMR level.  We use regular MultiGrid smoothing algorithm for this.
      m_mg[depth].cycle(0, a_correction, a_residual);
      m_mg[depth].m_depth = tmp;
    }
}
/************/
template <class T>
AMRMultiGrid<T>::~AMRMultiGrid()
{
  clear();
}

template <class T>
AMRLevelOp<T>&  AMRMultiGrid<T>::levelOp(int level)
{
  return *(m_op[level]);
}

/************/
template <class T>
Real AMRMultiGrid<T>::computeAMRResidual(Vector<T*>&       a_resid,
                                         Vector<T*>&       a_phi,
                                         const Vector<T*>& a_rhs,
                                         int               l_max, 
                                         int               l_base,
                                         bool              a_homogeneousBC)
{
  Real rnorm = 0.0;
  for(int ilev=l_base; ilev<= l_max; ilev++)
    {
      //always used at top level where bcs are inhomogeneous
      computeAMRResidualLevel(a_resid,
                              a_phi,
                              a_rhs,
                              l_max, l_base, ilev, a_homogeneousBC);
      Real localNorm = 0.0;
      if(ilev == l_max)
        {
          localNorm =m_op[ilev]->norm(*a_resid[ilev], 0);
        }
      else
        {
          int refRat = m_op[ilev+1]->refToCoarser();
          localNorm =m_op[ilev]->AMRNorm(*a_resid[ilev],*a_resid[ilev+1], refRat, 0);
        }
      rnorm = Max(Abs(localNorm), rnorm);
    }
  return rnorm;
}
/************/
template <class T>
void AMRMultiGrid<T>::computeAMRResidualLevel(Vector<T*>&       a_resid,
                                              Vector<T*>&       a_phi,
                                              const Vector<T*>& a_rhs,
                                              int l_max, int l_base, int ilev,
                                              bool a_homogeneousBC)
{
  if(l_max != l_base)
    {
      if(ilev == l_max)
        {
          m_op[l_max]->AMRResidualNF(*(a_resid[l_max]), *(a_phi[l_max]),
                                     *(a_phi[l_max-1]), *(a_rhs[l_max]),
                                     a_homogeneousBC);
        }
      else if(ilev == l_base)
        {
          if(l_base == 0)
            {
              m_op[l_base]->AMRResidualNC(*(a_resid[l_base]), *(a_phi[l_base+1]),
                                          *(a_phi[l_base]),  *(a_rhs[l_base]),
                                           a_homogeneousBC);
            }
          else
            {
              m_op[l_base]->AMRResidual(*a_resid[l_base], *a_phi[l_base+1], *a_phi[l_base],
                                        *a_phi[l_base-1], *a_rhs[l_base],
                                         a_homogeneousBC);
            }
        }
      else
        {
          m_op[ilev]->AMRResidual(*a_resid[ilev], *a_phi[ilev+1], *a_phi[ilev],
                                  *a_phi[ilev-1], *a_rhs[ilev],
                                   a_homogeneousBC);
        }
    }
  else
    {
      CH_assert(ilev == l_base);
      if(l_base == 0)
        {
          m_op[l_max]->residual(*a_resid[l_max], *a_phi[l_max], *a_rhs[l_max],a_homogeneousBC);
        }
      else
        {
          m_op[l_max]->AMRResidualNF(*(a_resid[l_max]), *(a_phi[l_max]),
                                     *(a_phi[l_max-1]), *(a_rhs[l_max]),
                                      a_homogeneousBC);
        }
    }

}

template<class T>
void AMRMultiGrid<T>::solve(Vector<T*>& a_phi, const Vector<T*>& a_rhs,
                            int l_max, int l_base)
{
  init(a_phi, a_rhs);

  int nlevels = a_phi.size();
  CH_assert(a_rhs.size() == nlevels);
  for(int ilev=0; ilev<m_mg.size(); ilev++)
    {
      m_mg[ilev].m_pre=1;
      m_mg[ilev].m_post=1;
      m_mg[ilev].m_bottomSolver = &m_nosolve;
    }

  m_mg[l_base].m_pre  = m_pre;
  m_mg[l_base].m_post = m_post;
  m_mg[l_base].setBottomSolver(m_bottomSolver);

  //these correspond to the residual and correction
  //that live in AMRSolver
  Vector<T*> uberCorrection(a_rhs.size());
  Vector<T*> uberResidual(a_rhs.size());

  for(int ilev=0; ilev<m_op.size(); ilev++)
    {
      uberCorrection[ilev] = new T();
      uberResidual[ilev] = new T();
      m_op[ilev]->create(*uberCorrection[ilev], *a_phi[ilev]);
      m_op[ilev]->create(*uberResidual[ilev], *a_rhs[ilev]);
      m_op[ilev]->setToZero(*(uberCorrection[ilev]));
    }

  //compute initial residual and initialize internal residual to it
  Real initial_rnorm = computeAMRResidual(uberResidual, a_phi, a_rhs, l_max, l_base);
  Real rnorm = initial_rnorm;
  Real norm_last = 2*initial_rnorm;


  int iter=0;
  if(m_verbosity>=3)
    {
      pout() << "    AMRMultiGrid:: iteration = " << iter << ", error norm = " << rnorm << std::endl;
    }
  bool goNorm = rnorm > m_normThresh;                        //iterate if norm is not small enough
  bool goRedu = rnorm > m_eps*initial_rnorm;                 //iterate if initial norm is not reduced enough
  bool goIter = iter < m_iterMax;                            //iterate if iter < max iteration count
  bool goHang = iter < m_imin || rnorm <(1-m_hang)*norm_last;//iterate if we didn't hang
  while (goIter && goRedu && goHang && goNorm)
    {
      for(int ilev=l_base; ilev<= l_max; ilev++)
        {
          m_op[ilev]->assign(*m_residual[ilev], *uberResidual[ilev]);
        }

      norm_last = rnorm;


      //this generates a correction from the current residual
      AMRVCycle(uberCorrection, uberResidual, l_max, l_max, l_base);

      //increment phi by correction and reset correction to zero
      for(int ilev=l_base; ilev<= l_max; ilev++)
        {
          m_op[ilev]->incr(*(a_phi[ilev]), *(uberCorrection[ilev]), 1.0);
          m_op[ilev]->setToZero(*(uberCorrection[ilev]));
        }

      //recompute residual
      rnorm = computeAMRResidual(uberResidual, a_phi, a_rhs, l_max, l_base);
      iter++;
      if(m_verbosity>=4)
        {
          pout() << "    AMRMultiGrid:: iteration = " << iter << ", error norm = " << rnorm << std::endl;
        }
      
      goNorm = rnorm > m_normThresh;                        //keep iterating if norm is not small enough
      goRedu = rnorm > m_eps*initial_rnorm;                 //keep iterating if initial norm is not reduced enough
      goIter = iter < m_iterMax;                            //keep iterating if iter < max iteration count
      goHang = iter < m_imin || rnorm <(1-m_hang)*norm_last;//keep iterating if we didn't hang
    }
  m_exitStatus = int(!goRedu) + int(!goIter)*2 + int(!goHang)*4 + int(!goNorm)*8;
  if(m_verbosity==3)
    {
      pout() << "    AMRMultiGrid:: iteration = " << iter << ", error norm = " << rnorm << std::endl;
    }
  if(m_verbosity>1)
    {
      if(!goIter && goRedu && goNorm)
        {
          pout() << "    AMRMultiGrid:: WARNING: Exit because max iteration count exceeded" << std::endl;
        }
      if(!goHang && goRedu && goNorm)
        {
          pout() << "    AMRMultiGrid:: WARNING: Exit because of solver hang" << std::endl;
        }
      if(m_verbosity>4)
        {
          pout() << "    AMRMultiGrid:: exitStatus = " << m_exitStatus << std::endl;
        }
    }
  for(int i=0; i<m_op.size(); i++)
    {
      delete uberCorrection[i];
      delete uberResidual[i];
    }
}


template<class T>
void AMRMultiGrid<T>::relaxOnly(Vector<T*>& a_phi, const Vector<T*>& a_rhs,
                                int l_max, int l_base)
{
  CH_assert(l_max == l_base);
  CH_assert(l_max == 0);
  init(a_phi, a_rhs);

  int nlevels = a_phi.size();
  CH_assert(a_rhs.size() == nlevels);
  for(int ilev=0; ilev<m_mg.size(); ilev++)
    {
      m_mg[ilev].m_pre=1;
      m_mg[ilev].m_post=1;
      m_mg[ilev].m_bottomSolver = &m_nosolve;
    }

  m_mg[l_base].m_pre  = m_pre;
  m_mg[l_base].m_post = m_post;
  m_mg[l_base].setBottomSolver(m_bottomSolver);

  //these correspond to the residual and correction
  //that live in AMRSolver
  Vector<T*> uberCorrection(a_rhs.size());
  Vector<T*> uberResidual(a_rhs.size());

  for(int ilev=0; ilev<m_op.size(); ilev++)
    {
      uberCorrection[ilev] = new T();
      uberResidual[ilev] = new T();
      m_op[ilev]->create(*uberCorrection[ilev], *a_phi[ilev]);
      m_op[ilev]->create(*uberResidual[ilev], *a_rhs[ilev]);
      m_op[ilev]->setToZero(*(uberCorrection[ilev]));
    }

  //compute initial residual and initialize internal residual to it
  Real initial_rnorm = computeAMRResidual(uberResidual, a_phi, a_rhs, l_max, l_base);
  Real rnorm = initial_rnorm;
  Real norm_last = 2*initial_rnorm;


  int iter=0;
  if(m_verbosity>=3)
    {
      pout() << "    AMRMultiGrid::relaxOnly iteration = " << iter << ", error norm = " << rnorm << std::endl;
    }
  bool goNorm = rnorm > m_normThresh;                        //iterate if norm is not small enough
  bool goRedu = rnorm > m_eps*initial_rnorm;                 //iterate if initial norm is big enough
  bool goIter = iter < m_iterMax;                            //iterate if iter < max iteration count
  bool goHang = iter < m_imin || rnorm <(1-m_hang)*norm_last;//iterate if we didn't hang
  while (goIter && goRedu && goHang && goNorm)
    {

      norm_last = rnorm;

      relax(*(uberCorrection[0]), *(uberResidual[0]), 0);

      //increment phi by correction and reset correction to zero
      for(int ilev=l_base; ilev<= l_max; ilev++)
        {
          m_op[ilev]->incr(*(a_phi[ilev]), *(uberCorrection[ilev]), 1.0);
          m_op[ilev]->setToZero(*(uberCorrection[ilev]));
        }

      //recompute residual
      rnorm = computeAMRResidual(uberResidual, a_phi, a_rhs, l_max, l_base);
      iter++;
      if(m_verbosity>=4)
        {
          pout() << "    AMRMultiGrid::relaxOnly iteration = " << iter << ", error norm = " << rnorm << std::endl;
        }
      
      goNorm = rnorm > m_normThresh;                        //keep iterating if norm is not small enough
      goRedu = rnorm > m_eps*initial_rnorm;                 //keep iterating if initial norm is big enough
      goIter = iter < m_iterMax;                            //keep iterating if iter < max iteration count
      goHang = iter < m_imin || rnorm <(1-m_hang)*norm_last;//keep iterating if we didn't hang
    }
  m_exitStatus = int(!goRedu) + int(!goIter)*2 + int(!goHang)*4 + int(!goNorm)*8;
  if(m_verbosity==3)
    {
      pout() << "    AMRMultiGrid::relaxOnly iteration = " << iter << ", error norm = " << rnorm << std::endl;
    }
  if(m_verbosity>1)
    {
      if(!goIter && goRedu && goNorm)
        {
          pout() << "    AMRMultiGrid:: WARNING: Exit because max iteration count exceeded" << std::endl;
        }
      if(!goHang && goRedu && goNorm)
        {
          pout() << "    AMRMultiGrid:: WARNING: Exit because of solver hang" << std::endl;
        }
      if(m_verbosity>4)
        {
          pout() << "    AMRMultiGrid:: exitStatus = " << m_exitStatus << std::endl;
        }
    }
  for(int i=0; i<m_op.size(); i++)
    {
      delete uberCorrection[i];
      delete uberResidual[i];
    }
}

template <class T>
void AMRMultiGrid<T>::clear()
{
  for(int i=0; i<m_op.size(); i++)
    {
      delete m_corrSave[i]  ;
      delete m_correction[i];
      delete m_dCorr[i];
      delete m_residual[i];
      delete m_resC[i];
      delete m_op[i];
    }
}

template <class T>
void AMRMultiGrid<T>::init(const Vector<T*>& a_phi, const Vector<T*>& a_rhs)
{
  CH_assert(a_phi.size()<=m_op.size());
  for(int i=0; i<a_phi.size(); i++)
    {
      AMRLevelOp<T>& op = *(m_op[i]);
      op.create(*m_corrSave[i], *a_phi[i]);
      op.create(*m_correction[i], *a_phi[i]);
      op.create(*m_dCorr[i], *a_phi[i]);
      op.create(*m_residual[i], *a_rhs[i]);
      m_mg[i].init(*a_phi[i], *a_rhs[i]);

      int r = op.refToCoarser();
      op.createCoarsened(*m_resC[i], *a_rhs[i], r);
    }
}

template <class T>
void AMRMultiGrid<T>::define(const ProblemDomain& a_coarseDomain,
                             AMRLevelOpFactory<T>& a_factory,
                             LinearSolver<T>* a_bottomSolver,
                             int a_maxAMRLevels)
{
  this->clear();
  m_op.        resize( a_maxAMRLevels, NULL);
  m_mg.        resize( a_maxAMRLevels);

  m_corrSave.   resize( a_maxAMRLevels, NULL);
  m_correction.resize( a_maxAMRLevels, NULL);
  m_dCorr. resize( a_maxAMRLevels, NULL);
  m_residual.  resize( a_maxAMRLevels, NULL);
  m_resC.  resize( a_maxAMRLevels, NULL);
  m_bottomSolver = a_bottomSolver;

  ProblemDomain current = a_coarseDomain;
  for(int i=0; i<a_maxAMRLevels; i++)
    {
      m_corrSave[i]   = new T();
      m_correction[i] = new T();
      m_dCorr[i]  = new T();
      m_residual[i]   = new T();
      m_resC[i]       = new T();
      m_op[i] = a_factory.AMRnewOp(current);
      int maxDepth = -1; //coarsen all the way for now
      m_mg[i].define(a_factory, &m_nosolve, current, maxDepth);
      current.refine(a_factory.refToFiner(current));
    }
}

template<class T>
void AMRMultiGrid<T>::AMRVCycle(Vector<T*>& a_uberCorrection,
                                Vector<T*>& a_uberResidual,
                                int ilev, int l_max, int l_base)
{
  if(l_max == l_base)
    {
      CH_assert(ilev == l_base);
      m_mg[l_base].solve(*(a_uberCorrection[ilev]), *(a_uberResidual[ilev]));
    }
  else if(ilev == l_base)
    {
      int imaxBottom = 2;
      for(int i=0; i< imaxBottom; i++)
        {
          m_mg[l_base].solve(*(m_correction[ilev]), *(m_residual[ilev]));
        }
      m_op[ilev]->incr(*(a_uberCorrection[ilev]), *(m_correction[ilev]), 1.0);
    }
  else
    {
      m_op[ilev]->assign(*(m_corrSave[ilev]), *(a_uberCorrection[ilev]));
      m_op[ilev]->setToZero(*(m_correction[ilev]));

      //downsweep
      // in AMRLevelMG, it is a coarsened version of this level
      //but here we need to make it the same as the coarser
      //level so I will use increment instead of copyTo

      this->relax(*(m_correction[ilev]), *(m_residual[ilev]), ilev);
      m_op[ilev]->incr(*(a_uberCorrection[ilev]), *(m_correction[ilev]), 1.0);

      //set next level correction to zero
      m_op[ilev-1]->setToZero(*(m_correction[ilev-1]));

      // Compute residual for the next coarser level.
      //and average down  to next coarser level
      //this overwrites where  the coarse is covered by fine
      m_op[ilev]->AMRRestrict(*(m_resC[ilev]),
                              *(m_residual[ilev]),
                              *(m_correction[ilev]),
                              *(m_correction[ilev-1]));

      //recompute residual for next coarser level
      //need homogeneous BCs because in residual-correction form
      computeAMRResidualLevel(m_residual,
                              a_uberCorrection,
                              a_uberResidual,
                              l_max, l_base, ilev-1,
                              true);

      //overwrite next coarser's residual with coarsened residual from
      //this level
      m_op[ilev-1]->assign(*m_residual[ilev-1], *(m_resC[ilev]));

        for(int img = 0; img < m_numMG; img++)
          {
            AMRVCycle(a_uberCorrection, a_uberResidual, ilev-1, l_max, l_base);
          }

      //upsweep
      //increment the correction with coarser version
      m_op[ilev]->AMRProlong(*(m_correction[ilev]), *(m_correction[ilev-1]));
      //recompute residual
      m_op[ilev]->AMRUpdateResidual(*(m_residual[ilev]), *(m_correction[ilev]), *(m_correction[ilev-1]));

      //compute correction to the correction
      m_op[ilev]->setToZero(*(m_dCorr[ilev]));
      this->relax(*(m_dCorr[ilev]), *(m_residual[ilev]), ilev);

      //correct the correction with the correction to the correction
      m_op[ilev]->incr(*(m_correction[ilev]), *(m_dCorr[ilev]), 1.0);

      //put correction into saved correction and output correction
      m_op[ilev]->incr(*(m_corrSave[ilev]), *(m_correction[ilev]), 1.0);
      m_op[ilev-1]->assign(*(a_uberCorrection[ilev]), *(m_correction[ilev]));

    }
}

#endif
