/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "AMRPoissonOp.H"
#include "FORT_PROTO.H"
#include "AMRPoissonOpF_F.H"
#include "BoxIterator.H"
#include "AverageF_F.H"
#include "LayoutIterator.H"

void AMRPoissonOp::
createCoarsened(LevelData<FArrayBox>&       a_lhs,
                const LevelData<FArrayBox>& a_rhs,
                const int &                 a_refRat)
{
  int ncomp = a_rhs.nComp();
  IntVect ghostVect = a_rhs.ghostVect();

  DisjointBoxLayout dbl = a_rhs.disjointBoxLayout();
  CH_assert(dbl.coarsenable(a_refRat));

  //fill ebislayout
  DisjointBoxLayout dblCoarsenedFine;
  coarsen(dblCoarsenedFine, dbl, a_refRat);

  a_lhs.define(dblCoarsenedFine, ncomp, a_rhs.ghostVect());
}
/** full define function for AMRLevelOp with both coarser and finer levels */
void AMRPoissonOp::define(const DisjointBoxLayout& a_grids,
                          const DisjointBoxLayout& a_gridsFiner,
                          const DisjointBoxLayout& a_gridsCoarser,
                          const RealVect&          a_dxLevel,
                          int                      a_refRatio,
                          int                      a_refRatioFiner,
                          const ProblemDomain&     a_domain,
                          BCFunc                   a_bc,
                          int                      a_ncomp)
{
  this->define(a_grids, a_gridsCoarser, a_dxLevel, a_refRatio, a_domain, a_bc,
               a_ncomp);
  m_refToFiner = a_refRatioFiner;
  m_levfluxreg.define(a_gridsFiner, a_grids, a_domain, a_refRatioFiner, a_ncomp);
}

  /** full define function for AMRLevelOp with finer levels, but no coarser */
void AMRPoissonOp::define(const DisjointBoxLayout& a_grids,
                          const DisjointBoxLayout& a_gridsFiner,
                          const RealVect&          a_dxLevel,
                          int                      a_refRatio, // dummy argument
                          int                      a_refRatioFiner,
                          const ProblemDomain&     a_domain,
                          BCFunc                   a_bc,
                          int                      a_ncomp)
{
  CH_assert(a_refRatio == 1);
  this->define(a_grids, a_dxLevel, a_domain, a_bc); //calls the MG version of define
  m_refToFiner = a_refRatioFiner;
  m_levfluxreg.define(a_gridsFiner, a_grids, a_domain, a_refRatioFiner, a_ncomp);

}

void AMRPoissonOp::define(const DisjointBoxLayout& a_grids,
                          const DisjointBoxLayout& a_coarse,
                          const RealVect&          a_dxLevel,
                          int                      a_refRatio,
                          const ProblemDomain&     a_domain,
                          BCFunc                   a_bc,
                          int                      a_ncomp)
{
  this->define(a_grids, a_dxLevel, a_domain, a_bc); //calls the MG version of define
  m_quadCFI.define(a_grids, &a_coarse, a_dxLevel[0],
                   a_refRatio, a_ncomp, a_domain);  // then builds the additional material

  
  // m_averageOp.define(a_grids, a_coarse, a_ncomp, a_refRatio);

  m_refToCoarser = a_refRatio;
  m_dxCrse = a_refRatio*a_dxLevel;
  m_refToFiner = 1;

}

void AMRPoissonOp::define(const DisjointBoxLayout& a_grids, const RealVect& a_dx,
                          const ProblemDomain& a_domain, BCFunc a_bc)
{
  m_bc     = a_bc;
  m_domain = a_domain;
  m_dx     = a_dx;
  m_dxCrse = 2*a_dx;
  m_refToCoarser = 2; // redefined in AMRLevelOp::define virtual function.
  m_refToFiner   = 2;
  m_invDx  = 1.0/m_dx;
 
  m_invDx2 = m_invDx*m_invDx;
  m_alpha = 0.0;
  m_beta  = 1.0;
    
  m_exchangeCopier.define(a_grids, a_grids, IntVect::Unit, true);
  for(int i=0; i<CH_SPACEDIM; ++i)
    {
      LayoutData<NewCFIVS>& lo =  m_loCFIVS[i];
      LayoutData<NewCFIVS>& hi =  m_hiCFIVS[i];
      lo.define(a_grids);
      hi.define(a_grids);
      for(DataIterator dit(a_grids); dit.ok(); ++dit)
        {
          lo[dit].define(a_domain, a_grids.get(dit),a_grids, i, Side::Lo);
          hi[dit].define(a_domain, a_grids.get(dit),a_grids, i, Side::Hi);
        }
    }
}

void AMRPoissonOp::residual(LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_phi,
                            const LevelData<FArrayBox>& a_rhs, bool a_homogeneous)
{
  applyOp(a_lhs, a_phi, a_homogeneous);
  incr(a_lhs, a_rhs, -1);
  scale(a_lhs, -1.0);
}

/**************************/
// this preconditioner first initializes phihat to (IA)phihat = rhshat
// (diagonization of L -- A is the matrix version of L)
// then smooths with a couple of passes of levelGSRB
void AMRPoissonOp::preCond(LevelData<FArrayBox>& a_phi, const LevelData<FArrayBox>& a_rhs)
{
  // diagonal term of this operator is 4/h/h in 2D, 6/h/h in 3D,
  // so inverse of this is our initial multiplier
  Real mult = -m_dx[0]*m_dx[0]/(2.0*SpaceDim);
  Interval comps = a_phi.interval();
  CH_assert(a_phi.nComp() == a_rhs.nComp());

  // don't need to use a Copier -- plain copy will do
  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      a_phi[dit].copy(a_rhs[dit].box(), comps,
                                          a_rhs[dit].box(), a_rhs[dit],
                                          comps);

      a_phi[dit] *= mult;
    }

  relax(a_phi, a_rhs, 2);
}

void AMRPoissonOp::applyOp(LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_phi,
                           bool a_homogeneous )
{

  LevelData<FArrayBox>& phi = (LevelData<FArrayBox>&)a_phi;
  Real dx = m_dx[0];

  DataIterator dit = phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      m_bc(phi[dit], m_domain, dx, a_homogeneous);
    }

  phi.exchange(phi.interval(), m_exchangeCopier);

  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& region = a_lhs[dit].box();
      FORT_OPERATORLAP(CHF_FRA(a_lhs[dit]),
                       CHF_CONST_FRA(phi[dit]),
                       CHF_BOX(region),
                       CHF_CONST_REAL(dx));
    }
}

///compute norm over all cells on coarse not covered by finer
Real
AMRPoissonOp::AMRNorm(const LevelData<FArrayBox>& a_coarResid,
                      const LevelData<FArrayBox>& a_fineResid,
                      const int& a_refRat,
                      const int& a_ord)

{
  const DisjointBoxLayout& coarGrids = a_coarResid.disjointBoxLayout();
  const DisjointBoxLayout& fineGrids = a_fineResid.disjointBoxLayout();

  //create temp and zero out under finer grids
  LevelData<FArrayBox> coarTemp;
  m_levelOps.create(coarTemp, a_coarResid);
  m_levelOps.assign(coarTemp, a_coarResid);
  int ncomp = coarTemp.nComp();
  for(DataIterator dit = coarGrids.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& coarTempFAB = coarTemp[dit()];
      LayoutIterator litFine = fineGrids.layoutIterator();
      for (litFine.reset(); litFine.ok(); ++litFine)
        {
          Box overlayBox = coarTempFAB.box();
          Box coarsenedGrid = coarsen(fineGrids[litFine()], a_refRat);

          overlayBox &= coarsenedGrid;

          if (!overlayBox.isEmpty())
            {
              coarTempFAB.setVal(0.0,overlayBox,0, ncomp);
            }
        }
    }
  //return norm of temp
  return norm(coarTemp, a_ord);
}
void AMRPoissonOp::create(    LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_rhs)
{
  m_levelOps.create(a_lhs, a_rhs);
}

void AMRPoissonOp::assign(    LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_rhs)
{
  m_levelOps.assign(a_lhs, a_rhs);
}

Real AMRPoissonOp::dotProduct(const LevelData<FArrayBox>& a_1, const LevelData<FArrayBox>& a_2)
{
  return m_levelOps.dotProduct(a_1, a_2);
}
void AMRPoissonOp::incr( LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_x, Real a_scale)
{
  m_levelOps.incr(a_lhs, a_x, a_scale);
}
void AMRPoissonOp::axby( LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_x,
                         const LevelData<FArrayBox>& a_y, Real a_a, Real a_b)
{
  m_levelOps.axby(a_lhs, a_x, a_y, a_a, a_b);
}
void AMRPoissonOp::scale(LevelData<FArrayBox>& a_lhs, const Real& a_scale)
{
  m_levelOps.scale(a_lhs, a_scale);
}
void AMRPoissonOp::setToZero(LevelData<FArrayBox>& a_lhs)
{
  m_levelOps.setToZero(a_lhs);
}

void AMRPoissonOp::relax(LevelData<FArrayBox>& a_e,
                         const LevelData<FArrayBox>& a_residual,
                         int a_iterations)
{
  for(int i=0; i<a_iterations; i++){
    homogeneousCFInterp(a_e);
    levelJacobi(a_e, a_residual);
  }
}

void AMRPoissonOp::createCoarser(LevelData<FArrayBox>& a_coarse,
                                 const LevelData<FArrayBox>& a_fine,
                                 bool ghosted)
{
  // CH_assert(!ghosted);
  IntVect ghost = a_fine.ghostVect();
  DisjointBoxLayout dbl;
  CH_assert(dbl.coarsenable(2));
  coarsen(dbl, a_fine.disjointBoxLayout(), 2); //multigrid, so coarsen by 2
  a_coarse.define(dbl, a_fine.nComp(), ghost);
}

void AMRPoissonOp::restrictResidual(LevelData<FArrayBox>& a_resCoarse,
                                    LevelData<FArrayBox>& a_phiFine,
                                    const LevelData<FArrayBox>& a_rhsFine)
{
  homogeneousCFInterp(a_phiFine);
  Real dx = m_dx[0];
  for(DataIterator dit = a_phiFine.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& phi = a_phiFine[dit];
      m_bc(phi,  m_domain,  dx, true);
    }

  a_phiFine.exchange(a_phiFine.interval());

  for(DataIterator dit = a_phiFine.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& phi = a_phiFine[dit];
      const FArrayBox& rhs = a_rhsFine[dit];
      FArrayBox& res = a_resCoarse[dit];

      Box region = rhs.box();

      res.setVal(0.0);
      FORT_RESTRICTRES(CHF_CONST_FRA(phi),
                       CHF_CONST_FRA(rhs),
                       CHF_FRA(res),
                       CHF_BOX(region),
                       CHF_CONST_REAL(dx));

    }
}

Real AMRPoissonOp::norm(const LevelData<FArrayBox>& a_x, int a_ord)
{
  return ::norm(a_x, a_x.interval(), a_ord);
}

void AMRPoissonOp::prolongIncrement(LevelData<FArrayBox>& a_phiThisLevel,
                                    const LevelData<FArrayBox>& a_correctCoarse)
{
  for(DataIterator dit = a_phiThisLevel.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& phi =  a_phiThisLevel[dit];
      const FArrayBox& coarse = a_correctCoarse[dit];
      Box region = a_phiThisLevel.box(dit);
      const Box& cBox = a_correctCoarse.box(dit);
      CH_assert(cBox == coarsen(region, m_refToCoarser));

      FORT_PROLONG(CHF_FRA(phi),
                   CHF_CONST_FRA(coarse),
                   CHF_BOX(region),
                   CHF_CONST_INT(m_refToCoarser));

    }
}

void AMRPoissonOp::levelGSRB(LevelData<FArrayBox>& a_phi,
                             const LevelData<FArrayBox>& a_rhs)
{
  CH_assert(a_phi.isDefined());
  CH_assert(a_rhs.isDefined());
  CH_assert(a_phi.ghostVect() >= IntVect::Unit);
  CH_assert(a_phi.nComp() == a_rhs.nComp());

  // do first red, then black passes
  for (int whichPass =0; whichPass <= 1; whichPass++)
    {

      homogeneousCFInterp(a_phi);
      // now step through grids...
      Real dx = m_dx[0];
      DataIterator dit = a_phi.dataIterator();
      for (dit.begin(); dit.ok(); ++dit)
        {
          // invoke physical BC's where necessary
          m_bc(a_phi[dit],  m_domain,  dx, true);
        }
      //fill in intersection of ghostcells and a_phi's boxes
      a_phi.exchange(a_phi.interval(), m_exchangeCopier);

      //fill in intersection of ghostcells and a_phi's boxes
      // dfm -- for a 5 point stencil, this should not be necessary
      //a_phi.exchange(a_phi.interval(), m_exchangeCopier);

      for (dit.begin(); dit.ok(); ++dit)
        {

#ifndef NDEBUG
//          FArrayBox& thisPhi = a_phi[dit];
//          const FArrayBox& thisRhs = a_rhs[dit];
//          const Box& thisBox =  a_phi.box(dit);
//
#endif

          FORT_GSRBLEVELLAP(CHF_FRA(a_phi[dit]),
                            CHF_CONST_FRA(a_rhs[dit]),
                            CHF_BOX(a_rhs.box(dit)),
                            CHF_CONST_REAL(dx),
                            CHF_CONST_INT(whichPass));

        } // end loop through grids

          //      ArrayViewLDFFormatLabel(&a_phi,"%7.5e","phismooth3");
    } // end loop through red-black

}
void AMRPoissonOp::levelJacobi(LevelData<FArrayBox>& a_phi,
                               const LevelData<FArrayBox>& a_rhs)
{
  LevelData<FArrayBox> resid;
  create(resid, a_rhs);
  // Get the residual
  residual(resid,a_phi,a_rhs,true);

  // Create the weight
  Real weight = m_alpha;
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      weight += -2.0 * m_beta * m_invDx2[idir];
    }
  // Do the Jacobi relaxation
  //Real one = 1.0;
  //axby(a_phi,a_phi,resid,one,0.5/weight);
  incr(a_phi, resid, 0.5/weight);
}

void
AMRPoissonOp::homogeneousCFInterp(LevelData<FArrayBox>& a_phif)
{

  CH_assert( a_phif.ghostVect() >= IntVect::Unit);

  DataIterator dit = a_phif.dataIterator();
  for(dit.begin(); dit.ok(); ++dit)
    {
      const DataIndex& datInd = dit();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          SideIterator sit;
          for(sit.begin(); sit.ok(); sit.next())
            {
              homogeneousCFInterp(a_phif,datInd,idir,sit());
            }
        }
    }
}

void
AMRPoissonOp::homogeneousCFInterp(LevelData<FArrayBox>& a_phif,
                                  const DataIndex& a_datInd,
                                  int a_idir,
                                  Side::LoHiSide a_hiorlo)
{

  CH_assert((a_idir >= 0) && (a_idir  < SpaceDim));
  CH_assert((a_hiorlo == Side::Lo )||(a_hiorlo == Side::Hi ));
  CH_assert( a_phif.ghostVect() >= IntVect::Unit);
  //  CH_assert (m_ncomp == a_phif.nComp());

  const NewCFIVS* cfivs_ptr = NULL;
  if(a_hiorlo == Side::Lo)
    cfivs_ptr = &m_loCFIVS[a_idir][a_datInd];
  else
    cfivs_ptr = &m_hiCFIVS[a_idir][a_datInd];

  if(cfivs_ptr->isPacked())
    {
      int ihiorlo = sign(a_hiorlo);
      FORT_INTERPHOMO(CHF_FRA(a_phif[a_datInd]),
                      CHF_BOX(cfivs_ptr->packedBox()),
                      CHF_CONST_REAL(m_dx[a_idir]),
                      CHF_CONST_REAL(m_dxCrse[a_idir]),
                      CHF_CONST_INT(a_idir),
                      CHF_CONST_INT(ihiorlo));
    }
  else
    {
      const IntVectSet& interp_ivs = cfivs_ptr->getFineIVS();
      if(!interp_ivs.isEmpty())
        {
          // Assuming homogenous, interpolate on fine ivs
          interpOnIVSHomo(a_phif, a_datInd, a_idir,
                          a_hiorlo, interp_ivs);
        }
    }
}

void
AMRPoissonOp::interpOnIVSHomo(LevelData<FArrayBox>& a_phif,
                              const DataIndex& a_datInd,
                              const int a_idir,
                              const Side::LoHiSide a_hiorlo,
                              const IntVectSet& a_interpIVS)
{

  CH_assert((a_idir >= 0) && (a_idir  < SpaceDim));
  CH_assert((a_hiorlo == Side::Lo )||(a_hiorlo == Side::Hi ));
  CH_assert( a_phif.ghostVect() >= IntVect::Unit);
  IVSIterator fine_ivsit(a_interpIVS);
  FArrayBox& a_phi = a_phif[a_datInd];

  // much of these scalar values can be precomputed and stored if
  // we ever need to speed-up this function (ndk)
  Real x1 = m_dx[a_idir];
  Real x2 = 0.5*(3. * x1 + m_dxCrse[a_idir]);
  Real denom = 1.0-((x1+x2)/x1);
  Real idenom = 1/(denom); // divide is more expensive usually
  Real x = 2.*x1;
  Real xsquared = x*x;

  Real m1 = 1/(x1*x1);
  Real m2 = 1/(x1*(x1-x2));

  Real q1 = 1/(x1-x2);
  Real q2 = x1+x2;

  int ihilo = sign(a_hiorlo);
  Real pa,pb,a,b;
  IntVect ivf;
  for(fine_ivsit.begin(); fine_ivsit.ok(); ++fine_ivsit)
    {
      ivf = fine_ivsit();
      // quadratic interpolation
      for(int ivar = 0; ivar < a_phif.nComp(); ivar++)
        {
          ivf[a_idir]-=2*ihilo;
          pa = a_phi(ivf, ivar);
          ivf[a_idir]+=ihilo;
          pb = a_phi(ivf, ivar);

          a = ((pb-pa)*m1 - (pb)*m2)*idenom;
          b = (pb)*q1 - a*q2;

          ivf[a_idir]+=ihilo;
          a_phi(fine_ivsit(), ivar) = a*xsquared + b*x + pa;

        } //end loop over components
    } //end loop over fine intvects
}

void AMRPoissonOp::AMRResidualNF(LevelData<FArrayBox>& a_residual,
                                 const LevelData<FArrayBox>& a_phi,
                                 const LevelData<FArrayBox>& a_phiCoarse,
                                 const LevelData<FArrayBox>& a_rhs,
                                 bool a_homogeneousPhysBC)
{
  LevelData<FArrayBox>& phi = (LevelData<FArrayBox>&)a_phi;
  m_quadCFI.coarseFineInterp(phi, a_phiCoarse);
  this->residual(a_residual, a_phi, a_rhs, a_homogeneousPhysBC ); //apply boundary conditions
}

void AMRPoissonOp::getFlux(FArrayBox&       a_flux,
                           const FArrayBox& a_data,
                           int              a_dir,
                           int ref) const
{

  CH_assert(a_dir >= 0);
  CH_assert(a_dir <  SpaceDim);
  CH_assert(!a_data.box().isEmpty());

  Box edgebox = surroundingNodes(a_data.box(),a_dir);
  edgebox.grow(a_dir, -1);

  // if this fails, the data box was too small (one cell wide, in fact)
  CH_assert(!edgebox.isEmpty());

  a_flux.resize(edgebox, a_data.nComp());
  BoxIterator bit(edgebox);

  RealVect dx(m_dx);
  dx/=ref;
  for( bit.begin(); bit.ok(); bit.next())
    {
      IntVect iv = bit();
      IntVect shiftiv = BASISV(a_dir);
      IntVect ivlo = iv - shiftiv;
      IntVect ivhi = iv;

      CH_assert(a_data.box().contains(ivlo));
      CH_assert(a_data.box().contains(ivhi));

      for (int ivar = 0; ivar < a_data.nComp(); ivar++)
        {
          Real phihi = a_data(ivhi,ivar);
          Real philo = a_data(ivlo,ivar);
          Real gradphi =(phihi-philo)/dx[ivar];

          a_flux(iv,ivar) = gradphi;
        }
    }
}

void AMRPoissonOp::reflux(const LevelData<FArrayBox>& a_phiFine,
                          const LevelData<FArrayBox>& a_phi,
                          LevelData<FArrayBox>& residual)
{
  m_levfluxreg.setToZero();
  Interval interv(0,a_phi.nComp()-1);

  DataIterator dit = a_phi.dataIterator();
  for (dit.reset(); dit.ok(); ++dit)
    {
      const FArrayBox& coarfab = a_phi[dit];

      for (int idir = 0; idir < SpaceDim; idir++)
        {
          FArrayBox coarflux;
          getFlux(coarflux, coarfab, idir);

          Real scale = 1.0;
          m_levfluxreg.incrementCoarse(coarflux,
                                       scale,dit(),
                                       interv,interv,idir);
        }
    }
  LevelData<FArrayBox>& p = ( LevelData<FArrayBox>&)a_phiFine;
  p.exchange(a_phiFine.interval());

  DataIterator ditf = a_phiFine.dataIterator();
  for (ditf.reset(); ditf.ok(); ++ditf)
    {
      const FArrayBox& phifFab = a_phiFine[ditf];

      for (int idir = 0; idir < SpaceDim; idir++)
        {
          SideIterator sit;
          for (sit.begin(); sit.ok(); sit.next())
            {
              Side::LoHiSide hiorlo = sit();
              Box fabbox;

              if (sit() == Side::Lo)
                {
                  fabbox = adjCellLo(phifFab.box(),idir, 2);
                  fabbox.shift(idir, 2);
                }
              else
                {
                  fabbox = adjCellHi(phifFab.box(),idir, 2);
                  fabbox.shift(idir, -2);
                }

              CH_assert(!fabbox.isEmpty());

              FArrayBox phifab(fabbox, a_phi.nComp());
              phifab.copy(phifFab);

              FArrayBox fineflux;
              getFlux(fineflux, phifab, idir, m_refToFiner);

              Real scale = 1.0;
              m_levfluxreg.incrementFine(fineflux, scale, ditf(),
                                         interv, interv, idir, hiorlo);
            }
        }
    }

  Real scale = - 1.0/m_dx[0];
  m_levfluxreg.reflux(residual, scale);
}

void AMRPoissonOp::AMRResidual(LevelData<FArrayBox>& a_residual,
                               const LevelData<FArrayBox>& a_phiFine,
                               const LevelData<FArrayBox>& a_phi,
                               const LevelData<FArrayBox>& a_phiCoarse,
                               const LevelData<FArrayBox>& a_rhs,
                               bool a_homogeneousPhysBC)
{
  AMRResidualNF(a_residual, a_phi, a_phiCoarse, a_rhs, a_homogeneousPhysBC);
  reflux(a_phiFine, a_phi, a_residual);
}

void AMRPoissonOp::AMRResidualNC(LevelData<FArrayBox>& a_residual,
                                 const LevelData<FArrayBox>& a_phiFine,
                                 const LevelData<FArrayBox>& a_phi,
                                 const LevelData<FArrayBox>& a_rhs,
                                 bool a_homogeneousPhysBC)
{
  this->residual(a_residual, a_phi, a_rhs, a_homogeneousPhysBC); //apply boundary conditions
  reflux(a_phiFine, a_phi, a_residual);
}

void AMRPoissonOp::AMRRestrict(LevelData<FArrayBox>& a_resCoarse,
                               const LevelData<FArrayBox>& a_residual,
                               const LevelData<FArrayBox>& a_correction,
                               const LevelData<FArrayBox>& a_coarseCorrection)
{
  
  LevelData<FArrayBox> r;
  create(r, a_residual);
  AMRResidualNF(r, a_correction, a_coarseCorrection, a_residual, true);
  //m_averageOp.averageToCoarse(a_resCoarse, r);
  
  DataIterator dit = a_residual.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      FArrayBox& coarse = a_resCoarse[dit];
      const FArrayBox& fine = r[dit];
      const Box& b = coarse.box();
      Box refbox(IntVect::Zero,
                 (m_refToCoarser-1)*IntVect::Unit);
      FORT_AVERAGE( CHF_FRA(coarse),
                    CHF_CONST_FRA(fine),
                    CHF_BOX(b),
                    CHF_CONST_INT(m_refToCoarser),
                    CHF_BOX(refbox)
                    );
    }
  
}

/** a_correction += I[2h->h](a_coarseCorrection) */
void AMRPoissonOp::AMRProlong(LevelData<FArrayBox>& a_correction,
                              const LevelData<FArrayBox>& a_coarseCorrection)
{
  DisjointBoxLayout c;
  coarsen(c,  a_correction.disjointBoxLayout(), m_refToCoarser);
  LevelData<FArrayBox> e(c, a_correction.nComp(),a_coarseCorrection.ghostVect());
  a_coarseCorrection.copyTo(e.interval(), e, e.interval());
  prolongIncrement(a_correction, e);
}

void AMRPoissonOp::AMRUpdateResidual(LevelData<FArrayBox>& a_residual,
                                     const LevelData<FArrayBox>& a_correction,
                                     const LevelData<FArrayBox>& a_coarseCorrection)
{
  LevelData<FArrayBox> r;
  this->create(r, a_residual);
  this->AMRResidualNF(r, a_correction, a_coarseCorrection, a_residual, true);
  this->assign(a_residual, r);
}

//==========================================================
//
//==========================================================

// MultiGrid define function

void AMRPoissonOpFactory::define(const ProblemDomain& a_domain,
                                 const DisjointBoxLayout& a_grid,
                                 const RealVect& a_dx,
                                 BCFunc a_bc,
                                 int ncomp,
                                 int maxDepth)
{

  m_bc = a_bc;
  m_boxes.resize(1, a_grid);
  m_dx.resize(1, a_dx);
  m_ncomp = ncomp;
  m_domains.resize(1, a_domain);
}

AMRPoissonOp*
AMRPoissonOpFactory::MGnewOp(const ProblemDomain& a_indexSpace,
                             int depth,
                             bool homoOnly)
{
  // CH_assert(m_boxes.size()>depth);

  int ref=0;
  for(;ref< m_domains.size(); ref++)
    {
      if(a_indexSpace.domainBox() == m_domains[ref].domainBox()) break;
    }
  CH_assert(ref !=  m_domains.size()); // didn't find domain

  DisjointBoxLayout layout(m_boxes[ref]);
  ProblemDomain domain(m_domains[ref]);
  RealVect dx = m_dx[ref];

  for(int i=0; i<depth; i++)
    {
      if(!layout.coarsenable(4)) return NULL;
      DisjointBoxLayout dbl;
      coarsen_dbl(dbl, layout, 2);
      layout = dbl;
      dx*=2;
      domain.coarsen(2);
    }

  AMRPoissonOp* newOp = new AMRPoissonOp;
  newOp->define(layout, dx, domain, m_bc);
  return newOp;
}

AMRPoissonOp* AMRPoissonOpFactory::AMRnewOp(const ProblemDomain& a_indexSpace)
{

  AMRPoissonOp* newOp = new AMRPoissonOp;
  int ref = 0;
  for(;ref< m_domains.size(); ref++)
    {
      if(a_indexSpace.domainBox() == m_domains[ref].domainBox()) break;
    }

  if(ref == 0)
    {
      // coarsest AMR level
      newOp->define(m_boxes[0], m_boxes[1], m_dx[0], 1, m_refRatios[1],
                    a_indexSpace, m_bc,  m_ncomp);
    }
  else if (ref ==  m_domains.size()-1)
    {
      // finest AMR level
      newOp->define(m_boxes[ref], m_boxes[ref-1], m_dx[ref], m_refRatios[ref],
                   a_indexSpace, m_bc,  m_ncomp);
    }
  else if ( ref == m_domains.size())
    {
      MayDay::Error("Did not find a domain to match AMRnewOp(const ProblemDomain& a_indexSpace)");

    }
  else
    {
      // intermediate AMR level, full define
      newOp->define(m_boxes[ref], m_boxes[ref+1], m_boxes[ref-1], m_dx[ref], m_refRatios[ref],
                    m_refRatios[ref+1], a_indexSpace, m_bc,  m_ncomp);
    }

  return newOp;
}

//  AMR Factory define function
void AMRPoissonOpFactory::define(const ProblemDomain& a_coarseDomain,
                                 const Vector<DisjointBoxLayout>& a_grids,
                                 const Vector<int> a_refRatios,
                                 const RealVect& a_coarsedx,
                                 int a_ncomp,
                                 BCFunc a_bc)
{
  m_domains.resize(a_grids.size());
  m_boxes=a_grids;
  m_refRatios=a_refRatios;
  m_dx.resize(a_grids.size());
  m_ncomp = a_ncomp;
  m_bc = a_bc;
  m_domains[0] = a_coarseDomain;
  m_dx[0] = a_coarsedx;
  for(int i=1; i<a_grids.size(); i++)
    {
      m_dx[i] = m_dx[i-1]/m_refRatios[i] ;
      m_domains[i] = m_domains[i-1];
      m_domains[i].refine(m_refRatios[i]);
    }

}

int AMRPoissonOpFactory::refToFiner(const ProblemDomain& a_indexSpace) const

{

  int ref = 0;
  for(;ref< m_domains.size(); ref++)
    {
      if(a_indexSpace.domainBox() == m_domains[ref].domainBox())
        {
          if(ref == m_domains.size()-1) return 1;
          return m_refRatios[ref+1];
        }
    }
  MayDay::Error(" no domain found for AMRPoissonOpFactory::refToFiner");
  return 0;
}
