/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "BCFunc.H"
#include "RealVect.H"
//#include "BCFuncF_F.H"

//compute value of a such that a parabola fits constant spaced a, b, c and passes
//through that value ab at a midpoint between a and b.

// solves for value of "a" given:
// u(x=-1)   = *a
// u(x=-1/2) = ab
// u(x=0)    = *b
// u(x=1)    = *c

inline
void quadraticInterp(Real* a, const Real* b, const Real* c, const Real& ab)
{
  *a = (8.0/3.0)*ab + (1.0/3.0)*(*c) -2*(*b);
}

inline
void linearInterp(Real* a, const Real* b, const Real* c, const Real& ab)
{
  // ignore value at x=1
  *a = *b + 2*(ab - *b);
}

void NeumBC(FArrayBox& a_state,
            const Box& a_valid,
            Real dx,
            bool a_homogeneous,
            BCValueFunc a_value,
            int a_dir,
            Side::LoHiSide a_side)
{
  Vector<Real> values(a_state.nComp());
  int comps = a_state.nComp();
  RealVect pos;
  int s = sign(a_side);
  if(a_side == Side::Lo)
    {
      Box to_region = adjCellLo(a_valid, a_dir, 1);
      Box fromRegion = to_region;
      fromRegion.shift(a_dir, 1);
      a_state.copy(a_state, fromRegion, 0, to_region, 0, a_state.nComp());
      if(!a_homogeneous){
        ForAllXBNN(Real,a_state,to_region,0, comps)
          {
            D_TERM(pos[0]=dx*(iR+0.5);, pos[1]=dx*(jR+0.5);, pos[2]=dx*(kR+0.5));
            pos[a_dir]+=dx/2.0;
            a_value(&(pos[0]), &a_dir, &s, &(values[0]), &comps);
            a_stateR += dx*values[nR];
          } EndFor ;
      }

    }
  if(a_side == Side::Hi )
    {
      Box to_region = adjCellHi(a_valid, a_dir, 1);
      Box fromRegion = to_region;
      fromRegion.shift(a_dir, -1);
      a_state.copy(a_state, fromRegion, 0, to_region, 0, a_state.nComp());
      if(!a_homogeneous){
        ForAllXBNN(Real,a_state,to_region,0, comps)
          {
            D_TERM(pos[0]=dx*(iR+0.5);, pos[1]=dx*(jR+0.5);, pos[2]=dx*(kR+0.5));
            pos[a_dir]-=dx/2.0;
            a_value(&(pos[0]), &a_dir, &s, &(values[0]), &comps);
            a_stateR += dx*values[nR];
          } EndFor ;
      }

    }
}
void GenBC(FArrayBox& a_state,
           const Box& a_valid,
           Real dx,
           bool a_homogeneous,
           GenBCFunc a_func,
           int a_dir,
           Side::LoHiSide a_side)
{
  const Box& box  = a_state.box();
  Real *l1, *l2, *lghost;
  RealVect pos1;
  Vector<Real> values(a_state.nComp());
  int comps = a_state.nComp();
  Box edge ;
  int s = a_side;
  int hom = a_homogeneous;

  if(a_side == Side::Lo)
    { s=0; edge = adjCellLo(a_valid, a_dir, 1); edge.shift(a_dir, 1);}
  else
    { s=1; edge = adjCellHi(a_valid, a_dir, 1); edge.shift(a_dir,-1);}

  int incr;
  if(a_dir==0){
    incr     = -sign(a_side);
  }
  else if(a_dir==1){
    incr     = -sign(a_side)*box.size(0);
  }
  else {
    incr     = -sign(a_side)*box.size(0)*box.size(1);
  }
  ForAllXBNN(Real,a_state,edge,0, comps)
    {
      a_stateR += 0 * nR; // shuts up compiler "unused variable" warning
      l1  =_x_pp;
      l2 =_x_pp+incr;
      lghost = _x_pp-incr;
      D_TERM(pos1[0]=dx*(iR+0.5);, pos1[1]=dx*(jR+0.5);, pos1[2]=dx*(kR+0.5));
      pos1[a_dir]-=dx/2.0;
      a_func(&(pos1[0]), &a_dir, &s, &hom, l1, l2, lghost);
    } EndFor ;

}

void DiriBC(FArrayBox& a_state,
            const Box& valid,
            Real dx,
            bool homogeneous,
            BCValueFunc a_value,
            int order)
{
  const Box& box  = a_state.box();
  Real *h1, *h, *hghost, *l1, *l, *lghost;
  RealVect pos1, pos2;
  Real deltaPos;
  Vector<Real> values(a_state.nComp());
  int comps = a_state.nComp();

  CH_assert(order == 1 || order == 2);

  for(int i=0; i<CH_SPACEDIM; ++i)
    {
      Box lo = adjCellLo(valid, i, 1); lo.shift(i, 1);
      int hi, incr;
      if(i==0){
        incr     = 1;
        hi       = box.size(0)-2;
        deltaPos = box.size(0)*dx;
      }
      else if(i==1){
        incr     = box.size(0);
        hi       = box.size(0)*(box.size(1)-2);
        deltaPos = box.size(1)*dx;
      }
      else {
        incr     = box.size(0)*box.size(1);
        hi       = box.size(0)*box.size(1)*(box.size(2)-2);
        deltaPos = box.size(2)*dx;
      }
      if(!homogeneous){
        ForAllXBNN(Real,a_state,lo,0, comps)
          {
            a_stateR+=0;
            l  =_x_pp;
            l1 =_x_pp+incr;
            lghost = _x_pp-incr;
            h  = _x_pp+hi;
            h1 = h-incr;
            hghost = _x_pp+hi+incr;
            D_TERM(pos1[0]=dx*(iR+0.5);, pos1[1]=dx*(jR+0.5);, pos1[2]=dx*(kR+0.5));
            pos1[i]-=dx/2.0;
            pos2=pos1;
            pos2[i]+=deltaPos;
            int s = -1;
            a_value(&(pos1[0]), &i, &s, &(values[0]), &comps);
            if(order == 2)
              quadraticInterp(lghost, l, l1, values[nR]);
            else
              linearInterp(lghost, l, l1, values[nR]);
            s=1;
            a_value(&(pos2[0]),  &i, &s, &(values[0]), &comps);
            quadraticInterp(hghost, h, h1, values[nR]);
            //linearInterp(lghost, l, l1, values[nR]);
          } EndFor ;
      }
      else {
        ForAllXBNN(Real,a_state,lo,0, comps)
          {
            a_stateR+=0 *nR * iR;//shuts up "unused variable" warning
            l  =_x_pp;
            l1 =_x_pp+incr;
            lghost = _x_pp-incr;
            h  = _x_pp+hi;
            h1 = h-incr;
            hghost = _x_pp+hi+incr;
            if(order == 2){
              quadraticInterp(lghost, l, l1, 0);
              quadraticInterp(hghost, h, h1, 0);
            } else {
              linearInterp(lghost, l, l1, 0);
              linearInterp(hghost, h, h1, 0);
            }
          }EndFor ;
      }

    }
}

void DiriBC(FArrayBox& a_state,
            const Box& valid,
            Real dx,
            bool homogeneous,
            BCValueFunc a_value,
            int a_dir,
            Side::LoHiSide a_side,
            int order)
{
  CH_assert(order == 1 || order == 2);
  
  const Box& box  = a_state.box();
  Real *l1, *l, *lghost;
  RealVect pos1;
  Vector<Real> values(a_state.nComp());
  int comps = a_state.nComp();
  Box edge ;
  int s = sign(a_side);
  if(a_side == Side::Lo)
    { edge = adjCellLo(valid, a_dir, 1); edge.shift(a_dir, 1);}
  else
    { edge = adjCellHi(valid, a_dir, 1); edge.shift(a_dir,-1);}

  int incr;
  if(a_dir==0){
    incr     = -sign(a_side);
  }
  else if(a_dir==1){
     incr     = -sign(a_side)*box.size(0);
  }
  else {
    incr     = -sign(a_side)*box.size(0)*box.size(1);
  }
  if(!homogeneous){
    ForAllXBNN(Real,a_state,edge,0, comps)
      {
        a_stateR+=0;
        l  =_x_pp;
        l1 =_x_pp+incr;
        lghost = _x_pp-incr;
        D_TERM(pos1[0]=dx*(iR+0.5);, pos1[1]=dx*(jR+0.5);, pos1[2]=dx*(kR+0.5));
        pos1[a_dir]+=sign(a_side)*dx/2.0;
        a_value(&(pos1[0]), &a_dir, &s, &(values[0]), &comps);
        if(order == 2)
          quadraticInterp(lghost, l, l1, values[nR]);
        else
          linearInterp(lghost, l, l1, values[nR]);
      } EndFor ;
  }
  else {
    ForAllXBNN(Real,a_state,edge,0, comps)
      {
        a_stateR += 0 *nR * iR; //shuts up "unused variable" warning
        l  =_x_pp;
        l1 =_x_pp+incr;
        lghost = _x_pp-incr;
        if(order == 2)
          quadraticInterp(lghost, l, l1, 0);
        else
          linearInterp(lghost, l, l1, 0);
      }EndFor ;
  }
}
