/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// BVS, June 18, 2003

#ifndef _BICGSTABSOLVER_H_
#define _BICGSTABSOLVER_H_

#include "LinearSolver.H"

template <class T>
class BiCGStabSolver : public LinearSolver<T>
{
public:
  BiCGStabSolver();
  virtual ~BiCGStabSolver();

  virtual void setHomogeneous(bool a_homogeneous)
     {m_homogeneous = a_homogeneous;}
  virtual void define(LinearOp<T>* a_op, bool a_homogeneous);
  virtual void solve(T& a_phi, const T& a_rhs);

  bool m_homogeneous;
  LinearOp<T>* m_op;
  int m_imax;
  int m_verbosity;
  Real m_eps;
  Real m_hang;
  Real m_small;
  int m_numRestarts;
  int m_normType;

};

//*******************************************************
// BiCGStabSolver Implementation
//*******************************************************

template <class T>
BiCGStabSolver<T>::BiCGStabSolver()
  :m_homogeneous(false), m_op(NULL), m_imax(40), m_verbosity(3), m_eps(1.0E-6), m_hang(1E-8), 
   m_small(1.0E-30), m_numRestarts(5), m_normType(2){;}

template <class T>
BiCGStabSolver<T>::~BiCGStabSolver()
{
  m_op = NULL;
}

template <class T>
void BiCGStabSolver<T>::define(LinearOp<T>* a_operator, bool a_homogeneous)
{
  m_homogeneous = a_homogeneous;
  m_op = a_operator;
}

template <class T>
void BiCGStabSolver<T>::solve(T& a_phi, const T& a_rhs)
{
  T r, r_tilde, e, p, p_tilde, s_tilde, t, v;

  m_op->create(r, a_rhs);
  m_op->create(r_tilde, a_rhs);
  m_op->create(e, a_phi);
  m_op->create(p, a_rhs);
  m_op->create(p_tilde, a_phi);
  m_op->create(s_tilde, a_phi);
  m_op->create(t, a_rhs);
  m_op->create(v, a_rhs);

  int recount = 0;

 CH_assert(m_op != NULL);

  m_op->residual(r, a_phi, a_rhs, m_homogeneous);

  m_op->assign(r_tilde, r);
  m_op->setToZero(e);
  int i=0;

  Real rho[4]={0,0,0,0}; // rho[0] = r_i , rho[1] = r_(i-1), etc.
  Real norm[2];
  norm[0] = m_op->norm(r, m_normType);
  Real initial_norm = norm[0];
  norm[1] = norm[0];

  Real alpha[2] = {0,0};
  Real beta[2] = {0,0};
  Real omega[2] ={0,0};

  bool init = true;
  int restarts = 0;

  if (m_verbosity >= 3) 
    {
      pout() << "    BiCGStab:: initial Residual norm = " 
             << initial_norm << std::endl;
    }

  while((i<m_imax && norm[0] > m_eps*norm[1]) && (norm[1] > 0))
    {
      i++;

      norm[1] = norm[0];
      alpha[1]= alpha[0];
      beta[1] = beta[0];
      omega[1]= omega[0];

      rho[3] = rho[2];
      rho[2] = rho[1];
      rho[1] = m_op->dotProduct(r_tilde, r);
      if(rho[1] == 0)
        {
          // we are finished, we will not converge anymore
          m_op->incr(a_phi, e, 1.0);
          if (m_verbosity >= 3) 
            {
              pout() << "    BiCGStab:: rho = 0, returning"
                     << " -- Residual norm = " 
                     << norm[0] << std::endl;
            }

          return;
        }
      if(init)
        {
          m_op->assign(p, r);
          init = false;
        }
      else
        {
          beta[1] = (rho[1]/rho[2])*(alpha[1]/omega[1]);
          m_op->scale(p, beta[1]);
          m_op->incr(p, v, -beta[1]*omega[1]);
          m_op->incr(p, r, 1.0);
        }
      m_op->preCond(p_tilde, p);
      m_op->applyOp(v, p_tilde, true);
      Real m = m_op->dotProduct(r_tilde, v);
      alpha[0] = rho[1]/m;
      if(Abs(m) > m_small*Abs(rho[1]))
        {
          m_op->incr(r, v, -alpha[0]);
          norm[0] = m_op->norm(r, m_normType);
          m_op->incr(e, p_tilde, alpha[0]);
        }
      else
        {
          m_op->setToZero(r);
          norm[0] = 0.0;
          
        }
      if(norm[0] > m_eps*initial_norm)
        {
          m_op->preCond(s_tilde, r);
          m_op->applyOp(t, s_tilde, true);
          omega[0] = m_op->dotProduct(t, r)/m_op->dotProduct(t, t);
          m_op->incr(e, s_tilde, omega[0]);
          m_op->incr(r, t,      -omega[0]);
          norm[0] = m_op->norm(r, m_normType);
        }

      if(m_verbosity >= 4)
        {
          pout() << "    BiCGStab::    iteration = "  << i        << ", error norm = " << norm[0] << std::endl;
        }

      if(norm[0] < m_eps*initial_norm) // converged to tolerance
        {
          break;
        }
      if(omega[0] == 0.0              ||
         norm[0] > (1-m_hang)*norm[1])
        {
          if(recount == 0){
            recount = 1;
            
          }else {
            recount = 0;
            m_op->incr(a_phi, e, 1.0);
            if(restarts == m_numRestarts)
              {
                if (m_verbosity >= 3) 
                  {
                    pout() << "    BiCGStab:: max number of restarts reached"
                           << " -- Residual norm = "
                           << norm[0] << std::endl;
                  }
                
                return;
              }
            m_op->residual(r, a_phi, a_rhs, m_homogeneous);
            norm[0] = m_op->norm(r, m_normType);
            rho[1]=0.0; rho[1]=0.0; rho[2]=0.0; rho[3]=0.0;
            alpha[0]=0; beta[0]=0; omega[0]=0;
            m_op->assign(r_tilde, r);
            m_op->setToZero(e);
            
            restarts ++;
            if(m_verbosity >= 4)
              {
                pout() << "    BiCGStab::      restart =  " << restarts << std::endl;
              }
            init = true;
          }
        }
    }

  if (m_verbosity >= 3) 
    {
      pout() << "    BiCGStab:: " << i << " iterations, final Residual norm = "
             << norm[0] << std::endl;
    }

  m_op->incr(a_phi, e, 1.0);
}

#endif /*_BICGSTABSOLVER_H_*/
