/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// BVS, June 18, 2003

// We can assume that template class T has null construction.

#ifndef _MULTIGRID_H_
#define _MULTIGRID_H_

#include "LinearSolver.H"
#include "ProblemDomain.H"
#include "REAL.H"
#include "Box.H"

///
/**
   class MGLevelOp handles the additional tasks of coordinating operations
   between this level, and the level one coarser than this 'level'
*/
template <class T>
class MGLevelOp : public LinearOp<T>
{
public:
  ///
  /**
   */
  virtual ~MGLevelOp(){;}

  ///
  /**
   */
  virtual void createCoarser(T& a_coarse, const T& a_fine, bool ghosted) = 0;

  ///
  /**
   */
  virtual void relax(T& a_correction, const T& a_residual, int a_iterations) = 0 ;

  ///
  /**
     calculate restricted residual
     a_resCoarse[2h] = I[h->2h] (rhsFine[h] - L[h](phiFine[h])
  */
  virtual void restrictResidual(T& a_resCoarse, T& a_phiFine, const T& a_rhsFine) = 0;

  ///
  /**
     correct the fine solution based on coarse correction
     a_phiThisLevel += I[2h->h](a_correctCoarse)
  */
  virtual void prolongIncrement(T& a_phiThisLevel, const T& a_correctCoarse) = 0;
};

///
/**
 */
template <class T>
class MGLevelOpFactory /* :  public LinearOpFactory<T> */
{
public:
  virtual ~MGLevelOpFactory(){;}

  ///returns NULL if no such Multigrid level can be created at this a_depth
  virtual MGLevelOp<T>* MGnewOp(const ProblemDomain& a_FineindexSpace,
                                int                  a_depth,
                                bool                 a_homoOnly = true) = 0;
};

template <class T>
class MultiGrid
{
public:
  MultiGrid();
  virtual ~MultiGrid();

  //go all the way to the bottom if maxdepth = -1
  virtual void define(MGLevelOpFactory<T>& a_factory,
                      LinearSolver<T>* a_bottomSolver,
                      const ProblemDomain& a_domain,
                      int  a_maxDepth = -1);

  /** residual-correction form of solution
      L(a_e) = a_res
  */
  virtual void solve(T& a_e, const T& a_res);

  void init(const T& a_correction, const T& a_residual); // used by AMRMultiGrid

  void cycle(int a_depth, T& a_correction, const T& a_residual);

  void clear();

  void setBottomSolver(LinearSolver<T>* a_bottomSolver);

  int  m_depth, m_pre, m_post, m_cycle,m_numMG;
  bool m_homogeneous;
  LinearSolver<T>*  m_bottomSolver;

  //for changing coefficients 
  Vector< MGLevelOp<T>* > getAllOperators();

protected:
  bool m_defined;

  ProblemDomain m_topLevelDomain;
  Vector<MGLevelOp<T>*> m_op;
  Vector< T* >   m_residual;
  Vector< T* >   m_correction;
};

//*******************************************************
// MultiGrid Implementation
//*******************************************************

template <class T>
Vector< MGLevelOp<T>* > 
MultiGrid<T>::getAllOperators()
{
  Vector<MGLevelOp<T>*> retval = m_op;
  return retval;
}

template <class T>
MultiGrid<T>::MultiGrid():
   m_pre(3), m_post(3), m_cycle(1), m_numMG(1), m_homogeneous(true),
  m_defined(false){;}

template <class T>
MultiGrid<T>::~MultiGrid()
{
  clear();
}

template <class T>
void MultiGrid<T>::define(MGLevelOpFactory<T>& a_factory,
                          LinearSolver<T>*     a_bottomSolver,
                          const ProblemDomain& a_domain,
                          int                  a_maxDepth)
{

  m_depth = 0;
  m_bottomSolver = a_bottomSolver;
  m_residual.resize(m_depth);   // the zero'th entry is left null constructed
  m_correction.resize(m_depth); // the zero'th entry is left null constructed
  m_op.resize(m_depth);
  m_topLevelDomain = a_domain;

  MGLevelOp<T>* nextOp =  a_factory.MGnewOp(a_domain, 0, true);
  while(nextOp != NULL)
    {
      m_residual.push_back(new T());
      m_correction.push_back(new T());
      m_op.push_back(nextOp);
      m_depth++;
      if((m_depth < a_maxDepth) || (a_maxDepth < 0))
        {
          nextOp = a_factory.MGnewOp(a_domain , m_depth, true);
        }
      else
        {
          nextOp = NULL;
        }
    }
  m_bottomSolver->define(m_op[m_depth-1], true);
  m_defined = true;
}

template <class T>
void MultiGrid<T>::setBottomSolver(LinearSolver<T>* a_bottomSolver)
{
  m_bottomSolver = a_bottomSolver;
  m_bottomSolver->define(m_op[m_depth-1], true);
}

template <class T>
void MultiGrid<T>::init(const T& a_e, const T& a_residual)
{
    if (m_depth > 1)
  {
    m_op[0]->createCoarser(*(m_residual[1]), a_residual, false);
    m_op[0]->createCoarser(*(m_correction[1]), a_e, true);
  }

  for (int i = 2; i < m_depth; i++)
    {
      m_op[i-1]->createCoarser(*(m_residual[i]), *(m_residual[i-1]), false);
      m_op[i-1]->createCoarser(*(m_correction[i]), *(m_correction[i-1]), true);
    }
}

template <class T>
void MultiGrid<T>::solve(T& a_e, const T& a_residual)
{
  this->init(a_e, a_residual);

  if (m_homogeneous)
    {
      cycle(0, a_e, a_residual);
    }
  else
    {
      T correction, residual;
      m_op[0]->create(correction, a_e);
      m_op[0]->create(residual, a_residual);
      m_op[0]->residual(residual, a_e, a_residual);
      m_op[0]->setToZero(correction);
      this->cycle(0, correction, residual);
      m_op[0]->incr(a_e, correction, 1.0);
    }
}

template <class T>
void MultiGrid<T>::cycle(int depth, T& correction, const T& residual)
{
  if(depth == m_depth-1)
    {
      m_bottomSolver->solve(correction, residual);
    }
  else
    {

      MGLevelOp<T>* op = m_op[depth];
      int cycles = m_cycle;
      if(depth == m_depth-1) cycles=1;
      for(int c=0; c<cycles; ++c){

        op->relax(correction, residual, m_pre);
        op->restrictResidual(*(m_residual[depth+1]), correction, residual);
        m_op[depth+1]->setToZero(*(m_correction[depth+1]));

        for(int img = 0; img < m_numMG; img++)
          {
            cycle(depth+1, *(m_correction[depth+1]), *(m_residual[depth+1]));
          }

        op->prolongIncrement(correction, *(m_correction[depth+1]));
        op->relax(correction, residual, m_post);

      }
    }

}

template <class T>
void MultiGrid<T>::clear()
{
  for(int i=0; i<m_op.size(); ++i)
    {
      delete m_op[i];
      delete m_correction[i];
      delete m_residual[i];
    }
  m_op.resize(0);
  m_residual.resize(0);
  m_correction.resize(0);
  m_defined = false;
}

#endif /*_LINEARSOLVER_H_*/
