/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "NewPoissonOp.H"
#include "FORT_PROTO.H"
#include "AMRPoissonOpF_F.H"

void NewPoissonOp::define( const RealVect& a_dx,
                           const ProblemDomain& a_domain,
                           BCFunc a_bc)
{
  m_bc     = a_bc;
  m_domain = a_domain;
  m_dx     = a_dx;
}

void NewPoissonOp::residual(  FArrayBox& a_lhs, const FArrayBox& a_phi,
                              const FArrayBox& a_rhs, bool a_homogeneous)
{
  applyOp(a_lhs, a_phi, a_homogeneous);
  a_lhs*=-1.0;
  incr(a_lhs, a_rhs, 1);
}

/**************************/
// this preconditioner first initializes phihat to (IA)phihat = rhshat
// (diagonization of L -- A is the matrix version of L)
// then smooths with a couple of passes of levelGSRB
void NewPoissonOp::preCond(   FArrayBox& a_phi, const FArrayBox& a_rhs)
{
    // diagonal term of this operator is 4/h/h in 2D, 6/h/h in 3D,
  // so inverse of this is our initial multiplier
  Real mult = -m_dx[0]*m_dx[0]/(2.0*SpaceDim);
  Interval comps = a_phi.interval();
 CH_assert(a_phi.nComp() == a_rhs.nComp());

  a_phi.copy(a_rhs.box(), comps,
             a_rhs.box(), a_rhs,
             comps);

  a_phi *= mult;

  relax(a_phi, a_rhs, 2);
}

void NewPoissonOp::applyOp(   FArrayBox& a_lhs, const FArrayBox& a_phi,
                              bool a_homogeneous )
{

  FArrayBox& phi = (FArrayBox&)a_phi;
  Real dx = m_dx[0];

  m_bc(phi, m_domain, dx, a_homogeneous);

  const Box& region = a_lhs.box();
  FORT_OPERATORLAP(CHF_FRA(a_lhs),
                   CHF_CONST_FRA(phi),
                   CHF_BOX(region),
                   CHF_CONST_REAL(dx));

}

void NewPoissonOp::create(    FArrayBox& a_lhs, const FArrayBox& a_rhs)
{
  a_lhs.define(a_rhs.box(), a_rhs.nComp());
}

void NewPoissonOp::assign(    FArrayBox& a_lhs, const FArrayBox& a_rhs)
{
  a_lhs.copy(a_rhs);
}

Real NewPoissonOp::dotProduct(const FArrayBox& a_1, const FArrayBox& a_2)
{
  return a_1.dotProduct(a_2);
}
void NewPoissonOp::incr( FArrayBox& a_lhs, const FArrayBox& a_x, Real a_scale)
{
  a_lhs.plus(a_x, a_scale);
}
void NewPoissonOp::axby( FArrayBox& a_lhs, const FArrayBox& a_x,
                         const FArrayBox& a_y, Real a_a, Real a_b)
{
  a_lhs.copy(a_x);
  a_lhs*=a_a;
  a_lhs.plus(a_y, a_b);
}
void NewPoissonOp::scale(FArrayBox& a_lhs, const Real& a_scale)
{
  a_lhs*=a_scale;
}
void NewPoissonOp::setToZero(FArrayBox& a_lhs)
{
  a_lhs.setVal(0.0);
}

Real NewPoissonOp::norm(const FArrayBox& a_x, int a_ord)
{
  return a_x.norm(a_ord);
}

void NewPoissonOp::relax(FArrayBox& a_e,
                         const FArrayBox& a_residual,
                         int a_iterations)
{
  for(int i=0; i<a_iterations; i++){
    levelGSRB(a_e, a_residual);
  }
}

void NewPoissonOp::createCoarser(FArrayBox& a_coarse,
                                 const FArrayBox& a_fine,
                                 bool ghosted)
{
  Box c = a_fine.box();
  if(ghosted)
    {
      c.grow(-1);
    }
  c.coarsen(2);
  c.refine(2);
  if(ghosted) c.grow(1);
 CH_assert(c == a_fine.box());
  if(ghosted) c.grow(-1);
  c.coarsen(2);
  if(ghosted)
    {
      c.grow(1);
    }
  a_coarse.define(c, a_fine.nComp());

}

void NewPoissonOp::restrictResidual(FArrayBox& a_resCoarse,
                                    FArrayBox& a_phiFine,
                                    const FArrayBox& a_rhsFine)
{

  Real dx = m_dx[0];

  FArrayBox& phi = a_phiFine;
  m_bc(phi,  m_domain,  dx, true);

  const FArrayBox& rhs = a_rhsFine;
  FArrayBox& res = a_resCoarse;

  Box region = rhs.box();

  res.setVal(0.0);
  FORT_RESTRICTRES(CHF_CONST_FRA(phi),
                   CHF_CONST_FRA(rhs),
                   CHF_FRA(res),
                   CHF_BOX(region),
                   CHF_CONST_REAL(dx));

}

void NewPoissonOp::prolongIncrement(FArrayBox& a_phiThisLevel,
                                    const FArrayBox& a_correctCoarse)
{

  FArrayBox& phi =  a_phiThisLevel;
  const FArrayBox& coarse = a_correctCoarse;
  //FArrayBox& c = (FArrayBox&)coarse;
  Box region = a_phiThisLevel.box();
  region.grow(-1);
  Box cBox = a_correctCoarse.box();
  cBox.grow(-1);
 CH_assert(cBox == coarsen(region, 2));

  int r = 2;

  FORT_PROLONG(CHF_FRA(phi),
               CHF_CONST_FRA(coarse),
               CHF_BOX(region),
               CHF_CONST_INT(r));
}

void NewPoissonOp::levelGSRB(FArrayBox& a_phi,
                             const FArrayBox& a_rhs)
{

 CH_assert(a_phi.nComp() == a_rhs.nComp());

  Real dx = m_dx[0];
  m_bc(a_phi,  m_domain,  dx, true);

  // do first red, then black passes
  for (int whichPass =0; whichPass <= 1; whichPass++)
    {

      // now step through grids...
      //fill in intersection of ghostcells and a_phi's boxes
      // dfm -- for a 5 point stencil, this should not be necessary
      //a_phi.exchange(a_phi.interval(), m_exchangeCopier);

      // invoke physical BC's where necessary
      //m_bc(a_phi,  m_domain,  dx, true); //new approach to help with checker
#ifndef NDEBUG

#endif

          FORT_GSRBLEVELLAP(CHF_FRA(a_phi),
                            CHF_CONST_FRA(a_rhs),
                            CHF_BOX(a_rhs.box()),
                            CHF_CONST_REAL(dx),
                            CHF_CONST_INT(whichPass));

    } // end loop through red-black

}

NewPoissonOpFactory::NewPoissonOpFactory()
{
  m_dx  = RealVect::Unit;
  m_bc  = NULL;
}

NewPoissonOpFactory::NewPoissonOpFactory(RealVect& a_dx, BCFunc a_bc)
  :m_dx(a_dx), m_bc(a_bc)
{ }

void NewPoissonOpFactory::define(const RealVect& a_dx,  BCFunc a_bc)
{
  m_dx=a_dx;
  m_bc=a_bc;
}

void NewPoissonOp::
createCoarsened(FArrayBox&                  a_lhs,
                const FArrayBox&            a_rhs,
                const int &                 a_refRat)
{
  int ncomp = a_rhs.nComp();
  //fill ebislayout
  Box coarsenedBox = coarsen(a_rhs.box(), a_refRat);
  a_lhs.define(coarsenedBox, ncomp);
}
NewPoissonOp* NewPoissonOpFactory::MGnewOp(const ProblemDomain& a_FineindexSpace,
                                           int   a_depth,
                                           bool  a_homoOnly)
{
 CH_assert(a_depth >= 0 );
 CH_assert(m_bc != NULL);
  NewPoissonOp* newOp = new NewPoissonOp();
  RealVect dx = m_dx;
  ProblemDomain domain = a_FineindexSpace;
  for(int i=0; i<a_depth; i++)
    {
      Box d = domain.domainBox();
      d.coarsen(8);
      d.refine(8);
      if(domain.domainBox() == d)
        {
          dx/=2;
          domain.coarsen(2);
        }
      else
        {
          return NULL;
        }
    }
  newOp->define(dx, domain, m_bc);
  return newOp;
}

void NewPoissonOpFactory::MGreclaim(NewPoissonOp* a_reclaim)
{
  delete a_reclaim;
}
