/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// InterpOnFace.H
// petermc, 20 Aug 2004

#ifndef INTERPONFACE_H
#define INTERPONFACE_H

#include <cstdlib>
#include <iostream>
#include "SPACE.H"
#include <cmath>
#include "FArrayBox.H"
#include "REAL.H"

/// Interpolates on faces of a rectangular domain.
class InterpOnFace
/** Interpolates on faces of a rectangular domain.
*/
{

public:

  /**
     \name Constructors, destructor and defines
  */
  /*@{*/

  ///
  /** Default constructor leaves InterpOnFace undefined.
  */
  InterpOnFace();

  ///
  /**
     Full constructor.
     Calls full define() function with identical arguments.
  */
  InterpOnFace(/// NODE-centered face
               const Box&   a_fineFaceBox,
               /// normal direction of the face
               int          a_idir,
               /// coarsening ratio
               int          a_coarsening,
               /// width of border around face
               int          a_interpBorder);

  ///
  /** Destructor.
   */
  virtual ~InterpOnFace();

  ///
  /**
     Full define function.
  */
  virtual void define(/// NODE-centered face
                      const Box&   a_fineFaceBox,
                      /// normal direction of the face
                      int          a_idir,
                      /// coarsening ratio
                      int          a_coarsening,
                      /// width of border around face
                      int          a_interpBorder);

  /*@}*/

  /**
     \name Access functions
  */
  /*@{*/

  ///
  /** Returns <tt>true</tt> if this object was created with the defining
      constructor or if define() has been called.
  */
  bool isDefined() const;


  /*@}*/

  /**
     \name Parameter-setting functions
  */
  /*@{*/

  /// 
  /** Set verbosity.  Default is 0.
   */
  void setVerbose(int a_verbose);

  /*@}*/

  /**
     \name Data modification functions
  */
  /*@{*/

  ///
  /** Interpolates from coarse to fine data on the face.
  */
  void interpolate(/// fine data
                   FArrayBox&         a_fine,
                   /// coarse data
                   const FArrayBox&   a_coarse);

  /*@}*/

protected:

  /** has this InterpOnFace been defined yet?
   */
  bool m_isDefined;

  /** NODE-centered fine face
   */
  Box m_fineFaceBox;

  /** NODE-centered coarse face
   */
  Box m_coarseFaceBox;

  /** NODE-centered coarse face refined by m_coarsening
   */
  Box m_crseFaceBoxBasicRefined;

  /** whether m_fineFaceBox - m_fineCoarseShift is coarsenable by m_coarsening
   */
  bool m_coarsenedExact;

  /** amount by which to shift a_fine after interpolation
   */
  IntVect m_offLoExt;

  /** verbosity; default 0
   */
  int m_verbose;
  
  /** coarsening ratio
   */
  int m_coarsening;

  /** normal direction of the face
   */
  int m_idir;

  /** parallel directions of the face
   */
  int m_ipar1, m_ipar2;

  /** width of border around face
   */
  int m_interpBorder;

  /** lower corner of m_fineFaceBox
   */
  IntVect m_fineCoarseShift;

  /** weights for interpolating to nodes where first face coordinate
      is not divisible by coarsening ratio, but second is divisible
   */
  FArrayBox m_weights1d1;

  /** weights for interpolating to nodes where first face coordinate
      is divisible by coarsening ratio, but second is not divisible
   */
  FArrayBox m_weights1d2;

  // new weights

  Box m_cbx1, m_cbx2, m_icbx1, m_cbx1e, m_cbx1n, m_icbx2;

  Box m_ifbx1, m_ifbx2, m_cfbx2e, m_cfbx2n;

  FArrayBox m_weights1, m_weights2;

  FArrayBox m_work;

private:

  void operator = (const InterpOnFace&){;}

  InterpOnFace(const InterpOnFace&){;}

};

#endif
