/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// Normals.H
// petermc, 19 Aug 2004

#ifndef NORMALS_H
#define NORMALS_H

#include <cstdlib>
#include <iostream>
#include "SPACE.H"
#include <cmath>
#include "FArrayBox.H"
#include "RealVect.H"
#include "Vector.H"
#include "Tuple.H"
#include "REAL.H"

/// Computes normal derivative on faces of a rectangular domain.
class Normals
/** Computes normal derivative on faces of a rectangular domain.
*/
{

public:

  /**
     \name Constructors, destructor and defines
  */
  /*@{*/

  ///
  /** Default constructor leaves Normals undefined.
  */
  Normals();

  ///
  /**
     Full constructor.
     Calls full define() function with identical arguments.
  */
  Normals(/// NODE-centered domain
          const Box&        a_bx,
          /// grid spacing
          const RealVect&   a_dx,
          /// degree of normal derivative approximation
          int               a_degree);

  ///
  /** Destructor.
   */
  virtual ~Normals();

  ///
  /**
     Full define function.
  */
  virtual void define(/// NODE-centered domain
                      const Box&        a_bx,
                      /// grid spacing
                      const RealVect&   a_dx,
                      /// degree of normal derivative approximation
                      int               a_degree);

  /*@}*/

  /**
     \name Access functions
  */
  /*@{*/

  ///
  /** Returns <tt>true</tt> if this object was created with the defining
      constructor or if define() has been called.
  */
  bool isDefined() const;


  /*@}*/

  /**
     \name Parameter-setting functions
  */
  /*@{*/

  /// 
  /** Set verbosity.  Default is 0.
   */
  void setVerbose(int a_verbose);

  /*@}*/

  /**
     \name Data modification functions
  */
  /*@{*/

  ///
  /** Computes the discrete normal derivative on all faces of the domain.
  */
  void evalNormal(/// normal derivatives.  order xlo, ylo, lo, xhi, yhi, zhi
                  Vector<FArrayBox*>&   a_deriv,
                  /// data from which to take normal derivatives
                  const FArrayBox&      a_phi);

  ///
  /** Computes the discrete normal derivative on one face of the domain.
  */
  void evalNormal(/// normal derivatives on one face
                  FArrayBox&              a_deriv,
                  /// data from which to take normal derivative
                  const FArrayBox&        a_phi,
                  /// face
                  int                     a_idir,
                  /// face
                  const Side::LoHiSide&   a_side);

  ///
  /** Computes the discrete normal derivative on all faces of the domain.
  */
  void evalNormalExtra(/// normal derivatives.  order xlo, ylo, lo, xhi, yhi, zhi
                       Vector<FArrayBox*>&   a_deriv,
                       /// data from which to take normal derivatives
                       const FArrayBox&      a_phi,
                       /// extra layer of data beyond all faces.  order xlo, ylo, lo, xhi, yhi, zhi
                       const Vector<FArrayBox*>&   a_phiExtra);

  ///
  /** Computes the discrete normal derivative on one face of the domain.
  */
  void evalNormalExtra(/// normal derivatives on one face
                       FArrayBox&              a_deriv,
                       /// data from which to take normal derivative
                       const FArrayBox&        a_phi,
                       /// extra layer of data beyond face
                       const FArrayBox&        a_phiExtra,
                       /// face
                       int                     a_idir,
                       /// face
                       const Side::LoHiSide&   a_side);

  /*@}*/
  
protected:

  /** has this Normals been defined yet?
   */
  bool m_isDefined;

  /** NODE-centered grid
   */
  Box m_bx;

  /** mesh spacings
   */
  RealVect m_dx;

  /** verbosity; default 0
   */
  int m_verbose;

  /** degree of normal derivative approximation
   */
  int m_degree;

  /** coefficients for one-sided discrete normal derivatives
   */
  Tuple<Real*, SpaceDim> m_coeffs;

  /** coefficients for discrete normal derivatives with an extra point
   */
  Tuple<Real*, SpaceDim> m_coeffsExtra;

private:

  void operator = (const Normals&){;}

  Normals(const Normals&){;}

};

#endif
