/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// PoissonDirichlet.cpp
// petermc, 19 Aug 2004

#include "PoissonDirichlet.H"
#include "InfiniteDomainF.H"
#include "MayDay.H"
#include "EvalOperatorF_F.H"

using std::cout;
using std::endl;
using std::cerr;

// ---------------------------------------------------------
// default constructor
PoissonDirichlet::PoissonDirichlet()
{
  m_isDefined = false;
}


// ---------------------------------------------------------
PoissonDirichlet::PoissonDirichlet(const Box&                         a_bx,
                                   const RealVect&                    a_dx,
                                   OperatorType                       a_op)
{
  define(a_bx, a_dx, a_op);
}


// ---------------------------------------------------------
PoissonDirichlet::~PoissonDirichlet()
{
}


// ---------------------------------------------------------
void
PoissonDirichlet::define(const Box&                         a_bx,
                         const RealVect&                    a_dx,
                         OperatorType                       a_op)
{
  m_bx = a_bx;
  m_dx = a_dx;
  m_op = a_op;
  m_verbose = 0;
  m_isDefined = true;
}

// ---------------------------------------------------------
bool PoissonDirichlet::isDefined() const
{
  return m_isDefined;
}


// ---------------------------------------------------------
void PoissonDirichlet::setVerbose(int a_verbose)
{
  m_verbose = a_verbose;
}


// ---------------------------------------------------------
void PoissonDirichlet::solveHomogeneous(FArrayBox&         a_phi,
                                        const FArrayBox&   a_rhs)
{
 CH_assert(isDefined());
 CH_assert(a_phi.box() == m_bx);
 CH_assert(a_rhs.box() == m_bx);

  int n0 = m_bx.size(0) - 1;
  int n1 = m_bx.size(1) - 1;
  int n2 = m_bx.size(2) - 1;

  int intOp = (int) m_op;
  a_phi.copy(a_rhs);
  FORT_POISSON3D(a_phi.dataPtr(), &n0, &n1, &n2, m_dx.dataPtr(), &intOp);
}


// ---------------------------------------------------------
void PoissonDirichlet::solveHomogeneousInPlace(FArrayBox&   a_phirhs)
{
 CH_assert(isDefined());
 CH_assert(a_phirhs.box() == m_bx);

  int n0 = m_bx.size(0) - 1;
  int n1 = m_bx.size(1) - 1;
  int n2 = m_bx.size(2) - 1;

  int intOp = (int) m_op;
  FORT_POISSON3D(a_phirhs.dataPtr(), &n0, &n1, &n2, m_dx.dataPtr(), &intOp);
}


// ---------------------------------------------------------
void PoissonDirichlet::solveInhomogeneous(FArrayBox&         a_phi,
                                          const FArrayBox&   a_rhs)
{
 CH_assert(isDefined());
 CH_assert(a_phi.box() == m_bx);
 CH_assert(a_rhs.box() == m_bx);

  // Copy a_rhs to a_phi in interior.
  Box bxInner = grow(m_bx, -1);
  a_phi.copy(a_rhs, bxInner);

  int intOp = (int) m_op;
  int errcode = 0;
  const int *bxLo = m_bx.loVect();
  const int *bxHi = m_bx.hiVect();
  FORT_INHOMODIRICHLET3D(a_phi.dataPtr(),
                         &bxLo[0], &bxLo[1], &bxLo[2],
                         &bxHi[0], &bxHi[1], &bxHi[2],
                         &m_dx[0], &m_dx[1], &m_dx[2],
                         &intOp, &errcode);
  if (errcode != 0)
    {
      cerr << "inhomodirichlet3d returned error code " << errcode << endl;
      MayDay::Error("returning");
      return;
    }
}


// ---------------------------------------------------------
void PoissonDirichlet::solveInhomogeneousInPlace(FArrayBox&   a_phirhs)
{
 CH_assert(isDefined());
 CH_assert(a_phirhs.box() == m_bx);

  int intOp = (int) m_op;
  int errcode = 0;
  const int *bxLo = m_bx.loVect();
  const int *bxHi = m_bx.hiVect();
  FORT_INHOMODIRICHLET3D(a_phirhs.dataPtr(),
                         &bxLo[0], &bxLo[1], &bxLo[2],
                         &bxHi[0], &bxHi[1], &bxHi[2],
                         &m_dx[0], &m_dx[1], &m_dx[2],
                         &intOp, &errcode);
  if (errcode != 0)
    {
      cerr << "inhomodirichlet3d returned error code " << errcode << endl;
      MayDay::Error("returning");
      return;
    }
}


// ---------------------------------------------------------
void PoissonDirichlet::evalOperator(FArrayBox&         a_eval,
                                    const FArrayBox&   a_phi)
{
 CH_assert(isDefined());
 CH_assert(a_eval.box() == m_bx);
 CH_assert(a_phi.box() == m_bx);

  int intOp = (int) m_op;

  a_eval.setVal(0.);
  Box bxInner = grow(m_bx, -1);
  FORT_EVALOP(CHF_FRA1(a_eval, 0),
              CHF_CONST_INT(intOp),
              CHF_CONST_REALVECT(m_dx),
              CHF_BOX(bxInner),
              CHF_CONST_FRA1(a_phi, 0));
}
