/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// Projections.H
// petermc, 2 Sep 2004

#include "FArrayBox.H"

///
/**
   Returns the "outer" coarsening of a NODE-centered box:
   coarsenOuter(((l1, l2, l3):(u1, u2, u3)), r) ==
   (floor(l1/r, l2/r, l3/r):ceil(u1/r, u2/r, u3/r))

   If r divides all of l1, l2, l3, u1, u2, u3 then we have simply
   coarsenOuter(((l1, l2, l3):(u1, u2, u3)), r) ==
   coarsen(((l1, l2, l3):(u1, u2, u3)), r)
   == ((l1/r, l2/r, l3/r):(u1/r, u2/r, u3/r))
 */
Box coarsenOuter(/// NODE-centered fine box
                 const Box&   a_bx,
                 /// refinement ratio
                 int          a_refRatio);

///
/**
   Returns the "inner" coarsening of a NODE-centered box:
   coarsenInner(((l1, l2, l3):(u1, u2, u3)), r) ==
   (ceil(l1/r, l2/r, l3/r):floor(u1/r, u2/r, u3/r))

   If r divides all of l1, l2, l3, u1, u2, u3 then we have simply
   coarsenInner(((l1, l2, l3):(u1, u2, u3)), r) ==
   coarsen(((l1, l2, l3):(u1, u2, u3)), r)
   == ((l1/r, l2/r, l3/r):(u1/r, u2/r, u3/r))
 */
Box coarsenInner(/// NODE-centered fine box
                 const Box&   a_bx,
                 /// refinement ratio
                 int          a_refRatio);

///
/**
   Projects a fine FAB to a coarse FAB by
   a_coarse(ivCoarse) = a_fine(ivCoarse * a_refRatio + a_shift)
   where a_shift is optional, defaulting to zero.

   Data are projected from nodes in the intersection of the box of a_coarse
   and the inner coarsening (see coarsenInner above) by a_refRatio of
   the box of a_fine shifted by -a_shift.
   Both FABs must have only one component.
 */
void projectToCoarse(/// coarse FAB result of projection
                     FArrayBox&         a_coarse,
                     /// fine FAB of one component to be projected
                     const FArrayBox&   a_fine,
                     /// refinement ratio
                     int                a_refRatio,
                     /// amount to shift fine FAB before projecting
                     const IntVect&     a_shift = IntVect::Zero);

///
/**
   Projects a coarse FAB to a fine FAB by
   a_coarse(ivCoarse * a_refRatio + a_shift) = a_coarse(ivCoarse)
   where a_shift is optional, defaulting to zero.

   Data are projected from nodes in the intersection of the box of a_coarse
   and the inner coarsening (see coarsenInner above) by a_refRatio of
   the box of a_fine shifted by -a_shift.
   Both FABs must have only one component.
 */
void projectToFine(/// fine FAB result of projection
                   FArrayBox&         a_fine,
                   /// coarse FAB of one component to be projected
                   const FArrayBox&   a_coarse,
                   /// refinement ratio
                   int                a_refRatio,
                   /// amount to shift fine FAB after projecting
                   const IntVect&     a_shift = IntVect::Zero);
