/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// Projections.cpp
// petermc, 2 Sep 2004

#include "BoxIterator.H"
#include "Projections.H"

// ---------------------------------------------------------
Box coarsenOuter(const Box&   a_bx,
                 int          a_refRatio)
{
 CH_assert(a_bx.ixType() == IndexType::TheNodeType());
  Box bxCoarse = coarsen(a_bx, a_refRatio);
  if (refine(bxCoarse, a_refRatio) == a_bx)
    return bxCoarse;
  
  IntVect loFine = a_bx.smallEnd();
  IntVect hiFine = a_bx.bigEnd();
  IntVect loCoarse = loFine / a_refRatio;
  IntVect hiCoarse = hiFine / a_refRatio;
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      if (loCoarse[idir] * a_refRatio > loFine[idir])
        loCoarse.setVal(idir, loCoarse[idir]-1);
      if (hiCoarse[idir] * a_refRatio < hiFine[idir])
        hiCoarse.setVal(idir, hiCoarse[idir]+1);
    }
  Box bxCoarseOuter = Box(loCoarse, hiCoarse, IndexType::TheNodeType());
  return bxCoarseOuter;
}


// ---------------------------------------------------------
Box coarsenInner(const Box&   a_bx,
                 int          a_refRatio)
{
 CH_assert(a_bx.ixType() == IndexType::TheNodeType());
  Box bxCoarse = coarsen(a_bx, a_refRatio);
  if (refine(bxCoarse, a_refRatio) == a_bx)
    return bxCoarse;
  
  IntVect loFine = a_bx.smallEnd();
  IntVect hiFine = a_bx.bigEnd();
  IntVect loCoarse = loFine / a_refRatio;
  IntVect hiCoarse = hiFine / a_refRatio;
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      if (loCoarse[idir] * a_refRatio < loFine[idir])
        loCoarse.setVal(idir, loCoarse[idir]+1);
      if (hiCoarse[idir] * a_refRatio > hiFine[idir])
        hiCoarse.setVal(idir, hiCoarse[idir]-1);
    }
  Box bxCoarseInner = Box(loCoarse, hiCoarse, IndexType::TheNodeType());
  return bxCoarseInner;
}


// ---------------------------------------------------------
void projectToCoarse(FArrayBox&         a_coarse,
                     const FArrayBox&   a_fine,
                     int                a_refRatio,
                     const IntVect&     a_shift)
{
  // Only for FArrayBoxes of one component.
 CH_assert(a_coarse.nComp() == 1);
 CH_assert(a_fine.nComp() == 1);
  if (a_refRatio == 1)
    {
      a_coarse.shift(+a_shift);
      a_coarse.copy(a_fine);
      a_coarse.shift(-a_shift);
    }
  else
    {
      Box bxCoarse = a_coarse.box();
      Box bxFine = a_fine.box();
      Box bxFineCoarsenedShifted = coarsenInner(bxFine - a_shift, a_refRatio);
      Box bxCoarseIntersect = bxCoarse & bxFineCoarsenedShifted;
      for (BoxIterator bit(bxCoarseIntersect); bit.ok(); ++bit)
        {
          IntVect ivCoarse = bit();
          a_coarse(ivCoarse, 0) = a_fine(ivCoarse * a_refRatio + a_shift, 0);
        }
    }
}


// ---------------------------------------------------------
void projectToFine(FArrayBox&         a_fine,
                   const FArrayBox&   a_coarse,
                   int                a_refRatio,
                   const IntVect&     a_shift)
{
  // Only for FArrayBoxes of one component.
 CH_assert(a_coarse.nComp() == 1);
 CH_assert(a_fine.nComp() == 1);
  if (a_refRatio == 1)
    {
      a_fine.shift(-a_shift);
      a_fine.copy(a_coarse);
      a_fine.shift(+a_shift);
    }
  else
    {
      Box bxCoarse = a_coarse.box();
      Box bxFine = a_fine.box();
      Box bxFineCoarsenedShifted = coarsenInner(bxFine - a_shift, a_refRatio);
      Box bxCoarseIntersect = bxCoarse & bxFineCoarsenedShifted;
      for (BoxIterator bit(bxCoarseIntersect); bit.ok(); ++bit)
        {
          IntVect ivCoarse = bit();
          a_fine(ivCoarse * a_refRatio + a_shift, 0) = a_coarse(ivCoarse, 0);
        }
    }
}
