/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _BINITEM_H_
#define _BINITEM_H_

#include <iostream>

#include "RealVect.H"

/// Virtual base class for particle data.
/**
  BinItem is a base class to provide the necessary interface
  with BinFab for particle-in-bin/cell problems. Since the only attribute
  of a particle needed by the BinFab is position, it is the only attribute
  present in this class.  All other attributes (charge, velocity, etc)
  must appear in a derived class.
*/

class BinItem
{
public:
  /// Null constructor, copy constructor and operator= can be compiler defined.
  BinItem();

  /// Constructs a \p BinItem at \a a_position
  BinItem(const RealVect& a_position);

  /// Destructor
  virtual ~BinItem();

  /// Same as the constructor
  virtual void define(const RealVect& a_position);

  /// Change the position
  void setPosition(const RealVect& a_position);

  /// Change one component of the position
  void setPosition(const Real a_position, const int a_dimension);

  /// Return a reference to the position
  RealVect& position();

  /// Return a constant reference to the position
  const RealVect& position() const;

  /// Gets one component of the position
  Real position(const int a_dir) const;

  // Linearization functions

  /// Number of bytes used by \b linearIn()/linearOut()
  /** Returns the size, in number of bytes, of a flat linear
      representation of the data in this object.
  */
  virtual int size() const;

  /// Write a serialized (packed) representation into \a a_buf
  /** Write a linear representation of the internal data. Assumes that
      sufficient memory for the buffer has already been allocated by
      the caller.
  */
  virtual void linearOut(void* a_buf) const;

  /// Extract a serialized (packed) representation from \a a_buf
  /** Given a linear representation of the class data previously made
      using \b linearIn() in \a a_buf, set the data for this class.
  */
  virtual void linearIn(void* a_buf);

  ///
  int preAllocatable()
  {
    return 0;
  }

  // Text output functions

  /// Write a human-readable version to stdout (for debugging)
  virtual void print() const;
  friend std::ostream& operator<<( std::ostream& ostr, const BinItem& a_item );

protected:
  /// this is the position in space of this item
  RealVect m_position;
};

std::ostream& operator<<( std::ostream& ostr, const BinItem& a_item );

#endif
