/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <iostream>
#include <vector>

#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "Vector.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LayoutIterator.H"
#include "LevelData.H"
#include "LoadBalance.H"
#include "BoxIterator.H"
#include "PiecewiseLinearFillPatch.H"

/// Prototypes:
int
testPWL();

void
parseTestOptions(int argc ,char* argv[]) ;

void
initData(LevelData<FArrayBox>& data,
         const IntVect& ghost,
         Real dx,
         Real time_interp_coef);

/// Global variables for handling output:
static const char* pgmname = "testPWL" ;
static const char* indent = "   ";
static const char* indent2 = "      " ;
static bool verbose = true ;

int
main(int argc ,char* argv[])
{
#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif
  parseTestOptions( argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  int status = testPWL() ;
  cout << indent << pgmname ;
  if( status == 0 )
    {
      cout << " passed." << endl;
    }else{
      cout << " failed with result code " << status << endl ;
    }

#ifdef CH_MPI
  MPI_Finalize();
#endif
  return status ;
}

int
testPWL()
{
  const int ref_ratio = 2;
  if (verbose)
  {
    cout << indent2 << "refinement ratio = " << ref_ratio << endl;
  }
  const Box coarse_problem_domain(IntVect(D_DECL(  0,  0, 0)),
                                  IntVect(D_DECL( 11, 11, 3)));
  if (verbose)
  {
    cout << indent2 << "coarse problem domain = " << coarse_problem_domain << endl;
  }
  const Box fine_problem_domain = refine(coarse_problem_domain, ref_ratio);
  if (verbose)
  {
    cout << indent2 << "fine problem domain = " << fine_problem_domain << endl;
  }
  const int num_coarse_grids = 2;
  if (verbose)
  {
    cout << indent2 << "num coarse grids = " << num_coarse_grids << endl;
  }
  Vector<int> assign;
  Vector<Box> coarse_boxes(num_coarse_grids);
  coarse_boxes[0] = Box( IntVect( D_DECL (0,  0, 0) ),
                         IntVect( D_DECL (5,  9, 3) ));
  coarse_boxes[1] = Box(IntVect( D_DECL ( 6,  2, 0) ),
                        IntVect( D_DECL (11, 11, 3) ));
  LoadBalance(assign, coarse_boxes);
  const DisjointBoxLayout coarse_grids(coarse_boxes, assign);

  if (verbose)
  {
    cout << indent2 << "coarse grids: " << endl;
    LayoutIterator coarse_lit = coarse_grids.layoutIterator();
    for (coarse_lit.begin(); coarse_lit.ok(); ++coarse_lit)
    {
      cout << indent2 << coarse_lit().intCode() << ": " << coarse_grids[coarse_lit()] << endl;
    }
  }
  int num_fine_grids = 3;
  if (verbose)
  {
    cout << indent2 << "num fine grids = " << num_fine_grids << endl;
  }

  Vector<Box> fine_boxes(num_fine_grids);
  fine_boxes[0] = Box(IntVect( D_DECL (  2,  4, 0) ),
                      IntVect( D_DECL (  7, 11, 7) ));
  fine_boxes[1] = Box(IntVect( D_DECL (  6, 12, 0) ),
                      IntVect( D_DECL ( 23, 17, 7) ));
  fine_boxes[2] = Box(IntVect( D_DECL ( 20,  6, 0) ),
                      IntVect( D_DECL ( 23, 11, 7) ));
  LoadBalance(assign, fine_boxes);
  const DisjointBoxLayout fine_grids(fine_boxes, assign);

  if (verbose)
  {
    cout << indent2 << "fine grids: " << endl;
    LayoutIterator fine_lit = fine_grids.layoutIterator();
    for (fine_lit.begin(); fine_lit.ok(); ++fine_lit)
    {
      cout << indent2 << fine_lit().intCode() << ": " << fine_grids[fine_lit()] << endl;
    }
  }
  const int num_comps = 1;
  if (verbose)
  {
    cout << indent2 << "num comps = " << num_comps << endl;
  }
  const int interp_radius = 1;
  if (verbose)
  {
    cout << indent2 << "interp radius = " << interp_radius << endl;
  }
  PiecewiseLinearFillPatch pwl(fine_grids,
                               coarse_grids,
                               num_comps,
                               coarse_problem_domain,
                               ref_ratio,
                               interp_radius);
// uncomment the following functions for more verbose deubigging output
//  pwl.printIntVectSets();
//  pwl.writeIntVectSetsPlotfiles();

  const IntVect ghost_vect = interp_radius * IntVect::Unit;

  LevelData<FArrayBox> fine_data(fine_grids, num_comps, ghost_vect);
  LevelData<FArrayBox> old_coarse_data(coarse_grids, num_comps);
  LevelData<FArrayBox> new_coarse_data(coarse_grids, num_comps);

  const Real time_interp_coef = .25;
  if (verbose)
  {
    cout << indent2 << "time interp coef = " << time_interp_coef << endl;
  }
  const Real fine_dx = 1.;
  initData(fine_data,
           IntVect::Zero,
           fine_dx,
           time_interp_coef);
  const Real coarse_dx = ref_ratio;
  initData(old_coarse_data,
           IntVect::Zero,
           coarse_dx,
           0.);
  initData(new_coarse_data,
           IntVect::Zero,
           coarse_dx,
           1.);
//
// test piecewise linear interpolation
  const int src_comp = 0;
  const int dest_comp = 0;

  pwl.fillInterp(fine_data,
                 old_coarse_data,
                 new_coarse_data,
                 time_interp_coef,
                 src_comp,
                 dest_comp,
                 num_comps);

  fine_data.exchange(fine_data.interval());
//
// compare with exact
  LevelData<FArrayBox> fine_errs(fine_grids, num_comps, ghost_vect);
  initData(fine_errs,
           ghost_vect,
           fine_dx,
           time_interp_coef);
  DataIterator fine_dit = fine_grids.dataIterator();
  Real err_sum = 0.;
  Real data_sum = 0.;
  for (fine_dit.begin(); fine_dit.ok(); ++fine_dit)
  {
    FArrayBox& fine_err_fab = fine_errs[fine_dit()];
    const FArrayBox& fine_data_fab = fine_data[fine_dit()];
    fine_err_fab -= fine_data_fab;
    fine_err_fab /= fine_data_fab;
    const Box domain_interior_box = fine_err_fab.box() & fine_problem_domain;
    err_sum += fine_err_fab.norm(domain_interior_box, 1, 0, num_comps);
    data_sum += fine_data_fab.norm(domain_interior_box, 1, 0, num_comps);
  }
  if (verbose)
  {
    cout << indent2 << "data  sum = " << data_sum << endl;
    cout << indent2 << "error sum = " << err_sum << endl;
  }
#if defined(CH_USE_DOUBLE)
  Real err_sum_tol = 1e-10;
#elif defined(CH_USE_FLOAT)
  Real err_sum_tol = 1e-4;
#endif
  if (err_sum < err_sum_tol)
    {
      if (verbose)
        cout << indent << pgmname << ": piecewise linear coarse-fine interpolation test passed." << endl;
      return 0 ;
    }
  else
    {
      cout << indent << pgmname << ": piecewise linear coarse-fine interpolation test FAILED!!!" << endl;
    if (verbose)
    {
      cout << indent2 << "error sum tolerance = " << err_sum_tol << endl;
    }
    return 1 ;
  }
}

void
initData(LevelData<FArrayBox>& data,
         const IntVect& ghost,
         Real dx,
         Real time_interp_coef)
{
  std::vector<Real> linear_coefs(SpaceDim);
  linear_coefs[0] = 2.;
#if (CH_SPACEDIM > 1)
  linear_coefs[1] = .02;
#if (CH_SPACEDIM > 2)
  linear_coefs[2] = .0002;
#endif
#endif
  const int num_comps = data.nComp();
// set fine level data
  const BoxLayout& grids = data.boxLayout();
  DataIterator dit = grids.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
  {
    FArrayBox& fab = data[dit()];
    fab.setVal(-666.666);
  }
  for (int comp = 0; comp < num_comps; ++comp)
  {
    for (dit.begin(); dit.ok(); ++dit)
    {
      FArrayBox& fab = data[dit()];
      const Box box = grow(grids[dit()], ghost);
      BoxIterator bit(box);
      for (bit.begin(); bit.ok(); ++bit)
      {
        const IntVect& iv = bit();
        Real& val = fab(iv,comp);
        val = 0.;
        for (int dir = 0; dir < SpaceDim; ++dir)
        {
          val += dx * linear_coefs[dir] * (iv[dir]+.5);
        }
        val += comp;
      }
    }
  }
}

///
// Parse the standard test options (-p -q) out of the command line.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
  {
    if( argv[i][0] == '-' ) //if it is an option
    {
      // compare 3 chars to differentiate -x from -xx
      if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
      {
        verbose = true ;
        argv[i] = "" ;
      }
      else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
      {
        verbose = false ;
        argv[i] = "" ;
      }
    }
  }
  return ;
}
