/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// Purpose:
//  Test the infrastucture for periodic boundary conditions
//
// Usage:
//  <program-name> [-q|-v] ...
//
//  where:
//    -q means run quietly (only pass/fail messages printed)
//    -v means run verbosely (all messages printed)
//    -write means write out the boxes; otherwise will read in
//           boxes from a file and compare with boxes generated
//           in tests.
//    ... all non-option arguments are ignored (Chombo convention)
//
//  Default is `-v'
//
//  Unknown options are treated as errors.
//

// Include files:
#include <cstdio>

#include "parstream.H"
#include "LoadBalance.H"
#include "FArrayBox.H"
#include "FluxBox.H"
#include "LevelData.H"
#include "ProblemDomain.H"
#include "BoxIterator.H"
#include "BRMeshRefine.H"
#include "ArrayView.H"
#include "DebugDump.H"

#include "LevelFluxRegister.H"
#include "GhostBC.H"
#include "PoissonBC.H"

#ifdef CH_MPI
#include "mpi.h"
#endif

//////////////////////////////////////////////////////////////
using std::cout;
using std::endl;

void parseTestOptions(int argc, char* argv[]);

void initData(LevelData<FArrayBox>& a_data, const Real a_dx);

void initEdgeData(LevelData<FluxBox>& a_data, const Real a_dx);

/// Global variables for handling output

static const char *pgmname = "testPeriodic";
static const char *indent = "   " ,*indent2 = "      " ;

static bool verbose = true ;
// if true, use arrayview to display results
static bool display = false;

/// Code:

int
main(int argc ,char *argv[] )
{

#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif

  parseTestOptions(argc, argv);

  // establish periodic domain -- first multiply periodic in all directions
  int status = 0;
  int baseDomainSize = 8;
  int numGhost = 2;
  IntVect ghostVect(numGhost*IntVect::Unit);

  Box baseDomBox(IntVect::Zero, (baseDomainSize-1)*IntVect::Unit);
  ProblemDomain baseDomain(baseDomBox);

  // set periodic in all directions
  for (int dir=0; dir<SpaceDim; dir++) {
    baseDomain.setPeriodic(dir, true);
  }

  // set up physical BC class for partially periodic tests
  DomainGhostBC physBC;
  for (int dir=0; dir<SpaceDim; dir++) {
    SideIterator sit;
    for (sit.begin(); sit.ok(); ++sit)
      {
        // it should be easy to see if Neumann bcs are doing the right thing
        NeumannBC thisBC(dir,sit());
        physBC.setBoxGhostBC(thisBC);
      }
  }

  {
    // doubly periodic flux register test
    int nRef = 2;
    ProblemDomain fineDomain(baseDomain);
    fineDomain.refine(nRef);

    const Box domainBox = baseDomain.domainBox();
    //Vector<Box> crseBoxes(1, domainBox);
    Vector<Box> crseBoxes(4);
    crseBoxes[0] = Box(IntVect::Zero, 4*IntVect::Unit);
    crseBoxes[1] = Box(IntVect(D_DECL(0,5,0)),
               IntVect(D_DECL(4,7,7)));
    crseBoxes[2] = Box(IntVect(D_DECL(5,0,0)),
               IntVect(D_DECL(7,4,7)));
    crseBoxes[3] = Box(5*IntVect::Unit, 7*IntVect::Unit);

    Vector<int> crseProcAssign(crseBoxes.size(),0);
    DisjointBoxLayout crseGrids(crseBoxes, crseProcAssign, baseDomain);

    // note the lack of ghost cells on the base level
    LevelData<FArrayBox> crseData(crseGrids, 1, IntVect::Zero);
    LevelData<FluxBox> crseFlux(crseGrids, 1, IntVect::Zero);

    Real dxCrse = 1.0/(baseDomainSize);
    Real dxFine = dxCrse/nRef;

    // try multiple boxes for fine level
    Vector<Box> fineBoxes(4);
    Vector<int> fineProcAssign(4);

    fineBoxes[0] = Box(IntVect::Zero, 9*IntVect::Unit);
    fineBoxes[1] = Box(IntVect(D_DECL(6,10,10)), IntVect(D_DECL(11,15,15)));
    // this box should fail the disjointness test
    //fineBoxes[1] = Box(IntVect(D_DECL(7,10,10)), 17*IntVect::Unit);
    fineBoxes[2] = Box(IntVect(D_DECL(12,2,0)), IntVect(D_DECL(15,7,7)));
    fineBoxes[3] = Box(IntVect(D_DECL(12,12,0)), 15*IntVect::Unit);

    int loadbalancestatus = LoadBalance(fineProcAssign, fineBoxes);
    CH_assert (loadbalancestatus == 0);

    DisjointBoxLayout fineGrids;
    fineGrids.define(fineBoxes, fineProcAssign, fineDomain);
    LevelData<FluxBox> fineFlux(fineGrids, 1, IntVect::Zero);

    // set coarse fluxes to 1, fine fluxes to 2 to start
    // coarse-grid cell-centered data is set to 0 to see effect of reflux
    // this tests reflux locations...
    DataIterator fineDit = fineFlux.dataIterator();
    for (fineDit.begin(); fineDit.ok(); ++fineDit)
      {
        fineFlux[fineDit()].setVal(2.0);
      }

    DataIterator crseDit = crseData.dataIterator();
    for (crseDit.begin(); crseDit.ok(); ++crseDit)
      {
        crseFlux[crseDit()].setVal(1.0);
        crseData[crseDit()].setVal(0.0);
      }

    LevelFluxRegister testFR(fineGrids, crseGrids,
                             fineDomain, nRef, 1);

    testFR.setToZero();

    // increment coarse fluxes
    Real crseScale = 1.0;
    Interval comps = crseData.interval();
    for (crseDit.begin(); crseDit.ok(); ++crseDit)
      {
        FluxBox& thisCrseFlux = crseFlux[crseDit()];
        for (int dir=0; dir<SpaceDim; ++dir)
          {
            FArrayBox& thisCrseFluxDir = thisCrseFlux[dir];
            testFR.incrementCoarse(thisCrseFluxDir,
                                   crseScale,
                                   crseDit(),
                                   comps, comps, dir);
          }
      }

    // increment fine fluxes
    Real fineScale = 1.0;
    for (fineDit.begin(); fineDit.ok(); ++fineDit)
      {
        FluxBox& thisFineFlux = fineFlux[fineDit()];
        for (int dir=0; dir<SpaceDim; ++dir)
          {
            FArrayBox& thisFineFluxDir = thisFineFlux[dir];

            testFR.incrementFine(thisFineFluxDir,
                                 fineScale, fineDit(),
                                 comps, comps, dir,
                                 Side::Lo);

            testFR.incrementFine(thisFineFluxDir,
                                 fineScale, fineDit(),
                                 comps, comps, dir,
                                 Side::Hi);
          }
      }

    // now apply reflux correction
    Real refluxScale = 1.0;
    testFR.reflux(crseData, refluxScale);

    if (display)
      {
        MultiArrayViewFab(&crseData);
      }

    // now test accuracy (of locations) -- redo test with linear
    // profiles.  all flux contributions should cancel out, leaving
    // no refluxing contributions.

    initEdgeData(fineFlux, dxFine);
    initEdgeData(crseFlux, dxCrse);

    for (crseDit.begin(); crseDit.ok(); ++crseDit)
      {
        crseData[crseDit()].setVal(0.0);
      }

    testFR.setToZero();

    // increment coarse fluxes
        // increment coarse fluxes
    crseScale = 1.0;
    for (crseDit.begin(); crseDit.ok(); ++crseDit)
      {
        FluxBox& thisCrseFlux = crseFlux[crseDit()];
        for (int dir=0; dir<SpaceDim; ++dir)
          {
            FArrayBox& thisCrseFluxDir = thisCrseFlux[dir];
            testFR.incrementCoarse(thisCrseFluxDir,
                                   crseScale,
                                   crseDit(),
                                   comps, comps, dir);
          }
      }

    // increment fine fluxes
    fineScale = 1.0;
    for (fineDit.begin(); fineDit.ok(); ++fineDit)
      {
        FluxBox& thisFineFlux = fineFlux[fineDit()];
        for (int dir=0; dir<SpaceDim; ++dir)
          {
            FArrayBox& thisFineFluxDir = thisFineFlux[dir];

            testFR.incrementFine(thisFineFluxDir,
                                 fineScale, fineDit(),
                                 comps, comps, dir,
                                 Side::Lo);

            testFR.incrementFine(thisFineFluxDir,
                                 fineScale, fineDit(),
                                 comps, comps, dir,
                                 Side::Hi);
          }
      }

    // now apply reflux correction
    refluxScale = 1.0;
    testFR.reflux(crseData, refluxScale);

    if (display)
      {
        MultiArrayViewFab(&crseData);
      }
  }

  {
    // now test partially periodic case (also coarsen back to original domain
    baseDomain.setPeriodic(0,false);
    // try similar tests to the previous case
    int nRef = 2;
    ProblemDomain fineDomain(baseDomain);
    fineDomain.refine(nRef);

    const Box domainBox = baseDomain.domainBox();
    Vector<Box> crseBoxes(1, domainBox);
    Vector<int> crseProcAssign(1,0);
    DisjointBoxLayout crseGrids(crseBoxes, crseProcAssign, baseDomain);

    // note the lack of ghost cells on the base level
    LevelData<FArrayBox> crseData(crseGrids, 1, IntVect::Zero);
    LevelData<FluxBox> crseFlux(crseGrids, 1, IntVect::Zero);

    Real dxCrse = 1.0/(baseDomainSize);
    Real dxFine = dxCrse/nRef;

    // try multiple boxes for fine level
    // note grids are different in this case to allow for all periodic checking
    Vector<Box> fineBoxes(3);
    Vector<int> fineProcAssign(3);

    fineBoxes[0] = Box(IntVect::Zero, 9*IntVect::Unit);
    fineBoxes[1] = Box(IntVect(D_DECL(6,10,10)), IntVect(D_DECL(11,15,15)));
    // this box should fail the disjointness test
    //fineBoxes[1] = Box(IntVect(D_DECL(7,10,10)), 17*IntVect::Unit);
    fineBoxes[2] = Box(IntVect(D_DECL(12,2,0)), IntVect(D_DECL(15,7,7)));
    //fineBoxes[3] = Box(IntVect(D_DECL(12,12,0)), 15*IntVect::Unit);

    int loadbalancestatus = LoadBalance(fineProcAssign, fineBoxes);
    CH_assert (loadbalancestatus == 0);

    DisjointBoxLayout fineGrids;
    fineGrids.define(fineBoxes, fineProcAssign, fineDomain);
    LevelData<FluxBox> fineFlux(fineGrids, 1, IntVect::Zero);

    // set coarse fluxes to 1, fine fluxes to 2 to start
    // coarse-grid cell-centered data is set to 0 to see effect of reflux
    DataIterator fineDit = fineFlux.dataIterator();
    for (fineDit.begin(); fineDit.ok(); ++fineDit)
      {
        fineFlux[fineDit()].setVal(2.0);
      }

    DataIterator crseDit = crseData.dataIterator();
    for (crseDit.begin(); crseDit.ok(); ++crseDit)
      {
        crseFlux[crseDit()].setVal(1.0);
        crseData[crseDit()].setVal(0.0);
      }

    LevelFluxRegister testFR(fineGrids, crseGrids,
                             fineDomain, nRef, 1);

    testFR.setToZero();

    // increment coarse fluxes
    Real crseScale = 1.0;
    Interval comps = crseData.interval();
    for (crseDit.begin(); crseDit.ok(); ++crseDit)
      {
        FluxBox& thisCrseFlux = crseFlux[crseDit()];
        for (int dir=0; dir<SpaceDim; ++dir)
          {
            FArrayBox& thisCrseFluxDir = thisCrseFlux[dir];
            testFR.incrementCoarse(thisCrseFluxDir,
                                   crseScale,
                                   crseDit(),
                                   comps, comps, dir);
          }
      }

    // increment fine fluxes
    Real fineScale = 1.0;
    for (fineDit.begin(); fineDit.ok(); ++fineDit)
      {
        FluxBox& thisFineFlux = fineFlux[fineDit()];
        for (int dir=0; dir<SpaceDim; ++dir)
          {
            FArrayBox& thisFineFluxDir = thisFineFlux[dir];

            testFR.incrementFine(thisFineFluxDir,
                                 fineScale, fineDit(),
                                 comps, comps, dir,
                                 Side::Lo);

            testFR.incrementFine(thisFineFluxDir,
                                 fineScale, fineDit(),
                                 comps, comps, dir,
                                 Side::Hi);
          }
      }

    // now apply reflux correction
    Real refluxScale = 1.0;
    testFR.reflux(crseData, refluxScale);

    if (display)
      {
        MultiArrayViewFab(&crseData);
      }

  }

#ifdef CH_MPI
  MPI_Finalize();
#endif

  pout() << indent << pgmname << ": "
     << ( (status == 0) ? "passed all tests" : "failed at least one test,")
     << " status = " << status
     << endl;

  return status ;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-d" ,3 ) == 0 )
            {
              display = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-p" ,3 ) == 0 )
            {
              display = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}

void
initData(LevelData<FArrayBox>& a_data, const Real a_dx)
{

  DataIterator dit = a_data.dataIterator();
  const DisjointBoxLayout& interiorBoxes = a_data.getBoxes();

  for (dit.begin(); dit.ok(); ++dit)
    {
      // first set to a bogus value which will persist in ghost cells
      // after initialization
      a_data[dit()].setVal(1.0e9);

      // this will be slow, but who cares?
      FArrayBox& localData = a_data[dit()];
      BoxIterator boxIt(interiorBoxes[dit()]);
      Real localVal;
      for (boxIt.begin(); boxIt.ok(); ++boxIt)
        {
          const IntVect& loc = boxIt();
          // linear profile
          //localVal = a_dx*(D_TERM(loc[0]+0.5,
          //                      + loc[1]+0.5,
          //                      + loc[2]+0.5));

          // quadratic profile
          localVal = a_dx*a_dx*(D_TERM((loc[0]+0.5)*(loc[0]+0.5),
                                       + (loc[1]+0.5)*(loc[1]+0.5),
                                       + (loc[2]+0.5)*(loc[2]+0.5)));
          localData(loc) = localVal;
        }

    }
}

void
initEdgeData(LevelData<FluxBox>& a_data, const Real a_dx)
{
  DataIterator dit = a_data.dataIterator();
  const DisjointBoxLayout& interiorBoxes = a_data.getBoxes();

  for (dit.begin(); dit.ok(); ++dit)
    {
      // first set to a bogus value which will persist in ghost cells
      // after initialization
      a_data[dit()].setVal(1.0e9);

      for (int dir=0; dir<SpaceDim; dir++)
    {
      // this will be slow, but who cares?
      FArrayBox& localData = a_data[dit()][dir];
      Box edgeBox = interiorBoxes[dit()];
      edgeBox.surroundingNodes(dir);
      BoxIterator boxIt(edgeBox);
      Real localVal;
      for (boxIt.begin(); boxIt.ok(); ++boxIt)
        {
          const IntVect& loc = boxIt();
          Real realLoc[SpaceDim];
          D_TERM(realLoc[0] = a_dx*(loc[0]+0.5);,
             realLoc[1] = a_dx*(loc[1]+0.5);,
             realLoc[2] = a_dx*(loc[2]+0.5));

          // now correct for edge centering
          realLoc[dir] = realLoc[dir] -0.5*a_dx;

          // linear profile
          localVal = D_TERM(realLoc[0],
                + realLoc[1],
                + realLoc[2]);

          // quadratic profile
          //localVal = a_dx*a_dx*(D_TERM((loc[0]+0.5)*(loc[0]+0.5),
          //                   + (loc[1]+0.5)*(loc[1]+0.5),
          //                   + (loc[2]+0.5)*(loc[2]+0.5)));
          localData(loc) = localVal;
        }

    }
    }
}
