/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <iostream>
#include <cstring>
#include "CH_HDF5.H"
#include "SPMD.H"

using std::cout;
using std::endl;

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;

int
test();

/// Global variables for handling output:
static const char *pgmname = "HDF5attributes" ;
static const char *indent = "   ", *indent2 = "      " ;
static bool verbose = true ;

#ifdef CH_USE_HDF5
int compare(const HDF5HeaderData& lhs, const HDF5HeaderData& rhs);
#endif

/// Code:

int
main(int argc, char* argv[])
{
#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif

  parseTestOptions( argc ,argv ) ;

  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  ///
  // Run the tests
  ///
  if(test() != 0)
    {
      cout << indent << pgmname <<" failed"<<endl;
    }
  else
    {
      cout << indent << pgmname <<" passed"<<endl;
    }
#ifdef CH_MPI
  MPI_Finalize();
#endif
}

// returns 0 on all tests passed.

int test()
{
#ifdef CH_USE_HDF5

  int error;
  HDF5Handle testFile;

 CH_assert(!testFile.isOpen());

  error = testFile.open("attributeTest.h5", HDF5Handle::CREATE);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "File creation failed "<<error<<endl;
      return error;
    }

 CH_assert(testFile.isOpen());

  HDF5HeaderData set1;

  Real dx=0.004;
  Box b1(IntVect(D_DECL(1,2,1)), IntVect(D_DECL(4,4,4)));
  Box b2(IntVect(D_DECL(5,2,1)), IntVect(D_DECL(12,4,4)));
  int currentStep = 2332;

  set1.m_string["name"] = "set1";
  set1.m_real["dx"] = dx;
  set1.m_int["currentStep"] = currentStep;
  set1.m_intvect["some intvect or other"] = b1.smallEnd();
  set1.m_box["b1"] = b1;
  set1.m_box["b2"] = b2;

  error = set1.writeToFile(testFile); //write information to file root.
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "attribute write failed "<<error<<endl;
      return error;
    }
  testFile.setGroupToLevel(2);

  error = set1.writeToFile(testFile);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "attribute write failed "<<error<<endl;
      return error;
    }

  error = testFile.setGroup("/myspecialgroup");
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "set group failed "<<error<<endl;
      return error;
    }

  testFile.close();

 CH_assert(!testFile.isOpen());
  //=================================================================

  HDF5HeaderData readData;

  error = testFile.open("attributeTest.h5", HDF5Handle::OPEN_RDONLY);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "File open failed "<<error<<endl;
      return error;
    }
 CH_assert(testFile.isOpen());

  error = readData.readFromFile(testFile);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "attribute read failed "<<error<<endl;
      return error;
    }

  error = compare(set1, readData);
  if(error != 0)
    {
      if( verbose )
        {
          cout << indent2 << "attribute read does not match write in root"<<error<<"\n";
          cout << set1;
          cout <<"readData:\n";
          cout <<readData<<endl;
        }
      return error;
    }

  testFile.setGroupToLevel(2);

  readData.clear();

  if(readData.m_int.find("currentStep") != readData.m_int.end())
    {
      if( verbose )
        cout << indent2 << "HDF5HeaderData.clear() not functioning correct"<<endl;
      return 1;
    }

  testFile.close();

#endif  // CH_USE_HDF5
  return 0;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
  {
    if( argv[i][0] == '-' ) //if it is an option
    {
      // compare 3 chars to differentiate -x from -xx
      if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
      {
        verbose = true ;
        argv[i] = "" ;
      }
      else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
      {
        verbose = false ;
        argv[i] = "" ;
      }
      else
      {
        break ;
      }
    }
  }
  return ;
}

#ifdef CH_USE_HDF5
/** returns 0 if they are the same
            1 if an attribute exists in lhs that does not exist in rhs
            2 if an attribute exists in rhs that does not exist in lhs
            3 if a common attribute has a different value
*/
int compare(const HDF5HeaderData& lhs, const HDF5HeaderData& rhs)
{

#define LEFTVSRIGHT(left, right,  mapName, Ttype, code)                        \
  for(map<std::string, Ttype>::const_iterator p = left.mapName.begin();        \
      p!= left.mapName.end(); ++p){                                            \
      map<std::string, Ttype>::const_iterator  q = right.mapName.find(p->first);  \
      if(q == right.mapName.end())                                             \
      {                                                                        \
         return code;                                                          \
      }                                                                        \
      if(q->second != p->second) return 3; }

  LEFTVSRIGHT(lhs, rhs, m_real, Real, 1);
  LEFTVSRIGHT(lhs, rhs, m_int, int, 1);
  LEFTVSRIGHT(lhs, rhs, m_string, std::string, 1);
  LEFTVSRIGHT(lhs, rhs, m_intvect, IntVect, 1);
  LEFTVSRIGHT(lhs, rhs, m_box, Box, 1);

  LEFTVSRIGHT(rhs, lhs, m_real, Real, 2);
  LEFTVSRIGHT(rhs, lhs, m_int, int, 2);
  LEFTVSRIGHT(rhs, lhs, m_string, std::string, 2);
  LEFTVSRIGHT(rhs, lhs, m_intvect, IntVect, 2);
  LEFTVSRIGHT(rhs, lhs, m_box, Box, 2);

  return 0;
}
#endif // CH_USE_HDF5
