/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cmath>
#include "IntVect.H"
#include "Box.H"
#include "IntVectSet.H"
#include "DebugOut.H"

/// Global variables for test output:

static const char *pgmname = "ivsfabTest" ;
static const char *indent = "   " ,*indent2 = "      " ;
static bool verbose = true ;

/// Prototypes:

void parseTestOptions( int argc ,char* argv[] ) ;
int  sum( const IntVect & a_iv ){ return D_TERM(a_iv[0], +2*a_iv[1], +4*a_iv[2]); }

/**
   testIVS returns:
    0: all tests passed
*/
int testIVS(void);

int main(int argc, char* argv[])
{
#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif
  bool passed = true;
  int icode = 0;

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  icode = testIVS();
  if(icode != 0)
  {
    cout << indent << pgmname
         << ": failed with error code " << icode
         << endl;
    passed = false;
  }

  if(passed)
    cout << indent << pgmname
         << ": passed all tests"
         << endl;
#ifdef CH_MPI
  MPI_Finalize();
#endif

  return icode;
}

int testIVS()
{
  int errors = 0 ;
  Box box(IntVect::Zero, IntVect::Unit);
  IntVectSet  ivs(box);
  int ncomp = 3;
  IVSFAB<Real> fab(ivs, ncomp);

  fab.setVerbose(verbose);

  // test that the IVS is the same
  IntVectSet ivstest(fab.getIVS()) ;
  if( ! ( ivs == ivstest ) )
  {
    if( verbose ) cout << "error: " << pgmname << ": failed IntVectSet consistency test" << endl;
    ++errors ;
  }
  // test that numIvs() returns the expected result
  if( fab.numIvs() != box.volume() )
  {
    if( verbose ) cout << "error: " << pgmname << ": failed numIvs() test" << endl;
    ++errors ;
  }
  // test that setVal changes all the elements
  fab.setVal(7.5);
  for( BoxIterator bit(box) ; bit.ok() ; ++bit )
  {
    for(int icomp = 0; icomp < ncomp; icomp++)
    {
      if( fab(bit(),icomp) != 7.5 )
      {
        if( verbose )
          cout << "error: " << pgmname << ": value at (" << bit() << "," << icomp
               << ") is not what setVal() set" << endl;
        ++errors;
      }
    }
  }

  //[NOTE: no way to test the output of this, so it effectively only checks for abort.]
  if( verbose ) dumpIVSFAB(&fab);

  // visual check that indexing+iterator doesnt fail
  Real val = 0.0;
  for(IVSIterator ivsit(ivs); ivsit.ok(); ++ivsit)
  {
    for(int icomp = 0; icomp < ncomp; icomp++)
    {
          fab(ivsit(), icomp) = val;
          val += 1.0;
    }
  }
  //[NOTE: see above]
  if( verbose ) dumpIVSFAB(&fab);

  // test that indexing produces the expected results
  // by putting values based on the IntVect into the fab
  // and taking them back out
  for(IVSIterator ivsit(ivs); ivsit.ok(); ++ivsit)
  {
    for(int icomp = 0; icomp < ncomp; icomp++)
    {
      fab(ivsit(), icomp) = 10*sum(ivsit()) + icomp;
    }
  }
  for(IVSIterator ivsit(ivs); ivsit.ok(); ++ivsit)
  {
    for(int icomp = 0; icomp < ncomp; icomp++)
    {
      val = 10*sum(ivsit()) + icomp;
      if( fab(ivsit(), icomp) != val )
      {
        if( verbose )
          cout << "error: " << pgmname << ": value at (" << ivsit() << "," << icomp
               << ") is " << fab(ivsit(), icomp) << ") but should be " << val << endl ;
        ++errors;
      }
    }
  }

  return errors;
}
/*****************/
///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
/*****************/
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
  {
    if( argv[i][0] == '-' ) //if it is an option
    {
      // compare 3 chars to differentiate -x from -xx
      if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
      {
        verbose = true ;
        argv[i] = "" ;
      }
      else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
      {
        verbose = false ;
        argv[i] = "" ;
      }
      else
      {
        break ;
      }
    }
  }
  return ;
}
