/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cmath>
#include "IntVect.H"
#include "Box.H"
#include "IntVectSet.H"
#include "DebugOut.H"

/// Global variables for test output:

static const char *pgmname = "newIVSTest" ;
static const char *indent = "   " ,*indent2 = "      " ;
static bool verbose = true ;

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;

/**
   testIVS returns:
    0: all tests passed
 */
int testIVS(void);

int main(int argc, char* argv[])
{
#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif
  bool passed = true;
  int icode = 0;

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  icode = testIVS();
  if(icode != 0)
    {
      cout << indent << pgmname
           << ": failed with error code " << icode
           << endl;
      passed = false;
    }

  if(passed)
    cout << indent << pgmname
           << ": passed all tests"
           << endl;
#ifdef CH_MPI
  MPI_Finalize();
#endif

  return 0;
}

int testIVS()
{
  int eekflag = 0;
#if CH_SPACEDIM==2
  IntVect lo(0, 0);
  IntVect hi(0, 1);
  IntVect low(0, 0);
  IntVect mid(0, 2);
  IntVect high(0, 4);
#elif CH_SPACEDIM==3
  IntVect lo(0, 0, 0);
  IntVect hi(0, 1, 0);
  IntVect low(0, 0, 0);
  IntVect mid(0, 2, 0);
  IntVect high(0, 4, 0);
#else
  bogus_ch_spacedim();
#endif

  Box box(lo,hi);
  IntVectSet ivs0(box);
  ivs0.convert();
  for(IVSIterator ivsit(ivs0); ivsit.ok(); ++ivsit)
    {
      if(!box.contains(ivsit()))
        {
          eekflag = 4;
          return eekflag;
        }
    }
  IntVectSet ivs1(Box(low, high));
  ivs1 -= Box(mid, high);

  IntVectSet ivs2(ivs1);
  for(IVSIterator ivsit(ivs2); ivsit.ok(); ++ivsit)
    {
      if(!ivs1.contains(ivsit()))
        {
          eekflag = 2;
          return eekflag;
        }
    }

  if(verbose)
    {
      cout << "testIVS" << endl;
      cout << "ivs1:" << endl;
      dumpIVS(&ivs1);
      cout << "ivs2:" << endl;
      dumpIVS(&ivs2);
    }
  IntVectSet ivs3 = ivs1 | ivs2;
  for(IVSIterator ivsit(ivs2); ivsit.ok(); ++ivsit)
    {
      if(!ivs1.contains(ivsit()))
        {
          eekflag = 6;
          return eekflag;
        }
    }

  if(verbose)
    {
      cout << "ivs1 union with ivs2" << endl;
      dumpIVS(&ivs3);
    }

  for(IVSIterator ivsit(ivs3); ivsit.ok(); ++ivsit)
    {
      if(!ivs1.contains(ivsit()))
        {
          eekflag = 1;
          return eekflag;
        }
    }

  return eekflag;
}
/*****************/
///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
/*****************/
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
