/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <iostream>
using std::cout ;
using std::endl ;
#include "RealVect.H"
#ifdef CH_MPI
#include <mpi.h>
#endif

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;

int
testRealVect();

/// Global variables for handling output:
static const char *pgmname = "testRealVect" ;
static const char *indent = "   ", *indent2 = "      " ;
static bool verbose = true ;

/// Code:

int
main(int argc, char* argv[])
{
#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif
  parseTestOptions( argc ,argv ) ;

  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  ///
  // Run the tests
  ///
  int ret = testRealVect() ;

  if( ret == 0 )
    {
      cout << indent << pgmname << " passed all tests" << endl ;
    }
  else
    {
      cout << indent << pgmname << " failed " << ret << " test(s)" << endl ;
    }

#ifdef CH_MPI
  MPI_Finalize();
#endif
  return ret;
}

#ifdef CH_USE_FLOAT
const Real eps = 0.75e-6 ;
#else
const Real eps = 1e-14 ;
#endif

int
testRealVect()
{
  int errors = 0 ;

  { // operator RealVect / Real
    bool err = false ;
    RealVect testrv( D_DECL( 10.0 ,20.0 ,30.0 ) );
    RealVect testdiv = testrv / 5.0 ;
    for (int dir=0; dir<SpaceDim; dir++){
      Real relerror = Abs ( 1.0 - ( testdiv[dir] / ( 2.0 * (Real)( dir+1 ) ) ) ) ;
      err |= ( relerror > eps ) ;
    }
    if( err ){
      cout << indent << pgmname << ": RealVect failed the operator/(RV,R) test" << endl;
      cout << indent2 << "info: expected (" D_TERM( << 2.0 ,<< 4.0 ,<< 6.0 ) << "), got " << testdiv << endl;
      ++errors ;
    }else{
      if( verbose ){
        cout << indent << pgmname << ": RealVect passed the operator/(RV,R) test" << endl;
      }
    }
  }

  { // operator Real / RealVect
    bool err = false ;
    RealVect testrv( D_DECL( 10.0 ,20.0 ,30.0 ) );
    RealVect testdiv = 5.0 / testrv ;
    for (int dir=0; dir<SpaceDim; dir++){
      Real relerror = Abs ( 1.0 - ( testdiv[dir] * 2.0 * (Real)( dir+1 ) ) ) ;
      err |= ( relerror > eps ) ;
    }
    if( err ){
      cout << indent << pgmname << ": RealVect failed the operator/(R,RV) test" << endl;
      cout << indent2 << "info: expected (" D_TERM( << 0.5 ,<< 0.25 ,<< 1.0/6.0 ) << "), got " << testdiv << endl;
      ++errors ;
    }else{
      if( verbose ){
        cout << indent << pgmname << ": RealVect passed the operator/(R,RV) test" << endl;
      }
    }
  }

  { // operator RealVect / RealVect
    bool err = false ;
    RealVect testrv1( D_DECL( 10.0 ,20.0 ,30.0 ) );
    RealVect testrv2( D_DECL(  5.0 , 5.0 , 5.0 ) );
    const RealVect testdiv = testrv1 / testrv2 ;
    for (int dir=0; dir<SpaceDim; dir++){
      Real relerror = Abs ( 1.0 - ( testdiv[dir] / ( 2.0 * (Real)( dir+1 ) ) ) ) ;
      err |= ( relerror > eps ) ;
    }
    if( err ){
      cout << indent << pgmname << ": RealVect failed the operator/(RV,RV) test" << endl;
      cout << indent2 << "info: expected (" D_TERM( << 2.0 ,<< 4.0 ,<< 6.0 ) << "), got " << testdiv << endl;
      ++errors ;
    }else{
      if( verbose ){
        cout << indent << pgmname << ": RealVect passed the operator/(RV,RV) test" << endl;
      }
    }
  }

  return errors ;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
