/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

/// This tests how std::complex is implemented to make it easier to
/// figure out how to write code for Chombo's Complex::re() and ::im()
/// functions.  Since these need to be as fast as possible, they are
/// compiler dependent and can be whatever ugly hack gets the job done.
/// D.B.Serafini   Nov2004

// Usage:
//  <program-name> [-hqv] ...
//
//  where:
//    -h means print the help message and exit
//    -q means run quietly (only pass/fail messages printed)
//    -v means run verbosely (all messages printed)
//    ... all non-option arguments are ignored (Chombo convention)
//
//  Default is `-v'
//
//  Unknown options are treated as errors.

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <complex>
#include <cstring> //for strncmp()

#ifdef CH_MPI
#include <mpi.h>
#endif

#include "MayDay.H"

/// Global variables for handling output

static const char *pgmname = "testComplex" ;
static const char *indent = "   " ,*indent2 = "      " ;
static const char *usage = "[-hqv]" ;
static bool verbose = true ;

/// functions
extern "C" void check_complex_datarep( double * x );
void parseTestOptions( int argc ,char* argv[] );
void print_usage()
{
  cerr << "usage: " << pgmname << " " << usage << endl;
  MayDay::Abort();
}

////////////////////////////////////////////////////////////////

int main(int argc, char *argv[]){
#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif
  int status = 0 ;
  parseTestOptions( argc ,argv ) ;
  if( verbose ) cout << indent << "Beginning " << pgmname << " ..." << endl ;

  std::complex<double> c(123.4567,765.4321);
  double r1,i1,r2,i2 ;
  r1 = c.real() ; i1 = c.imag() ;
  if( verbose ) cout << indent2 << "before: c = " << c << endl ;

  // see if the data is stored in the expected way
  check_complex_datarep( (double*)&c );

  size_t csize = sizeof(c) ,dsize=2*sizeof(r1) ;
  if( csize != dsize ){
    cout << "error: " << pgmname
         << ": sizeof std::complex [" << csize << "] is not twice the sizeof double ["
         << dsize << "]" << endl;
    ++status;
  }

  if( verbose ) cout << indent2 << "after: c = " << c << endl ;
  r2 = c.real() ; i2 = c.imag() ;
  if( r1 != i2 || i1 != r2 ){
    cout << "error: " << pgmname
         << ": the real,imag parts of std::complex do not have the usual format" << endl;
    ++status ;
  }

  cout << indent << pgmname << ": "
       << ( (status == 0) ? "passed all tests," : "failed at least one test,")
       << " status = " << status
       << endl;

#ifdef CH_MPI
  MPI_Finalize();
#endif
  return status ;
}

////////////////////////////////////////////////////////////////

// see if the complex is stored in memory in two consecutive
// words, real part first
void check_complex_datarep( double * x ){
  double tmp = x[0] ;
  x[0] = x[1] ;
  x[1] = tmp ;
}

///
// Parse the standard test options (-v -q) out of the command line.
// The non-standard option -w means write an output file instead of reading input.
// Stop parsing when a non-option argument is found.
///
void parseTestOptions( int argc ,char* argv[] ){
  for( int i = 1 ; i < argc ; ++i ){
    if( argv[i][0] == '-' ){ //if it is an option
      // compare 3 chars to differentiate -x from -xx
      if( strncmp( argv[i] ,"-v" ,3 ) == 0 ){
        verbose = true ;
        argv[i] = "" ;
      }else if( strncmp( argv[i] ,"-q" ,3 ) == 0 ){
        verbose = false ;
        argv[i] = "" ;
      }else if( strncmp( argv[i] ,"-h" ,3 ) == 0 ){
        print_usage();
      }else{
        break ;
      }
    }
  }
  return ;
}
