/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// This program tests ChomboFortran.

// Note: the standard_io tests are not done by default because
//       they can't automatically determine if the result is
//       correct and they might fail by core-dumping.
// Note: the MayDay tests are not done by default because they
//       are _intended_ to core dump if they succeed.

// Usage: a.out [-mqsv] [-E num] [args...]*
//
// Where:
//   -m selects the mayday tests
//   -q selects `quiet' execution: only pass/fail messages are printed
//   -s selects the standard_io tests
//   -v selects `verbose' execution: all messages are printed
//   -E specifies the threshold value (aka epsilon) to use for
//      comparing floating point numbers (should be <1)
//   args... are ignored.
//
// Defaults:  -q
//
// Returns:
//  The exit code of the program is the result code from the first
//  test routine that fails.

#include <cmath> //for pow()
#include <cstring>
#include <iostream>
using std::cout;
using std::endl;
#include <memory>
#ifdef CH_MPI
#include <mpi.h>
#endif

#include "Vector.H"
#include "FArrayBox.H"
#include "FORT_PROTO.H"
#include "ChFSubs_F.H"
#include "ChFIOF_F.H"
#include "test_fm_F.H"
//[NOTE: this is a hand-generated header for a .F file, not a .ChF. <dbs>]
#include "test_fm2.H"

/// Prototypes:
void
parseTestOptions(bool *option_s ,bool *option_m ,int argc ,char* argv[]) ;

/// Global variables for handling output:
static const char* pgmname = "testChFIO" ;
static const char* indent = "   ";
static const char* indent2 = "      " ;
static bool verbose = true ;
#ifdef CH_USE_DOUBLE
static Real Epsilon = 1.0e-15 ;
#else
static Real Epsilon = 1.0e-6 ;
#endif

int
main(int argc ,char* argv[])
{
  int status = 0 ;
#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif
  {//scoping trick to make sure everything is destructed before MPI_FINALIZE is called

    bool test_stdio = false ; //dont do this test by default
    bool test_mayday = false ; //dont do this by default either

    parseTestOptions( &test_stdio ,&test_mayday ,argc ,argv ) ;
    if( verbose )
      cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

    { // Test File input/output in Chombo Fortran
      int i1 = 23 ;
      Real r1 = 23.45 ;
      int status2;

      FORT_CHF_FIO( CHF_CONST_INT( i1 ) ,CHF_CONST_REAL( r1 ), CHF_INT(status2) ) ;

      if( status2 == 0 ){
        if( verbose )
          cout << indent << pgmname << " passed the file I/O tests." << endl ;
      }
      else
      {
        cout << indent << pgmname << " failed the file I/O tests with return code " << status2 << endl ;
        if( status == 0 ) status = status2 ;
      }
    }

    { // Test standard input/output in ChF
      if( test_stdio )
      {
        int i1 = 23 ;
        Real r1 = 23.45 ;
        cout << indent << pgmname << ": running std I/O tests.  You should see some output" << endl ;
        int status2;
        FORT_CHF_STDIO( CHF_CONST_INT( i1 ) ,CHF_CONST_REAL( r1 ), CHF_INT(status2) ) ;
        if( status2 == 0 )
        {
          if( verbose )
            cout << indent << pgmname << " passed the std I/O tests." << endl ;
        }
        else
        {
          cout << indent << pgmname << " failed the std I/O tests with return code " << status << endl ;
          if( status == 0 ) status = status2 ;
        }
      }
    }

    { // Test ChomboFortran argument list macros for Vector<>
      // Test conditions: (from ChFSubs.ChF)
      //  vectorI length 1, value 12345678, modified to 87654321
      //  vectorCI length 2, values 23456789,98765432, not modified
      //  vectorR length 3, values 1.1,22.22,333.333, modified to 1/<value>
      //  vectorCR length 4, values 1+1/2^20,1/<value[0]>,1+1/2^40,1/<value[2]>
      //  NOTE: vectorCR[2:3] will be 1 when PRECISION=FLOAT

      const int IVerbose = verbose ? 1 : 0 ;
      int status2 = 0 ;

      const int I1 = 12345678, I1r = 87654321, I20 = 23456789, I21 = 98765432 ;
      const Real R10 = 1.1, R11 = 22.22, R12 = 333.333, R10r = 1.0/R10, R11r = 1.0/R11, R12r = 1.0/R12 ;
      const Real TwoTo20 = 1<<20 ;
      const Real R20 = 1.0+(1.0/TwoTo20), R21 = 1.0/R20 ;
      const Real R22 = 1.0+(1.0/(TwoTo20*TwoTo20)), R23 = 1.0/R22 ;

      Vector<int>  vectorI(1)  ; vectorI[0]  = I1  ;
      Vector<int>  vectorCI(2) ; vectorCI[0] = I20 ; vectorCI[1] = I21 ;
      Vector<Real> vectorR(3)  ; vectorR[0]  = R10 ; vectorR[1]  = R11 ; vectorR[2] = R12 ;
      Vector<Real> vectorCR(4) ; vectorCR[0] = R20 ; vectorCR[1] = R21 ; vectorCR[2] = R22 ; vectorCR[3] = R23 ;

      // test chfptr class
      chfptr p( vectorI.size()-1 );
      if( p.i != 0 )
      {
       if( verbose )
         cout << indent2 << "error: chfptr.i returned " << p.i << " instead of 0" << endl ;
        ++status2 ;
      }
      if( (const int *)p != &p.i )
      {
        if( verbose )
          cout << indent2 << "error: (const int * const)chfptr returned " << (const int *)p
               << "instead of " << &p.i << endl ;
        ++status2 ;
      }

      FORT_TESTVIR_1( CHF_VI( vectorI ), CHF_CONST_VI( vectorCI ),
                      CHF_VR( vectorR ), CHF_CONST_VR( vectorCR ),
                      CHF_CONST_INT( IVerbose ) ,CHF_CONST_REAL( Epsilon ),CHF_INT( status2 ) );
      // check return values
      if( vectorI[0] != I1r )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned int arg 1 = " << vectorI[0]
               << " instead of " << I1r << endl;
        ++status2 ;
      }
      if( vectorCI[0] != I20 )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned const int arg 2 [0] = " << vectorCI[0]
               << " instead of " << I20 << endl ;
        ++status2 ;
      }
      if( vectorCI[1] != I21 )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned const int arg 2 [1] = " << vectorCI[1]
               << " instead of " << I21 << endl ;
        ++status2 ;
      }
      if( vectorR[0] != R10r )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned Real arg 3 [0] = " << vectorR[0]
               << " instead of " << R10r << endl;
        ++status2 ;
      }
      if( vectorR[1] != R11r )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned Real arg 3 [1] = " << vectorR[1]
               << " instead of " << R11r << endl;
        ++status2 ;
      }
      if( vectorR[2] != R12r )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned Real arg 3 [2] = " << vectorR[2]
               << " instead of " << R12r << endl;
        ++status2 ;
      }
      if( vectorCR[0] != R20 )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned const Real arg 4 [0] = " << vectorCR[0]
               << " instead of " << R20 << endl;
        ++status2 ;
      }
      if( vectorCR[1] != R21 )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned const Real arg 4 [1] = " << vectorCR[1]
               << " instead of " << R21 << endl;
        ++status2 ;
      }
      if( vectorCR[2] != R22 )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned const Real arg 4 [2] = " << vectorCR[2]
               << " instead of " << R22 << endl;
        ++status2 ;
      }
      if( vectorCR[3] != R23 )
      {
        if( verbose )
          cout << indent2 << "error: TESTVIR_1 returned const Real arg 4 [3] = " << vectorCR[3]
               << " instead of " << R23 << endl;
        ++status2 ;
      }
      // done with tests
      if( status2 == 0 )
      {
        if( verbose )
          cout << indent << pgmname << " passed the ChF Vector<> tests." << endl ;
      }
      else
      {
        if( verbose )
          cout << indent << pgmname << " failed " << status2 << " of the ChF Vector<> tests." << endl ;
        if( status == 0 ) status = status2 ;
      }
    }

    { // Test ChomboFortran 1D array macros
      // Test conditions: (from ChFSubs.ChF)
      //  arrayI length 1, value 12345678, modified to 87654321
      //  arrayCI length 2, values 23456789,98765432, not modified
      //  arrayR length 3, values 1.1,22.22,333.333, modified to 1/<value>
      //  arrayCR length 4, values 1+1/2^20,1/<value[0]>,1+1/2^40,1/<value[2]>
      //  NOTE: arrayCR[2:3] will be 1 when PRECISION=FLOAT

      const int IVerbose = verbose ? 1 : 0 ;
      int status2 = 0 ;

      const int I1 = 12345678, I1r = 87654321, I20 = 23456789, I21 = 98765432 ;
      const Real R10 = 1.1, R11 = 22.22, R12 = 333.333, R10r = 1.0/R10, R11r = 1.0/R11, R12r = 1.0/R12 ;
      const Real TwoTo20 = 1<<20 ;
      const Real R20 = 1.0+(1.0/TwoTo20), R21 = 1.0/R20 ;
      const Real R22 = 1.0+(1.0/(TwoTo20*TwoTo20)), R23 = 1.0/R22 ;

      int arrayI[1]  ; arrayI[0]  = I1  ;
      const int arrayCI[2] = { I20,I21 } ;
      Real arrayR[3]  ; arrayR[0]  = R10 ; arrayR[1]  = R11 ; arrayR[2]  = R12 ;
      Real arrayR2[3] ; arrayR2[0] = R10 ; arrayR2[1] = R11 ; arrayR2[2] = R12 ;
      const Real arrayCR[4] = { R20 ,R21 ,R22 ,R23 } ;
      const Real arrayCR2[4] = { R20 ,R21 ,R22 ,R23 } ;

      FORT_TESTIR1D_1( CHF_I1D( arrayI,1 ), CHF_CONST_I1D( arrayCI,2 ),
                       CHF_R1D( arrayR,3 ), CHF_CONST_R1D( arrayCR,4 ),
                       CHF_R1D( arrayR2,3 ), CHF_CONST_R1D( arrayCR2,4 ),
                      CHF_CONST_INT( IVerbose ) ,CHF_CONST_REAL( Epsilon ),CHF_INT( status2 ) );
      // check return values
      if( arrayI[0] != I1r )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned int arg 1 = " << arrayI[0]
               << " instead of " << I1r << endl;
        ++status2 ;
      }
      if( arrayCI[0] != I20 )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned const int arg 2 [0] = " << arrayCI[0]
               << " instead of " << I20 << endl ;
        ++status2 ;
      }
      if( arrayCI[1] != I21 )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned const int arg 2 [1] = " << arrayCI[1]
               << " instead of " << I21 << endl ;
        ++status2 ;
      }
      if( arrayR[0] != R10r )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned Real arg 3 [0] = " << arrayR[0]
               << " instead of " << R10r << endl;
        ++status2 ;
      }
      if( arrayR[1] != R11r )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned Real arg 3 [1] = " << arrayR[1]
               << " instead of " << R11r << endl;
        ++status2 ;
      }
      if( arrayR[2] != R12r )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned Real arg 3 [2] = " << arrayR[2]
               << " instead of " << R12r << endl;
        ++status2 ;
      }
      if( arrayCR[0] != R20 )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned const Real arg 4 [0] = " << arrayCR[0]
               << " instead of " << R20 << endl;
        ++status2 ;
      }
      if( arrayCR[1] != R21 )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned const Real arg 4 [1] = " << arrayCR[1]
               << " instead of " << R21 << endl;
        ++status2 ;
      }
      if( arrayCR[2] != R22 )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned const Real arg 4 [2] = " << arrayCR[2]
               << " instead of " << R22 << endl;
        ++status2 ;
      }
      if( arrayCR[3] != R23 )
      {
        if( verbose )
          cout << indent2 << "error: TESTIR1D_1 returned const Real arg 4 [3] = " << arrayCR[3]
               << " instead of " << R23 << endl;
        ++status2 ;
      }
      // done with tests
      if( status2 == 0 )
      {
        if( verbose )
          cout << indent << pgmname << " passed the ChF *1D tests." << endl ;
      }
      else
      {
        if( verbose )
          cout << indent << pgmname << " failed " << status2 << " of the ChF *1D tests." << endl ;
        if( status == 0 ) status = status2 ;
      }
    }

    { // Test calling ChomboFortran from ChomboFortran, using the 1D array macros
      //
      //[NOTE: all the code here is the same as above for TESTIR1D except the
      //       subr name and the error msgs. <dbs>]
      //
      // Test conditions: (from ChFSubs.ChF)
      //  arrayI length 1, value 12345678, modified to 87654321
      //  arrayCI length 2, values 23456789,98765432, not modified
      //  arrayR length 3, values 1.1,22.22,333.333, modified to 1/<value>
      //  arrayCR length 4, values 1+1/2^20,1/<value[0]>,1+1/2^40,1/<value[2]>
      //  NOTE: arrayCR[2:3] will be 1 when PRECISION=FLOAT

      const int IVerbose = verbose ? 1 : 0 ;
      int status2 = 0 ;

      const int I1 = 12345678, I1r = 87654321, I20 = 23456789, I21 = 98765432 ;
      const Real R10 = 1.1, R11 = 22.22, R12 = 333.333, R10r = 1.0/R10, R11r = 1.0/R11, R12r = 1.0/R12 ;
      const Real TwoTo20 = 1<<20 ;
      const Real R20 = 1.0+(1.0/TwoTo20), R21 = 1.0/R20 ;
      const Real R22 = 1.0+(1.0/(TwoTo20*TwoTo20)), R23 = 1.0/R22 ;

      int arrayI[2]  ; arrayI[0]  = I1  ;
      const int arrayCI[2] = { I20,I21 } ;
      Real arrayR[3]  ; arrayR[0]  = R10 ; arrayR[1]  = R11 ; arrayR[2]  = R12 ;
      Real arrayR2[3] ; arrayR2[0] = R10 ; arrayR2[1] = R11 ; arrayR2[2] = R12 ;
      const Real arrayCR[4] = { R20 ,R21 ,R22 ,R23 } ;
      const Real arrayCR2[4] = { R20 ,R21 ,R22 ,R23 } ;

      FORT_TESTCALL( CHF_I1D( arrayI,2 ), CHF_CONST_I1D( arrayCI,2 ),
                     CHF_R1D( arrayR,3 ), CHF_CONST_R1D( arrayCR,4 ),
                     CHF_R1D( arrayR2,3 ), CHF_CONST_R1D( arrayCR2,4 ),
                     CHF_CONST_INT( IVerbose ) ,CHF_CONST_REAL( Epsilon ),CHF_INT( status2 ) );
      // check return values
      if( arrayI[0] != I1r )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned int arg 1 = " << arrayI[0]
               << " instead of " << I1r << endl;
        ++status2 ;
      }
      if( arrayCI[0] != I20 )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned const int arg 2 [0] = " << arrayCI[0]
               << " instead of " << I20 << endl ;
        ++status2 ;
      }
      if( arrayCI[1] != I21 )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned const int arg 2 [1] = " << arrayCI[1]
               << " instead of " << I21 << endl ;
        ++status2 ;
      }
      if( arrayR[0] != R10r )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned Real arg 3 [0] = " << arrayR[0]
               << " instead of " << R10r << endl;
        ++status2 ;
      }
      if( arrayR[1] != R11r )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned Real arg 3 [1] = " << arrayR[1]
               << " instead of " << R11r << endl;
        ++status2 ;
      }
      if( arrayR[2] != R12r )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned Real arg 3 [2] = " << arrayR[2]
               << " instead of " << R12r << endl;
        ++status2 ;
      }
      if( arrayCR[0] != R20 )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned const Real arg 4 [0] = " << arrayCR[0]
               << " instead of " << R20 << endl;
        ++status2 ;
      }
      if( arrayCR[1] != R21 )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned const Real arg 4 [1] = " << arrayCR[1]
               << " instead of " << R21 << endl;
        ++status2 ;
      }
      if( arrayCR[2] != R22 )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned const Real arg 4 [2] = " << arrayCR[2]
               << " instead of " << R22 << endl;
        ++status2 ;
      }
      if( arrayCR[3] != R23 )
      {
        if( verbose )
          cout << indent2 << "error: TESTCALL returned const Real arg 4 [3] = " << arrayCR[3]
               << " instead of " << R23 << endl;
        ++status2 ;
      }
      // done with tests
      if( status2 == 0 )
      {
        if( verbose )
          cout << indent << pgmname << " passed the ChF calls ChF tests." << endl ;
      }
      else
      {
        if( verbose )
          cout << indent << pgmname << " failed " << status2 << " of the ChF calls ChF tests." << endl ;
        if( status == 0 ) status = status2 ;
      }
    }

    { // Test ChomboFortran CHF_DSELECT macro
      const int IVerbose = verbose ? 1 : 0 ;
      int status2 = 0 ;
      FORT_TESTDSEL_1( CHF_CONST_INT( SpaceDim ),
                       CHF_CONST_INT( IVerbose ), CHF_INT( status2 ) );
      // done with tests
      if( status2 == 0 )
      {
        if( verbose )
          cout << indent << pgmname << " passed the CHF_DSELECT tests." << endl ;
      }
      else
      {
        if( verbose )
          cout << indent << pgmname << " failed " << status2 << " of the CHF_DSELECT tests." << endl ;
        if( status == 0 ) status = status2 ;
      }
    }

    { // Test ChomboFortran CHF_MULTIDO macro
      Box          ibox(IntVect::Zero,IntVect::Unit*3) ;
      BaseFab<int> ifab(ibox,1) ;
      const int ivol = (int)ibox.volume() ;
      const int IVerbose = verbose ? 1 : 0 ;
      int status2 = 0 ;
      ifab.setVal(1) ;
      FORT_TESTMDO_1( CHF_FIA1( ifab,0 ) ,CHF_BOX( ibox ), CHF_CONST_INT(ivol),
                      CHF_CONST_INT( IVerbose ), CHF_INT( status2 ) );
      if( status2 == 0 ){
        if( verbose )
          cout << indent << pgmname << " passed the CHF_MULTIDO tests." << endl ;
      }else{
        if( verbose )
          cout << indent << pgmname << " failed " << status2 << " of the CHF_MULTIDO tests." << endl ;
        if( status == 0 ) status = status2 ;
      }
    }

    { // Test calling MayDay routines (testing only one is enough).
      //[NOTE: this causes the program to fail, so it's not done by default (use -m).]
      if( test_mayday )
      {
        // this shouldn't return, but if it does, something went wrong
        cout << indent << pgmname << ": running MayDay tests. An Error or Abort should happen." << endl ;
        int status2;
        FORT_CHF_MAYDAY(CHF_INT(status2) ) ;
        cout << indent << pgmname << " failed the MayDay tests with return code " << status << endl ;
        if( status == 0 ) status = status2 ;
      }
    }

    { // Regression test for parsing end-of-line comments.
      // If this compiles ok, then probably it works.
      int status2 ;
      FORT_TEST_PARSE1( CHF_INT( status2 ) ) ;
      if( verbose ){
        cout << indent << pgmname ;
        if( status2 == 0 ) cout << " passed " ;
        else               cout << " failed " << status2 << " of " ;
        cout << "the parse tests." << endl ;
      }
      if( status == 0 ) status = status2 ;
    }

    { // Regression test for statement label bug found by Francesco Minati in gcc v3.x
      // If this compiles ok, then it works.
      int i ,status2 = 0 ;
      i = 1 ; FORT_TEST_FM1(CHF_INT(i)) ;
      if( i < 100 ) ++status2 ;
      i = 2 ; FORT_TEST_FM1(CHF_INT(i)) ;
      if( i < 100 ) ++status2 ;
      i = 3 ; FORT_TEST_FM1(CHF_INT(i)) ;
      if( i < 100 ) ++status2 ;
      i = 4 ; FORT_TEST_FM1(CHF_INT(i)) ;
      if( i < 100 ) ++status2 ;
      i = 5 ; FORT_TEST_FM1(CHF_INT(i)) ;
      if( i < 100 ) ++status2 ;
      if( verbose ){
        cout << indent << pgmname ;
        if( status2 == 0 ) cout << " passed " ;
        else               cout << " failed " << status2 << " of " ;
        cout << "the FM1 tests." << endl ;
      }
      if( status == 0 ) status = status2 ;
    }

    { // Regression test for leading [Tab] bug found by Francesco.
      int status2;
      FORT_TEST_FM2( CHF_INT(status2) ) ;
      if( verbose ){
        cout << indent << pgmname ;
        if( status2 == 0 ) cout << " passed " ;
        else               cout << " failed " << status2 << " of " ;
        cout << "the FM2 tests." << endl ;
      }
      if( status == 0 ) status = status2 ;
    }
  }
#ifdef CH_MPI
  MPI_Finalize();
#endif
  return status ;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Also parse the `-s' option to set the first argument
///
void
parseTestOptions(bool *option_s, bool *option_m,int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-s" ,3 ) == 0 )
            {
              *option_s = true ;
            }
          else if( strncmp( argv[i] ,"-m" ,3 ) == 0 )
            {
              *option_m = true ;
            }
          else if( strncmp( argv[i] ,"-E" ,3 ) == 0 )
            {
              Epsilon = atof( argv[++i] );
            }
        }
    }
  return ;
}
