/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "LoadBalance.H"
#include "EBArith.H"

#include "EBAMRNSviscOp.H"
#include "EBEllipticLoHiCenter.H"
#include "EBQuadCFInterp.H"
#include "EBFluxRegister.H"

#include "EBAMRNSviscOpF_F.H"
#include "DirichletNSviscEBBC.H"
#include "ArrayToRaw.H"
#include "EBCoarseAverage.H"
/////////////
/**
  coarsen the dbl if we can
  otherwise,
  check to see if the domain is covered. if so
  make one big grid.  otherwise we fail
  returns false if not coarsenable
*/


//-----------------------------------------------------------------------
EBAMRNSviscOp::~EBAMRNSviscOp()
{

}


//-----------------------------------------------------------------------
EBAMRNSviscOp::EBAMRNSviscOp()
{

}


//-----------------------------------------------------------------------
EBAMRNSviscOp::EBAMRNSviscOp(
	       const DisjointBoxLayout&            a_dblFine,
               const DisjointBoxLayout&            a_dbl,
               const DisjointBoxLayout&            a_dblCoar,
               const EBISLayout&                   a_ebislFine,
               const EBISLayout&                   a_ebisl,
               const EBISLayout&                   a_ebislCoar,
               const RefCountedPtr<DirichletNSviscDomainBC>&  a_domainBCFine,
               const RefCountedPtr<DirichletNSviscDomainBC>&  a_domainBC,
               const RefCountedPtr<DirichletNSviscDomainBC>&  a_domainBCCoar,
               const RefCountedPtr<DirichletNSviscEBBC>&      a_ebBCFine,
               const RefCountedPtr<DirichletNSviscEBBC>&      a_ebBC,
               const RefCountedPtr<DirichletNSviscEBBC>&      a_ebBCCoar,
               const RealVect&                     a_dxFine,
               const RealVect&                     a_dx,
               const RealVect&                     a_dxCoar,
               const ProblemDomain&                a_domain,
               const RealVect&                     a_origin,
               const int&                          a_refToFine,
               const int&                          a_refToCoar,
               const bool&                         a_hasFine,
               const bool&                         a_hasCoar,
               const int&                          a_relaxType,
               const int&                          a_orderEB,
	       const Real                          a_c,
	       const Real                          a_mu,
               const Real                          a_lambda,
	       const LevelData<EBCellFAB>&         a_rho,
               const bool&                         a_rz,
               const Real                          a_time )
{

  for(int idir = 1; idir < SpaceDim; idir++)
    {
      if(Abs((a_dx[idir]-a_dx[0])/a_dx[0]) > 0.01)
        {
          MayDay::Warning("Anisotropic problem");
          cout << "dx = " << a_dx << endl;
        }
    }
  m_time      = a_time;
  m_hasFine   = a_hasFine  ;
  m_hasCoar   = a_hasCoar  ;
  m_relaxType = a_relaxType;
  m_orderEB   = a_orderEB  ;
  m_domain    = a_domain   ;
  m_dbl       = a_dbl      ;
  m_domainBC  = a_domainBC ;
  m_ebBC      = a_ebBC     ;
  m_dx        = a_dx       ;
  m_origin    = a_origin   ;
  m_ebisl     = a_ebisl    ;
  m_c         = a_c        ;
  m_rz        = a_rz       ;
  m_homogeneous_relax = true;

  EBCellFactory cfactory( m_ebisl );
  
  //  m_dbl.print();
  //  cout << "a_rho ghosts " << a_rho.ghostVect() << endl;

  m_mu.define(     m_dbl, 1, a_rho.ghostVect(), cfactory );
  m_lambda.define( m_dbl, 1, a_rho.ghostVect(), cfactory );
  m_rho.define(    m_dbl, 1, a_rho.ghostVect(), cfactory );

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  a_rho.copyTo( Interval(0,0), m_rho, Interval(0,0) );
  
  for( DataIterator di(m_dbl) ; di.ok() ; ++di )
  {
     m_mu[di()].setVal(a_mu);
     m_lambda[di()].setVal(a_lambda);
     //     cout << "lambda box " << m_lambda[di()].box() << endl;
     //     cout << "mu box " << m_mu[di()].box() << endl;
  }

  m_invDx  = 1.0/m_dx;
  for( int idir = 0; idir < SpaceDim; idir++ )
  {
     m_invDx2[idir] = m_invDx[idir]*m_invDx[idir];
  }

  //dxScale is used for unscaling the boundary area (see GeometryShop)
  // (it would be better if we could just use m_dx[0])
  Real maxDx = 0.0;
  Real dv    = 1.0;
  for (int idir = 0; idir < SpaceDim; idir++)
  {
      dv *= m_dx[idir];
      if ( m_dx[idir] > maxDx )
          maxDx = m_dx[idir];
  }
  m_dxScale = pow(maxDx,SpaceDim-1)/dv;


  if(m_hasFine)
  {
      m_dblFine        = a_dblFine;
      m_dxFine         = a_dxFine;
      m_ebislFine      = a_ebislFine;
      m_domainBCFine   = a_domainBCFine;
      m_ebBCFine       = a_ebBCFine;
      m_refToFine      = a_refToFine;
      m_domainFine     = refine(m_domain, m_refToFine);

  }
  else
  {
      m_ebBCFine     = NULL;
      m_domainBCFine = NULL;
      m_refToFine    = 1;
  }

  if(m_hasCoar)
  {
      m_dblCoar        = a_dblCoar     ;
      m_dxCoar         = a_dxCoar;
      m_ebislCoar      = a_ebislCoar   ;
      m_domainBCCoar   = a_domainBCCoar;
      m_ebBCCoar       = a_ebBCCoar    ;
      m_refToCoar      = a_refToCoar   ;
      m_domainCoar     = coarsen(m_domain, m_refToCoar);

      for (int idir = 0; idir < SpaceDim; idir++)
        {
          m_loCFIVS[idir].define(m_dbl);
          m_hiCFIVS[idir].define(m_dbl);

          for (DataIterator dit = m_dbl.dataIterator();  dit.ok(); ++dit)
            {
              m_loCFIVS[idir][dit()].define(m_domain, m_dbl.get(dit()),
                                            m_dbl, idir,Side::Lo);
              m_hiCFIVS[idir][dit()].define(m_domain, m_dbl.get(dit()),
                                            m_dbl, idir,Side::Hi);
            }
        }

      //if this fails, then the AMR grids violate proper nesting.
      Box domainCoarsenedFine;
      bool hasCoarser = getCoarserLayouts(m_dblCoarsenedFine,
                                          m_ebislCoarsenedFine,
                                          domainCoarsenedFine,
                                          m_dbl,
                                          m_domain.domainBox(),
                                          m_refToCoar,ebisPtr);

      //should follow from coarsenable
      if(hasCoarser)
        {
          m_hasInterpAve = true;
          int ncomp = SpaceDim;
          m_ebInterp.define( m_dbl,     m_dblCoar,
                             m_ebisl, m_ebislCoar,
                             domainCoarsenedFine, m_refToCoar, ncomp,ebisPtr);
          m_ebAverage.define(m_dbl,     m_dblCoarsenedFine,
                             m_ebisl, m_ebislCoarsenedFine,
                             domainCoarsenedFine, m_refToCoar, ncomp,ebisPtr);
        }
      else
        {
          m_hasInterpAve = false;
        }
  }
  else
  {
      m_ebBCCoar     = NULL;
      m_domainBCCoar = NULL;
      m_refToCoar    = 1;
      m_domainCoar   = m_domain;
  }

  //special mg objects for when we do not have
  //a coarser level or when the refinement to coarse
  //is greater than two
  //flag for when we need special MG objects
  m_hasMGObjects = false;
  if((!m_hasCoar) || (m_refToCoar > 2))
  {
      int mgRef = 2;
      //don't want to go to 1x1
      int ncomp = SpaceDim;
      bool hasCoarser = getCoarserLayouts(m_dblCoarMG,
                                          m_ebislCoarMG,
                                          m_domainCoarMG,
                                          m_dbl,
                                          m_domain.domainBox(),
                                          mgRef,ebisPtr);
      if(hasCoarser)
      {
          m_hasMGObjects = true;
          m_ebInterpMG.define( m_dbl,     m_dblCoarMG,
                               m_ebisl, m_ebislCoarMG,
                               m_domainCoarMG, mgRef, ncomp,ebisPtr);
          m_ebAverageMG.define(m_dbl,     m_dblCoarMG,
                               m_ebisl, m_ebislCoarMG,
                               m_domainCoarMG, mgRef, ncomp,ebisPtr);
      }
      else
         m_hasMGObjects = false;
  }

   get_irregular();

   m_split_u = false;
   define_flux_stencils();

   m_weight.define( m_dbl );
   for (DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
   {
     //      IntVectSet notRegular = m_ebisl[dit()].getIrregIVS(m_dbl.get(dit()));
     //      IntVectSet notRegular = m_ebisl[dit()].getIrregIVS(m_mu[dit()].box());

     //      const IntVectSet& notRegular = m_irreg_w_ghost[dit()];
      const IntVectSet& notRegular = m_irreg[dit()];
      m_weight[dit()].define( notRegular, m_ebisl[dit()].getEBGraph(),
			      SpaceDim );
      //      cout << "mubox " << m_mu[dit()].box() << endl;
      m_weight[dit()].setVal(0.0);
      //      cout << "m_weight box " << m_weight[dit()].getIVS() << endl;

      if( !m_rz )
      {
         get_ns_uweight( m_weight[dit()], m_ebisl[dit()], m_iflux_stencil[dit()],
		         m_mu[dit()], m_lambda[dit()], 0 );
         get_ns_uweight( m_weight[dit()], m_ebisl[dit()], m_jflux_stencil[dit()],
		       m_mu[dit()], m_lambda[dit()], 1 );
         if( SpaceDim == 3 )      
            get_ns_uweight( m_weight[dit()], m_ebisl[dit()], m_kflux_stencil[dit()],
	  	            m_mu[dit()], m_lambda[dit()], 2 );
      }
      else
      {

         for( VoFIterator vi(notRegular,m_ebisl[dit()].getEBGraph());vi.ok();++vi)
         {
	    RealVect c = m_ebisl[dit()].centroid(vi());
            Real r = (c[0]+vi().gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
	    m_weight[dit()](vi(),0) = -2*m_mu[dit()](vi(),0)/(r*r);
	 }
         get_ns_uweight_rz( m_weight[dit()], m_ebisl[dit()], m_iflux_stencil[dit()],
		       m_mu[dit()], m_lambda[dit()], 0, notRegular );
         get_ns_uweight_rz( m_weight[dit()], m_ebisl[dit()], m_jflux_stencil[dit()],
		       m_mu[dit()], m_lambda[dit()], 1, notRegular );
      }
   }


   if( m_split_u )
      define_ustencil();

   //   m_wgh.define( m_dbl, 2, a_rho.ghostVect(), cfactory );
   m_wgh.define( m_dbl, 2, IntVect::Zero, cfactory );
   for (DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
   {
     //      if( m_rz )
      {
	//         m_wgh[dit()].define( m_ebisl[dit()], a_rho[dit()].getRegion(), SpaceDim );   
         get_uweight( dit(), m_wgh );
      }
   }

}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::get_irregular( )
{
   Vector<IntVect> ofi(3,IntVect::Zero), ofj(3,IntVect::Zero);
   Vector<IntVect> ofil(3,IntVect::Zero), ofjl(3,IntVect::Zero);
   Vector<IntVect> ofih(3,IntVect::Zero), ofjh(3,IntVect::Zero);

   ofi[0][0]=-1;
   ofi[2][0]= 1;

   ofil[1][0]=1;
   ofil[2][0]=2;

   ofih[0][0]=-2;
   ofih[1][0]=-1;
   
   ofj[0][1]=-1;
   ofj[2][1]= 1;

   ofjl[1][1]=1;
   ofjl[2][1]=2;

   ofjh[0][1]=-2;
   ofjh[1][1]=-1;


   m_irreg.define( m_dbl );
   for( DataIterator d=m_dbl.dataIterator() ; d.ok() ; ++d )
   {
      const DataIndex dex=d();
      IntVectSet& irs = m_irreg[dex];
      Vector<IntVect> of1(3,IntVect::Zero), of2(3,IntVect::Zero);
      Box itbox  =  m_dbl.get(dex) & m_domain.domainBox();
      //      cout << "irreg " << itbox << endl;
      for( BoxIterator b(itbox) ; b.ok(); ++b )
      {
	 const IntVect& iv=b();
         if( iv[0] == m_domain.domainBox().smallEnd(0) || iv[0] == itbox.smallEnd(0) )
	    of1 = ofil;
	 else if( iv[0] == m_domain.domainBox().bigEnd(0) || iv[0] == itbox.bigEnd(0) )
	    of1 = ofih;
	 else
	    of1 = ofi;

         if( iv[1] == m_domain.domainBox().smallEnd(1) || iv[1] == itbox.smallEnd(1) )
	    of2 = ofjl;
	 else if( iv[1] == m_domain.domainBox().bigEnd(1) || iv[1] == itbox.bigEnd(1) )
	    of2 = ofjh;
	 else
	    of2 = ofj;
	 
	 if( m_ebisl[dex].isRegular( iv ) )
	 {
	    for( int d1=0 ; d1 < 3 ; d1++ )
	       for( int d2=0 ; d2 < 3 ; d2++ )
	       {
		 if( itbox.contains(iv+of1[d1]+of2[d2]) &&
		                    !m_ebisl[dex].isCovered(iv+of1[d1]+of2[d2]))
		 {
	           Vector<VolIndex> vovar = m_ebisl[dex].getVoFs(iv+of1[d1]+of2[d2]);
	           if( vovar.size() != 1 )
	           {  
  	              irs |= iv;
	              break;
		   }
                 }
		 else
	           {  
  	              irs |= iv;
	              break;
		   }
		   
	       }
	 }
	 else if( m_ebisl[dex].isIrregular( iv ) )
	 {
	    irs |= iv;
	 }
      }
      irs &= m_domain.domainBox();
   }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::residual(LevelData<EBCellFAB>&       a_residual,
			     const LevelData<EBCellFAB>& a_phi,
			     const LevelData<EBCellFAB>& a_rhs,
			     bool                        a_homogeneousPhysBC)
{
  //this is a multigrid operator so only homogeneous CF BC
  //and null coar level
  applyOp( a_residual, a_phi, NULL, a_homogeneousPhysBC, true );
  scale( a_residual, -1.0 );
  incr( a_residual, a_rhs, 1.0 );
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::preCond(LevelData<EBCellFAB>&       a_lhs,
			    const LevelData<EBCellFAB>& a_rhs)
{
  EBLevelDataOps::assign(a_lhs,a_rhs);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::applyOp(LevelData<EBCellFAB>&             a_opPhi,
			    const LevelData<EBCellFAB>&       a_phi,
			    bool                              a_homogeneousPhysBC)
{
  //homogeneous CFBCs because that is all we can do.
  applyOp( a_opPhi, a_phi, NULL, a_homogeneousPhysBC, true );
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::applyOp(LevelData<EBCellFAB>&                    a_opPhi,
			    const LevelData<EBCellFAB>&              a_phi,
			    const LevelData<EBCellFAB>* const        a_phiCoar,
			    bool                                     a_homogeneousPhysBC,
			    bool                                     a_homogeneousCFBC)
{
  //apply operator without input EB inhomogeneous BC leveldata
  applyOp(a_opPhi, a_phi, a_phiCoar, a_homogeneousPhysBC, a_homogeneousCFBC, NULL);
}



//-----------------------------------------------------------------------
//get EB flux from leveldata if we have one.  otherwise use m_ebBC
//ignore input data in the case of homogeneous Phys BC
void EBAMRNSviscOp::applyOp(LevelData<EBCellFAB>&                    a_res,
			    const LevelData<EBCellFAB>&              a_u,
			    const LevelData<EBCellFAB>* const        a_uCoar,
			    bool                                     a_homogeneousPhysBC,
			    bool                                     a_homogeneousCFBC,
			    const LevelData<BaseIVFAB<Real> >* const a_ebFluxBCLD)
{
  //if input EBFlux leveldata is null, then we have no input eb flux
  bool hasNoEBLevelData = (a_ebFluxBCLD == NULL);

  CH_assert(a_u.nComp() == SpaceDim );
  CH_assert(a_res.nComp() == SpaceDim);
  //  ArrayToRaw atr;

  LevelData<EBCellFAB>& au = (LevelData<EBCellFAB>&)a_u;
  //  if( a_uCoar != NULL )
  //  {
  //      atr.output_array( *a_uCoar, m_ebislCoar, "u0bef.bin" );
  //      atr.output_array( au, m_ebisl, "ubef.bin" );
  //  }
  applyCFBCs( au, a_uCoar, a_homogeneousCFBC );
  au.exchange(au.interval());

  //  if( a_uCoar != NULL )
  //      atr.output_array( au, m_ebisl, "uaft.bin" );


  int nComps = SpaceDim;
  
// Save slices of u, can not find any function to do this 
//  Vector< LevelData<EBCellFAB>* > uu(SpaceDim);
//  for( int c = 0 ; c <SpaceDim ; c++ )
//  {
//    EBCellFactory datafactory( m_ebisl );
//    uu[c] = new LevelData<EBCellFAB>;
//    uu[c]->define( a_u.getBoxes(), 1, a_u.ghostVect(), datafactory );
//    a_u.copyTo( Interval(c,c), *(uu[c]), Interval(0,0) );
//  }
  
  for (DataIterator dit = a_u.dataIterator(); dit.ok(); ++dit)
  {
      DataIndex dex(dit());

      const EBCellFAB&     u              = au[dex];

      const BaseFab<Real>& u_single_valued= u.getSingleValuedFAB();
   //      const EBISBox& curEBISBox = m_ebisl[dex];

      EBCellFAB&           res = a_res[dex];
      BaseFab<Real>&       res_single_valued = res.getSingleValuedFAB();

      //      IntVectSet notRegular = m_ebisl[dex].getIrregIVS(m_dbl.get(dex));

      IntVectSet notRegular = m_irreg[dex];

      //      cout << "irregular set in applyop " << notRegular << endl;
      //      cout << "irregular box in applyop " << m_dbl.get(dex) << endl;
      res.setVal(0.0);
      //      for( int i=0; i<2 ;i++ )
      //	for( int j=5; j<15 ;j++ )
      //	{
      //           Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(IntVect(i,j));
      //           if( vofs.size() > 0 )
      //	     cout << i << " " << j<< " vol frac = " << m_ebisl[dex].volFrac(vofs[0]) << endl;
      //           cout << i << " " << j  << " covered " << m_ebisl[dex].isCovered(IntVect(i,j))<<endl;
      //           cout << i << " " << j  << " irregular " << m_ebisl[dex].isIrregular(IntVect(i,j))<<endl;
      //	}


// Find out about boundary conditions
      Box loBox[SpaceDim],hiBox[SpaceDim];
      int hasLo[SpaceDim],hasHi[SpaceDim];
      Box center, entire;
      Vector<Box> resid_boxes(3);
      for( int dir=0 ; dir <SpaceDim ; dir++ )
      {
         Box centerBox,entireBox;
         ebEllipticLoHiCenter(loBox[dir], hasLo[dir], hiBox[dir], hasHi[dir],
                              center, entire, res.box(), u.box(), m_domain, dir);
         center.growLo(dir);
         resid_boxes[dir] = center;
      }
// Navier-Stokes viscosity at regular cells.

//      cout << "bef" << res_single_valued(IntVect(0,10),0) << endl;
//      cout << "box1" << resid_boxes[0] << endl;
//      cout << "box2" << resid_boxes[1] << endl;
//      cout << "qbox" << u.box() << endl;
//      cout << "rbox" << res.box() << endl;

//      for( BoxIterator b(u.getRegion()); b.ok(); ++b )
//      {
//         if( !m_ebisl[dex].isCovered(b()) )
//	 {
//	     Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
//             if( vofs.size() > 1 )
//	     {
	       //	       for( int iv=0 ; iv < vofs.size() ; iv++ )
		 //		  for( int c=0 ; c < SpaceDim ; c++ )
		    //                    cout << u(vofs[iv],c) << endl;
		  //		    u_single_valued(b(),c) = u(vofs[0],c);
	       //	       cout << u_single_valued(b(),0) << " " << u_single_valued(b(),1) << endl;
      //	     }
      //	 }
      //      }
		  //      for( int i=0 ; i<4 ;i++ )
		  //	for( int j=0 ;j<4 ;j++ )
		  //	{
		  //	  Vector<VolIndex> vf= m_ebisl[dex].getVoFs(IntVect(i,j));
		  //	  cout << "before " << i << " " << j << " " << u_single_valued(IntVect(i,j),0) << endl;
		  //	}
      
      if( m_rz )
         FORT_NSVISCOP_RZ( CHF_FRA(res_single_valued),
			   CHF_CONST_FRA(u_single_valued),
                           CHF_CONST_REALVECT(m_dx),
			   CHF_CONST_FRA1(m_mu[dex].getSingleValuedFAB(),0), 
                           CHF_CONST_FRA1(m_lambda[dex].getSingleValuedFAB(),0),
			   CHF_CONST_REALVECT(m_origin),
			   CHF_BOX(resid_boxes[0]),
			   CHF_BOX(resid_boxes[1]),
			   CHF_BOX(resid_boxes[2]),
			   CHF_BOX(u.box()),
			   CHF_BOX(res.box()) );
      else
         FORT_NSVISCOP( CHF_FRA(res_single_valued),
			CHF_CONST_FRA(u_single_valued),
			CHF_CONST_REALVECT(m_dx),
			CHF_CONST_FRA1(m_mu[dex].getSingleValuedFAB(),0), 
			CHF_CONST_FRA1(m_lambda[dex].getSingleValuedFAB(),0),
			CHF_BOX(resid_boxes[0]),
			CHF_BOX(resid_boxes[1]),
			CHF_BOX(resid_boxes[2]),
			CHF_BOX(u.box()),
			CHF_BOX(res.box()) );


      //      for( int i=0 ; i<4 ;i++ )
      //	for( int j=0 ;j<4 ;j++ )
      //	{
      //	  Vector<VolIndex> vf= m_ebisl[dex].getVoFs(IntVect(i,j));
      //	  cout << "after op" << i << " " << j << " " << res_single_valued(IntVect(i,j),0) << "  ";          cout << m_ebisl[dex].isIrregular(IntVect(i,j)) << " " << m_ebisl[dex].isRegular(IntVect(i,j)) << "  " << m_ebisl[dex].isCovered(IntVect(i,j)) << " " << vf.size() << endl;
      //	}
     
      for( int dir=0 ; dir <SpaceDim ; dir++ )
      {
// Navier-Stokes domain boundary viscous fluxes
         Box centerBox,entireBox;
         ebEllipticLoHiCenter(loBox[dir], hasLo[dir], hiBox[dir], hasHi[dir],
                              centerBox, entireBox, res.box(), u.box(), m_domain, dir);

         if( hasLo[dir] == 1 )
         {
             BaseFab<Real> loFaceFlux( loBox[dir], nComps );
             int side = -1;
             CH_assert(m_domainBC != NULL);
             for( BoxIterator b(loBox[dir]); b.ok(); ++b )
	     {
		Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
                if( vofs.size() > 1 )
                    MayDay::Warning("EBAMRNS::Multiple vofs on domain BC");
                if( vofs.size() > 0 )
		{
                   Vector<Real> face_flux(nComps);
                   if( m_rz )
                      m_domainBC->getFaceFlux_rz( face_flux, 
                                        vofs[0], u, m_origin, m_dx,
					dir, Side::Lo, a_homogeneousPhysBC,
                                        m_mu[dex](vofs[0],0), m_lambda[dex](vofs[0],0), m_time );
	   	   else
                      m_domainBC->getFaceFlux( face_flux, vofs[0], u, m_origin, m_dx,
					dir, Side::Lo, a_homogeneousPhysBC,
                                        m_mu[dex](vofs[0],0), m_lambda[dex](vofs[0],0), m_time );
                   for( int c = 0 ; c < nComps ; c++ )
	              loFaceFlux(b(),c) = face_flux[c];
		}
	     }
             FORT_NS_ADD_BCFLUX( CHF_FRA(res_single_valued),
				 CHF_CONST_FRA( loFaceFlux ),
				 CHF_CONST_REAL( m_dx[dir] ),
				 CHF_CONST_INT( side ),
				 CHF_CONST_INT( dir ),
				 CHF_BOX( loBox[dir] ) );
	     //            cout << "after Lo" << res_single_valued(IntVect(0,10),0) << endl;
	 }
         if( hasHi[dir] == 1 )
         {

             BaseFab<Real> hiFaceFlux( hiBox[dir], nComps );
             int side = 1;
             CH_assert(m_domainBC != NULL);
             for( BoxIterator b(hiBox[dir]); b.ok(); ++b )
	     {
		Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
                if( vofs.size() > 1 )
                    MayDay::Warning("EBAMRNS::Multiple vofs in domain BC");
                if( vofs.size() > 0 )
		{
                   Vector<Real> face_flux(nComps);
                   if( m_rz )
                      m_domainBC->getFaceFlux_rz( face_flux, vofs[0], u, m_origin, m_dx,
                                            dir, Side::Hi, a_homogeneousPhysBC,
                                       m_mu[dex](vofs[0],0),m_lambda[dex](vofs[0],0), m_time );
                   else
                      m_domainBC->getFaceFlux( face_flux, vofs[0], u, m_origin, m_dx,
                                            dir, Side::Hi, a_homogeneousPhysBC,
                                       m_mu[dex](vofs[0],0),m_lambda[dex](vofs[0],0), m_time );
                   for( int c = 0 ; c < nComps ; c++ )
		      hiFaceFlux(b(),c)=face_flux[c];
		}
	     }
             FORT_NS_ADD_BCFLUX( CHF_FRA(res_single_valued),
				 CHF_CONST_FRA( hiFaceFlux ),
				 CHF_CONST_REAL( m_dx[dir] ),
				 CHF_CONST_INT( side ),
				 CHF_CONST_INT( dir ),
				 CHF_BOX( hiBox[dir] ) );
	     //            cout << "after Hi" << res_single_valued(IntVect(0,10),0) << endl;
	 }
      }

      FORT_NS_ADD_LOWORDER( CHF_FRA(res_single_valued),
	                    CHF_CONST_FRA(u_single_valued),
                            CHF_CONST_FRA1(m_rho[dex].getSingleValuedFAB(),0),
			    CHF_CONST_REAL( m_c ),
                            CHF_BOX(res.box()) );

      //      cout << "after lowoder" << res_single_valued(IntVect(0,10),0) << endl;
      for( VoFIterator vi(notRegular,m_ebisl[dex].getEBGraph());vi.ok();++vi)
      {
   	 for( int nc= 0 ; nc<nComps ; nc++ )
	    res(vi(),nc) = 0;
	 if( m_rz )
	 {
	    RealVect c = m_ebisl[dex].centroid(vi());
            Real r = (c[0]+vi().gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
	    res(vi(),0) = -2*m_mu[dex](vi(),0)*u(vi(),0)/(r*r);
	 }
      }
      //      for( int i=0 ; i<4 ;i++ )
      //	for( int j=0 ;j<4 ;j++ )
      //	  cout << "after op+bc" << i << " " << j << " " << res_single_valued(IntVect(i,j),0) << endl;
     
// Navier-Stokes irregular flux update
      for( int dir = 0 ; dir < SpaceDim ; dir++ )
      {
	 if( m_rz )
	 {
	   //	      	           cout << "tan stencils points : " << endl;
	   //	   	   	   IVSIterator ivsit(m_irreg[dex]);
	   //	   	   	   for( ivsit.begin() ; ivsit.ok() ; ++ivsit )
	   //	   		   {
	   //	   		     cout << " ---- at point " << ivsit() << endl;
	   //	     	              Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(ivsit());
	   //	   		      for( int iv=0 ; iv < vofs.size() ;iv++ )
	   //	   			for( int s=0 ; s<m_uxz_stencil[dex](vofs[iv],0).size();s++ )
	   //	   			  cout << m_uxz_stencil[dex](vofs[iv],0).vof(s).gridIndex() << " " << m_uxz_stencil[dex](vofs[iv],0).weight(s) << endl;
	   //	   		   }

	   //	      cout << ivsit() << endl;
	   //           cout << "End of irregular points           " << endl;
	   //           cout << "----------------------------------" << endl;
            if( dir == 0 )
 	       update_irregular_rz( res, u, m_ebisl[dex], m_iflux_stencil[dex],
			         m_uyz_stencil[dex], m_mu[dex], m_lambda[dex], dir,
				 m_irreg[dex], m_irreg[dex] );
            else if( dir == 1 )
 	       update_irregular_rz( res, u, m_ebisl[dex], m_jflux_stencil[dex],
			      m_uxz_stencil[dex], m_mu[dex], m_lambda[dex], dir,
				 m_irreg[dex], m_irreg[dex] );
            else
 	       update_irregular_rz( res, u, m_ebisl[dex], m_kflux_stencil[dex],
			      m_uxy_stencil[dex], m_mu[dex], m_lambda[dex], dir,
				 m_irreg[dex], m_irreg[dex] );
         }
	 else
	 {
            if( dir == 0 )
 	       update_irregular( res, u, m_ebisl[dex], m_iflux_stencil[dex],
				 m_uyz_stencil[dex], m_mu[dex], m_lambda[dex], dir,
				 m_irreg[dex], m_irreg[dex] );
            else if( dir == 1 )
 	       update_irregular( res, u, m_ebisl[dex], m_jflux_stencil[dex],
				 m_uxz_stencil[dex], m_mu[dex], m_lambda[dex], dir,
				 m_irreg[dex], m_irreg[dex] );
            else
 	       update_irregular( res, u, m_ebisl[dex], m_kflux_stencil[dex],
				 m_uxy_stencil[dex], m_mu[dex], m_lambda[dex], dir,
				 m_irreg[dex], m_irreg[dex] );
	 }
      }
      //      cout << "after irreg" << res_single_valued(IntVect(0,10),0) << endl;
// EB-fluxes, and domain boundary fluxes in irregular cells:
//      for( int i=0 ; i<4 ;i++ )
//	for( int j=0 ;j<4 ;j++ )
//	  cout << "after updirreg" << i << " " << j << " " << res_single_valued(IntVect(i,j),0) << endl;

      Real cofi=1/m_c;
      for( VoFIterator vi(notRegular,m_ebisl[dex].getEBGraph());vi.ok();++vi)       
      {
// EB-fluxes:
	  VolIndex vof=vi();

// Assume constant viscosity for now
	  Real mu=m_mu[dex](vof,0);
	  Real la=m_lambda[dex](vof,0);

          Vector<Real> ebFlux(nComps,0);
          if( hasNoEBLevelData )
          {
	     if( m_rz )
	        m_ebBC->getEBFlux_rz( ebFlux, vof, au, dex, m_origin, m_dx, a_homogeneousPhysBC,mu, la,m_time );
	     else
	        m_ebBC->getEBFlux( ebFlux, vof, au, dex, m_origin, m_dx, a_homogeneousPhysBC,
			       mu, la,m_time );
          }
          else
          {
            MayDay::Warning("EBAMRNS::applyOp. EB alternative not implemented");
          }
	  //          cout << "ebf " << vof.gridIndex() << " " << ebFlux[0] << " " << ebFlux[1] << endl;
          for( int comp=0 ; comp < nComps ; comp++ )
             res(vof,comp) += ebFlux[comp]*m_dxScale;

	  //      cout << "after EB" << res_single_valued(IntVect(0,10)) << endl;
// Domain BC for irregular cells:
          for( int dir = 0 ; dir < SpaceDim ; dir++ )
	  {
             if( hasLo[dir] == 1 )
             {
                if( loBox[dir].contains(vof.gridIndex()))
                {
                    Vector<Real> flux(nComps);
                    if( m_rz )
                       m_domainBC->getFaceFlux_rz( flux, vof, u, m_origin, m_dx, dir,
					     Side::Lo, a_homogeneousPhysBC,
					     m_mu[dex](vof,0), m_lambda[dex](vof,0),m_time);
                    else
                       m_domainBC->getFaceFlux( flux, vof, u, m_origin, m_dx, dir,
					     Side::Lo, a_homogeneousPhysBC,
					     m_mu[dex](vof,0), m_lambda[dex](vof,0),m_time);
                    for( int comp=0 ; comp < nComps ; comp++ )
                       res(vof,comp) -= flux[comp]*m_invDx[dir];
                }
	     }
             if( hasHi[dir] == 1 )
             {
                if( hiBox[dir].contains(vof.gridIndex()) )
                {
                   Vector<Real> flux(nComps);
                   if( m_rz )
                      m_domainBC->getFaceFlux_rz( flux, vof, u, m_origin, m_dx, dir,
					     Side::Hi, a_homogeneousPhysBC,
					     m_mu[dex](vof,0), m_lambda[dex](vof,0),m_time);
                   else
                      m_domainBC->getFaceFlux( flux, vof, u, m_origin, m_dx, dir,
					     Side::Hi, a_homogeneousPhysBC,
					     m_mu[dex](vof,0), m_lambda[dex](vof,0),m_time);
                   for( int comp=0 ; comp < nComps ; comp++ )
                      res(vof,comp) += flux[comp]*m_invDx[dir];
                 }
             }
	  }
	  //      cout << "after EB bc " << res_single_valued(IntVect(7,50),0) << endl;
// Low order term
          Real hhterm=0;
          for( int comp= 0 ; comp < nComps ; comp++ )
	  {
	     hhterm = 0;
             if( m_split_u )
	     {
	       //	       if( !m_irreg_w_ghost[dex].contains(vof.gridIndex()) )
	       //	       {
	       //		 cout << "ustencil does not contain vof" << endl;
	       //		 cout << m_irreg[dex].contains(vof.gridIndex()) << " " <<
	       //		   m_irreg_w_ghost[dex].contains(vof.gridIndex()) << endl;
	       //	       }
                const VoFStencil& u_stencil = m_uStencil[dex](vof,0);
                for (int i = 0; i < u_stencil.size(); i++)
                   hhterm += u_stencil.weight(i)*u(u_stencil.vof(i),comp)*m_rho[dex](u_stencil.vof(i),0);
	     }
             else
	        hhterm = u(vof,comp)*m_rho[dex](vof,0);
             res(vof,comp) -= cofi*hhterm*(m_ebisl[dex].volFrac(vof));
	  }
	  //      cout << "after low order term" << res_single_valued(IntVect(7,50),0) << endl;
      }

  }
}
	     
	     

//-----------------------------------------------------------------------
void EBAMRNSviscOp::create( LevelData<EBCellFAB>&       a_lhs,
			    const LevelData<EBCellFAB>& a_rhs )
{
  int ncomp = a_rhs.nComp();
  EBCellFactory ebcellfact(m_ebisl);
  a_lhs.define(m_dbl, ncomp, a_rhs.ghostVect(), ebcellfact );
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::createCoarsened(LevelData<EBCellFAB>&       a_lhs,
				    const LevelData<EBCellFAB>& a_rhs,
				    const int &                 a_refRat)
{
  //  cout << "create coarsened " << endl;
  int ncomp = a_rhs.nComp();
  IntVect ghostVect = a_rhs.ghostVect();

  CH_assert(m_dbl.coarsenable(a_refRat));

  //fill ebislayout
  DisjointBoxLayout dblCoarsenedFine;
  coarsen(dblCoarsenedFine, m_dbl, a_refRat);

  EBISLayout ebislCoarsenedFine;
  IntVect ghostVec = a_rhs.ghostVect();
  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  ebisPtr->fillEBISLayout(ebislCoarsenedFine, dblCoarsenedFine, m_domainCoar.domainBox(), ghostVec[0]);
  if(m_refToCoar > 1)
    {
      ebislCoarsenedFine.setMaxRefinementRatio(m_refToCoar,ebisPtr);
    }

  //create coarsened data
  EBCellFactory ebcellfactCoarsenedFine(ebislCoarsenedFine);
  a_lhs.define(dblCoarsenedFine, ncomp, ghostVec, ebcellfactCoarsenedFine);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::assign(LevelData<EBCellFAB>&       a_lhs,
			   const LevelData<EBCellFAB>& a_rhs)
{
  EBLevelDataOps::assign(a_lhs,a_rhs);
}


//-----------------------------------------------------------------------
Real EBAMRNSviscOp::dotProduct(const LevelData<EBCellFAB>& a_1,
			       const LevelData<EBCellFAB>& a_2)
{
  CH_assert(a_1.nComp() == a_2.nComp() );

  ProblemDomain domain;
  Real volume=0;
  Real dprod = 0;

  for( int c = 0 ; c < a_1.nComp() ; c++ )
  {
     for( DataIterator d=a_1.getBoxes().dataIterator() ; d.ok() ; ++d )
        for( BoxIterator b(a_1[d()].getRegion()) ; b.ok() ; ++b )
           if( !a_1[d()].getEBISBox().isCovered(b()) )
           {
	      Vector<VolIndex> vofs=a_1[d()].getEBISBox().getVoFs(b());
              for( int vi=0 ; vi < vofs.size() ; vi++ )
	      {
                 dprod  += a_1[d()](vofs[vi],c)*a_2[d()](vofs[vi],c)*a_1[d()].getEBISBox().volFrac(vofs[vi]);
                 volume += a_1[d()].getEBISBox().volFrac(vofs[vi]);
	      }
	   }
  }
  volume /= a_1.nComp();
  if( volume > 0 )
    dprod /= volume;
  return dprod;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::incr(LevelData<EBCellFAB>&       a_lhs,
			 const LevelData<EBCellFAB>& a_x,
			 Real                        a_scale)
{
  EBLevelDataOps::incr(a_lhs,a_x,a_scale);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::axby(LevelData<EBCellFAB>&       a_lhs,
			 const LevelData<EBCellFAB>& a_x,
			 const LevelData<EBCellFAB>& a_y,
			 Real                        a_a,
			 Real                        a_b)
{
  EBLevelDataOps::axby(a_lhs,a_x,a_y,a_a,a_b);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::scale(LevelData<EBCellFAB>& a_lhs,
			  const Real&           a_scale)
{
  EBLevelDataOps::scale(a_lhs,a_scale);
}


//-----------------------------------------------------------------------
Real EBAMRNSviscOp::norm(const LevelData<EBCellFAB>& a_rhs,
			 int                         a_ord)
{
  Vector<Real> nrm(a_rhs.nComp());
  Real volume = 0;
  for( int c = 0 ; c < a_rhs.nComp() ; c++ )
  {
     nrm[c] = 0;
     for( DataIterator d=a_rhs.getBoxes().dataIterator() ; d.ok() ; ++d )
        for( BoxIterator b(a_rhs[d()].box()) ; b.ok() ; ++b )
           if( !a_rhs[d()].getEBISBox().isCovered(b()) )
           {
	      Vector<VolIndex> vofs=a_rhs[d()].getEBISBox().getVoFs(b());
              for( int vi=0 ; vi < vofs.size() ; vi++ )
              {
		 if( c == 0 )
		     volume += a_rhs[d()].getEBISBox().volFrac(vofs[vi]);
		 if( a_ord == 0 )
		 {
		   if( nrm[c] < fabs(a_rhs[d()](vofs[vi],c)) )
		       nrm[c] = fabs(a_rhs[d()](vofs[vi],c));
		 }
		 else
	            nrm[c] += pow(a_rhs[d()](vofs[vi],c),a_ord)*a_rhs[d()].getEBISBox().volFrac(vofs[vi]);
	      }
	   }
  }


  if( volume > 0 && a_ord != 0 )
    for( int c = 0 ; c < a_rhs.nComp() ; c++ )
      nrm[c] /= volume;

  int norm_alternative = 0;
  double retval = 0;
  if( norm_alternative == 1 )
  {
     if( a_ord > 0 )
     {
        for( int c = 0 ; c < a_rhs.nComp() ; c++ )
           nrm[c] = pow(nrm[c],1.0/a_ord);
     }
     for( int c = 0 ; c < a_rhs.nComp() ; c++ )
     {
        if( retval < nrm[c] )
           retval = nrm[c];
     }
  }
  else
  {
     if( a_ord > 0 )
     {
        for( int c = 0 ; c < a_rhs.nComp() ; c++ )
	  retval += nrm[c];
        retval = pow(retval,1.0/a_ord);
     }
     else
        for( int c = 0 ; c < a_rhs.nComp() ; c++ )
	   if( retval < nrm[c] )
	      retval = nrm[c];
  }
  return retval;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::setToZero(LevelData<EBCellFAB>& a_lhs)
{
  EBLevelDataOps::setToZero(a_lhs);
}


// MGLevelOp functions
//-----------------------------------------------------------------------
void EBAMRNSviscOp::createCoarser(LevelData<EBCellFAB>&       a_coar,
				  const LevelData<EBCellFAB>& a_fine,
				  bool                        a_ghosted )
{
  //  CH_assert(a_fine.nComp() == 1);
  //  cout << "create coarser " << endl;
  DisjointBoxLayout dbl;
  Box coarDom = coarsen(m_domain.domainBox(), 2);
  // This is multigrid, so coarsen by 2
  if(a_fine.disjointBoxLayout().coarsenable(2))
  {
      coarsen(dbl,a_fine.disjointBoxLayout(),2);
  }
  else
  {
      //otherwise check to see if we are covering domain.
      //if so, make one box filling domain.
      //otherwise, SOL
      IntVectSet ivsDomain(m_domain.domainBox());
      for(LayoutIterator lit = m_dbl.layoutIterator(); lit.ok(); ++lit)
        {
          ivsDomain -= m_dbl.get(lit());
        }
      if(ivsDomain.isEmpty())
        {
          Vector<Box> boxes(1, coarDom);
          Vector<int> procs(1, 0);
          dbl.define(boxes, procs);
        }
      else
        {
          MayDay::Error("EBAMRP::createCoarse error---dbl not coarsenable");
        }
    }

  int nghost = a_fine.ghostVect()[0];
  EBISLayout coarEBISL;

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  ebisPtr->fillEBISLayout(coarEBISL, dbl, coarDom, nghost);

  EBCellFactory ebcellfact(coarEBISL);
  if( a_ghosted )
      a_coar.define(dbl, a_fine.nComp(), a_fine.ghostVect(), ebcellfact);
  else
      a_coar.define(dbl, a_fine.nComp(), IntVect::Zero, ebcellfact);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::relax(LevelData<EBCellFAB>&       a_e,
			  const LevelData<EBCellFAB>& a_residual,
			  int                         a_iterations)
{
  //  CH_assert(a_e.nComp() == 1);
  //  CH_assert(a_residual.nComp() == 1);
  //debug!
  //m_relaxType = 0;
  //end debug
  if (m_relaxType == 0)
    {
      for (int i = 0; i < a_iterations; i++)
        {
          levelJacobi(a_e,a_residual);
        }
    }
  else  if (m_relaxType == 1)
    {
      for (int i = 0; i < a_iterations; i++)
        {
          levelMulticolorGS(a_e,a_residual);
        }
    }
  else
    {
      MayDay::Error("EBAMRNSviscOp::relax - invalid relaxation type");
    }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::restrictResidual(LevelData<EBCellFAB>&       a_resCoar,
				     LevelData<EBCellFAB>&       a_phiThisLevel,
				     const LevelData<EBCellFAB>& a_rhsThisLevel)
{

  //  CH_assert(a_resCoar.nComp() == 1);
  //  CH_assert(a_phiThisLevel.nComp() == 1);
  //  CH_assert(a_rhsThisLevel.nComp() == 1);

  //  cout << "restrict resid" << endl;

  int nc = a_rhsThisLevel.nComp();
  CH_assert( a_resCoar.nComp() == nc );
  CH_assert( a_phiThisLevel.nComp() == nc );

  LevelData<EBCellFAB> resThisLevel;
  bool homogeneous = true;

  EBCellFactory ebcellfactTL(m_ebisl);
  IntVect ghostVec = a_rhsThisLevel.ghostVect();

  resThisLevel.define(m_dbl, nc, ghostVec, ebcellfactTL);

  // Get the residual on the fine grid
  residual(resThisLevel,a_phiThisLevel,a_rhsThisLevel,homogeneous);

  // now use our nifty averaging operator
  Interval variables(0, nc-1);
  if(m_hasMGObjects)
    {
      m_ebAverageMG.average(a_resCoar, resThisLevel, variables);
    }
  else
    {
      CH_assert(m_hasInterpAve);
      m_ebAverage.average(a_resCoar, resThisLevel, variables);
    }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::prolongIncrement(LevelData<EBCellFAB>&       a_phiThisLevel,
				     const LevelData<EBCellFAB>& a_correctCoar)
{
  CH_assert( a_phiThisLevel.nComp() == a_correctCoar.nComp() );

  //  cout << "prolinc" << endl;

  Interval vars(0, a_correctCoar.nComp()-1);
  if(m_hasMGObjects)
    {
      m_ebInterpMG.pwcInterp(a_phiThisLevel, a_correctCoar, vars);


// debug
//      for (DataIterator dit = a_phiThisLevel.dataIterator(); dit.ok(); ++dit)
//      {
//          DataIndex dex(dit());
//          const EBCellFAB& u = a_phiThisLevel[dex];
//          const EBCellFAB& d = a_correctCoar[dex];
	  //          int ni = d.box().size(0);
	  //          int nj = d.box().size(1);
	  //          double dut* = new double[ni*nj*2];

      //          for( BoxIterator b(d.box()); b.ok(); ++b )
      //	  if( !d.getEBISBox().isCovered(b()) )
      //	  {
      //	      Vector<VolIndex> vofs = d.getEBISBox().getVoFs(b());
      //              for( int vi=0 ; vi < vofs.size() ; vi++ )
      //	      {
      //                 if( vofs.size() > 1 )
      //		    cout << "voff voff" << endl;
      //                  int i=b()[0], j=b()[1];
      //                  cout << "p(" << i << "," << j << ")= " << d(vofs[vi],1) << endl;

		  //                  dut[i+2*ni*j  ]   = d(vofs[vi],0);
		  //                  dut[i+2*ni*j+1]   = d(vofs[vi],1);
      //	       }
      //	    }


	  //      int fd=open("dut.bin", O_CREAT | O_TRUNC | O_WRONLY, 0660 );
	  //      write(fd,&ni,sizeof(int));
	  //      write(fd,&nj,sizeof(int));
	  //      write(fd,dut,sizeof(double)*ni*nj);
	  //      close(fd);
	  //	  delete[] dut;
      //      }
    }
  else
    {
      CH_assert(m_hasInterpAve);
      m_ebInterp.pwcInterp(a_phiThisLevel, a_correctCoar, vars);
    }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::applyCFBCs( LevelData<EBCellFAB>&             a_phi,
			        const LevelData<EBCellFAB>* const a_phiCoar,
			        bool a_homogeneousCFBC )
{
   int ncomp=a_phi.nComp();
   if( m_hasCoar )
   {
      if( !a_homogeneousCFBC )
      {
         if( a_phiCoar==NULL )
            MayDay::Error("cannot enforce inhomogeneous CFBCs with NULL coar");

          CH_assert(a_phiCoar->nComp() == SpaceDim );

          EBQuadCFInterp quadCFIWithCoar(m_dbl,     m_dblCoar,
                                         m_ebisl, m_ebislCoar,
                                         m_domainCoar.domainBox(),
                                         m_refToCoar, ncomp);

          Interval interv(0,ncomp-1);
          quadCFIWithCoar.interpolate(a_phi, *a_phiCoar, interv);
          corner_cfbcs( a_phi, m_ebisl );
      }
      else
      {
         applyHomogeneousCFBCs(a_phi);
      }
   }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::corner_cfbcs( LevelData<EBCellFAB>& a_phi,
				  const EBISLayout& a_eb)
{
#if CH_SPACEDIM == 2
   int ic, jc, il, jl;
   for( DataIterator di=a_phi.dataIterator() ; di.ok() ; ++di )
   {

      Box bx=a_phi[di()].box();
      BaseFab<Real>& phi = a_phi[di()].getSingleValuedFAB();

      for( int corn = 0 ; corn < 4 ; corn++ )
      {
	 if( corn == 0 )
	 {
            ic = bx.smallEnd(0);
            jc = bx.smallEnd(1);
            il = 1;
            jl = 1;
	 }
	 else if( corn == 1 )
	 {
            ic = bx.bigEnd(0);
            jc = bx.smallEnd(1);
            il =-1;
            jl = 1;
	 }
	 else if( corn == 2 )
	 {
            ic = bx.bigEnd(0);
            jc = bx.bigEnd(1);
            il = -1;
            jl = -1;
	 }
         else
	 {
            ic = bx.smallEnd(0);
            jc = bx.bigEnd(1);
            il = 1;
            jl = -1;
	 }
	 int order=3;
//    phi(ic,jc,c) = 2*phi(ic+il,jc+jl,c)-phi(ic+2*il,jc+2*jl,c);
//         IntVect co(ic,jc), copi(ic+il,jc), copj(ic,jc+jl);
//         IntVect coppi(ic+2*il,jc), coppj(ic,jc+2*jl);
         IntVect co(ic,jc), ofi(il,0), ofj(0,jl);

         int s = 0;
         while( (a_eb[di()].getVoFs(co+s*ofi).size()==1 &&
		      a_eb[di()].getVoFs(co+s*ofj).size()==1)&& s<4 )
           s++;

	 order = s-1;
         for( int c=0 ; c < SpaceDim ; c++ )
	   if( order == 1 )
	      phi(co,c) = 0.5*( phi(co+ofi,c) + phi(co+ofj,c) );
	   else if( order == 2 )
              phi(co,c) = 0.5*( 2*phi(co+ofi,c) - phi(co+2*ofi,c) +
				2*phi(co+ofj,c) - phi(co+2*ofj,c) );
	   else if( order == 3 )
              phi(co,c) = 0.5*( 3*phi(co+ofi,c)-3*phi(co+2*ofi,c) + phi(co+3*ofi,c) +
                                3*phi(co+ofj,c)-3*phi(co+2*ofj,c) + phi(co+3*ofj,c) );
           else
 	      phi(co,c) = 2*phi(co+ofi+ofj,c) - phi(co+2*ofi+2*ofj,c);
	     //	     phi(co,c) = phi(co+ofi+ofj,c);
      }
   }
#endif
}

//-----------------------------------------------------------------------
void EBAMRNSviscOp::applyHomogeneousCFBCs( LevelData<EBCellFAB>& a_phi )
{
//  CH_assert(a_phi.nComp() == 1);
   CH_assert( a_phi.ghostVect() >= IntVect::Unit);
   for( DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit )
      for( int idir = 0; idir < SpaceDim; idir++ )
          for( SideIterator sit; sit.ok(); sit.next() )
              applyHomogeneousCFBCs( a_phi[dit()], dit(), idir, sit() );
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::applyHomogeneousCFBCs(EBCellFAB&            a_phi,
					  const DataIndex&      a_datInd,
					  int                   a_idir,
					  Side::LoHiSide        a_hiorlo)
{
  CH_assert((a_idir >= 0) && (a_idir  < SpaceDim));
  CH_assert((a_hiorlo == Side::Lo )||(a_hiorlo == Side::Hi ));
  //  CH_assert(a_phi.nComp() == 1);
  //  int ivar = 0;
  int ncomp = a_phi.nComp();
  
  const CFIVS* cfivsPtr = NULL;

  if (a_hiorlo == Side::Lo)
    {
      cfivsPtr = &m_loCFIVS[a_idir][a_datInd];
    }
  else
    {
      cfivsPtr = &m_hiCFIVS[a_idir][a_datInd];
    }

  const IntVectSet& interpIVS = cfivsPtr->getFineIVS();
  if (!interpIVS.isEmpty())
    {

      int ihilo = sign(a_hiorlo);

      Real halfdxcoar = m_dxCoar[a_idir]/2.0;
      Real halfdxfine = m_dx[a_idir]/2.0;
      Real xg = halfdxcoar -   halfdxfine;
      Real xc = halfdxcoar +   halfdxfine;
      Real xf = halfdxcoar + 3*halfdxfine;
      Real hf = m_dx[a_idir];
      Real denom = xf*xc*hf;

      const EBGraph&  ebgraph = m_ebisl[a_datInd].getEBGraph();
      for(VoFIterator vofit(interpIVS, ebgraph); vofit.ok(); ++vofit)
        {
          const VolIndex& VoFGhost = vofit();

          IntVect ivGhost  = VoFGhost.gridIndex();
          IntVect ivClose =  ivGhost;
          IntVect ivFar   =  ivGhost;

          ivClose[a_idir] -=  ihilo;
          ivFar[a_idir]   -= 2*ihilo;

          Vector<VolIndex> closeVoFs = ebgraph.getVoFs(ivClose);
          Vector<VolIndex>   farVoFs = ebgraph.getVoFs(ivFar);

          bool hasClose = (closeVoFs.size() == 1);
          bool hasFar   = (farVoFs.size()   == 1);

          Vector<Real> phiGhost(ncomp);

          if(hasClose && hasFar)
            {
              const VolIndex& VoFClose = closeVoFs[0];
              const VolIndex& VoFFar = farVoFs[0];

              for( int c=0 ; c<ncomp ; c++ )
	      {
                 Real phic = a_phi(VoFClose, c);
                 Real phif = a_phi(VoFFar, c);

              // quadratic interpolation  phi = ax^2 + bx + c
                 Real A = (phif*xc - phic*xf)/denom;
                 Real B = (phic*hf*xf - phif*xc*xc + phic*xf*xc)/denom;

                 phiGhost[c] = A*xg*xg + B*xg;
	      }
            }
          else if(hasClose)
            {
              const VolIndex& VoFClose = closeVoFs[0];
              for( int c=0 ; c<ncomp ; c++ )
	      {
                 Real phic = a_phi(VoFClose, c);
              //linear interpolation
                 Real slope =  phic/xc;
                 phiGhost[c]   =  slope*xg;
	      }
            }
          else
            {
              for( int c=0 ; c<ncomp ; c++ )
                 phiGhost[c] = 0.0; //nothing to interpolate from
            }
          for( int c=0 ; c<ncomp ; c++ )      
             a_phi(VoFGhost, c) = phiGhost[c];
        }
    }
}


//-----------------------------------------------------------------------
int EBAMRNSviscOp::refToCoarser()
{
  return m_refToCoar;
}


//-----------------------------------------------------------------------
int EBAMRNSviscOp::refToFiner()
{
  return m_refToFine;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::AMRResidual(LevelData<EBCellFAB>&       a_residual,
				const LevelData<EBCellFAB>& a_phiFine,
				const LevelData<EBCellFAB>& a_phi,
				const LevelData<EBCellFAB>& a_phiCoar,
				const LevelData<EBCellFAB>& a_rhs,
				bool a_homogeneousPhysBC )
{
   CH_assert(a_residual.nComp() == SpaceDim );
   CH_assert(a_phi.nComp() == SpaceDim );
   CH_assert(a_rhs.nComp() == SpaceDim );

   setToZero( a_residual );
   applyOp( a_residual, a_phi, &a_phiCoar,  a_homogeneousPhysBC, false);
   if( m_hasFine )
      reflux(a_residual, a_phiFine, a_phi, 0 );

  //  a_residual  :=  rhs - L(phi)
   scale(a_residual,-1.0);
   incr(a_residual,a_rhs,1.0);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::AMRResidual_eeq(LevelData<EBCellFAB>&   a_residual,
				    const LevelData<EBCellFAB>* a_phiFine,
				    const LevelData<EBCellFAB>& a_phi,
				    const LevelData<EBCellFAB>* a_phiCoar,
				    bool a_homogeneousPhysBC)
{
   CH_assert(a_residual.nComp() == 1);
   CH_assert(a_phi.nComp() == SpaceDim );

   setToZero( a_residual );
   {
      get_energyeq_viscosity( a_residual, a_phi, a_phiCoar,
			      a_homogeneousPhysBC, false, NULL );
      //      ArrayToRaw atr;
      //      atr.chk_nan(a_residual,m_ebisl," res in EBAMRNSvisc is nan ");
      if( m_hasFine && a_phiFine != NULL )
         reflux( a_residual, *a_phiFine, a_phi, 1 );
   }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::reflux( LevelData<EBCellFAB>& a_residual,
			    const LevelData<EBCellFAB>& a_phiFine,
			    const LevelData<EBCellFAB>& a_phi,
			    int a_energy_eq )
{
  int ncomp = a_residual.nComp();
  if( a_energy_eq == 1 )
     ncomp = 1;

  Interval interv(0,ncomp-1);
  if(!m_fluxReg.isDefined())
  {
     const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
     m_fluxReg.define(m_dblFine,   m_dbl,
                      m_ebislFine, m_ebisl,
                      m_domain.domainBox(),
                      m_refToFine, ncomp, ebisPtr );
  }

  m_fluxReg.setToZero();
  incrementFRCoar( m_fluxReg, a_phiFine, a_phi, a_energy_eq );
  incrementFRFine( m_fluxReg, a_phiFine, a_phi, a_energy_eq );
  for(int idir = 0; idir < SpaceDim; idir++)
  {
      Real scale = 1.0/m_dx[idir];
      m_fluxReg.reflux( a_residual, interv, scale, idir );
  }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::incrementFRCoar(EBFluxRegister&             a_fluxReg,
				    const LevelData<EBCellFAB>& a_phiFine,
				    const LevelData<EBCellFAB>& a_phi,
				    int a_energy_eq )
{
  CH_assert(a_phiFine.nComp() == SpaceDim );
  CH_assert(a_phi.nComp() == SpaceDim );

  int ncomp = SpaceDim;
  Interval interv(0,ncomp-1);

  if( a_energy_eq == 1 )
     interv.define(0,0);

  for(DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
    {
      const EBCellFAB& coarfab = a_phi[dit()];
      const EBISBox& ebisBox = m_ebisl[dit()];
      const Box&  box = m_dbl.get(dit());

      BoxIterator b(box);
      while( b.ok() && ebisBox.isCovered(b()) )
	 ++b;

      Real mu=0, la=0;
      if( b.ok() )
      {
         Vector<VolIndex> vofs = ebisBox.getVoFs(b());
   
         mu = m_mu[dit()](vofs[0],0);
         la = m_lambda[dit()](vofs[0],0);
      }

      for (int idir = 0; idir < SpaceDim; idir++)
      {
          EBFaceFAB coarflux(ebisBox, box, idir, ncomp);

          if( m_rz )
             getFlux_rz( coarflux, coarfab, box, ebisBox, m_dx, idir, mu, la, a_energy_eq );
          else
             getFlux( coarflux, coarfab, box, ebisBox, m_dx, idir, mu, la, a_energy_eq );

          Real scale = 1.0;
          if( m_rz && idir == 0 )
	     incrementCoarseBothRZ( a_fluxReg, coarflux, scale,
                                   dit(), interv, idir, m_dx[0] );
          else
             a_fluxReg.incrementCoarseBoth(coarflux, scale,
                                           dit(), interv, idir);
      }
    }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::incrementFRFine(EBFluxRegister&             a_fluxReg,
				    const LevelData<EBCellFAB>& a_phiFine,
				    const LevelData<EBCellFAB>& a_phi,
				    int a_energy_eq )
{
  CH_assert(a_phiFine.nComp() == SpaceDim );
  CH_assert(a_phi.nComp() == SpaceDim);
  CH_assert(m_hasFine);
  int ncomp = SpaceDim;
  Interval interv(0,ncomp-1);
  //  int ncomp_interp = ncomp;
  //  if( a_energy_eq == 1 )
  //  {
  //     interv.define(0,0);
  //     ncomp_interp = 1;
  //  }

  // need to do cfinterpolation in aggregate before
  // any getFineData calls are made
  //define CF interpolation object with finer level
  //define coarse fine interpolation object on the fly
  //because most operators do not need it
  EBQuadCFInterp quadCFIWithFine(m_dblFine,     m_dbl,
                                 m_ebislFine, m_ebisl,
                                 m_domain.domainBox(),
                                 m_refToFine, ncomp );

  //ghost cells of phiFine need to be filled
  LevelData<EBCellFAB>& phiFine = (LevelData<EBCellFAB>&) a_phiFine;
  //   ArrayToRaw atr;
   //  for(DataIterator di = m_dbl.dataIterator(); di.ok(); ++di)
   //    {
   //      const EBCellFAB& u = a_phi[di()];
   //      atr.print_ebcellfab(u,"ua");
   //    }

   //  for(DataIterator di = m_dbl.dataIterator(); di.ok(); ++di)
   //    {
   //      const EBCellFAB& u = a_phi[di()];
   //      atr.print_ebcellfab(u,"ua");
   //    }


  quadCFIWithFine.interpolate(phiFine, a_phi, interv);

  //  for(DataIterator di = m_dbl.dataIterator(); di.ok(); ++di)
  //    {
  //      const EBCellFAB& u = a_phi[di()];
  //      atr.print_ebcellfab(u,"ub");
  //    }

  phiFine.exchange(interv);


  DataIterator d = a_phi.dataIterator();
  d.reset();

  BoxIterator b(a_phi[d()].box());
  while( b.ok() && m_ebisl[d()].isCovered(b()) )
    ++b;

  Real mu=0, la=0;
  if( b.ok() )
  {
     Vector<VolIndex> vofs = m_ebisl[d()].getVoFs(b());
     mu = m_mu[d()](vofs[0],0);
     la = m_lambda[d()](vofs[0],0);
  }

  DataIterator ditf = a_phiFine.dataIterator();
  for (ditf.reset(); ditf.ok(); ++ditf)
  {
      const Box&     boxFine = m_dblFine.get(ditf());
      const EBISBox& ebisBoxFine = m_ebislFine[ditf()];
      const EBCellFAB& phiFine = a_phiFine[ditf()];

      for (int idir = 0; idir < SpaceDim; idir++)
        {
          SideIterator sit;
          for (sit.begin(); sit.ok(); sit.next())
            {
              Box fabBox = adjCellBox(boxFine, idir, sit(), 1);
              fabBox.shift(idir, -sign(sit()));

              EBFaceFAB fluxFine(ebisBoxFine, fabBox, idir, ncomp);
              if( m_rz )
                 getFlux_rz(fluxFine, phiFine, fabBox, ebisBoxFine, m_dxFine,
		      idir, mu, la, a_energy_eq );
              else
                 getFlux(fluxFine, phiFine, fabBox, ebisBoxFine, m_dxFine,
		      idir, mu, la, a_energy_eq );

              Real scale = 1.0;
              if( a_energy_eq == 1 )
	      {
		 Interval first(0,0);
                 EBFaceFAB fluxFine_eeq(ebisBoxFine, fabBox, idir, 1);
		 fluxFine_eeq.copy(fabBox,first,fabBox,fluxFine,first);
                 if( m_rz && idir == 0 )
                    incrementFineBothRZ( a_fluxReg, fluxFine_eeq, scale, 
                                ditf(), first,  idir,    sit(), m_dxFine[0]);
                 else
                    a_fluxReg.incrementFineBoth(fluxFine_eeq, scale, ditf(),
                                             first,  idir,    sit());
	      }
              else if( m_rz && idir == 0 )
 	          incrementFineBothRZ( a_fluxReg, fluxFine, scale,
                                        ditf(), interv, idir, sit(), m_dxFine[0] );

	      else 
                 a_fluxReg.incrementFineBoth(fluxFine, scale, ditf(),
                                             interv,  idir,    sit());
            }
        }
    }
}
//-----------------------------------------------------------------------
void EBAMRNSviscOp::incrementCoarseBothRZ(
				    EBFluxRegister& a_fluxReg,
				    const EBFaceFAB& a_coarFlux,
                                    const Real&      a_scale,
                                    const DataIndex& a_coarDatInd,
                                    const Interval&  a_variables,
                                    const int&       a_dir,
				    const Real&      a_dx)
{
  a_fluxReg.incrementCoarseRegulRZ(a_coarFlux, a_scale, a_coarDatInd, 
				     a_variables, a_dir, a_dx);
  
  IntVectSet ivs(a_coarFlux.getCellRegion());
  const EBISBox& ebisBoxCoar = a_coarFlux.getEBISBox();
  BaseIFFAB<Real> irregFlux(ivs, ebisBoxCoar.getEBGraph(), a_dir, a_coarFlux.nComp());

  for(FaceIterator faceit(ivs, ebisBoxCoar.getEBGraph(), a_dir,
                          FaceStop::SurroundingWithBoundary);
      faceit.ok(); ++faceit)
    {
      for(int icomp = 0; icomp < a_coarFlux.nComp(); icomp++)
        {
          irregFlux(faceit(), icomp) = a_coarFlux(faceit(), icomp);
        }
    }
  a_fluxReg.incrementCoarseIrregulRZ(irregFlux, a_scale, a_coarDatInd,
				       a_variables, a_dir, a_dx );
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::incrementFineBothRZ(
				  EBFluxRegister& a_fluxReg,
				  const EBFaceFAB&      a_fineFlux,
                                  const Real&           a_scale,
                                  const DataIndex&      a_fineDatInd,
                                  const Interval&       a_variables,
                                  const int&            a_dir,
                                  const Side::LoHiSide& a_sd,
                                  const Real&           a_dx )
{
  a_fluxReg.incrementFineRegulRZ(a_fineFlux, a_scale, a_fineDatInd,
				   a_variables, a_dir, a_sd, a_dx);

  IntVectSet ivs(a_fineFlux.getCellRegion());
  const EBISBox& ebisBoxFine = a_fineFlux.getEBISBox();
  BaseIFFAB<Real> irregFlux(ivs, ebisBoxFine.getEBGraph(), a_dir, a_fineFlux.nComp());

  for(FaceIterator faceit(ivs, ebisBoxFine.getEBGraph(), a_dir,
                          FaceStop::SurroundingWithBoundary);
      faceit.ok(); ++faceit)
    {
      for(int icomp = 0; icomp < a_fineFlux.nComp(); icomp++)
        {
          irregFlux(faceit(), icomp) = a_fineFlux(faceit(), icomp);
        }
    }
  a_fluxReg.incrementFineIrregulRZ(irregFlux, a_scale, a_fineDatInd,
				     a_variables, a_dir, a_sd, a_dx);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::getFlux_rz(EBFaceFAB&       a_flux,
			    const EBCellFAB& a_phi,
			    const Box&       a_box,
			    const EBISBox&   a_ebisBox,
			    const RealVect&  a_dx,
			    const int&       a_idir,
			    const Real       a_mu,
			    const Real       a_lambda,
			    int              a_energy_eq )
{
  CH_assert( a_flux.nComp() == SpaceDim );
  CH_assert( a_phi.nComp() == SpaceDim );

  Box ubox = a_phi.getRegion();
  int n1, t1, t2;

  Real dxi = 1/a_dx[a_idir];

  if( a_idir == 0 )
  {
     n1 = 0;
     t1 = 1;
     t2 = 2;
  }
  else if( a_idir == 1 )
  {
     n1 = 1;
     t1 = 0;
     t2 = 2;
  }
  else if( a_idir == 2 )
  {
     n1 = 2;
     t1 = 0;
     t2 = 1;
  }
  IntVectSet ivsCell(a_flux.getCellRegion());
  for(FaceIterator faceit(ivsCell, a_ebisBox.getEBGraph(), a_idir,
                          FaceStop::SurroundingNoBoundary);
      faceit.ok(); ++faceit)
    {
      const FaceIndex& face = faceit();
      VolIndex voh = face.getVoF(Side::Hi);
      VolIndex vol = face.getVoF(Side::Lo);
// Compute NS-momentum viscosity fluxes:

// Build stencil, compute tangential derivatives
      Vector<VolIndex> vohpv = a_ebisBox.getVoFs(voh,t1,Side::Hi,1);
      Vector<VolIndex> vohmv = a_ebisBox.getVoFs(voh,t1,Side::Lo,1);
      Vector<VolIndex> volpv = a_ebisBox.getVoFs(vol,t1,Side::Hi,1);
      Vector<VolIndex> volmv = a_ebisBox.getVoFs(vol,t1,Side::Lo,1);

      VolIndex vohp, vohm, volp, volm;

      Real idx=1/(2*a_dx[t1]);
      if( vohpv.size() > 0 )
	 vohp = vohpv[0];
      else
      {
 	 vohp = voh;
         idx=1/(a_dx[t1]);
      }
      if( vohmv.size() > 0 )
	 vohm = vohmv[0];
      else
      {
 	 vohm = voh;
         idx=1/(a_dx[t1]);
      }
      Real dun_dt1=0, dut1_dt1=0;
    
      if( ubox.contains(vohp.gridIndex()) && ubox.contains(vohm.gridIndex()) )
      {
         dun_dt1  = (a_phi(vohp,n1)-a_phi(vohm,n1))*idx;
         dut1_dt1 = (a_phi(vohp,t1)-a_phi(vohm,t1))*idx;
      }

      idx=1/(2*a_dx[t1]);
      if( volpv.size() > 0 )
	 volp = volpv[0];
      else
      {
 	 volp = vol;
         idx=1/(a_dx[t1]);
      }
      if( volmv.size() > 0 )
	 volm = volmv[0];
      else
      {
 	 volm = vol;
         idx=1/(a_dx[t1]);
      }

      if( ubox.contains(volp.gridIndex()) && ubox.contains(volm.gridIndex()) )
      {
         dun_dt1  = 0.5*( dun_dt1  + (a_phi(volp,n1)-a_phi(volm,n1))*idx );
         dut1_dt1 = 0.5*( dut1_dt1 + (a_phi(volp,t1)-a_phi(volm,t1))*idx );
      }


      Real flxr1=0;
      Real flxr2=0;
      Real flxz=0;

      bool volinbox = ubox.contains(vol.gridIndex());
      bool vohinbox = ubox.contains(voh.gridIndex());

      Real dun_dn = 0;
       if( volinbox && vohinbox )
           dun_dn  = (a_phi(voh,n1)-a_phi(vol,n1))*dxi;

       Real rp= (voh.gridIndex()[0]+0.5)*a_dx[0]+m_origin[0];
       Real r = (vol.gridIndex()[0]+0.5)*a_dx[0]+m_origin[0];

       Real drun_dn = 0;
       if( volinbox && vohinbox )
           drun_dn = (rp*a_phi(voh,0)-r*a_phi(vol,0))*dxi;

       Real dut1_dn = 0;
       if( volinbox && vohinbox )
            dut1_dn = (a_phi(voh,t1)-a_phi(vol,t1))*dxi;

       Real rh =(r+rp)*0.5;

       if( a_energy_eq == 1 )
       {
	  Real uavg1=0, uavg2=0;
          if( ubox.contains(voh.gridIndex()) && ubox.contains(vol.gridIndex()))
	  {
             uavg1 = (a_phi(vol,0)+a_phi(voh,0))*0.5;
             uavg2 = (a_phi(vol,1)+a_phi(voh,1))*0.5;
	  }
	  if( a_idir == 0 )
	  {
 	     a_flux(face,0) = (2*rh*a_mu*dun_dn +
			       a_lambda*(drun_dn+rh*dut1_dt1) )*uavg1 +
	                       rh*a_mu*(dun_dt1+dut1_dn)*uavg2;
	  }
	  else
	  {
	     Real drut1_dt1or = uavg1/rh+dut1_dt1;
	     a_flux(face,0) = a_mu*(dun_dt1+dut1_dn)*uavg1+
                    (a_lambda*drut1_dt1or+(2*a_mu+a_lambda)*dun_dn)*uavg2;
	  }
       }
       else
       {
          if( a_idir == 0 )
          {
	     flxr1 = 2*a_mu*rh*dun_dn;
   	     flxr2 = a_lambda*(drun_dn/rh+dut1_dt1);
	     flxz  = a_mu*rh*(dun_dt1+dut1_dn);
             a_flux(face,n1) = flxr1 + flxr2*rh;
             a_flux(face,t1) = flxz;
          }
          else
          {
	     flxr2 = a_mu*(dut1_dn+dun_dt1);
	     flxz  = ((2*a_mu+a_lambda)*dun_dn + a_lambda*dut1_dt1/rh);
             a_flux(face,n1) = flxr2;
             a_flux(face,t1) = flxz;
          }
       }
    }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::getFlux(EBFaceFAB&       a_flux,
			    const EBCellFAB& a_phi,
			    const Box&       a_box,
			    const EBISBox&   a_ebisBox,
			    const RealVect&  a_dx,
			    const int&       a_idir,
			    const Real       a_mu,
			    const Real       a_lambda,
			    int              a_energy_eq )
{
  CH_assert( a_flux.nComp() == SpaceDim );
  CH_assert( a_phi.nComp() == SpaceDim );

  Box ubox = a_phi.getRegion();
  int n1, t1, t2;

  if( a_idir == 0 )
  {
     n1 = 0;
     t1 = 1;
     t2 = 2;
  }
  else if( a_idir == 1 )
  {
     n1 = 1;
     t1 = 0;
     t2 = 2;
  }
  else if( a_idir == 2 )
  {
     n1 = 2;
     t1 = 0;
     t2 = 1;
  }
  IntVectSet ivsCell(a_flux.getCellRegion());
  for(FaceIterator faceit(ivsCell, a_ebisBox.getEBGraph(), a_idir,
                          FaceStop::SurroundingNoBoundary);
      faceit.ok(); ++faceit)
    {
      const FaceIndex& face = faceit();
      VolIndex voh = face.getVoF(Side::Hi);
      VolIndex vol = face.getVoF(Side::Lo);

// Build stencil, compute tangential derivatives
      Vector<VolIndex> vohpv = a_ebisBox.getVoFs(voh,t1,Side::Hi,1);
      Vector<VolIndex> vohmv = a_ebisBox.getVoFs(voh,t1,Side::Lo,1);
      Vector<VolIndex> volpv = a_ebisBox.getVoFs(vol,t1,Side::Hi,1);
      Vector<VolIndex> volmv = a_ebisBox.getVoFs(vol,t1,Side::Lo,1);

      VolIndex vohp, vohm, volp, volm;
      Real idx=1/(2*a_dx[t1]);

      if( vohpv.size() > 0 )
	 vohp = vohpv[0];
      else
      {
 	 vohp = voh;
         idx=1/(a_dx[t1]);
      }
      if( vohmv.size() > 0 )
	 vohm = vohmv[0];
      else
      {
 	 vohm = voh;
         idx=1/(a_dx[t1]);
      }
      Real dun_dt1=0, dut1_dt1=0;
    
      if( ubox.contains(vohp.gridIndex()) && ubox.contains(vohm.gridIndex()) )
      {
         dun_dt1  = (a_phi(vohp,n1)-a_phi(vohm,n1))*idx;
         dut1_dt1 = (a_phi(vohp,t1)-a_phi(vohm,t1))*idx;
      }
      idx=1/(2*a_dx[t1]);
      if( volpv.size() > 0 )
	 volp = volpv[0];
      else
      {
 	 volp = vol;
         idx=1/(a_dx[t1]);
      }
      if( volmv.size() > 0 )
	 volm = volmv[0];
      else
      {
 	 volm = vol;
         idx=1/(a_dx[t1]);
      }

      if( ubox.contains(volp.gridIndex()) && ubox.contains(volm.gridIndex()) )
      {
         dun_dt1  = 0.5*( dun_dt1  + (a_phi(volp,n1)-a_phi(volm,n1))*idx );
         dut1_dt1 = 0.5*( dut1_dt1 + (a_phi(volp,t1)-a_phi(volm,t1))*idx );
      }

      Real dun_dt2 = 0, dut2_dt2 = 0;
      if( SpaceDim == 3 )
      {
         vohpv = a_ebisBox.getVoFs(voh,t2,Side::Hi,1);
         vohmv = a_ebisBox.getVoFs(voh,t2,Side::Lo,1);
	 volpv = a_ebisBox.getVoFs(vol,t2,Side::Hi,1);
	 volmv = a_ebisBox.getVoFs(vol,t2,Side::Lo,1);
         Real idx=1/(2*a_dx[t2]);
         if( vohpv.size() > 0 )
	    vohp = vohpv[0];
         else
         {
 	    vohp = voh;
            idx=1/(a_dx[t2]);
         }
         if( vohmv.size() > 0 )
	    vohm = vohmv[0];
         else
         {
 	    vohm = voh;
            idx=1/(a_dx[t2]);
         }
         dun_dt2  = (a_phi(vohp,n1)-a_phi(vohm,n1))*idx;
         dut2_dt2 = (a_phi(vohp,t2)-a_phi(vohm,t2))*idx;
         idx=1/(2*a_dx[t2]);
         if( volpv.size() > 0 )
   	    volp = volpv[0];
         else
         {
 	    volp = vol;
            idx=1/(a_dx[t2]);
         }
         if( volmv.size() > 0 )
   	    volm = volmv[0];
         else
         {
 	    volm = vol;
            idx=1/(a_dx[t2]);
         }
         dun_dt2  = 0.5*( dun_dt2  + (a_phi(volp,n1)-a_phi(volm,n1))*idx );
         dut2_dt2 = 0.5*( dut2_dt2 + (a_phi(volp,t2)-a_phi(volm,t2))*idx );
      }
      Real dun_dn=0,dut1_dn=0;
      if( ubox.contains(voh.gridIndex()) && ubox.contains(vol.gridIndex()))
      {
         dun_dn  = (a_phi(voh,n1)-a_phi(vol,n1))/a_dx[a_idir];
         dut1_dn = (a_phi(voh,t1)-a_phi(vol,t1))/a_dx[a_idir];
      }

      //      Real la = (a_lambda(vol,0)+a_lambda(voh,0))*half*dxi;
      //      Real mu = (a_mu(vol,0)    +a_mu(voh,0))*half*dxi;
      Real la = a_lambda;
      Real mu = a_mu;

      a_flux(face,n1) = (2*mu+la)*dun_dn + la*dut1_dt1;
      a_flux(face,t1) = mu*( dun_dt1 + dut1_dn );
      if( SpaceDim == 3 )
      {
	 Real dut2_dn = (a_phi(voh,t2)-a_phi(vol,t2))/a_dx[a_idir];
	 a_flux(face,t2) =  mu*(dun_dt2+dut2_dn);
	 a_flux(face,n1) += la*dut2_dt2;
      }
      if( a_energy_eq == 1 )
      {
         Real flx = 0;
         if( ubox.contains(voh.gridIndex()) && ubox.contains(vol.gridIndex()))
	    for( int d = 0 ; d < SpaceDim ; d++ )
 	       flx += a_flux(face,d)*(a_phi(vol,d)+a_phi(voh,d))*0.5;
         a_flux(face,0) = flx;
      }
    }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::AMRResidualNC(LevelData<EBCellFAB>&       a_residual,
				  const LevelData<EBCellFAB>& a_phiFine,
				  const LevelData<EBCellFAB>& a_phi,
				  const LevelData<EBCellFAB>& a_rhs,
				  bool a_homogeneousPhysBC )
{
  //dummy. there is no coarse when this is called
  //  LevelData<EBCellFAB> phiC;
  //  AMRResidual(a_residual, a_phiFine, a_phi, phiC, a_rhs, a_homogeneousPhysBC);

  CH_assert(a_residual.nComp() == SpaceDim );
  CH_assert(a_phi.nComp() == SpaceDim );
  CH_assert(a_rhs.nComp() == SpaceDim );

  setToZero( a_residual );
  applyOp( a_residual, a_phi, NULL,  a_homogeneousPhysBC, false );
  if( m_hasFine )
     reflux(a_residual, a_phiFine, a_phi, 0 );

  //  a_residual  :=  rhs - L(phi)
   scale(a_residual,-1.0);
   incr(a_residual,a_rhs,1.0);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::AMRResidualNF(LevelData<EBCellFAB>&       a_residual,
				  const LevelData<EBCellFAB>& a_phi,
				  const LevelData<EBCellFAB>& a_phiCoar,
				  const LevelData<EBCellFAB>& a_rhs,
				  bool a_homogeneousPhysBC)
{
  applyOp( a_residual, a_phi, &a_phiCoar,  a_homogeneousPhysBC, false);
  scale(a_residual,-1.0);
  incr(a_residual,a_rhs,1.0);
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::AMRRestrict(LevelData<EBCellFAB>&       a_resCoar,
				const LevelData<EBCellFAB>& a_residual,
				const LevelData<EBCellFAB>& a_correction,
				const LevelData<EBCellFAB>& a_coarCorrection)
{
   CH_assert(a_residual.nComp() == SpaceDim );
   CH_assert(a_resCoar.nComp() == SpaceDim );
   CH_assert(a_correction.nComp() == SpaceDim );

  int ncomp = a_residual.nComp();

  LevelData<EBCellFAB> resThisLevel;
  bool homogeneousPhys = true;
  bool homogeneousCF =   false;

  EBCellFactory ebcellfactTL(m_ebisl);
  IntVect ghostVec = a_residual.ghostVect();

  resThisLevel.define(m_dbl, ncomp, ghostVec, ebcellfactTL);

  //API says that we must average(a_residual - L(correction, coarCorrection))
  applyOp(resThisLevel, a_correction, &a_coarCorrection, homogeneousPhys, homogeneousCF);
  incr(resThisLevel, a_residual, -1.0);
  scale(resThisLevel,-1.0);

  //use averaging operator
  Interval variables(0, ncomp-1);
  CH_assert(m_hasInterpAve);
  m_ebAverage.average(a_resCoar, resThisLevel, variables);

}


//-----------------------------------------------------------------------
///compute norm over all cells on coarse not covered by finer
Real EBAMRNSviscOp::AMRNorm(const LevelData<EBCellFAB>& a_coarResid,
			    const LevelData<EBCellFAB>& a_fineResid,
			    const int& a_refRat,
			    const int& a_ord)

{
  int ncomp = a_coarResid.nComp();
  Interval interv(0, ncomp-1);
  const DisjointBoxLayout& coarGrids = a_coarResid.disjointBoxLayout();
  const DisjointBoxLayout& fineGrids = a_fineResid.disjointBoxLayout();
  CH_assert(coarGrids == m_dbl);

  //create temp and zero out under finer grids
  EBCellFactory ebcellfact(m_ebisl);
  LevelData<EBCellFAB> coarTemp(coarGrids, ncomp, IntVect::Zero, ebcellfact);
  a_coarResid.copyTo(interv, coarTemp, interv);

  for(DataIterator dit = coarGrids.dataIterator(); dit.ok(); ++dit)
    {
      EBCellFAB& coarTempFAB = coarTemp[dit()];
      const EBISBox& ebisBox = m_ebisl[dit()];

      LayoutIterator litFine = fineGrids.layoutIterator();
      for (litFine.reset(); litFine.ok(); ++litFine)
        {
          Box overlayBox = coarTempFAB.box();
          Box coarsenedGrid = coarsen(fineGrids[litFine()], a_refRat);

          overlayBox &= coarsenedGrid;
          if (!overlayBox.isEmpty())
            {
              IntVectSet ivsZero(overlayBox);

              for(VoFIterator vofit(ivsZero, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
                {
                  for(int ivar =0; ivar < ncomp; ivar++)
                    {
                      coarTempFAB(vofit(), ivar) = 0.0;
                    }
                }
            }
        }
    }
  //return norm of temp
  return norm(coarTemp, a_ord);
}
///////////
void EBAMRNSviscOp::
AMRProlong(LevelData<EBCellFAB>&       a_correction,
           const LevelData<EBCellFAB>& a_coarCorrection)
{
  //use cached interpolation object
  Interval variables(0, a_coarCorrection.nComp()-1);
  CH_assert(m_hasInterpAve);
  m_ebInterp.pwcInterp(a_correction, a_coarCorrection, variables);
}

///////////
void EBAMRNSviscOp::
AMRUpdateResidual(LevelData<EBCellFAB>&       a_residual,
                  const LevelData<EBCellFAB>& a_correction,
                  const LevelData<EBCellFAB>& a_coarCorrection)
{
  LevelData<EBCellFAB> lcorr;
  bool homogeneousPhys = true;
  bool homogeneousCF   = false;

  EBCellFactory ebcellfactTL(m_ebisl);
  IntVect ghostVec = a_residual.ghostVect();

  lcorr.define(m_dbl, a_coarCorrection.nComp(), ghostVec, ebcellfactTL);

  applyOp(lcorr, a_correction, &a_coarCorrection, homogeneousPhys, homogeneousCF);

  incr(a_residual, lcorr, -1);
}

void EBAMRNSviscOp::
levelMulticolorGS(LevelData<EBCellFAB>&       a_phi,
                  const LevelData<EBCellFAB>& a_rhs )
{
  int ncomp = a_rhs.nComp();

  LevelData<EBCellFAB> residLD;
  //  bool homogeneous = true;
  //  bool homogeneous = false;

  EBCellFactory ebcellfact(m_ebisl);
  residLD.define( m_dbl, a_rhs.nComp(),a_rhs.ghostVect(), ebcellfact );

  IntVect color = IntVect::Zero;
  IntVect limit = IntVect::Unit;

  // Loop over all possibilities (in all dimensions)
  while (color[SpaceDim-1] <= limit[SpaceDim-1])
  {
      // Get the residual everywhere
      residual(residLD,a_phi,a_rhs,m_homogeneous_relax);

      //      cout << "GS it " << norm( residLD, 0 ) << endl;
      int ibox = 0;
      for (DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
        {
          DataIndex dex=dit();

          EBCellFAB& phi = a_phi[dit()];

          const EBISBox& ebisBox = m_ebisl[dit()];
          const Box& box   = m_dbl.get(dit());

          EBCellFAB phiSave(phi.getEBISBox(),
                            box,
                            ncomp);

          phiSave.copy(phi);

          const EBCellFAB& resid = residLD[dit()];

	  //          EBCellFAB wgh(phi.getEBISBox(), box, ncomp);

	  //          Real cofi = 1/m_c;
	  //          const BaseFab<Real>& la_singv = m_lambda[dex].getSingleValuedFAB();
	  //          const BaseFab<Real>& mu_singv = m_mu[dex].getSingleValuedFAB();
	  //          const BaseFab<Real>& rho_singv = m_rho[dex].getSingleValuedFAB();
	  //          BaseFab<Real>& wgh_singv = wgh.getSingleValuedFAB();
	  
	  //          if( !m_rz )
	  //             FORT_GETWGH( CHF_CONST_FRA1(la_singv,0), CHF_CONST_FRA1(mu_singv,0),
	  //		       CHF_CONST_FRA1(rho_singv,0), CHF_FRA(wgh_singv),
	  //		       CHF_CONST_REAL(cofi), CHF_CONST_REALVECT(m_dx),
	  //		       CHF_BOX(m_dbl.get(dex)) );
	  //          else
	    //             FORT_GETWGH_RZ( CHF_CONST_FRA1(la_singv,0), CHF_CONST_FRA1(mu_singv,0),
	    //		       CHF_CONST_FRA1(rho_singv,0), CHF_FRA(wgh_singv),
	    //		       CHF_CONST_REAL(cofi), CHF_CONST_REALVECT(m_dx),
	    //		       CHF_BOX(m_dbl.get(dex)), CHF_CONST_REALVECT(m_origin) );


	  //          for( VoFIterator v(m_ebisl[dex].getIrregIVS(m_ebisl[dex].getRegion()),m_ebisl[dex].getEBGraph()); v.ok(); ++v )
	  //	  {
	  //	     VolIndex vof = v();
	  //	     for( int comp = 0 ; comp < ncomp ; comp++ )
	  //	     {
	  //		Real weight = m_weight[dex](vof,comp);
	  //                if( m_split_u  )
	  //                    weight -= cofi*m_rho[dex](vof,0)*m_uweight[dex](vof,0)*m_ebisl[dex].volFrac(vof);
	  //	        else
	  //		    weight -= cofi*m_rho[dex](vof,0)*m_ebisl[dex].volFrac(vof);
	  //                wgh(vof,comp) = 0.5/weight;
	  //	     }
	  //	  }


          // Do the regular cells
          BaseFab<Real>& phiFAB         = phi.getSingleValuedFAB();
          BaseFab<Real>& phiSaveFAB     = phiSave.getSingleValuedFAB();
          const BaseFab<Real>& residFAB = resid.getSingleValuedFAB();
          const BaseFab<Real>& wghFAB   = m_wgh[dex].getSingleValuedFAB();

          IntVect loIV = box.smallEnd();
          IntVect hiIV = box.bigEnd();

          for (int idir = 0; idir < SpaceDim; idir++)
            {
              if (loIV[idir] % 2 != color[idir])
                {
                  loIV[idir]++;
                }
            }

          if (loIV <= hiIV)
            {
              Box coloredBox(loIV,hiIV);

              FORT_MULTICOLORGSNSV(CHF_FRA(phiFAB),
                                   CHF_CONST_FRA(phiSaveFAB),
                                   CHF_CONST_FRA(wghFAB),
                                   CHF_CONST_FRA(residFAB),
                                   CHF_BOX(coloredBox) );
            }

          // Do the irregular cells
          const BaseIVFAB<Real>& curWeightBaseIVFAB = m_weight[dit()];
          const IntVectSet& curIrreg = curWeightBaseIVFAB.getIVS();
          for (VoFIterator vofit(curIrreg,ebisBox.getEBGraph()); vofit.ok(); ++vofit)
          {
              const VolIndex& VoF = vofit();
              const IntVect& iv = VoF.gridIndex();

              bool ok = true;
              for (int idir = 0; idir < SpaceDim; idir++)
              {
                  if (iv[idir] % 2 != color[idir])
                    {
                      ok = false;
                      break;
                    }
              }
              if (ok)
              {
		  //                  const Real& weight = curWeightBaseIVFAB(VoF,comp);
                for( int comp=0 ; comp < ncomp ; comp++ )
                   phi(VoF,comp) = phiSave(VoF,comp)
                                + m_wgh[dex](VoF,comp) * resid(VoF,comp);
              }
          }
          ibox++;
        }

      color[0]++;

      // If greater that the maximum it can be...
      if (color[0] > limit[0])
        {
          // For all dimensions...
          for (int i = 0; i < SpaceDim-1; i++)
            {
              color[i] = 0;
              color[i+1]++;

              // If the next dimension's power is okay go on
              if (color[i+1] <= limit[i+1])
                {
                  break;
                }
            }
        }
    }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::levelJacobi(LevelData<EBCellFAB>&       a_phi,
				const LevelData<EBCellFAB>& a_rhs )
{
  LevelData<EBCellFAB> resid;
  EBCellFactory ebcellfact(m_ebisl);
  resid.define( m_dbl, a_rhs.nComp(), IntVect::Zero, ebcellfact );
  LevelData<EBCellFAB> tmp(m_dbl,a_phi.nComp(), IntVect::Zero, ebcellfact );
  for( int nit = 0 ; nit < 5 ; nit++ )
  {
     residual(resid,a_phi,a_rhs,m_homogeneous_relax);
     //     cout << "Jacobi it " << norm( resid, 0 ) << endl;
     //     DataIterator d=m_dbl.dataIterator();
     //     DataIndex dex = d();
     //     Vector<VolIndex> vovvar=m_ebisl[dex].getVoFs(IntVect(2,2));
     //     cout <<"BEfore " << endl;
     //     for( int i=0 ; i < vovvar.size() ; i++ )
     //     {
     //       cout << "res " << resid[dex](vovvar[i],0) << endl;
     //       cout << "wgh " << m_wgh[dex](vovvar[i],0) << endl;
     //       cout << "phi " << a_phi[dex](vovvar[i],0) << endl;
     //     }
     EBLevelDataOps::product(tmp,m_wgh,resid);
     EBLevelDataOps::sum(a_phi,a_phi,tmp);
     //     cout <<"After " << endl;
     //     for( int i=0 ; i < vovvar.size() ; i++ )
     //     {
     //       cout << "res " << resid[dex](vovvar[i],0) << endl;
     //       cout << "wgh " << m_wgh[dex](vovvar[i],0) << endl;
     //       cout << "phi " << a_phi[dex](vovvar[i],0) << endl;
     //     }
  }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::set_homogeneous_relax()
{
  m_homogeneous_relax = true;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::set_inhomogeneous_relax()
{
  m_homogeneous_relax = false;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::define_flux_stencils( )
{
  m_iflux_stencil.define( m_dbl );
  m_jflux_stencil.define( m_dbl );
  m_kflux_stencil.define( m_dbl );
  m_uyz_stencil.define( m_dbl );
  m_uxz_stencil.define( m_dbl );
  m_uxy_stencil.define( m_dbl );

  for (DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
  {
     DataIndex dex(dit());
     Box fluxbox = m_mu[dex].box(); //m_dbl.get(dex);
     for( int dir=0; dir < SpaceDim; dir++ )
     {
        if( dir == 0 )
	   define_flux_stencil( dir, m_iflux_stencil[dex], m_uyz_stencil[dex],
			        m_ebisl[dex], fluxbox, m_irreg[dex] );
        else if( dir == 1 )
           define_flux_stencil( dir, m_jflux_stencil[dex], m_uxz_stencil[dex],
				m_ebisl[dex], fluxbox, m_irreg[dex] );
        else
           define_flux_stencil( dir, m_kflux_stencil[dex], m_uxy_stencil[dex],
				m_ebisl[dex], fluxbox, m_irreg[dex] );
     }
  }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::define_flux_stencil( int dir, BaseIFFAB<FaceStencil>& flux_stencil,
					 BaseIVFAB<VoFStencil>& utan_stencil,
					 const EBISBox& eb,
					 const Box& ubox,
					 const IntVectSet& notRegular )
{
     const EBGraph& curEBGraph = eb.getEBGraph();
     //     IntVectSet notRegular     = eb.getIrregIVS(ubox);
     //     IntVectSet& notRegular = m_irreg[dex];

     flux_stencil.define( notRegular, curEBGraph, dir, 1 );
     IntVectSet utan;

// Set up stencils for the flux tangential interpolation
     for( FaceIterator it(notRegular,curEBGraph,dir,FaceStop::SurroundingNoBoundary); it.ok(); ++it)
     {
        FaceIndex fi=it();

        flux_stencil(fi,0).clear();

        RealVect center = eb.centroid(fi);
        RealVect normal = eb.normal(fi.getVoF(Side::Lo));
        if( eb.volFrac(fi.getVoF(Side::Hi)) < eb.volFrac(fi.getVoF(Side::Lo)) )
	   normal = eb.normal(fi.getVoF(Side::Hi));
	
        RealVect wgh= RealVect::Zero;
        IntVect  of = IntVect::Zero;
        for( int d=0 ; d< SpaceDim ; d++ )
           if( d != dir )
  	   {
	      of[d]  =  normal[d] > 0 ? 1:-1;
              wgh[d] = fabs(center[d]);
	   }

        for( int d1=0 ; d1 < 2 ; d1++ )
           for( int d2=0 ; d2 < 2 ; d2++ )
              for( int d3=0 ; d3 < 2 ; d3++ )
       	      {
	         IntVect of2=IntVect::Zero;
  		 of2[0] = d1==0 ? 0 :of[0];
		 Real w1 = d1==0 ? 1-wgh[0]:wgh[0];
		 of2[1] = d2==0 ? 0 :of[1];
		 Real w2 = d2==0 ? 1-wgh[1]:wgh[1];
                 Real w3;
                 if( SpaceDim == 3 )
		 {
		    of2[2] = d3==0 ? 0 :of[2];
   		    w3 = d3==0 ? 1-wgh[2]:wgh[2];
		 }
		 else
		 {
		   w3 = 1-d3;
		 }
                 if( w1*w2*w3 != 0 )
		 {
                   Vector<VolIndex> vl, vu;
		   if( eb.getRegion().contains(fi.gridIndex(Side::Lo)+of2) &&
                       eb.getRegion().contains(fi.gridIndex(Side::Hi)+of2) )
		   {
		      vl=eb.getVoFs(fi.gridIndex(Side::Lo)+of2);
		      vu=eb.getVoFs(fi.gridIndex(Side::Hi)+of2);
                   }
		   else
		   {
		      vl=eb.getVoFs(fi.gridIndex(Side::Lo));
		      vu=eb.getVoFs(fi.gridIndex(Side::Hi));
                   }
                       
		       //                       Vector<VolIndex> vl=m_ebisl[dex].getVoFs(fi.getVoF(Side::Lo),dir,Side::Lo,1);
		       //		       Vector<VolIndex> vu=m_ebisl[dex].getVoFs(fi.getVoF(Side::Hi),dir,Side::Hi,1);
		   if( vl.size() > 0 && vu.size() > 0 )
		   {
 		      FaceIndex fif(vl[0],vu[0],dir);
 		      flux_stencil(fi,0).add(fif,w1*w2*w3);
		      utan |= fif.getVoF(Side::Hi).gridIndex();
		      utan |= fif.getVoF(Side::Lo).gridIndex();
		   }
		 }
	      }
        flux_stencil(fi,0) *= eb.areaFrac(fi);
     }

// Set up stencils for the tangential velocity derivatives
     utan_stencil.define( utan, curEBGraph, SpaceDim-1 );

        
     IntVect t1=IntVect::Zero;
     IntVect t2=IntVect::Zero;
     Real dyi, dzi;
	// note: In 2D, t2 and dzi are never used. Values at index SpaceDim-1
	// just dummy to prevent index out of range.
     if( dir == 0 )
     {
	t1[1] = 1;
        dyi = 1/m_dx[1];
        t2[SpaceDim-1] = 1;
        dzi = 1/m_dx[SpaceDim-1];
     }
     else if( dir == 1 )
     {
	t1[0] = 1;
	dyi = 1/m_dx[0];
        t2[SpaceDim-1] = 1;
        dzi = 1/m_dx[SpaceDim-1];
     }
     else
     {
        t1[0] = 1;
	dyi = 1/m_dx[0];
	t2[SpaceDim-2] = 1;
        dzi = 1/m_dx[SpaceDim-2];
     }
     for( VoFIterator vi(utan,curEBGraph); vi.ok(); ++vi )
     {
        const VolIndex& VoF = vi();
//  first tangential derivative:
	tangential_derivative( t1, VoF, eb, utan_stencil(vi(),0), dyi, ubox );
//  second tangential derivative:
        if( SpaceDim == 3 )
   	   tangential_derivative( t2, VoF, eb, utan_stencil(vi(),1), dzi, ubox );


     }
}


//-----------------------------------------------------------------------
//void EBAMRNSviscOp::tangential_derivative( IntVect of,
//					   const VolIndex& vof,
//					   const EBISBox& eb,
//					   VoFStencil& vs,
//					   const Real di,
//					   const Box ubox )
//{
//  // try six cases.
//
//  bool found = true;
//  vs.clear();
//  IntVect v=vof.gridIndex();
//  
//  Vector<VolIndex> allMonotoneVoFs;
//  int radius = 1;
//  IntVect timesMoved = IntVect::Zero;
//  IntVect pathSign   = IntVect::Zero;
//  EBArith::getAllVoFsInMonotonePath(allMonotoneVoFs,timesMoved,pathSign,
//                                    vof,eb,radius);

//  VolIndex vmm, vm, vp, vpp;  
//  if( EBArith::isVoFHere(vp,allMonotoneVoFs,v+of) && 
//      EBArith::isVoFHere(vm,allMonotoneVoFs,v-of) && 
//      ubox.contains(vp.gridIndex()) &&
//      ubox.contains(vm.gridIndex())   )
//    //  if( !eb.isCovered(v+of) && !eb.isCovered(v-of) )
//  {
//    vs.add(vp,0.5);
//    vs.add(vm,-0.5);
//  }
//  else if( EBArith::isVoFHere(vpp,allMonotoneVoFs,v+2*of) && 
//           EBArith::isVoFHere(vp,allMonotoneVoFs,v+of) &&
//      ubox.contains(vp.gridIndex()) &&
//      ubox.contains(vpp.gridIndex())   )
//  //  else if( !eb.isCovered(v+of) && !eb.isCovered(v+2*of) )
//  {
//    vs.add(vof,-1.5);
//    vs.add(vp,2.0);
//    vs.add(vpp,-0.5);
//  }
//  else if( EBArith::isVoFHere(vmm,allMonotoneVoFs,v-2*of) && 
//           EBArith::isVoFHere(vm,allMonotoneVoFs,v-of) &&
//      ubox.contains(vm.gridIndex()) &&
//      ubox.contains(vmm.gridIndex())   )
//  //  else if( !eb.isCovered(v-of) && !eb.isCovered(v-2*of) )
//  {
//    vs.add(vmm,0.5);
//    vs.add(vm,-2.0);
//    vs.add(vof,1.5);
//  }
//  else if( EBArith::isVoFHere(vp,allMonotoneVoFs,v+of) &&
//	   ubox.contains(vp.gridIndex())  )
//
//  //  else if( !isCovered(v+of) )
//  {
//    vs.add(vp,1.0);
//    vs.add(vof,-1.0);
//  }
//  else if( EBArith::isVoFHere(vm,allMonotoneVoFs,v-of) &&
//	   ubox.contains(vm.gridIndex())  )
//    //  else if( !isCovered(v-of) )
//  {
//    vs.add(vm,-1.0);
//    vs.add(vof,1.0);
//  }
//  else
//  {
//    //     MayDay::Warning("EBAMRNSviscOp::tangential_derivative Could not find tangential velocity stencil");
//     //     vs.add(vof,0.0);
//     found =false;
//     //     cout << "error at " << vof.gridIndex()  << endl;
//     //     cout << "neighbors " << vmm.gridIndex() << " " << vm.gridIndex() << " " << vp.gridIndex() << " " << vpp.gridIndex() << endl;
//     //     cout << "box " << ubox << endl;
//     //     MayDay::Error("exiting..");
//  }
//  if( found )
//     vs *= di;
//}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::tangential_derivative( IntVect of,
					   const VolIndex& vof,
					   const EBISBox& eb,
					   VoFStencil& vs,
					   const Real di,
					   const Box ubox )
{
  bool found=true;
  int dir = 0;
  if( of[1] != 0 )
    dir = 1;

  if( SpaceDim > 2 )
     if( of[2] !=0 )
        dir = 2;

  Vector<VolIndex> vofh = eb.getVoFs( vof, dir, Side::Hi, 1 );
  Vector<VolIndex> vofl = eb.getVoFs( vof, dir, Side::Lo, 1 );

  //  cout << " ----------------- one tangential derivative ---------------------" << endl;
  //  cout << dir << " " << vof.gridIndex() << endl;
  //  cout << "vofhsize " << vofh.size() << endl;
  //  cout << "volhsize " << vofl.size() << endl;
  //  if( vofh.size() == 0 && vofl.size() == 0 )
  //  {
    //    cout << "vof at " << vof.gridIndex() << endl;
    //    cout << "ubox = " <<  ubox << endl;
    //    BoxIterator b(ubox);
    //    for( b.begin() ;b.ok() ;++b)
    //    {
    //      cout << b() << "c= " << eb.isCovered(b()) << "i= " << eb.isIrregular(b()) << endl;
    //    }
  //  }
  
  if( vofh.size() > 0 && vofl.size() > 0 && (ubox.contains(vofh[0].gridIndex())&&ubox.contains(vofl[0].gridIndex())))
  {
     vs.add(vofh[0],0.5);
     vs.add(vofl[0],-0.5);
  }
  else if( vofh.size() > 0 && ubox.contains(vofh[0].gridIndex()))
  {
     Vector<VolIndex> vofh2 = eb.getVoFs( vof, dir, Side::Hi, 2 );
     //     cout << "vofh2 " << vofh2.size() << endl;
     if( vofh2.size() > 0 && ubox.contains(vofh2[0].gridIndex()))
     {
       vs.add(vofh2[0],-0.5);
       vs.add(vofh[0],2.0);
       vs.add(vof,-1.5);
     }
     else
     {
       vs.add(vofh[0],1.0);
       vs.add(vof,-1.0);
     }
  }
  else if( vofl.size() > 0 && ubox.contains(vofl[0].gridIndex()))
  {
     Vector<VolIndex> vofl2 = eb.getVoFs( vof, dir, Side::Lo, 2 );
     //     cout << "vofl2 " << vofl2.size() << endl;
     if( vofl2.size() > 0 && ubox.contains(vofl2[0].gridIndex()))
     {
       vs.add(vofl2[0],0.5);
       vs.add(vofl[0],-2.0);
       vs.add(vof,1.5);
     }
     else
     {
       vs.add(vofl[0],-1.0);
       vs.add(vof,1.0);
     }
  }
  else
  {
    //    cout << "utan_derivative not found " << endl;
    found = false;
    vs.clear();
  }
  if( found )
     vs *= di;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::define_ustencil( )
{
   m_uStencil.define( m_dbl );
   m_uweight.define( m_dbl );
   for (DataIterator dit = m_dbl.dataIterator(); dit.ok(); ++dit)
   {
     //      const Box& curBox = m_dbl.get(dit());
     //      const Box& curBox = m_mu[dit()].box();
      const EBISBox& curEBISBox = m_ebisl[dit()];
      const EBGraph& curEBGraph = curEBISBox.getEBGraph();
      //      IntVectSet notRegular = curEBISBox.getIrregIVS(curBox);
      IntVectSet &notRegular = m_irreg[dit()];
      m_uStencil[dit()].define(notRegular,curEBGraph,1);
      m_uweight[dit()].define(notRegular,curEBGraph,1);
      for (VoFIterator vofit(notRegular,curEBGraph); vofit.ok(); ++vofit)
      {
         const VolIndex& VoF = vofit();
         get_ustencil( m_uStencil[dit()](VoF,0), m_uweight[dit()](VoF,0), curEBISBox, VoF );
      }
   }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::get_ustencil(
		VoFStencil&     a_stencil,
                Real&           a_wgh,
                const EBISBox&  a_ebisbox,
                const VolIndex& a_VoF )
{
  a_stencil.clear();
  Vector<VolIndex> allMonotoneVoFs;
  int radius = 1;
  IntVect timesMoved = IntVect::Zero;
  IntVect pathSign   = IntVect::Zero;

  EBArith::getAllVoFsInMonotonePath( allMonotoneVoFs,timesMoved,pathSign,
                                    a_VoF,a_ebisbox,radius);
  a_wgh  = 1;
  int no = 0;

  RealVect dists = a_ebisbox.centroid(a_VoF);
  for( int dir=0 ; dir < SpaceDim ; dir++ )   
  {
     Vector<VolIndex> vofl = a_ebisbox.getVoFs( a_VoF, dir, Side::Lo,1);
     Vector<VolIndex> vofh = a_ebisbox.getVoFs( a_VoF, dir, Side::Hi,1);
     if( vofl.size() > 0 && dists[dir] < 0 )
     {
	a_stencil.add(vofl[0],-dists[dir]);
	a_wgh += dists[dir];
        no++;
     }
     else if( vofh.size() > 0 && dists[dir] >= 0 )
     {
	a_stencil.add(vofh[0],dists[dir]);
	a_wgh -= dists[dir];
        no++;
     }  
  }
  //  if( no != SpaceDim )
  //  {
  //     MayDay::Warning("EBAMRNSviscOp::get_ustencil. Could not find u-stencil");
  //     cout << "no =    " << no << endl;
  //     cout << "a_wgh = " << a_wgh << endl;
  //  }
  a_stencil.add(a_VoF,a_wgh);
  //  cout << "-------------------------------------" << endl;
  //  for( int i=0 ; i<a_stencil.size() ; i++ )
  //  {
  //    cout << a_VoF.gridIndex() << " i-elem = " << a_stencil.weight(i) << " at " << a_stencil.vof(i).gridIndex() << endl;
  //  }
  //  cout << "-------------------------------------" << endl;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::update_irregular( EBCellFAB& a_res,
				      const EBCellFAB& a_u,
				      const EBISBox& eb,
				      const BaseIFFAB<FaceStencil>& flux_stencil,
				      const BaseIVFAB<VoFStencil>& utan_stencil,
	                              const EBCellFAB& a_mu,
				      const EBCellFAB& a_lambda,
				      int dir,
				      const IntVectSet& notRegular,
				      const IntVectSet& irreg_res )
{
   const Real half=0.5;
  //   for( DataIterator li=m_dbl.dataIterator() ; li.ok() ; ++li )
  //   {
     //      DataIndex dex(li());

     //      EBCellFAB& res = a_res[dex];
     //      EBCellFAB& u   = a_u[dex];

     //      VoFStencil&  utan_stencil;
     //      FaceStencil& flux_stencil;

      int n1, t1, t2;
      //      for( int dir=0 ; dir < SpaceDim ; dir++ )
      //      {
         if( dir == 0 )
         {
	   //            utan_stencil = m_uyz_stencil[dex];
	   //            flux_stencil = m_iflux_stencil[dex];
            n1 = 0;
            t1 = 1;
            t2 = 2;
         }
         else if( dir == 1 )
         {
	   //            utan_stencil = m_uxz_stencil[dex];
	   //            flux_stencil = m_jflux_stencil[dex];
            n1 = 1;
            t1 = 0;
            t2 = 2;
         }
         else
         {
	   //            utan_stencil = m_uxy_stencil[dex];
	   //            flux_stencil = m_kflux_stencil[dex];
            t1 = 0;
            t2 = 1;
            n1 = 2;
         }

         Real dxi = 1/m_dx[dir];

         const EBGraph& curEBGraph = eb.getEBGraph();
	 //         IntVectSet notRegular     = eb.getIrregIVS(a_u.box());
	 //         IntVectSet irreg_res      = eb.getIrregIVS(a_res.box());

	 //         IntVectSet& notRegular = m_irreg_w_ghost[dex];
	 //         IntVectSet& irreg_res  = m_irreg[dex];

         for(FaceIterator it(notRegular,curEBGraph,dir,FaceStop::SurroundingNoBoundary); it.ok();++it)
         {
            FaceIndex fi=it();
// Compute NS-momentum viscosity fluxes:
            RealVect flx = RealVect::Zero;

            for( int ii=0 ; ii < flux_stencil(fi,0).size() ; ii++ )
            {
               FaceIndex fic = flux_stencil(fi,0).face(ii);

               VolIndex voh = fic.getVoF(Side::Hi);
               VolIndex vol = fic.getVoF(Side::Lo);

               Real dut1_dt1 = 0;
               Real dun_dt1  = 0;
               const VoFStencil& vs=utan_stencil(voh,0);
               for( int i=0 ; i<vs.size() ; i++ )
       	       {
       	           dut1_dt1 += a_u(vs.vof(i),t1)*vs.weight(i);
	           dun_dt1  += a_u(vs.vof(i),n1)*vs.weight(i);
               }
               const VoFStencil& vsl = utan_stencil(vol,0);
	       for( int i=0 ; i<vsl.size() ; i++ )
	       {
	          dut1_dt1 += a_u(vsl.vof(i),t1)*vsl.weight(i);
	          dun_dt1  += a_u(vsl.vof(i),n1)*vsl.weight(i);
	       }
	       dut1_dt1 *= 0.5;
	       dun_dt1  *= 0.5;

               Real dut2_dt2 = 0;
               Real dun_dt2  = 0;
               if( SpaceDim == 3 )
	       {
	           const VoFStencil& vsh=utan_stencil(voh,1);
                   for( int i=0 ; i<vsh.size() ; i++ )
	           {
	              dut2_dt2 += a_u(vsh.vof(i),t1)*vsh.weight(i);
 	              dun_dt2  += a_u(vsh.vof(i),n1)*vsh.weight(i);
                   }
                   const VoFStencil& vsl = utan_stencil(vol,1);
	           for( int i=0 ; i<vsl.size() ; i++ )
	           {
	              dut2_dt2 += a_u(vsl.vof(i),t1)*vsl.weight(i);
	              dun_dt2  += a_u(vsl.vof(i),n1)*vsl.weight(i);
	           }
	           dut2_dt2 *= 0.5;
	           dun_dt2  *= 0.5;
       	       }

               Real la, mu, dun_dn=0, dut1_dn=0;
	       
               bool inl =a_u.box().contains(vol.gridIndex());
               bool inh =a_u.box().contains(voh.gridIndex());
	       if( inl && inh )
	       {
	          la = (a_lambda(vol,0)+a_lambda(voh,0))*half*dxi;
	          mu = (a_mu(vol,0)    +a_mu(voh,0))*half*dxi;
                  dun_dn  = (a_u(voh,n1)-a_u(vol,n1))*dxi;
                  dut1_dn = (a_u(voh,t1)-a_u(vol,t1))*dxi;
	       }
	       else if( inl )
	       {
	          la = a_lambda(vol,0)*dxi;
	          mu = a_mu(vol,0)*dxi;
	       }
	       else
	       {
	          la = a_lambda(voh,0)*dxi;
	          mu = a_mu(voh,0)*dxi;
	       }
               RealVect flxc;

               flxc[n1] =  (2*mu+la)*dun_dn+la*(dut1_dt1);
	       flxc[t1] =  mu*( dun_dt1 + dut1_dn );
               if( SpaceDim == 3 )
	       {
	          Real dut2_dn = (a_u(voh,t2)-a_u(vol,t2))*dxi;
   	          flxc[t2] = mu*(dun_dt2+dut2_dn);
                  flxc[n1] += la*dut2_dt2;
	       }
               for( int d = 0 ; d < SpaceDim ; d++ )
                  flx[d] += flxc[d]*flux_stencil(fi,0).weight(ii);
            }
            if( irreg_res.contains(fi.getVoF(Side::Hi).gridIndex()) )
            {
               for( int d = 0 ; d < SpaceDim ; d++ )
 	          a_res(fi.getVoF(Side::Hi),d) -= flx[d];
            }
	    if( irreg_res.contains(fi.getVoF(Side::Lo).gridIndex()) )
	    { 
               for( int d = 0 ; d < SpaceDim ; d++ )
 	          a_res(fi.getVoF(Side::Lo),d) += flx[d];
	    }
	 }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::update_irregular_rz( EBCellFAB& a_res,
					 const EBCellFAB& a_u,
					 const EBISBox& eb,
					 const BaseIFFAB<FaceStencil>& flux_stencil,
					 const BaseIVFAB<VoFStencil>& utan_stencil,
					 const EBCellFAB& a_mu,
					 const EBCellFAB& a_lambda,
					 int dir,
					 const IntVectSet& notRegular,
					 const IntVectSet& irreg_res )
{
   const Real half=0.5;
  //   for( DataIterator li=m_dbl.dataIterator() ; li.ok() ; ++li )
  //   {
     //      DataIndex dex(li());

     //      EBCellFAB& res = a_res[dex];
     //      EBCellFAB& u   = a_u[dex];

     //      VoFStencil&  utan_stencil;
     //      FaceStencil& flux_stencil;

      int n1, t1, t2;
      //      for( int dir=0 ; dir < SpaceDim ; dir++ )
      //      {
         if( dir == 0 )
         {
	   //            utan_stencil = m_uyz_stencil[dex];
	   //            flux_stencil = m_iflux_stencil[dex];
            n1 = 0;
            t1 = 1;
            t2 = 2;
         }
         else if( dir == 1 )
         {
	   //            utan_stencil = m_uxz_stencil[dex];
	   //            flux_stencil = m_jflux_stencil[dex];
            n1 = 1;
            t1 = 0;
            t2 = 2;
         }
         else
         {
	   //            utan_stencil = m_uxy_stencil[dex];
	   //            flux_stencil = m_kflux_stencil[dex];
            t1 = 0;
            t2 = 1;
            n1 = 2;
         }

         Real dxi = 1/m_dx[dir];

         const EBGraph& curEBGraph = eb.getEBGraph();
	 //         IntVectSet notRegular     = eb.getIrregIVS(a_u.box());
	 //         IntVectSet irreg_res      = eb.getIrregIVS(a_res.box());
         for(FaceIterator it(notRegular,curEBGraph,dir,FaceStop::SurroundingNoBoundary); it.ok();++it)
         {
            FaceIndex fi=it();

// Compute NS-momentum viscosity fluxes:
            Real flxr1=0;
            Real flxr2=0;
            Real flxz=0;

	    //            cout << dir << " Face index " << fi.gridIndex(Side::Lo) << " and " << fi.gridIndex(Side::Hi) << " uses " << endl;

            for( int ii=0 ; ii < flux_stencil(fi,0).size() ; ii++ )
            {
               FaceIndex fic = flux_stencil(fi,0).face(ii);
	       //               cout << fic.gridIndex(Side::Lo) << " and " << fic.gridIndex(Side::Hi) << " weight " << flux_stencil(fi,0).weight(ii) << endl;

               VolIndex voh = fic.getVoF(Side::Hi);
               VolIndex vol = fic.getVoF(Side::Lo);

               Real dut1_dt1 = 0;
               Real dun_dt1  = 0;

               const VoFStencil& vs=utan_stencil(voh,0);
               for( int i=0 ; i<vs.size() ; i++ )
       	       {
		   if( a_u.box().contains(vs.vof(i).gridIndex()) )
		   {
                   if( dir == 0 )
       	              dut1_dt1 += a_u(vs.vof(i),t1)*vs.weight(i);
		   else
		   {
		      double r = (vs.vof(i).gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
		      dut1_dt1 += a_u(vs.vof(i),t1)*r*vs.weight(i);
		   }
  	           dun_dt1  += a_u(vs.vof(i),n1)*vs.weight(i);
		   }
               }
               const VoFStencil& vsl = utan_stencil(vol,0);
	       for( int i=0 ; i < vsl.size() ; i++ )
	       {
		  if( a_u.box().contains(vsl.vof(i).gridIndex()) )
		  {
		  if( dir == 0 )
	             dut1_dt1 += a_u(vsl.vof(i),t1)*vsl.weight(i);
		  else
		  {
		      double r = (vsl.vof(i).gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
		      dut1_dt1 += a_u(vsl.vof(i),t1)*r*vsl.weight(i);
		  }
	          dun_dt1  += a_u(vsl.vof(i),n1)*vsl.weight(i);
		  }
	       }
	       dut1_dt1 *= 0.5;
	       dun_dt1  *= 0.5;
	       //               Real dut2_dt2 = 0;
	       //               Real dun_dt2  = 0;
               bool volinbox = a_lambda.box().contains(vol.gridIndex());
               bool vohinbox = a_lambda.box().contains(voh.gridIndex());
               Real la=0, mu=0;
               if(  volinbox && vohinbox )
	       {
	          la = (a_lambda(vol,0)+a_lambda(voh,0))*half*dxi;
   	          mu = (a_mu(vol,0)    +a_mu(voh,0))*half*dxi;
	       }
	       else if( vohinbox )
	       {
	          la = a_lambda(voh,0)*dxi;
   	          mu = a_mu(voh,0)*dxi;
	       }
	       else if( volinbox )
	       {
	          la = a_lambda(vol,0)*dxi;
   	          mu = a_mu(vol,0)*dxi;
	       }
		 
               Real dun_dn = 0;
               if( volinbox && vohinbox )
                  dun_dn  = (a_u(voh,n1)-a_u(vol,n1))*dxi;

               Real rp= (voh.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
               Real r = (vol.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];

               Real drun_dn = 0;
	       if( volinbox && vohinbox )
                   drun_dn = (rp*a_u(voh,0)-r*a_u(vol,0))*dxi;

               Real dut1_dn = 0;
	       if( volinbox && vohinbox )
                   dut1_dn = (a_u(voh,t1)-a_u(vol,t1))*dxi;

	       //              RealVect flxc;
	       //              flxc[n1] =  (2*mu+la)*dun_dn+la*(dut1_dt1);
	       //	       flxc[t1] =  mu*( dun_dt1 + dut1_dn );

               Real wgh = flux_stencil(fi,0).weight(ii);
               Real rh =(r+rp)*0.5;

               if( dir == 0 )
	       {
		  flxr1 += wgh*2*mu*rh*dun_dn;
		  flxr2 += wgh*la*(drun_dn/rh+dut1_dt1);
		  flxz  += wgh*mu*rh*(dun_dt1+dut1_dn);
	       }
	       else
	       {
		  flxr2 += wgh*mu*(dut1_dn+dun_dt1);
		  flxz  += wgh*((2*mu+la)*dun_dn + la*dut1_dt1/rh);
	       }
	       //               for( int d = 0 ; d < SpaceDim ; d++ )
	       //                  flx[d] += flxc[d]*flux_stencil(fi,0).weight(ii);
            }

            VolIndex vohi=fi.getVoF(Side::Hi);
            if( irreg_res.contains(vohi.gridIndex()) )
            {
	       if( dir == 0 )
	       {
                  RealVect c=eb.centroid(vohi);
                  Real r=(c[0]+vohi.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
                  a_res(vohi,0) -= (flxr1/r+flxr2);
 	          a_res(vohi,1) -= flxz/r;
	       }
	       else
	       {
                  a_res(vohi,0) -= flxr2;
	          a_res(vohi,1) -= flxz;
	       }
            }
            VolIndex volo=fi.getVoF(Side::Lo);
	    if( irreg_res.contains(volo.gridIndex()) )
	    { 
	       if( dir == 0 )
	       {
                  RealVect c=eb.centroid(volo);
                  Real r=(c[0]+volo.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
                  a_res(volo,0) += (flxr1/r+flxr2);
 	          a_res(volo,1) += flxz/r;
	       }
	       else
	       {
                  a_res(volo,0) += flxr2;
	          a_res(volo,1) += flxz;
	       }
	    }

	    //            if( dir == 1 )
	    //	    {
	    //	      cout << volo.gridIndex() << " " << vohi.gridIndex() << endl;
	    //              cout << "flxr1 " << flxr1 << "flxr2 " << flxr2 << "  flxz = " << flxz << endl;

	    //	    }

	 }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::update_irregular_eeq( EBCellFAB& a_res,
					  const EBCellFAB& a_u,
					  const EBISBox& eb,
					  const BaseIFFAB<FaceStencil>& flux_stencil,
					  const BaseIVFAB<VoFStencil>& utan_stencil,
					  const EBCellFAB& a_mu,
					  const EBCellFAB& a_lambda,
					  int dir )
{
   CH_assert( a_res.nComp() == 1 );

   const Real half=0.5;
  //   for( DataIterator li=m_dbl.dataIterator() ; li.ok() ; ++li )
  //   {
     //      DataIndex dex(li());

     //      EBCellFAB& res = a_res[dex];
     //      EBCellFAB& u   = a_u[dex];

     //      VoFStencil&  utan_stencil;
     //      FaceStencil& flux_stencil;

      int n1, t1, t2;
      //      for( int dir=0 ; dir < SpaceDim ; dir++ )
      //      {
         if( dir == 0 )
         {
	   //            utan_stencil = m_uyz_stencil[dex];
	   //            flux_stencil = m_iflux_stencil[dex];
            n1 = 0;
            t1 = 1;
            t2 = 2;
         }
         else if( dir == 1 )
         {
	   //            utan_stencil = m_uxz_stencil[dex];
	   //            flux_stencil = m_jflux_stencil[dex];
            n1 = 1;
            t1 = 0;
            t2 = 2;
         }
         else
         {
	   //            utan_stencil = m_uxy_stencil[dex];
	   //            flux_stencil = m_kflux_stencil[dex];
            t1 = 0;
            t2 = 1;
            n1 = 2;
         }

         Real dxi = 1/m_dx[dir];

         const EBGraph& curEBGraph = eb.getEBGraph();
         IntVectSet notRegular     = eb.getIrregIVS(a_u.box());
         for(FaceIterator it(notRegular,curEBGraph,dir,FaceStop::SurroundingNoBoundary); it.ok();++it)
         {
            FaceIndex fi=it();
// Compute NS-momentum viscosity fluxes:
            RealVect flx = RealVect::Zero;

            for( int ii=0 ; ii < flux_stencil(fi,0).size() ; ii++ )
            {
               FaceIndex fic = flux_stencil(fi,0).face(ii);

               VolIndex voh = fic.getVoF(Side::Hi);
               VolIndex vol = fic.getVoF(Side::Lo);

               Real dut1_dt1 = 0;
               Real dun_dt1  = 0;
               const VoFStencil& vs=utan_stencil(voh,0);
               for( int i=0 ; i<vs.size() ; i++ )
       	       {
       	           dut1_dt1 += a_u(vs.vof(i),t1)*vs.weight(i);
	           dun_dt1  += a_u(vs.vof(i),n1)*vs.weight(i);
               }
               const VoFStencil& vsl = utan_stencil(vol,0);
	       for( int i=0 ; i<vsl.size() ; i++ )
	       {
	          dut1_dt1 += a_u(vsl.vof(i),t1)*vsl.weight(i);
	          dun_dt1  += a_u(vsl.vof(i),n1)*vsl.weight(i);
	       }
	       dut1_dt1 *= 0.5;
	       dun_dt1  *= 0.5;

               Real dut2_dt2 = 0;
               Real dun_dt2  = 0;
               if( SpaceDim == 3 )
	       {
	           const VoFStencil& vsh=utan_stencil(voh,1);
                   for( int i=0 ; i<vsh.size() ; i++ )
	           {
	              dut2_dt2 += a_u(vsh.vof(i),t1)*vsh.weight(i);
 	              dun_dt2  += a_u(vsh.vof(i),n1)*vsh.weight(i);
                   }
                   const VoFStencil& vsl = utan_stencil(vol,1);
	           for( int i=0 ; i<vsl.size() ; i++ )
	           {
	              dut2_dt2 += a_u(vsl.vof(i),t1)*vsl.weight(i);
	              dun_dt2  += a_u(vsl.vof(i),n1)*vsl.weight(i);
	           }
	           dut2_dt2 *= 0.5;
	           dun_dt2  *= 0.5;
       	       }

	       Real la = (a_lambda(vol,0)+a_lambda(voh,0))*half*dxi;
	       Real mu = (a_mu(vol,0)    +a_mu(voh,0))*half*dxi;

               Real dun_dn  = (a_u(voh,n1)-a_u(vol,n1))*dxi;
               Real dut1_dn = (a_u(voh,t1)-a_u(vol,t1))*dxi;
               RealVect flxc;

               flxc[n1] =  (2*mu+la)*dun_dn+la*(dut1_dt1);
	       flxc[t1] =  mu*( dun_dt1 + dut1_dn );
               if( SpaceDim == 3 )
	       {
	          Real dut2_dn = (a_u(voh,t2)-a_u(vol,t2))*dxi;
   	          flxc[t2] = mu*(dun_dt2+dut2_dn);
                  flxc[n1] += la*dut2_dt2;
	       }
               for( int d = 0 ; d < SpaceDim ; d++ )
                  flx[d] += flxc[d]*flux_stencil(fi,0).weight(ii);
            }
            Real flxeeq = 0;
            VolIndex vlo = fi.getVoF(Side::Lo);
            VolIndex vhi = fi.getVoF(Side::Hi);
            for( int d=0 ; d < SpaceDim ; d++ )
	        flxeeq += flx[d]*(a_u(vlo,d)+a_u(vhi,d))*half;
            if( notRegular.contains(vhi.gridIndex()) )
	       a_res(vhi,0) -= flxeeq;
	    if( notRegular.contains(vlo.gridIndex()) )
               a_res(vlo,0) += flxeeq;
	 }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::update_irregular_eeq_rz(
			      EBCellFAB& a_res,
			      const EBCellFAB& a_u,
			      const EBISBox& eb,
			      const BaseIFFAB<FaceStencil>& flux_stencil,
			      const BaseIVFAB<VoFStencil>& utan_stencil,
	                      const EBCellFAB& a_mu,
			      const EBCellFAB& a_lambda,
			      int dir,
			      const IntVectSet& irreg )
{
   const Real half=0.5;
   int n1, t1, t2;
   if( dir == 0 )
   {
	   //            utan_stencil = m_uyz_stencil[dex];
	   //            flux_stencil = m_iflux_stencil[dex];
      n1 = 0;
      t1 = 1;
      t2 = 2;
   }
   else if( dir == 1 )
   {
	   //            utan_stencil = m_uxz_stencil[dex];
	   //            flux_stencil = m_jflux_stencil[dex];
      n1 = 1;
      t1 = 0;
      t2 = 2;
   }
   else
   {
	   //            utan_stencil = m_uxy_stencil[dex];
	   //            flux_stencil = m_kflux_stencil[dex];
      t1 = 0;
      t2 = 1;
      n1 = 2;
   }

   Real dxi = 1/m_dx[dir];

   const EBGraph& curEBGraph = eb.getEBGraph();
   //   IntVectSet notRegular     = eb.getIrregIVS(a_u.box());
   for(FaceIterator it(irreg,curEBGraph,dir,FaceStop::SurroundingNoBoundary); it.ok();++it)
   {
       FaceIndex fi=it();
// Compute NS-momentum viscosity fluxes:
//            RealVect flx = RealVect::Zero;
       Real flxr1=0;
       Real flxr2=0;
       Real flxz=0;

            for( int ii=0 ; ii < flux_stencil(fi,0).size() ; ii++ )
            {
               FaceIndex fic = flux_stencil(fi,0).face(ii);

               VolIndex voh = fic.getVoF(Side::Hi);
               VolIndex vol = fic.getVoF(Side::Lo);

               Real dut1_dt1 = 0;
               Real dun_dt1  = 0;

               const VoFStencil& vs=utan_stencil(voh,0);
               for( int i=0 ; i<vs.size() ; i++ )
       	       {
		   if( a_u.box().contains(vs.vof(i).gridIndex()) )
		   {
                   if( dir == 0 )
       	              dut1_dt1 += a_u(vs.vof(i),t1)*vs.weight(i);
		   else
		   {
		      double r = (vs.vof(i).gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
		      dut1_dt1 += a_u(vs.vof(i),t1)*r*vs.weight(i);
		   }
  	           dun_dt1  += a_u(vs.vof(i),n1)*vs.weight(i);
		   }
               }
               const VoFStencil& vsl = utan_stencil(vol,0);
	       for( int i=0 ; i < vsl.size() ; i++ )
	       {
		  if( a_u.box().contains(vsl.vof(i).gridIndex()) )
		  {
		  if( dir == 0 )
	             dut1_dt1 += a_u(vsl.vof(i),t1)*vsl.weight(i);
		  else
		  {
		      double r = (vsl.vof(i).gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
		      dut1_dt1 += a_u(vsl.vof(i),t1)*r*vsl.weight(i);
		  }
	          dun_dt1  += a_u(vsl.vof(i),n1)*vsl.weight(i);
		  }
	       }
	       dut1_dt1 *= 0.5;
	       dun_dt1  *= 0.5;
	       //               Real dut2_dt2 = 0;
	       //               Real dun_dt2  = 0;
               bool volinbox = a_lambda.box().contains(vol.gridIndex());
               bool vohinbox = a_lambda.box().contains(voh.gridIndex());
               Real la, mu;
               if(  volinbox && vohinbox )
	       {
	          la = (a_lambda(vol,0)+a_lambda(voh,0))*half*dxi;
   	          mu = (a_mu(vol,0)    +a_mu(voh,0))*half*dxi;
	       }
	       else if( !volinbox )
	       {
	          la = a_lambda(voh,0)*dxi;
   	          mu = a_mu(voh,0)*dxi;
	       }
	       else
	       {
	          la = a_lambda(vol,0)*dxi;
   	          mu = a_mu(vol,0)*dxi;
	       }
		 
               Real dun_dn = 0;
               if( volinbox && vohinbox )
                  dun_dn  = (a_u(voh,n1)-a_u(vol,n1))*dxi;

               Real rp= (voh.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
               Real r = (vol.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];

               Real drun_dn = 0;
               Real uav = 0;
               Real wav = 0;
               Real dut1_dn = 0;
	       if( volinbox && vohinbox )
	       {
                   drun_dn = (rp*a_u(voh,0)-r*a_u(vol,0))*dxi;
                   uav = (a_u(voh,0)+a_u(vol,0))*half;
                   wav = (a_u(voh,1)+a_u(vol,1))*half;
                   dut1_dn = (a_u(voh,t1)-a_u(vol,t1))*dxi;
	       }

	       //              RealVect flxc;
	       //              flxc[n1] =  (2*mu+la)*dun_dn+la*(dut1_dt1);
	       //	       flxc[t1] =  mu*( dun_dt1 + dut1_dn );

               Real wgh = flux_stencil(fi,0).weight(ii);
               Real rh =(r+rp)*0.5;

               if( dir == 0 )
	       {
		  flxr1 += wgh*2*mu*rh*dun_dn*uav;
		  flxr2 += wgh*la*rh*(drun_dn/rh+dut1_dt1)*uav;
		  flxz  += wgh*mu*rh*(dun_dt1+dut1_dn)*wav;
	       }
	       else
	       {
		  flxr2 += rh*wgh*mu*(dut1_dn+dun_dt1)*uav;
		  flxz  += rh*wgh*((2*mu+la)*dun_dn + la*dut1_dt1/rh)*wav;
	       }
	       //               for( int d = 0 ; d < SpaceDim ; d++ )
	       //                  flx[d] += flxc[d]*flux_stencil(fi,0).weight(ii);
            }
	    //            if( dir == 0 )
	    //	       cout << "flxr1 " << flxr1;
	    //            cout << "flxr2 " << flxr2 << "  flxz = " << flxz << endl;


            VolIndex vohi=fi.getVoF(Side::Hi);
            VolIndex volo=fi.getVoF(Side::Lo);

	    //            if( a_res.box().contains(vohi.gridIndex()) )
            if( irreg.contains(vohi.gridIndex()) )
            {
               RealVect c=eb.centroid(vohi);
               Real r=(c[0]+vohi.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
	       if( dir == 0 )
	       {
                  a_res(vohi,0) -= ( flxr1 + flxr2 + flxz )/r;
	       }
	       else
	       {
		  a_res(vohi,0) -= (flxr2+flxz)/r;
	       }
            }

	    //	    if( a_res.box().contains(volo.gridIndex()) )
	    if( irreg.contains(volo.gridIndex()) )
	    { 
               RealVect c=eb.centroid(volo);
               Real r=(c[0]+volo.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
	       if( dir == 0 )
	       {
                  a_res(volo,0) += ( flxr1 + flxr2 + flxz )/r;
	       }
	       else
	       {
		  a_res(volo,0) += ( flxr2+flxz)/r;
	       }
	    }
	 }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::get_ns_uweight( BaseIVFAB<Real>& weight,
				    const EBISBox& eb,
				    const BaseIFFAB<FaceStencil>& flux_stencil, 
				    const EBCellFAB& a_mu,
				    const EBCellFAB& a_lambda,
				    int dir )
{
   int n1, t1, t2;
   if( dir == 0 )
   {
      n1 = 0;
      t1 = 1;
      t2 = 2;
   }
   else if( dir == 1 )
   {
      n1 = 1;
      t1 = 0;
      t2 = 2;
   }
   else
   {
      t1 = 0;
      t2 = 1;
      n1 = 2;
   }

   Real dxi = 1/m_dx[dir];
   Real half = 1.0/2;

   const EBGraph& curEBGraph = eb.getEBGraph();
   IntVectSet notRegular     = eb.getIrregIVS(a_mu.box());
   //   IntVectSet notRegular = weight.getIVS();

   for(FaceIterator it(notRegular,curEBGraph,dir,FaceStop::SurroundingNoBoundary); it.ok();++it)
   {
       FaceIndex fi=it();

       VolIndex vof_actual_h = fi.getVoF(Side::Hi);
       VolIndex vof_actual_l = fi.getVoF(Side::Lo);

       for( int ii=0 ; ii < flux_stencil(fi,0).size() ; ii++ )
       {
           FaceIndex fic = flux_stencil(fi,0).face(ii);

           VolIndex voh = fic.getVoF(Side::Hi);
           VolIndex vol = fic.getVoF(Side::Lo);

           bool inh = notRegular.contains(voh.gridIndex());
           bool inl = notRegular.contains(vol.gridIndex());
           Real mu, la;
           if( inh && inl )
	   {
              la = (a_lambda(vol,0)+a_lambda(voh,0))*half*dxi;
	      mu = (a_mu(vol,0)    +a_mu(voh,0))*half*dxi;
	   }
	   else if( inh )
	   {
              la = a_lambda(voh,0)*dxi;
	      mu = a_mu(voh,0)*dxi;
	   }
           else if( inl )
	   {
              la = a_lambda(vol,0)*dxi;
	      mu = a_mu(vol,0)*dxi;
	   }

           if( vof_actual_h == voh  && inh )
	   {
	      weight(voh,n1) -= dxi*(2*mu+la)*flux_stencil(fi,0).weight(ii);
	      weight(voh,t1) -= dxi*mu*flux_stencil(fi,0).weight(ii);
              if( SpaceDim == 3 )
		 weight(voh,t2) -= dxi*mu*flux_stencil(fi,0).weight(ii);
	   }
	   if(  vof_actual_l == vol && inl )
	   {
	      weight(vol,n1) -= dxi*(2*mu+la)*flux_stencil(fi,0).weight(ii);
	      weight(vol,t1) -= dxi*mu*flux_stencil(fi,0).weight(ii);
              if( SpaceDim == 3 )
		 weight(vol,t2) -= dxi*mu*flux_stencil(fi,0).weight(ii);
	   }
       }
   }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::get_ns_uweight_rz( BaseIVFAB<Real>& weight,
				     const EBISBox& eb,
				     const BaseIFFAB<FaceStencil>& flux_stencil, 
 	                             const EBCellFAB& a_mu,
				     const EBCellFAB& a_lambda,
				     int dir,
				     const IntVectSet& notRegular )
{
   int n1, t1, t2;
   if( dir == 0 )
   {
      n1 = 0;
      t1 = 1;
      t2 = 2;
   }
   else if( dir == 1 )
   {
      n1 = 1;
      t1 = 0;
      t2 = 2;
   }
   else
   {
      t1 = 0;
      t2 = 1;
      n1 = 2;
   }

   Real dxi  = 1/m_dx[dir];
   Real half = 1.0/2;

   const EBGraph& curEBGraph = eb.getEBGraph();
   //   IntVectSet notRegular     = eb.getIrregIVS(a_mu.box());
   //   IntVectSet notRegular = weight.getIVS();

   for(FaceIterator it(notRegular,curEBGraph,dir,FaceStop::SurroundingNoBoundary); it.ok();++it)
   {
       FaceIndex fi=it();

       VolIndex vof_actual_h = fi.getVoF(Side::Hi);
       VolIndex vof_actual_l = fi.getVoF(Side::Lo);

       for( int ii=0 ; ii < flux_stencil(fi,0).size() ; ii++ )
       {
           FaceIndex fic = flux_stencil(fi,0).face(ii);

           VolIndex voh = fic.getVoF(Side::Hi);
           VolIndex vol = fic.getVoF(Side::Lo);

           Real la=0,mu=0;
           bool inh = notRegular.contains(voh.gridIndex());
           bool inl = notRegular.contains(vol.gridIndex());
           if( inh && inl )
	   {
              la = (a_lambda(vol,0)+a_lambda(voh,0))*half*dxi;
	      mu = (a_mu(vol,0)    +a_mu(voh,0))*half*dxi;
	   }
	   else if( inl )
	   {
              la = a_lambda(vol,0)*dxi;
              mu = a_mu(vol,0)*dxi;
	   }
           else if( inh )
	   {
              la = a_lambda(voh,0)*dxi;
              mu = a_mu(voh,0)*dxi;
	   }

           Real rp= (voh.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];
           Real r = (vol.gridIndex()[0]+0.5)*m_dx[0]+m_origin[0];

	   //           Real drun_dn = (rp*a_u(voh,0)-r*a_u(vol,0))*dxi;
	   //           Real dut1_dn = (a_u(voh,t1)-a_u(vol,t1))*dxi;

           Real wgh = flux_stencil(fi,0).weight(ii);
           Real rh =(r+rp)*0.5;

           Real rpc = rp + eb.centroid(voh)[0]*m_dx[0];
           Real rc  = r  + eb.centroid(vol)[0]*m_dx[0];

           if( vof_actual_h == voh  && inh )
	   {
              
	     if( dir == 0 )
	     {
	        weight(voh,n1) -= dxi*wgh*(2*mu*rh/rpc+la*rp/rh);
   	        weight(voh,t1) -= dxi*mu*wgh*rh/rpc;
	     }
             else
	     {
	        weight(voh,n1) -= dxi*wgh*(2*mu+la);
	        weight(voh,t1) -= dxi*wgh*mu;
	     }
	   }
	   if(  vof_actual_l == vol && inl )
	   {
	     if( dir == 0 )
	     {
	         weight(vol,n1) -= dxi*wgh*(2*mu*rh/rc+la*r/rh);
   	         weight(vol,t1) -= dxi*mu*wgh*rh/rc;
	     }
             else
	     {
	        weight(vol,n1) -= dxi*wgh*(2*mu+la);
	        weight(vol,t1) -= dxi*wgh*mu;
	     }
	   }
           if( vof_actual_h == vol || vof_actual_l == voh )
	     cout << "XXX This should not happen XXX" << endl;
       }
   }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::get_uweight( const DataIndex dex, LevelData<EBCellFAB>& wgh )
		
{
   Real cofi = 1/m_c;
   const BaseFab<Real>& la_singv  = m_lambda[dex].getSingleValuedFAB();
   const BaseFab<Real>& mu_singv  = m_mu[dex].getSingleValuedFAB();
   const BaseFab<Real>& rho_singv = m_rho[dex].getSingleValuedFAB();
   BaseFab<Real>& wgh_singv = wgh[dex].getSingleValuedFAB();
	  
   if( !m_rz )
       FORT_GETWGH( CHF_CONST_FRA1(la_singv,0), CHF_CONST_FRA1(mu_singv,0),
	       CHF_CONST_FRA1(rho_singv,0), CHF_FRA(wgh_singv),
	       CHF_CONST_REAL(cofi), CHF_CONST_REALVECT(m_dx),
	       CHF_BOX(m_dbl.get(dex)) );
   else
       FORT_GETWGH_RZ( CHF_CONST_FRA1(la_singv,0), CHF_CONST_FRA1(mu_singv,0),
               CHF_CONST_FRA1(rho_singv,0), CHF_FRA(wgh_singv),
	       CHF_CONST_REAL(cofi), CHF_CONST_REALVECT(m_dx),
	       CHF_BOX(m_dbl.get(dex)), CHF_CONST_REALVECT(m_origin) );


   Box loBox[SpaceDim],hiBox[SpaceDim];
   int hasLo[SpaceDim],hasHi[SpaceDim];

   for( int dir=0 ; dir <SpaceDim ; dir++ )
   {
// Navier-Stokes domain boundary weights
      Box centerBox,entireBox;
      ebEllipticLoHiCenter(loBox[dir], hasLo[dir], hiBox[dir], hasHi[dir],
                           centerBox, entireBox, wgh[dex].box(), wgh[dex].box(), m_domain, dir);
      if( hasLo[dir] == 1 )
      {
          int side = -1;
          CH_assert(m_domainBC != NULL);
          for( BoxIterator b(loBox[dir]); b.ok(); ++b )
	  {
	     Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
             if( vofs.size() > 1 )
                 MayDay::Warning("EBAMRNS::get_uweight:Multiple vofs on domain BC");
             if( vofs.size() >0 )
	     {
                Vector<Real> face_wgh(SpaceDim);
                if( m_rz )
                   m_domainBC->get_uweight_rz( face_wgh, vofs[0], m_ebisl[dex], 
                                        m_origin, m_dx, dir, Side::Lo, 
                                        m_mu[dex](vofs[0],0), m_lambda[dex](vofs[0],0));
	        else
                   m_domainBC->get_uweight( face_wgh, vofs[0], m_ebisl[dex], 
                    		        m_origin, m_dx, dir, Side::Lo,  
					m_mu[dex](vofs[0],0), m_lambda[dex](vofs[0],0));
                for( int c = 0 ; c < SpaceDim ; c++ )
	           wgh[dex](vofs[0],c) =  wgh[dex](vofs[0],c)+side*face_wgh[c]/m_dx[dir];
	     }
	  }
      }
      if( hasHi[dir] == 1 )
      {
         int side = 1;
         CH_assert(m_domainBC != NULL);
         for( BoxIterator b(hiBox[dir]); b.ok(); ++b )
	 {
	    Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
            if( vofs.size() > 1 )
                MayDay::Warning("EBAMRNS::Multiple vofs in domain BC");
            if( vofs.size() > 0 )
	    {
               Vector<Real> face_wgh(SpaceDim);
               if( m_rz )
		 {
                  m_domainBC->get_uweight_rz( face_wgh, vofs[0], m_ebisl[dex], 
                                        m_origin, m_dx, dir, Side::Hi,
                                        m_mu[dex](vofs[0],0),m_lambda[dex](vofs[0],0));
		 }
               else
                  m_domainBC->get_uweight( face_wgh, vofs[0], m_ebisl[dex], 
                    		        m_origin, m_dx, dir, Side::Hi,  
					m_mu[dex](vofs[0],0), m_lambda[dex](vofs[0],0));
                 
	       //               cout << dir << " Hi wgh " << face_wgh[0] << " " << face_wgh[1]<<endl;
               for( int c = 0 ; c < SpaceDim ; c++ )
		   wgh[dex](vofs[0],c)=wgh[dex](vofs[0],c)+side*face_wgh[c]/m_dx[dir];
	    }
	 }
      }
   }

   //   for( VoFIterator v(m_ebisl[dex].getIrregIVS(m_ebisl[dex].getRegion()),m_ebisl[dex].getEBGraph()); v.ok(); ++v )
   //   for( VoFIterator v(m_ebisl[dex].getIrregIVS(wgh[dex].box()),m_ebisl[dex].getEBGraph()); v.ok(); ++v )
   //   for( VoFIterator v(m_ebisl[dex].getIrregIVS(m_dbl.get(dex)),m_ebisl[dex].getEBGraph()); v.ok(); ++v )
   //   for( VoFIterator v(m_irreg_w_ghost[dex],m_ebisl[dex].getEBGraph()); v.ok(); ++v )
   for( VoFIterator v(m_irreg[dex],m_ebisl[dex].getEBGraph()); v.ok(); ++v )
   {
       VolIndex vof = v();
       for( int c=0 ;c < SpaceDim ; c++ )
          wgh[dex](vof,c) = 0;       
// Domain BC for irregular cells:
       for( int dir = 0 ; dir < SpaceDim ; dir++ )
       {
          if( hasLo[dir] == 1 )
          {
             if( loBox[dir].contains(vof.gridIndex()))
             {

                 Vector<Real> face_wgh(SpaceDim);
                 if( m_rz )
                   m_domainBC->get_uweight_rz( face_wgh, vof, m_ebisl[dex], 
                                        m_origin, m_dx, dir, Side::Lo, 
                                        m_mu[dex](vof,0), m_lambda[dex](vof,0));
	        else
                   m_domainBC->get_uweight( face_wgh, vof, m_ebisl[dex], 
                    		        m_origin, m_dx, dir, Side::Lo,  
					m_mu[dex](vof,0), m_lambda[dex](vof,0));
		 //		 cout << "lowr bndr"<< vof.gridIndex() << " add " << face_wgh[0]/m_dx[dir] << ","<< face_wgh[1]/m_dx[dir] <<endl;
                for( int c = 0 ; c < SpaceDim ; c++ )
	           wgh[dex](vof,c) =  wgh[dex](vof,c)-face_wgh[c]/m_dx[dir];
	     }
	  }
          if( hasHi[dir] == 1 )
          {
             if( hiBox[dir].contains(vof.gridIndex()) )
             {
                Vector<Real> face_wgh(SpaceDim);
                if( m_rz )
                  m_domainBC->get_uweight_rz( face_wgh, vof, m_ebisl[dex], 
                                        m_origin, m_dx, dir, Side::Hi,
                                        m_mu[dex](vof,0),m_lambda[dex](vof,0));
                 else
                  m_domainBC->get_uweight( face_wgh, vof, m_ebisl[dex], 
                    		        m_origin, m_dx, dir, Side::Hi,  
					m_mu[dex](vof,0), m_lambda[dex](vof,0));
               for( int c = 0 ; c < SpaceDim ; c++ )
		   wgh[dex](vof,c)=wgh[dex](vof,c)+face_wgh[c]/m_dx[dir];
	    }
	  }
       }
       for( int comp = 0 ; comp < SpaceDim ; comp++ )
       {
	  Real weight = m_weight[dex](vof,comp);
	  //          if( vof.gridIndex() == IntVect(2,2) )
	    //             cout << comp << "  " << weight << " " << vof.gridIndex() << endl;
          if( m_split_u  )
              weight -= cofi*m_rho[dex](vof,0)*m_uweight[dex](vof,0)*
		                                  m_ebisl[dex].volFrac(vof);
	  else
	      weight -= cofi*m_rho[dex](vof,0)*m_ebisl[dex].volFrac(vof);

          if( vof.gridIndex() == IntVect::Unit*2 )
	    {
	      //          cout << "aft" << comp << "  " << weight << " " << vof.gridIndex() << endl;
	  //          cout << "add to " << comp << " " << wgh[dex](vof,comp) << endl;
	    }
          wgh[dex](vof,comp) += weight;
       }
       //       if( hasNoEBLevelData )
       //       {

       Vector<Real> e_wgh(SpaceDim,0);
       if( m_rz )
       {
	 // 	   Real mu=m_mu[dex](vof,0);
	   // 	   Real la=m_lambda[dex](vof,0);
	   //	   m_ebBC->get_uweight_rz( e_wgh, vof, m_ebisl[dex], dex, m_origin, m_dx, mu, la );
       }
       else
	    //	     m_ebBC->getEBFlux( ebFlux, vof, au, dex, m_origin, m_dx, a_homogeneousPhysBC,
	    //			       mu, la );
	    {}
       for( int c = 0 ; c < SpaceDim ; c++ )
	  wgh[dex](vof,c) +=  e_wgh[c]*m_dxScale;

	  //       }
   }

   Real relpar;
   if( m_relaxType == 0 )
      relpar = 0.5;
   else
      relpar = 1.0;

   for( BoxIterator b(wgh[dex].box()); b.ok(); ++b )
   {
     if( !m_ebisl[dex].isCovered(b()) )
     {
        Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
        for( int iv=0 ; iv < vofs.size() ; iv++ )
           for( int comp = 0 ; comp < SpaceDim ; comp++ )          
              wgh[dex](vofs[iv],comp) = relpar/wgh[dex](vofs[iv],comp);
     }
   }

}


//void EBAMRNSviscOp::dirichlet_tangential( RealVect nv, RealVect tv, RealVect bv, 
//                                          const VolIndex &a_vof, int side,
//                                          const EBCellFAB& la, const EBCellFAB& mu,
//                                          BaseFab<Real>& tterm, const Real faceAreaFrac,
//                                      RealVect dx, RealVect x0, bool homogeneousPhysBC )
//{
//  if( homogeneousPhysBC )
//     for( int c = 0 ; c < SpaceDim ; c++ )
//       tterm(a_vof.gridIndex(),c) = 0;
//  else
//  {
//     RealVect x=static_cast<RealVect>(a_vof.gridIndex());
//     x[0] = (x[0]+0.5)*dx[0]+x0[0]+0.5*side*dx[0];
//     x[1] = (x[1]+0.5)*dx[1]+x0[1]+0.5*side*dx[1];

     //     double u =  sin(x[0])*cos(x[1]);
//     double ux = cos(x[0])*sin(x[1]);
//     double uy =-sin(x[0])*sin(x[1]);

     //     double v =  cos(x[0])*sin(x[1]);
//     double vx =-sin(x[0])*cos(x[1]);
//     double vy = cos(x[0])*cos(x[1]);

// 2D formulas
//     double dutdt = tv[0]*(tv[0]*ux+tv[1]*vx) + tv[1]*(tv[0]*uy+tv[1]*vy);
//     double dubdb = bv[0]*(bv[0]*ux+bv[1]*vx) + bv[1]*(bv[0]*uy+bv[1]*vy);
//     double dundt = tv[0]*(nv[0]*ux+nv[1]*vx) + tv[1]*(nv[0]*uy+nv[1]*vy);
//     double dundb = bv[0]*(nv[0]*ux+nv[1]*vx) + bv[1]*(nv[0]*uy+nv[1]*vy);
//
//     for( int c = 0 ; c < SpaceDim ; c++ )
//       tterm(a_vof.gridIndex(),c) = faceAreaFrac*( la(a_vof,0)*(dutdt+dubdb)*nv[c]+
//                      mu(a_vof,0)*(dundt*tv[c]+dundb*bv[c]) );
//  }
//}

//void EBAMRNSviscOp::dirichlet_tangential( RealVect nv, RealVect tv, RealVect bv,
//                                          const VolIndex& vof, const EBCellFAB& la,
//                                          const EBCellFAB& mu, const EBISBox& ebis_box,
//                                          RealVect tt, RealVect x0, RealVect dx,
//					  bool homogeneousPhysBC )
//{
//  if( homogeneousPhysBC )
//    for( int c = 0 ; c<SpaceDim ; c++ )
//      tt[c] = 0;
//  else
//  {
//      RealVect x = static_cast<RealVect>(vof.gridIndex())+ebis_box.bndryCentroid(vof);
//      x[0] = (x[0]+0.5)*dx[0]+x0[0];
//      x[1] = (x[1]+0.5)*dx[1]+x0[1];
      
      //      double u =  sin(x[0])*cos(x[1]);
//      double ux = cos(x[0])*sin(x[1]);
//      double uy =-sin(x[0])*sin(x[1]);

      //      double v =  cos(x[0])*sin(x[1]);
//      double vx =-sin(x[0])*cos(x[1]);
//      double vy = cos(x[0])*cos(x[1]);

// 2D formulas
//      double dutdt = tv[0]*(tv[0]*ux+tv[1]*vx) + tv[1]*(tv[0]*uy+tv[1]*vy);
//      double dubdb = bv[0]*(bv[0]*ux+bv[1]*vx) + bv[1]*(bv[0]*uy+bv[1]*vy);
//      double dundt = tv[0]*(nv[0]*ux+nv[1]*vx) + tv[1]*(nv[0]*uy+nv[1]*vy);
//      double dundb = bv[0]*(nv[0]*ux+nv[1]*vx) + bv[1]*(nv[0]*uy+nv[1]*vy);

//      for( int c = 0 ; c < SpaceDim ; c++ )
//         tt[c] = ebis_box.bndryArea(vof)*( la(vof,0)*(dutdt+dubdb)*nv[c]+
//                      mu(vof,0)*(dundt*tv[c]+dundb*bv[c]) );
//  }
//}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::get_energyeq_viscosity( LevelData<EBCellFAB>& a_res,
					    const LevelData<EBCellFAB>& a_u,
        const LevelData<EBCellFAB>* const        a_uCoar,
        bool                                     a_homogeneousPhysBC,
        bool                                     a_homogeneousCFBC,
        const LevelData<BaseIVFAB<Real> >* const a_ebFluxBCLD )
{
  //if input EBFlux leveldata is null, then we have no input eb flux
  bool hasNoEBLevelData = (a_ebFluxBCLD == NULL);

  CH_assert(a_u.nComp() == SpaceDim );
  CH_assert(a_res.nComp() == 1);

  LevelData<EBCellFAB>& au = (LevelData<EBCellFAB>&)a_u;
  applyCFBCs(au, a_uCoar, a_homogeneousCFBC);
  au.exchange(au.interval());

// Save slices of u, can not find any function to do this 
//  Vector< LevelData<EBCellFAB>* > uu(SpaceDim);
//  for( int c = 0 ; c <SpaceDim ; c++ )
//  {
//    EBCellFactory datafactory( m_ebisl );
//    uu[c] = new LevelData<EBCellFAB>;
//    uu[c]->define( a_u.getBoxes(), 1, a_u.ghostVect(), datafactory );
//    a_u.copyTo( Interval(c,c), *(uu[c]), Interval(0,0) );
//  }
//  ArrayToRaw atr;

  //  atr.chk_nan(a_u,m_ebisl," u before regupdate is nan ");  

  for (DataIterator dit = a_u.dataIterator(); dit.ok(); ++dit)
  {
      DataIndex dex(dit());

      const EBCellFAB&     u              = au[dex];

      const BaseFab<Real>& u_single_valued= u.getSingleValuedFAB();
      //      const EBISBox& curEBISBox = m_ebisl[dex];
      
      EBCellFAB&           res = a_res[dex];
      BaseFab<Real>&       res_single_valued = res.getSingleValuedFAB();

      //      IntVectSet notRegular = m_ebisl[dex].getIrregIVS(m_dbl.get(dex));
      IntVectSet notRegular = m_irreg[dex];
  
      res.setVal(0.0);

      Box loBox[SpaceDim],hiBox[SpaceDim];
      int hasLo[SpaceDim],hasHi[SpaceDim];
      Box center, entire;
      Vector<Box> resid_boxes(3);
      for( int dir=0 ; dir <SpaceDim ; dir++ )
      {
         Box centerBox,entireBox;
         ebEllipticLoHiCenter(loBox[dir], hasLo[dir], hiBox[dir], hasHi[dir],
                              center, entire, res.box(), u.box(), m_domain, dir);
         center.growLo(dir);
         resid_boxes[dir] = center;
      }


// Navier-Stokes viscosity at regular cells.
      if( m_rz )
         FORT_NSEEQ_RZ( CHF_FRA1(res_single_valued,0),
			CHF_CONST_FRA(u_single_valued),
                        CHF_CONST_REALVECT(m_dx),
			CHF_CONST_FRA1(m_mu[dex].getSingleValuedFAB(),0), 
                        CHF_CONST_FRA1(m_lambda[dex].getSingleValuedFAB(),0),
			CHF_CONST_REALVECT(m_origin),
		        CHF_BOX(resid_boxes[0]),
			CHF_BOX(resid_boxes[1]),
		        CHF_BOX(resid_boxes[2]),
			CHF_BOX(u.box()),
			CHF_BOX(res.box()) );
      else
         FORT_NSEEQ( CHF_FRA1(res_single_valued,0),
		     CHF_CONST_FRA(u_single_valued),
                     CHF_CONST_REALVECT(m_dx),
		     CHF_CONST_FRA1(m_mu[dex].getSingleValuedFAB(),0), 
                     CHF_CONST_FRA1(m_lambda[dex].getSingleValuedFAB(),0),
		     CHF_BOX(resid_boxes[0]),
		     CHF_BOX(resid_boxes[1]),
		     CHF_BOX(resid_boxes[2]),
		     CHF_BOX(u.box()),
		     CHF_BOX(res.box()) );



      for( int dir=0 ; dir <SpaceDim ; dir++ )
      {
// Navier-Stokes domain boundary viscous fluxes
         Box centerBox,entireBox;
         ebEllipticLoHiCenter(loBox[dir], hasLo[dir], hiBox[dir], hasHi[dir],
                              centerBox, entireBox, res.box(), u.box(), m_domain, dir);
         if( hasLo[dir] == 1 )
         {
             BaseFab<Real> loFaceFlux( loBox[dir], 1 );
             int side = -1;
             CH_assert(m_domainBC != NULL);
             for( BoxIterator b(loBox[dir]); b.ok(); ++b )
	     {
		Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
                if( vofs.size() > 1 )
                    MayDay::Warning("EBAMRNSviscOp::eeq Multiple vofs on domain BC");
		{
                if( vofs.size() > 0 )
		{
                   Real face_flux;
                   if( m_rz )
                      m_domainBC->getFaceFlux_eeq_rz( face_flux, vofs[0], u, m_origin,
					   m_dx, dir, Side::Lo, a_homogeneousPhysBC,
                                           m_mu[dex](vofs[0],0),
					   m_lambda[dex](vofs[0],0),m_time );
		   else
                      m_domainBC->getFaceFlux_eeq( face_flux, vofs[0], u, m_origin, 
					   m_dx, dir, Side::Lo, a_homogeneousPhysBC,
                                           m_mu[dex](vofs[0],0), 
                                           m_lambda[dex](vofs[0],0),m_time );
                   loFaceFlux(b(),0) = face_flux;
		}
		else
		  loFaceFlux(b(),0) = 0;
		}
	     }
             FORT_NS_ADD_BCFLUX_EEQ( CHF_FRA1(res_single_valued,0),
				 CHF_CONST_FRA1( loFaceFlux,0 ),
				 CHF_CONST_REAL( m_dx[dir] ),
				 CHF_CONST_INT( side ),
				 CHF_CONST_INT( dir ),
				 CHF_BOX( loBox[dir] ) );
	 }
         if( hasHi[dir] == 1 )
         {
             BaseFab<Real> hiFaceFlux( hiBox[dir], 1 );
             int side = 1;
             CH_assert(m_domainBC != NULL);
             for( BoxIterator b(hiBox[dir]); b.ok(); ++b )
	     {
		Vector<VolIndex> vofs = m_ebisl[dex].getVoFs(b());
                if( vofs.size() > 1 )
                    MayDay::Warning("EBAMRNS::Multiple vofs in domain BC");
		{
                if( vofs.size() > 0 )
		{
                   Real face_flux;
                   if( m_rz )
                      m_domainBC->getFaceFlux_eeq_rz( face_flux, vofs[0], u, m_origin, m_dx,
                                       dir, Side::Hi, a_homogeneousPhysBC,
                                       m_mu[dex](vofs[0],0),m_lambda[dex](vofs[0],0),m_time);
                   else
                      m_domainBC->getFaceFlux_eeq( face_flux, vofs[0], u, m_origin, m_dx,
                                       dir, Side::Hi, a_homogeneousPhysBC,
                                       m_mu[dex](vofs[0],0),m_lambda[dex](vofs[0],0),m_time);
		   hiFaceFlux(b(),0)=face_flux;
		}
		else
		  hiFaceFlux(b(),0) = 0;
		}
	     }
             FORT_NS_ADD_BCFLUX_EEQ( CHF_FRA1(res_single_valued,0),
				 CHF_CONST_FRA1( hiFaceFlux,0 ),
				 CHF_CONST_REAL( m_dx[dir] ),
				 CHF_CONST_INT( side ),
				 CHF_CONST_INT( dir ),
				 CHF_BOX( hiBox[dir] ) );
	 }
      }

      for( VoFIterator vi(notRegular,m_ebisl[dex].getEBGraph());vi.ok();++vi)
      {
         res(vi(),0) = 0;
      }
      //      cout << "-------------------------" << endl;
      //      for( BoxIterator b(res.getRegion());b.ok();++b)
      //      {
      //        cout << "ar(" << b() << ")= ";
      //	if( notRegular.contains(b()) )
      //	  cout << " irr ";
      //	if( m_ebisl[dex].isRegular(b()))
      //	  cout << " r " ;
      //	if( m_ebisl[dex].isIrregular(b()))
      //	  cout << " i " ;
      //	if( m_ebisl[dex].isCovered(b()))
      //	  cout << " c " ;
      //        cout << endl;
      //      }

      //      cout <<"Nan chk"<<endl;
      //      atr.chk_nan(a_res,m_ebisl," res after regupdate is nan ");


// Irregular flux update
      for( int dir = 0 ; dir < SpaceDim ; dir++ )
      {
         if( dir == 0 )
	 {
	    if( m_rz )
               update_irregular_eeq_rz( res, u, m_ebisl[dex], m_iflux_stencil[dex],
	     m_uyz_stencil[dex], m_mu[dex], m_lambda[dex], dir, m_irreg[dex] );
            else
               update_irregular_eeq( res, u, m_ebisl[dex], m_iflux_stencil[dex],
		     m_uyz_stencil[dex], m_mu[dex], m_lambda[dex], dir );
	 }
         else if( dir == 1 )
	 {
	    if( m_rz )
 	       update_irregular_eeq_rz( res, u, m_ebisl[dex], m_jflux_stencil[dex],
	      m_uxz_stencil[dex], m_mu[dex], m_lambda[dex], dir, m_irreg[dex] );
	    else
 	       update_irregular_eeq( res, u, m_ebisl[dex], m_jflux_stencil[dex],
			      m_uxz_stencil[dex], m_mu[dex], m_lambda[dex], dir );
	 }
         else
	 {
	    if( m_rz )
 	       update_irregular_eeq_rz( res, u, m_ebisl[dex], m_kflux_stencil[dex],
	      m_uxy_stencil[dex], m_mu[dex], m_lambda[dex], dir, m_irreg[dex] );
            else
 	       update_irregular_eeq( res, u, m_ebisl[dex], m_kflux_stencil[dex],
			      m_uxy_stencil[dex], m_mu[dex], m_lambda[dex], dir );
	 }
      }
      //      atr.chk_nan(a_res,m_ebisl," res after irregupdate is nan ");

// EB-fluxes, and domain boundary fluxes in irregular cells:
      for( VoFIterator vi(notRegular,m_ebisl[dex].getEBGraph());vi.ok();++vi)       
      {
// EB-fluxes:
	  VolIndex vof=vi();

// Assume constant viscosity for now
	  Real mu=m_mu[dex](vof,0);
	  Real la=m_lambda[dex](vof,0);

          Real ebFlux = 0;
          if( hasNoEBLevelData )
          {
	     if( m_rz )
	        m_ebBC->getEBFlux_eeq_rz( ebFlux, vof, au, dex, m_origin, m_dx,
	                              a_homogeneousPhysBC, mu, la, m_time );
	     else
	        m_ebBC->getEBFlux_eeq( ebFlux, vof, au, dex, m_origin, m_dx,
	                              a_homogeneousPhysBC, mu, la, m_time );
          }
          else
          {
            MayDay::Warning("EBAMRNS::applyOp. EB alternative not implemented");
          }
          res(vof,0) += ebFlux*m_dxScale;

// Domain BC for irregular cells:
          for( int dir = 0 ; dir < SpaceDim ; dir++ )
	  {
             if( hasLo[dir] == 1 )
             {
                if( loBox[dir].contains(vof.gridIndex()))
                {
                    Real flux=0;
                    if( m_rz )
                       m_domainBC->getFaceFlux_eeq_rz( flux, vof, u, m_origin, m_dx, dir,
					     Side::Lo, a_homogeneousPhysBC,
					     m_mu[dex](vof,0), m_lambda[dex](vof,0),m_time);
                    else
                       m_domainBC->getFaceFlux_eeq( flux, vof, u, m_origin, m_dx, dir,
					     Side::Lo, a_homogeneousPhysBC,
					     m_mu[dex](vof,0), m_lambda[dex](vof,0),m_time);
                    res(vof,0) -= flux*m_invDx[dir];
                }
	     }
             if( hasHi[dir] == 1 )
             {
                if( hiBox[dir].contains(vof.gridIndex()) )
                {
                   Real flux=0;
                   if( m_rz )
                      m_domainBC->getFaceFlux_eeq_rz( flux, vof, u, m_origin, m_dx, dir,
					     Side::Hi, a_homogeneousPhysBC,
					     m_mu[dex](vof,0), m_lambda[dex](vof,0),m_time);
                   else
                      m_domainBC->getFaceFlux_eeq( flux, vof, u, m_origin, m_dx, dir,
					     Side::Hi, a_homogeneousPhysBC,
					     m_mu[dex](vof,0), m_lambda[dex](vof,0),m_time);
                   res(vof,0) += flux*m_invDx[dir];
                 }
             }
          }
      }
   }
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::set_rho_zero( )
{
  EBLevelDataOps::setToZero( m_rho );

}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::set_coeff( Real a_c )
{
  m_c = a_c;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::set_time( Real a_time )
{
  m_time = a_time;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOp::set_rho( const LevelData<EBCellFAB>& a_rho )
{
   EBLevelDataOps::assign( m_rho, a_rho );
}


//-----------------------------------------------------------------------
EBAMRNSviscOpFactory::~EBAMRNSviscOpFactory()
{
  for(int ilev = 0; ilev < m_numLevels; ilev++)
    delete m_rhoPtr[ilev];
}


//-----------------------------------------------------------------------
int EBAMRNSviscOpFactory::refToFiner(const ProblemDomain& a_domain) const
{
  int retval = -1;
  bool found = false;
  for(int ilev = 0; ilev < m_domainVec.size(); ilev++)
    {
      if(m_domainVec[ilev].domainBox() == a_domain.domainBox())
        {
          retval = m_refRatioVec[ilev];
          found = true;
        }
    }
  if(!found)
    {
      MayDay::Error("Domain not found in AMR hierarchy");
    }
  return retval;
}


//-----------------------------------------------------------------------
EBAMRNSviscOpFactory::EBAMRNSviscOpFactory(
                      const Vector<DisjointBoxLayout>&      a_grids,
                      const Vector<EBISLayout>&             a_ebisl,
                      const Vector<int>&                    a_refRatio,
                      const ProblemDomain&                  a_domainCoar,
                      const RealVect&                       a_dxCoar,
                      const RealVect&                       a_origin,
                      const int&                            a_orderEB,
                      const int&                            a_relaxType,
                      RefCountedPtr<DirichletNSviscDomainBCFactory>    a_domainBCFactory,
		      RefCountedPtr<DirichletNSviscEBBCFactory>        a_ebBCFactory,
   	              const Vector<Real>                   a_c,
	              const Real                           a_mu,
                      const Real                           a_lambda,
	              const Vector<LevelData<EBCellFAB>*>& a_rho,
                      const bool&                          a_rz )
{
   CH_assert(a_grids.size() <= a_refRatio.size());
   m_numLevels = a_grids.size();

   m_domainBCFactory = a_domainBCFactory;
   m_ebBCFactory     = a_ebBCFactory;

   m_gridVec      = a_grids;
   m_ebislVec     = a_ebisl;
   m_refRatioVec  = a_refRatio;
   m_orderEB      = a_orderEB;
   m_relaxType    = a_relaxType;
   m_origin       = a_origin;

// non-constant data
   m_rhoPtr.resize(m_numLevels);
   m_c.resize(m_numLevels);
   for( int i= 0 ; i<m_numLevels ; i++ )
   {  
      EBCellFactory cfactory( a_ebisl[i] );
      m_rhoPtr[i] = new LevelData<EBCellFAB>();
      m_rhoPtr[i]->define(a_grids[i],1,IntVect::Unit,cfactory);
      EBLevelDataOps::assign( *(m_rhoPtr[i]), *(a_rho[i]) );
      m_c[i]     = a_c[i];
   }

// grids and domains
   m_dxVec.resize(m_numLevels);
   m_domainVec.resize(m_numLevels);
   m_domainVec[0] = a_domainCoar;
   m_dxVec[0] = a_dxCoar;
   for(int ilev = 1; ilev < m_numLevels; ilev++)
   {
      m_domainVec[ilev] = refine(m_domainVec[ilev-1], m_refRatioVec[ilev-1]);
      m_dxVec[ilev]     = m_dxVec[ilev-1]/m_refRatioVec[ilev-1];
   }

// constant data
   m_mu     = a_mu;
   m_lambda = a_lambda;
   m_rz     = a_rz;

}


//-----------------------------------------------------------------------
EBAMRNSviscOp* EBAMRNSviscOpFactory::MGnewOp(const ProblemDomain& a_domainFine,
					     int                  a_depth,
					     bool                 a_homoOnly)
{
  //find out if there is a real starting point here.
  int whichlev;
  bool found = false;
  for(int ilev = 0; ilev < m_numLevels && !found; ilev++)
  {
     if(a_domainFine.domainBox() == m_domainVec[ilev].domainBox() )
     {
        found = true;
        whichlev = ilev ;
     }
  }
  if(!found)
     MayDay::Error("No corresponding AMRLevel to starting point of MGnewOp");


//multigrid operator.  coarn by two from depth  no
//coarse or finer to worry about.
  ProblemDomain  domainMGLevel;
  DisjointBoxLayout dblMGLevel;
  EBISLayout      ebislMGLevel;
  RealVect           dxMGLevel;
  if(a_depth == 0)
  {
     domainMGLevel = m_domainVec[whichlev];
     dblMGLevel    = m_gridVec[whichlev];
     ebislMGLevel  = m_ebislVec[whichlev];
     dxMGLevel     = m_dxVec[whichlev];
  }
  else
  {
     int icoar = 1;
     for(int idep = 0; idep < a_depth; idep++)
     {
        icoar *= 2;
     }
     Box domainBoxMGLevel;
     bool coarsenable = true;
      //find out how much we coarsen
      //should never coarsen beyond next AMR level
      //only really works for lbase = 0
      //need to fix this
      //if((whichlev > 0) && (icoar >= m_refRatioVec[whichlev-1]))
      //  {
      //    coarsenable = false;
      //  }
     if(coarsenable)
     {
        dxMGLevel = m_dxVec[whichlev];
        dxMGLevel *= Real(icoar);
        const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
        coarsenable= getCoarserLayouts(dblMGLevel,
                                       ebislMGLevel,
                                       domainBoxMGLevel,
                                       m_gridVec[whichlev],
                                       m_domainVec[whichlev].domainBox(),
                                       icoar,ebisPtr);
     }
     if(!coarsenable)
     {
          //not coarsenable.
          //return null
          return NULL;
     }
     domainMGLevel = ProblemDomain(domainBoxMGLevel);
  }

//creates coarse and finer info and bcs and all that
  EBAMRNSviscOp* op = createOperator(dblMGLevel, ebislMGLevel, domainMGLevel, dxMGLevel, whichlev);
  return op;
}


//-----------------------------------------------------------------------
EBAMRNSviscOp* EBAMRNSviscOpFactory::AMRnewOp(const ProblemDomain& a_domainFine)
{
  //figure out which level we are at.
  int whichlev;
  bool found = false;

  for(int ilev = 0; ilev < m_numLevels && !found; ilev++)
    {
      if(a_domainFine.domainBox() == m_domainVec[ilev].domainBox())
        {
          found = true;
          whichlev = ilev;
        }
    }
  if(!found)
    {
      MayDay::Error("No corresponding AMRLevel to starting point of AMRnewOp");
    }
  //creates coarse and finer info and bcs and all that
  ProblemDomain      domainMGLevel = m_domainVec[whichlev];
  DisjointBoxLayout  dblMGLevel    = m_gridVec[whichlev];
  EBISLayout         ebislMGLevel  = m_ebislVec[whichlev];
  RealVect           dxMGLevel     = m_dxVec[whichlev];

  EBAMRNSviscOp* op = createOperator(dblMGLevel, ebislMGLevel, domainMGLevel, dxMGLevel, whichlev);
  return op;
}


//-----------------------------------------------------------------------
EBAMRNSviscOp* EBAMRNSviscOpFactory::createOperator(
				      const DisjointBoxLayout& a_dblMGLevel,
                                      const EBISLayout&        a_ebislMGLevel,
                                      const ProblemDomain&     a_domainMGLevel,
                                      const RealVect&          a_dxMGLevel,
                                      const int&               a_whichLevel)
{

  RefCountedPtr<DirichletNSviscEBBC>     ebbcMGLevel  = m_ebBCFactory->create(a_domainMGLevel, a_ebislMGLevel, a_dxMGLevel);
  RefCountedPtr<DirichletNSviscDomainBC> dombcMGLevel = m_domainBCFactory->create(a_domainMGLevel, a_ebislMGLevel, a_dxMGLevel);

  //fine and coarse stuff undefined because this is a MG level
  DisjointBoxLayout  dblFine,     dblCoar;
  EBISLayout       ebislFine,   ebislCoar;
  ProblemDomain    domainFine, domainCoar;
  RealVect             dxFine,     dxCoar;
  RefCountedPtr<DirichletNSviscEBBC>      ebbcCoar = NULL;
  RefCountedPtr<DirichletNSviscEBBC>       ebbcFine  = NULL;
  RefCountedPtr<DirichletNSviscDomainBC> dombcCoar = NULL;
  RefCountedPtr<DirichletNSviscDomainBC>  dombcFine  = NULL;
  int refToFine   = 1;
  int refToCoar = 1;
  bool hasFine   = ((a_whichLevel < m_numLevels - 1)  && (a_domainMGLevel.domainBox() == m_domainVec[a_whichLevel].domainBox()));
  bool hasCoar   = ((a_whichLevel > 0) && (a_domainMGLevel.domainBox() == m_domainVec[a_whichLevel].domainBox()));
  if(hasFine)
    {
      refToFine = m_refRatioVec[a_whichLevel];
      dblFine   =     m_gridVec[a_whichLevel+1];
      ebislFine =    m_ebislVec[a_whichLevel+1];
      domainFine =  m_domainVec[a_whichLevel+1];
      dxFine    =       m_dxVec[a_whichLevel+1];
      ebbcFine  =     m_ebBCFactory->create(domainFine, ebislFine, dxFine);
      dombcFine = m_domainBCFactory->create(domainFine, ebislFine, dxFine);
    }
  if(hasCoar)
    {
      refToCoar  =  m_refRatioVec[a_whichLevel-1];
      dblCoar   =       m_gridVec[a_whichLevel-1];
      ebislCoar =      m_ebislVec[a_whichLevel-1];
      domainCoar =    m_domainVec[a_whichLevel-1];
      dxCoar     =        m_dxVec[a_whichLevel-1];
      ebbcCoar   =     m_ebBCFactory->create(domainCoar, ebislCoar, dxCoar);
      dombcCoar  = m_domainBCFactory->create(domainCoar, ebislCoar, dxCoar);
    }

  LevelData<EBCellFAB>* rho;
  bool alloced=false;
  if( a_domainMGLevel.domainBox() != m_domainVec[a_whichLevel].domainBox() )
  {
     int rat=1;
     Box bx = a_domainMGLevel.domainBox();
     while( bx != m_domainVec[a_whichLevel].domainBox() )
     {
        bx.refine(IntVect::Unit*2);
        rat *= 2;
     }
     rho = new LevelData<EBCellFAB>(a_dblMGLevel, 1, IntVect::Unit,
				    EBCellFactory(a_ebislMGLevel));
     alloced =true;
     interp_frfine( *rho, *(m_rhoPtr[a_whichLevel]), a_domainMGLevel.domainBox(),
		    a_ebislMGLevel, m_ebislVec[a_whichLevel], rat );
  }
  else
  {
     rho = m_rhoPtr[a_whichLevel];
  }

  EBAMRNSviscOp* op = new EBAMRNSviscOp(dblFine ,      a_dblMGLevel,   dblCoar,
                                          ebislFine,   a_ebislMGLevel, ebislCoar,
                                          dombcFine,     dombcMGLevel, dombcCoar,
                                          ebbcFine ,      ebbcMGLevel,  ebbcCoar,
                                          dxFine ,        a_dxMGLevel,    dxCoar,
                                          a_domainMGLevel,  m_origin,
                                          refToFine, refToCoar,
                                          hasFine, hasCoar,
                                          m_relaxType, m_orderEB,
					  m_c[a_whichLevel], m_mu, m_lambda,
					  *rho, m_rz, m_time );
  if( alloced )
     delete rho;
  return op;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOpFactory::reclaim(MGLevelOp<LevelData<EBCellFAB> >* a_reclaim)
{
   delete a_reclaim;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOpFactory::AMRreclaim(EBAMRNSviscOp* a_reclaim)
{
   delete a_reclaim;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOpFactory::set_coeff( Real a_cof, int a_level )
{
   m_c[a_level] = a_cof;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOpFactory::set_rho( const LevelData<EBCellFAB>& a_rho, int a_level )
{
   EBLevelDataOps::assign( *(m_rhoPtr[a_level]), a_rho );
}


//-----------------------------------------------------------------------
void EBAMRNSviscOpFactory::set_time( Real a_time )
{
  m_time = a_time;
}


//-----------------------------------------------------------------------
void EBAMRNSviscOpFactory::interp_frfine( LevelData<EBCellFAB>& a_uc,
					     const LevelData<EBCellFAB>& a_uf,
                                             const Box& a_domainc,
                                             const EBISLayout& a_ec,
                                             const EBISLayout& a_ef,
					     const int a_rat )
{
   int ncomp = a_uc.nComp();
   CH_assert( ncomp == a_uf.nComp() );
   Interval interval(0, ncomp-1);
   const EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
   if( a_uf.getBoxes().coarsenable(a_rat) )
   {
      EBCoarseAverage ebaverage(a_uf.getBoxes(), a_uc.getBoxes(),
                                a_ef, a_ec, a_domainc,
                                a_rat, ncomp, ebisPtr);
      ebaverage.average(a_uc, a_uf, interval);
   }
   else
   {
      Box domainFine = refine( a_domainc, a_rat );
      Vector<Box> fineBoxes(1,  domainFine);
      Vector<Box> coarBoxes(1,a_domainc);
      Vector<int> procs(1, 0);
      DisjointBoxLayout oneBoxDBLFine(fineBoxes, procs);
      DisjointBoxLayout oneBoxDBLCoar(coarBoxes, procs);

      CH_assert(oneBoxDBLFine.coarsenable(a_rat));
       
      //make the one box data holders
      LevelData<EBCellFAB> u_box_coar,u_box_fine;
      const IntVect& ghostVecCoar = a_uc.ghostVect();
      const IntVect& ghostVecFine = a_uf.ghostVect();
      CH_assert(ghostVecCoar == ghostVecFine);
      int nghost = 1;
      EBISLayout oneBoxEBISLCoar, oneBoxEBISLFine;
      ebisPtr->fillEBISLayout( oneBoxEBISLCoar, oneBoxDBLCoar, a_domainc,  nghost );
      ebisPtr->fillEBISLayout( oneBoxEBISLFine, oneBoxDBLFine, domainFine, nghost );
     u_box_coar.define(oneBoxDBLCoar,ncomp,ghostVecCoar, EBCellFactory(oneBoxEBISLCoar));
     u_box_fine.define(oneBoxDBLFine,ncomp,ghostVecFine, EBCellFactory(oneBoxEBISLFine));

      //copy fine data over to the single box fine data
      a_uf.copyTo(u_box_fine);
 
      //now do the coarsening
      EBCoarseAverage ebaverage(oneBoxDBLFine,
                                oneBoxDBLCoar,
                                oneBoxEBISLFine,
                                oneBoxEBISLCoar,
                                a_domainc,
                                a_rat,
                                ncomp,
                                ebisPtr);
      ebaverage.average(u_box_coar, u_box_fine, interval );
      u_box_coar.copyTo(a_uc);
   }
}

